Type: Package
Package: camtrapR
Title: Camera Trap Data Management and Analysis Framework
Version: 3.1.0
Date: 2026-07-25
Authors@R: c(
    person("Juergen", "Niedballa", email = "camtrapr@gmail.com", role = c("aut", "cre"),
           comment = c(ORCID = "0000-0002-9187-2116")),
    person("Alexandre", "Courtiol", role = "aut",
           comment = c(ORCID = "0000-0003-0637-2959")),
    person("Rahel", "Sollmann", role = "aut",
           comment = c(ORCID = "0000-0002-1607-2039")),
    person("John", "Mathai", role = "ctb"),
    person("Seth Timothy", "Wong", role = "ctb",
           comment = c(ORCID = "0000-0001-8083-9268")),
    person("An The Truong", "Nguyen", role = "ctb",
           comment = c(ORCID = "0009-0000-2861-2672")),
    person("Azlan bin", "Mohamed", role = "ctb",
           comment = c(ORCID = "0000-0003-3788-4383")),
    person("Andrew", "Tilker", role = "ctb",
           comment = c(ORCID = "0000-0003-3630-8691")),
    person("Roshan", "Guharajan", role = "ctb",
           comment = c(ORCID = "0000-0001-8124-5461")),
    person("Ioannis", "Alexiou", role = "ctb",
           comment = c(ORCID = "0000-0001-5095-4767")),
    person("Andreas", "Wilting", role = c("ctb", "ths"),
           comment = c(ORCID = "0000-0001-5073-9186"))
  )
Description: Management and analysis of camera trap wildlife data through
    an integrated workflow. Provides functions for image/video
    organization and metadata extraction, species/individual
    identification. Creates detection histories for occupancy and spatial
    capture-recapture analyses, with support for multi-season studies.
    Includes tools for fitting community occupancy models in JAGS and
    NIMBLE, and an interactive dashboard for survey data visualization and
    analysis. Features visualization of species distributions and activity
    patterns, plus export capabilities for GIS and reports. Emphasizes
    automation and reproducibility while maintaining flexibility for
    different study designs.
License: GPL (>= 2)
URL: https://github.com/jniedballa/camtrapR,
        https://jniedballa.github.io/camtrapR/,
        https://groups.google.com/forum/#!forum/camtrapr
BugReports: https://groups.google.com/forum/#!forum/camtrapr
Depends: R (>= 4.1.0)
Imports: checkmate, crayon, data.table, dplyr, DT, generics, ggplot2,
        htmltools, leaflet, lubridate, methods, reshape2, secr, sf,
        shiny, shinyBS, shinydashboard, shinyjs, terra, tibble
Suggests: abind, bayesplot, callr, camtrapdp, coda, corrplot, elevatr,
        iNEXT, jsonify, jsonlite, knitr, lattice, magick, mapview,
        mockery, nimble, nimbleEcology, overlap, parallel, patchwork,
        pbapply, plotly, psych, R.rsp, raster, ritis, rjags, rlang,
        rmarkdown, RSQLite, rstudioapi, scales, shinytest2,
        shinyWidgets, stringr, taxize, tesseract, testthat, ubms,
        units, unmarked, viridisLite, withr, zip
VignetteBuilder: R.rsp
Encoding: UTF-8
SystemRequirements: ExifTool (https://exiftool.org/)
Config/roxygen2/version: 8.0.0
NeedsCompilation: no
Packaged: 2026-07-28 10:35:24 UTC; juergen
Author: Juergen Niedballa [aut, cre] (ORCID:
    <https://orcid.org/0000-0002-9187-2116>),
  Alexandre Courtiol [aut] (ORCID:
    <https://orcid.org/0000-0003-0637-2959>),
  Rahel Sollmann [aut] (ORCID: <https://orcid.org/0000-0002-1607-2039>),
  John Mathai [ctb],
  Seth Timothy Wong [ctb] (ORCID:
    <https://orcid.org/0000-0001-8083-9268>),
  An The Truong Nguyen [ctb] (ORCID:
    <https://orcid.org/0009-0000-2861-2672>),
  Azlan bin Mohamed [ctb] (ORCID:
    <https://orcid.org/0000-0003-3788-4383>),
  Andrew Tilker [ctb] (ORCID: <https://orcid.org/0000-0003-3630-8691>),
  Roshan Guharajan [ctb] (ORCID: <https://orcid.org/0000-0001-8124-5461>),
  Ioannis Alexiou [ctb] (ORCID: <https://orcid.org/0000-0001-5095-4767>),
  Andreas Wilting [ctb, ths] (ORCID:
    <https://orcid.org/0000-0001-5073-9186>)
Maintainer: Juergen Niedballa <camtrapr@gmail.com>
Repository: CRAN
Date/Publication: 2026-07-28 11:20:02 UTC
Built: R 4.6.1; ; 2026-07-28 14:24:12 UTC; unix

# Generated by roxygen2: do not edit by hand

S3method("[",camOp)
S3method("[",detHist)
S3method("[",records)
S3method(dplyr_reconstruct,records)
S3method(plot,camOp)
S3method(print,camOp)
S3method(print,cams)
S3method(print,cams_dp)
S3method(print,detHist)
S3method(print,records)
S3method(summary,camOp)
S3method(summary,records)
export(OCRdataFields)
export(PPC.community)
export(PPC.residuals)
export(activityDensity)
export(activityHistogram)
export(activityOverlap)
export(activityRadial)
export(addCopyrightTag)
export(addToPath)
export(aggregateStations)
export(appendSpeciesNames)
export(cameraOperation)
export(checkSpeciesIdentification)
export(checkSpeciesNames)
export(communityModel)
export(createCovariates)
export(createSpeciesFolders)
export(createStationFolders)
export(detectionHistory)
export(detectionMaps)
export(exifTagNames)
export(filterRecordTable)
export(fixDateTimeOriginal)
export(getSpeciesImages)
export(imageRename)
export(readCamtrapDP)
export(readWildlifeInsights)
export(recordTable)
export(recordTableIndividual)
export(simulateCamtrapData)
export(spatialDetectionHistory)
export(speciesAccum)
export(surveyDashboard)
export(surveyReport)
export(timeShiftImages)
export(writeDateTimeOriginal)
exportClasses(commOccu)
exportMethods(fit)
exportMethods(plot_coef)
exportMethods(plot_effects)
exportMethods(predict)
exportMethods(summary)
importFrom(DT,DTOutput)
importFrom(DT,datatable)
importFrom(DT,renderDT)
importFrom(checkmate,assert)
importFrom(checkmate,assert_character)
importFrom(checkmate,assert_choice)
importFrom(checkmate,assert_data_frame)
importFrom(checkmate,assert_logical)
importFrom(checkmate,checkClass)
importFrom(checkmate,makeAssertCollection)
importFrom(checkmate,reportAssertions)
importFrom(data.table,":=")
importFrom(data.table,foverlaps)
importFrom(data.table,rbindlist)
importFrom(data.table,setDF)
importFrom(data.table,setDT)
importFrom(data.table,setkey)
importFrom(dplyr,"%>%")
importFrom(dplyr,arrange)
importFrom(dplyr,bind_rows)
importFrom(dplyr,count)
importFrom(dplyr,distinct)
importFrom(dplyr,dplyr_reconstruct)
importFrom(dplyr,group_by)
importFrom(dplyr,left_join)
importFrom(dplyr,mutate)
importFrom(dplyr,n)
importFrom(dplyr,n_distinct)
importFrom(dplyr,pull)
importFrom(dplyr,select)
importFrom(dplyr,summarise)
importFrom(dplyr,summarize)
importFrom(dplyr,sym)
importFrom(generics,fit)
importFrom(ggplot2,aes)
importFrom(ggplot2,element_blank)
importFrom(ggplot2,element_rect)
importFrom(ggplot2,element_text)
importFrom(ggplot2,expr)
importFrom(ggplot2,facet_grid)
importFrom(ggplot2,facet_wrap)
importFrom(ggplot2,geom_abline)
importFrom(ggplot2,geom_boxplot)
importFrom(ggplot2,geom_col)
importFrom(ggplot2,geom_line)
importFrom(ggplot2,geom_linerange)
importFrom(ggplot2,geom_point)
importFrom(ggplot2,geom_pointrange)
importFrom(ggplot2,geom_ribbon)
importFrom(ggplot2,geom_violin)
importFrom(ggplot2,geom_vline)
importFrom(ggplot2,ggplot)
importFrom(ggplot2,ggsave)
importFrom(ggplot2,ggtitle)
importFrom(ggplot2,labs)
importFrom(ggplot2,median_hilow)
importFrom(ggplot2,scale_color_manual)
importFrom(ggplot2,scale_x_continuous)
importFrom(ggplot2,scale_y_discrete)
importFrom(ggplot2,stat_summary)
importFrom(ggplot2,theme)
importFrom(ggplot2,theme_bw)
importFrom(ggplot2,theme_void)
importFrom(ggplot2,vars)
importFrom(ggplot2,xlab)
importFrom(ggplot2,xlim)
importFrom(ggplot2,ylab)
importFrom(ggplot2,ylim)
importFrom(grDevices,col2rgb)
importFrom(grDevices,colorRampPalette)
importFrom(grDevices,dev.off)
importFrom(grDevices,extendrange)
importFrom(grDevices,hcl.colors)
importFrom(grDevices,heat.colors)
importFrom(grDevices,png)
importFrom(grDevices,rgb)
importFrom(graphics,abline)
importFrom(graphics,axis)
importFrom(graphics,box)
importFrom(graphics,grconvertX)
importFrom(graphics,grconvertY)
importFrom(graphics,hist)
importFrom(graphics,image)
importFrom(graphics,layout)
importFrom(graphics,legend)
importFrom(graphics,lines)
importFrom(graphics,mtext)
importFrom(graphics,pairs)
importFrom(graphics,par)
importFrom(graphics,plot)
importFrom(graphics,plot.default)
importFrom(graphics,plot.new)
importFrom(graphics,points)
importFrom(graphics,polygon)
importFrom(graphics,rect)
importFrom(graphics,segments)
importFrom(graphics,strheight)
importFrom(graphics,strwidth)
importFrom(graphics,text)
importFrom(graphics,title)
importFrom(leaflet,addCircleMarkers)
importFrom(leaflet,addLayersControl)
importFrom(leaflet,addPolygons)
importFrom(leaflet,addTiles)
importFrom(leaflet,clearGroup)
importFrom(leaflet,layersControlOptions)
importFrom(leaflet,leaflet)
importFrom(leaflet,leafletOutput)
importFrom(leaflet,leafletProxy)
importFrom(leaflet,renderLeaflet)
importFrom(lubridate,"%within%")
importFrom(lubridate,as_date)
importFrom(lubridate,as_datetime)
importFrom(lubridate,days)
importFrom(lubridate,ddays)
importFrom(lubridate,dhours)
importFrom(lubridate,dseconds)
importFrom(lubridate,hour)
importFrom(lubridate,int_end)
importFrom(lubridate,int_overlaps)
importFrom(lubridate,int_start)
importFrom(lubridate,interval)
importFrom(lubridate,is.Date)
importFrom(lubridate,is.POSIXt)
importFrom(lubridate,parse_date_time)
importFrom(lubridate,stamp)
importFrom(lubridate,time_length)
importFrom(lubridate,tz)
importFrom(lubridate,ymd_hms)
importFrom(methods,.hasSlot)
importFrom(methods,hasArg)
importFrom(methods,is)
importFrom(methods,new)
importFrom(reshape2,melt)
importFrom(secr,RMarkInput)
importFrom(secr,covariates)
importFrom(secr,make.capthist)
importFrom(secr,read.traps)
importFrom(secr,traps)
importFrom(secr,usage)
importFrom(sf,st_as_sf)
importFrom(sf,st_bbox)
importFrom(sf,st_buffer)
importFrom(sf,st_convex_hull)
importFrom(sf,st_coordinates)
importFrom(sf,st_drop_geometry)
importFrom(sf,st_intersection)
importFrom(sf,st_make_valid)
importFrom(sf,st_set_crs)
importFrom(sf,st_transform)
importFrom(sf,st_union)
importFrom(sf,st_write)
importFrom(shiny,HTML)
importFrom(shiny,actionButton)
importFrom(shiny,checkboxGroupInput)
importFrom(shiny,checkboxInput)
importFrom(shiny,column)
importFrom(shiny,conditionalPanel)
importFrom(shiny,dateRangeInput)
importFrom(shiny,div)
importFrom(shiny,fileInput)
importFrom(shiny,fluidRow)
importFrom(shiny,h1)
importFrom(shiny,h2)
importFrom(shiny,h3)
importFrom(shiny,h4)
importFrom(shiny,h5)
importFrom(shiny,helpText)
importFrom(shiny,hr)
importFrom(shiny,htmlOutput)
importFrom(shiny,icon)
importFrom(shiny,incProgress)
importFrom(shiny,isolate)
importFrom(shiny,modalButton)
importFrom(shiny,modalDialog)
importFrom(shiny,need)
importFrom(shiny,numericInput)
importFrom(shiny,observe)
importFrom(shiny,observeEvent)
importFrom(shiny,outputOptions)
importFrom(shiny,plotOutput)
importFrom(shiny,radioButtons)
importFrom(shiny,reactive)
importFrom(shiny,reactiveTimer)
importFrom(shiny,reactiveVal)
importFrom(shiny,reactiveValues)
importFrom(shiny,removeModal)
importFrom(shiny,removeNotification)
importFrom(shiny,renderPlot)
importFrom(shiny,renderPrint)
importFrom(shiny,renderTable)
importFrom(shiny,renderText)
importFrom(shiny,renderUI)
importFrom(shiny,req)
importFrom(shiny,selectInput)
importFrom(shiny,showModal)
importFrom(shiny,showNotification)
importFrom(shiny,sliderInput)
importFrom(shiny,span)
importFrom(shiny,tabPanel)
importFrom(shiny,tableOutput)
importFrom(shiny,tabsetPanel)
importFrom(shiny,tagList)
importFrom(shiny,tags)
importFrom(shiny,textInput)
importFrom(shiny,textOutput)
importFrom(shiny,uiOutput)
importFrom(shiny,updateActionButton)
importFrom(shiny,updateCheckboxGroupInput)
importFrom(shiny,updateCheckboxInput)
importFrom(shiny,updateDateRangeInput)
importFrom(shiny,updateNumericInput)
importFrom(shiny,updateSelectInput)
importFrom(shiny,updateSelectizeInput)
importFrom(shiny,updateSliderInput)
importFrom(shiny,updateTextInput)
importFrom(shiny,validate)
importFrom(shiny,varSelectizeInput)
importFrom(shiny,verbatimTextOutput)
importFrom(shiny,wellPanel)
importFrom(shiny,withProgress)
importFrom(shinyBS,bsTooltip)
importFrom(shinydashboard,dropdownMenu)
importFrom(shinydashboard,dropdownMenuOutput)
importFrom(shinydashboard,renderMenu)
importFrom(stats,aggregate)
importFrom(stats,dbinom)
importFrom(stats,dpois)
importFrom(stats,end)
importFrom(stats,na.omit)
importFrom(stats,quantile)
importFrom(stats,rbinom)
importFrom(stats,rmultinom)
importFrom(stats,rnorm)
importFrom(stats,rpois)
importFrom(stats,runif)
importFrom(stats,sd)
importFrom(stats,setNames)
importFrom(stats,start)
importFrom(stats,window)
importFrom(terra,"values<-")
importFrom(terra,mask)
importFrom(terra,nlyr)
importFrom(terra,project)
importFrom(terra,rast)
importFrom(terra,resample)
importFrom(terra,vect)
importFrom(utils,capture.output)
importFrom(utils,head)
importFrom(utils,menu)
importFrom(utils,modifyList)
importFrom(utils,object.size)
importFrom(utils,read.csv)
importFrom(utils,read.table)
importFrom(utils,sessionInfo)
importFrom(utils,str)
importFrom(utils,unzip)
importFrom(utils,write.csv)
importFrom(utils,zip)

# camtrapR

<!-- badges: start -->

[![CRAN
status](https://www.r-pkg.org/badges/version-ago/camtrapR)](https://CRAN.R-project.org/package=camtrapR)
![](https://cranlogs.r-pkg.org/badges/grand-total/camtrapR)
[![](http://cranlogs.r-pkg.org/badges/camtrapR)](https://cran.r-project.org/package=camtrapR)
[![Codecov test
coverage](https://codecov.io/gh/jniedballa/camtrapR/graph/badge.svg)](https://app.codecov.io/gh/jniedballa/camtrapR)
[![R-CMD-check](https://github.com/jniedballa/camtrapR/actions/workflows/R-CMD-check.yaml/badge.svg)](https://github.com/jniedballa/camtrapR/actions/workflows/R-CMD-check.yaml)
<!-- badges: end -->

camtrapR is a package for camera trap data management in R.

## Features

- **Data management** — Build record tables from raw camera trap
  images/videos using EXIF metadata
- **Data import** — Import data from external platforms (Wildlife
  Insights, Camtrap DP, csv files)
- **Survey dashboard** — `surveyDashboard()` provides a code-free
  graphical user interface for data exploration and occupancy modelling
- **Activity patterns** — Kernel density estimation and comparison of
  temporal activity
- **Occupancy models** — Prepare and run single-species and Bayesian
  community occupancy models (via `unmarked`, `ubms`, JAGS or NIMBLE)
- **Spatial capture-recapture** — Generate input for SCR analyses (with
  the `secr` package)
- **Spatial covariates** — Extract environmental covariates and predict
  species occupancy across space
- **Visualisation** — Maps, activity plots, and survey summary figures

## Installation

Install the release version from CRAN:

``` r
install.packages("camtrapR")
```

Install the GitHub master version:

``` r
# install.packages("remotes")
# install.packages("R.rsp")
remotes::install_github("jniedballa/camtrapR", build_vignettes = TRUE)
```

Install the GitHub development version (most recent changes and new
features):

``` r
remotes::install_github("jniedballa/camtrapR", ref = "dev", build_vignettes = TRUE)
```

> **Note:** For installation from GitHub, package `R.rsp` is required.
> `build_vignettes = TRUE` is necessary due to the use of static
> vignettes.

You can also install a specific previous version (e.g., if a recent
update broke code you need):

``` r
remotes::install_version(package = "camtrapR", version = "3.0.1")
```

## Exiftool

Many camtrapR functions read EXIF metadata from JPG images and videos
via [**Exiftool**](https://exiftool.sourceforge.net/), a free and
open-source tool by Phil Harvey (available for Windows, macOS, and
Linux).

To make full use of camtrapR, you will need Exiftool on your system. See
the installation instructions in [vignette
1](https://CRAN.R-project.org/package=camtrapR/vignettes/camtrapr1.pdf)
and the [Exiftool website](https://exiftool.sourceforge.net/).

> If you only use camtrapR to create model input from existing record
> tables (without working with raw image files), Exiftool is not
> required.

## Vignettes

camtrapR includes six vignettes covering the entire workflow:

1.  **Camera trap data management**
2.  **Species and individual identification**
3.  **Record tables and model input**
4.  **Maps, species activity and reports**
5.  **Community Occupancy models**
6.  **Sample data sets and the surveyDashboard()**

Browse them online at <https://cran.r-project.org/package=camtrapR> or
locally after installation with `browseVignettes("camtrapR")`.

## Help and Support

- **Google Group**:
  [camtrapR](https://groups.google.com/forum/#!forum/camtrapr) — for all
  questions and problems around the use of `camtrapR`
- **GitHub Issues**: [Report bugs or request
  features](https://github.com/jniedballa/camtrapR/issues)

## Further reading

Niedballa, J., Sollmann, R., Courtiol, A., Wilting, A. (2016). camtrapR:
an R package for efficient camera trap data management. *Methods in
Ecology and Evolution*, 7(12), 1457–1462.
<https://doi.org/10.1111/2041-210X.12600>

Niedballa, J., Sollmann, R., Wilting, A. (2025) The camtrapR R package:
From data management to interactive ecological analysis of camera trap
data. *bioRxiv* 2025.09.26.678697; doi:
<https://doi.org/10.1101/2025.09.26.678697>

## Citation

To cite camtrapR in publications, please use:

> Niedballa, J., Sollmann, R., Courtiol, A., Wilting, A. (2016).
> camtrapR: an R package for efficient camera trap data management.
> *Methods in Ecology and Evolution* 7(12), 1457–1462.
> <https://doi.org/10.1111/2041-210X.12600>

or run `citation("camtrapR")` in R.

## License

camtrapR is released under the GPL-2 license.

# Version Check for .onAttach()  ####
# adapted from http://thecoatlessprofessor.com/programming/automatically-check-if-r-package-is-the-latest-version-on-package-load/. Thank you!


.pkgVersionCRAN <- function(pkg, cran_url="http://cran.r-project.org/web/packages/")
{
  # Create URL
  cran_pkg_loc <- paste0(cran_url,pkg)
  
  # Try to establish a connection
  suppressWarnings( conn <- try( url(cran_pkg_loc) , silent=TRUE ) )
  
  # If connection, try to parse values, otherwise return NULL
  if (!inherits(conn, "try-error") ) {
    suppressWarnings( cran_pkg_page <- try( readLines(conn) , silent=TRUE ) )
    close(conn)
  } else {
    return(NULL)
  }
  
  # Extract version info
  version_line = cran_pkg_page[grep("Version:",cran_pkg_page)+1]
  gsub("<(td|\\/td)>","",version_line)
}


# for all functions in which user specifies column names: error if spaces in column names    ####
checkForSpacesInColumnNames <- function(...){
  
  z <- list(...)
  
  # if all arguments are of length 1, do
  if(all(sapply(z, FUN = length) == 1)){
    if(any(grepl(pattern = " ", x = unlist(z), fixed = TRUE))) stop("column names may not contain spaces: \n ",
                                                                    paste(names(z)[which(grepl(pattern = " ", x = unlist(z), fixed = TRUE))], "=",
                                                                          z[which(grepl(pattern = " ", x = unlist(z), fixed = TRUE))], collapse = "\n "),
                                                                    call. = FALSE)
  }
  
  # if the argument is of length >1, do
  if(any(sapply(z, FUN = length) > 1)){
    if(length(z) == 1) {
      if(any(grepl(pattern = " ", x = unlist(z[[1]]), fixed = TRUE))) stop("column names in '", names(z) ,"' may not contain spaces: \n ",
                                                                           paste(names(unlist(z))[which(grepl(pattern = " ", x = unlist(z), fixed = TRUE))], "=",
                                                                                 z[[1]][which(grepl(pattern = " ", x = unlist(z), fixed = TRUE))], collapse = "\n "),
                                                                           call. = FALSE)
    } 
    if(length(z) > 1) {
      which_is_the_culprit <- which(sapply(z, FUN = length) > 1)
      stop(paste("Argument", names(z)[which_is_the_culprit], "is not of length 1"), call. = FALSE)
    }
  }
}



# run exiftool for functions reading out and tabulating image metadata    ####

runExiftool <- function(command.tmp,
                        colnames.tmp)
{
  tmp1 <- strsplit(system(command.tmp, intern=TRUE), split = "\t")
  
  if(length(tmp1) == 0) return(NULL) # if nothing returned (no images, no metadata)
  if(length(tmp1) == 1) {
    if(grepl(tmp1[[1]], pattern = "^Error: ", perl = TRUE)){
      warning(paste(unlist(tmp1)), call. = FALSE)
      return(NULL)
      }
    }
  # if first entry is exiftool warning about FileName encoding remove (happens when there's special characters in directory name)
  if(any(grepl(pattern = "FileName encoding not specified", tmp1[[1]]))) tmp1[[1]] <- NULL
  
  metadata.tmp <- as.data.frame(matrix(unlist(lapply(tmp1, FUN = function(X){X[2]})),
                                       ncol = length(colnames.tmp),
                                       byrow = TRUE),
                                stringsAsFactors = FALSE)
  colnames(metadata.tmp) <- colnames.tmp
  
  # find and remove ._ files created on Macs
  strangeMacFiles <- grep("^[._]", metadata.tmp$FileName, fixed = FALSE)
  if(length(strangeMacFiles) >= 1)  {
    warning(paste("found", length(strangeMacFiles), "JPG files beginning with '._' in", paste(unique(metadata.tmp$Directory[strangeMacFiles]), collapse = "\n"), ". \nThese images will be ignored."), call. = FALSE, immediate. = TRUE)
    metadata.tmp <- metadata.tmp[-strangeMacFiles,]
  }
  return(metadata.tmp)
}


# add image metadata as columns in record table ####
addMetadataAsColumns <- function(intable,
                                 metadata.tagname,
                                 metadataHierarchyDelimitor,
                                 multiple_tag_separator)
{
  intable[,metadata.tagname] <- as.character(intable[,metadata.tagname])
  tmp2 <- strsplit(intable[,metadata.tagname], split = ",")      # split items of "HierarchicalSubject" at comma
  tmp3 <- lapply(tmp2, FUN = function(X){X[grep(pattern = metadataHierarchyDelimitor, x = X, fixed = TRUE)]})   # get only the ones with values
  
  
  # find all metadata categories, remove spaces
  list.tmp <- vector()
  
  # loop over images, get name of tag groups
    for(xy in 1:length(tmp3)){
    list.tmp <- c(list.tmp, gsub(pattern = " ",
                                 replacement = "",
                                 x =  sapply(strsplit(tmp3[[xy]],
                                                             split = metadataHierarchyDelimitor,
                                                             fixed = TRUE),
                                                    FUN = function(Y){Y = Y[1]})))
  }
  cols2add <- unique(list.tmp)    # these are the columns to add
  
  # add empty columns
  if(length(cols2add) >= 1){    # if anything to add
    intable <- data.frame(intable, 
                          matrix(NA, ncol = length(cols2add), nrow = nrow(intable)), 
                          check.names = FALSE)
    colnames(intable)[seq((ncol(intable) - length(cols2add) + 1),ncol(intable))] <- cols2add
    
    
    # fill metadata columns
    for(index_new_col in 1:length(cols2add)){                 # loop over metadata tag groups
      
      taggroup <- paste0(cols2add[index_new_col],
                         metadataHierarchyDelimitor)
      

      for(img in 1:length(tmp3)) {
        
        current <- tmp3[[img]]
        if(length(current) == 1) if(current == "") intable[img, cols2add[index_new_col]] <- NA
        
        # remove leading space
        current <- gsub("^ ", "", current)
        
        # subset to current tag group
        current_subset <- subset_taggroup(current, taggroup)
        
        use_these <- vector(length = length(current_subset))
        # check if current tag is included in another tag at a deeper level in the hierarchy
        for(tag in 1:length(current_subset)){
          use_these[tag] <- ifelse(is.na(charmatch(current_subset[tag], current_subset[-tag])), TRUE, FALSE)
        }
          
          #get the deepest hierarchical level of the tag
          current_subset_split <- sapply(current_subset[use_these], 
                                         strsplit,
                                         split = metadataHierarchyDelimitor,
                                         fixed = TRUE)
          tags_to_use <- sapply(current_subset_split, FUN = function(x) x [length(x)]) 
          intable[img, cols2add[index_new_col]] <- paste(tags_to_use, collapse = multiple_tag_separator)
          
          rm(tags_to_use, current_subset_split, current_subset, current)
        }
    } # end for index_new_col
  }  # end if(length(cols2add) >= 1){ 
  
  which_cols_to_rename <- which(colnames(intable) %in% cols2add)
  
  # rename metadata columns with prefix "metadata_"
  colnames(intable)[which_cols_to_rename] <- paste("metadata_", colnames(intable)[which_cols_to_rename], sep = "")
  
  return(intable)
}




subset_taggroup <- function(x, parent) {
  x[grep(x = x,
         pattern = parent,
         fixed = TRUE)]
}




# assign species IDs from metadata tags or directory names   ####
assignSpeciesID <- function(intable,
                            IDfrom,
                            metadataSpeciesTag,
                            speciesCol,
                            dirs_short,
                            i_tmp,
                            multiple_tag_separator,
                            returnFileNamesMissingTags,
                            parent = "recordTable")
{
  
  if(parent == "recordTable") {
    parent_name <- "species"
    parent_argname <- "metadataSpeciesTag"
  }
  if(parent == "recordTableIndividual") {
    parent_name <- "individual" 
    parent_argname <- "metadataIDTag"
  }
  
  file.sep <- .Platform$file.sep
  
  
  if(IDfrom == "directory"){
    intable[,speciesCol] <-  sapply(strsplit(intable$Directory, split = file.sep, fixed = TRUE), FUN = function(X){X[length(X)]})
    return(intable)
  } else {           # if IDfrom = "metadata"
    if(hasArg(metadataSpeciesTag)){
      metadataSpeciesTag2 <- paste("metadata", metadataSpeciesTag, sep = "_")
      
      # if the metadata_Species tag is found in metadata.tmp
      if(metadataSpeciesTag2 %in% colnames(intable)){
        
        # copy to species column proper
        intable[,speciesCol] <- intable[,metadataSpeciesTag2]
        
        # find records without proper species tag, to be removed
        species_records_to_remove <- c(which(is.na(intable[,speciesCol])),
                                       which(intable[,speciesCol] == "NA"))
        
        # if there's records to remove
        if(length(species_records_to_remove) >= 1){    
          
          # give warnings
          if(isTRUE(returnFileNamesMissingTags)){
            warning(paste(paste( dirs_short[i_tmp],":  removed", length(species_records_to_remove), "images out of", nrow(intable),
                                 "because of missing", parent_name, "metadata tag:\n"),
                          paste(head(paste(intable$Directory[species_records_to_remove], intable$FileName[species_records_to_remove], sep = file.sep)), collapse = "\n")),
                    call. = FALSE, immediate. = TRUE)
          } else {
            warning(paste(paste( dirs_short[i_tmp],":  removed", length(species_records_to_remove), "images out of", nrow(intable),
                                 "because of missing", parent_name, "metadata tag")),
                    call. = FALSE, immediate. = TRUE)
          }
          #remove records without species tag
          intable <- intable[-species_records_to_remove,]      
        }
        
        # duplicate records with multiple species (separate row for each species)
        intable <- separateMultipleSpecies (intable                = intable,
                                            speciesCol             = speciesCol,
                                            multiple_tag_separator = multiple_tag_separator)
        
        return(intable)
      } else {
        warning(paste(dirs_short[i_tmp], ":   metadataSpeciesTag '", metadataSpeciesTag, "' not found in image metadata tag 'HierarchicalSubject'.", 
                      ifelse(grepl(" ", metadataSpeciesTag, fixed = T), paste0("This is probably because of an empty space in '", metadataSpeciesTag, "'"), ""), 
                      sep = ""), call. = FALSE, immediate. = TRUE)
        
        return("found no species tag")
      }
    } else {
      stop(paste("station", dirs_short[i_tmp], ":   cannot figure out", parent_name, "names. Is", parent_argname, "defined?"), call. = FALSE)
    }
  }
}


# find and separate multiple species in same image (only if using metadata ID)    ####
separateMultipleSpecies <- function(intable,
                                    speciesCol,
                                    multiple_tag_separator)
{
  
  records0                 <- intable[,speciesCol]
  records_duplicate        <- strsplit(intable[,speciesCol], split = multiple_tag_separator, fixed = TRUE)
  records_duplicate_length <- sapply(records_duplicate, length)
  
  if(any(records_duplicate_length > 1)){
    intable <- intable[rep(row.names(intable), records_duplicate_length), ]                # replicate rows with >1 species / individual
    intable[,speciesCol] <- unlist(strsplit (records0, split = multiple_tag_separator))    # assign species anew
  }
  return(intable)
}




# add station and camera id to metadata table   ####

addStationCameraID <- function(intable,
                               dirs_short,
                               stationCol,
                               cameraCol,
                               cameraID,
                               hasStationFolders,
                               i,
                               IDfrom)
{
  
  file.sep <- .Platform$file.sep
  
  # append station ID
  
  if(isTRUE(hasStationFolders)) {       # take station ID from station directories
    intable <- cbind(intable, dirs_short[i])
    colnames(intable)[ncol(intable)] <- stationCol
    
  } else  {                             # take station ID from image filenames
    
    station.tmp  <- try(sapply(strsplit(as.character(intable$FileName), split = "__"), FUN = function(X){X[1]}))      # assumes filenames: STATION__Camera__Date/Time(Number).JPG)
    if(length(station.tmp) == nrow(intable)){
      intable <- cbind(intable, station.tmp)
      colnames(intable)[ncol(intable)] <- stationCol
    } else {
      stop(paste(dirs_short[i], ": numbers of images and station ID extracted from image names do not match. Do image filenames begin with station IDs?"))
    }
  }
  
  # append camera ID
  
  if(hasArg(cameraID)){
    if(cameraID == "filename"){
      camera.tmp  <- try(sapply(strsplit(as.character(intable$FileName), split = "__"), FUN = function(X){X[2]}))      # assumes filenames: Station__CAMERA__Date/Time(Number).JPG)
      if(length(camera.tmp) == nrow(intable)){
        intable <- cbind(intable, camera.tmp)
        colnames(intable)[ncol(intable)]     <- cameraCol
      }
    }
    if(cameraID == "directory"){            # this can only happen in recordTable. Not in recordTableIndividual
      if(IDfrom == "directory"){             # assumes directory structure: Station/Camera/Species
        intable <- cbind(intable,
                         sapply(strsplit(intable$Directory, split = file.sep, fixed = TRUE), FUN = function(X){X[length(X) - 1]}))
      } else {                                    # assumes directory structure: Station/Camera
        intable <- cbind(intable,
                         sapply(strsplit(intable$Directory, split = file.sep, fixed = TRUE), FUN = function(X){X[length(X)]}))
      }
      colnames(intable)[ncol(intable)]     <- cameraCol
    }
  }
  return(intable)
}

# check if date/time information is present and was readable    ####

checkDateTimeOriginal <- function (intable, dirs_short, i, stationCol, recordDateTimeCol = "DateTimeOriginal"){
  
  if(any(is.na(intable[, recordDateTimeCol]))){
    which_na_time <- which(is.na(intable[, recordDateTimeCol]))
    if(hasArg(dirs_short)) {
      warning(paste0(dirs_short[i], ": Removing ", length(which_na_time), " out of ",
                     nrow(intable)," images because date/time is NA:\n",
                     paste("  ", file.path(intable$Directory, intable$FileName)[which_na_time], collapse = "\n")), call. = FALSE,  immediate. = TRUE)
    } else {
      warning(paste0("Removing ", length(which_na_time), " out of ",
                     nrow(intable)," images because date/time is NA."), 
              call. = FALSE,  immediate. = TRUE)
    }
    intable <- intable[-which_na_time,]
  }
    
    
  # if all date/time information is missing, go to next station
  if(all(intable[, recordDateTimeCol] == "-")){
    if(hasArg(dirs_short)) {
      warning(paste(dirs_short[i], ": no readable date/time information. Skipping"), call. = FALSE,  immediate. = TRUE)
    } 
    
    if(hasArg(stationCol)) {
      warning(paste(unique(intable[, stationCol]), ": no readable date/time information. Skipping"), call. = FALSE,  immediate. = TRUE)
    }
    intable <- NULL
  } else {
    
    # if date/time information is missing for some records only
    if(any(intable[, recordDateTimeCol] == "-")){
      which_no_time <- which(intable[, recordDateTimeCol] == "-")
      if(hasArg(dirs_short)) {
        warning(paste(dirs_short[i], ": omitting", length(which_no_time), "images because of missing/unreadable date/time information."), 
                call. = FALSE,  immediate. = FALSE)
      }
      
      if(hasArg(stationCol)) {
        warning(paste(unique(intable[, stationCol]), ": omitting", length(which_no_time), "images because of missing/unreadable date/time information."), 
                call. = FALSE,  immediate. = FALSE)
      }
      
      intable <-  intable[-which_no_time,]    # removing rows with missing date/time information
    }
  }
  return(intable)
}

# remove duplicate records of same species taken in same second at the same station (by the same camera, if relevant)   ####
# Note to self: this may also be done outside the station loop, after the final record table is assembled. Saves a few executions of this function.
# edit 2019-12-04: but then the messages are useless, so just leave it as is for now

removeDuplicatesOfRecords <- function(metadata.tmp, 
                                      removeDuplicateRecords, 
                                      camerasIndependent, 
                                      stationCol, 
                                      speciesCol, 
                                      cameraCol, 
                                      recordDateTimeCol = "DateTimeOriginal",
                                      current, 
                                      total,
                                      max_nchar_station,
                                      quiet = FALSE){
  
  metadata.tmp0 <- metadata.tmp
  

  
  if(isTRUE(removeDuplicateRecords)){
    if(isTRUE(camerasIndependent)){
      remove.tmp <- which(duplicated(metadata.tmp[,c(recordDateTimeCol, stationCol, speciesCol, cameraCol)]))
      if(length(remove.tmp >= 1)){
        metadata.tmp <- metadata.tmp[-remove.tmp,]
      }
    } else {
      remove.tmp <- which(duplicated(metadata.tmp[,c(recordDateTimeCol, stationCol, speciesCol)]))
      if(length(remove.tmp >= 1)) {
        metadata.tmp <- metadata.tmp[-remove.tmp,]
      }
    }
    
    if(!quiet) {
      pb <- makeProgressbar(current = current, total = total)
      
      if(length(unique(metadata.tmp[,stationCol])) == 1) {                  # 1 station per exiftool call
        message(formatC(as.character(unique(metadata.tmp[,stationCol])), 
                        width = max_nchar_station, 
                        flag = "-"), ":  ",
                formatC(nrow(metadata.tmp0), width = 5), " images ", 
                formatC(length(remove.tmp),  width = 4), " duplicates removed",
                pb)
      } else {                                                               # > 1 station per exiftool call (recordTableIndividual)
        message(paste(unique(metadata.tmp[,stationCol]), collapse = ", "), ":  ",
                formatC(nrow(metadata.tmp0), width = 5), " images ", 
                formatC(length(remove.tmp),  width = 4), " duplicates removed",
                pb)
      }
    } 
    
    
    if(isFALSE(removeDuplicateRecords)){
      if(length(unique(metadata.tmp[,stationCol])) == 1) {                  # 1 station per exiftool call
        message(formatC(as.character(unique(metadata.tmp[,stationCol])), 
                        width = max_nchar_station, 
                        flag = "-"), ":  ",
                formatC(nrow(metadata.tmp0), width = 5), " images", 
                pb)
      } else {                                                               # > 1 station per exiftool call (recordTableIndividual)
        message(paste(unique(metadata.tmp[,stationCol]), collapse = ", "), ":  ", 
                formatC(nrow(metadata.tmp0), width = 5), " images", 
                pb)
      }
    }
  }
  return(metadata.tmp)
}


# assess temporal independence between records   ####

assessTemporalIndependence <- function(intable,
                                       deltaTimeComparedTo,
                                       columnOfInterest,     # species/individual column
                                       recordDateTimeCol = "DateTimeOriginal",
                                       cameraCol,
                                       camerasIndependent,
                                       stationCol,
                                       minDeltaTime,
                                       eventSummaryColumn,
                                       eventSummaryFunction)
{
  # check if all Exif DateTimeOriginal tags were read correctly
  if(any(is.na(intable[, recordDateTimeCol]))){
    which.tmp <- which(is.na(intable[, recordDateTimeCol]))
    if(length(which.tmp) == nrow(intable)) stop("Could not read any Exif DateTimeOriginal tag at station: ", paste(unique(intable[which.tmp, stationCol])), " Consider checking for corrupted Exif metadata.")
    warning(paste("Could not read Exif DateTimeOriginal tag of", length(which.tmp),"image(s) at station", paste(unique(intable[which.tmp, stationCol]), collapse = ", "), ". Will omit them.\nConsider checking for corrupted Exif metadata. Or does your selected time zone have daylight saving time and the image(s) fall in the missing hour at spring forward (cameras don't usually record DST)?. \n",
                  paste(file.path(intable[which.tmp, "Directory"],
                                  intable[which.tmp, "FileName"]), collapse = "\n")), call. = FALSE, immediate. = TRUE)
    intable <- intable[-which.tmp ,]
    rm(which.tmp)
  }
  
  # prepare to add time difference between observations columns
  intable <- data.frame(intable,
                        delta.time.secs  = NA,
                        delta.time.mins  = NA,
                        delta.time.hours = NA,
                        delta.time.days  = NA,
                        independent      = ifelse(minDeltaTime == 0, TRUE, NA),   # all independent if no temporal filtering
                        stringsAsFactors = FALSE,
                        check.names      = FALSE)        # to prevent ":" being converted to ".", e.g. in EXIF:Make
  
  # sort records by station, species, then time
  intable <- intable[order(intable[, stationCol], intable[, columnOfInterest], intable[, recordDateTimeCol]),]
  
  for(xy in 1:nrow(intable)){     # for every record
    
    
    which.columnOfInterest <- which(intable[, columnOfInterest]  == intable[xy, columnOfInterest])          # same species/individual
    which.stationCol       <- which(intable[, stationCol]        == intable[xy, stationCol])                # at same station
    which.independent      <- which(intable$independent          == TRUE)                                   # independent (first or only record of a species at a station)
    which.earlier          <- which(intable[, recordDateTimeCol]     <  intable[xy, recordDateTimeCol])          # earlier than record xy (takes long)
    #which.earlier          <- 1: (xy-1)                                                                  # earlier than record xy  (fast alternative, relies on table being sorted by date/time before anything else)
    if(camerasIndependent) {
      which.cameraCol      <- which(intable[, cameraCol]  == intable[xy, cameraCol])                        # at same camera
    }
    
    # set independent = TRUE and delta.time = 0 if it is the 1st/only  record of a species / individual
    
    if(camerasIndependent == TRUE){
      which.tmp <- Reduce(intersect, list(which.columnOfInterest, 
                                          which.stationCol, 
                                          which.cameraCol))
      if(intable[xy, recordDateTimeCol]  == min(intable[which.tmp, recordDateTimeCol])){    # cameras at same station assessed independently
        intable$independent[xy]       <- TRUE
        intable$delta.time.secs[xy]   <- 0
      }
    } else {
      which.tmp <- Reduce(intersect, list(which.columnOfInterest, 
                                          which.stationCol))
      if(intable[xy, recordDateTimeCol]  == min(intable[which.tmp, recordDateTimeCol])){
        intable$independent[xy]       <- TRUE
        intable$delta.time.secs[xy]   <- 0
      }
    }
    
    # calculate time difference to previous records of same species at this station (if not the 1st/only record)
    if(is.na(intable$delta.time.secs[xy])) {
      
      if(deltaTimeComparedTo == "lastIndependentRecord"){
        
        if(camerasIndependent == TRUE){
          which_time2 <- Reduce(intersect, list(which.columnOfInterest, 
                                              which.stationCol,
                                              which.cameraCol,
                                              which.independent,
                                              which.earlier))
        } else {
          
          which_time2 <- Reduce(intersect, list(which.columnOfInterest, 
                                                which.stationCol,
                                                which.independent,
                                                which.earlier))
        }
      }  else {    # if(deltaTimeComparedTo == "lastRecord"){'
        if(camerasIndependent  == TRUE){
          which_time2 <- Reduce(intersect, list(which.columnOfInterest, 
                                                which.stationCol,
                                                which.cameraCol,
                                                which.earlier))
        } else {
          which_time2 <- Reduce(intersect, list(which.columnOfInterest, 
                                                which.stationCol,
                                                which.earlier))
        }
      }
      
      # time difference to last (independent) record
      diff_tmp <- min(na.omit(difftime(time1 = intable[xy, recordDateTimeCol],            # delta time to last independent record
                                       time2 = intable[which_time2, recordDateTimeCol],
                                       units = "secs")))
      
      # save delta time in seconds
      intable$delta.time.secs[xy] <-  diff_tmp
      if(intable$delta.time.secs[xy] >= (minDeltaTime * 60) | intable$delta.time.secs[xy] == 0){
        intable$independent[xy] <- TRUE
      } else {
        intable$independent[xy] <- FALSE
      }
      
    }   # end   if(intable$DateTimeOriginal[xy] == min(...)} else {...}
  }     # end for(xy in 1:nrow(intable))
  
  # summarise some column by independent event
  
  n_imagesColumn <- "n_images"
  intable[, n_imagesColumn] <- NA
  
  which_independent <- which(intable$independent)
  
  if(hasArg(eventSummaryColumn)){
    if(all(eventSummaryColumn %in% colnames(intable))){
      if(length(eventSummaryColumn) != length(eventSummaryFunction)) stop('"eventSummaryColumn" and "eventSummaryFunction" must have same length', call. = FALSE)
      
    } else {
      warning(paste(unique(intable[, stationCol]), ": eventSummaryColumn(s) " ,
                    paste(eventSummaryColumn[!eventSummaryColumn %in% colnames(intable)], collapse = ", "),
                    " not found in column names of recordtable"), call. = FALSE)
    }
    
    summary_column_name <- paste(eventSummaryColumn, eventSummaryFunction, sep = "_")
    intable[, summary_column_name[eventSummaryColumn %in% colnames(intable)]] <- NA
    
    }    # end    if(hasArg(eventSummaryColumn)){
  
  
  for(xy in 1:length(which_independent)){     # for every independent record (the ones that end up in record table)
    
    current_row <- which_independent[xy]
    
    if(camerasIndependent){
      which_records_to_group <- which(intable[, columnOfInterest]     == intable[current_row, columnOfInterest] &   # same species
                                      intable[, stationCol]           == intable[current_row, stationCol]  &        # same station
                                      intable[, cameraCol]            == intable[current_row, cameraCol]   &        # same camera
                                      intable[, recordDateTimeCol]    >= intable[current_row, recordDateTimeCol] &    # later than current record
                                      !isTRUE(intable$independent))                                        # not independent
      
    } else {
      which_records_to_group <- which(intable[, columnOfInterest]     == intable[current_row, columnOfInterest] &   # same species
                                      intable[, stationCol]           == intable[current_row, stationCol]  &        # same station
                                      intable[, recordDateTimeCol]    >= intable[current_row, recordDateTimeCol] &
                                      !isTRUE(intable$independent))                                                 # not independent
    }

    # subset to records before the next independent record
    if(xy < length(which_independent)){
      which_records_to_group <- which_records_to_group[which_records_to_group < which_independent[xy + 1]] #which_records_to_group[which_records_to_group %in% seq(current_row, (which_independent[xy + 1] - 1))]
    } else {
      which_records_to_group <-  which_records_to_group[which_records_to_group <= nrow(intable)]   # which_records_to_group[which_records_to_group %in% seq(current_row, nrow(intable))]
    }
    
    if(hasArg(eventSummaryColumn)){
      
      for(eventSummaryIndex in 1:length(eventSummaryColumn)) {
        
        if(eventSummaryColumn[eventSummaryIndex] %in% colnames(intable)){
          
          summary_value <- do.call(what = eventSummaryFunction[eventSummaryIndex], 
                                   args = list(intable[which_records_to_group, eventSummaryColumn[eventSummaryIndex]], 
                                               na.rm = TRUE))
          if(!is.infinite(summary_value)){
            intable[current_row, summary_column_name[eventSummaryIndex]] <- ifelse(length(summary_value) > 1, 
                                                                                   paste(summary_value, collapse = ", "),
                                                                                   summary_value)
          }
          rm(summary_value)
        }
        
        
      }    # end for(eventSummaryIndex in 1:length(eventSummaryColumn)
    }      # end if(hasArg(eventSummaryColumn))
    
    
    # quick and dirty solution to prevent n_images = 0 when length(which_records_to_group) = 0 (can happen if only 1 image in a directory)
    
    if(length(which_records_to_group) == 0) {
      intable[ current_row, n_imagesColumn] <- 1
    } else {
      intable[ current_row, n_imagesColumn] <- length(which_records_to_group)
    }
    
    rm(which_records_to_group, current_row)
    
  }        # end for(xy in 1:length(which_independent))
  
  
  # keep only independent records
  outtable <- intable[intable$independent,]
  
  
  # compute delta time in hours and days
  outtable$delta.time.secs  <- round(outtable$delta.time.secs, digits = 0)
  outtable$delta.time.mins  <- round(outtable$delta.time.secs  / 60, digits = 1)
  outtable$delta.time.hours <- round(outtable$delta.time.mins  / 60, digits = 1)
  outtable$delta.time.days  <- round(outtable$delta.time.hours / 24, digits = 1)
  
  # remove "independent" column
  
  outtable <- outtable[, !colnames(outtable) %in% "independent"]
  
  return(outtable)
}


# add potential new columns to global record.table   ####

addNewColumnsToGlobalTable <- function(intable,
                                       i,
                                       record.table)
{
  
  if( nrow(record.table) >= 1){    # if there is a record table already (i.e., it is not the first station with images)
    which_cols_to_add_to_d1 <- seq(1, ncol(record.table))[-which(colnames(record.table) %in% colnames(intable))]   # columns in record.table but not in intable
    
    # if intable lacks columns present in record.table, add them here (filled with NA)
    if(length(which_cols_to_add_to_d1) >= 1){
      intable <- data.frame(intable, as.list(rep(NA, each = length(which_cols_to_add_to_d1))))
      colnames(intable)[(ncol(intable) - length(which_cols_to_add_to_d1) + 1) :  ncol(intable)] <- colnames(record.table)[which_cols_to_add_to_d1]
    }
    
    # now check which columns are present in intable but not in record.table (new tag groups) and add these (filled with NA)
    which_cols_to_add_to_record.table <- seq(1, ncol(intable))[-which(colnames(intable) %in% colnames(record.table))]  # columns present in intable but not in record.table
    if(length(which_cols_to_add_to_record.table) >= 1){
      record.table <- data.frame(record.table, 
                                 as.list(rep(NA, each = length(which_cols_to_add_to_record.table))),
                                 check.names = FALSE)
      colnames(record.table)[(ncol(record.table) - length(which_cols_to_add_to_record.table) + 1) :  ncol(record.table)] <- colnames(intable)[which_cols_to_add_to_record.table]
    }
    outtable <- intable[,match(colnames(record.table), colnames(intable))]
  } else {
    outtable <- intable
  }
  return(list(outtable, record.table))
}



#####################################################

# for detectionHistory functions

# check column names of camera operation matrix  ####
# if there's a time shift from cameraOperation, extract return number as attribute
checkCamOpColumnNames <- function(cameraOperationMatrix){
  
  if(any(is.na(colnames(cameraOperationMatrix)))) stop("There are NAs in the column names of camOp", call. = FALSE)
  
  if(!all(diff(as.Date(colnames(cameraOperationMatrix))) == 1)) stop("Column names in camop must be a continuous sequence of dates without gaps", call. = FALSE)
  
  # check if camera operration matrix has time shift
  if(all(grepl(pattern = "+", colnames(cameraOperationMatrix), fixed = TRUE))){
    colnames_as_dates <- sapply(strsplit(colnames(cameraOperationMatrix), split = "+", fixed = TRUE), FUN = function(x)x[1])
    camopTest <- try(as.Date(colnames_as_dates), silent = TRUE)
    occasionStartTime <- as.numeric(gsub("h", "", unique(sapply(strsplit(colnames(cameraOperationMatrix), split = "+", fixed = TRUE), FUN = function(x)x[2]))))
  } else {
    occasionStartTime <- 0
    camopTest <- try(as.Date(colnames(cameraOperationMatrix)), silent = TRUE)
  }
  
  if(inherits(camopTest, "try-error")) stop(paste('Could not interpret column names in camOp as Dates. Desired format is YYYY-MM-DD (e.g. "2016-12-31") YYYY-MM-DD+Xh (X being a number, e.g. "2016-12-31+12h"). First column name in your camera operation matrix is "', colnames(cameraOperationMatrix)[1], '"', sep = '' ), call. = FALSE)
  
  
  colnames(cameraOperationMatrix) <- as.character(camopTest)
  
  attr(cameraOperationMatrix, "occasionStartTime") <- occasionStartTime   # return time shift extracted from the column names
  return(cameraOperationMatrix)    
}


# create table of station / camera deployment and image data ranges  ####
createDateRangeTable <- function(cam.op,
                                 subset_species_tmp,
                                 buffer_tmp,
                                 stationCol_tmp,
                                 day1_tmp,
                                 occasionStartTime_tmp,
                                 maxNumberDays_tmp,
                                 timeZone_tmp)
{
  
  # first day of each station
  cam.tmp.min <- apply(cam.op, MARGIN = 1, function(X){min(which(!is.na(X)))})
  # last day of each station
  cam.tmp.max <- apply(cam.op, MARGIN = 1, function(X){max(which(!is.na(X)))})
  
  # date of first / last record by station
  rec.tmp.min  <- aggregate(as.Date(subset_species_tmp$DateTime2, tz = timeZone_tmp),
                            list(subset_species_tmp[,stationCol_tmp]),
                            FUN = min)
  rec.tmp.max  <- aggregate(as.Date(subset_species_tmp$DateTime2, tz = timeZone_tmp),
                            list(subset_species_tmp[,stationCol_tmp]),
                            FUN = max)
  
  
  # combine record dates and camera operation dates in one table
  date_ranges <- data.frame(rec.min = rec.tmp.min[match(rownames(cam.op), rec.tmp.min[,1]), 2],       # first record
                            rec.max = rec.tmp.max[match(rownames(cam.op), rec.tmp.max[,1]), 2],       # last record
                            cam.min = as.POSIXct(colnames(cam.op)[cam.tmp.min], tz = timeZone_tmp),   # station setup date
                            cam.max = as.POSIXct(colnames(cam.op)[cam.tmp.max], tz = timeZone_tmp)    # station retrieval date
  )
  
  rownames(date_ranges) <- rownames(cam.op)
  
  # check if images were taken between setup and retrieval dates (warning if images outside station date range)
  if(any(date_ranges$rec.min < as.Date(date_ranges$cam.min, tz = timeZone_tmp), na.rm = TRUE)){
    warning(paste("\nAt", sum(date_ranges$rec.min < as.Date(date_ranges$cam.min, tz = timeZone_tmp), na.rm = TRUE), "stations",
                  "there were records before camera operation date range: ",
                  paste(rownames(date_ranges)[which(date_ranges$rec.min < as.Date(date_ranges$cam.min, tz = timeZone_tmp))], 
                        sep = "\n", collapse = ", " )), call. = FALSE)
  }
  if(any(date_ranges$rec.max > as.Date(date_ranges$cam.max, tz = timeZone_tmp), na.rm = TRUE)) {
    warning(paste("\nAt", sum(date_ranges$rec.max > as.Date(date_ranges$cam.max, tz = timeZone_tmp), na.rm = TRUE), "stations",
                  "there were records after camera operation date range: ",
                  paste(rownames(date_ranges)[which(date_ranges$rec.max > as.Date(date_ranges$cam.max, tz = timeZone_tmp))], 
                        sep = "\n", collapse = ", " )), call. = FALSE)
  }
  
  # define when first occasion begins (to afterwards remove prior records in function    cleanSubsetSpecies)
  if(!hasArg(buffer_tmp)) buffer_tmp <- 0
  
  
  date_ranges$start_first_occasion <- date_ranges$cam.min + buffer_tmp * 86400 + occasionStartTime_tmp * 3600     #each stations setup  + buffer + starttime
  date_ranges$start_first_occasion_survey <- min(date_ranges$cam.min) + buffer_tmp * 86400 + occasionStartTime_tmp * 3600    # first station's setup  + buffer + starttime

  
  if(day1_tmp %in% c("survey", "station") == FALSE) {
    if(as.Date(day1_tmp, tz = timeZone_tmp) < min(as.Date(date_ranges$cam.min,  tz = timeZone_tmp))) stop(paste("day1 (", day1_tmp, ") is before the first station's setup date (",  min(as.Date(date_ranges$cam.min,  tz = timeZone_tmp)), ")", sep = ""))
    if(as.Date(day1_tmp, tz = timeZone_tmp) > max(as.Date(date_ranges$cam.max,  tz = timeZone_tmp))) stop(paste("day1 (", day1_tmp, ") is after the last station's retrieval date (",  max(as.Date(date_ranges$cam.max,  tz = timeZone_tmp)), ")", sep = ""))
    date_ranges$start_first_occasion <- as.POSIXlt(day1_tmp, tz = timeZone_tmp) + occasionStartTime_tmp * 3600
  }
  
  
  
  # # define when last occasion ends 
  # the old way:
  date_ranges$end_of_retrieval_day <- as.POSIXct(paste(date_ranges$cam.max, "23:59:59"), tz = timeZone_tmp, format = "%Y-%m-%d %H:%M:%S")    # end of retrieval day
  
  # new possible solution:
  # # Option 1: end of retrieval day
  # end_of_retrieval_day <- as.POSIXct(paste(date_ranges$cam.max, "23:59:59"), tz = timeZone_tmp, format = "%Y-%m-%d %H:%M:%S")    # end of retrieval day
  # # Option 2: exactly at retrieval hour (this only makes sense if value in camOp on retrieval day is fraction (if time of day was provided))
  # # if there was a problem on the last day, it will be incorrect (return 0:0:0 if effort = 0)
  # end_of_retrieval_day_hour <- date_ranges$cam.max + dhours(apply(cam.op, MARGIN = 1, function(X){X[max(which(!is.na(X)))] * 24}))
  # 
  # # define end of survey: exact retrieval time (if effort is a fraction of day)
  # if(all(end_of_retrieval_day_hour < end_of_retrieval_day)) {
  #   date_ranges$end_of_retrieval_day <- end_of_retrieval_day_hour
  # } else{
  #   date_ranges$end_of_retrieval_day <- end_of_retrieval_day
  # }
  
  # if maxNumberDays is defined, find which is earlier: start + maxNumberDays or station retrieval?
  if(hasArg(maxNumberDays_tmp)) {
    
    if(day1_tmp %in% c("survey", "station") == FALSE){
      # count maximum number days from the beginning of each station's 1st occasion
      date_ranges$start_first_occasion_plus_maxNumberDays <- date_ranges$start_first_occasion_survey + (maxNumberDays_tmp * 86400) - 1   # -1 second ensures that last occasion does not spill into next day if occasionStartTime = 0
    } else {
      # count maximum number days from the beginning of survey's 1st occasion
      date_ranges$start_first_occasion_plus_maxNumberDays <- date_ranges$start_first_occasion + (maxNumberDays_tmp * 86400) - 1   # -1 second ensures that last occasion does not spill into next day if occasionStartTime = 0
    }
    
    # end of last occasion by staion (either end of retrieval day, or after maximum number of days)
    for(xy in 1:nrow(date_ranges)){
      date_ranges$end_last_occasion[xy] <- min(date_ranges$end_of_retrieval_day[xy], date_ranges$start_first_occasion_plus_maxNumberDays[xy])   # use smaller value
    }
    # assign the attributes: POSIX + time zone (to convert from numeric value back to date/time)
    attributes(date_ranges$end_last_occasion) <- attributes(date_ranges$start_first_occasion)   
  } 
  
  # if maxNumberDays is not defined, occasions end on station retrieval?
  if(!hasArg(maxNumberDays_tmp)) {
    date_ranges$end_last_occasion <- date_ranges$end_of_retrieval_day
  }
  
  return(date_ranges)
  
}




# check camera operation matrix with date range table  ####
adjustCameraOperationMatrix <- function(cam.op,
                                        date_ranges2,
                                        timeZone_tmp,
                                        day1_2
){
  # remove stations where occasions begin after end of last occasion (if buffer argument is too large)
  if(any(date_ranges2$start_first_occasion > date_ranges2$end_last_occasion)){
    remove.these.stations <- which(date_ranges2$start_first_occasion > date_ranges2$end_last_occasion)
    if(length(remove.these.stations) == nrow(date_ranges2)) stop("In all stations, the occasions begin after retrieval. Choose a smaller buffer argument.")
    cam.op [remove.these.stations, ] <- NA
  }
  
  
  ##################################
  # set values before beginning of first occasion NA in camera operation matrix (so effort is 0 before): only relevant if buffer was used
  first_col_to_keep2  <- match(as.character(as.Date(date_ranges2$start_first_occasion, tz = timeZone_tmp)), colnames(cam.op))
  
  
  for(xxx in 1:length(first_col_to_keep2)){
    if(first_col_to_keep2[xxx] > 1){         # if it does not begin on 1st day of camera operation matrix
      cam.op[xxx, seq(1, first_col_to_keep2[xxx]-1)] <- NA
    }
  }
  
  # set values after end of last occasion NA in camera operation matrix (so effort is 0 afterwards)
  last_col_to_keep2  <- match(as.character(as.Date(date_ranges2$end_last_occasion, tz = timeZone_tmp)), colnames(cam.op))
  
  
  for(yyy in 1:length(last_col_to_keep2)){
    if(last_col_to_keep2[yyy]+1 < ncol(cam.op)){   # if it does not end on last day of camera operation matrix
      cam.op[yyy, seq(last_col_to_keep2[yyy]+1, ncol(cam.op))] <- NA
    }
  }
  
  
  ####################################
  # trim camera operation matrix (taking into account buffer, occasionStartTime(?), maxNumberDays)
  # based on data frame   "date_ranges"   computed by function   createDateRangeTable
  
  if(day1_2 == "station") {    # 1st day of each station OR some specified date
    
    cam.tmp.min <- apply(cam.op, MARGIN = 1, function(X){min(which(!is.na(X)))})    # first occasion of each station
    cam.tmp.max <- apply(cam.op, MARGIN = 1, function(X){max(which(!is.na(X)))})    # last occasion of each station
    
    diff.days.tmp <- cam.tmp.max - cam.tmp.min
    
    cam.op2 <- matrix(NA,
                      nrow = nrow(cam.op),
                      ncol = max(diff.days.tmp)+1)
    
    # make all stations begin in 1st column
    for(l in 1:nrow(cam.op)){
      if(is.finite(diff.days.tmp[l])){
        cam.op2[l, 1:(diff.days.tmp[l]+1)] <- as.vector(cam.op[l,cam.tmp.min[l]:cam.tmp.max[l]])
      }
    }
    
    if(day1_2 == "station") {
      colnames(cam.op2) <- paste("day", 1:ncol(cam.op2), sep = "")
    }
    
    rownames(cam.op2) <- rownames(cam.op)
    
    cam.op <- cam.op2
    
    
  } else {
    
    
    # remove all columns of cam.op that were before beginning of 1st occasion
    first_col_to_keep <- match(as.character(min(as.Date(date_ranges2$start_first_occasion, tz = timeZone_tmp))), colnames(cam.op))
    if(!is.na(first_col_to_keep)){
      if(first_col_to_keep != 1){
        cam.op <- cam.op[,-seq(1, (first_col_to_keep-1))]
      }
    }
    
    # remove all columns of cam.op that were after end of last occasion / after retrieval of last camera
    last_col_to_keep  <- match(as.character(max(as.Date(date_ranges2$end_last_occasion, tz = timeZone_tmp))), colnames(cam.op))
    if(!is.na(last_col_to_keep)){
      if(last_col_to_keep != ncol(cam.op)){
        cam.op <- cam.op[,-seq((last_col_to_keep + 1), ncol(cam.op))]
      }
    }
    
  }
  return(cam.op)
}


# check consistency of species record table before creating detection history (remove records outside date range etc)  ####
cleanSubsetSpecies <- function(subset_species2,
                               stationCol2,
                               date_ranges2
){
  
  nrow_subset_species2 <- nrow(subset_species2)
  
  
  # remove records that were taken before beginning of first occasion (because of buffer, occasionStartTime, day1)
  corrected_start_time_by_record <- date_ranges2$start_first_occasion[match(subset_species2[,stationCol2], rownames(date_ranges2))]
  
  remove.these <- which(subset_species2$DateTime2 < corrected_start_time_by_record)
  if(length(remove.these) >= 1){
    
    warning(paste(length(remove.these), 
                  " records (out of ", 
                  nrow_subset_species2, 
                  ") were removed because they were taken before day1 (if a date was specified),  within the buffer period, or before occasionStartTime on the 1st day, e.g.:\n", sep = ""),
            paste(head(subset_species2[remove.these, stationCol2]), 
                       head(subset_species2$DateTime2[remove.these]), 
                       collapse = "\n", 
                       sep = ": "),
            call. = FALSE)
    
    subset_species2 <- subset_species2[-remove.these,]
    if(nrow(subset_species2) == 0) stop("No more records after removing records before survey begin. The detection history would be empty.")
    
    rm(corrected_start_time_by_record, remove.these)
  }
  
  # remove records that were taken after end of last occasion (because of maxNumberDays)
  corrected_end_time_by_record <- date_ranges2$end_last_occasion[match(subset_species2[,stationCol2], rownames(date_ranges2))]
  
  remove.these2 <- which(subset_species2$DateTime2 > corrected_end_time_by_record)
  if(length(remove.these2) >= 1){
    
    warning(paste("\n", 
                  paste(length(remove.these2), 
                        " records (out of ", 
                        nrow_subset_species2, 
                        ") were removed because they were taken after the end of the last occasion, e.g.:", sep = ""),
                  paste(head(subset_species2[remove.these2, stationCol2]), 
                        head(subset_species2$DateTime2[remove.these2]), 
                        collapse = "\n", 
                        sep = ": "), 
                  sep = "\n"), 
            call. = FALSE)
    
    subset_species2 <- subset_species2[-remove.these2,]
    
    if(nrow(subset_species2) == 0) stop("No more records left. The detection history would be empty.")
    rm(corrected_end_time_by_record, remove.these2)
  }
  
  return(subset_species2)
}


# calculate trapping effort matrix by day   ####
calculateTrappingEffort <- function(cam.op,
                                    occasionLength2,
                                    scaleEffort2,
                                    includeEffort2,
                                    minActiveDaysPerOccasion2,
                                    occasionStartTime2){
  
  ######################
  # calculate trapping effort by station and occasion
  
  if(occasionLength2 == 1){
    effort <- cam.op          # if occasionLength2 = 1 day, it is identical
  } else {
    effort <- matrix(NA, nrow = nrow(cam.op), ncol = ceiling(ncol(cam.op) / occasionLength2 ))
    
    index <- 1
    for(m in 1:ncol(effort)){    # for every occasion in the effort matrix
      # index for columns in camera operation matrix to aggregate
      if(index + occasionLength2 <= ncol(cam.op)){
        index.tmp <- index : (index + occasionLength2 - 1)
      } else {
        index.tmp <- index : ncol(cam.op)
      }
      
      # calculate effort as sum of active days per occasion
      effort[, m] <- apply(as.matrix(cam.op[,index.tmp]), MARGIN = 1, FUN = sum, na.rm = TRUE)
      # if full occasion NA in cam.op, make effort NA
      effort[, m] <- ifelse(apply(as.matrix(cam.op[,index.tmp]), MARGIN = 1, FUN = function(X) {sum(is.na(X))}) == length(index.tmp), NA, effort[,m])
      # if full occasion = 0 in cam.op, make effort NA
      effort[, m] <- ifelse(apply(as.matrix(cam.op[,index.tmp]), MARGIN = 1, FUN = function(X) {all(X == 0)}),   NA, effort[,m])
      
      ## if full occasion is smaller than 1 (i.e. all 0 or NA), set effort NA
      ## EDIT: 2020-09-12: This can cause problems if cameraOperation calculated fraction of days.
      ## Also, the cases with 0 and NA are covered above. So I comment it for now.
      #effort[, m] <- ifelse(apply(as.matrix(cam.op[,index.tmp]), MARGIN = 1, FUN = function(X) {all(X < 1)}),   NA, effort[,m])
      
      # set cells in effort matrix NA (according to input arguments)
      # this is later used to adjust the detection/non-detection matrix
      if(includeEffort2 == FALSE){
        if(hasArg(minActiveDaysPerOccasion2)){   # includeEffort = FALSE and minActiveDays is defined
          # if occasion has less active days than minActiveDays, set NA
          effort[, m] <- ifelse(apply(as.matrix(cam.op[,index.tmp]), MARGIN = 1, FUN = function(X) {sum(X, na.rm = TRUE) < minActiveDaysPerOccasion2}), NA, effort[,m])
        } else {                                 # includeEffort = FALSE and minActiveDays not defined
          # if any day NA in cam.op, make  occasion effort NA
          effort[, m] <- ifelse(apply(as.matrix(cam.op[,index.tmp]), MARGIN = 1, FUN = function(X) {any(is.na(X))}), NA, effort[,m])
          # if any day = 0 in cam.op, make occasion effort NA
          effort[, m] <- ifelse(apply(as.matrix(cam.op[,index.tmp]), MARGIN = 1, FUN = function(X) {any(X == 0)}),   NA, effort[,m])
          
          if(length(index.tmp) < occasionLength2){  # if occasion is shorter than occasionLength (i.e. last occasion), set NA.
            effort[, m] <- NA
          }
          
        }
        
        ## includeEffort = TRUE
      } else {    
        # if minActiveDays is defined
        if(hasArg(minActiveDaysPerOccasion2)){   
          # if occasion has less actice days than minActiveDays, set NA
          effort[, m] <- ifelse(apply(as.matrix(cam.op[,index.tmp]), MARGIN = 1, FUN = function(X) {sum(X, na.rm = TRUE) < minActiveDaysPerOccasion2}), NA, effort[,m])
        } else {                                 # includeEffort = TRUE and minActiveDays not defined
          # if all days of occasion NA in cam.op, make  occasion effort NA
          #effort[, m] <- ifelse(apply(as.matrix(cam.op[,index.tmp]), MARGIN = 1, FUN = function(X) {all(is.na(X))}), NA, effort[,m])
          # if all days of occasion = 0 in cam.op, make occasion effort NA
          
        }
      }
      index <- index + occasionLength2
    }
    rm(index, index.tmp)
  }
  
  # scale effort, if required
  if(isTRUE(scaleEffort2)){
    if(occasionLength2 == 1) stop("cannot scale effort if occasionLength is 1", call. = FALSE)
    if(length(table(effort)) == 1) stop(paste("all values of effort are identical (", names(table(effort)), "). Cannot scale effort", sep = ""), call. = FALSE)
    
    scale.eff.tmp <- scale(as.vector(effort))                       # scale effort (as a vector, not matrix)
    scale.eff.tmp.attr <- data.frame(effort.scaled.center = NA,     # prepare empty data frame
                                     effort.scaled.scale = NA)
    scale.eff.tmp.attr$effort.scaled.center[1] <- attr(scale.eff.tmp, which = "scaled:center")   # write scaling parameters to data frame
    scale.eff.tmp.attr$effort.scaled.scale [1] <- attr(scale.eff.tmp, which = "scaled:scale")
    effort <- matrix(scale.eff.tmp, nrow = nrow(effort), ncol = ncol(effort))                    # convert effort vector to matrix again
  }
  
  rownames(effort) <- rownames(cam.op)
  
  # return objects
  if(isTRUE(scaleEffort2)){
    return(list(effort, scale.eff.tmp.attr))
  } else {
    return(list(effort))
  }
}



##########################################################################################################
# for function surveyReport

makeSurveyZip <- function(output,
                          recordTable,
                          CTtable ,
                          speciesCol,
                          stationCol,
                          cameraCol,
                          setupCol,
                          retrievalCol,
                          CTDateFormat,
                          CTHasProblems,
                          Xcol,
                          Ycol,
                          recordDateTimeCol,
                          recordDateTimeFormat,
                          sinkpath,
                          usePackageZip){
  
  wd0 <- getwd()
  on.exit(setwd(wd0))
  
  dir.tmp <- tempdir()
  
  file.sep <- .Platform$file.sep
  
  
  dir.zip <- file.path(dir.tmp, paste("surveyReport_", Sys.Date(), sep = ""))
  dir.zip.short <- paste("surveyReport_", Sys.Date(), sep = "")
  unlink(dir.zip, recursive = TRUE)
  dir.create(dir.zip, showWarnings = FALSE, recursive = TRUE)
  
  # create directories
  invisible(sapply(file.path(dir.zip, c("surveyReport", "activity", "scripts", "detectionMaps")), dir.create, showWarnings = FALSE))
  
  ######
  # save input tables
  
  write.csv(recordTable, file = file.path(dir.zip, "recordTable.csv"), row.names = FALSE)
  write.csv(CTtable, file = file.path(dir.zip, "CTtable.csv"), row.names = FALSE)
  
  ######
  # save surveyReport tables
  
  dir.tmp2 <- file.path(dir.zip, "surveyReport")
  
  for(xyz in 1:length(output)){
    write.csv(output[[xyz]],
              file = file.path(dir.tmp2, paste(names(output)[[xyz]], ".csv", sep = "")),
              row.names = FALSE)
  }
  rm(xyz)
  
  ######
  # make activity plots
  
  activityDensity(recordTable          = recordTable,
                  allSpecies           = TRUE,
                  speciesCol           = speciesCol,
                  recordDateTimeCol    = recordDateTimeCol,
                  recordDateTimeFormat = recordDateTimeFormat,
                  plotR                = FALSE,
                  writePNG             = TRUE,
                  plotDirectory        = file.path(dir.zip, "activity"))
  
  ######
  # make detection maps
  
  if(hasArg(Xcol) & hasArg(Ycol)){
    detectionMaps(CTtable       = CTtable,
                  recordTable   = recordTable,
                  Xcol          = Xcol,
                  Ycol          = Ycol,
                  stationCol    = stationCol,
                  speciesCol    = speciesCol,
                  richnessPlot  = TRUE,
                  speciesPlots  = TRUE,
                  printLabels   = TRUE,
                  plotR         = FALSE,
                  writePNG      = TRUE,
                  plotDirectory = file.path(dir.zip, "detectionMaps"),
                  pngMaxPix     = 1000
    )
  }
  
  ########################################################################################
  # prepare scripts
  
  scriptfile <- file.path(dir.zip, "scripts", "camtrapR_scripts.R")
  file.create(scriptfile, showWarnings = FALSE)
  
  # load basic data
  
  sink(file = scriptfile)
  cat("###  load data tables  ### \n\n")
  
  cat("directory.data <- PLEASE_SPECIFY        # this is the directory you got after unzipped the zip file to (e.g. .../surveyReport_2016-02-29/)  \n\n")
  
  
  cat("CTtable     <- read.csv(paste(directory.data, 'CTtable.csv', sep = '/'))\n")
  cat("recordTable <- read.csv(paste(directory.data, 'recordTable.csv', sep = '/'))\n\n\n")
  
  sink()
  
  
  # make detection maps   # no, because of coordinate columns
  
  sink(file = scriptfile, append = TRUE)
  cat("###  plot species detections  ### \n\n")
  if(hasArg(Xcol) & hasArg(Ycol)){
    cat(paste("Xcol <- '", Xcol, "'\n", sep = ""))
    cat(paste("Ycol <- '", Ycol, "'\n\n", sep = ""))
  } else {
    cat("Xcol <- PLEASE_SPECIFY\n")
    cat("Ycol <- PLEASE_SPECIFY\n\n")
  }
  
  cat(paste("detections <- detectionMaps(CTtable = CTtable,
            recordTable  = recordTable,
            Xcol         = Xcol,
            Ycol         = Ycol,
            stationCol   = '", stationCol, "',
            speciesCol   = '", speciesCol, "',
            writePNG     = FALSE,
            plotR        = TRUE,
            printLabels  = TRUE,
            richnessPlot = TRUE,
            addLegend    = TRUE
  ) \n\n\n", sep = ""))

  sink()
  
  
  # camera operation matrix
  
  # Conditionally create the string for each optional argument
  
  # 1. Handle cameraCol
  if (hasArg(cameraCol)) {
    # If cameraCol was provided, create the active line of code
    cameraCol_line <- sprintf("            cameraCol = '%s',\n", cameraCol)
  } else {
    # If cameraCol was not provided, create a commented-out placeholder
    cameraCol_line <- "            #cameraCol = 'Not defined',\n"
  }
  
  # 2. Handle byCamera
  if (hasArg(cameraCol)) {
    # If cameraCol was provided, set byCamera = FALSE (aggregate to station)
    byCamera_line <- "            byCamera = FALSE,\n"
  } else {
    byCamera_line <- "            #byCamera,\n"
  }
  
  # 3. Handle allCamsOn
  if (hasArg(cameraCol)) {
    # If cameraCol was provided, set allCamsOn = FALSE (one out of several cameras active is acceptable)
    allCamsOn_line <- "            allCamsOn = FALSE,\n"
  } else {
    allCamsOn_line <- "            #allCamsOn,\n"
  }
  
  # 4. Handle camerasIndependent
  if (hasArg(cameraCol)) {
    camerasIndependent_line <- "            camerasIndependent = TRUE,  #  please check \n"
  } else {
    camerasIndependent_line <- "            #camerasIndependent,\n"
  }
  
  
  sink(file = scriptfile, append = TRUE)
  cat("###  camera operation matrix  ### \n\n")
  
  cat(paste("camOp <- cameraOperation(CTtable = CTtable,
            stationCol = '", stationCol, "',\n",
            cameraCol_line, 
            "            setupCol  = '", setupCol, "',
            retrievalCol  = '", retrievalCol, "',
            hasProblems = '", CTHasProblems, "',\n",
            byCamera_line,
            allCamsOn_line,
            camerasIndependent_line,
            "            dateFormat  = '", CTDateFormat, "' #,
            #writecsv  = FALSE,
            #outDir
  ) \n\n\n", sep = ""))

  sink()
  
  
  # make detection histories
  
  sink(file = scriptfile, append = TRUE)
  cat("###  detection histories  ### \n\n")
  
  cat("day1              <- PLEASE_SPECIFY \n")
  cat("occasionLength    <- PLEASE_SPECIFY\n")
  cat("speciesOfInterest <- PLEASE_SPECIFY\n")
  cat("timeZone          <- PLEASE_SPECIFY \n\n")
  
  cat(paste("detHist <- detectionHistory(recordTable = recordTable,
            species                                  = speciesOfInterest,
            camOp                                    = cameraOperation,
            stationCol                               = '", stationCol, "',
            speciesCol                               = '", speciesCol, "',
            recordDateTimeCol                        = '", recordDateTimeCol, "',
            recordDateTimeFormat                     = '", recordDateTimeFormat, "',
            occasionLength                           = occasionLength,
            #maxNumberDays,
            day1                                     = day1,
            #buffer                                  = 0,
            includeEffort                            = TRUE,
            scaleEffort                              = FALSE,
            datesAsOccasionNames                     = FALSE,
            timeZone                                 = timeZone,
            writecsv                                 = FALSE #,
            # outDir
  ) \n\n\n", sep = ""))

  sink()
  
  
  ### write description text file ###
  
  readmefile <- file.path(dir.zip, "readme.txt")
  file.create(readmefile, showWarnings = FALSE)
  
  sink(file = readmefile)
  
  
  cat(paste("this zip file contains a summary of a camera trapping survey:\n\n"))
  cat(paste("total survey period:              ", min(output$survey_dates$setup), "-", max(output$survey_dates$retrieval), "\n"))
  cat(paste("Total number of stations:         ", nrow(output$survey_dates), "\n"))
  cat(paste("Number of operational stations:   ", length(which(output$survey_dates$n_nights_active >= 1)), "\n"))
  cat(paste("Total number of cameras:          ", sum(output$survey_dates$n_cameras), "\n"))
  cat(paste("Total number of active trap days: ", sum(output$survey_dates$n_nights_active), "\n\n"))
  
  cat("\n-------------------------------------------------------\n\n")
  
  cat(paste("the following table shows a summary of survey period for each station \n\n"))
  print(output$survey_dates, row.names = FALSE)
  cat("\n-------------------------------------------------------\n\n")
  
  cat("legend to the data structure in the zip file:\n\n")
  cat(".../activity/              plots of activity density estimations for each species created with activityDensity\n")
  if(hasArg(Xcol) & hasArg(Ycol)){cat(".../detectionMaps/         maps of observed species richness and numbers of species records\n")}
  cat(".../scripts/               a prepared R script for occupancy analyses\n")
  cat(".../surveyReport/          the tables created by surveyReport summarising the survey\n")
  cat(".../CTtable.csv            table of camera trap station IDs, operation times and coordinates\n")
  cat(".../recordTable.csv        table of species records\n")
  cat(".../readme.txt             this file\n\n")
  
  cat("\n-------------------------------------------------------\n\n")
  
  cat(paste("species images are located in:\n\n"))
  cat(paste(as.character(recordTable$Directory)[1], "          # the first record\n"))
  cat(paste(as.character(recordTable$Directory)[nrow(recordTable)], "          # the last record\n"))
  
  cat("\n-------------------------------------------------------\n\n")
  
  cat("information about who created this report:\n\n")
  print(as.data.frame(Sys.info()))
  
  cat(paste("\n\n\n\n ***   report created by function surveyReport on",  Sys.time(),  "   ***\n\n\n"))
  
  
  sink()
  
  ######
  # make final zip file
  
  # list files
  files2zip <- dir(path = dir.zip, full.names = FALSE, recursive = TRUE)
  files2zip <- file.path(dir.zip.short, files2zip)
  
  # write zip
  setwd(dir.tmp)
  
  if(isFALSE(usePackageZip)) {
  zip(zipfile = file.path(sinkpath,
                          paste(dir.zip.short, ".zip", sep = "")),
                       files   = files2zip,
                       flags   = "")
  }
  
  if(isTRUE(usePackageZip)) {
    zip::zipr(zipfile = file.path(sinkpath,
                                  paste(dir.zip.short, ".zip", sep = "")),
              files   = files2zip)
  }
  
  # check if output was created
  if(file.exists(file.path(sinkpath,
                           paste(dir.zip.short, ".zip", sep = "")))){
  message("zip file compiled \n",
      paste(sinkpath, paste(dir.zip.short, ".zip\n\n", sep = ""), sep = file.sep))
  } else {
    message("zip file creation failed")
  }
}



##########################################################################################################
# assign session IDs to records in recordTableIndividual based on a session column in camera trap table    ####
# This is for when there are several entries for each station, one for each session
# in other words, if stations were operated continuously and session start/end dates are defined in camera trap table

assignSessionIDtoRecordTableIndividual <- function(recordTableIndividual_tmp,
                                                   cameraTrapTable_tmp,
                                                   stationCol,
                                                   sessionCol,
                                                   setup_date_col,
                                                   retrieval_date_col){
  
  # check input
  if(sessionCol %in% colnames(recordTableIndividual_tmp)) stop("recordTableIndividual has a session column already. Cannot assign new session ID",
                                                               call. = FALSE)
  if(!setup_date_col %in% colnames(cameraTrapTable_tmp)) stop("CTtable setup column not found",
                                                              call. = FALSE)
  if(!retrieval_date_col %in% colnames(cameraTrapTable_tmp)) stop("CTtable retrieval column not found",
                                                                  call. = FALSE)
  
  # define "between" function
  `%between%` <- function(x, interval) x >= interval[1] & x <= interval[2]
  
  # add an empty session column to the record table
  recordTableIndividual_tmp[, sessionCol] <- NA
  
  # for each station, find the respective sessions and assign the session IDs to the records based on their date.
  for(i in 1:length(unique(cameraTrapTable_tmp[,stationCol]))){
    
    # find index of records in station i and respective dates
    record_rows_id_tmp <- which(recordTableIndividual_tmp[,stationCol] == unique(cameraTrapTable_tmp[,stationCol])[i])
    rec_dates_tmp <- recordTableIndividual_tmp$Date [record_rows_id_tmp]
    
    # NOTE: output may be assigned to wrong session if occasionStartTime != 0!!!
    # There is also an issue when station ID is a combination of station and session IDs.
    
    # extract session start/end dates (by session)
    session_id    <- cameraTrapTable_tmp[, sessionCol]         [cameraTrapTable_tmp[,stationCol] == unique(cameraTrapTable_tmp[,stationCol])[i]]
    session_start <- cameraTrapTable_tmp[, setup_date_col]     [cameraTrapTable_tmp[,stationCol] == unique(cameraTrapTable_tmp[,stationCol])[i]]
    session_end   <- cameraTrapTable_tmp[, retrieval_date_col] [cameraTrapTable_tmp[,stationCol] == unique(cameraTrapTable_tmp[,stationCol])[i]]
    
    # assign session ID to recordTable
    for(session_id_tmp in session_id){
      record_rows_session_id_tmp <- which(rec_dates_tmp %between% c(session_start[session_id_tmp], session_end[session_id_tmp]))
      recordTableIndividual_tmp[record_rows_id_tmp, sessionCol]  [record_rows_session_id_tmp] <- session_id_tmp
    }
    rm(session_id, session_start, session_end, record_rows_id_tmp, rec_dates_tmp)
  }
  
  # create new station columns based on station + session (and create backup first)
  
  recordTableIndividual_tmp$station_backup <- recordTableIndividual_tmp[, stationCol]
  cameraTrapTable_tmp$station_backup       <- cameraTrapTable_tmp[, stationCol]
  
  recordTableIndividual_tmp[, stationCol] <- paste(recordTableIndividual_tmp[, stationCol], recordTableIndividual_tmp[, sessionCol], sep = "_session")
  cameraTrapTable_tmp[, stationCol]       <- paste(cameraTrapTable_tmp[, stationCol], cameraTrapTable_tmp[, sessionCol], sep = "_session")
  
  
  return(list(recordTable = recordTableIndividual_tmp,
              CTtable     = cameraTrapTable_tmp))
}



assignSessionIDtoRecordTable <- function(recordTable_tmp,
                                         camOp,
                                         dateTimeCol,
                                         stationCol,
                                         sessionCol
){
  
  
  camop.info.df <- deparseCamOpRownames(camOp)
  
  # if there's records at stations that are not in the camop row names
  if(!any(recordTable_tmp[, stationCol] %in% unique(camop.info.df$station))) {
    stop("Stations in record table station column don't match stations in row names of camera operation matrix.")
  }
    
  if(!all(recordTable_tmp[, stationCol] %in% unique(camop.info.df$station))) {
    warning(paste("Some stations in the record table are not matched by stations in camOp row names:\n",
                  paste(unique(recordTable_tmp[!recordTable_tmp[, stationCol] %in% unique(camop.info.df$station), stationCol]), collapse = ", ")))
  }
  
  recordTable_tmp[, sessionCol] <- NA
  
  record_dates_tmp <- as.Date(recordTable_tmp[, dateTimeCol])
  if(all(is.na(record_dates_tmp))) stop("Could not interpret record date/time. Output is all NA.")
  
  record_dates_backup <- record_dates_tmp
  camOp_dates_tmp  <- as.Date(colnames(camOp))
  
  # remove records before / after camOp date range
  records_too_early <- which(sapply(record_dates_tmp, FUN = function(x) x < min(camOp_dates_tmp)))
  
  if(length(records_too_early) > 0) {
    recordTable_tmp <- recordTable_tmp[-records_too_early, ]
    warning(paste(length(records_too_early), "records were removed because they were before the first day in the camera operation matrix"), call. = FALSE)
    record_dates_tmp <- as.Date(recordTable_tmp[, dateTimeCol])
  }
  
  records_too_late  <- which(sapply(record_dates_tmp, FUN = function(x) x > max(camOp_dates_tmp)))
  
  if(length(records_too_late) > 0) {
    recordTable_tmp <- recordTable_tmp[- records_too_late, ]
    warning(paste(length(records_too_late), "records were removed because they were after the last day in the camera operation matrix"), call. = FALSE)
    record_dates_tmp <- as.Date(recordTable_tmp[, dateTimeCol])
  }
  
  if(nrow(recordTable_tmp) == 0) stop(paste0("Date range mismatch between records (", paste(range(record_dates_backup), collapse = " - "),
                                            ") and camera operation matrix (", paste(range(camOp_dates_tmp), collapse = " - "), ")"))
  
  for(i in 1:nrow(camop.info.df)){
    # for every row in camOp, find all records at that station  
    which_tmp <- which(recordTable_tmp[, stationCol] == camop.info.df$station[i])
    
    if(length(which_tmp) >= 1){
      # of the records at that station, which are not NA in that row of the camOp (there are different rows for different sessions)
      which_tmp2 <- which(!is.na(camOp[i, as.character(record_dates_tmp[which_tmp])]))
      
      recordTable_tmp[which_tmp[which_tmp2], sessionCol] <- camop.info.df$session[i]
    }
  }
  
  which_na <- which(is.na(recordTable_tmp[, sessionCol]))
  
  if(length(which_na) > 0) {
    recordTable_tmp_na <- recordTable_tmp[which_na, ]
    recordTable_tmp <- recordTable_tmp[-which_na, ]
    if(length(which_na) == nrow(recordTable_tmp)) {
      stop("Could not assign session to any record. Please check format of column names of camOp and station column of record table")
    } else {
      warning(paste0(length(which_na), " records were removed because they could not be assigned to an active station / session:\n",
                    paste(recordTable_tmp_na[, stationCol], recordTable_tmp_na[, dateTimeCol], sep = ": ", collapse = "\n")), call. = FALSE)
    }
  }
  
  separatorSession <- "__SESS_"
  stationCol_backup <- paste(stationCol, "backup", sep = "_")
  
  recordTable_tmp[, stationCol_backup] <- recordTable_tmp[, stationCol]
  
  recordTable_tmp[, stationCol] <- paste(recordTable_tmp[, stationCol], recordTable_tmp[, sessionCol], sep = separatorSession)
  
  return(recordTable_tmp)
}



# checks if input is data.frame or tibble. If tibble, either convert to data.frame or stop  ####

dataFrameTibbleCheck <- function(df, 
                                 tibble_allowed = TRUE, 
                                 data_table_allowed = TRUE,
                                 message = FALSE){
  
  # check if it is a data.frame at all
  if(!is.data.frame(df)) stop(paste(substitute(df), "must be a data.frame"), call. = FALSE)
  
  # handling tibbles (tidyverse)
  if (requireNamespace("tibble", quietly = TRUE)) {
    
    if(tibble::is_tibble(df)) {
      if(tibble_allowed) {
        if(message) message (paste(substitute(df), "was converted from tibble to data.frame"))
        df <- as.data.frame(df)
      } else {
        stop (paste(substitute(df), "is a tibble. Please provide a data.frame instead (use read.csv() or as.data.frame())"), call. = FALSE)
      }
    }
  }
  
  # handling data.tables
  
  if (requireNamespace("data.table", quietly = TRUE)) {
    
    if(data.table::is.data.table(df)) {
      if(data_table_allowed) {
        message (paste(substitute(df), "was converted from data.table to data.frame"))
        #df <- as.data.frame(df)
        df <- setDF(df)
      } else {
        stop (paste(substitute(df), "is a data.table Please provide a data.frame instead (use read.csv() or as.data.frame())"), call. = FALSE)
      }
    }
  }
  return(df)
}


# check and convert dates (character) to date objects, either with base functions or lubridate    ####

# CHECK: timeZone argument needed in surveyReport?

parseDateObject <- function(inputColumn,     
                            dateFormat,   
                            checkNA,      # throw error if there are NAs in input (only setup / retrieval, not problems)
                            checkEmpty,    # throw error if there are blank values in input  (only setup / retrieval, not problems)
                            allowEmptyOutput = FALSE,
                            returndatetime = FALSE
){
  
  #if(!class(inputColumn) %in% c("factor", "character")) stop(paste("date column must be a factor or character:", deparse(substitute(inputColumn))), call. = FALSE)
  
  if(checkNA & any(is.na(inputColumn)))   stop(paste("there are NAs in", deparse(substitute(inputColumn))), call. = FALSE)
  if(checkEmpty & any(inputColumn == "")) stop(paste("there are blank values in", deparse(substitute(inputColumn))), call. = FALSE)
  
  if(all(is.na(inputColumn)) & allowEmptyOutput) return(NA)
  if(all(inputColumn == "")  & allowEmptyOutput) return(NA)
  
  inputColumn.char <- as.character(inputColumn)
  
  # option 1: base functions for dates as per strptime (identified by "%")
  if(grepl(pattern = "%", x = dateFormat, fixed = TRUE)){
    if(any(grepl(pattern = " ", inputColumn.char))) warning(paste0("There are spaces in ", deparse(substitute(inputColumn)), ", but not in dateFormat\n"), call. = FALSE)
    out <- as.Date(inputColumn.char,     format = dateFormat)
  } else {
  # option 2: lubridate functions (identified by absence of "%")
    if(!requireNamespace("lubridate", quietly = TRUE)) stop(paste("package 'lubridate' is required for the specified dateFormat", dateFormat))
    
    out <- lubridate::date(lubridate::parse_date_time(inputColumn.char, orders = dateFormat))
  }
  
  if(all(is.na(out))) stop(paste0("Cannot read date format in ", deparse(substitute(inputColumn)), ". Output is all NA.\n",
                                  "expected:  ", dateFormat, "\nactual:    ", inputColumn[1]), call. = FALSE)
  
  if(checkNA & any(is.na(out))) stop(paste("At least one entry in", deparse(substitute(inputColumn)), "cannot be interpreted using dateFormat:", dateFormat, "\n",
                                           "rows", paste(which(is.na(out)), collapse = ", ")), call. = FALSE)
  if(isTRUE(returndatetime))  return(as_datetime(out, tz = "UTC"))
  if(isFALSE(returndatetime)) return(out)
}


# check and convert date - time (character) to datetime objects (POSIXct), either with base functions or lubridate    ####

parseDateTimeObject <- function(inputColumn,   
                                dateTimeFormat,
                                timeZone,
                                checkNA = TRUE,      # throw error if there are NAs in input (only setup / retrieval, not problems)
                                checkEmpty = TRUE,   # throw error if there are blank values in input  (only setup / retrieval, not problems)
                                checkNA_out = TRUE,  # throw error when there is NAs in output (FALSE so reporting is done by detectionHistory, which returns correct row numbers)
                                allowEmptyOutput = FALSE,  
                                quiet = FALSE
){
  
  if(inherits(inputColumn, c("POSIXct", "POSIXlt"))){
    if(inherits(inputColumn, "POSIXlt")) {
      # warning(paste("datetime column is in POSIXlt format. Converting to character:", deparse(substitute(inputColumn)), ""), call. = FALSE)
    }
    
    if(inherits(inputColumn, "POSIXct")){
      # message(paste("datetime column is in POSIXct format. Converting to character:", deparse(substitute(inputColumn)), ""), call. = FALSE)
    }
    # inputColumn <- as.character(inputColumn)  # converts date-time to date if time = 00:00:00
    inputColumn <- format(inputColumn, format = "%Y-%m-%d %H:%M:%S")
  } else {
    if(!inherits(inputColumn, "character")) stop(paste("datetime column must be a character:", deparse(substitute(inputColumn))), call. = FALSE)
  }
  
  if(checkNA & any(is.na(inputColumn)))   stop(paste("there are NAs in", deparse(substitute(inputColumn))), call. = FALSE)
  if(checkEmpty & any(inputColumn == "")) stop(paste("there are blank values in", deparse(substitute(inputColumn))), call. = FALSE)
  
  if(all(is.na(inputColumn)) & allowEmptyOutput) return(NA)
  if(all(inputColumn == "") & allowEmptyOutput) return(NA)
  
  
  inputColumn.char <- as.character(inputColumn)
  
  # option 1: base functions for dates as per strptime (identified by "%")
  if(grepl(pattern = "%", x = dateTimeFormat[1], fixed = TRUE)){
    out <- as.POSIXct(inputColumn.char, tz = timeZone, format = dateTimeFormat)
  } else {
    # option 2: lubridate functions (identified by absence of "%")
    if(!requireNamespace("lubridate", quietly = TRUE)) stop(paste("package 'lubridate' is required for the specified dateTimeFormat", dateTimeFormat))
    
    out <- lubridate::parse_date_time(inputColumn.char, orders = c(dateTimeFormat, 
                                                                   "ymd"),   # to catch records exactly on midnight where R drops the time
                                      tz = timeZone, 
                                      quiet = quiet,
                                      truncated = 3)    # allow time to be missing, defaulting to 00:00:00
  }
  
  if(all(is.na(out))) stop(paste0("Cannot read datetime format in ", deparse(substitute(inputColumn)), ". Output is all NA.\n",
  "expected:  ", dateTimeFormat, "\nactual:    ", inputColumn[1]), call. = FALSE)
  
  if(checkNA_out & any(is.na(out))) stop(paste(sum(is.na(out)), "out of", length(out), "records in",
                                               deparse(substitute(inputColumn)), "cannot be interpreted using dateTimeFormat:", dateTimeFormat, "\n",
                                               "rows", paste(which(is.na(out)), collapse = ", ")), call. = FALSE)
  
 
  if(inherits(inputColumn, c("POSIXct", "POSIXlt"))) stop("couldn't interpret recordDateTimeCol of recordTable using specified recordDateTimeFormat. Output is not POSIX object")
  return(out)
}

# make empty matrix, a row for each unique station / camera combination ----
stationSessionCamMatrix <- function(CTtable,
                                    stationCol,
                                    cameraCol, 
                                    sessionCol, 
                                    setupCol,
                                    retrievalCol
){
  
  separatorCam     <- "__CAM_"
  separatorSession <- "__SESS_"
  
  double_underscore <- "__"
  
  if(length(grep(pattern = double_underscore, x = CTtable[,stationCol])) >= 1) stop(paste("Station IDs may not contain double underscores", double_underscore), call. = FALSE)
  
  
  # convert setup and retrievalCol to Dates (in case they contain times)
  CTtable[,setupCol]     <- as.Date(as.character(CTtable[,setupCol]))
  CTtable[,retrievalCol] <- as.Date(as.character(CTtable[,retrievalCol]))
  
  
  if(hasArg(sessionCol)){
    if(any(CTtable[,sessionCol] == "")) stop("there are empty cells in sessionCol Please provide camera IDs for all cameras",
                                             call. = FALSE)    
    if(any(is.na(CTtable[,sessionCol]))) stop("there are NAs in sessionCol Please provide camera IDs for all cameras",
                                              call. = FALSE)    
    if(length(grep(pattern = double_underscore, x = CTtable[,sessionCol])) >= 1)  stop(paste("Session IDs may not contain double underscores", double_underscore),  call. = FALSE)
    
    stationsession <- paste(CTtable[,stationCol], CTtable[,sessionCol], sep = separatorSession)
    m <- matrix(ncol = abs(as.integer(max(CTtable[,retrievalCol]) - min(CTtable[,setupCol]))) + 1,
                nrow  = length(stationsession))
    colnames(m) <- as.character(as.Date(min(CTtable[,setupCol]):max(CTtable[,retrievalCol]), origin = "1970-01-01"))
    rownames(m) <- stationsession
  }
  
  if(hasArg(cameraCol)){
    if(any(CTtable[,cameraCol] == "")) stop("there are empty cells in cameraCol. Please provide camera IDs for all cameras",
                                            call. = FALSE)    
    if(any(is.na(CTtable[,cameraCol]))) stop("there are NAs in cameraCol. Please provide camera IDs for all cameras",
                                             call. = FALSE)    
    if(length(grep(pattern = double_underscore, x = CTtable[,cameraCol])) >= 1)  stop(paste("Camera IDs may not contain double underscores", double_underscore),  call. = FALSE)
    
    
    stationcam <- paste(CTtable[,stationCol], CTtable[,cameraCol], sep = separatorCam)
    m <- matrix(ncol = abs(as.integer(max(CTtable[,retrievalCol]) - min(CTtable[,setupCol]))) + 1,
                nrow  = length(stationcam))
    colnames(m) <- as.character(as.Date(min(CTtable[,setupCol]):max(CTtable[,retrievalCol]), origin = "1970-01-01"))
    rownames(m) <- stationcam
  }
  
  if(hasArg(sessionCol) & hasArg(cameraCol)){
    
    stationsessioncam <- paste(CTtable[,stationCol], separatorSession, CTtable[,sessionCol], separatorCam, CTtable[,cameraCol], sep = "")
    m <- matrix(ncol = abs(as.integer(max(CTtable[,retrievalCol]) - min(CTtable[,setupCol]))) + 1,
                nrow  = length(stationsessioncam))
    colnames(m) <- as.character(as.Date(min(CTtable[,setupCol]):max(CTtable[,retrievalCol]), origin = "1970-01-01"))
    rownames(m) <- stationsessioncam
  }
  
  if(!hasArg(sessionCol) & !hasArg(cameraCol)){
    
    m <- matrix(ncol = abs(as.integer(max(CTtable[,retrievalCol]) - min(CTtable[,setupCol]))) + 1,
                nrow  = length(CTtable[, stationCol]))
    colnames(m) <- as.character(as.Date(min(CTtable[,setupCol]):max(CTtable[,retrievalCol]), origin = "1970-01-01"))
    rownames(m) <- CTtable[, stationCol]
  }

  return(m)
}

# analyse row names of camera operation matrix to extract station, camera and session information (the latter two only if applicable)
deparseCamOpRownames <- function(camOp){
  
  separatorCam <- "__CAM_"
  separatorSession <- "__SESS_"
  
  if(is.data.frame(camOp)) camOp <- as.matrix(camOp)
  
  # extract rownames, if it is a matrix (if not I assume it is a vector, the result of rownames(camOp))
  if(is.matrix(camOp)){
    x <- rownames(camOp)
  } else {
    x <- camOp
  }
  
  if(any(grepl(separatorSession, x)) &
     any(grepl(separatorCam, x))){
    x2 <- strsplit(x, split = separatorSession)
    
    stationIDs <- sapply(x2, FUN = function(x)x[1])
    
    sessionCameraIDs <- sapply(x2, FUN = function(x)x[2])
    
    x3 <- strsplit(sessionCameraIDs, split = separatorCam)
    
    sessionIDs <- sapply(x3, FUN = function(x)x[1])
    cameraIDs  <- sapply(x3, FUN = function(x)x[2])
    
    return(data.frame(station = stationIDs,
                      session = sessionIDs,
                      camera = cameraIDs, 
                      stringsAsFactors = FALSE))
  }
  
  if(any(grepl(separatorSession, x))){
    x2 <- strsplit(x, split = separatorSession)
    
    stationIDs <- sapply(x2, FUN = function(x)x[1])
    sessionIDs <- sapply(x2, FUN = function(x)x[2])
    
    return(data.frame(station = stationIDs,
                      session = sessionIDs, 
                      stringsAsFactors = FALSE))
  }
  
  if(any(grepl(separatorCam, x))){
    x2 <- strsplit(x, split = separatorCam)
    
    stationIDs <- sapply(x2, FUN = function(x)x[1])
    cameraIDs  <- sapply(x2, FUN = function(x)x[2])
    
    return(data.frame(station = stationIDs,
                      camera = cameraIDs, 
                      stringsAsFactors = FALSE))
  }
  return(data.frame(station = x, 
                    stringsAsFactors = FALSE))
}



# add NA columns to matrix to achieve a certain number of columns (for creating multi-seasons unmarked frames)
padMatrixWithNA <- function(mat, ncol_desired){
  
  if(ncol_desired < ncol(mat)){
    mat.out <- matrix(c(mat, 
                        rep(NA, times = nrow(mat) * (ncol_desired - ncol(mat)))),
                      ncol = ncol_desired, 
                      nrow = nrow(mat)
    )  
    rownames(mat.out) <- rownames(mat)
    colnames(mat.out) <- colnames(mat)
    return(mat.out)
  } else {
    return(mat)
  }
}

# for progress indicator in user messages (28 charcters wide in total)
makeProgressbar <- function(current, 
                            total){
  progress_bar_width <- 20
  perc <- current / total
  
  pb <- paste("      |", 
              paste(rep("=", times = round(perc * progress_bar_width)), collapse = ""), 
              paste(rep(" ", times = progress_bar_width - round(perc * progress_bar_width)), collapse = ""), 
              "|", 
              "  ", formatC(round(perc * 100), width = 3), "%", sep = "")
}


# Accessing / manipulating digiKam database ----

# access digiKam database and provide tables to extract species tags for videos
# call before exiftool

accessDigiKamDatabase <- function(db_directory,   # database directory 
                                  db_filename     # database filename
)
{
  # ensure database directory and file exist
  if(!dir.exists(db_directory)) stop("Could not find directory 'db_directory'", call. = FALSE)
  if(!file.exists(file.path(db_directory, db_filename))) stop("Could not find db_filename in db_directory", call. = FALSE)
  
  # establish database connection
  con <- RSQLite::dbConnect(RSQLite::SQLite(), file.path(db_directory, db_filename))
  
  # read tables
  Images      <- RSQLite::dbReadTable(con, 'Images')
  Tags        <- RSQLite::dbReadTable(con, 'Tags')
  ImageTags   <- RSQLite::dbReadTable(con, 'ImageTags')
  Albums      <- RSQLite::dbReadTable(con, 'Albums')
  AlbumRoots  <- RSQLite::dbReadTable(con, 'AlbumRoots')
  
#  ImageInformation  <- RSQLite::dbReadTable(con, 'ImageInformation')
#  ImageMetadata     <- RSQLite::dbReadTable(con, 'ImageMetadata')
  
  # disconnect database  
  RSQLite::dbDisconnect(con)
  
  # make output
  return(list(Albums           = Albums,
              AlbumRoots       = AlbumRoots,
              Images           = Images, 
              Tags             = Tags,
              ImageTags        = ImageTags#,
#              ImageInformation = ImageInformation,
#              ImageMetadata    = ImageMetadata
))
}



# extract species tags of videos from digiKam database tables

digiKamVideoHierarchicalSubject <- function(stationDir,
                                    digiKamTablesList,    # output of accessDigiKamDatabase
                                    videoFormat           # character vector of desired video formats
)
{
  
  Albums           <- digiKamTablesList$Albums
  AlbumRoots       <- digiKamTablesList$AlbumRoots
  Images           <- digiKamTablesList$Images
  Tags             <- digiKamTablesList$Tags
  ImageTags        <- digiKamTablesList$ImageTags

  
  # add platform file separator to stationDir
  stationDir0 <- stationDir
  stationDir <- paste0(stationDir, .Platform$file.sep)
  
  # combine album root and album path (match by albumRoot id, not index position)
  Albums <- merge(Albums, AlbumRoots, by.x = "albumRoot", by.y = "id", sort = FALSE)
  Albums$albumPath_full <- paste0(Albums$specificPath, Albums$relativePath)
  
  # add drive letter (only relevant on Windows, and can potentially be wrong if there's Album roots on different drives)
  # also not sure if this works on Mac / Linux due to missing drive letters
  if(.Platform$OS.type == "windows"){
    Albums$albumPath_full2 <- paste(substr(stationDir, 1,2),   # the Drive letter, digiKam doesn't return it
                                    Albums$albumPath_full,
                                    sep = "")
  }
  
  # Linux/Mac solution?
  if(.Platform$OS.type == "unix"){
    Albums$albumPath_full2 <- Albums$albumPath_full#paste(substr(stationDir, 1,2),
                                    #Albums$albumPath_full,
                                    #sep = "")
  }
  
  
  
  # add "/" to ensure Station1 doesn't include Station10 also. Also ensure all folders end with one / only
  Albums$albumPath_full2 <- ifelse(endsWith(Albums$albumPath_full2, .Platform$file.sep),
                                   Albums$albumPath_full2,
                                   paste0(Albums$albumPath_full2, .Platform$file.sep)) 
  
  pathColumn <- "albumPath_full2"
  
  # see if stationDir exists in database
  if(!stationDir %in% Albums[, pathColumn]){
    # stop(paste("station directory", stationDir,  "was not found in digiKam albums. Skipping"), call. = FALSE)
    # try to handle with a warning instead
    warning(paste("station directory", stationDir,  "was not found in digiKam albums. Skipping"), call. = FALSE, immediate. = T)
    # next
    return(NULL)
  }
  
  # find current station in albums
  #album_of_interest <-Albums [which(Albums[, pathColumn] == stationDir),]   # only return the station directory, not camera subdirectories
  album_of_interest <- Albums [grep(pattern = stationDir, Albums[, pathColumn]),]   # This one returns Station directory and camera subdirectories
  if(nrow(album_of_interest) == 0) {
    warning("Could not locate album for ", stationDir, ". Skipping", call. = FALSE)   # NOTE TO SELF: DOESN'T SKIP OR BREAK. CHANGE?
  }
  
  # keep only images in the current album
  image_subset <- Images[Images$album %in% album_of_interest$id,]  # returns matches for all directories. Also, no NAs apparently
  
  # add stationDirectory 
  image_subset$stationDir <- stationDir0
  
  # to do: handle situation where there's no images in image_subset
  
  # NAs are possible, so remove them
  if(any(is.na(image_subset$id))) {
    image_subset <- image_subset[!is.na(image_subset$id),]
  }
  
  # keep only desired video files
  image_subset2 <- image_subset[tolower(substr(image_subset$name, 
                                               nchar(image_subset$name) - 3, 
                                               nchar(image_subset$name))) %in% 
                                  paste(".", videoFormat, sep = ""),]
  
  image_subset_others <- image_subset[!tolower(substr(image_subset$name, 
                                               nchar(image_subset$name) - 3, 
                                               nchar(image_subset$name))) %in% 
                                  paste(".", videoFormat, sep = ""),]
  
  #warning if no videos found
  if(nrow(image_subset2) == 0) {
    warning("Could not find any ", paste(videoFormat, collapse = "/"), " files in ", stationDir, call. = FALSE)
  }
  
  # find "Species" tag group and its children
   
   # subset image tags
   ImageTags <- ImageTags[ImageTags$tagid %in% Tags$id,]
   
   # get proper labels for image tags (and their parent labels = tag group names)
   ImageTags$cleartext_child  <- Tags$name [match(ImageTags$tagid, Tags$id)]
   
   Tags$parent_name <- Tags$name[match(Tags$pid, Tags$id)]
   ImageTags$cleartext_parent <- Tags$parent_name[match(ImageTags$tagid, Tags$id)]    #alternative to above, seems to work (and above seems wrong suddenly)
   
   
   # # # solution by Joel Ruprecht (Google group 2020-07-21) - if a station has no tags with parent ID, doesn't seem to work yet, but should be identical to above solution
   # parentNA <- which(is.na(Tags$name.parent))
   # Tags$name.parent[parentNA] <- Tags$name[parentNA]
   # ImageTags$cleartext_parent_Joel <- Tags$name.parent[match(ImageTags$tagid, Tags$id)]
   
   
   
   # combine parent and child to create HierarchicalTags
   ImageTags$cleartext_full <- paste(ImageTags$cleartext_parent, ImageTags$cleartext_child, sep = "|")
   
   
   # remove unnecessary (internal) tags (not essential)
    remove1 <- grep(Tags$name, pattern = "_Digikam_Internal_Tags_")
    remove2 <- grep(Tags$name, pattern = "Color Label ")
    remove3 <- grep(Tags$name, pattern = "Pick Label ")

    Tags <- Tags[!Tags$id  %in% c(remove1, remove2, remove3),]
    Tags <- Tags[!Tags$pid %in% c(remove1, remove2, remove3),]
   
   ImageTags <- ImageTags[!ImageTags$tagid %in% c(remove1, remove2, remove3),]

   # combine multiple tags for images into single field "HierarchicalSubject"
   ImageTags_aggregate <- aggregate(ImageTags$cleartext_full,
                                    by = list(ImageTags$imageid),
                                    FUN = paste, sep  = "", collapse = ", ")

   # assign column names to output
   colnames(ImageTags_aggregate) <- c("imageid", "HierarchicalSubject")
   
   # assign HierarchicalSubject to matching images
   image_subset2$HierarchicalSubject <- ImageTags_aggregate$HierarchicalSubject[match(image_subset2$id, ImageTags_aggregate$imageid)]
  
   return(image_subset2) 
}

# process the "video" argument of recordTable

processVideoArgument <- function(IDfrom = IDfrom,
                                 video = video){
  
  if(!exists("file_formats",  where = video)) stop("'file_formats' is missing in argument 'video'", call. = FALSE)
  if(!exists("dateTimeTag",   where = video)) stop("'dateTimeTag' is missing in argument 'video'", call. = FALSE)
  
  file_formats <- video$file_formats
  
  if(any(duplicated(file_formats))) stop("There are duplicates in file_formats (in the video argument")
  
  # check file_formats argument
  if(!is.character(file_formats)) stop("'file_formats' in argument 'video' must be of class 'character'", call. = FALSE)
  file_formats <- tolower(file_formats)
  
  # access digiKam database, if required
  if(IDfrom == "metadata"){
    if(!exists("db_directory", where = video)) stop("'db_directory' is missing in argument 'video'", call. = FALSE)
    if(!exists("db_filename",  where = video)) stop("'db_filename' is missing in argument 'video'", call. = FALSE)
    if(!dir.exists(video$db_directory)) stop("directory 'db_directory' does not exist", call. = FALSE)
    if(!file.exists(file.path(video$db_directory, video$db_filename))) stop(paste("file 'db_filename' in directory 'db_directory' does not exist\n",
                                                                                  file.path(video$db_directory, video$db_filename)), call. = FALSE)
    
    if (!requireNamespace("RSQLite", quietly = TRUE)) {
      stop("the package 'RSQLite' is needed for extracting data from digiKam database,  you can install it via: install.packages('RSQLite')")
    } 
   
    # read digiKam database with helper function
    digiKam_data <- accessDigiKamDatabase (db_directory = video$db_directory,
                                           db_filename = video$db_filename)
  } else {    # if IDfrom != "metadata"
    digiKam_data <- NULL
  }
  return(list(digiKam_data = digiKam_data,
              file_formats = file_formats))
}


# if video files extracted, add DateTimeOriginal

addVideoDateTimeOriginal <- function(metadata.tmp,
                                     video){
  # if there's missing entries in DateTimeOriginal that are present in the video date time tag, copy the video tags over
  rows_of_interest1 <- which(metadata.tmp$DateTimeOriginal == "-" & 
                               metadata.tmp[,video$dateTimeTag] != "-")
  if(length(rows_of_interest1) >= 1) {
    metadata.tmp$DateTimeOriginal[rows_of_interest1] <- metadata.tmp[rows_of_interest1, video$dateTimeTag] 
  }
  metadata.tmp[, video$dateTimeTag] <- NULL
  
  return(metadata.tmp) 
}

# add HierachicalSubject for video files
addVideoHierarchicalSubject <- function(metadata.tmp,
                                        video,
                                        digiKamVideoMetadata,
                                        digiKamTablesList,
                                        videoFormat){
  
  if(nrow(digiKamVideoMetadata) == 0) return(metadata.tmp)
  if(is.null(digiKamVideoMetadata))   return(metadata.tmp)
  # add HierarchialSubject for video files (match by filename, must be unique)

  # new version, should match filenames AND paths in digiKamVideoMetadata with metadata.tmp (can deal with duplicate file names in separate folders)
  # get (partial) album path for videos in digiKamVideoMetadata (for station / camera directory)
  # NOTE: no idea how to get the drive letter from volumeid / uuid in AlbumRoots, so it is missing here
  album_name_tmp  <- digiKamTablesList$Albums[match(digiKamVideoMetadata$album, digiKamTablesList$Albums$id), ]
  album_name_tmp2 <- merge(digiKamTablesList$AlbumRoots, album_name_tmp, by.x = "id", by.y = "albumRoot")
  album_name_tmp3 <- paste0(album_name_tmp2$specificPath, album_name_tmp2$relativePath)
  
  # combine partial path and filename of extracted videos 
  digiKamVideoMetadata$almostFullPath <-  file.path(album_name_tmp3, digiKamVideoMetadata$name)
  
  
  # subset metadata.tmp to video files only
  metadata.tmp.video <- metadata.tmp[tolower(substr(metadata.tmp$FileName, 
                                                    nchar(metadata.tmp$FileName) - 3, 
                                                    nchar(metadata.tmp$FileName))) %in% 
                                       paste(".", videoFormat, sep = ""), ]
  
  metadata.tmp.notvideo <- metadata.tmp[!tolower(substr(metadata.tmp$FileName, 
                                                    nchar(metadata.tmp$FileName) - 3, 
                                                    nchar(metadata.tmp$FileName))) %in% 
                                       paste(".", videoFormat, sep = ""), ]
  
  
  metadata.tmp.video$fullPath <- file.path(metadata.tmp.video$Directory, metadata.tmp.video$FileName)
  
  # order metadata.tmp.video.fullPath and metadata.tmp.video.fullPath
  
  metadata.tmp.video <- metadata.tmp.video[order(metadata.tmp.video$fullPath),]
  digiKamVideoMetadata <- digiKamVideoMetadata[order(digiKamVideoMetadata$almostFullPath),]
  
  
  # in metadata.tmp.video, strip drive letter (everything left of first forward slash)
  metadata.tmp.almostFullPath <- substr(metadata.tmp.video$fullPath, 
        start = (nchar(metadata.tmp.video$fullPath) - nchar(digiKamVideoMetadata$almostFullPath) + 1),
        stop = nchar(metadata.tmp.video$fullPath))

  # add metadata from digiKamVideoMetadata to metadata.tmp
  metadata.tmp.video$HierarchicalSubject_video <- digiKamVideoMetadata$HierarchicalSubject [match(metadata.tmp.almostFullPath, 
                                                                                              digiKamVideoMetadata$almostFullPath)]
  
  
  # find rows that have video metadata
  rows_of_interest2 <- which(!is.na(metadata.tmp.video$HierarchicalSubject_video) & 
                               metadata.tmp.video$HierarchicalSubject == "-")
  # write HierarchicalSubject of videos to the normal HierarchicalSubject column
  if(length(rows_of_interest2) >= 1) {
    metadata.tmp.video$HierarchicalSubject[rows_of_interest2] <- metadata.tmp.video$HierarchicalSubject_video[rows_of_interest2] 
  }
  
  
  metadata.tmp.out <- data.frame(rbindlist(list(metadata.tmp.video, metadata.tmp.notvideo), fill = TRUE))
  
  # remove column HierarchicalSubject_video
  metadata.tmp.out$HierarchicalSubject_video <- NULL
  metadata.tmp.out$fullPath <- NULL
  
  return(metadata.tmp.out)
}


# function to return fraction of day that has passed already at a given time  ----
# also does the inverse, fraction of day remaining

fractionOfDay <- function(time, type) {
  # time difference between time and midnight that day (fraction of the day that has passed already)
  delta <- as.numeric(difftime(time, as.Date(time), units = "days"))
  
  # return fraction of day that remains (for setup day)
  if(type == "after") return( 1 - delta)
  # return fraction of day that has passed (for retrieval day)
  if(type == "before")  return(delta)
}


# plot camera operation matrix (function from vignette 3) ----
camopPlot <- function(camOp, 
                      palette = "viridis",
                      lattice = FALSE){
  
  which.tmp <- grep(as.Date(colnames(camOp)), pattern = "01$")
  label.tmp <- format(as.Date(colnames(camOp))[which.tmp], "%Y-%m")
  at.tmp <- which.tmp / ncol(camOp)
  
  values_tmp <- sort(na.omit(unique(c(camOp))))
  
  # hcl.colors only exists in R >3.6.0, use heat.colors for earlier versions
  if(getRversion() >= "3.6.0") {
    image_colors <- grDevices::hcl.colors(n = length(values_tmp), palette = palette, rev = TRUE)
  } else {
    image_colors <- heat.colors(n = length(values_tmp), rev = TRUE)
  }

  
  # transpose and reverse x axis to make sure it plots in the same row/column order as toe original matrix
  if(nrow(camOp) > 1) {
    camop_for_plotting <- t(as.matrix(camOp)[seq(nrow(camOp) ,1),])
  } else {
    camop_for_plotting <- t(as.matrix(camOp))
  }
  
  # lattice::levelplot
  if(isTRUE(lattice)) {
    
    if (!requireNamespace("tibble", quietly = TRUE)) stop("package 'lattice' is required for levelplot (lattice = TRUE)")
    # generate color ramp 
    image_colors_lattice <- grDevices::hcl.colors(n = 100, palette = palette, rev = TRUE)
    
    
    # levelplot in lattice
    # to be improved: y axis labels (station IDs) are all drawn and can overlap. 
    # this may help: https://stat.ethz.ch/R-manual/R-devel/library/lattice/html/axis.default.html
    lattice::levelplot(camop_for_plotting,
              col.regions = image_colors_lattice,
              xlab = "", ylab ="",
              scales = list(x = list(at = which.tmp,
                                     labels = label.tmp, rot = 45)),
              aspect = "fill", 
    )

  } else {
    
    # grDevices::image()    quite basic and no legend
    image(camop_for_plotting, xaxt = "n", yaxt = "n", col = image_colors)
    axis(1, at = at.tmp, labels = label.tmp)
    axis(2, at = seq(from = 0, to = 1, length.out = nrow(camOp)), labels = rev(rownames(camOp)), las = 1)
    abline(v = at.tmp, col = rgb(0,0,0, 0.2))
    box()
  }
}

# intersect intervals, fast (adapted from lubridate) ----
# https://github.com/tidyverse/lubridate/blob/master/R/intervals.r

intersect.Interval.fast <- function(int1, int2, ...) {  # (x, y, ...) {

  starts <- pmax(int1@start, int2@start)
  ends <- pmin(int1@start + int1@.Data, int2@start + int2@.Data)
  spans <- as.numeric(ends) - as.numeric(starts)
  
  spans[spans < 0] <- NA
  spans
}



# split file names (created by imageRename) into their components ----

deparseFilename <- function(x, cameras) {
  x2 <- strsplit(x, split = "__")
  
  if(isTRUE(cameras)) {
    out <- data.frame(Station = sapply(x2, FUN = function(x) x[[1]]),
                      Camera = sapply(x2, FUN = function(x) x[[2]]),
                      Date = as.Date(sapply(x2, FUN = function(x) x[[3]])),
                      Time = sapply(x2, FUN = function(x) x[[4]]))
  }
  
  if(isFALSE(cameras)) {
    out <- data.frame(Station = sapply(x2, FUN = function(x) x[[1]]),
                      Camera = NA,
                      Date = as.Date(sapply(x2, FUN = function(x) x[[2]])),
                      Time = sapply(x2, FUN = function(x) x[[3]]))
  }
  
  time_split <- strsplit(out$Time, split = "(", fixed = T)
  out$Time <- sapply(time_split, FUN = function(y) y[1])
  
  
  
  # out$DateTimeOriginal <- ymd_hms(paste(out$Date, out$Time))
  out$DateTimeOriginal <- paste(out$Date, out$Time)
  
  out$number <- as.numeric(sapply(strsplit(sapply(time_split, FUN = function(y) y[2]), 
                                           split = ")", fixed = T), 
                                  FUN = function(y) y[1]))
  
  out
}


# Probe internet connection  ----

# check if connection to Google's DNS server can be established
# caveat: specific IP/port which could theoretically be blocked even if internet is otherwise fine
# 8.8.8.8:53 should be reliable though

has_internet_socket <- function() {
  is_connected <- FALSE
  # Use a known public DNS server or a reliable website's IP
  # Google DNS: 8.8.8.8, Port 53 (DNS) or 80 (HTTP)
  host <- "8.8.8.8"
  port <- 53 # DNS port is good for basic connectivity check
  timeout_seconds <- 3 # Prevent hanging
  
  suppressWarnings({
    tryCatch({
      con <- socketConnection(host = host, port = port, blocking = FALSE, open = "w+", timeout = timeout_seconds)
      if (isOpen(con)) {
        is_connected <- TRUE
        close(con)
      }
    }, error = function(e) {
      is_connected <- FALSE
    })
  })
  return(is_connected)
}


# Check total / available system memory ----

.get_system_memory <- function() {
  out <- list(total = NA, available = NA)
  
  try({
    if (.Platform$OS.type == "windows") {
      # Use a direct PowerShell call to get raw numbers only
      cmd <- "powershell -command \"$m = Get-CimInstance Win32_OperatingSystem; $m.TotalVisibleMemorySize; $m.FreePhysicalMemory\""
      res <- system(cmd, intern = TRUE, ignore.stderr = TRUE)
      res <- res[nzchar(gsub("[^0-9]", "", res))] # Remove non-numeric lines
      if (length(res) >= 2) {
        out$total <- as.numeric(res[1]) / 1e6
        out$available <- as.numeric(res[2]) / 1e6
      }
      
    } else if (Sys.info()["sysname"] == "Darwin") {
      # macOS: Use absolute path for sysctl
      sysctl_path <- "/usr/sbin/sysctl"
      if (file.exists(sysctl_path)) {
        res <- system2(sysctl_path, "hw.memsize", stdout = TRUE, stderr = FALSE)
        # Extract digits only, ignore labels
        digits <- gsub("[^0-9]", "", res)
        if (nzchar(digits)) out$total <- as.numeric(digits) / 1e9
        # available is difficult to get in script on Mac
      }
      
    } else if (Sys.info()["sysname"] == "Linux") {
      if (file.exists("/proc/meminfo")) {
        lines <- readLines("/proc/meminfo")
        parse_kb <- function(tag) {
          line <- grep(tag, lines, value = TRUE)
          if (length(line) > 0) return(as.numeric(gsub("[^0-9]", "", line)) / 1e6)
          return(NA)
        }
        out$total <- parse_kb("MemTotal")
        out$available <- parse_kb("MemAvailable")
        if (is.na(out$available)) out$available <- parse_kb("MemFree")
      }
    }
  }, silent = TRUE)
  
  return(out)
}



#' validate memory before allocation
#' @param ram_usage_estimate Estimated size of the object in GB
#' 
.check_memory_usage <- function(ram_usage_estimate) {
  
  mem <- .get_system_memory()
  
  # Only proceed if we actually successfully retrieved memory info
  if (!is.na(mem$total)) {
    
    # A: HARD ERROR - Physical hardware limit
    if (ram_usage_estimate > mem$total) {
      stop(sprintf(
        "Cannot allocate array. Predicted size (%.2f GB) exceeds total system RAM (%.2f GB).", 
        ram_usage_estimate, mem$total), call. = FALSE)
    }
    
    # B: AVAILABLE MEMORY CHECKS
    if (!is.na(mem$available)) {
      # ERROR: If it exceeds what's currently free
      if (ram_usage_estimate > mem$available) {
        stop(sprintf(
          "Insufficient memory: Predicted array requires %.2f GB, but only %.2f GB is currently available. Free up memory, reduce dimensions, or use 'batch = TRUE'.", 
          ram_usage_estimate, mem$available), call. = FALSE)
        
        # WARNING: If it uses more than 80% of what's left
      } else if (ram_usage_estimate > (mem$available * 0.8)) {
        warning(sprintf(
          "High memory pressure: Predicted array (%.2f GB) will use %.0f%% of available RAM. This may slow down your system.", 
          ram_usage_estimate, (ram_usage_estimate / mem$available) * 100), call. = FALSE)
      }
      
    } else {
      # C: FALLBACK (When Available is unknown, e.g., macOS)
      # Stop if > 90% total, warn if > 70% total
      if (ram_usage_estimate > (mem$total * 0.9)) {
        stop(sprintf(
          "Predicted array (%.2f GB) exceeds safe threshold (90%%) of total RAM (%.2f GB).", 
          ram_usage_estimate, mem$total), call. = FALSE)
      } else if (ram_usage_estimate > (mem$total * 0.7)) {
        warning(sprintf(
          "High memory usage: Predicted array (%.2f GB) uses a large portion of total RAM.", 
          ram_usage_estimate), call. = FALSE)
      }
    }
  }
  
  return(invisible(NULL))
}

context("surveyReport")

# --- Test Setup ---
# Create shared data and camera operation matrices once.
data(camtraps)
data(recordTableSample)

# Camera operation matrix WITHOUT problems
camop_no_problem <- cameraOperation(
  CTtable      = camtraps,
  stationCol   = "Station",
  setupCol     = "Setup_date",
  retrievalCol = "Retrieval_date",
  hasProblems  = FALSE,
  dateFormat   = "dmy"
)

# Camera operation matrix WITH problems
camop_problem <- cameraOperation(
  CTtable      = camtraps,
  stationCol   = "Station",
  setupCol     = "Setup_date",
  retrievalCol = "Retrieval_date",
  hasProblems  = TRUE,
  dateFormat   = "dmy"
)

# camtrapR:::camopPlot(camop_problem)

# --- Test Suite ---

testthat::describe("Core Functionality and Return Value", {
  
  test_that("it returns a list of 5 data frames with correct structure", {
    report <- surveyReport(
      recordTable  = recordTableSample,
      CTtable      = camtraps,
      camOp        = camop_no_problem,
      setupCol     = "Setup_date",
      retrievalCol = "Retrieval_date",
      CTDateFormat = "dmy"
    )
    
    # 1. Check the main output structure
    expect_true(class(report) == "list")
    expect_length(report, 5)
    expected_names <- c("survey_dates", "species_by_station", "events_by_species",
                        "events_by_station", "events_by_station2")
    expect_named(report, expected_names)
    
    # 2. Check that each element is a data frame
    expect_s3_class(report$survey_dates, "data.frame")
    expect_s3_class(report$species_by_station, "data.frame")
    expect_s3_class(report$events_by_species, "data.frame")
    expect_s3_class(report$events_by_station, "data.frame")
    expect_s3_class(report$events_by_station2, "data.frame")
  })
  
  test_that("calculations are correct for a survey without problems", {
    report <- surveyReport(
      recordTable  = recordTableSample,
      CTtable      = camtraps,
      setupCol     = "Setup_date",
      retrievalCol = "Retrieval_date",
      camOp        = camop_no_problem,
      CTDateFormat = "dmy"
    )
    
    # Check key calculations in the survey_dates table
    survey_dates <- report$survey_dates
    stationA_dates <- survey_dates[survey_dates$Station == "StationA", ]
    
    # Setup: 2009-04-02, Retrieval: 2009-05-14. Total duration: 43 calendar days
    expect_equal(stationA_dates$n_calendar_days_total, 43)
    # No problems, so active days should be the same
    expect_equal(stationA_dates$n_calendar_days_active, 43)
    expect_equal(stationA_dates$n_calendar_days_inactive, 0)
    # Trap nights = active days * number of cameras (1)
    expect_equal(stationA_dates$n_trap_nights_active, 42)
    
    # Check events_by_species
    events_species <- report$events_by_species
    pbe_events <- events_species[events_species$species == "PBE", ]
    expect_equal(as.numeric(pbe_events$n_events), 18) # 18 records for PBE
    expect_equal(as.numeric(pbe_events$n_stations), 3) # Detected at all three stations
  })
  
  test_that("calculations are correct for a survey WITH problems", {
    report_problem <- surveyReport(
      recordTable  = recordTableSample,
      CTtable      = camtraps,
      setupCol     = "Setup_date",
      retrievalCol = "Retrieval_date",
      camOp        = camop_problem,
      CTDateFormat = "dmy"
    )
    
    survey_dates_prob <- report_problem$survey_dates
    stationC_dates_prob <- survey_dates_prob[survey_dates_prob$Station == "StationC", ]
    
    # Station1 had a problem from May 12 to May 17 (6 days)
    # Total days = 44
    # Inactive days = 6
    # Active days = 38
    expect_equal(stationC_dates_prob$n_calendar_days_total, 44)
    expect_equal(stationC_dates_prob$n_calendar_days_active, 38)
    expect_equal(stationC_dates_prob$n_calendar_days_inactive, 6)
    expect_equal(stationC_dates_prob$n_trap_nights_active, 37.5)
  })
  
})


testthat::describe("File Writing and Side Effects", {
  
  tmpdir <- tempdir()
  
    test_that("sinkpath argument creates a text report", {
      report <- surveyReport(
        recordTable  = recordTableSample,
        CTtable      = camtraps,
        setupCol     = "Setup_date",
        retrievalCol = "Retrieval_date",
        camOp        = camop_no_problem,
        CTDateFormat = "dmy",
        sinkpath     = tmpdir #"." # Write to the temp directory
      )
      
      # Check for the report file
      report_files <- list.files(tmpdir, pattern = paste0("survey_report_", Sys.Date(), ".txt"), full.names = T)
      expect_equal(length(report_files), 1)
      
      
      # Check that the file has content
      report_content <- readLines(report_files[1])
      expect_gt(length(report_content), 10)
      expect_true(any(grepl("Total number of stations", report_content)))
    })
    
    
    
    test_that("makezip = TRUE creates a valid zip file with expected contents", {

      testthat::skip_if_not_installed("overlap")

      report <- surveyReport(
        recordTable  = recordTableSample,
        CTtable      = camtraps,
        setupCol     = "Setup_date",
        retrievalCol = "Retrieval_date",
        camOp        = camop_no_problem,
        CTDateFormat = "dmy",
        Xcol         = "utm_x",
        Ycol         = "utm_y",
        sinkpath     = tmpdir, #".",
        makezip      = TRUE
      )


    # 1. Check if the zip file was created
    zip_files <- list.files(tmpdir, pattern = paste0("surveyReport_", Sys.Date(), ".zip"), full.names = T)
    expect_equal(length(zip_files), 1)

    # 2. Check the contents of the zip file
    zip_contents <- utils::unzip(zip_files[1], list = TRUE)

    # Expected files/folders
    expect_true("camtrapR_scripts.R" %in% basename(zip_contents$Name))

    # Check for specific files
    expect_true("recordTable.csv" %in% basename(zip_contents$Name))
    n_species <- length(unique(recordTableSample$Species))
    expect_equal(sum(grepl("activity_", zip_contents$Name)), n_species)
    expect_equal(sum(grepl("Presence", zip_contents$Name)), n_species)
  })
})



testthat::describe("Date/time column checks", {
  
  camtraps_setup_blank    <- camtraps
  camtraps_retrieval_blank <- camtraps
  camtraps_setup_blank$Setup_date[1] <- ""
  camtraps_retrieval_blank$Retrieval_date[1] <- ""
  
  recordTableSample_datetime_blank <- recordTableSample
  recordTableSample_datetime_blank$DateTimeOriginal[1] <- ""
  
  camtraps_setup_NA <- camtraps_setup_blank
  camtraps_retrieval_NA <- camtraps_retrieval_blank
  recordTableSample_datetime_NA <- recordTableSample_datetime_blank
  
  camtraps_setup_NA$Setup_date[1] <- NA
  camtraps_retrieval_NA$Retrieval_date[1] <- NA
  recordTableSample_datetime_NA$DateTimeOriginal[1] <- NA
  
  # camera operation matrices
  camop_no_problem <- cameraOperation(CTtable      = camtraps,
                                      stationCol   = "Station",
                                      setupCol     = "Setup_date",
                                      retrievalCol = "Retrieval_date",
                                      writecsv     = FALSE,
                                      hasProblems  = FALSE,
                                      dateFormat   = "%d/%m/%Y"
  )
  
  # with problems/malfunction
  camop_problem <- cameraOperation(CTtable      = camtraps,
                                   stationCol   = "Station",
                                   setupCol     = "Setup_date",
                                   retrievalCol = "Retrieval_date",
                                   writecsv     = FALSE,
                                   hasProblems  = TRUE,
                                   dateFormat   = "%d/%m/%Y"
  )
  
  test_that("errors about blank values in Date/Time are correct", {
    expect_error(surveyReport (recordTable          = recordTableSample,
                               CTtable              = camtraps_setup_blank,
                               camOp                = camop_no_problem,
                               speciesCol           = "Species",
                               stationCol           = "Station",
                               setupCol             = "Setup_date",
                               retrievalCol         = "Retrieval_date",
                               CTDateFormat         = "%d/%m/%Y"), 
                 "there are blank values in CTtable[, setupCol]", fixed = TRUE)
    expect_error(surveyReport (recordTable          = recordTableSample,
                               CTtable              = camtraps_retrieval_blank,
                               camOp                = camop_no_problem,
                               speciesCol           = "Species",
                               stationCol           = "Station",
                               setupCol             = "Setup_date",
                               retrievalCol         = "Retrieval_date",
                               CTDateFormat         = "%d/%m/%Y"), 
                 "there are blank values in CTtable[, retrievalCol]", fixed = TRUE)
    expect_error(surveyReport (recordTable          = recordTableSample_datetime_blank,
                               CTtable              = camtraps,
                               camOp                = camop_no_problem,
                               speciesCol           = "Species",
                               stationCol           = "Station",
                               setupCol             = "Setup_date",
                               retrievalCol         = "Retrieval_date",
                               CTDateFormat         = "%d/%m/%Y"), 
                 "there are blank values in recordTable[, recordDateTimeCol]", fixed = TRUE)
  })
  
  
  
  test_that("errors about NAs in Date/Time are correct", {
    expect_error(surveyReport (recordTable          = recordTableSample,
                               CTtable              = camtraps_setup_NA,
                               camOp                = camop_no_problem,
                               speciesCol           = "Species",
                               stationCol           = "Station",
                               setupCol             = "Setup_date",
                               retrievalCol         = "Retrieval_date",
                               CTDateFormat         = "%d/%m/%Y"), 
                 "there are NAs in CTtable[, setupCol]", fixed = TRUE)
    expect_error(surveyReport (recordTable          = recordTableSample,
                               CTtable              = camtraps_retrieval_NA,
                               camOp                = camop_no_problem,
                               speciesCol           = "Species",
                               stationCol           = "Station",
                               setupCol             = "Setup_date",
                               retrievalCol         = "Retrieval_date",
                               CTDateFormat         = "%d/%m/%Y"), 
                 "there are NAs in CTtable[, retrievalCol]", fixed = TRUE)
    expect_error(surveyReport (recordTable          = recordTableSample_datetime_NA,
                               CTtable              = camtraps,
                               camOp                = camop_no_problem,
                               speciesCol           = "Species",
                               stationCol           = "Station",
                               setupCol             = "Setup_date",
                               retrievalCol         = "Retrieval_date",
                               CTDateFormat         = "%d/%m/%Y"), 
                 "there are NAs in recordTable[, recordDateTimeCol]", fixed = TRUE)
  })
  
  
  
  test_that("errors about all NAs in Date/Time are correct", {
    expect_error(surveyReport (recordTable          = recordTableSample,
                               CTtable              = camtraps,
                               camOp                = camop_no_problem,
                               speciesCol           = "Species",
                               stationCol           = "Station",
                               setupCol             = "Setup_date",
                               retrievalCol         = "Retrieval_date",
                               CTDateFormat         = "%d-%m-%Y"),            # this is intentionally wrong
                 "Cannot read date format in CTtable[, setupCol]. Output is all NA.
expected:  %d-%m-%Y
actual:    02/04/2009", fixed = TRUE)
    
    expect_error(surveyReport (recordTable          = recordTableSample,
                               CTtable              = camtraps,
                               camOp                = camop_no_problem,
                               speciesCol           = "Species",
                               stationCol           = "Station",
                               setupCol             = "Setup_date",
                               retrievalCol         = "Retrieval_date",
                               CTDateFormat         = "%d/%m/%Y",
                               recordDateTimeFormat = "%Y/%m/%d %H:%M:%S"),      # this is intentionally wrong
                 "Cannot read datetime format in recordTable[, recordDateTimeCol]. Output is all NA.
expected:  %Y/%m/%d %H:%M:%S
actual:    2009-04-21 00:40:00", fixed = TRUE)
    
  })
})

#' Write values to DateTimeOriginal tag in image metadata
#' 
#' 
#' This function assigns values to the DateTimeOriginal tag in image's EXIF
#' metadata using Exiftool. It can be used when the original DateTimeOriginal
#' values in the metadata were lost for whatever reason. In order to first read
#' the Date/Time values from the data fields in the images, see the function
#' \code{\link{OCRdataFields}}. After running \code{OCRdataFields} and checking
#' its output you can run \code{writeDateTimeOriginal}.
#' 
#' 
#' The first value in \code{DateTimeOriginal} will be assigned to the first
#' image in \code{fileNames}, and so on. Both \code{DateTimeOriginal} and
#' \code{fileNames} can be obtained from the output of
#' \code{\link{OCRdataFields}}.  \code{DateTimeOriginal} uses the standard
#' "YYYY-MM-SS HH:MM:SS"" notation. If the values extracted via
#' \code{\link{OCRdataFields}} are in a different format you'll need to
#' reformat them first. Please provide them as character. Also, before using
#' this function, make sure that the date/time values read by
#' \code{\link{OCRdataFields}} are correct (sometimes OCR misreads values, so
#' check carefully).
#' 
#' Parallel processing is advised since the function is rather slow (due to
#' calling Exiftool separately on every, so about 1 second per image). If you
#' know how to batch-assign DateTimeOriginal values in one Exiftool call,
#' please let me know.
#' 
#' The function only works on JPG images, not video files.
#' 
#' @param DateTimeOriginal character. DateTimeOriginal values to write into
#' EXIF:DateTimeOriginal tag of the files in fileNames.
#' @param fileNames character. Full file names (including directories) of
#' images to process.
#' @param parallel A parallel cluster object. Specify if you wish to run this
#' function in parallel. Provide a cluster object (output of makeCluster()) -
#' optional.
#' @param overwrite logical. Overwrite existing files (TRUE) or create new
#' files while saving the original data as jpg_original files as a backup
#' (FALSE)?
#' 
#' @return Invisible NULL. The actual output is the JPG images which now have a
#' DateTimeOriginal tag.
#' 
#' @author Juergen Niedballa
#' 
#' @seealso \code{\link{OCRdataFields}}
#' 
#' @examples
#' 
#' 
#' \dontrun{
#' # dontrun is to avoid forcing users to install additional dependencies
#' 
#' wd_images_OCR <- system.file("pictures/full_size_for_ocr", package = "camtrapR")
#' 
#' library(magick)
#' 
#' # define geometries
#' geometry1 <- geometry_area(x_off = 0, y_off = 0, width = 183, height = 37)
#' geometry2 <- geometry_area(x_off = 196, y_off = 0, width = 200, height = 17)
#' geometry3 <- geometry_area(x_off = 447, y_off = 0, width = 63, height = 17)
#' geometry4 <- geometry_area(x_off = 984, y_off = 0, width = 47, height = 17)
#' geometry5 <- geometry_area(x_off = 0, y_off = 793, width = 320, height = 17)
#' 
#' # combine geometries into list
#' geometries <- list(date = geometry1, 
#'                    time = geometry2, 
#'                    sequence_id = geometry3,
#'                    temperature = geometry4,
#'                    camera_model = geometry5)
#' 
#' df_image_data  <- OCRdataFields(inDir = wd_images_OCR,
#'                                 geometries = geometries, 
#'                                 invert = TRUE)       
#' df_image_data
#' 
#' 
#' 
#' library(parallel)
#' library(lubridate)
#' 
#' # prepare DateTimeOriginal column (ymd_hms() automatically respects the PM indicator)
#' df_image_data$DateTimeOriginal <- paste(df_image_data$date, df_image_data$time)
#' df_image_data$DateTimeOriginal <- as.character(ymd_hms(df_image_data$DateTimeOriginal))
#' 
#' # create cluster (3 cores)
#' cl <- makeCluster(3)
#' 
#' # assign new DateTimeOriginal
#' writeDateTimeOriginal(DateTimeOriginal = df_image_data$DateTimeOriginal,
#'                       fileNames = df_image_data$filename_full,
#'                       parallel = cl)
#'                       
#' stopCluster(cl)
#'                    
#' }
#' 
#' 
#' 
#' @export writeDateTimeOriginal
#' 
writeDateTimeOriginal <- function(DateTimeOriginal, 
                                  fileNames,
                                  parallel,
                                  overwrite = FALSE) {
  
  stopifnot(is.character(DateTimeOriginal))
  stopifnot(is.character(fileNames))
  
  if(length(DateTimeOriginal) != length(fileNames)) stop("DateTimeOriginal and fileNames must have same length")
  
  exif_call <- paste0('exiftool -datetimeoriginal="', DateTimeOriginal, '" ', 
                      ifelse(overwrite, "-overwrite_original ", " "), fileNames)
  
  
  if(hasArg(parallel)){
    if (!requireNamespace("parallel", quietly = TRUE)) {
      stop("Please install the package parallel to run this function in parallel")
    }
    if(!is(parallel, "cluster")) stop("parallel must be a cluster object, i.e. output of parallel::makeCluster()", call. = FALSE)
    
    out <- parallel::parLapply(parallel, exif_call, system)
    
  } else {
    out <- lapply(exif_call, system)
  }
  return(invisible(NULL))
}

#' Create a camera trap station operation matrix
#' 
#' Construct a matrix of daily camera trap station operation for use in
#' \code{\link{detectionHistory}} and \code{\link{spatialDetectionHistory}},
#' where it is needed for calculating trapping effort per occasion. It is also
#' used in \code{\link{surveyReport}} to calculate the number of trap nights
#' durig a survey. If several cameras were deployed per station, the matrix can
#' contain camera- or station-specific trap operation information, or
#' information about sessions during repeated surveys.
#' 
#' \code{cameraCol} is NULL by default, meaning the function assumes there was
#' 1 camera per station in \code{CTtable}. If more than 1 camera was deployed
#' per station, \code{cameraCol} needs to be specified to identify individual
#' cameras within a station. Likewise, \code{sessionCol} can be used to if
#' camera trap stations were operated during multiple sessions / trapping
#' seasons.
#' 
#' \code{dateFormat} defaults to "YYYY-MM-DD", e.g. "2014-10-31", but can be
#' any other date format or date-time also. It can be specified either in the
#' format required by \code{\link[base]{strptime}} or the 'orders' argument in
#' \code{\link[lubridate]{parse_date_time}} in \pkg{lubridate}. In the example
#' above, "YYYY-MM-DD" would be specified as "\%Y-\%m-\%d" in base R or "ymd"
#' in \pkg{lubridate}.
#' 
#' Since version 2.1, dateFormat can be a date-time. That makes it possible to
#' specify the exact time cameras were set up / retrieved / malfunctioned /
#' worked again. This information is used to calculate the daily trapping
#' effort more precisely on days with incomplete effort.
#' 
#' Previously, setup and retrival day were counted as 1, indicating a whole day
#' of effort on those days. Since version 2.1, setup and retrieval are assumed
#' to have happened at 12 noon (resulting in daily effort of 0.5 instead of 1).
#' Users can also specify the exact time cameras were set up (by providing a
#' date-time in the setup / retrieval / problem columns). See vignette 3 for
#' more details.
#' 
#' If \code{hasProblems} is TRUE, the function tries to find columns
#' \code{ProblemX_from} and \code{ProblemX_to} in \code{CTtable}. \code{X} is a
#' consecutive number from 1 to n, specifying periods in which a camera or
#' station was not operational. If \code{hasProblems} is FALSE, cameras are
#' assumed to have been operational uninterruptedly from setup to retrieval
#' (see \code{\link{camtraps}} for details).
#' 
#' \code{allCamsOn} only has an effect if there was more than 1 camera at a
#' station. If TRUE, for the station to be considered operational, all cameras
#' at a station need to be operational. If FALSE, at least 1 active camera
#' renders the station operational. Argument \code{camerasIndependent} defines
#' if cameras record animals independently (it thus only has an effect if there
#' was more than 1 camera at a station). This is the case if an observation at
#' one camera does not increase the probability for detection at another camera
#' (cameras face different trails at a distance of one another).
#' Non-independence occurs if an animal is likely to trigger both camers (as
#' would be the case with 2 cameras facing each other).
#' 
#' If \code{camerasIndependent} is TRUE, 2 active cameras at a station will
#' result in a station operation value of 2 in the resulting matrix, i.e., 2
#' independent trap days at 1 station and day. If \code{camerasIndependent} is
#' FALSE, 2 active cameras will return value 1, i.e., 1 trap night at 1 station
#' per day.
#' 
#' Row names depend on the input arguments and contain the station name and
#' potentially session and camera names (if \code{sessionCol} and/or
#' \code{cameraCol} are defined).
#' 
#' Naming convention is (since version 1.2) \bold{Bold} information are from
#' the columns \code{stationCol}, \code{sessionCol} and \code{cameraCol} in
#' \code{CTtable}:
#' 
#' \bold{Station} \cr \bold{Station}__SESS_\bold{SessionID} \cr
#' \bold{Station}__CAM_\bold{CameraID} \cr
#' \bold{Station}__SESS_\bold{SessionID}__CAM_\bold{CameraID}
#' 
#' Session are designated with prefix "__SESS_", cameras with prefix "__CAM_".
#' Therefore, these are reserved words and may not be part of station, session
#' or camera names. Here's what it may look like in real life:
#' 
#' \bold{Station1} \cr \bold{Station1}__SESS_\bold{2019} \cr
#' \bold{Station1}__CAM_\bold{1024152} \cr
#' \bold{Station1}__SESS_\bold{2019}__CAM_\bold{1024152}
#' 
#' Functions detectionHistory and spatialDetectionHistory recognize these and
#' use the information accordingly.
#' 
#' @param CTtable data.frame containing information about location and trapping
#' period of camera trap stations
#' @param stationCol character. name of the column specifying Station ID in
#' \code{CTtable}
#' @param cameraCol character. name of the column specifying Camera ID in
#' \code{CTtable} (optional). If empty, 1 camera per station is assumed.
#' @param sessionCol character. name of the column specifying session ID in
#' \code{CTtable} (optional). Use it for creating multi-session / multi-season
#' detection histories (unmarked: \code{\link[unmarked]{unmarkedMultFrame}};
#' secr: \code{\link[secr]{capthist}} )
#' @param setupCol character. name of the column containing camera setup dates
#' in \code{CTtable}
#' @param retrievalCol character. name of the column containing camera
#' retrieval dates in \code{CTtable}
#' @param hasProblems logical. If TRUE, function will look for columns
#' specifying malfunction periods in \code{CTtable} (naming convention:
#' \code{ProblemX_from} and \code{ProblemX_to}, where X is a number)
#' @param byCamera logical. If TRUE, camera operation matrix is computed by
#' camera, not by station (requires \code{cameraCol})
#' @param allCamsOn logical. Takes effect only if \code{cameraCol} is defined
#' and if \code{byCamera} is FALSE. If \code{allCamsOn = TRUE}, all cameras at
#' a station need to be operational for the station to be operational (e.g. 1
#' camera out of 2 malfunctioning renders the station inoperational). Output
#' values can be 1/0/NA only (all cameras at a station operational/ at least 1
#' camera not operational/ no camera set up). If \code{allCamsOn = FALSE}, at
#' least 1 active camera makes a station operational.
#' @param camerasIndependent logical. Return number of active camera traps by
#' station? Only if \code{byCamera} is FALSE and \code{allCamsOn} is FALSE. If
#' \code{camerasIndependent} is TRUE, output values will be the number of
#' operational cameras at a station. If \code{camerasIndependent} is FALSE, the
#' value is 1 if at least 1 camera was operational, otherwise 0. In both cases,
#' values are NA if no camera was set up.
#' @param dateFormat character. The format of columns \code{setupCol} and
#' \code{retrievalCol} (and potential problem columns) in \code{CTtable}. Must
#' be interpretable by either \code{as.Date} or the "orders" argument
#' \code{\link[lubridate]{parse_date_time}} in \pkg{lubridate}. Can be a date
#' or (since version 2.1) a date-time.
#' @param occasionStartTime integer. time of day (the full hour) at which to
#' begin occasions. Replaces \code{occasionStartTime} from
#' \code{\link{detectionHistory}} and \code{\link{spatialDetectionHistory}}.
#' @param writecsv logical. Should the camera operation matrix be saved as a
#' .csv?
#' @param outDir character. Directory into which csv is saved
#' 
#' @return A matrix. Row names always indicate Station IDs. If
#' \code{sessionCol} and/or \code{cameraCol} are defined, they are contained in
#' the row names also (camera ID only if \code{byCamera = TRUE}). Column names
#' are dates. \cr Legend: NA: camera(s) not set up, 0: camera(s) not
#' operational, 1 (or higher): number of operational camera(s) or an indicator
#' for whether the station was operational (depending on
#' \code{camerasIndependent} and \code{allCamsOn})
#' 
#' @note Setting \code{camerasIndependent} according to the sampling situation
#' is important for the functions \code{\link{detectionHistory}} and
#' \code{\link{spatialDetectionHistory}}, if sampling effort (the number of
#' active trap nights in a occasion) is to be computed and returned.
#' 
#' @author Juergen Niedballa
#' 
#' @examples
#' 
#' 
#' data(camtraps)
#' 
#' # no problems/malfunction
#' camop_no_problem <- cameraOperation(CTtable      = camtraps,
#'                                     stationCol   = "Station",
#'                                     setupCol     = "Setup_date",
#'                                     retrievalCol = "Retrieval_date",
#'                                     writecsv     = FALSE,
#'                                     hasProblems  = FALSE,
#'                                     dateFormat   = "dmy"
#' )
#' 
#' # with problems/malfunction
#' camop_problem <- cameraOperation(CTtable      = camtraps,
#'                                  stationCol   = "Station",
#'                                  setupCol     = "Setup_date",
#'                                  retrievalCol = "Retrieval_date",
#'                                  writecsv     = FALSE,
#'                                  hasProblems  = TRUE,
#'                                  dateFormat   = "dmy"
#' )
#' 
#' # The examples above specified dateFormat using lubdridate package, which is more intuitive. 
#' # Alternatively one can used in base-R date conversions (strptime) as below:
#' 
#' # with problems/malfunction / dateFormat in strptime format
#' camop_problem_oldformat <- cameraOperation(CTtable      = camtraps,
#'                                            stationCol   = "Station",
#'                                            setupCol     = "Setup_date",
#'                                            retrievalCol = "Retrieval_date",
#'                                            writecsv     = FALSE,
#'                                            hasProblems  = TRUE,
#'                                            dateFormat   = "%d/%m/%Y"
#' )
#' 
#' camop_no_problem
#' camop_problem
#' camop_problem_oldformat
#' 
#' @importFrom data.table rbindlist setDF setDT setkey foverlaps ":="
#' @importFrom lubridate as_date as_datetime ddays dhours dseconds interval int_start int_end int_overlaps time_length "%within%" 
#' @importFrom methods hasArg is new
#' @importFrom stats aggregate na.omit start end rnorm window quantile
#' @importFrom utils capture.output modifyList write.csv zip head menu read.table
#' @export cameraOperation
#' 
cameraOperation <- function(CTtable,
                            stationCol = "Station",
                            cameraCol,
                            sessionCol,
                            setupCol,
                            retrievalCol,
                            hasProblems = FALSE,
                            byCamera,
                            allCamsOn,
                            camerasIndependent,
                            dateFormat = "ymd",
                            occasionStartTime = 0,
                            writecsv = FALSE,
                            outDir){
  
  
  # check and prepare input
  wd0 <- getwd()
  on.exit(setwd(wd0))
  
  checkForSpacesInColumnNames(stationCol = stationCol, setupCol = setupCol, retrievalCol = retrievalCol)
  
  CTtable <- dataFrameTibbleCheck(df = CTtable)
  
  if(!stationCol   %in% colnames(CTtable))  stop(paste('stationCol = "',   stationCol,     '" is not a column name in CTtable', sep = ''), call. = FALSE)
  if(!setupCol     %in% colnames(CTtable))  stop(paste('setupCol = "',     setupCol,       '" is not a column name in CTtable', sep = ''), call. = FALSE)
  if(!retrievalCol %in% colnames(CTtable))  stop(paste('retrievalCol = "', retrievalCol,   '" is not a column name in CTtable', sep = ''), call. = FALSE)
  
  
  # flag for whether to use fraction of days (if hours are provided in dateFormat).
  # If not setup and retrieval are assumed to be at 12 noon
  effortAsFraction <- grepl("H", dateFormat)
  
  stopifnot(length(stationCol) == 1)
  CTtable[,stationCol] <- as.character(CTtable[,stationCol])
  
  stopifnot(length(setupCol) == 1)
  stopifnot(length(retrievalCol) == 1)
  
  # if(effortAsFraction) {
  #   CTtable[,setupCol] <- format(CTtable[, setupCol], format = "%Y-%m-%d %H:%M:%S")
  #   CTtable[,retrievalCol] <- format(CTtable[, retrievalCol], format = "%Y-%m-%d %H:%M:%S")
  #   dateFormat <- "ymd HMS"
  # } else {
  #   CTtable[,setupCol] <- as.character(CTtable[,setupCol])   # parse_date_time gives error if in Date format
  #   CTtable[,retrievalCol] <- as.character(CTtable[,retrievalCol])
  # }
  

  stopifnot(is.logical(writecsv))
  stopifnot(is.logical(hasProblems))
  
  if(hasArg(byCamera)) {
    stopifnot(is.logical(byCamera))
    if(isTRUE(byCamera) & hasArg(cameraCol) == FALSE){
      stop("if 'byCamera' is TRUE, 'cameraCol' needs to be specified")
    }
  } 
  
  if(length(occasionStartTime) != 1) stop("occasionStartTime must have length 1")
  occasionStartTime    <- as.integer(round(occasionStartTime))
  if(occasionStartTime != 0 & !is.integer(occasionStartTime)) stop ("occasionStartTime must be an integer between 0 and 23", call. = FALSE)
  if(occasionStartTime < 0 | occasionStartTime >= 24)         stop ("occasionStartTime must be between 0 and 23", call. = FALSE)
  
  myArgs <- match.call()
  cameraColInArgs  <- "cameraCol" %in% names(myArgs)
  sessionColInArgs <- "sessionCol" %in% names(myArgs)
  
  
  if(cameraColInArgs){
    checkForSpacesInColumnNames(cameraCol = cameraCol)
    if(!cameraCol %in% colnames(CTtable)) stop(paste('cameraCol = "', cameraCol, '" is not a column name in CTtable', sep = ''), call. = FALSE)
    if(!hasArg(byCamera)) stop("if cameraCol is set, byCamera must be specified")
    stopifnot(is.logical(byCamera))
    CTtable[,cameraCol] <- as.character(CTtable[,cameraCol])
    
    if(isFALSE(byCamera)){
      if(!hasArg(allCamsOn)) stop("if cameraCol is set and byCamera is FALSE, allCamsOn must be specified")
      stopifnot(is.logical(allCamsOn))
      if(!allCamsOn){
        if(!hasArg(camerasIndependent)) stop("if cameraCol is set, byCamera is FALSE and allCamsOn is FALSE, camerasIndependent must be specified")
        stopifnot(is.logical(camerasIndependent))
      }
    } else {    # if byCamera = TRUE
      if(hasArg(allCamsOn))          warning("if cameraCol is set and byCamera is TRUE, allCamsOn will have no effect", call. = FALSE)
      if(hasArg(camerasIndependent)) warning("if cameraCol is set and byCamera is TRUE, camerasIndependent will have no effect", call. = FALSE)
    }
  } else {
    cameraCol <- "camera"
    CTtable[, cameraCol] <-  paste(CTtable[, stationCol], "Cam1", sep = "")     # add a dummy camera column
    if(hasArg(byCamera)) warning("If cameraCol is not defined, byCamera will have no effect")
  }
  
  # check argument sessionCol
  if(isTRUE(sessionColInArgs)){
    checkForSpacesInColumnNames(sessionCol = sessionCol)
    if(!sessionCol %in% colnames(CTtable)) stop(paste('sessionCol = "', sessionCol, '" is not a column name in CTtable', sep = ''), call. = FALSE)
    if(!is.numeric(CTtable[, sessionCol])) stop("Values of sessionCol must be numeric", call. = FALSE)
  } else{
    sessionCol <- "session"
    CTtable[, sessionCol] <- 1   # add a dummy session column
  }
  
  # return error if duplicate stations (i.e. more than 1 row per station)
  if(isFALSE(cameraColInArgs) & isFALSE(sessionColInArgs)){
    if(any(duplicated(CTtable[,stationCol]))){
      tmp <- table(CTtable[,stationCol])
      stop(paste(sum(tmp >= 2)," stations have more than 1 item in CTtable. Please specify 'cameraCol' or 'sessionCol'\n", 
                 paste(names(tmp[tmp >= 2]),
                       tmp[tmp >= 2], sep = ": ", collapse = "\n"),
                 sep = ""), 
           call. = FALSE)
    }
  }
  
  if(isTRUE(cameraColInArgs) & isFALSE(sessionColInArgs)){
    if(any(duplicated(CTtable[,c(stationCol, cameraCol)]))){
      tmp <- table(paste(CTtable[,stationCol], " - ", cameraCol, " ", CTtable[, cameraCol], sep = ""))
      stop(paste(sum(tmp >= 2), " station/camera combinations have more than 1 item in CTtable. Consider specifying 'sessionCol' if you have multiple sessions / seasons\n",
                 paste(names(tmp[tmp >= 2]), tmp[tmp >= 2], sep = ": ", collapse = "\n"), sep = ""),
           call. = FALSE)
    }
  }
  
  if(isFALSE(cameraColInArgs) & isTRUE(sessionColInArgs)){
    if(any(duplicated(CTtable[,c(stationCol, sessionCol)]))){
      tmp <- table(paste(CTtable[,stationCol], " - ", sessionCol, " ", CTtable[, sessionCol], sep = ""))
      stop(paste(sum(tmp >= 2)," station/session combinations have more than 1 item in CTtable. Consider specifying 'cameraCol' if you have multiple cameras per station\n",
                 paste(names(tmp[tmp >= 2]), tmp[tmp >= 2], sep = ": ", collapse = "\n"), sep = ""),
           call. = FALSE)
    }
  }
  
  if(isTRUE(cameraColInArgs) & isTRUE(sessionColInArgs)){
    if(any(duplicated(CTtable[,c(stationCol, cameraCol, sessionCol)]))){
      tmp <- table(paste(CTtable[,stationCol], " - ", cameraCol, " ", CTtable[, cameraCol], " - ", sessionCol, " ", CTtable[, sessionCol], sep = ""))
      stop(paste(sum(tmp >= 2), " station/camera/session combination have more than 1 item in CTtable.\n",
                 paste(names(tmp[tmp >= 2]), tmp[tmp >= 2], sep = ": ", collapse = "\n"), sep = ""),
           call. = FALSE)
    }
  }
  
  if(hasArg(outDir)){
    if(!is.character(outDir)){stop("outDir must be of class 'character'")}
    if(file.exists(outDir) == FALSE) stop("outDir does not exist")
  }
  
  # Convert Date columns to character
  date_columns <- which(sapply(CTtable, is.Date))
  if(length(date_columns) >= 1) {
    CTtable[date_columns] <- lapply(CTtable[date_columns], 
                                    as.character
                                    # FUN = function(col) format(col, format = "%Y-%m-%d")
                                    )
    # dateFormat <- "%Y-%m-%d %H:%M:%S"
    dateFormat <- "%Y-%m-%d"
  }
  
  # Convert POSIX columns to character
  posix_columns <- which(sapply(CTtable, is.POSIXt))
  if(length(posix_columns) >= 1) {
    CTtable[posix_columns] <- lapply(CTtable[posix_columns], 
                                     # as.character
                                     FUN = function(col) format(col, format = "%Y-%m-%d %H:%M:%S")
                                     )
    dateFormat <- "%Y-%m-%d %H:%M:%S"
  }
  
  # if(exists("dateFormat_new")) dateFormat <- dateFormat_new
  
  # if dateFormat contains H (hour), use parseDateTimeObject, otherwise parseDateObject
  if(effortAsFraction) {
    CTtable[,setupCol]     <- parseDateTimeObject(inputColumn = CTtable[,setupCol],     
                                                  dateTimeFormat = dateFormat, 
                                                  checkNA = TRUE, 
                                                  checkEmpty = TRUE, 
                                                  timeZone = "UTC")
    CTtable[,retrievalCol] <- parseDateTimeObject(inputColumn = CTtable[,retrievalCol], 
                                                  dateTimeFormat = dateFormat, 
                                                  checkNA = TRUE, 
                                                  checkEmpty = TRUE, 
                                                  timeZone = "UTC")
  } else {
    CTtable[,setupCol]     <- parseDateObject(inputColumn = CTtable[,setupCol],     
                                              dateFormat = dateFormat, 
                                              checkNA = TRUE, 
                                              checkEmpty = TRUE)
    CTtable[,retrievalCol] <- parseDateObject(inputColumn = CTtable[,retrievalCol], 
                                              dateFormat = dateFormat, 
                                              checkNA = TRUE, 
                                              checkEmpty = TRUE)
  }
  
  # if setup time was not defined, assume 12 noon (so effort on setup/retrieval day = 0.5)
  if(isFALSE(effortAsFraction)){
    CTtable[,setupCol]     <- CTtable[,setupCol]     + dhours(12)
    CTtable[,retrievalCol] <- CTtable[,retrievalCol] + dhours(12) + dseconds(ifelse(occasionStartTime==12 && effortAsFraction==FALSE, 1 ,0))
    if(occasionStartTime==12 && effortAsFraction==FALSE) message("occasionStartTime = 12 and retrieval time is noon. Occasions beginning on retrieval day (at noon) thus have no effort, but are set to 0 instead of NA to prevent omission of records on retrieval day. Include effort as a detection covariate in your models to account for this.")
  }
  # * prevent omission of records on retrieval day when occasionStartTime = 12 and occasions end at noon (if retrieval time undefined)
  # can happen because otherwise effort on retrieval day = NA when setup/retrieval at noon. With this fix it is 0
  # that may influence detection probabilities in model very slightly, especially if not using effort in models
  
  
  if(any(CTtable[,setupCol] == CTtable[,retrievalCol])) stop(paste0("row ", paste(which(CTtable[,setupCol] == CTtable[,retrievalCol]), collapse = ", "), ": setup is identical to retrieval"), call. = FALSE)
  
  # check if dates make sense
  if(any(CTtable[,setupCol]     < as.Date("1970-01-01"))) warning("setup dates begin before 1970. If this is not intended please check dateFormat", call. = FALSE)
  if(any(CTtable[,retrievalCol] < as.Date("1970-01-01"))) warning("retrieval dates are before 1970. If this is not intended please check dateFormat", call. = FALSE)
  
  if(any(CTtable[,setupCol]     > Sys.Date())) warning("setup date is in the future. If this is not intended please check dateFormat", call. = FALSE)
  if(any(CTtable[,retrievalCol] > Sys.Date())) warning("retrieval date is in the future. If this is not intended please check dateFormat", call. = FALSE)
  
  
  
  # ensure setup is before retrieval (treating sessions independently)
  CTtable_split <- split(CTtable, f = CTtable[, sessionCol])
  
  lapply(CTtable_split, FUN = function(x) {
    if(any(x[,setupCol] > x[,retrievalCol])){
      stop(paste("Setup Date after Retrieval Date:   "),
           paste(x[which(x[,setupCol] > x[,retrievalCol]), stationCol],
                 collapse = ", "), 
           ifelse(sessionColInArgs, paste0(" (Session ", unique(x[, sessionCol]), ")"), ""), call. = FALSE)
    }
  }
  )

  # get start / retrieval dates of all cameras
  date0 <- sapply(CTtable[, setupCol],     FUN =  function(x) as.character(min(x)))
  # if(occasionStartTime != 12) {
  date1 <- sapply(CTtable[, retrievalCol], FUN =  function(x) as.character(max(x) - dseconds(1)))  # if not removing 1 second, last day with end on midnight when the day begins, leading to retrieval day being 0
  # } else {
    # date1 <- sapply(CTtable[, retrievalCol], FUN =  function(x) as.character(max(x)))  # if removing 1 second, there will be no overlap between last day and last occasion, resulting in NA and dropped records on last day
    # if(!effortAsFraction) message("occasionStartTime = 12 and retrieval time is noon (since retrieval time is not defined). The occasion beginning at noon on retrieval day thus has no effort. Effort is set to 0 instead of NA though to prevent omission of records on retrieval day in detectionHistory.")
  # }
    
  # create interval from start to end of each camera
  start_to_end <- interval(date0, date1)
  
  if(isTRUE(hasProblems)){
    
    # find problem columns
    cols.prob.from <- grep(colnames(CTtable), pattern = "Problem[0-9]+\\Sfrom")
    cols.prob.to   <- grep(colnames(CTtable), pattern = "Problem[0-9]+\\Sto")
    
    # convert problem column entries to character
    for(problem_col_index in c(cols.prob.from, cols.prob.to)){
      CTtable[, problem_col_index] <- as.character(CTtable[, problem_col_index])
    }
    
    # error if no Problem columns
    if(length(cols.prob.from) == 0) stop("could not find column ProblemX_from")
    if(length(cols.prob.to) == 0)   stop("could not find column ProblemX_to")
    
    if(length(cols.prob.from) != length(cols.prob.to)){
      stop("length of 'Problem..._from' and 'Problem..._to' columns differs. Check column names. Sample: 'Problem1_from', 'Problem1_to'")
    }
    
    if(effortAsFraction) {
      for(problemFromColumn in cols.prob.from){
        CTtable[, problemFromColumn] <- parseDateTimeObject(inputColumn = CTtable[, problemFromColumn], dateFormat, quiet = TRUE,
                                                            checkNA = FALSE, checkEmpty = FALSE, checkNA_out = FALSE, timeZone = "UTC",
                                                            allowEmptyOutput = T)
      }
      for(problemToColumn in cols.prob.to){
        CTtable[, problemToColumn]   <- parseDateTimeObject(inputColumn = CTtable[, problemToColumn],   dateFormat, quiet = TRUE,
                                                            checkNA = FALSE, checkEmpty = FALSE, checkNA_out = FALSE, timeZone = "UTC",
                                                            allowEmptyOutput = T)
      }
    } else {
      for(problemFromColumn in cols.prob.from){
        CTtable[, problemFromColumn] <- parseDateObject(inputColumn = CTtable[, problemFromColumn], dateFormat, 
                                                        checkNA = FALSE, checkEmpty = FALSE, returndatetime = TRUE,
                                                        allowEmptyOutput = TRUE)
      }
      for(problemToColumn in cols.prob.to){
        CTtable[, problemToColumn] <- parseDateObject(inputColumn = CTtable[, problemToColumn], dateFormat, 
                                                      checkNA = FALSE, checkEmpty = FALSE, returndatetime = TRUE,
                                                      allowEmptyOutput = TRUE)
      }
    }
    

    # check that there are some problems at all (since hasProblems = TRUE)
    if(all(is.na(CTtable[, cols.prob.from]))) warning("in problemFromColumn column(s), all values are NA", call. = FALSE)
    if(all(is.na(CTtable[, cols.prob.to])))   warning("in Problem_to column(s), all values are NA", call. = FALSE)
    
    # if problems begin on setup day, make sure it's the same time as setup (if only dates are specified)
    if(isFALSE(effortAsFraction)){
      problem_begin_on_setup_day <- which(as.Date(CTtable[, setupCol]) == as.Date(CTtable[, cols.prob.from[1]]))
      if(length(problem_begin_on_setup_day) >= 1) {
        CTtable[problem_begin_on_setup_day, cols.prob.from[1]] <- CTtable[problem_begin_on_setup_day, setupCol]
      }
      problem_ends_on_retrieval_day <- which(as.Date(CTtable[, retrievalCol]) == as.Date(CTtable[, cols.prob.to[length(cols.prob.to)]]))
      if(length(problem_ends_on_retrieval_day) >= 1) {
        CTtable[problem_ends_on_retrieval_day, cols.prob.to[length(cols.prob.to)]] <- CTtable[problem_ends_on_retrieval_day, retrievalCol]
      }
    }
    
    # check that problems begin after setup
    for(cols.prob.from.index in cols.prob.from){
      if(isTRUE(effortAsFraction)){
        if(any(CTtable[,setupCol] > CTtable[,cols.prob.from.index], na.rm = TRUE)) {
          stop(paste(paste(CTtable[which(CTtable[,setupCol] > CTtable[,cols.prob.from.index]), stationCol], collapse = ", "), ": Problem begins before Setup"), call. = FALSE)
        }
      }
      if(isFALSE(effortAsFraction)){
        if(any(as.Date(CTtable[,setupCol]) > CTtable[,cols.prob.from.index], na.rm = TRUE)){
          stop(paste(paste(CTtable[which(CTtable[,setupCol] > CTtable[,cols.prob.from.index]), stationCol], collapse = ", "), ": Problem begins before Setup"), call. = FALSE)
        } 
      }
    }
    
    # check that problems end before (or on) retrieval
    for(cols.prob.to.index in cols.prob.to){
      if(any(CTtable[,retrievalCol] < CTtable[,cols.prob.to.index], na.rm = TRUE)){
        stop(paste(paste(CTtable[which(CTtable[,retrievalCol] < CTtable[,cols.prob.to.index]), stationCol], collapse = ", "), ": Problem ends after retrieval"), call. = FALSE)
      }
    }
    
    
    # make list of Problem groups
    
    # loop over problem groups, save problem start and end date(time)
    # this is a bit hacky, but ensures that it works even if Problem columns are not ordered 
    # list item = Problem index (Problem1, Problem2, ...); vector items within = cameras
    problem_colnames_index_list <- vector(mode = "list", length = length(cols.prob.from))
    
    for(Problem_group in 1:length(problem_colnames_index_list)){
      problem_colnames_index_list[[Problem_group]]$prob.from <- CTtable[, cols.prob.from [order(colnames(CTtable) [cols.prob.from])] [Problem_group]]
      problem_colnames_index_list[[Problem_group]]$prob.to   <- CTtable[, cols.prob.to   [order(colnames(CTtable) [cols.prob.to])]   [Problem_group]]
    }
    
    # enure that no problem pairs are all NA
    lapply(1:length(problem_colnames_index_list),
           FUN = function(x) {
             if(all(is.na(unlist(problem_colnames_index_list[x])))) stop(paste0("Problem", x, " columns are both NA"), call. = F)
           })
    
    # loop over cameras, make and concatenate the problem intervals by station
    problem_intervals_by_row <- list()
    for(row_index in 1:nrow(CTtable)){
      
      # make data frame of problem start / end times
      tmp <- lapply(problem_colnames_index_list, FUN = function(x) {
        
        if(is.na( x$prob.from  [row_index]) & is.na( x$prob.to  [row_index])){
          return(data.frame(start = NA, end   = NA))
        }
        if(is.na( x$prob.from  [row_index]) & !is.na( x$prob.to  [row_index])){
          stop("row ", row_index, ": Problem_from is NA, but Problem_to is ", x$prob.to  [row_index], call. = FALSE)
        }
        if(!is.na( x$prob.from  [row_index]) & is.na( x$prob.to  [row_index])){
          stop("row ", row_index, ": Problem_from is ", x$prob.from  [row_index], " but Problem_to is NA", call. = FALSE)
        }
        
        # when using ifelse, output is numeric, not POSIXct
        if(effortAsFraction)  prob.to <- x$prob.to  [row_index] #- dseconds(1)           
        # if Problem_to is defined as date, add 1 day - 1 second (so problem ends just before midnight the same day)
        if(!effortAsFraction) {
          prob.to <- x$prob.to  [row_index] + ddays(1) - dseconds(1) 
          # if problem period were to end after end of camera trapping period, replace the end of Problem with end of camera trapping period
          if(prob.to > int_end(start_to_end[row_index])){
            prob.to <- int_end(start_to_end[row_index])
          }
        }
        data.frame(start = x$prob.from[row_index],
                   end   = prob.to)
      })
      
      # combine date.frames for problem start end times (if multiple problem periods defines)
      tmp.rbind <- do.call(rbind, tmp)
      
      
      # make intervals for the problem periods
      if(!all(is.na(tmp.rbind))) {
        problem_intervals_by_row[[row_index]] <- interval(start = tmp.rbind$start, tmp.rbind$end  - dseconds(1))   # -1 to avoid problems matching with camera operation date-time
        if(any(time_length(problem_intervals_by_row[[row_index]]) < 0, na.rm = TRUE)) stop("row", row_index, ": Problem ends before it starts.")
      } else {
        problem_intervals_by_row[[row_index]] <- NA
      }
    }
    
    rm(problemFromColumn, problemToColumn, cols.prob.from.index, cols.prob.to.index)
  }
  
  # create empty matrix with desired dimensions (depending on presence of camera / session columns)
  arg_list <- list(CTtable = CTtable,
                   stationCol = stationCol,
                   setupCol = setupCol,
                   retrievalCol = retrievalCol
  )
  
  if(cameraColInArgs)  arg_list <- c(arg_list, "cameraCol" = cameraCol)
  if(sessionColInArgs) arg_list <- c(arg_list, "sessionCol" = sessionCol)
  
  camOp_empty <- do.call(stationSessionCamMatrix, args = arg_list) 
  
  
  
  
  
  
  #  trapping intervals for all cameras (setup to retrieval)
  # for each day in camop, make an interval covering the entire day
  camop_daily_intervals <- lapply(as.Date(colnames(camOp_empty)),
                                  FUN = function(x){
                                    start <- x + dhours(occasionStartTime)
                                    end <- start + ddays(1) - dseconds(1)
                                    interval(start = start,  
                                             end =  end)
                                  }
                                  )
  names(camop_daily_intervals) <- colnames(camOp_empty)
  
  # get start / end of the days covered by the study (+ occasionStartTime, if defined)
  int_start_daily <- lapply(camop_daily_intervals, FUN = function(x) int_start(x))
  int_end_daily   <- lapply(camop_daily_intervals, FUN = function(x) int_end(x))
  
  # get start / end of camera trapping period by camera
  int_start_total <- int_start(start_to_end)
  int_end_total   <- int_end(start_to_end)
  
  # alternative to data.table madness below. Find overlapping intervals between camera traps and days
  camop_binary <- sapply(camop_daily_intervals, int_overlaps, start_to_end)
  
  # handle case of 1-camera input table
  if(nrow(CTtable) == 1) {
    camop_binary <- t(as.matrix(camop_binary))
    dimnames(camop_binary) <- dimnames(camOp_empty)
  }
  
  rownames(camop_binary) <- rownames(camOp_empty)
  
  
  
  
  # identify overlapping intervals with data.table
  # https://www.howtobuildsoftware.com/index.php/how-do/bzA9/r-intervals-lubridate-r-and-lubridate-do-the-intervals-in-x-fit-into-any-of-the-intervals-in-y
  
  # # start and end of each row in camera trap table (setup to retrieval)
  # time_intervals_cttable <- data.frame(start = sapply(int_start(start_to_end), as.POSIXct), 
  #                                      end   = sapply(int_end(start_to_end), as.POSIXct))
  # 
  # # start and end of each day
  # time_intervals_each_day <- data.frame(start = sapply(int_start_daily, as.POSIXct), 
  #                                       end   = sapply(int_end_daily, as.POSIXct))
  # 
  # setDT(time_intervals_cttable)[, `:=`(start = start,
  #                                      end   = end)]
  # setkey(setDT(time_intervals_each_day)[, `:=`(start = start,
  #                                              end   = end)], 
  #        start, end)
  # # find matching time intervals (between days and cameras set up)
  # intervals_matched <- foverlaps(time_intervals_cttable, 
  #                                time_intervals_each_day, 
  #                                type = "any", 
  #                                which = TRUE)
  
  # loop over cameras (= rows)
  for(i in 1:nrow(camOp_empty)){
    
    run_these <- which(camop_binary[i,])    #intervals_matched[intervals_matched$xid == i,]$yid
    
    # intersect daily intervals with interval from setup to retrieval
    interval.tmp <- sapply(camop_daily_intervals[run_these], intersect.Interval.fast, start_to_end[i])

    # assign values to camera operation matrix
    camOp_empty[i, run_these] <- time_length(interval.tmp, unit = "days")
    
    
    # if problems are defined, subtract those from the camera operation values
    if(hasProblems) {
      if(any(!is.na(problem_intervals_by_row[[i]]))){
        if(!all(problem_intervals_by_row[[i]] %within% start_to_end[i], na.rm = TRUE)) stop(paste(CTtable[i,stationCol], ": problem intervals are not within interval from setup to retrieval"))
        
        interval.tmp.prob <- sapply(camop_daily_intervals[run_these], intersect.Interval.fast, problem_intervals_by_row[[i]])   # intersection of day and total trapping period
        # total Problem value per day
        if(inherits(interval.tmp.prob, "array")) {
          fraction_to_remove <- time_length(colSums(interval.tmp.prob, na.rm = TRUE), unit = "days")   # if mutliple problem periods are defined, they show up as rows here and are combined with colSums
        } else {
          fraction_to_remove <- time_length(interval.tmp.prob, unit = "days")
        }
        # replace NA with 0 
        fraction_to_remove <- ifelse(is.na(fraction_to_remove), 0, fraction_to_remove)
        # assign values to camera operation matrix
        camOp_empty[i,run_these] <- camOp_empty[i,run_these] - fraction_to_remove
      }
    }
    
    if(any(camOp_empty[i, run_these] < 0)) stop(paste("Negative effort calculated in", rownames(camOp_empty)[i], "on:",
                                                      paste(names(which(camOp_empty[i, run_these] < 0)), collapse = ", "), "\n",
                                                      "Check for overlapping dates in Problem columns"))
  }
  
  #camOp_filled <- camOp_empty #
  camOp_filled <- round(camOp_empty, 4)   # to account for slight imprecision because daily interval is 86399 seconds, not 86400s
  
  
  if(occasionStartTime != 0) colnames(camOp_filled) <- paste0(colnames(camOp_filled), "+", occasionStartTime, "h")
  
  if(isTRUE(cameraColInArgs)){   # there is a camera column, i.e., potentially > 1 cameras per station
    
    if(isFALSE(byCamera)){        # if aggregate to station level (byCamera = FALSE)
      
      separatorSession <- "__SESS_"
      
      # byCamera = FALSE, allCamsOn = TRUE, camerasIndependent = TRUE
      if(allCamsOn){
        if(camerasIndependent){
          # if all cameras have full effort, value = sum. 
          if(effortAsFraction){
            # If any camera is not active whole day, use minimum value x n_cameras (fraction of day that all cameras were active)
            # doesn't make much sense, but included for completeness sake
            dat2 <- aggregate(camOp_filled, by = list(CTtable[, c(stationCol)],
                                                      CTtable[, c(sessionCol)]), 
                              FUN = function(x) ifelse(all(x == 1), sum(x), min(x) * length(x)))
            
          } else {
            # if all cameras have full effort, value = sum. If any camera is not active whole day, count station as inactive (0)
            dat2 <- aggregate(camOp_filled, by = list(CTtable[, c(stationCol)],
                                                      CTtable[, c(sessionCol)]), 
                              FUN = function(x) ifelse(all(x == 1), sum(x), 0))
          }
        }
        
        # byCamera = FALSE, allCamsOn = TRUE, camerasIndependent = FALSE
        if(isFALSE(camerasIndependent)) {
          
          if(effortAsFraction){
            # if all cameras at station active, effort = 1. If any camera inactive, 0. If any camera partially active, the fraction of day for that particular camera.
            dat2 <- aggregate(camOp_filled, by = list(CTtable[, c(stationCol)],
                                                      CTtable[, c(sessionCol)]),
                              FUN = min)
          } else {
            # if all cameras at station active, effort = 1. If any camera not active, 0
            dat2 <- aggregate(camOp_filled, by = list(CTtable[, c(stationCol)],
                                                      CTtable[, c(sessionCol)]), 
                              FUN = function(x) ifelse(all(x == 1), 1, 0))  
          }
        }
        
        # asssign row names (adding session ID, if applicable)
        if(sessionColInArgs)  row.names(dat2) <- paste(dat2[,1], dat2[,2], sep = separatorSession)
        if(!sessionColInArgs) row.names(dat2) <- dat2[,1]
        
        # remove station & session columns
        dat2[,c(1,2)] <- NULL
      } 
      
      # byCamera = FALSE, allCamsOn = FALSE,
      if(isFALSE(allCamsOn)){
        # allCamsOn = FALSE means that not all cameras need to be active simultaneously for the station to be considered active.
        
        # byCamera = FALSE, allCamsOn = FALSE, camerasIndependent = TRUE
        if(camerasIndependent){
          # sum of daily effort from multiple cameras by station (complete days or fraction days, depending on input dateFormat)
          # same for effortAsFraction = TRUE or FALSE
          dat2    <- aggregate(camOp_filled, by = list(CTtable[, c(stationCol)],
                                                       CTtable[, c(sessionCol)]), 
                               FUN = sum, na.rm = TRUE)
        } 
        
        # byCamera = FALSE, allCamsOn = FALSE, camerasIndependent = FALSE
        if(isFALSE(camerasIndependent)){
          
          if(effortAsFraction){
            # if days as fraction, effort per station = mean fraction per day of cameras
            dat2    <- aggregate(camOp_filled, by = list(CTtable[, c(stationCol)],
                                                         CTtable[, c(sessionCol)]), 
                                 FUN = mean, na.rm = TRUE)
          } else {
            # if days NOT as fraction, effort per station = sum of cameras days
            dat2    <- aggregate(camOp_filled, by = list(CTtable[, c(stationCol)],
                                                         CTtable[, c(sessionCol)]), 
                                 FUN = function(x) ifelse(sum(x, na.rm = TRUE) >= 1, 1, sum(x, na.rm = TRUE)))
          }
        }
        
        # find cells that are NA for all cameras
        dat2.na <- aggregate(camOp_filled, by = list(CTtable[, c(stationCol)],
                                                     CTtable[, c(sessionCol)]), 
                             FUN = function(X){all(is.na(X))})     # TRUE if no camera was set up that day
        
        # asssign row names (adding session ID, if applicable)
        if(sessionColInArgs)  row.names(dat2) <- row.names(dat2.na) <- paste(dat2[,1], dat2[,2], sep = separatorSession)
        if(!sessionColInArgs) row.names(dat2) <- row.names(dat2.na) <- dat2[,1]
        
        # remove station & session columns
        dat2[,c(1,2)] <- dat2.na[,c(1,2)] <- NULL
        
        dat2     <- as.matrix(dat2)
        dat2.na  <- as.matrix(dat2.na)
        
        # fill in NAs in matrix (when all cameras were NA)
        if(any(dat2.na))  dat2[which(dat2.na)] <- NA
        
        # if cameras are not independent and effort by day (not hour), replace 2 or higher with 1
        if(camerasIndependent == FALSE & !effortAsFraction){
          dat2 <- ifelse(dat2 >= 2, 1, dat2)
        }
        dat2 <- as.data.frame(dat2)
      }
    } else {   # belongs to:    if(isFALSE(byCamera)){  
      dat2 <- as.data.frame(camOp_filled)
    }   # end if(cameraColInArgs)
  } else {    # belongs to: if(cameraColInArgs)
    # if only station information, no camera level information
    dat2 <- as.data.frame(camOp_filled)
  } # end if(cameraColInArgs)
  
  if(writecsv == TRUE){
    
    # assemble parts of outfile name (according to function arguments)
    hasProblemsString <- ifelse(isTRUE(hasProblems), "with_problems_", "")
    
    if(cameraColInArgs){
      byCameraString <- ifelse(byCamera, "by_camera", "by_station")
    } else {
      byCameraString <- "by_station"
    }
    if(sessionColInArgs) byCameraString <- paste(byCameraString, "_by_session", sep = "")
    
    filename.out <- paste("CameraOperationMatrix_", byCameraString, "_", hasProblemsString, Sys.Date(), ".csv", sep = "")
    
    if(hasArg(outDir) == FALSE){
      setwd(getwd())
      write.csv(dat2, file = filename.out,
                row.names = TRUE)
    } else {
      setwd(outDir)
      write.csv(dat2, file = filename.out,
                row.names = TRUE)
    }
    if(missing(outDir)) message(paste("writecsv was TRUE, but outDir was not defined. Saved camera operation matrix in:", getwd(), sep = "   "))
  }
  return(as.matrix(dat2))
}

context("spatialDetectionHistory")


library(testthat)
library(secr)

# --- Test Setup ---

# 1. Load single-session data
data(recordTableIndividualSample)
data(camtraps)

# 2. Create a standard single-session camera operation matrix
camop_problem <- cameraOperation(CTtable      = camtraps,
                                 stationCol   = "Station",
                                 setupCol     = "Setup_date",
                                 retrievalCol = "Retrieval_date",
                                 hasProblems  = TRUE,
                                 dateFormat   = "dmy")

# 3. Load and prepare multi-session data
data(camtrapsMultiSeason)
# The function expects session IDs to be numeric (1, 2, ...)
camtrapsMultiSeason$session[camtrapsMultiSeason$session == 2009] <- 1
camtrapsMultiSeason$session[camtrapsMultiSeason$session == 2010] <- 2
data(recordTableIndividualSampleMultiSeason)

# 4. Create a multi-session camera operation matrix
camop_session <- cameraOperation(CTtable      = camtrapsMultiSeason,
                                 stationCol   = "Station",
                                 setupCol     = "Setup_date",
                                 sessionCol   = "session",
                                 retrievalCol = "Retrieval_date",
                                 hasProblems  = TRUE,
                                 dateFormat   = "dmy")


# --- Test Suite ---

testthat::describe("Core Functionality (Single Session)", {
  
  # Skip all tests in this file if secr is not installed
  skip_if_not_installed("secr")
  
  test_that("it creates a valid single-session capthist object", {
    sdh <- spatialDetectionHistory(
      recordTableIndividual = recordTableIndividualSample,
      species               = "LeopardCat",
      camOp                 = camop_problem,
      CTtable               = camtraps,
      output                = "binary",
      stationCol            = "Station",
      Xcol                  = "utm_x",
      Ycol                  = "utm_y",
      individualCol         = "Individual",
      recordDateTimeCol     = "DateTimeOriginal",
      recordDateTimeFormat  = "ymd HMS",
      occasionLength        = 10,
      day1                  = "survey",
      includeEffort         = TRUE,
      timeZone              = "Asia/Kuala_Lumpur"
    )
    
    # 1. Check main object class and structure
    expect_s3_class(sdh, "capthist")
    expect_false(is.list(sdh))  # Single-session is not a list
    
    # 2. Check dimensions and content
    expect_equal(dim(sdh), c(3,5,3))
    
    # 3. Check traps attribute
    traps_obj <- secr::traps(sdh)
    expect_s3_class(traps_obj, "traps")
    expect_equal(nrow(traps_obj), nrow(camtraps))
    
    # 4. Check usage (effort) attribute
    expect_true(!is.null(secr::usage(traps_obj)))
    expect_equal(dim(secr::usage(traps_obj)), c(nrow(camtraps), 5)) # stations x occasions
  })
})


testthat::describe("Multi-Session Functionality", {
  
  skip_if_not_installed("secr")
  
  test_that("it creates a valid multi-session capthist object", {
    sdh_multi <- spatialDetectionHistory(
      recordTableIndividual = recordTableIndividualSampleMultiSeason,
      species               = "LeopardCat",
      camOp                 = camop_session,
      CTtable               = camtrapsMultiSeason,
      sessionCol            = "session",
      output                = "binary",
      stationCol            = "Station",
      Xcol                  = "utm_x",
      Ycol                  = "utm_y",
      individualCol         = "Individual",
      occasionLength        = 10,
      timeZone              = "Asia/Kuala_Lumpur",
      day1                  = "station"
    )
    
    # 1. Check main object class and structure
    expect_s3_class(sdh_multi, "capthist")
    expect_true(is.list(sdh_multi))  # Multi-session IS a list
    expect_length(sdh_multi, 2)      # Two sessions in the data
    
    # 2. Check that list elements are valid capthist objects
    expect_s3_class(sdh_multi[[1]], "capthist")
    expect_s3_class(sdh_multi[[2]], "capthist")
    
  })
})


testthat::describe("Covariate Handling", {
  
  skip_if_not_installed("secr")
  
  test_that("it correctly attaches station and individual covariates", {
    sdh_cov <- spatialDetectionHistory(
      recordTableIndividual = recordTableIndividualSampleMultiSeason,
      species               = "LeopardCat",
      camOp                 = camop_session,
      CTtable               = camtrapsMultiSeason,
      sessionCol            = "session",
      Xcol                  = "utm_x",
      Ycol                  = "utm_y",
      individualCol         = "Individual",
      occasionLength        = 10,
      timeZone              = "Asia/Kuala_Lumpur",
      stationCovariateCols  = c("utm_x", "utm_y"), # Use a column from CTtable as a station covariate
      individualCovariateCols = c("Individual", "Species"), # Use columns as individual covariates
      day1                  = "survey"
    )
    
    # 1. Check for station covariates
    # In multi-session, check one of the sessions' traps
    station_covs <- secr::covariates(secr::traps(sdh_cov[[1]]))
    expect_s3_class(station_covs, "data.frame")
    expect_true(all(names(station_covs) %in% c("utm_y", "utm_x")))
    
    # For single covariate, column name is "value". Fix!
    
    
    # 2. Check for individual covariates
    ind_covs <- secr::covariates(sdh_cov)
    expect_s3_class(ind_covs[[1]], "data.frame")
    expect_true(all(c("Individual", "Species") %in% names(ind_covs[[1]])))
  })
})


testthat::describe("Output Modes and Other Options", {
  
  skip_if_not_installed("secr")
  
  test_that("output = 'count' works correctly", {
    sdh_count <- spatialDetectionHistory(
      recordTableIndividual = recordTableIndividualSample,
      species               = "LeopardCat",
      camOp                 = camop_problem,
      CTtable               = camtraps,
      output                = "count", # Key argument
      occasionLength        = 10,
      Xcol = "utm_x", Ycol = "utm_y", 
      individualCol = "Individual",
      timeZone = "Asia/Kuala_Lumpur",
      day1 = "station"
    )
    
    # The detector attribute should be 'count'
    expect_equal(secr::detector(secr::traps(sdh_count)), "count")
    
    # The sample data has an individual detected twice in one occasion,
    # so at least one cell value should be > 1.
    expect_true(any(sdh_count > 1))
  })
  
  test_that("makeRMarkInput = TRUE produces a data.frame", {
    rmark_df <- spatialDetectionHistory(
      recordTableIndividual = recordTableIndividualSample,
      species               = "LeopardCat",
      camOp                 = camop_problem,
      CTtable               = camtraps,
      makeRMarkInput        = TRUE, # Key argument
      occasionLength        = 10,
      Xcol                  = "utm_x", 
      Ycol                  = "utm_y", 
      individualCol         = "Individual",
      timeZone              = "Asia/Kuala_Lumpur",
      day1                  = "2009-04-05" 
    )
    
    # 1. Check output class
    expect_s3_class(rmark_df, "data.frame")
    expect_false(inherits(rmark_df, "capthist"))
    
    # 2. Check structure
    expect_true("ch" %in% names(rmark_df)) # Must have a capture history column
    expect_equal(nrow(rmark_df), 3) # One row per individual
    
    # 3. Check capture history string format
    expect_true(all(nchar(rmark_df$ch) == 5))
    expect_true(is.character(rmark_df$ch))
  })
})


testthat::describe("Input Validation", {
  
  test_that("it stops for missing or invalid inputs", {
    # Missing camOp
    expect_error(spatialDetectionHistory(recordTableIndividual = recordTableIndividualSample,
                                         species = "LeopardCat",
                                         CTtable = camtraps,
                                         occasionLength = 10,
                                         Xcol = "utm_x", 
                                         Ycol = "utm_y",
                                         individualCol = "Individual"),
                 "'camOp' is not defined")
    
    # Wrong column name
    expect_error(spatialDetectionHistory(recordTableIndividual = recordTableIndividualSample,
                                         species = "LeopardCat",
                                         camOp = camop_problem,
                                         CTtable = camtraps,
                                         occasionLength = 10,
                                         Xcol = "WRONG_X_COL", Ycol = "utm_y",
                                         individualCol = "Individual"),
                 "is not a column name in CTtable")
  })
})

context("cameraOperation")
library(camtrapR)
library(lubridate)


# load station information
data(camtraps)

# create camera operation matrix
# dates are specified as character
camop_no_problem <- cameraOperation(CTtable      = camtraps,
                                    stationCol   = "Station",
                                    setupCol     = "Setup_date",
                                    retrievalCol = "Retrieval_date",
                                    hasProblems  = FALSE,
                                    dateFormat   = "%d/%m/%Y"
)

# with problems/malfunction
camop_problem <- cameraOperation(CTtable      = camtraps,
                                 stationCol   = "Station",
                                 setupCol     = "Setup_date",
                                 retrievalCol = "Retrieval_date",
                                 writecsv     = FALSE,
                                 hasProblems  = TRUE,
                                 dateFormat   = "%d/%m/%Y"
)

# table with duplicate rows
camtraps_duplicate <- camtraps
camtraps_duplicate[4,]    <- camtraps_duplicate[1,]

camtraps_duplicate$camera <- "cam1"
camtraps_duplicate[c(5,6),] <- camtraps_duplicate[3,]
camtraps_duplicate$session <- 1


# convert setup / retrieval / problem to Date class
camtraps_date <- camtraps
camtraps_date$Setup_date      <- dmy(camtraps$Setup_date)
camtraps_date$Retrieval_date  <- dmy(camtraps$Retrieval_date )
camtraps_date$Problem1_from   <- dmy(camtraps$Problem1_from )
camtraps_date$Problem1_to     <- dmy(camtraps$Problem1_to )


# dates are specified as Date class
camop_dates <- cameraOperation(CTtable      = camtraps_date,
                               stationCol   = "Station",
                               setupCol     = "Setup_date",
                               retrievalCol = "Retrieval_date",
                               #writecsv     = FALSE,
                               hasProblems  = TRUE,
                               dateFormat   = "ymd"
)



# Date-time 

# add time of day to setup / retrieval / problem
# date-time is provided as POXIXct

camtraps_date_time <- camtraps
camtraps_date_time$Setup_date      <- dmy_h(paste(camtraps$Setup_date, "12"))
camtraps_date_time$Retrieval_date  <- dmy_h(paste(camtraps$Retrieval_date, "15" ))
camtraps_date_time$Problem1_from   <- dmy_h(paste(camtraps$Problem1_from, "06" ), quiet = TRUE)
camtraps_date_time$Problem1_to     <- dmy_h(paste(camtraps$Problem1_to, "15" ), quiet = TRUE)

camop_dates_time <- cameraOperation(CTtable      = camtraps_date_time,
                                    stationCol   = "Station",
                                    setupCol     = "Setup_date",
                                    retrievalCol = "Retrieval_date",
                                    hasProblems  = TRUE,
                                    dateFormat   = "ymd HMS"
)


# date-time is provided as character
camtraps_date_time_char <- camtraps_date_time
camtraps_date_time_char$Setup_date      <- as.character(camtraps_date_time$Setup_date)
camtraps_date_time_char$Retrieval_date  <- as.character(camtraps_date_time$Retrieval_date)
camtraps_date_time_char$Problem1_from   <- as.character(camtraps_date_time$Problem1_from)
camtraps_date_time_char$Problem1_to     <- as.character(camtraps_date_time$Problem1_to)

camop_dates_time_char <- cameraOperation(CTtable      = camtraps_date_time_char,
                                    stationCol   = "Station",
                                    setupCol     = "Setup_date",
                                    retrievalCol = "Retrieval_date",
                                    hasProblems  = TRUE,
                                    dateFormat   = "ymd HMS"
)


# set one time to midnight. It should still work.
# issue is that as.character(POSIXct) removes the time if it's exactly midnight
# input is character
camtraps_date_time_char_midnight <- camtraps_date_time_char
camtraps_date_time_char_midnight$Setup_date[1] <- "2009-04-02 00:00:00"

camop_dates_time_char_midnight <- cameraOperation(CTtable      = camtraps_date_time_char_midnight,
                                    stationCol   = "Station",
                                    setupCol     = "Setup_date",
                                    retrievalCol = "Retrieval_date",
                                    hasProblems  = TRUE,
                                    dateFormat   = "ymd HMS"
)

# input is POSIXct
# camtraps_date_time_char <- camtraps_date_time
# camtraps_date_time_char$Setup_date <- format(camtraps_date_time_char$Setup_date, 
#                                         format = "%Y-%m-%d %H:%M:%S")

camtraps_date_time_midnight <- camtraps_date_time_char_midnight
camtraps_date_time_midnight$Setup_date <- ymd_hms(camtraps_date_time_midnight$Setup_date)
camtraps_date_time_midnight$Retrieval_date <- ymd_hms(camtraps_date_time_midnight$Retrieval_date)
camtraps_date_time_midnight$Problem1_from <- ymd_hms(camtraps_date_time_midnight$Problem1_from)
camtraps_date_time_midnight$Problem1_to <- ymd_hms(camtraps_date_time_midnight$Problem1_to)



camop_dates_time_midnight <- cameraOperation(CTtable      = camtraps_date_time_midnight,
                                     stationCol   = "Station",
                                     setupCol     = "Setup_date",
                                     retrievalCol = "Retrieval_date",
                                     hasProblems  = TRUE,
                                     dateFormat   = "ymd HMS"
)


# change datetime format
camtraps_date_time_format <- camtraps_date_time
stupid_date_format <- "%H:%M:%S %d/%m/%Y"    #"%d/%m/%Y %H:%M:%S"
camtraps_date_time_format$Setup_date <- format(camtraps_date_time_format$Setup_date, 
                                        format = stupid_date_format)
camtraps_date_time_format$Retrieval_date <- format(camtraps_date_time_format$Retrieval_date, 
                                        format = stupid_date_format)
camtraps_date_time_format$Problem1_from <- format(camtraps_date_time_format$Problem1_from, 
                                        format = stupid_date_format)
camtraps_date_time_format$Problem1_to <- format(camtraps_date_time_format$Problem1_to, 
                                            format = stupid_date_format)

camop_dates_time_format <- cameraOperation(CTtable      = camtraps_date_time_format,
                                     stationCol   = "Station",
                                     setupCol     = "Setup_date",
                                     retrievalCol = "Retrieval_date",
                                     hasProblems  = TRUE,
                                     dateFormat   = "HMS dmy"
)



# add empty Problem columns
camtraps2 <- camtraps
camtraps2$Problem2_from <- NA
camtraps2$Problem2_to <- NA



# Test section
test_that("POSIX and character give same output", {
  expect_equal(camop_dates_time, camop_dates_time_char)
  expect_equal(camop_dates_time_char_midnight, camop_dates_time_midnight)
})

test_that("Different date formats work", {
  expect_equal(camop_dates_time_format, camop_dates_time)
})

camop_dates_time_format


test_that("duplicate row error messages work", {
  expect_error(cameraOperation(CTtable      = camtraps_duplicate,
                               stationCol   = "Station",
                               #sessionCol = "session",
                               #cameraCol = "camera",
                               #byCamera = TRUE,
                               setupCol     = "Setup_date",
                               retrievalCol = "Retrieval_date",
                               writecsv     = FALSE,
                               hasProblems  = FALSE,
                               dateFormat   = "%d/%m/%Y"
  ),
  "2 stations have more than 1 item in CTtable. Please specify 'cameraCol' or 'sessionCol'
StationA: 2
StationC: 3", fixed = TRUE)
  
  expect_error(cameraOperation(CTtable      = camtraps_duplicate,
                               stationCol   = "Station",
                               #sessionCol = "session",
                               cameraCol = "camera",
                               byCamera = TRUE,
                               setupCol     = "Setup_date",
                               retrievalCol = "Retrieval_date",
                               writecsv     = FALSE,
                               hasProblems  = FALSE,
                               dateFormat   = "%d/%m/%Y"
  ), 
  "2 station/camera combinations have more than 1 item in CTtable. Consider specifying 'sessionCol' if you have multiple sessions / seasons
StationA - camera cam1: 2
StationC - camera cam1: 3")
  
  expect_error(cameraOperation(CTtable      = camtraps_duplicate,
                               stationCol   = "Station",
                               sessionCol = "session",
                               #cameraCol = "camera",
                               #byCamera = TRUE,
                               setupCol     = "Setup_date",
                               retrievalCol = "Retrieval_date",
                               writecsv     = FALSE,
                               hasProblems  = FALSE,
                               dateFormat   = "%d/%m/%Y"
  ), 
  "2 station/session combinations have more than 1 item in CTtable. Consider specifying 'cameraCol' if you have multiple cameras per station
StationA - session 1: 2
StationC - session 1: 3")
  
  expect_error(cameraOperation(CTtable      = camtraps_duplicate,
                               stationCol   = "Station",
                               sessionCol = "session",
                               cameraCol = "camera",
                               byCamera = TRUE,
                               setupCol     = "Setup_date",
                               retrievalCol = "Retrieval_date",
                               writecsv     = FALSE,
                               hasProblems  = FALSE,
                               dateFormat   = "%d/%m/%Y"
  )
  , 
  "2 station/camera/session combination have more than 1 item in CTtable.
StationA - camera cam1 - session 1: 2
StationC - camera cam1 - session 1: 3")
  
}
)


test_that("Input columns in class 'Date' work", {
  
  expect_identical(camop_dates, camop_problem)
}
)


test_that("Input columns with hours (POSIXct) work", {

  expect_true(inherits(camop_dates_time, "matrix"))
  expect_true(!identical(camop_dates_time, camop_dates))
}
)

test_that("Error when Problem ends after retrieval", {
  
  # Date-time as POSIX
  camtraps_date_time$Problem1_to[3] <- camtraps_date_time$Problem1_to[3] + 1
  expect_error(cameraOperation(CTtable  = camtraps_date_time,
                                      stationCol   = "Station",
                                      setupCol     = "Setup_date",
                                      retrievalCol = "Retrieval_date",
                                      hasProblems  = TRUE,
                                      dateFormat   = "ymd HMS"),
               "StationC : Problem ends after retrieval"
  )
  
  # Dates as character
  camtraps$Problem1_to[3] <- "18/05/2009"
  expect_error(cameraOperation(CTtable      = camtraps,
                                   stationCol   = "Station",
                                   setupCol     = "Setup_date",
                                   retrievalCol = "Retrieval_date",
                                   writecsv     = FALSE,
                                   hasProblems  = TRUE,
                                   dateFormat   = "%d/%m/%Y"),
               "StationC : Problem ends after retrieval"
  )
}
)

test_that("setting time to 00:00:00 works", {
  
  expect_true(camop_dates_time_midnight[1,1] == 1)
  expect_true(camop_dates_time_char_midnight[1,1] == 1)
  expect_true(!identical(camop_dates_time_midnight, camop_dates))
}
)


test_that("Error when all NA in problem columns", {
  
  expect_error(cameraOperation(CTtable      = camtraps2,
                               stationCol   = "Station",
                               setupCol     = "Setup_date",
                               retrievalCol = "Retrieval_date",
                               writecsv     = FALSE,
                               hasProblems  = TRUE,
                               dateFormat   = "dmy"
  ),
  "Problem2 columns are both NA"
  )
}
)

context("detectionHistory")
library(camtrapR)


# load station information
data(camtraps)

# create camera operation matrix
camop_no_problem <- cameraOperation(CTtable      = camtraps,
                                    stationCol   = "Station",
                                    setupCol     = "Setup_date",
                                    retrievalCol = "Retrieval_date",
                                    hasProblems  = FALSE,
                                    dateFormat   = "%d/%m/%Y"
)

data(recordTableSample)    # load the record history

# compute detection history for a species

# without trapping effort
DetHist1 <- detectionHistory(recordTable          = recordTableSample,
                             camOp                = camop_no_problem,
                             stationCol           = "Station",
                             speciesCol           = "Species",
                             recordDateTimeCol    = "DateTimeOriginal",
                             species              = "VTA",
                             occasionLength       = 7,
                             day1                 = "station",
                             datesAsOccasionNames = FALSE,
                             includeEffort        = FALSE,
                             timeZone             = "Asia/Kuala_Lumpur"
)


# with effort / using base R to define recordDateTimeFormat
DetHist2 <- detectionHistory(recordTable          = recordTableSample,
                             camOp                = camop_no_problem,
                             stationCol           = "Station",
                             speciesCol           = "Species",
                             recordDateTimeCol    = "DateTimeOriginal",
                             species              = "VTA",
                             occasionLength       = 7,
                             day1                 = "station",
                             datesAsOccasionNames = FALSE,
                             includeEffort        = TRUE,
                             scaleEffort          = FALSE,
                             timeZone             = "Asia/Kuala_Lumpur"
)

DetHist2a <- detectionHistory(recordTable          = recordTableSample,
                              camOp                = camop_no_problem,
                              stationCol           = "Station",
                              speciesCol           = "Species",
                              recordDateTimeCol    = "DateTimeOriginal",
                              species              = "VTA",
                              occasionLength       = 7,
                              day1                 = "station",
                              datesAsOccasionNames = FALSE,
                              includeEffort        = TRUE,
                              scaleEffort          = TRUE,
                              timeZone             = "Asia/Kuala_Lumpur"
)


# with effort / using lubridate package to define recordDateTimeFormat
DetHist2_lub <- detectionHistory(recordTable          = recordTableSample,
                                 camOp                = camop_no_problem,
                                 stationCol           = "Station",
                                 speciesCol           = "Species",
                                 recordDateTimeCol    = "DateTimeOriginal",
                                 recordDateTimeFormat = "ymd HMS",
                                 species              = "VTA",
                                 occasionLength       = 7,
                                 day1                 = "station",
                                 datesAsOccasionNames = FALSE,
                                 includeEffort        = TRUE,
                                 scaleEffort          = FALSE,
                                 timeZone             = "Asia/Kuala_Lumpur"
)



DetHist1day <- detectionHistory(recordTable          = recordTableSample,
                                camOp                = camop_no_problem,
                                stationCol           = "Station",
                                speciesCol           = "Species",
                                recordDateTimeCol    = "DateTimeOriginal",
                                species              = "VTA",
                                occasionLength       = 1,
                                day1                 = "station",
                                datesAsOccasionNames = F,
                                includeEffort        = T,
                                scaleEffort          = FALSE,
                                timeZone             = "Asia/Kuala_Lumpur"
)

# day1 = "survey", datesAsOccasionNames = TRUE
DetHist1day_survey <- detectionHistory(recordTable          = recordTableSample,
                                       camOp                = camop_no_problem,
                                       stationCol           = "Station",
                                       speciesCol           = "Species",
                                       recordDateTimeCol    = "DateTimeOriginal",
                                       species              = "VTA",
                                       occasionLength       = 1,
                                       day1                 = "survey",
                                       datesAsOccasionNames = TRUE,
                                       includeEffort        = TRUE,
                                       scaleEffort          = FALSE,
                                       timeZone             = "Asia/Kuala_Lumpur"
)

# multi-season detection history

# load multi-season data
data(camtrapsMultiSeason)
data(recordTableSampleMultiSeason)

# multi-season camera operation matrix
camop_season <- cameraOperation(CTtable      = camtrapsMultiSeason,
                                stationCol   = "Station",
                                setupCol     = "Setup_date",
                                sessionCol   = "session",
                                retrievalCol = "Retrieval_date",
                                hasProblems  = TRUE,
                                dateFormat   = "%d/%m/%Y"
)

# multi-season detection history
DetHist_multi <- detectionHistory(recordTable            = recordTableSampleMultiSeason,
                                  camOp                  = camop_season,
                                  stationCol             = "Station",
                                  speciesCol             = "Species",
                                  species                = "VTA",
                                  occasionLength         = 10,
                                  day1                   = "station",
                                  recordDateTimeCol      = "DateTimeOriginal",
                                  includeEffort          = TRUE,
                                  scaleEffort            = FALSE,
                                  timeZone               = "UTC",
                                  unmarkedMultFrameInput = TRUE
)

# multi, unmarkedMultFrameInput = FALSE  
DetHist_multi2 <- detectionHistory(recordTable            = recordTableSampleMultiSeason,
                                   camOp                  = camop_season,
                                   stationCol             = "Station",
                                   speciesCol             = "Species",
                                   species                = "VTA",
                                   occasionLength         = 10,
                                   day1                   = "station",
                                   recordDateTimeCol      = "DateTimeOriginal",
                                   includeEffort          = TRUE,
                                   scaleEffort            = FALSE,
                                   timeZone               = "UTC",
                                   unmarkedMultFrameInput = FALSE  
)

# multi, occasionLength = 1, day1 = "survey"
DetHist_multi3 <- detectionHistory(recordTable            = recordTableSampleMultiSeason,
                                   camOp                  = camop_season,
                                   stationCol             = "Station",
                                   speciesCol             = "Species",
                                   species                = "VTA",
                                   occasionLength         = 1,
                                   day1                   = "survey",
                                   recordDateTimeCol      = "DateTimeOriginal",
                                   includeEffort          = TRUE,
                                   scaleEffort            = FALSE,
                                   timeZone               = "UTC",
                                   datesAsOccasionNames   = TRUE,
                                   unmarkedMultFrameInput = TRUE
)



# Test section

test_that("recordTable output has correct class and names", {
  expect_is(DetHist1,   "list")
  expect_is(DetHist2,   "list")
  expect_is(DetHist2a,   "list")
  expect_is(DetHist2_lub, "list")
  expect_is(DetHist_multi, "list")
  
  expect_equal(length(DetHist1), 1)
  expect_equal(length(DetHist2),   2)
  expect_equal(length(DetHist2a),   3)
  expect_equal(length(DetHist2_lub), 2)
  expect_equal(length(DetHist_multi), 2)
  
  expect_equal(names(DetHist1),   "detection_history")
  expect_equal(names(DetHist2),   c("detection_history", "effort"))
  expect_equal(names(DetHist2a),   c("detection_history", "effort",  "effort_scaling_parameters"))
  expect_equal(names(DetHist2_lub), c("detection_history", "effort"))
  expect_equal(names(DetHist_multi), c("detection_history", "effort"))
  
})

test_that("detectionHistory output has correct dimensions", {
  expect_equal(dim(DetHist1$detection_history),   c(3,7))
  expect_equal(dim(DetHist2$detection_history),   c(3,7))
  expect_equal(dim(DetHist2$effort), c(3,7))
  expect_equal(dim(DetHist_multi$detection_history), c(4,16))
  expect_equal(dim(DetHist_multi$effort),  c(4,16))
  expect_equal(dim(DetHist1day$detection_history),  c(3,44))
  expect_equal(dim(DetHist1day_survey$detection_history),  c(3,46))
  expect_equal(dim(DetHist_multi2$detection_history),  c(7,8))
  expect_equal(dim(DetHist_multi3$detection_history),  c(4,884))
})

test_that("detectionHistory output has correct row names", {
  
  expect_rownames <- c("StationA", "StationB", "StationC")
  expect_rownames_multi <- c("StationA", "StationB", "StationC", "StationD")
  expect_rownames_multi2 <- c("StationA__SESS_2009", "StationB__SESS_2009", "StationC__SESS_2009", 
                              "StationA__SESS_2010", "StationB__SESS_2010", "StationC__SESS_2010", "StationD__SESS_2010")
  
  expect_equal(rownames(DetHist1$detection_history),  expect_rownames)
  expect_equal(rownames(DetHist2$detection_history),   expect_rownames)
  expect_equal(rownames(DetHist2$effort), expect_rownames)
  expect_equal(rownames(DetHist1day$detection_history),  expect_rownames)
  expect_equal(rownames(DetHist1day_survey$detection_history),  expect_rownames)
  expect_equal(rownames(DetHist_multi$detection_history), expect_rownames_multi)
  expect_equal(rownames(DetHist_multi$effort),  expect_rownames_multi)
  expect_equal(rownames(DetHist_multi2$detection_history),  expect_rownames_multi2)
  expect_equal(rownames(DetHist_multi3$detection_history),  expect_rownames_multi)
})


test_that("occasionLength error message works", {
  expect_error(detectionHistory(recordTable          = recordTableSample,
                                camOp                = camop_no_problem,
                                stationCol           = "Station",
                                speciesCol           = "Species",
                                recordDateTimeCol    = "DateTimeOriginal",
                                species              = "VTA",
                                occasionLength       = 47,
                                day1                 = "station",
                                datesAsOccasionNames = FALSE,
                                includeEffort        = FALSE,
                                timeZone             = "Asia/Kuala_Lumpur"), 
               "occasionLength must be smaller than the total number of days in camOp")
  
  expect_error(detectionHistory(recordTable          = recordTableSample,
                                camOp                = camop_no_problem,
                                stationCol           = "Station",
                                speciesCol           = "Species",
                                recordDateTimeCol    = "DateTimeOriginal",
                                species              = "VTA",
                                occasionLength       = 0,
                                day1                 = "station",
                                datesAsOccasionNames = FALSE,
                                includeEffort        = FALSE,
                                timeZone             = "Asia/Kuala_Lumpur"), 
               "occasionLength must be a positive integer and not 0")
})






test_that("Warnings when day1 is a date and there are records before day1", {
  expect_warning(detectionHistory(recordTable          = recordTableSample,
                                  camOp                = camop_no_problem,
                                  stationCol           = "Station",
                                  speciesCol           = "Species",
                                  recordDateTimeCol    = "DateTimeOriginal",
                                  species              = "VTA",
                                  occasionLength       = 7,
                                  day1                 = "2009-05-01",
                                  datesAsOccasionNames = FALSE,
                                  includeEffort        = FALSE,
                                  timeZone             = "Asia/Kuala_Lumpur"),
                 "4 records (out of 5) were removed because they were taken before day1", fixed = TRUE)
}                             
)


test_that("Error when day1 is after camera retrieval", {
  expect_error(detectionHistory(recordTable          = recordTableSample,
                                camOp                = camop_no_problem,
                                stationCol           = "Station",
                                speciesCol           = "Species",
                                recordDateTimeCol    = "DateTimeOriginal",
                                species              = "VTA",
                                occasionLength       = 7,
                                day1                 = "2009-06-01",
                                datesAsOccasionNames = FALSE,
                                includeEffort        = FALSE,
                                timeZone             = "Asia/Kuala_Lumpur"),
               "day1 (2009-06-01) is after the last station's retrieval date (2009-05-17)", fixed = TRUE)
}                             
)


test_that("Error when day1 is before camera setup", {
  expect_error(detectionHistory(recordTable          = recordTableSample,
                                camOp                = camop_no_problem,
                                stationCol           = "Station",
                                speciesCol           = "Species",
                                recordDateTimeCol    = "DateTimeOriginal",
                                species              = "VTA",
                                occasionLength       = 7,
                                day1                 = "2009-04-01",
                                maxNumberDays        = 10,
                                datesAsOccasionNames = FALSE,
                                includeEffort        = FALSE,
                                timeZone             = "Asia/Kuala_Lumpur"),
               "day1 (2009-04-01) is before the first station's setup date (2009-04-02)", fixed = TRUE)
}                             
)


test_that("Warning when timeZone not defined", {
  expect_warning(detectionHistory(recordTable          = recordTableSample,
                                  camOp                = camop_no_problem,
                                  stationCol           = "Station",
                                  speciesCol           = "Species",
                                  recordDateTimeCol    = "DateTimeOriginal",
                                  species              = "VTA",
                                  occasionLength       = 7,
                                  day1                 = "station", 
                                  includeEffort        = FALSE),
                 "timeZone is not specified. Assuming UTC", fixed = TRUE)
}                             
)



test_that("Error when camOp is not a matrix", {
  expect_error(detectionHistory(recordTable          = recordTableSample,
                                camOp                = "camop_no_problem",
                                stationCol           = "Station",
                                speciesCol           = "Species",
                                recordDateTimeCol    = "DateTimeOriginal",
                                species              = "VTA",
                                occasionLength       = 7,
                                day1                 = "station", 
                                includeEffort        = TRUE,
                                timeZone             = "Asia/Kuala_Lumpur"),
               "camOp must be a matrix", fixed = TRUE)
}                             
)


test_that("Error when NAs/blanks in DateTimeOriginal", {
  recordTableSample_withNA <- recordTableSample_with_blank <- recordTableSample
  recordTableSample_withNA$DateTimeOriginal[1] <- NA
  recordTableSample_with_blank$DateTimeOriginal[1] <- ""
  
  expect_error(detectionHistory(recordTable          = recordTableSample_withNA,
                                camOp                = camop_no_problem,
                                stationCol           = "Station",
                                speciesCol           = "Species",
                                recordDateTimeCol    = "DateTimeOriginal",
                                species              = "PBE",
                                occasionLength       = 7,
                                day1                 = "station",
                                datesAsOccasionNames = FALSE,
                                includeEffort        = FALSE,
                                timeZone             = "Asia/Kuala_Lumpur"
  ),
               "there are NAs in subset_species[, recordDateTimeCol]", fixed = TRUE)
  
  expect_error(detectionHistory(recordTable          = recordTableSample_with_blank,
                                camOp                = camop_no_problem,
                                stationCol           = "Station",
                                speciesCol           = "Species",
                                recordDateTimeCol    = "DateTimeOriginal",
                                species              = "PBE",
                                occasionLength       = 7,
                                day1                 = "station", 
                                includeEffort        = FALSE,
                                timeZone             = "Asia/Kuala_Lumpur"),
               "there are blank values in subset_species[, recordDateTimeCol]", fixed = TRUE)
}                             
)

test_that("Error when DateTimeOriginal can't be interpreted (format wrong)", {
  # suppressWarnings is because parse_date_time() gives a warning too:  All formats failed to parse. No formats found. 
  expect_error(suppressWarnings(detectionHistory(recordTable  = recordTableSample,
                                                 camOp                = camop_no_problem,
                                                 stationCol           = "Station",
                                                 speciesCol           = "Species",
                                                 recordDateTimeCol    = "DateTimeOriginal",
                                                 recordDateTimeFormat = "mdy HMS",  # wrong format
                                                 species              = "PBE",
                                                 occasionLength       = 7,
                                                 day1                 = "station", 
                                                 includeEffort        = FALSE,
                                                 timeZone             = "Asia/Kuala_Lumpur")),
               "Cannot read datetime format in subset_species[, recordDateTimeCol]. Output is all NA.\nexpected:  mdy HMS\nactual:    2009-04-21 00:40:00", fixed = TRUE)
}                             
)





# PART 2: camera operation matrix with fraction days!

camtraps_hrs <- camtraps

camtraps_hrs$Setup_date <- paste(camtraps_hrs$Setup_date, "12")
camtraps_hrs$Retrieval_date <- paste(camtraps_hrs$Retrieval_date, "18")
camtraps_hrs$Problem1_from  <- paste(camtraps_hrs$Problem1_from, "12")
camtraps_hrs$Problem1_to  <- paste(camtraps_hrs$Problem1_to, "18")


# create camera operation matrix
camop_no_problem_frac <- cameraOperation(CTtable      = camtraps_hrs,
                                         stationCol   = "Station",
                                         setupCol     = "Setup_date",
                                         retrievalCol = "Retrieval_date",
                                         hasProblems  = FALSE,
                                         dateFormat   = "dmy H"
)

# add some records on incomplete days
recordTableSample_extra <- recordTableSample_extra2 <- recordTableSample[5,]
recordTableSample_extra$DateTimeOriginal  <- "2009-04-02 05:07:00"  # on first day / StationA
recordTableSample_extra2$DateTimeOriginal <- "2009-05-14 05:07:00"  # on last day / StationA


recordTableSample2 <- rbind(recordTableSample_extra, recordTableSample, recordTableSample_extra2)

test_that("it works when camera operation matrix uses fraction of days & there's records on those days", {

DetHist2_lub_frac <- detectionHistory(recordTable          = recordTableSample2,
                                      camOp                = camop_no_problem_frac,
                                      stationCol           = "Station",
                                      speciesCol           = "Species",
                                      recordDateTimeCol    = "DateTimeOriginal",
                                      recordDateTimeFormat = "ymd HMS",
                                      species              = "VTA",
                                      occasionLength       = 7,
                                      day1                 = "station",
                                      datesAsOccasionNames = FALSE,
                                      includeEffort        = TRUE,
                                      scaleEffort          = FALSE,
                                      timeZone             = "Asia/Kuala_Lumpur"
)
expect_type(DetHist2_lub_frac, type = "list")                                      
})

context("communityModels")
library(camtrapR)
library(terra)


# Use local() to create a self-contained environment for all related tests.

local({
  
  # --- SETUP: Define all shared data objects ONCE ---
  
  data("camtraps")
  data("recordTableSample")
  
  # create camera operation matrix
  camop_no_problem <- cameraOperation(
    CTtable      = camtraps,
    stationCol   = "Station",
    setupCol     = "Setup_date",
    retrievalCol = "Retrieval_date",
    hasProblems  = FALSE,
    dateFormat   = "dmy"
  )
  
  # make list of detection histories
  species_to_include <- unique(recordTableSample$Species)
  DetHist_list <- detectionHistory(
    recordTable         = recordTableSample,
    camOp               = camop_no_problem,
    stationCol          = "Station",
    speciesCol          = "Species",
    recordDateTimeCol   = "DateTimeOriginal",
    species             = species_to_include,
    occasionLength      = 7,
    day1                = "station",
    datesAsOccasionNames= FALSE,
    includeEffort       = TRUE,
    scaleEffort         = TRUE,
    timeZone            = "Asia/Kuala_Lumpur"
  )
  
  # create covariates
  sitecovs <- camtraps[, c(1:3)]
  sitecovs$elevation <- c(300, 500, 600)
  sitecovs[, c(2:4)] <- scale(sitecovs[, -1])
  sitecovs$some_factor <- factor(c("A", "A", "B"))
  
  sitecovs$elevation_squared <- sitecovs$elevation^2
  
  effort_categ <- ifelse(DetHist_list$effort > 0, "high", "low")
  
  # bundle input data
  data_list <- list(
    ylist = DetHist_list$detection_history,
    siteCovs = sitecovs,
    obsCovs = list(effort = DetHist_list$effort,
                   effort_categ = effort_categ)
  )
  
  # create community model
  modelfile1   <- tempfile(fileext = ".txt")
  modelfile1_RN <- tempfile(fileext = ".txt")
  
  
  mod.jags <- communityModel(
    data_list,
    occuCovs = list(fixed = c("some_factor", "utm_y"), ranef = "elevation"),
    detCovsObservation = list(fixed = "effort"),
    intercepts = list(det = "fixed", occu = "ranef"),
    modelFile = modelfile1
  )
  
  mod.jags_RN <- communityModel(
    data_list,
    model = "RN",
    occuCovs = list(fixed = c("utm_y"), ranef = "elevation"),
    detCovsObservation = list(fixed = "effort"),
    intercepts = list(det = "fixed", occu = "ranef"),
    modelFile = modelfile1_RN
  )
  
  
  # create prediction raster
  
  # 1. Create the base raster template
  template <- rast(nrows = 10, ncols = 10)
  
  # 2. Create the 'utm_y' layer
  layer_utm_y <- setValues(template, rnorm(ncell(template)))
  names(layer_utm_y) <- "utm_y"
  
  # 3. Create the 'elevation' layer
  layer_elevation <- setValues(template, rnorm(ncell(template)))
  names(layer_elevation) <- "elevation"
  
  # 4. Create the 'some_factor' (categorical) layer
  factor_values <- sample(c(1, 2), ncell(template), replace = TRUE)
  layer_factor <- setValues(template, factor_values) 
  names(layer_factor) <- "some_factor"
  
  # 5. Combine the individual layers into a single SpatRaster
  prediction_raster <- c(layer_utm_y, layer_elevation, layer_factor)
  
  
  
  # create various models for checks downstream
  
  # various effect types
  mod.jags_fixed <- communityModel(
    data_list,
    occuCovs = list(fixed = c("some_factor", "utm_y", "elevation")),
    detCovsObservation = list(fixed = c("effort", "effort_categ")),
    intercepts = list(det = "fixed", occu = "fixed"),
    modelFile = tempfile(fileext = "txt")
  )
  
  mod.jags_ranef <- communityModel(
    data_list,
    occuCovs = list(ranef = c("some_factor", "elevation")),
    detCovsObservation = list(ranef =  c("effort", "effort_categ")),
    intercepts = list(det = "ranef", occu = "ranef"),
    modelFile = tempfile(fileext = "txt")
  )
  
  mod.jags_indep <- communityModel(
    data_list,
    occuCovs = list(independent  = c("utm_y")),
    intercepts = list(det = "independent", occu = "independent"),
    modelFile = tempfile(fileext = "txt")
  )
  
  mod.jags_fixed_specsiteranef <- communityModel(
    data_list,
    speciesSiteRandomEffect = list(det = T, occu = F),
    modelFile = tempfile(fileext = "txt")
  )
  
  
  mod.jags_fixed2 <- communityModel(
    data_list,
    # occuCovs = list(fixed = c("some_factor", "utm_y", "elevation")),
    detCovsObservation = list(fixed = "effort_categ"),
    intercepts = list(det = "fixed", occu = "fixed"),
    modelFile = tempfile(fileext = "txt")
  )
  
  # quadratic effects
  mod.jags_quad1 <- communityModel(
    data_list,
    occuCovs = list(fixed = c("elevation", "elevation_squared")),
    modelFile = tempfile(fileext = "txt")
  )
  mod.jags_quad2 <- communityModel(
    data_list,
    occuCovs = list(ranef = c("elevation", "elevation_squared")),
    modelFile = tempfile(fileext = "txt")
  )
  
  # data augmentation
  
  mod.jags_aug1 <- communityModel(
    data_list,
    modelFile = tempfile(fileext = "txt"),
    augmentation = c(maxknown = 7) 
  )
  
  mod.jags_aug2 <- communityModel(
    data_list,
    modelFile = tempfile(fileext = "txt"),
    augmentation = c(full = 8) 
  )
  
  
  # --- TESTS:  ---
  
  test_that("communityModel output has correct structure for JAGS occupancy model", {
    # Objects like mod.jags are available from the parent local() environment
    expect_s4_class(mod.jags, "commOccu")
    expect_equal(slotNames(mod.jags), c("modelText", "params", "inits_fun", "data", "input", "nimble", "modelFile",
                                        "covariate_info", "model"))
    expect_false(mod.jags@nimble)
    expect_true(file.exists(mod.jags@modelFile))
    expect_equal(mod.jags@model, "Occupancy")
  })
  
  test_that("communityModel Royle-Nichols output has correct structure", {
    # Objects like mod.jags are available from the parent local() environment
    expect_s4_class(mod.jags_RN, "commOccu")
    expect_equal(slotNames(mod.jags_RN), c("modelText", "params", "inits_fun", "data", "input", "nimble", "modelFile",
                                        "covariate_info", "model"))
    expect_false(mod.jags_RN@nimble)
    expect_true(file.exists(mod.jags_RN@modelFile))
    expect_equal(mod.jags_RN@model, "RN")
  })
  
  test_that("summary works", {
    expect_output(summary(mod.jags), "=== Community Occupancy Model Summary ===")
    expect_output(summary(mod.jags_RN), "=== Community Occupancy Model Summary ===")
  })
  
  test_that("JAGS workflow (fit, plot, predict) runs correctly", {
    
    # Skips first
    skip_if_not_installed("rjags")
    skip_if(Sys.which("jags") == "", message = "JAGS executable not found. Skipping test.")

    
    # 1. Fit the model
    
    fit.jags    <- fit(mod.jags, n.iter = 100, n.burnin = 50, chains = 3)
    fit.jags_RN <- fit(mod.jags_RN, n.iter = 100, n.burnin = 50, chains = 3)
    
    
    
    # Check fit object
    expect_s3_class(fit.jags, "mcmc.list")
    expect_length(fit.jags, 3)
    
    expect_s3_class(fit.jags_RN, "mcmc.list")
    expect_length(fit.jags_RN, 3)
    
    # 2. Test plotting
    plot_eff_state       <- plot_effects(mod.jags, fit.jags, submodel = "state")
    plot_coef_state_comb <- plot_coef(mod.jags, fit.jags, submodel = "state", combine = TRUE, ordered = FALSE)
    
    expect_length(plot_eff_state, 3)
    expect_s3_class(plot_coef_state_comb, "ggplot")
    
    
    plot_eff_state_RN       <- plot_effects(mod.jags_RN, fit.jags_RN, submodel = "state")
    plot_coef_state_RN_comb <- plot_coef(mod.jags_RN, fit.jags_RN, submodel = "state", combine = TRUE, ordered = FALSE)
    
    expect_length(plot_eff_state_RN, 2)
    expect_s3_class(plot_coef_state_RN_comb, "ggplot")
    
    
    # 3. Test predictions
    draws <- 20
    
    
    p_array <- predict(object = mod.jags, mcmc.list = fit.jags, type = "p_array", draws = draws)
    psi_array <- predict(object = mod.jags, mcmc.list = fit.jags, type = "psi_array", draws = draws)
    
    rich <- predict(object = mod.jags, mcmc.list = fit.jags, type = "richness", draws = draws)
    psi <- predict(object = mod.jags, mcmc.list = fit.jags, type = "psi", draws = draws)
    pao <- predict(object = mod.jags, mcmc.list = fit.jags, type = "pao", draws = draws)
    
    rich_batch <- predict(object = mod.jags, mcmc.list = fit.jags, type = "richness", draws = draws,
                          batch = T)
    
    psi_array_rast <- predict(object = mod.jags, mcmc.list = fit.jags, type = "psi_array", draws = draws)
    
    rich_rast <- predict(object = mod.jags, mcmc.list = fit.jags, type = "richness", draws = draws, 
                         x = prediction_raster)
    psi_rast <- predict(object = mod.jags, mcmc.list = fit.jags, type = "psi", draws = draws,
                   x = prediction_raster)
    pao <- predict(object = mod.jags, mcmc.list = fit.jags, type = "pao", draws = draws,
                   x = prediction_raster)
    
    # predictions with confidence intervals
    
    
    rich_rast_conf <- predict(object = mod.jags, mcmc.list = fit.jags, type = "richness", draws = draws, 
                         x = prediction_raster, interval = "confidence")
    psi_rast_conf <- predict(object = mod.jags, mcmc.list = fit.jags, type = "psi", draws = draws,
                        x = prediction_raster, interval = "confidence")
    pao_conf <- predict(object = mod.jags, mcmc.list = fit.jags, type = "pao", draws = draws,
                   x = prediction_raster, interval = "confidence")
    
    expect_equal(dim(rich), c(3, 2))
    expect_equal(dim(psi), c(15, 4))
    
    expect_length(pao, 3)
    expect_named(pao, c("pao_summary", "pao_matrix",  "pao_df"))
    expect_equal(dim(pao$pao_summary), c(5,8))
    expect_equal(dim(pao$pao_df), c(100,3))
    expect_equal(dim(pao$pao_matrix), c(5,20))
    
    expect_true(class(rich_rast) == "SpatRaster")
    expect_equal(dim(rich_rast), c(10, 10, 2))
    expect_named(rich_rast, c("mean", "sd"))
    expect_equal(dim(rich_rast_conf), c(10, 10, 4))
    expect_named(rich_rast_conf, c("mean", "sd", "lower", "upper"))
    
    
    # 4. Test Posterior predictive checks
    ppc_comm1a <- PPC.community(p = p_array, psi = psi_array, 
                              y = mod.jags@input$ylist, 
                              model = "Occupancy", 
                              type = "FT",
                              z.cond = T)
    
    
    expect_length(ppc_comm1a, 2)
    expect_named(ppc_comm1a, c("BP", "residuals"))
    expect_equal(dim(ppc_comm1a$BP), c(6, 2))
    expect_named(ppc_comm1a$residuals, species_to_include)
    
    ppc_comm1b <- PPC.community(p = p_array, psi = psi_array, 
                                y = mod.jags@input$ylist, 
                                model = "Occupancy", 
                                type = "FT",
                                z.cond = F)
    
    
    expect_length(ppc_comm1b, 2)
    expect_named(ppc_comm1b, c("BP", "residuals"))
    expect_equal(dim(ppc_comm1b$BP), c(6, 2))
    expect_named(ppc_comm1b$residuals, species_to_include)
    
    
    # 
    ppc_comm2a <- PPC.community(p = p_array, psi = psi_array, 
                                y = mod.jags@input$ylist, 
                                model = "Occupancy", 
                                type = "PearChi2",
                                z.cond = T)
    
    
    expect_length(ppc_comm2a, 2)
    expect_named(ppc_comm2a, c("BP", "residuals"))
    expect_equal(dim(ppc_comm2a$BP), c(6, 2))
    expect_named(ppc_comm2a$residuals, species_to_include)
    
    ppc_comm2b <- PPC.community(p = p_array, psi = psi_array, 
                                y = mod.jags@input$ylist, 
                                model = "Occupancy", 
                                type = "PearChi2",
                                z.cond = F)
    
    
    expect_length(ppc_comm2b, 2)
    expect_named(ppc_comm2b, c("BP", "residuals"))
    expect_equal(dim(ppc_comm2b$BP), c(6, 2))
    expect_named(ppc_comm2b$residuals, species_to_include)
    
    

    
    # 5. Test predictions (Royle-Nichols model)
    
    p_array <- predict(object = mod.jags_RN, mcmc.list = fit.jags_RN, type = "p_array", draws = draws)
    psi_array <- predict(object = mod.jags_RN, mcmc.list = fit.jags_RN, type = "psi_array", draws = draws)
    
    rich <- predict(object = mod.jags_RN, mcmc.list = fit.jags_RN, type = "richness", draws = draws)
    psi <- predict(object = mod.jags_RN, mcmc.list = fit.jags_RN, type = "psi", draws = draws)
    abundance <- predict(object = mod.jags_RN, mcmc.list = fit.jags_RN, type = "abundance", draws = draws)
    abundance_array <- predict(object = mod.jags_RN, mcmc.list = fit.jags_RN, type = "lambda_array", draws = draws)
    pao <- predict(object = mod.jags_RN, mcmc.list = fit.jags_RN, type = "pao", draws = draws)
    
    abundance_rast <- predict(object = mod.jags_RN, mcmc.list = fit.jags_RN, type = "abundance", draws = draws,
                              x = prediction_raster)
    
    rich_conf <- predict(object = mod.jags_RN, mcmc.list = fit.jags_RN, type = "richness", draws = draws, 
                         interval = "confidence")
    abundance_conf <- predict(object = mod.jags_RN, mcmc.list = fit.jags_RN, type = "abundance", draws = draws,
                              interval = "confidence")
    
    
    expect_equal(dim(rich), c(3, 2))
    expect_equal(dim(rich_conf), c(3, 4))
    expect_equal(dim(psi), c(15, 4))
    expect_equal(dim(abundance), c(15, 4))
    expect_equal(dim(abundance_array), c(3,5,20))
    
    expect_length(pao, 3)
    expect_named(pao, c("pao_summary", "pao_matrix",  "pao_df"))
    expect_equal(dim(pao$pao_summary), c(5,8))
    expect_equal(dim(pao$pao_df), c(100,3))
    expect_equal(dim(pao$pao_matrix), c(5,20))
    
    
    # 6. Test PPC (RN model)
    
    # returns warning in each call of PPC.residualy
    ppc_comm <- expect_warning(PPC.community(p = p_array, psi = abundance_array, 
                              y = mod.jags_RN@input$ylist, model = "RN", type = "FT"))
    
    
    
    expect_length(ppc_comm, 2)
    expect_named(ppc_comm, c("BP", "residuals"))
    expect_equal(dim(ppc_comm$BP), c(6, 2))
    expect_named(ppc_comm$residuals, species_to_include)
    
    ppc_comm2 <- expect_warning(PPC.community(p = p_array, psi = abundance_array, 
                               y = mod.jags_RN@input$ylist, model = "RN", type = "PearChi2"))
    
    
  })
  
 
  test_that("Other model specifications work", {
    
   
    
    expect_equal(length(mod.jags_fixed@input$ylist), 5)
    expect_equal(length(mod.jags_aug1@input$ylist), 7)
    expect_equal(length(mod.jags_aug2@input$ylist), 8)
  })
  
  
  test_that("Nimble models work",  {
    
    skip_if_not_installed("nimble")
    skip_if_not_installed("nimbleEcology")
    
    library(nimbleEcology)
    
    # nimble models
    mod.jags_fixed_nimble <- communityModel(
      data_list,
      occuCovs = list(fixed = c("some_factor", "utm_y"), fixed = c("some_factor", "elevation")),
      detCovsObservation = list(fixed = "effort"),
      intercepts = list(det = "fixed", occu = "fixed"),
      modelFile = tempfile(fileext = "txt"),
      nimble = TRUE
    )
    
    mod.jags_ranef_nimble <- communityModel(
      data_list,
      occuCovs = list(ranef = c("some_factor", "elevation")),
      detCovsObservation = list(ranef = "effort"),
      intercepts = list(det = "ranef", occu = "ranef"),
      modelFile = tempfile(fileext = "txt"),
      nimble = TRUE
    )
    
    mod.jags_indep_nimble <- communityModel(
      data_list,
      occuCovs = list(independent  = c("utm_y")),
      intercepts = list(det = "independent", occu = "independent"),
      modelFile = tempfile(fileext = "txt"),
      nimble = TRUE
    )
    
    # check that nimble models can be fit
    
    fit.nimble1 <- fit(mod.jags_fixed_nimble, n.iter = 100, n.burnin = 50, chains = 3)
    
    expect_s3_class(fit.nimble1, "mcmc.list")
  })
  
  test_that("plot function works", {
    
    # skips first
    skip_if_not_installed("rjags")
    
    
    # prevent warning about unused variable "effort categ"
    mod.jags_fixed@data$effort_categ <- NULL
    mod.jags_ranef@data$effort_categ <- NULL
    
    # first fit the models above
    fit.mod.jags.fixed <- fit(mod.jags_fixed, n.iter = 100, n.burnin = 50, chains = 3)
    fit.mod.jags.ranef <- fit(mod.jags_ranef, n.iter = 100, n.burnin = 50, chains = 3)
    fit.mod.jags.indep <- fit(mod.jags_indep, n.iter = 100, n.burnin = 50, chains = 3)
    fit.mod.jags.quad1 <- fit(mod.jags_quad1, n.iter = 100, n.burnin = 50, chains = 3)
    fit.mod.jags.quad2 <- fit(mod.jags_quad2, n.iter = 100, n.burnin = 50, chains = 3)
    
    
    #  plot
    p1a <- plot_effects(mod.jags_fixed, 
                       fit.mod.jags.fixed, 
                       submodel = "state")
    p1b <- plot_coef(mod.jags_fixed, 
                        fit.mod.jags.fixed, 
                        submodel = "state")
    
    expect_equal(class(p1a), "list")
    expect_equal(class(p1b), "list")
    
    expect_equal(length(p1a), 3)
    expect_equal(length(p1b), 3)
    
    expect_named(p1a, c("some_factor", "utm_y", "elevation" ))
    
    expect_s3_class(p1a$utm_y, "ggplot")
    
    
    
    p2a <- plot_effects(mod.jags_ranef, 
                        fit.mod.jags.ranef, 
                        submodel = "state")
    p2b <- plot_coef(mod.jags_ranef, 
                     fit.mod.jags.ranef, 
                     submodel = "state")
    
    p3a <- plot_effects(mod.jags_indep, 
                        fit.mod.jags.indep, 
                        submodel = "state")
    p3b <- plot_coef(mod.jags_indep, 
                     fit.mod.jags.indep, 
                     submodel = "state")
    
    # p4a <- plot_effects(mod.jags_quad1,    # error currently
    #                     fit.mod.jags.quad1, 
    #                     submodel = "state")
    p4b <- plot_coef(mod.jags_quad1, 
                     fit.mod.jags.quad1, 
                     submodel = "state")
    
    p5a <- plot_effects(mod.jags_quad2, 
                        fit.mod.jags.quad2, 
                        submodel = "state")
    p5b <- plot_coef(mod.jags_quad2, 
                     fit.mod.jags.quad2, 
                     submodel = "state")
  
    
    p1b1 <- plot_coef(mod.jags_ranef, 
                      fit.mod.jags.ranef, 
                      submodel = "state", 
                      ordered = F)
    
    p1b2 <- plot_coef(mod.jags_ranef, 
                      fit.mod.jags.ranef, 
                      submodel = "state", 
                      combine = T,
                      ordered = F)
    
    p1b3 <- plot_coef(mod.jags_ranef, 
                      fit.mod.jags.ranef, 
                      submodel = "state",
                      level = c(0.9, 0.5))
    
    p1b4 <- plot_coef(mod.jags_ranef, 
                      fit.mod.jags.ranef, 
                      submodel = "state", 
                      community_lines = T)
    
    p1b5 <- plot_coef(mod.jags_ranef, 
                      fit.mod.jags.ranef, 
                      submodel = "state", 
                      community_lines = T,
                      combine = T,
                      ordered = F)
    
  })
  
  
  test_that("Other error conditions work", {
    
    expect_error(
      mod.jags_indep <- communityModel(
        data_list,
        occuCovs = list(independent  = c("some_factor", "utm_y")),
        intercepts = list(det = "independent", occu = "independent"),
        modelFile = tempfile(fileext = "txt")
      ),
      "independent effects of categorical site covariates are currently not supported"
    )
    
    expect_error(
    mod.jags_fixed_specsiteranef <- communityModel(
      data_list,
      speciesSiteRandomEffect = list(det = T, occu = T),
      modelFile = tempfile(fileext = "txt")
    ),
    fixed = "speciesSiteRandomEffect$occu must be FALSE"
    )
    
  })
  
})

#' Generate a species record table from camera trap images and videos
#' 
#' Generates a record table from camera trap images or videos. Images/videos
#' must be sorted into station directories at least. The function can read
#' species identification from a directory structure (Station/Species or
#' Station/Camera/Species) or from image metadata tags.
#' 
#' The function can handle a number of different ways of storing images, and
#' supports species identification by moving images into species directories as
#' well as metadata tagging. In every case, images need to be stored into
#' station directories. If images are identified by moving them into species
#' directories, a camera directory is optional: "Station/Species/XY.JPG" or
#' "Station/Camera/Species/XY.JPG". Likewise, if images are identified using
#' metadata tagging, a camera directory can be used optionally:
#' "Station/XY.JPG" or "Station/Camera/XY.JPG".
#' 
#' If images are identified by metadata tagging, \code{metadataSpeciesTag}
#' specifies the metadata tag group name that contains species identification
#' tags. \code{metadataHierarchyDelimitor} is "|" for images tagged in DigiKam
#' and images tagged in Adobe Bridge / Lightroom with the default settings. It
#' is only necessary to change it if the default was changed in these programs.
#' 
#' \code{minDeltaTime} is a criterion for temporal independence of species
#' recorded at the same station. Setting it to 0 will make the function return
#' all records. Setting it to a higher value will remove records that were
#' taken less than \code{minDeltaTime} minutes after the last record
#' (\code{deltaTimeComparedTo = "lastRecord"}) or the last independent record
#' (\code{deltaTimeComparedTo = "lastIndependentRecord"}).
#' 
#' \code{removeDuplicateRecords} determines whether duplicate records
#' (identical station, species, date/time, (and camera if applicable)) are
#' all returned (FALSE) or collapsed into a single unique record (TRUE).
#' 
#' \code{camerasIndependent} defines if the cameras at a station are to be
#' considered independent. If \code{TRUE}, records of the same species taken by
#' different cameras are considered independent (e.g. if they face different
#' trails). Use \code{FALSE} if both cameras face each other and possibly
#' \code{TRUE} ).
#' 
#' \code{exclude} can be used to exclude "species" directories containing
#' irrelevant images (e.g. "team", "blank", "unidentified"). \code{stationCol}
#' can be set to match the station column name in the camera trap station table
#' (see \code{\link{camtraps}}).
#' 
#' Many digital images contain Exif metadata tags such as "AmbientTemperature"
#' or "MoonPhase" that can be extracted if specified in \code{metadataTags}.
#' Because these are manufacturer-specific and not standardized, function
#' \code{\link{exifTagNames}} provides a vector of all available tag names.
#' Multiple names can be specified as a character vector as: \code{c(Tag1,
#' Tag2, ...)}. The metadata tags thus extracted may be used as covariates in
#' modelling species distributions.
#' 
#' \code{eventSummaryColumn} and \code{eventSummaryFunction} can be used to
#' extract summary statistics for independent sampling events. For example, you
#' assigned a "count" tag to your images, indicating the number of individuals
#' in a picture. In a sequence of pictures taken within 1 minute, most pictures
#' show one individual, but one image shows two individuals. You tagged the
#' images accordingly (count = 1 or count = 2) and run \code{recordTable}. Set
#' \code{eventSummaryColumn = "count"} and \code{eventSummaryFunction = "max"}
#' to obtain the maximum number of \code{count} in all images within
#' \code{minDeltaTime} minutes of a given record. The results is in a new
#' column, in this example \code{count_max}. You can also calculate several
#' statistics at the same time, by supplying vectors of values, e.g.
#' \code{eventSummaryColumn = c("count", "count", "camera")} and
#' \code{eventSummaryFunction = c("min", "max", "unique")} to get minimum and
#' maximum count and all unique camera IDs for that event. Note that
#' \code{eventSummaryColumn} and \code{eventSummaryFunction} must be of same
#' length.
#' 
#' Argument \code{video} is a named list with 2 or 4 items. 2 items
#' (\code{file_formats}, \code{dateTimeTag}) are always required, and are
#' sufficent if \code{IDfrom = "directory"}. In that case, no digiKam tags will
#' be returned.  To return digiKam tags, two additional items are required
#' (\code{db_directory}, \code{db_filename}). This is essential when using
#' \code{IDfrom = "metadata"}. When using \code{IDfrom = "directory"}, it is
#' optional, but allows to extract metadata tags assigned to videos in digiKam.
#' This workaround is necessary because digiKam tags are not written into video
#' metadata, but are only saved in the digiKam database. So in contrast to JPG
#' images, they can not be extracted with ExifTool. It also requires that
#' \code{inDir} is in your digiKam database.
#' 
#' The items of argument \code{video} are:
#' 
#' \tabular{ll}{ \code{file_formats} \tab The video formats to extract (include
#' "jpg" if you want .JPG image metadata) \cr \code{dateTimeTag} \tab the
#' metadata tag to extract date/time from (use \code{\link{exifTagNames}} to
#' find out which tag is suitable) \cr \code{db_directory} \tab The directory
#' containing digiKam database (optional if \code{IDfrom = "directory"}) \cr
#' \code{db_filename} \tab The digiKam database file in \code{db_directory}
#' (optional if \code{IDfrom = "directory"}) \cr }
#' 
#' See the examples below for for how to specify the argument \code{video}.
#' 
#' @param inDir character. Directory containing station directories. It must
#' either contain images in species subdirectories (e.g.
#' inDir/StationA/SpeciesA) or images with species metadata tags (without
#' species directories, e.g. inDir/StationA).
#' @param IDfrom character. Read species ID from image metadata ("metadata") of
#' from species directory names ("directory")?
#' @param cameraID character. Where should the function look for camera IDs:
#' 'filename', 'directory'. 'filename' requires images renamed with
#' \code{\link{imageRename}}. 'directory' requires a camera subdirectory within
#' station directories (station/camera/species). Can be missing.
#' @param camerasIndependent logical. If \code{TRUE}, species records are
#' considered to be independent between cameras at a station.
#' @param exclude character. Vector of species names to be excluded from the
#' record table
#' @param minDeltaTime integer. Time difference between records of the same
#' species at the same station to be considered independent (in minutes)
#' @param deltaTimeComparedTo character. For two records to be considered
#' independent, must the second one be at least \code{minDeltaTime} minutes
#' after the last independent record of the same species
#' (\code{"lastIndependentRecord"}), or \code{minDeltaTime} minutes after the
#' last record (\code{"lastRecord"})?
#' @param timeZone character. Must be a value returned by
#' \code{\link[base:timezones]{OlsonNames}}
#' @param stationCol character. Name of the camera trap station column.
#' Assuming "Station" if undefined.
#' @param writecsv logical. Should the record table be saved as a .csv?
#' @param outDir character. Directory to save csv to. If NULL and
#' \code{writecsv = TRUE}, recordTable will be written to \code{inDir}.
#' @param metadataHierarchyDelimitor character. The character delimiting
#' hierarchy levels in image metadata tags in field "HierarchicalSubject".
#' Either "|" or ":".
#' @param metadataSpeciesTag character. In custom image metadata, the species
#' ID tag name.
#' @param additionalMetadataTags character. Additional camera model-specific
#' metadata tags to be extracted. (If possible specify tag groups as returned
#' by \code{\link{exifTagNames}})
#' @param removeDuplicateRecords logical. If there are several records of the
#' same species at the same station (also same camera if cameraID is defined)
#' at exactly the same time, show only one?
#' @param returnFileNamesMissingTags logical. If species are assigned with
#' metadata and images are not tagged, return a few file names of these images
#' as a message?
#' @param eventSummaryColumn character. A column in the record table (e.g. from
#' a metadata tag) by to summarise non-independent records (those within
#' \code{minDeltaTime} of a given record) with a user-defined function
#' (\code{eventSummaryFunction})
#' @param eventSummaryFunction character. The function by which to summarise
#' \code{eventSummaryColumn} of non-independent records, e.g. "sum", "max"
#' (optional)
#' @param video list. Contains information on how to handle video data
#' (optional). See details.
#' 
#' @return A data frame containing species records and additional information
#' about stations, date, time and (optionally) further metadata.
#' 
#' @note The results of a number of other function will depend on the output of
#' this function (namely on the arguments \code{exclude} for excluding species
#' and \code{minDeltaTime}/ \code{deltaTimeComparedTo} for temporal
#' independence):
#' 
#' \tabular{l}{ \code{\link{detectionMaps}} \cr \code{\link{detectionHistory}}
#' \cr \code{\link{activityHistogram}} \cr \code{\link{activityDensity}} \cr
#' \code{\link{activityRadial}} \cr \code{\link{activityOverlap}} \cr
#' \code{\link{activityHistogram}} \cr \code{\link{surveyReport}} \cr }
#' @section Warning : Custom image metadata must be organised hierarchically
#' (tag group - tag; e.g. "Species" - "Leopard Cat"). Detailed information on
#' how to set up and use metadata tags can be found in
#' \href{https://CRAN.R-project.org/package=camtrapR/vignettes/camtrapr2.pdf}{vignette
#' 2: Species and Individual Identification}.
#' 
#' Custom image metadata tags must be written to the images. The function
#' cannot read tags from .xmp sidecar files. Make sure you set the preferences
#' accordingly. In DigiKam, go to Settings/Configure digiKam/Metadata. There,
#' make sure "Write to sidecar files" is unchecked.
#' 
#' Please note the section about defining argument \code{timeZone} in the
#' vignette on data extraction (accessible via
#' \code{vignette("DataExtraction")} or online
#' (\url{https://cran.r-project.org/package=camtrapR/vignettes/camtrapr3.pdf})).
#' 
#' @author Juergen Niedballa
#' 
#' @references Phil Harvey's ExifTool \url{https://exiftool.org/}
#' 
#' @examples
#' 
#' 
#' 
#' \dontrun{   # the examples take too long to pass CRAN tests
#' 
#' 
#' # set directory with camera trap images in station directories
#' wd_images_ID_species <- system.file("pictures/sample_images_species_dir", 
#'                                     package = "camtrapR")
#' 
#' if (Sys.which("exiftool") != ""){        # only run these examples if ExifTool is available
#' 
#' 
#' rec_table1 <- recordTable(inDir               = wd_images_ID_species,
#'                        IDfrom                 = "directory",
#'                        minDeltaTime           = 60,
#'                        deltaTimeComparedTo    = "lastRecord",
#'                        writecsv               = FALSE,
#'                        additionalMetadataTags = c("EXIF:Model", "EXIF:Make")
#' )
#' # note argument additionalMetadataTags: it contains tag names as returned by function exifTagNames
#' 
#' rec_table2 <- recordTable(inDir               = wd_images_ID_species,
#'                        IDfrom                 = "directory",
#'                        minDeltaTime           = 60,
#'                        deltaTimeComparedTo    = "lastRecord",
#'                        exclude                = "UNID",
#'                        writecsv               = FALSE,
#'                        timeZone               = "Asia/Kuala_Lumpur",
#'                        additionalMetadataTags = c("EXIF:Model", "EXIF:Make", "NonExistingTag"),
#'                        eventSummaryColumn     = "EXIF:Make",
#'                        eventSummaryFunction   = "unique"
#'                        )
#'                        
#' # note the warning that the last tag in "additionalMetadataTags" ("NonExistingTag") was not found
#' 
#' 
#' any(rec_table1$Species == "UNID")    # TRUE
#' any(rec_table2$Species == "UNID")    # FALSE
#' 
#' 
#' # here's how the removeDuplicateRecords argument works
#' 
#' rec_table3a <- recordTable(inDir              = wd_images_ID_species,
#'                        IDfrom                 = "directory",
#'                        minDeltaTime           = 0,
#'                        exclude                = "UNID",
#'                        timeZone               = "Asia/Kuala_Lumpur",
#'                        removeDuplicateRecords = FALSE
#' )
#' 
#' rec_table3b <- recordTable(inDir              = wd_images_ID_species,
#'                        IDfrom                 = "directory",
#'                        minDeltaTime           = 0,
#'                        exclude                = "UNID",
#'                        timeZone               = "Asia/Kuala_Lumpur",
#'                        removeDuplicateRecords = TRUE
#' )
#' 
#' 
#' anyDuplicated(rec_table3a[, c("Station", "Species", "DateTimeOriginal")])   # got duplicates
#' anyDuplicated(rec_table3b[, c("Station", "Species", "DateTimeOriginal")])   # no duplicates
#' 
#' # after removing duplicates, both are identical:
#' whichAreDuplicated <- which(duplicated(rec_table3a[,c("Station", "Species", "DateTimeOriginal")]))
#' all(rec_table3a[-whichAreDuplicated,] == rec_table3b)
#' 
#' 
#' ### extracting species IDs from metadata
#' 
#' wd_images_ID_species_tagged <- system.file("pictures/sample_images_species_tag", 
#'                                            package = "camtrapR")
#' 
#' rec_table4 <- recordTable(inDir               = wd_images_ID_species_tagged,
#'                        IDfrom                 = "metadata",
#'                        metadataSpeciesTag     = "Species",
#'                        exclude                = "unidentified")
#' 
#' 
#' ###  Including videos
#' # sample videos are not included in package
#' 
#' # with videos, IDfrom = "directory", not extracting digiKam metadata
#' 
#' rec_table4 <- recordTable(inDir  = wd_images_ID_species,
#'                           IDfrom = "directory",
#'                           video  = list(file_formats = c("jpg", "mp4"),
#'                                         dateTimeTag  = "QuickTime:CreateDate")
#' )
#' 
#' # with videos, IDfrom = "metadata", extracting digiKam metadata
#' 
#' rec_table5 <- recordTable(inDir  = wd_images_ID_species,
#'                           IDfrom = "metadata",
#'                           metadataSpeciesTag = "Species",
#'                           video  = list(file_formats = c("jpg", "mp4", "avi", "mov"),
#'                                         dateTimeTag  = "QuickTime:CreateDate",
#'                                         db_directory = "C:/Users/YourName/Pictures",
#'                                         db_filename = "digikam4.db")
#' )
#' 
#' } else {  
#' # show function output if ExifTool is not available
#' message("ExifTool is not available. Cannot test function. Loading recordTableSample instead")
#' data(recordTableSample)
#' }
#' 
#' }
#' 
#' @export recordTable
#' 
recordTable <- function(inDir,
                        IDfrom,
                        cameraID,
                        camerasIndependent,
                        exclude,
                        minDeltaTime = 0,
                        deltaTimeComparedTo,
                        timeZone,
                        stationCol,
                        writecsv = FALSE,
                        outDir,
                        metadataHierarchyDelimitor = "|",
                        metadataSpeciesTag,
                        additionalMetadataTags,
                        removeDuplicateRecords = TRUE,
                        returnFileNamesMissingTags = FALSE,
                        eventSummaryColumn,
                        eventSummaryFunction,
                        video
)
{

  wd0 <- getwd()
  on.exit(setwd(wd0))

  if(!hasArg(stationCol)) stationCol <- "Station"
  if(!is.character(stationCol)) stop("stationCol must be of class 'character'")
  checkForSpacesInColumnNames(stationCol = stationCol)
  
  speciesCol <- "Species"

  if(!is.character(IDfrom)) stop("IDfrom must be of class 'character'")
  IDfrom <- match.arg(IDfrom, choices = c("metadata", "directory"))  

 if(IDfrom == "metadata"){
    metadataHierarchyDelimitor <- match.arg(metadataHierarchyDelimitor, choices = c("|", ":"))
    if(!hasArg(metadataSpeciesTag))       stop("'metadataSpeciesTag' must be defined if IDfrom = 'metadata'")
    if(!is.character(metadataSpeciesTag)) stop("metadataSpeciesTag must be of class 'character'")
    if(length(metadataSpeciesTag) != 1)   stop("metadataSpeciesTag must be of length 1")
  }

  multiple_tag_separator <- "_&_"

  # check input
  if(!hasArg(timeZone)) {
    message("timeZone is not specified. Assuming UTC")
    timeZone <- "UTC"
  }
  if(!is.element(timeZone , OlsonNames())){
    stop("timeZone must be an element of OlsonNames()", call. = FALSE)
  }
  if(Sys.which("exiftool") == "") stop("cannot find ExifTool", call. = FALSE)

  if(hasArg(cameraID)){
    if(!is.character(cameraID))         stop("cameraID must be of class 'character'", call. = FALSE)
    cameraID <- match.arg(cameraID, choices = c("filename", "directory"))
    if(!hasArg(camerasIndependent))     stop("camerasIndependent is not defined. It must be defined if cameraID is defined", call. = FALSE)
    if(!is.logical(camerasIndependent)) stop("camerasIndependent must be of class 'logical'", call. = FALSE)
  } else { 
    camerasIndependent <- FALSE
  }

  cameraCol <- "Camera"


  if(hasArg(outDir)){
    if(!is.character(outDir))         stop("outDir must be of class 'character'", call. = FALSE)
    if(isFALSE(file.exists(outDir)))  stop("outDir does not exist", call. = FALSE)
  }

  if(hasArg(exclude)){
    if(!is.character(exclude)) stop("exclude must be of class 'character'", call. = FALSE)
  }

  if(!is.logical(removeDuplicateRecords))     stop("'removeDuplicateRecords' must be logical (TRUE / FALSE)", call. = FALSE)
  if(!is.logical(returnFileNamesMissingTags)) stop("'returnFileNamesMissingTags' must be logical (TRUE / FALSE)", call. = FALSE)


  metadata.tagname <- "HierarchicalSubject"    # for extracting metadata assigned in tagging software

  if(hasArg(additionalMetadataTags)){
    if(!is.character(additionalMetadataTags)) stop("additionalMetadataTags must be of class 'character'", call. = FALSE)
    if(any(grep(pattern = " ", x = additionalMetadataTags, fixed = TRUE))) stop("In argument additionalMetadataTags, spaces are not allowed")
    if("HierarchicalSubject" %in% additionalMetadataTags & IDfrom == "metadata")  {
      message("'HierarchicalSubject' may not be in 'additionalMetadataTags' if IDfrom = 'metadata'. It will be ignored because the function returns it anyway.", call. = FALSE)
      additionalMetadataTags <- additionalMetadataTags[-grep(pattern = "HierarchicalSubject", x = additionalMetadataTags)]  # remove HierarchicalSubject from additionalMetadataTags
    }
  }

  minDeltaTime <- as.integer(minDeltaTime)
  if(!is.integer(minDeltaTime)) stop("'minDeltaTime' must be an integer", call. = FALSE)

  if(minDeltaTime != 0){
    if(isFALSE(removeDuplicateRecords)){
      warning("minDeltaTime is > 0. Therefore, removeDuplicateRecords was set to TRUE (otherwise there may be records taken at the same time)", call. = FALSE, immediate. = TRUE)
      removeDuplicateRecords <- TRUE
    }
    
    deltaTimeComparedTo < match.arg(deltaTimeComparedTo, choices = c("lastRecord", "lastIndependentRecord"))
    
    if(!hasArg(deltaTimeComparedTo)) {
      stop(paste("minDeltaTime is not 0. deltaTimeComparedTo must be defined"), call. = FALSE)
    }
  } else {
    if(hasArg(deltaTimeComparedTo)) {
      warning(paste("minDeltaTime is 0. deltaTimeComparedTo = '", deltaTimeComparedTo, "' will have no effect", sep = ""), call. = FALSE, immediate. = TRUE)
    } else {
      deltaTimeComparedTo <- "lastRecord"
    }
  }

  if(!is.logical(writecsv))  stop("writecsv must be logical (TRUE or FALSE)", call. = FALSE)
  if(!is.character(inDir))   stop("inDir must be of class 'character'", call. = FALSE)
  if(length(inDir) != 1)     stop("inDir may only consist of 1 element only", call. = FALSE)
  if(!dir.exists(inDir))     stop("Could not find inDir:\n", inDir, call. = FALSE)

  if(hasArg(eventSummaryColumn)) {
    if(!is.character(eventSummaryColumn))     stop("eventSummaryColumn must be of class 'character'", call. = FALSE)
    if(!is.character(eventSummaryFunction))   stop("eventSummaryFunction must be of class 'character'", call. = FALSE)
  }
  
  # find image directories
  dirs       <- list.dirs(inDir, full.names = TRUE, recursive = FALSE)
  dirs_short <- list.dirs(inDir, full.names = FALSE, recursive = FALSE)
  
  if(length(dirs) == 0) stop("inDir contains no station directories", call. = FALSE)
  max_nchar_station <- max(nchar(dirs_short))
  
  # Ignore the ".dtrash" folder if present (on Mac)
  if(".dtrash" %in% dirs_short) {
    # message("Ignoring .dtrash folder.")
    dirs <- dirs[dirs_short != ".dtrash"]
    dirs_short <- dirs_short[dirs_short != ".dtrash"]
  }
  
  # Ignore  the ".mysql.digikam" folder if present
  if(".mysql.digikam" %in% dirs_short) {
    # message("Ignoring .mysql.digikam folder.")
    dirs <- dirs[dirs_short != ".mysql.digikam"]
    dirs_short <- dirs_short[dirs_short != ".mysql.digikam"]
  }
  
  
  # process video argument (if present)
  if(hasArg(video)){
    video_out <- processVideoArgument(IDfrom = IDfrom,
                                      video  = video)
    digiKam_data <- video_out$digiKam_data
    file_formats <- video_out$file_formats
    # if(isFALSE("jpg" %in% file_formats)) file_formats <- c(file_formats, "jpg")
  } else {
    file_formats <- "jpg"   # jpg, as the default, if video not requested
  }
  
  # empty list for metadata output
  record.table.list <- list()
  
  # create command line calls
  command.tmp  <- paste('exiftool -q -f -t -r -Directory -FileName -EXIF:DateTimeOriginal', 
                        ifelse(hasArg(video), 
                               paste(" -", video$dateTimeTag, sep = ""), 
                               ""),    # if video requested, video date time tag
                        ' -HierarchicalSubject',
                        ifelse(hasArg(additionalMetadataTags), paste(" -",additionalMetadataTags,  collapse = "", sep = ""), ""),
                        paste(" -ext", file_formats, collapse = " ", sep = " "),    # requested file extensions
                        ' "', dirs, '"', sep = "")
  
  # construct column names for metadata table
  colnames.tmp <- c("Directory", "FileName", "DateTimeOriginal")
  if(hasArg(video)) colnames.tmp <- c(colnames.tmp, video$dateTimeTag)
  colnames.tmp <- c(colnames.tmp, "HierarchicalSubject")
  if(hasArg(additionalMetadataTags)) colnames.tmp <- c(colnames.tmp, additionalMetadataTags [nchar(additionalMetadataTags) >= 2]) 
  # only add as new column if entry of additionalMetadataTag has more than 2 characters (to allow users to specify exiftool commands, e.g. -L)
  
  
  for(i in 1:length(dirs)){   # loop through station directories

    # execute exiftool
    metadata.tmp <- runExiftool(command.tmp = command.tmp[i], colnames.tmp = colnames.tmp)

    if(is.null(class(metadata.tmp))){            # omit station if no images found

      length.tmp <- length(list.files(dirs[i], pattern = paste(".", file_formats, "$", collapse = "|", sep = ""), 
                                      ignore.case = TRUE, recursive = TRUE))
      message(paste(formatC(dirs_short[i], width = max_nchar_station, flag = "-"),  ":  ",
                    formatC(length.tmp, width = 5), " files      Skipping", sep = ""))
      warning(paste(dirs_short[i],  ":  contains no files of interest and was omitted\n"), call. = FALSE,  immediate. = FALSE)
    } else {

      # if video files extracted, add DateTimeOriginal and HierarchicalSubject
      if(hasArg(video)){
        metadata.tmp <- addVideoDateTimeOriginal(metadata.tmp = metadata.tmp, video = video)
        
        # add HierachicalSubject for video files
        if(!is.null(digiKam_data)){
          digiKamVideoMetadata <- digiKamVideoHierarchicalSubject(stationDir = dirs[i],
                                                                    digiKamTablesList = digiKam_data,    # output of accessDigiKamDatabase
                                                                    videoFormat = file_formats[!grepl(file_formats, pattern = "jpg")])
          # add HierarchialSubject for video files (match by filename and path)
          metadata.tmp <- addVideoHierarchicalSubject (metadata.tmp = metadata.tmp,
                                                       video = video,
                                                       digiKamVideoMetadata = digiKamVideoMetadata,
                                                       digiKamTablesList = digiKam_data,
                                                       videoFormat = file_formats[!grepl(file_formats, pattern = "jpg")])
        }
      }
      
      # check presence / consistency of DateTimeOriginal column, go to next station or remove records if necessary
      metadata.tmp <- checkDateTimeOriginal (intable    = metadata.tmp,
                                             dirs_short = dirs_short,
                                             i          = i)
      if(is.null(metadata.tmp)) next

      
      # now split HierarchicalSubject tags and add as columns to table
      metadata.tmp <- addMetadataAsColumns (intable                    = metadata.tmp,
                                            metadata.tagname           = metadata.tagname,
                                            metadataHierarchyDelimitor = metadataHierarchyDelimitor,
                                            multiple_tag_separator     = multiple_tag_separator)

      # add species names to metadata table (from folders or metadata, otherwise NA)
      metadata.tmp <- assignSpeciesID(intable                    = metadata.tmp,
                                      IDfrom                     = IDfrom,
                                      metadataSpeciesTag         = metadataSpeciesTag,
                                      speciesCol                 = speciesCol,
                                      dirs_short                 = dirs_short,
                                      i_tmp                      = i,
                                      multiple_tag_separator     = multiple_tag_separator,
                                      returnFileNamesMissingTags = returnFileNamesMissingTags)
      
      # if images in station contain no metadata species tags, skip that station
      if(!is.data.frame(metadata.tmp)){
        if(metadata.tmp == "found no species tag") {
          # warning(paste(dirs_short[i], ":   metadataSpeciesTag '", metadataSpeciesTag, "' not found in image metadata tag 'HierarchicalSubject'. Skipping", sep = ""), call. = FALSE, immediate. = TRUE)
        } else {
          warning(paste(dirs_short[i], ":   error in species tag extraction. Skipping. Please report this bug", sep = ""), call. = FALSE, immediate. = TRUE)
        }
        next
      }

      # remove empty metadata columns (if HierarchicalSubject is all empty or if additionalMetadataTags were not found)
      empty_cols <- which(apply(metadata.tmp, MARGIN = 2, FUN = function(X){all(X == "-")}))
      if(length(empty_cols) >= 1){
        metadata.tmp <-  metadata.tmp[, -empty_cols]
      }

      # add station and camera id to metadata table
      arg.list0 <- list(intable = metadata.tmp, dirs_short = dirs_short, stationCol = stationCol, hasStationFolders = TRUE, cameraCol = cameraCol, i = i, IDfrom = IDfrom)  # assumes station directories

      if(!hasArg(cameraID)) metadata.tmp <- do.call(addStationCameraID, arg.list0)
      if( hasArg(cameraID)) metadata.tmp <- do.call(addStationCameraID, c(arg.list0, cameraID = cameraID))

      # remove species in argument "excluded"
      if(hasArg (exclude)){
        if(any(tolower(metadata.tmp[,speciesCol]) %in% tolower(exclude))) {  # if there is anything to remove
          metadata.tmp <- metadata.tmp[-which(tolower(metadata.tmp[,speciesCol]) %in% tolower(exclude)),]
        }
      }

      if(nrow(metadata.tmp) >= 1){   # if anything left after excluding species, do

        # convert character vector extracted from images to time object and format for outfilename
        metadata.tmp$DateTimeOriginal <- as.POSIXct(strptime(x = metadata.tmp$DateTimeOriginal, format = "%Y:%m:%d %H:%M:%S", tz = timeZone))

        # sort by station, (camera), species and time
        if(camerasIndependent == TRUE) {
          if(all(is.na(metadata.tmp[, cameraCol]))) stop(paste("Failed to extract camera ID from", cameraID))
          metadata.tmp <- metadata.tmp[order(metadata.tmp[,stationCol], metadata.tmp[,speciesCol], metadata.tmp[,cameraCol], metadata.tmp$DateTimeOriginal),]
        } else {
          metadata.tmp <- metadata.tmp[order(metadata.tmp[,stationCol], metadata.tmp[,speciesCol], metadata.tmp$DateTimeOriginal),]
        }



        #remove duplicate records of same species taken in same second at the same station (by the same camera, if relevant)
        metadata.tmp2 <- removeDuplicatesOfRecords(metadata.tmp          = metadata.tmp,
                                                  removeDuplicateRecords = removeDuplicateRecords,
                                                  camerasIndependent     = camerasIndependent,
                                                  stationCol             = stationCol,
                                                  speciesCol             = speciesCol,
                                                  cameraCol              = cameraCol,
                                                  current                = i, 
                                                  total                  = length(dirs),
                                                  max_nchar_station      = max_nchar_station)


        # assess independence between records and calculate time differences
        args.assessTemporalIndependence <- list(intable             = metadata.tmp2,
                                                deltaTimeComparedTo = deltaTimeComparedTo,
                                                columnOfInterest    = speciesCol,
                                                cameraCol           = cameraCol,
                                                camerasIndependent  = camerasIndependent,
                                                minDeltaTime        = minDeltaTime,
                                                stationCol          = stationCol)
        
        if(hasArg(eventSummaryColumn)) {
          args.assessTemporalIndependence <- c(args.assessTemporalIndependence,
                                               list(eventSummaryColumn   = eventSummaryColumn,
                                                    eventSummaryFunction = eventSummaryFunction))
        }
        
        d1 <- do.call(assessTemporalIndependence, args = args.assessTemporalIndependence)

        record.table.list[[i]] <- d1

        suppressWarnings(rm(d1))
      }  # end      if(nrow(metadata.tmp) >= 1){} else {...}   # i.e. not all species were excluded
    }    # end      if(nrow(metadata.tmp) == 0){} else {...}   # i.e. directory i contained images
  }      # end      for(i in 1:length(dirs)){   # loop through station directories

  
  # combine all data frames from list into one data frame
  record.table <- as.data.frame(data.table::rbindlist(record.table.list, fill = TRUE, use.names = TRUE))
  
  if(nrow(record.table) == 0){
    stop(paste("something went wrong. I looked through all those", length(dirs)  ,"folders and now your table is empty. Were date/time information unreadable? Could species tags not be extracted (if IDfrom = 'metadata')? Or did you exclude too many species?"), call. = FALSE)
  }

  # rearrange table, add date and time as separate columns. add additional column names as needed.

  record.table2  <-  data.frame(record.table[,c(stationCol, speciesCol, "DateTimeOriginal")],
                                Date = as.Date (record.table$DateTimeOriginal, format = "%Y/%M/%d", tz = timeZone),
                                Time = strftime(record.table$DateTimeOriginal, format = "%H:%M:%S", tz = timeZone),
                                record.table[,c("delta.time.secs", "delta.time.mins", "delta.time.hours", "delta.time.days",
                                                "Directory", "FileName")])

  metadata_columns <- which(!colnames(record.table) %in% colnames(record.table2))

  # add metadata columns
  if(length(metadata_columns) >= 1){
    record.table3 <- cbind(record.table2, record.table[,metadata_columns])
    colnames(record.table3)[(ncol(record.table2) + 1) : ncol(record.table3)] <- colnames(record.table)[metadata_columns]
  } else {record.table3 <- record.table2}


  # add camera column (if present)
  if(hasArg(cameraID)){
    record.table3 <- data.frame(record.table3[,stationCol],
                                record.table[,cameraCol],
                                record.table3[,-which(colnames(record.table3) %in% c(stationCol, cameraCol))])
    colnames(record.table3)[1] <- stationCol
    colnames(record.table3)[2] <- cameraCol
  }

  rownames(record.table3) <- NULL


  # warning if additionalMetadataTags were not found
  if(hasArg(additionalMetadataTags)){
    #whichAdditionalMetadataTagsFound <- which(gsub(additionalMetadataTags, pattern = ":", replacement = ".") %in% colnames(record.table3))   # replace : in additionalMetadataTags (if specifying tag groups) with . as found in column names
    whichAdditionalMetadataTagsFound <- which(additionalMetadataTags %in% colnames(record.table3))   # replace : in additionalMetadataTags (if specifying tag groups) with . as found in column names
    if(length(whichAdditionalMetadataTagsFound) < length(additionalMetadataTags)){
      if(length(whichAdditionalMetadataTagsFound) == 0) {  # if none of the additionalMetadataTags was found
        warning("metadata tag(s)  not found in image metadata:  ", paste(additionalMetadataTags, collapse = ", "), call. = FALSE)
        } else {                                                            # if only some of the additionalMetadataTags was found
        warning("metadata tag(s)  not found in image metadata:  ", paste(additionalMetadataTags[-whichAdditionalMetadataTagsFound], collapse = ", "), call. = FALSE)
      }
    }
  }

  # make column "HierarchicalSubject" the last column
  col_to_move <- which(colnames(record.table3) %in% metadata.tagname)
  if(length(col_to_move) >= 1){
     record.table3 <- cbind(record.table3, record.table3[,col_to_move])
	  record.table3 <- record.table3[,-col_to_move]
    colnames(record.table3)[ncol(record.table3)] <- metadata.tagname
  }

  # convert to data.frame, in order to get all the column names right (: becomes .) 
  # NOTE: doesn't work unless data.frame is called. : in column names means they must be adressed as e.g. record.table3$`IPTC:Keywords`
  record.table3 <- data.frame(record.table3, stringsAsFactors = FALSE, check.names = TRUE)
  
  # save table
  if(writecsv == TRUE){
    outtable_filename <- paste("record_table_", minDeltaTime, "min_deltaT_", Sys.Date(), ".csv", sep = "")
    message("saving csv to  ", file.path(inDir, outtable_filename))
    if(hasArg(outDir) == FALSE){
      setwd(inDir)
    } else {
      setwd(outDir)
    }
  write.csv(record.table3, file = outtable_filename)
  }
  return(record.table3)
}

context("recordTable")
library(camtrapR)

# run recordTable

wd_images_ID_species <- system.file("pictures/sample_images_species_dir", package = "camtrapR")

rec_table0 <- recordTable(inDir               = wd_images_ID_species,
                          IDfrom                 = "directory",
                          timeZone = "UTC")


rec_table1 <- recordTable(inDir               = wd_images_ID_species,
                          IDfrom                 = "directory",
                          minDeltaTime           = 60,
                          deltaTimeComparedTo    = "lastRecord",
                          writecsv               = FALSE,
                          additionalMetadataTags = c("EXIF:Model", "EXIF:Make"),
                          timeZone               = "Asia/Kuala_Lumpur"
)


# with additionalMetadataTags, event Summary and exclude

rec_table2_1 <- recordTable(inDir               = wd_images_ID_species,
                            IDfrom                 = "directory",
                            minDeltaTime           = 60,
                            deltaTimeComparedTo    = "lastRecord",
                            exclude                = "UNID",
                            writecsv               = FALSE,
                            timeZone               = "Asia/Kuala_Lumpur",
                            additionalMetadataTags = c("EXIF:Model", "EXIF:Make"),
                            eventSummaryColumn     = "EXIF:Make",
                            eventSummaryFunction   = "unique"
)

# with additionalMetadataTags and multiple event summaries

rec_table2_2 <- recordTable(inDir               = wd_images_ID_species,
                            IDfrom                 = "directory",
                            minDeltaTime           = 60,
                            deltaTimeComparedTo    = "lastRecord",
                            exclude                = "UNID",
                            writecsv               = FALSE,
                            timeZone               = "Asia/Kuala_Lumpur",
                            additionalMetadataTags = c("EXIF:Model", "EXIF:Make"),
                            eventSummaryColumn     = c("EXIF:Make", "EXIF:Model"),
                            eventSummaryFunction   = c("unique", "unique")
)


# with removeDuplicateRecords

rec_table3a <- recordTable(inDir              = wd_images_ID_species,
                           IDfrom                 = "directory",
                           minDeltaTime           = 0,
                           exclude                = "UNID",
                           timeZone               = "Asia/Kuala_Lumpur",
                           removeDuplicateRecords = FALSE
)

rec_table3b <- recordTable(inDir              = wd_images_ID_species,
                           IDfrom                 = "directory",
                           minDeltaTime           = 0,
                           exclude                = "UNID",
                           timeZone               = "Asia/Kuala_Lumpur",
                           removeDuplicateRecords = TRUE
)


rec_table4 <- recordTable(inDir                  = wd_images_ID_species,
                          IDfrom                 = "directory",
                          timeZone               = "UTC",
                          video = list(file_formats = c("jpg", "mp4"), 
                                       dateTimeTag = "QuickTime:CreateDate"
                                       )
)

# Test section
  
test_that("recordTable output has correct class", {
  expect_is(rec_table0,   "data.frame")
  expect_is(rec_table1,   "data.frame")
  expect_is(rec_table2_1, "data.frame")
  expect_is(rec_table2_2, "data.frame")
  expect_is(rec_table3a,  "data.frame")
  expect_is(rec_table3b,  "data.frame")
  expect_is(rec_table4,  "data.frame")
})

test_that("recordTable output has correct dimensions", {
  expect_equal(dim(rec_table0),   c(56,12))
  expect_equal(dim(rec_table1),   c(40,14))
  expect_equal(dim(rec_table2_1), c(39,15))
  expect_equal(dim(rec_table2_2), c(39,16))
  expect_equal(dim(rec_table3a),  c(67,12))
  expect_equal(dim(rec_table3b),  c(55,12))
  expect_equal(dim(rec_table4),   c(57,12))
})

test_that("removeDuplicateRecords works", {
  expect_equal(anyDuplicated(rec_table3a[, c("Station", "Species", "DateTimeOriginal")]), 12)   # 12 duplicates
  expect_equal(anyDuplicated(rec_table3b[, c("Station", "Species", "DateTimeOriginal")]), 0)    # 0 duplicates
})

test_that("errors are correct", {
  expect_error(recordTable(inDir               = wd_images_ID_species,
                           IDfrom                 = "Directory",
                           timeZone               = "Asia/Kuala_Lumpur"), 
               "'arg' should be one of")
  expect_error(recordTable(inDir               = wd_images_ID_species,
                           IDfrom              = "directory",
                           cameraID            = "directory",
                           timeZone               = "Asia/Kuala_Lumpur"), 
               "camerasIndependent is not defined. It must be defined if cameraID is defined")
  expect_error(recordTable(inDir               = wd_images_ID_species,
                           IDfrom              = "directory",
                           cameraID            = "Filename",
                           camerasIndependent  = TRUE,
                           timeZone               = "Asia/Kuala_Lumpur"), 
               "'arg' should be one of")
})

test_that("time zone message works", {
  expect_message(recordTable(inDir     = wd_images_ID_species,
                             IDfrom    = "directory"),
                 "Warning: timeZone is not specified. Assuming UTC
StationA:      8 images    0 duplicates removed      |=======             |   33%
StationB:     23 images    6 duplicates removed      |=============       |   67%
StationC:     37 images    6 duplicates removed      |====================|  100%")
})


test_that("warnings are correct", {
  
  expect_warning(recordTable(inDir               = wd_images_ID_species,
                             IDfrom                 = "directory",
                             additionalMetadataTags = c("EXIF:Model", "NonExistingTag"),
                             timeZone               = "Asia/Kuala_Lumpur"),
                 " not found in image metadata:  NonExistingTag")
  
})

context("filterRecordTable")

# Load necessary library
library(testthat)

# --- Test Setup ---
# This setup code prepares the data needed for the tests.
# It will be run inside the test blocks to ensure independence.

# Define the path to the sample images
wd_images_ID_species <- system.file("pictures/sample_images_species_dir",
                                    package = "camtrapR")

# --- Test Suite ---

testthat::describe("Core Filtering Logic", {
  
  # Skip all tests in this describe block if ExifTool is not available
  skip_if(Sys.which("exiftool") == "", message = "ExifTool not found, skipping tests.")
  
  # Create the initial, unfiltered record table once for this block
  rec_table <- recordTable(inDir       = wd_images_ID_species,
                           IDfrom      = "directory",
                           minDeltaTime= 0,
                           exclude     = "UNID",
                           timeZone    = "Asia/Kuala_Lumpur")
  
  test_that("filtering with 'lastIndependentRecord' matches the example", {
    # 1. Verify the initial state matches the example
    expect_equal(nrow(rec_table), 55)
    
    # 2. Apply the filtering from the example
    rec_table_filt <- filterRecordTable(recordTable         = rec_table,
                                        minDeltaTime        = 60,
                                        deltaTimeComparedTo = "lastIndependentRecord")
    
    # 3. Verify the final state matches the example
    expect_equal(nrow(rec_table_filt), 39)
    
    # 4. Perform additional checks on the output
    expect_s3_class(rec_table_filt, "data.frame")
    expect_true(all(names(rec_table_filt) %in% names(rec_table))) # Columns should be identical
  })
  
  test_that("filtering with 'lastRecord' is stricter", {
    # 'lastRecord' compares against every previous record, not just independent ones.
    # This should result in fewer (or equal) records than 'lastIndependentRecord'.
    rec_table_filt_last <- filterRecordTable(recordTable         = rec_table,
                                             minDeltaTime        = 60,
                                             deltaTimeComparedTo = "lastRecord")
    
    # For this specific dataset, the result is also 39
    expect_equal(nrow(rec_table_filt_last), 39)
    
    # In general, it must be less than or equal to the other method
    rec_table_filt_independent <- filterRecordTable(recordTable = rec_table,
                                                    minDeltaTime = 60,
                                                    deltaTimeComparedTo = "lastIndependentRecord")
    expect_lte(nrow(rec_table_filt_last), nrow(rec_table_filt_independent))
  })
})


testthat::describe("Edge Cases", {
  
  skip_if(Sys.which("exiftool") == "", message = "ExifTool not found, skipping tests.")
  
  rec_table <- recordTable(inDir = wd_images_ID_species, IDfrom = "directory", minDeltaTime = 0)
  
  test_that("minDeltaTime = 0 returns all records", {
    # A delta time of 0 means no filtering should occur.
    rec_table_filt_zero <- filterRecordTable(recordTable  = rec_table,
                                             minDeltaTime = 0)
    
    expect_equal(nrow(rec_table_filt_zero), nrow(rec_table))
  })
  
  test_that("a very large minDeltaTime returns one record per station/species", {
    # A huge time gap should result in only the very first record for each
    # combination of Station and Species being kept.
    rec_table_filt_large <- filterRecordTable(recordTable  = rec_table,
                                              minDeltaTime = 999999,
                                              deltaTimeComparedTo = "lastRecord")
    
    # Calculate the expected number of rows: one for each unique combo
    expected_rows <- nrow(unique(rec_table[, c("Station", "Species")]))
    
    expect_equal(nrow(rec_table_filt_large), expected_rows)
  })
})


testthat::describe("Input Validation", {
  
  # This describe block does NOT need ExifTool, as it tests inputs, not logic.
  # We create a minimal dummy data frame for these tests.
  dummy_rec_table <- data.frame(
    Station = "S1",
    Species = "A",
    DateTimeOriginal = "2023-01-01 12:00:00"
  )
  
  test_that("it stops for invalid inputs", {
    
    
    # Wrong class for recordTable
    expect_error(
      filterRecordTable(recordTable = as.matrix(dummy_rec_table),
                        minDeltaTime = 60,
                        deltaTimeComparedTo = "lastRecord"),
      regexp = "recordTable must be a data frame"
    )
    
    # Negative minDeltaTime
    expect_error(
      filterRecordTable(recordTable = dummy_rec_table, 
                        minDeltaTime = -10,
                        deltaTimeComparedTo = "lastRecord", 
                        timeZone = "UTC"),
      regexp = "'minDeltaTime' must be a positive number"
    )
    
    # Invalid stationCol
    expect_error(
      filterRecordTable(recordTable = dummy_rec_table, 
                        minDeltaTime = 60, 
                        deltaTimeComparedTo = "lastRecord", 
                        stationCol = "WRONG_COL"),
      regexp = "stationCol is not a column name in recordTable"
    )
    
    # Invalid deltaTimeComparedTo
    expect_error(
      filterRecordTable(recordTable = dummy_rec_table, 
                        minDeltaTime = 60, 
                        deltaTimeComparedTo = "invalid_option",
                        timeZone = "UTC"),
      regexp = "'arg' should be one of"
    )
  })
})

globalVariables(c(
  "PAO", "Observed", "Predicted", "Records", "Species", "Stations",
  "covariateValue", "lower", "upper", "median",
  "modelEffects",
  "generateFormula",
  "export_table",
  "plot_effect_det",
  "plot_effect_occu",
  "generate_spatial_predictions",
  "create_species_detections_sf",
  "all_models",
  "p",
  ".data",
  
  # speciesAccum:
  "survey_date",  "survey_day",  "survey_day_by_station", "station_setup_date"
  
  
))

#' Species detection histories for occupancy analyses
#' 
#' This function generates species detection histories that can be used in
#' single-species occupancy analyses with packages \link[unmarked:unmarked-package]{unmarked} 
#' and \pkg{ubms}, as well as multi-species/community occupancy models via 
#' \code{\link{communityModel}}. It generates detection histories in different formats, 
#' with adjustable occasion length and occasion start time.
#'  
#' 
#' The function creates species detection matrices in two possible formats:
#' detection-by-date or detection-by-occasion. The start of detection histories 
#' is controlled by \code{day1}:
#' \itemize{
#'   \item \code{"station"}: Each station's history begins on its setup day
#'   \item \code{"survey"}: All stations begin on the first day of the survey
#'   \item A specific date (e.g., \code{"2015-12-31"}): All stations begin on this date
#' }
#' Dates must be in "YYYY-MM-DD" format if specified directly.
#' 
#' 
#' Two output formats are available via the \code{output} parameter:
#' \itemize{
#'   \item \code{"binary"}: Records detection (1) or non-detection (0)
#'   \item \code{"count"}: Records the number of detections per occasion
#' }
#' 
#' 
#' The \code{includeEffort} parameter determines how camera operation affects the output:
#' \itemize{
#'   \item If \code{FALSE}: Periods when cameras were not operational or only 
#'   partly operational appear as NA in the detection history. This may lose 
#'   species record from incomplete occasions.
#'   \item If \code{TRUE}: Incomplete occasions are retained. Outputs contain a 
#'   separate effort matrix that can be used as an observation covariate in occupancy models.
#' }
#' It is generally advisable to include effort as a covariate to account for 
#' uneven sampling effort.
#' 
#' 
#' \code{occasionLength} controls how many days are aggregated into each sampling occasion.
#' Note that \code{occasionStartTime} has moved to \code{\link{cameraOperation}} to ensure
#' proper calculation of daily effort.
#' 
#' The values of \code{stationCol} in \code{recordTable} must be matched by the
#' row names of \code{camOp} (case-insensitive), otherwise an error is raised.
#' 
#' For date/time formatting, \code{recordDateTimeFormat} accepts two syntax styles:
#' \itemize{
#'   \item Base R style (using \%): e.g., "\%Y-\%m-\%d \%H:\%M:\%S"
#'   \item lubridate style: e.g., "ymd HMS"
#' }
#' \pkg{lubridate} will be used if there are no "\%" characters in
#' \code{recordDateTimeFormat}. The default and recommended format is 
#' "YYYY-MM-DD HH:MM:SS" (e.g., "2014-09-30 22:59:59").
#' 
#' 
#' For multi-season studies where \code{sessionCol} was used in \code{\link{cameraOperation}},
#' the function automatically detects this structure. Set \code{unmarkedMultFrameInput = TRUE}
#' to format output for \code{unmarkedMultFrame}, with rows representing sites and columns
#' ordered by season-major, occasion-minor (e.g., season1-occasion1, season1-occasion2, etc.).
#' 
#' 
#' @param recordTable data.frame. the record table created by
#' \code{\link{recordTable}}
#' @param species character. species name(s) for which to compute detection histories. Can 
#' be either a single species name (for use with \pkg{unmarked}/\pkg{ubms}) or a vector 
#' of multiple species names (for input to \code{\link{communityModel}})
#' @param camOp The camera operation matrix as created by
#' \code{\link{cameraOperation}}
#' @param output character. Return binary detections ("binary") or counts of
#' detections ("count")
#' @param stationCol character. name of the column specifying Station ID in
#' \code{recordTable}
#' @param speciesCol character. name of the column specifying species in
#' \code{recordTable}
#' @param recordDateTimeCol character. name of the column specifying date and
#' time in \code{recordTable}
#' @param recordDateTimeFormat character. Format of column
#' \code{recordDateTimeCol} in \code{recordTable}
#' @param occasionLength integer. occasion length in days
#' @param minActiveDaysPerOccasion integer. minimum number of active trap days
#' for occasions to be included (optional)
#' @param maxNumberDays integer. maximum number of trap days per station
#' (optional)
#' @param day1 character. When should occasions begin: station setup date
#' ("station"), first day of survey ("survey"), a specific date (e.g.
#' "2015-12-31")?
#' @param buffer integer. Makes the first occasion begin a number of days after
#' station setup. (optional)
#' @param includeEffort logical. Compute trapping effort (number of active
#' camera trap days per station and occasion)?
#' @param scaleEffort logical. scale and center effort matrix to mean = 0 and
#' sd = 1?
#' @param occasionStartTime (DEPRECATED) integer. time of day (the full hour)
#' at which to begin occasions. Please use argument \code{occasionStartTime} in
#' \code{\link{cameraOperation}} instead.
#' @param datesAsOccasionNames If \code{day1 = "survey"}, occasion names in the
#' detection history will be composed of first and last day of that occasion.
#' @param timeZone character. Must be a value returned by
#' \code{\link[base:timezones]{OlsonNames}}
#' @param writecsv logical. Should the detection history be saved as a .csv?
#' @param outDir character. Directory into which detection history .csv file is
#' saved
#' @param unmarkedMultFrameInput logical. Return input for multi-season
#' occupancy models in unmarked (argument "y" in
#' \code{\link[unmarked]{unmarkedMultFrame}}?
#' 
#' @return If a single species is provided (typical for \pkg{unmarked}/\pkg{ubms} analyses), 
#' returns a list with either 1, 2 or 3 elements depending on the value of \code{includeEffort} 
#' and \code{scaleEffort}:
#' \item{detection_history}{A species detection matrix}
#' \item{effort}{A matrix giving the number of active camera trap days per station and 
#' occasion (= camera trapping effort). Only returned if \code{includeEffort = TRUE}}
#' \item{effort_scaling_parameters}{Scaling parameters of the effort matrix. Only 
#' returned if \code{includeEffort} and \code{scaleEffort} are \code{TRUE}}
#'
#' If multiple species are provided (for use with \code{\link{communityModel}}), returns 
#' a similar list structure but with \code{detection_history} containing a named list of 
#' detection matrices, one for each species. The effort matrix is identical for all 
#' species and thus returned only once.
#' 
#' 
#' @section Warning: Setting \code{output = "count"} returns a count of
#' detections, not individuals. These counts are not suitable for abundance modeling
#' (e.g., N-mixture models) as they do not represent individual animals.
#' 
#' For important information about the \code{timeZone} parameter, please refer to
#' the "Data Extraction" vignette (\code{vignette("DataExtraction")} or online at
#' \url{https://cran.r-project.org/package=camtrapR/vignettes/camtrapr3.pdf}).
#' 
#' @author Juergen Niedballa
#' 
#' @examples
#' 
#' 
#' # define image directory
#' wd_images_ID <- system.file("pictures/sample_images_species_dir", package = "camtrapR")
#' 
#' # load station information
#' data(camtraps)
#' 
#' # create camera operation matrix
#' camop_no_problem <- cameraOperation(CTtable      = camtraps,
#'                                     stationCol   = "Station",
#'                                     setupCol     = "Setup_date",
#'                                     retrievalCol = "Retrieval_date",
#'                                     hasProblems  = FALSE,
#'                                     dateFormat   = "dmy"
#' )
#' 
#' \dontrun{
#' if (Sys.which("exiftool") != ""){        # only run this function if ExifTool is available
#' recordTableSample <- recordTable(inDir               = wd_images_ID,
#'                                  IDfrom              = "directory",
#'                                  minDeltaTime        = 60,
#'                                  deltaTimeComparedTo = "lastRecord",
#'                                  exclude             = "UNID",
#'                                  timeZone            = "Asia/Kuala_Lumpur"
#' )
#' }
#' }
#' data(recordTableSample)    # load the record history, as created above
#' 
#' 
#' # compute detection history for a species
#' 
#' # without trapping effort
#' DetHist1 <- detectionHistory(recordTable         = recordTableSample,
#'                             camOp                = camop_no_problem,
#'                             stationCol           = "Station",
#'                             speciesCol           = "Species",
#'                             recordDateTimeCol    = "DateTimeOriginal",
#'                             species              = "VTA",
#'                             occasionLength       = 7,
#'                             day1                 = "station",
#'                             datesAsOccasionNames = FALSE,
#'                             includeEffort        = FALSE,
#'                             timeZone             = "Asia/Kuala_Lumpur"
#' )
#' 
#' DetHist1                     # this is a list with 1 element
#' DetHist1$detection_history   # this is the contained detection/non-detection matrix
#' 
#' 
#' # with effort / using base R to define recordDateTimeFormat
#' DetHist2 <- detectionHistory(recordTable          = recordTableSample,
#'                              camOp                = camop_no_problem,
#'                              stationCol           = "Station",
#'                              speciesCol           = "Species",
#'                              recordDateTimeCol    = "DateTimeOriginal",
#'                              species              = "VTA",
#'                              occasionLength       = 7,
#'                              day1                 = "station",
#'                              datesAsOccasionNames = FALSE,
#'                              includeEffort        = TRUE,
#'                              scaleEffort          = FALSE,
#'                              timeZone             = "Asia/Kuala_Lumpur"
#' )
#' 
#' DetHist2$detection_history  # detection history  (alternatively, use: DetHist2[[1]])
#' DetHist2$effort             # effort (alternatively, use: DetHist2[[2]])
#' 
#' # with effort / using lubridate package to define recordDateTimeFormat
#' DetHist2_lub <- detectionHistory(recordTable          = recordTableSample,
#'                                  camOp                = camop_no_problem,
#'                                  stationCol           = "Station",
#'                                  speciesCol           = "Species",
#'                                  recordDateTimeCol    = "DateTimeOriginal",
#'                                  recordDateTimeFormat = "ymd HMS",
#'                                  species              = "VTA",
#'                                  occasionLength       = 7,
#'                                  day1                 = "station",
#'                                  datesAsOccasionNames = FALSE,
#'                                  includeEffort        = TRUE,
#'                                  scaleEffort          = FALSE,
#'                                  timeZone             = "Asia/Kuala_Lumpur"
#' )    
#' 
#' DetHist2_lub$detection_history  # detection history  (alternatively, use: DetHist2_lub[[1]])
#' DetHist2_lub$effort             # effort (alternatively, use: DetHist2_lub[[2]])
#' 
#' 
#' # multi-season detection history
#' 
#' # load multi-season data
#' data(camtrapsMultiSeason)
#' data(recordTableSampleMultiSeason)
#' 
#' # multi-season camera operation matrix
#' camop_season <- cameraOperation(CTtable          = camtrapsMultiSeason,
#'                                     stationCol   = "Station",
#'                                     setupCol     = "Setup_date",
#'                                     sessionCol   = "session",
#'                                     retrievalCol = "Retrieval_date",
#'                                     hasProblems  = TRUE,
#'                                     dateFormat   = "dmy"
#' )
#' 
#' # multi-season detection history
#' DetHist_multi_season <- detectionHistory(recordTable      = recordTableSampleMultiSeason,
#'                             camOp                  = camop_season,
#'                             stationCol             = "Station",
#'                             speciesCol             = "Species",
#'                             species                = "VTA",
#'                             occasionLength         = 10,
#'                             day1                   = "station",
#'                             recordDateTimeCol      = "DateTimeOriginal",
#'                             includeEffort          = TRUE,
#'                             scaleEffort            = FALSE,
#'                             timeZone               = "UTC",
#'                             unmarkedMultFrameInput = TRUE
#' )
#' 
#' DetHist_multi_season
#' 
#' 
#' # Multi-species example for community occupancy analysis with communityModel()
#' DetHist_multi_species <- detectionHistory(recordTable = recordTableSample,
#'                                  species = c("VTA", "PBE", "EGY"),   # multiple species
#'                                  camOp = camop_no_problem,
#'                                  stationCol = "Station",
#'                                  speciesCol = "Species",
#'                                  recordDateTimeCol = "DateTimeOriginal",
#'                                  occasionLength = 7,
#'                                  day1 = "station",
#'                                  includeEffort = TRUE,
#'                                  scaleEffort = FALSE,
#'                                  timeZone = "Asia/Kuala_Lumpur"
#' )
#' 
#' 
#' # bundle input data for communityModel
#' data_list <- list(ylist = DetHist_multi_species$detection_history,
#'                   siteCovs = camtraps,
#'                   obsCovs = list(effort = DetHist_multi_species$effort))
#' 
#' 
#' \dontrun{
#' 
#' # create community model
#' mod.jags <- communityModel(data_list,
#'                            ...) # model specification 
#' }
#' 
#' @export detectionHistory
#' 
detectionHistory <- function(recordTable,
                             species,
                             camOp,
                             output = c("binary", "count"),
                             stationCol = "Station",
                             speciesCol = "Species",
                             recordDateTimeCol = "DateTimeOriginal",
                             recordDateTimeFormat = "ymd HMS",
                             occasionLength,
                             minActiveDaysPerOccasion,
                             maxNumberDays,
                             day1 = "survey",
                             buffer,
                             includeEffort = TRUE,
                             scaleEffort = FALSE,
                             occasionStartTime = "deprecated",
                             datesAsOccasionNames = FALSE,
                             timeZone,
                             writecsv = FALSE,
                             outDir,
                             unmarkedMultFrameInput)
{
  wd0 <- getwd()
  on.exit(setwd(wd0))
  
  
  # Check if species is vector
  is_multispecies <- length(species) > 1
  
  
  # single-species case ----
  if(!is_multispecies) { 
    
    
  #################
  # check input
  
  # check column names
  checkForSpacesInColumnNames(stationCol = stationCol, speciesCol = speciesCol, recordDateTimeCol = recordDateTimeCol)
  
  recordTable <- dataFrameTibbleCheck(df = recordTable)

  if(!stationCol %in% colnames(recordTable))  stop(paste('stationCol = "', stationCol, '" is not a column name in recordTable', sep = ''), call. = FALSE)
  if(!speciesCol %in% colnames(recordTable))  stop(paste('speciesCol = "', speciesCol, '" is not a column name in recordTable', sep = ''), call. = FALSE)
  if(!recordDateTimeCol %in% colnames(recordTable))  stop(paste('recordDateTimeCol = "', recordDateTimeCol,  '" is not a column name in recordTable', sep = ''), call. = FALSE)
  
  output <- match.arg (output)
  
  if(!hasArg(species)) stop("'species' is not defined", call. = FALSE)
  if(!is.character(species))  stop("species must be of class character", call. = FALSE)
  # if(!length(species) == 1)   stop("species may only contain one value", call. = FALSE)
  
  if(!hasArg(occasionLength)) stop("'occasionLength' is not defined", call. = FALSE)
  
  if(!hasArg(recordTable))         stop("'recordTable' is not defined", call. = FALSE)
  if(!is.data.frame(recordTable))  stop("'recordTable' must be a data.frame", call. = FALSE)
  if(!hasArg(camOp))               stop("'camOp' is not defined", call. = FALSE)
  if(!inherits(camOp, "matrix"))   stop ("camOp must be a matrix", call. = FALSE)
  
  if(!length(stationCol) == 1)     stop("stationCol may only contain one value", call. = FALSE)
  recordTable[,stationCol] <- as.character(recordTable[,stationCol])
  if(!is.character(stationCol))    stop("stationCol must be a character", call. = FALSE)
  
  if(!length(speciesCol) == 1)    stop("speciesCol may only contain one value", call. = FALSE)
  recordTable[,speciesCol] <- as.character(recordTable[,speciesCol])
  if(!is.character(speciesCol))   stop("speciesCol must be a character", call. = FALSE)
  
  if(!length(recordDateTimeCol) == 1)  stop("recordDateTimeCol may only contain one value", call. = FALSE)
  recordTable[,recordDateTimeCol] <- as.character(recordTable[,recordDateTimeCol])   # make character to get rid of attributes. Will later assign time zone again
  if(!is.character(recordDateTimeCol)) stop("recordDateTimeCol must be a character", call. = FALSE)
  
  
  if(hasArg(timeZone) == FALSE) {
    warning("timeZone is not specified. Assuming UTC", call. = FALSE)
    timeZone <- "UTC"
  }
  if(!is.element(timeZone , OlsonNames())){
    stop("timeZone must be an element of OlsonNames()")
  }
  if(!is.logical(writecsv)) stop("writecsv must be logical (TRUE / FALSE)", call. = FALSE)
  
  if(hasArg(occasionStartTime)) warning("'occasionStartTime' is deprecated and will be removed in the next release. It is now found in cameraOperation()")
  # if(length(occasionStartTime) != 1) stop("occasionStartTime must have length 1")
  # occasionStartTime <- as.integer(round(occasionStartTime))
  # if(occasionStartTime != 0 & !is.integer(occasionStartTime)) {stop ("occasionStartTime must be between 0 and 23", call. = FALSE)}
  # if(occasionStartTime < 0 | occasionStartTime >= 24){         stop ("occasionStartTime must be between 0 and 23", call. = FALSE)}
  
  occasionLength <- as.integer(round(occasionLength))
  if(length(occasionLength) != 1)  stop("occasionLength may only contain one value", call. = FALSE)
  if(!is.numeric(occasionLength))  stop("occasionLength must be a positive integer", call. = FALSE)
  if(occasionLength <= 0)          stop("occasionLength must be a positive integer and not 0", call. = FALSE)
  if(occasionLength > ncol(camOp)) stop("occasionLength must be smaller than the total number of days in camOp", call. = FALSE)
  
  
  if(hasArg(maxNumberDays)){
    maxNumberDays <- as.integer(maxNumberDays)
    if(maxNumberDays > ncol(camOp))    stop("maxNumberDays is larger than the number of columns of camOp", call. = FALSE)
    if(maxNumberDays < occasionLength) stop("maxNumberDays must be larger than or equal to occasionLength", call. = FALSE)
  }
  
  if(hasArg(buffer)) {
    if(!is.numeric(buffer)) stop("buffer must be number", call. = FALSE)
    buffer <- round(buffer)
    if(!buffer >= 1)        stop("if buffer is defined, it must be 1 or higher", call. = FALSE)
  }
  
  
  if(!speciesCol %in% colnames(recordTable))        stop(paste("speciesCol", speciesCol, "is not a column name in recordTable"), call. = FALSE)
  if(!recordDateTimeCol %in% colnames(recordTable)) stop(paste("recordDateTimeCol", recordDateTimeCol, "is not a column name in recordTable"), call. = FALSE)
  if(!stationCol %in% colnames(recordTable))        stop(paste("stationCol", stationCol, "is not a column name in recordTable"), call. = FALSE)
  
  
  if(!species %in% recordTable[,speciesCol]) stop("species ", species, " not found in speciesCol of recordTable", call. = FALSE)
  
  if(writecsv == TRUE){
    if(!file.exists(outDir)){stop("outDir does not exist", call. = FALSE)}
  }
  
  if(includeEffort){
    # if(!hasArg(scaleEffort))     stop("scaleEffort must be defined if includeEffort is TRUE", call. = FALSE)
    if(!is.logical(scaleEffort)) stop("scaleEffort must be logical (TRUE or FALSE)", call. = FALSE)
  } else {scaleEffort <- FALSE}
  
  if(hasArg(minActiveDaysPerOccasion)){
    if(!is.numeric(minActiveDaysPerOccasion))       stop("minActiveDaysPerOccasion must be a number", call. = FALSE)
    if(!minActiveDaysPerOccasion <= occasionLength) stop("minActiveDaysPerOccasion must be smaller than or equal to occasionLength", call. = FALSE)
  }
  
  #############
  # bring date, time, station ids into shape
  
  subset_species           <- recordTable[recordTable[,speciesCol] == species,]
  subset_species$DateTime2 <- parseDateTimeObject(inputColumn = subset_species[,recordDateTimeCol],
                                                  dateTimeFormat = recordDateTimeFormat,
                                                  timeZone = timeZone, 
                                                  checkNA_out = FALSE)
  
  if(any(is.na(subset_species$DateTime2))) stop(paste(sum(is.na(subset_species$DateTime2)), "out of",
                                                      nrow(subset_species),
                                                      "entries in recordDateTimeCol of recordTable could not be interpreted using recordDateTimeFormat (NA). row",
                                                      paste(rownames(subset_species)[which(is.na(subset_species$DateTime2))], collapse = ", ")))
  
  
  # check consistency of argument day1
  if(!is.character(day1)) stop("day1 must be a character", call. = FALSE)
  day1 <- tolower(day1)
  if(day1 == "survey") {day1switch <- 1} else {
    if(day1 == "station") {day1switch <- 2} else {
      try(date.test <- as.Date(day1), silent = TRUE)
      if(!exists("date.test"))       stop("day1 is not specified correctly. It can only be 'station', 'survey', or a date formatted as 'YYYY-MM-DD', e.g. '2016-12-31'")
      if(!inherits(date.test, "Date")) stop('could not interpret argument day1: can only be "station", "survey" or a specific date (e.g. "2015-12-31")')
      if(hasArg(buffer))             stop("if buffer is defined, day1 can only be 'survey' or 'station'")
      suppressWarnings(rm(date.test))
      day1switch <- 3
    }
  }
  
  # check that column names can be interpreted as day, as save
  camOp <- checkCamOpColumnNames (cameraOperationMatrix = camOp)
  # check if there's conflict in occasionStartTime. Error if so. 
  if(hasArg(occasionStartTime)){
    if(occasionStartTime != attributes(camOp)$occasionStartTime){
     stop(paste("occasionStartTime", occasionStartTime, "differs from occasionStartTime in camOp", attributes(camOp)$occasionStartTime)) 
    }
  }
  occasionStartTime <- attributes(camOp)$occasionStartTime
  
  # extract station / camera / session information from row names of camera operation matrix
  camop.info.df <- deparseCamOpRownames(camOp)
  
  # get information about station / session / camera IDs
  if("session" %in% colnames(camop.info.df)){
    
    if("camera" %in% colnames(camop.info.df)) stop("currently detectionHistory does not support camera-specific camera operation matrices when session/seasons are defined. Please run cameraOperation with byCamera = FALSE.", call. = FALSE)
    
    if(!is.logical(unmarkedMultFrameInput)) stop("unmarkedMultFrameInput must be logical (TRUE / FALSE)", call. = FALSE)
    
    # assign session IDs to record table (this will also change station ID to station__session)
    subset_species <- assignSessionIDtoRecordTable(recordTable_tmp = subset_species,
                                                    camOp = camOp,
                                                    dateTimeCol = "DateTime2",  #recordDateTimeCol,
                                                    stationCol = stationCol,
                                                    sessionCol = "session")
  } else {
    if(hasArg(unmarkedMultFrameInput)) warning("'unmarkedMultFrameInput' is defined, but I cannot find session IDs in the rownames of camOp. Check the row names format and maybe run camtrapR:::deparseCamOpRownames(camOp) to see if it can be interpreted", call. = FALSE)
  }
  
  cam.op.worked0 <- as.matrix(camOp)
  
  if(all(as.character(unique(subset_species[,stationCol])) %in% rownames(cam.op.worked0)) == FALSE){
  #if(all(as.character(unique(subset_species[,stationCol])) %in% camop.info.df$station) == FALSE){
    (stop("Not all values of stationCol in recordTable are matched by rownames of camOp"))
  }
  
  ################################################
  # compute date range of stations and records
  arg.list0 <- list(cam.op                = cam.op.worked0, 
                    subset_species_tmp    = subset_species, 
                    stationCol_tmp        = stationCol, 
                    day1_tmp              = day1, 
                    occasionStartTime_tmp = occasionStartTime, 
                    timeZone_tmp          = timeZone)
  
  if(hasArg(maxNumberDays))  arg.list0 <- c(arg.list0, maxNumberDays_tmp = maxNumberDays)
  if(hasArg(buffer))         arg.list0 <- c(arg.list0, buffer_tmp =  buffer)
  
  date_ranges <- do.call(createDateRangeTable, arg.list0)
  
  rm(arg.list0)
  
  #######################
  # adjust camera operation matrix
  
  cam.op.worked <- adjustCameraOperationMatrix(cam.op       = cam.op.worked0, 
                                               date_ranges2 = date_ranges, 
                                               timeZone_tmp = timeZone, 
                                               day1_2       = day1)
  
  # append occasionStartTime (if != 0) to column names for output table
  if(occasionStartTime != 0){
    colnames(cam.op.worked) <- paste(colnames(cam.op.worked), "+", occasionStartTime, "h", sep = "")
  }
  
  ######################
  # calculate trapping effort by station and occasion
  arg.list0 <- list(cam.op          = cam.op.worked, 
                    occasionLength2 = occasionLength, 
                    scaleEffort2    = scaleEffort, 
                    includeEffort2  = includeEffort,
                    occasionStartTime2 = occasionStartTime)
  
  if(hasArg(minActiveDaysPerOccasion))  arg.list0 <- c(arg.list0, minActiveDaysPerOccasion2 = minActiveDaysPerOccasion)
  
  effort.tmp <- do.call(calculateTrappingEffort, arg.list0)
  
  rm(arg.list0)
  
  effort <- effort.tmp[[1]]
  if(isTRUE(scaleEffort))  scale.eff.tmp.attr <- effort.tmp[[2]]
  
  ###################
  # remove records that fall into buffer period or were taken after maxNumberDays
  
  subset_species <- cleanSubsetSpecies(subset_species2 = subset_species, 
                                       stationCol2     = stationCol, 
                                       date_ranges2    = date_ranges)
  
  ############
  #  define the 1st day of the effective survey period.
  if(day1 %in% c("survey")){
    time2 <- date_ranges$start_first_occasion_survey[match(subset_species[,stationCol], rownames(date_ranges))]
  } else {
    time2 <- date_ranges$start_first_occasion[match(subset_species[,stationCol], rownames(date_ranges))]
  }
  
  # calculate the occasion each record belongs into from the time difference between records and beginning of the first occasion
  subset_species$occasion <- as.numeric(ceiling((difftime(time1  = subset_species$DateTime2,
                                                          time2  =  time2,
                                                          units  = "secs",
                                                          tz     = timeZone)
                                                 / (occasionLength * 86400))))
  
  # check if occasions are valid
  if(max(subset_species$occasion) > ncol(effort)) stop("Occasions exceeding total number of occasions calculated. This is a bug. I'm Sorry. Please report it.", call. = FALSE)
  if(any(subset_species$occasion == 0)) {
    # this is old code that works on POSIXlt objects, not POSIXct. Need to convert first
    subset_species$DateTime2_POSIXlt <- as.POSIXlt(subset_species$DateTime2)
    
    # identify records with time 00:00:00
    idx_records_without_time <- which(subset_species$DateTime2_POSIXlt$hour == 0 &
                                        subset_species$DateTime2_POSIXlt$min  == 0 &
                                        subset_species$DateTime2_POSIXlt$sec  == 0)
    
    # add 1 second and recalculate occasion
    if(length(idx_records_without_time) >= 1) { 
      warning(paste(length(idx_records_without_time), " records out of ", nrow(subset_species),
                    " seem to have no time information. Please provide time also. I will assume time of these records was occasionStartTime + 1 second (",
              occasionStartTime, ":00:01), but please treat this as experimental.", sep = ""), call. = FALSE)
      subset_species$DateTime2_POSIXlt$sec  <- 1
      subset_species$DateTime2_POSIXlt$hour <- occasionStartTime
      subset_species$DateTime2 <- as.POSIXct(subset_species$DateTime2_POSIXlt)
      # recalculate occasions with new date/time
      subset_species$occasion <- as.numeric(ceiling((difftime(time1  = subset_species$DateTime2,
                                                              time2  =  time2,
                                                              units  = "secs",
                                                              tz     = timeZone)
                                                     / (occasionLength * 86400))))
      if(any(subset_species$occasion == 0)) stop("I tried to fix the date-only recordDateTimeCol, but there is still occasions with value 0. Please provide date and time in recordDateTimeCol.", call. = FALSE)
    } else {
      stop("Occasion 0 calculated for at least one record. If recordDateTimeCol contains date and time (it seems they do), this is likely a bug. I am sorry. Please report it.", call. = FALSE)}
  }
  if(min(subset_species$occasion) < 0) {stop("Negative occasions calculated for at least one record. This is a bug. I'm Sorry. Please report it.", call. = FALSE)}
  
  ############
  # make detection history
  
  record.hist <- effort                     # start with effort (including NAs)
  record.hist <- ifelse(!is.na(record.hist), 0, record.hist)     # set all cells 0 (unless they are NA)
  rownames(record.hist) <- rownames(cam.op.worked)
  
  # identify occasions during which there were records (and how many)
  # binary: values = occasion
  # count: values = number of records, names = occasion
  if(output == "binary")  occasions.by.station <- tapply(X = subset_species$occasion, INDEX = subset_species[,stationCol], FUN = unique, simplify = FALSE)
  if(output == "count")   occasions.by.station <- tapply(X = subset_species$occasion, INDEX = subset_species[,stationCol], FUN = table, simplify = FALSE)
  
  
  
 # if(occasionLength == 1){                    # occasion length = 1
    # record.hist <- cam.op.worked
    # record.hist <- ifelse(record.hist == 0, NA, record.hist)    # if cameras not operational, set NA
    # record.hist <- ifelse(record.hist >= 1, 0,  record.hist)    # if cameras operational, set to 0
    # 
    # # identify occasions during which there were records (and how many)
    # if(output == "binary")  occasions.by.station <- tapply(X = subset_species$occasion, INDEX = subset_species[,stationCol], FUN = unique, simplify = FALSE)
    # if(output == "count")   occasions.by.station <- tapply(X = subset_species$occasion, INDEX = subset_species[,stationCol], FUN = table, simplify = FALSE)
    # 
    # fill detection matrix with 1 (or count) in appropriate cells
    for(xyz in which(sapply(occasions.by.station, FUN = function(x){!is.null(x)}))){
      
      if(output == "binary"){
        if(any(occasions.by.station[[xyz]] < 0)) stop("this is a bug in the function (DH_error1.binary). Please report it.", call. = FALSE)
        if(any(occasions.by.station[[xyz]] > ncol(record.hist))) stop("this is a bug in the function (DH_error2.binary). Please report it.", call. = FALSE)
        record.hist[match(names(occasions.by.station)[xyz], rownames(record.hist)), 
                    occasions.by.station[[xyz]]] <- 1
      } 
      if(output == "count") {
        if(any(as.numeric(names(occasions.by.station[[xyz]])) < 0)) stop("this is a bug in the function (DH_error1.count). Please report it.", call. = FALSE)
        if(any(as.numeric(names(occasions.by.station[[xyz]])) > ncol(record.hist))) stop("this is a bug in the function (DH_error2.count). Please report it.", call. = FALSE)
        record.hist[match(names(occasions.by.station)[xyz], rownames(record.hist)), 
                    as.numeric(names(occasions.by.station[[xyz]]))] <- occasions.by.station[[xyz]]
      } 
    }
    

   # ensure that occasions with NA or 0 effort are NA in detectionHistory
  if(occasionLength == 1){
    record.hist[is.na(cam.op.worked)] <- NA   # remove the records that were taken when cams were NA (redundant with above:   # remove records taken after day1 + maxNumberDays)
    if(isFALSE(scaleEffort)) record.hist[cam.op.worked == 0] <- NA
    
  } else {
    record.hist[is.na(effort)] <- NA     # just to make sure NA stays NA
    if(isFALSE(scaleEffort)) record.hist[effort == 0] <- NA
  }
    
   # rm(occasions.by.station, xyz)
    
#  } else {                                    # occasion length > 1
    # record.hist <- effort                     # start with effort (including NAs)
    # record.hist <- ifelse(!is.na(record.hist), 0, record.hist)     # set all cells 0 (unless they are NA)
    # rownames(record.hist) <- rownames(cam.op.worked)
    # 
    # # identify occasions during which there were records (and how many)
    # # binary: values = occasion
    # # count: values = number of records, names = occasion
    # if(output == "binary")  occasions.by.station <- tapply(X = subset_species$occasion, INDEX = subset_species[,stationCol], FUN = unique, simplify = FALSE)
    # if(output == "count")   occasions.by.station <- tapply(X = subset_species$occasion, INDEX = subset_species[,stationCol], FUN = table, simplify = FALSE)
    # 
    
    # fill detection matrix with "1" (or count) in appropriate cells
    # for(xyz in which(sapply(occasions.by.station, FUN = function(x){!is.null(x)}))){
    #   if(output == "binary")  record.hist[match(names(occasions.by.station)[xyz], rownames(record.hist)), 
    #                                       occasions.by.station[[xyz]]] <- 1
    #   if(output == "count")   record.hist[match(names(occasions.by.station)[xyz], rownames(record.hist)), 
    #                                       as.numeric(names(occasions.by.station[[xyz]]))] <- occasions.by.station[[xyz]]
    # }
    # record.hist[is.na(effort)] <- NA     # just to make sure NA stays NA
 # }
  
  # assign row names to output
  row.names(record.hist) <- row.names(effort) <- rownames(cam.op.worked)
  
  
  # assign column names
  if(isTRUE(datesAsOccasionNames)){
    
    seq.tmp <- seq(from = 1, by = occasionLength, length.out = ncol(record.hist))
    if(occasionStartTime != 0){
      colnames.tmp <- paste(colnames(cam.op.worked)[seq.tmp],
                            colnames(cam.op.worked)[seq.tmp + occasionLength], sep = "_")
    } else {
      colnames.tmp <- paste(colnames(cam.op.worked)[seq.tmp],
                            colnames(cam.op.worked)[seq.tmp + occasionLength - 1], sep = "_")
    }
    # make sure the last occasion ends on the last day
    
    if(day1 == "station"){
      colnames.tmp[length(colnames.tmp)] <- paste(colnames(cam.op.worked)[max(seq.tmp)],
                                                  colnames(cam.op.worked)[ncol(cam.op.worked)],
                                                  sep = "_")
    } else {   # if day1 = "survey" or date
      colnames.tmp[length(colnames.tmp)] <- paste(colnames(cam.op.worked)[max(seq.tmp)],
                                                  max(as.Date(date_ranges$end_last_occasion)),
                                                  sep = "_")
    }
    
    colnames(record.hist) <- colnames(effort) <- colnames.tmp
    
  }  else {
    colnames(record.hist) <- colnames(effort) <-  paste("o", seq(1,ncol(record.hist), by = 1), sep = "")
  }
  
  
  
  ## if multi-season object, make input for unmarkedMultFrame if desired
  if("session" %in% colnames(camop.info.df)){
    if(unmarkedMultFrameInput){
      
      record.hist.season.list <- split(as.data.frame(record.hist), f = camop.info.df$session)
      effort.season.list      <- split(as.data.frame(effort),      f = camop.info.df$session)
      
      all_stations <- unique(camop.info.df$station)
      
      # add missing stations for individual seasons
      record.hist.season.list2 <- lapply(record.hist.season.list, FUN = function(x){
        deparse_tmp <- deparseCamOpRownames(x)
          x2 <- matrix(NA, ncol = ncol(x), nrow = length(all_stations))
          rownames(x2) <- all_stations
          colnames(x2) <- colnames(x)
          x2[match(deparse_tmp$station, all_stations), ] <- as.matrix(x)
          x2
      })
      
      effort.season.list2 <- lapply(effort.season.list, FUN = function(x){
        deparse_tmp <- deparseCamOpRownames(x)
        x2 <- matrix(NA, ncol = ncol(x), nrow = length(all_stations))
        rownames(x2) <- all_stations
        colnames(x2) <- colnames(x)
        x2[match(deparse_tmp$station, all_stations), ] <- as.matrix(x)
        x2
      })
      
      # # get detection history and effort into separate lists
      # detHist_season_list_det    <- lapply(detHist_season_list, FUN = function(x) {x$detection_history})
      # detHist_season_list_effort <- lapply(detHist_season_list, FUN = function(x) {x$effort})
      
      
      # add NA columns to make sure all detection histories / effort matrices have same number of occasions (= identical number of columns)
      ncol_by_season <- sapply(record.hist.season.list, FUN = ncol)
      
      #detHist_season_list_det    <- lapply(detHist_season_list, FUN = function(x) {x$detection_history})
      #detHist_season_list_effort <- lapply(detHist_season_list, FUN = function(x) {x$effort})
      
      detHist_season_list_det_padded    <- lapply(record.hist.season.list2, 
                                                  FUN = padMatrixWithNA, 
                                                  ncol_desired = max(ncol_by_season))
      detHist_season_list_effort_padded <- lapply(effort.season.list2, 
                                                  FUN = padMatrixWithNA, 
                                                  ncol_desired = max(ncol_by_season))
      
      # combine detection histories for all seasons (effort also) into wide format
      record.hist <- do.call(what = "cbind",
                                  args = detHist_season_list_det_padded) 
      effort <- do.call(what = "cbind",
                                 args = detHist_season_list_effort_padded)
      
    }
  }
    
  
  
  
  ################################################
  # save output as table
  if(day1switch == 1) day1string <- "_first_day_from_survey"
  if(day1switch == 2) day1string <- "_first_day_by_station"
  if(day1switch == 3) day1string <- paste("_first_day", day1, sep = "_")
  
  effortstring <- ifelse(isTRUE(includeEffort), "with_effort__", "no_effort__")
  maxNumberDaysstring <- ifelse(hasArg(maxNumberDays), paste("max",maxNumberDays,"days_", sep = ""), "")
  if(isTRUE(includeEffort)){
    scaleEffortstring <- ifelse(isTRUE(scaleEffort), "scaled_", "not_scaled_")
  } else {
    scaleEffortstring <- ""
  }
  
  # create names for the csv files
  outtable.name <- paste(species, "__detection_history__", effortstring,
                         occasionLength, "_days_per_occasion_",
                         maxNumberDaysstring,
                         "_occasionStart", occasionStartTime,"h_",
                         day1string, "__",
                         Sys.Date(),
                         ".csv", sep = "")
  
  outtable.name.effort <- paste(species, "__effort__",
                                scaleEffortstring,
                                occasionLength, "_days_per_occasion_",
                                maxNumberDaysstring,
                                "_occasionStart", occasionStartTime,"h_",
                                day1string, "__",
                                Sys.Date(),
                                ".csv", sep = "")
  
  outtable.name.effort.scale <- paste(species, "__effort_scaling_parameters__",
                                      occasionLength, "_days_per_occasion_",
                                      maxNumberDaysstring,
                                      "_occasionStart", occasionStartTime,"h_",
                                      day1string, "__",
                                      Sys.Date(),
                                      ".csv", sep = "")
  
  if(writecsv){
    setwd(outDir)
    write.csv(record.hist, file = outtable.name)
    if(isTRUE(includeEffort)){
      write.csv(effort, file = outtable.name.effort)
      if(hasArg(scaleEffort)){
        if(scaleEffort)  write.csv(scale.eff.tmp.attr, file = outtable.name.effort.scale)
      }
    }
  }
  
  if(includeEffort){
    if(scaleEffort){
      return(list(detection_history = record.hist,
                  effort = effort,
                  effort_scaling_parameters = scale.eff.tmp.attr))
    } else {
      return(list(detection_history = record.hist,
                  effort = effort))
    }
  } else {
    return(list(detection_history = record.hist))
  }
  
  }  # end is_multispecies
  
  
  
  # Multi-species case ----

  # Check if all species are in the record table
  species_not_found <- species[!species %in% recordTable[[speciesCol]]]
  if(length(species_not_found) > 0) {
    stop(paste("The following species were not found in recordTable:", 
               paste(species_not_found, collapse = ", ")), call. = FALSE)
  }
  
  # Get the original call minus the species argument which we're modifying
  call_args <- match.call(expand.dots = FALSE)
  call_args$species <- NULL  # remove species as we'll modify it
  
  # Convert to list for manipulation
  call_list <- as.list(call_args)[-1]  # -1 removes the function name
  
  # This function uses recursion with do.call, which can cause issues with
  # finding objects in non-standard environments (like during testing).
  # We must explicitly evaluate the camOp argument to get the actual object
  # and replace the symbol in our call list before recursing.
  if (!is.null(call_list$camOp)) {
    call_list$camOp <- eval(call_list$camOp, envir = parent.frame())
  }
  
  
  # Create the lapply call with only defined arguments
  results <- lapply(species, function(sp) {
    # Add current species to argument list
    call_list$species <- sp
    
    # Execute call with only the arguments that were provided
    do.call(detectionHistory, call_list)[[1]]  
  })
  
  names(results) <- species
  

  # Get effort matrix from first species result (they're identical)
  call_list$species <- species[1]
  full_result <- do.call(detectionHistory, call_list)
  
  # Return modified structure
  if(includeEffort) {
    if(scaleEffort) {
      return(list(
        detection_history = results,
        effort = full_result$effort,
        effort_scaling_parameters = full_result$effort_scaling_parameters
      ))
    } else {
      return(list(
        detection_history = results,
        effort = full_result$effort
      ))
    }
  } else {
    return(list(detection_history = results))
  }
}

#' Collect all images of a species
#' 
#' This function will fetch all images of a particular species from all camera
#' trap stations and copies these images to a new location. The images which
#' are to be copied are found in one of 2 possible ways, 1) by providing an
#' existing record table (created with \code{\link{recordTable}}) or 2) by
#' reading species IDs from species directories or from metadata (calling
#' ExifTool). Earlier in the workflow, i.e., before running this function,
#' images should have been renamed (with \code{\link{imageRename}}) to give
#' images unique file names based on station ID and date/time.
#' 
#' 
#' The function finds the images to be copied by either consulting a record
#' table created with \code{\link{recordTable}} or by reading species IDs from
#' images. The former is considerable faster because ExifTool is not called,
#' but requires images to be in precisely the location given by the columns
#' \code{Directory} and \code{FileName} in \code{recordTable}. To use this
#' feature, provide the function with a record table in argument
#' \code{recordTable}.
#' 
#' If you'd rather read species IDs from images within the function (to make
#' sure all file paths are correct), images need to be in the directory
#' structure required by the package, e.g.
#' 
#' > inDir/Station/Species
#' 
#' or
#' 
#' > inDir/Station/Camera/Species
#' 
#' if using species directories for species IDs, and
#' 
#' > inDir/Station
#' 
#' or
#' 
#' > inDir/Station/Camera
#' 
#' if reading IDs from species metadata tags. In the latter case, only station
#' directories are needed. In any case, the argument \code{species} must match
#' species IDs (either the \code{speciesCol} in \code{recordTable}, species
#' directory names or species metadata tags).
#' 
#' Before running the function, first rename the images using function
#' \code{\link{imageRename}} to provide unique file names and prevent several
#' images from having the same name (if generic names like "IMGP0001.jpg" are
#' used). The function will not copy images if there are duplicate filenames to
#' prevent overwriting images unintentionally.
#' 
#' @param species character. Species whose images are to be fetched
#' @param recordTable data frame. A data frame as returned by function
#' \code{\link{recordTable}}. If you specify this argument, do not specify
#' \code{inDir}
#' @param speciesCol character. Name of the column specifying species ID in
#' \code{recordTable}. Only required if \code{recordTable} is defined
#' @param stationCol character. Name of the column specifying station ID in
#' \code{recordTable}. Only required if \code{recordTable} is defined
#' @param inDir character. Directory containing identified (species level)
#' camera trap images sorted into station subdirectories (e.g.
#' inDir/StationA/). If you specify this argument, do not specify
#' \code{recordTable}.
#' @param outDir character. Directory in which to save species images. A
#' species subdirectory will be created in \code{outDir} automatically.
#' @param createStationSubfolders logical. Save images in station directories
#' within the newly created species directory in \code{outDir}?
#' @param IDfrom character. Read species ID from image metadata ("metadata") of
#' from species directory names ("directory")? Only required if \code{inDir} is
#' defined.
#' @param metadataSpeciesTag character. The species ID tag name in image
#' metadata (if IDfrom = "metadata"). Only required if \code{inDir} is defined.
#' @param metadataHierarchyDelimitor character. The character delimiting
#' hierarchy levels in image metadata tags in field "HierarchicalSubject".
#' Either "|" or ":" (if IDfrom = "metadata"). Only required if \code{inDir} is
#' defined and \code{IDfrom = "metadata"}.
#' 
#' @return A \code{data.frame} with old and new directories and file names and
#' the copy status (\code{copy_ok}; TRUE if copying was successful, FALSE if
#' not).
#' 
#' @author Juergen Niedballa
#' 
#' @examples
#' 
#' \dontrun{
#' # define image directory
#' wd_images_ID <- system.file("pictures/sample_images_species_dir", package = "camtrapR")
#' wd_images_ID_copy <- file.path(tempdir(), "sample_images_species_dir")
#' 
#' species_to_copy <- "VTA"    # = Viverra tangalunga, Malay Civet
#' 
#' specImagecopy <- getSpeciesImages(species                 = species_to_copy,
#'                                   inDir                   = wd_images_ID,
#'                                   outDir                  = wd_images_ID_copy,
#'                                   createStationSubfolders = FALSE,
#'                                   IDfrom                  = "directory"
#'                                   )
#' }
#' 
#' @export getSpeciesImages
#' 
getSpeciesImages <- function(species,
                             recordTable,
                             speciesCol = "Species",
                             stationCol = "Station",
                             inDir,
                             outDir,
                             createStationSubfolders = FALSE,
                             IDfrom,
                             metadataSpeciesTag,
                             metadataHierarchyDelimitor = "|")
{

  file.sep <- .Platform$file.sep

  if(hasArg(recordTable) & hasArg(inDir)) stop("recordTable and inDir cannot be both defined. Please define one only.", call. = FALSE)
    
  
  stopifnot(is.logical(createStationSubfolders))

  if(hasArg(inDir)){
    if(!dir.exists(inDir)) stop("Could not find inDir:\n", inDir, call. = FALSE)
    if (isTRUE(all(unlist(strsplit(tolower(inDir), split = file.sep)) %in%
                   unlist(strsplit(tolower(outDir), split = file.sep)))))    stop("outDir may not be identical to or a subdirectory of inDir", call. = FALSE)

    if(!is.character(IDfrom)) stop("IDfrom must be of class 'character'", call. = FALSE)
    if(IDfrom %in% c("metadata", "directory") == FALSE) stop("'IDfrom' must be 'metadata' or 'directory'", call. = FALSE)

    if(IDfrom == "metadata"){
      stopifnot(metadataHierarchyDelimitor %in% c("|", ":"))
      metadata.tagname <- "HierarchicalSubject"
      if(Sys.which("exiftool") == "")       stop("cannot find ExifTool")
    }
  }

  if(hasArg(recordTable)){
    recordTable <- dataFrameTibbleCheck(df = recordTable)
    checkForSpacesInColumnNames(speciesCol = speciesCol)
  #if(!is.data.frame(recordTable)) stop("recordTable must be a data frame", call. = FALSE)
  
  if(!speciesCol %in% colnames(recordTable))  stop(paste('speciesCol = "', speciesCol, '" is not a column name in recordTable', sep = ''), call. = FALSE)

    if(createStationSubfolders == TRUE &
       stationCol %in% colnames(recordTable) == FALSE)  {
          stop(paste('stationCol = "', stationCol, '" is not a column name in recordTable', sep = ''), call. = FALSE)
          checkForSpacesInColumnNames(stationCol = stationCol)
          }
    if(species %in% recordTable[,speciesCol] == FALSE)  stop("species was not found in speciesCol of recordTable", call. = FALSE)
  }

  if (length(list.files(file.path(outDir, species),
                        recursive = TRUE, full.names = TRUE, pattern = ".jpg$|.JPG$")) !=   0){
    stop(paste("target directory ", file.path(outDir, species), " is not empty"), call. = FALSE)
  }


  if(hasArg(metadataSpeciesTag)){
    if(!is.character(metadataSpeciesTag)){stop("metadataSpeciesTag must be of class 'character'", call. = FALSE)}
    if(length(metadataSpeciesTag) != 1){stop("metadataSpeciesTag must be of length 1", call. = FALSE)}
  }



  multiple_tag_separator = "_&_"

  # define directories to save images to
  dirs.out0 <- file.path(outDir, species)


  if(hasArg(inDir)){

    copy.table <- data.frame()

    if(IDfrom == "directory"){

      dir.list <- list.dirs(inDir, recursive = TRUE, full.names = TRUE)         # recursive to also find species directories in camera subdirectories
      dir.list.short <- list.dirs(inDir, recursive = TRUE, full.names = FALSE)

      # find directories of requested species
      spec.index <- grep(species, dir.list, ignore.case = TRUE)

      if (length(spec.index) == 0)      stop(paste("found no directory named  '", species, "' in ", inDir, sep = ""), call. = FALSE)
      dir.list2 <- dir.list[spec.index]
      dir.list.short2 <- dir.list.short[spec.index]

      # find the station directories that have species records

      for (i in 1:length(dir.list2)) {
        files.tmp <- list.files(dir.list2[i], recursive = FALSE,
                                full.names = TRUE, include.dirs = FALSE, pattern = ".jpg$|.JPG$")
        files.tmp.short <- list.files(dir.list2[i], recursive = FALSE,
                                      full.names = FALSE, include.dirs = FALSE, pattern = ".jpg$|.JPG$")


        copy.table.tmp <- as.data.frame(matrix(nrow = length(files.tmp),
                                               ncol = 5))
        colnames(copy.table.tmp) <- c("DirectoryOrig", "FilenameOrig", "DirectoryCopy",  "FilenameCopy", "copy_ok")

        copy.table.tmp$DirectoryOrig <- dir.list2[i]
        copy.table.tmp$FilenameOrig <- files.tmp.short

        if(isTRUE(createStationSubfolders))  copy.table.tmp$DirectoryCopy <- file.path(dirs.out0, unlist(strsplit(dir.list.short2[i], split = file.sep))[1])
        if(isFALSE(createStationSubfolders)) copy.table.tmp$DirectoryCopy <- dirs.out0
        copy.table.tmp$FilenameCopy <- files.tmp.short

        copy.table <- rbind(copy.table, copy.table.tmp)

        rm(copy.table.tmp)
      }

    } else {

      dirs <- list.dirs(inDir, recursive = FALSE, full.names = TRUE)     # exiftool will work through directory recursively, so set to FALSE here
      dirs_short <- list.dirs(inDir, recursive = FALSE, full.names = FALSE)

      if(length(dirs) == 0) stop("inDir contains no station directories", call. = FALSE)


      for(i in 1:length(dirs_short)){

        command.tmp  <- paste('exiftool -t -q -r -f -Directory -FileName -HierarchicalSubject -ext JPG "', dirs[i], '"', sep = "")
        colnames.tmp <- c("Directory", "FileName", "HierarchicalSubject")

        # run exiftool and make data frame
        metadata.tmp <- runExiftool(command.tmp = command.tmp, colnames.tmp = colnames.tmp)


        if(is.data.frame(metadata.tmp)){

          message(paste(dirs_short[i], ": ", formatC(nrow(metadata.tmp), width = 4), " images", 
                        makeProgressbar(current = i, total = length(dirs_short)), sep = ""))

          # add metadata
          metadata.tmp <- addMetadataAsColumns (intable                    = metadata.tmp,
                                                metadata.tagname           = metadata.tagname,
                                                metadataHierarchyDelimitor = metadataHierarchyDelimitor,
                                                multiple_tag_separator     = multiple_tag_separator)

          # assign species ID
          metadata.tmp <- assignSpeciesID (intable                = metadata.tmp,
                                           IDfrom                 = IDfrom,
                                           metadataSpeciesTag     = metadataSpeciesTag,
                                           speciesCol             = "species",
                                           dirs_short             = dirs_short,
                                           i_tmp                  = i,
                                           multiple_tag_separator = multiple_tag_separator,
                                           returnFileNamesMissingTags = FALSE
          )

          # filter by species
          metadata.tmp <- metadata.tmp[metadata.tmp$species == species,]

          if(nrow(metadata.tmp) >= 1){
            copy.table.tmp <- data.frame(DirectoryOrig = metadata.tmp$Directory,
                                         FilenameOrig  = metadata.tmp$FileName,
                                         DirectoryCopy = NA,
                                         FilenameCopy  = metadata.tmp$FileName)

            if(isTRUE(createStationSubfolders))  copy.table.tmp$DirectoryCopy <- file.path(dirs.out0, dirs_short[i])
            if(isFALSE(createStationSubfolders)) copy.table.tmp$DirectoryCopy <- dirs.out0

            copy.table <- rbind(copy.table, copy.table.tmp)
          }

        } # end if(is.data.frame(metadata.tmp)){
      }   # end for(i in 1:length(dirs_short)){
    }     # end else (IDfrom)

    # if species not found, exit
    if(nrow(copy.table) == 0) {stop(paste("did not find species '", species, "' in  ", inDir, sep = ""), call. = FALSE)}

  }       # end if(hasArg(inDir))

  if(hasArg(recordTable)){

    # species subset from record Table
    species_subset <- recordTable[recordTable[,speciesCol] == species,]

    if(isTRUE(createStationSubfolders))  DirectoryCopy.tmp <- file.path(dirs.out0, species_subset[,stationCol])
    if(isFALSE(createStationSubfolders)) DirectoryCopy.tmp <- dirs.out0

    # fill copy.table
    copy.table <- data.frame(DirectoryOrig = species_subset$Directory,
                             FilenameOrig  = species_subset$FileName,
                             DirectoryCopy = DirectoryCopy.tmp,
                             FilenameCopy  = species_subset$FileName,
                             stringsAsFactors = FALSE)

    missing_files <- !file.exists(file.path(copy.table$DirectoryOrig, copy.table$FilenameOrig))

    if(any(missing_files)) stop("Did not find images from recordTable, e.g.: \n",
                                paste(head(
                                  file.path(copy.table[missing_files, 1],
                                            copy.table[missing_files, 2])
                                ), collapse = "\n"),
                                call. = FALSE)

  }

  # check if there are any duplicate file names (would cause overwriting of images)
  if(any(tapply(copy.table$FilenameCopy,
                INDEX = copy.table$DirectoryCopy,
                FUN   = function(X){any(duplicated(X))}))){
    message(table(copy.table$FilenameCopy)[table(copy.table$FilenameCopy) >= 2])
    stop("Cannot copy. there are duplicate filenames", call. = FALSE)
  }

  # create directories
  sapply(unique(copy.table$DirectoryCopy), FUN = dir.create, recursive = TRUE, showWarnings = FALSE)

  # copy images
  copy.table$copy_ok <- file.copy(from = file.path(copy.table$DirectoryOrig, copy.table$FilenameOrig),
                                  to   = file.path(copy.table$DirectoryCopy, copy.table$FilenameCopy))

  # little summary line
  if(hasArg(inDir)){
    message(paste(toupper(species), " - copied", sum(copy.table$copy_ok),
                  "out of", nrow(copy.table), "images in", inDir, "  to  ",
                  file.path(outDir, species)))
  }
  if(hasArg(recordTable)){
    message(paste(toupper(species), " - copied", sum(copy.table$copy_ok),
                  "out of", nrow(copy.table), "images to  ",
                  file.path(outDir, species)))
  }

  copy.table <- cbind(species = species, copy.table)
  return(copy.table)
}

context("Image Management Workflow")

# Load necessary libraries for testing
library(testthat)
library(withr)

# --- Test Suite ---

# =========================================================================
# Tests for createStationFolders()
# =========================================================================
testthat::describe("createStationFolders()", {
  
  test_that("it creates station directories correctly", {
    with_tempdir({ # All file operations happen in a temporary, self-cleaning directory
      
      stations <- c("StationA", "StationB", "StationC")
      
      # Call the function
      result <- createStationFolders(inDir = ".", stations = stations)
      
      # 1. Check the returned data frame
      expect_s3_class(result, "data.frame")
      expect_equal(nrow(result), 3)
      expect_named(result, c("station", "directory", "created", "exists"))
      expect_true(all(result$created))
      expect_true(all(result$exists))
      
      # 2. Check the actual file system
      created_dirs <- list.dirs(recursive = FALSE)
      expect_equal(sort(basename(created_dirs)), sort(stations))
    })
  })
  
  test_that("it creates nested station/camera directories correctly", {
    with_tempdir({
      stations <- c("StationA", "StationA", "StationB")
      cameras <- c("Cam1", "Cam2", "Cam1")
      
      result <- createStationFolders(inDir = ".", stations = stations, cameras = cameras)
      
      # 1. Check the returned data frame
      expect_true("camera" %in% names(result))
      expect_true(all(result$created))
      
      # 2. Check the actual file system
      expect_true(dir.exists("StationA/Cam1"))
      expect_true(dir.exists("StationA/Cam2"))
      expect_true(dir.exists("StationB/Cam1"))
      expect_false(dir.exists("StationB/Cam2")) # Ensure it didn't create extra dirs
    })
  })
  
  test_that("it handles existing directories and createinDir correctly", {
    with_tempdir({
      # 1. Test createinDir = TRUE
      new_root <- "new_root_dir"
      expect_false(dir.exists(new_root))
      createStationFolders(inDir = new_root, stations = "StationA", createinDir = TRUE)
      expect_true(dir.exists(new_root))
      
      # 2. Test idempotency (running twice)
      dir.create("existing_station")
      result_rerun <- createStationFolders(inDir = ".", stations = "existing_station")
      expect_false(result_rerun$created) # Should report it was not created this time
      expect_true(result_rerun$exists)   # But that it exists
    })
  })
})


# =========================================================================
# Tests for imageRename()
# =========================================================================
testthat::describe("imageRename()", {
  
  # All tests in this block require ExifTool
  skip_if(Sys.which("exiftool") == "", message = "ExifTool not found, skipping tests.")
  
  # Path to the raw sample images included with the package
  wd_images_raw <- system.file("pictures/raw_images", package = "camtrapR")
  
  test_that("it renames correctly with hasCameraFolders = FALSE", {
    with_tempdir({
      # Set up input and output directories
      in_dir <- "raw_data"
      out_dir <- "renamed_data"
      dir.create(in_dir)
      
      # Copy the sample data into our temporary input directory
      file.copy(list.files(wd_images_raw, full.names = TRUE), in_dir, recursive = TRUE)
      
      # Run the function
      renaming_table <- imageRename(
        inDir            = in_dir,
        outDir           = out_dir,
        hasCameraFolders = FALSE,
        copyImages       = TRUE
      )
      
      # 1. Check the returned table
      expect_s3_class(renaming_table, "data.frame")
      expect_true(all(renaming_table$CopyStatus))
      
      # 2. Check the output files
      renamed_files <- list.files(out_dir, recursive = TRUE)
      expect_gt(length(renamed_files), 0)
      
      # 3. Check the filename pattern (Station__Date__Time(n).JPG)
      # Example: "StationA__2009-04-21__15-49-26(1).JPG"
      expect_true(all(grepl("^Station[A-C]__\\d{4}-\\d{2}-\\d{2}__\\d{2}-\\d{2}-\\d{2}\\(\\d+\\)\\.JPG$", basename(renamed_files))))
    })
  })
  
  test_that("copyImages = FALSE produces a table but no files", {
    with_tempdir({
      in_dir <- "raw_data"
      out_dir <- "renamed_data" # This directory will not be created
      dir.create(in_dir)
      file.copy(list.files(wd_images_raw, full.names = TRUE), in_dir, recursive = TRUE)
      
      renaming_table <- imageRename(
        inDir            = in_dir,
        outDir           = out_dir,
        hasCameraFolders = FALSE,
        copyImages       = FALSE # Key argument
      )
      
      # 1. Check returned table
      expect_true(all(!renaming_table$CopyStatus))
      expect_s3_class(renaming_table, "data.frame")
      
      # 2. Assert that NO output directory or files were created
      expect_false(dir.exists(out_dir))
    })
  })
})


# =========================================================================
# Tests for appendSpeciesNames()
# =========================================================================
testthat::describe("appendSpeciesNames()", {
  
  test_that("it can append and then remove species names (IDfrom = directory)", {
    with_tempdir({
      # 1. Setup a directory structure that mimics the state *after* species ID
      species_id_dir <- "species_identified"
      dir.create(file.path(species_id_dir, "StationA", "PBE"), recursive = TRUE)
      dir.create(file.path(species_id_dir, "StationB", "VTA"), recursive = TRUE)
      
      # Create some dummy files with the post-imageRename format
      file.create(file.path(species_id_dir, "StationA", "PBE", "StationA__2023-01-01__12-00-00(1).JPG"))
      file.create(file.path(species_id_dir, "StationB", "VTA", "StationB__2023-01-02__14-00-00(1).JPG"))
      
      # 2. Append the names
      append_table <- appendSpeciesNames(
        inDir            = species_id_dir,
        IDfrom           = "directory",
        hasCameraFolders = FALSE,
        removeNames      = FALSE
      )
      
      # 3. Verify the appended files exist
      files_after_append <- list.files(species_id_dir, recursive = TRUE)
      expect_true(any(grepl("StationA__2023-01-01__12-00-00\\(1\\)__PBE\\.JPG", files_after_append)))
      expect_true(any(grepl("StationB__2023-01-02__14-00-00\\(1\\)__VTA\\.JPG", files_after_append)))
      
      # 4. Remove the names
      remove_table <- appendSpeciesNames(
        inDir            = species_id_dir,
        IDfrom           = "directory", # Needs to know the species to remove the name
        hasCameraFolders = FALSE,
        removeNames      = TRUE
      )
      
      # 5. Verify the files are back to their original names
      files_after_remove <- list.files(species_id_dir, recursive = TRUE)
      expect_false(any(grepl("__PBE\\.JPG", files_after_remove)))
      expect_true(any(grepl("StationA__2023-01-01__12-00-00\\(1\\)\\.JPG", files_after_remove)))
    })
  })
})


# =========================================================================
# Integrated Workflow Test
# =========================================================================
testthat::describe("Full Image Management Workflow", {
  
  skip_if(Sys.which("exiftool") == "", message = "ExifTool not found, skipping tests.")
  
  test_that("the full workflow from creation to renaming works as expected", {
    with_tempdir({
      # --- STAGE 1: createStationFolders ---
      raw_image_root <- "1_raw_images"
      stations <- c("StationA", "StationB")
      createStationFolders(inDir = raw_image_root, stations = stations, createinDir = TRUE)
      
      # Manually copy sample images into the newly created structure
      raw_images_path <- system.file("pictures/raw_images", package = "camtrapR")
      # Put the first 2 sample station dirs into our new structure
      file.copy(file.path(raw_images_path, "StationA"), raw_image_root, recursive = TRUE)
      file.copy(file.path(raw_images_path, "StationB"), raw_image_root, recursive = TRUE)
      
      # --- STAGE 2: imageRename ---
      renamed_image_root <- "2_renamed_images"
      imageRename(
        inDir            = raw_image_root,
        outDir           = renamed_image_root,
        hasCameraFolders = FALSE,
        copyImages       = TRUE
      )
      
      # Verify that renamed files exist
      renamed_files <- list.files(renamed_image_root, recursive = TRUE)
      expect_gt(length(renamed_files), 0)
      
      # --- STAGE 3: Simulate Manual Species ID ---
      # Create species subdirectories and move the renamed files into them
      species_id_root <- "3_species_id_images"
      # We know from sample data that StationA has PBE, VTA, and SUS. We'll just ID them all as PBE for simplicity.
      dir.create(file.path(species_id_root, "StationA", "PBE"), recursive = TRUE)
      # And StationB has VTA
      dir.create(file.path(species_id_root, "StationB", "VTA"), recursive = TRUE)
      
      # Move files
      files_to_move_A <- list.files(file.path(renamed_image_root, "StationA"), full.names = TRUE)
      file.copy(files_to_move_A, file.path(species_id_root, "StationA", "PBE"))
      
      files_to_move_B <- list.files(file.path(renamed_image_root, "StationB"), full.names = TRUE)
      file.copy(files_to_move_B, file.path(species_id_root, "StationB", "VTA"))
      
      # --- STAGE 4: appendSpeciesNames ---
      appendSpeciesNames(inDir = species_id_root, IDfrom = "directory", hasCameraFolders = FALSE)
      
      # Final verification
      final_files <- list.files(species_id_root, recursive = TRUE)
      expect_true(all(grepl("__PBE\\.JPG$|__VTA\\.JPG$", final_files)))
    })
  })
})



# =========================================================================
# Tests for getSpeciesImages()
# =========================================================================
testthat::describe("getSpeciesImages()", {
  
  # Path to sample images included with the package
  wd_images_ID <- system.file("pictures/sample_images_species_dir", 
                              package = "camtrapR") 
  
  # create a mock record table and file structure for robustness
  mock_record_table <- data.frame(
    Station = c("StationA", "StationA", "StationB"),
    Species = c("PBE", "VTA", "PBE"),
    Directory = c("path/to/stationA", "path/to/stationA", "path/to/stationB"),
    FileName = c("img1.jpg", "img2.jpg", "img3.jpg"),
    stringsAsFactors = FALSE
  )
  
  test_that("it copies correct files using a recordTable", {
    with_tempdir({ # All file operations happen in a temporary directory
      
      # 1. Create the source directory structure and dummy files
      dir.create("path/to/stationA", recursive = TRUE)
      dir.create("path/to/stationB", recursive = TRUE)
      file.create(file.path("path/to/stationA", c("img1.jpg", "img2.jpg")))
      file.create(file.path("path/to/stationB", "img3.jpg"))
      
      # Define the output directory
      out_dir <- "species_images"
      
      # 2. Call the function to copy images for species "PBE"
      copy_table <- getSpeciesImages(
        species     = "PBE",
        recordTable = mock_record_table,
        outDir      = out_dir
      )
      
      # 3. Verify the output
      expect_s3_class(copy_table, "data.frame")
      expect_equal(nrow(copy_table), 2) # Should find 2 images of PBE
      expect_true(all(copy_table$copy_ok))
      
      # 4. Check the file system
      output_path <- file.path(out_dir, "PBE")
      expect_true(dir.exists(output_path))
      copied_files <- list.files(output_path)
      expect_equal(length(copied_files), 2)
      expect_equal(sort(copied_files), c("img1.jpg", "img3.jpg"))
    })
  })
  
  test_that("it copies correct files using inDir and IDfrom = 'directory'", {
    with_tempdir({
      # 1. Create a source directory structure with species subfolders
      in_dir <- "source_images"
      dir.create(file.path(in_dir, "StationA", "PBE"), recursive = TRUE)
      dir.create(file.path(in_dir, "StationA", "VTA"), recursive = TRUE)
      dir.create(file.path(in_dir, "StationB", "PBE"), recursive = TRUE)
      
      # Create dummy image files
      file.create(file.path(in_dir, "StationA/PBE/img1.jpg"))
      file.create(file.path(in_dir, "StationA/VTA/img2.jpg"))
      file.create(file.path(in_dir, "StationB/PBE/img3.jpg"))
      
      out_dir <- "species_images"
      
      # 2. Call the function
      copy_table <- getSpeciesImages(
        species = "PBE",
        inDir   = in_dir,
        outDir  = out_dir,
        IDfrom  = "directory"
      )
      
      # 3. Verify output and file system
      expect_equal(nrow(copy_table), 2) # 2 images of PBE
      
      output_path <- file.path(out_dir, "PBE")
      copied_files <- list.files(output_path)
      expect_equal(length(copied_files), 2)
      expect_equal(sort(copied_files), c("img1.jpg", "img3.jpg"))
    })
  })
  
  test_that("createStationSubfolders = TRUE works correctly", {
    with_tempdir({
      # Setup is the same as the previous test
      in_dir <- "source_images"
      dir.create(file.path(in_dir, "StationA", "PBE"), recursive = TRUE)
      dir.create(file.path(in_dir, "StationB", "PBE"), recursive = TRUE)
      file.create(file.path(in_dir, "StationA/PBE/img1.jpg"))
      file.create(file.path(in_dir, "StationB/PBE/img3.jpg"))
      
      out_dir <- "species_images"
      
      # Call with createStationSubfolders = TRUE
      copy_table <- getSpeciesImages(
        species                 = "PBE",
        inDir                   = in_dir,
        outDir                  = out_dir,
        IDfrom                  = "directory",
        createStationSubfolders = TRUE # Key argument
      )
      
      # Verify the nested output directory structure
      expect_true(dir.exists(file.path(out_dir, "PBE", "StationA")))
      expect_true(dir.exists(file.path(out_dir, "PBE", "StationB")))
      expect_true(file.exists(file.path(out_dir, "PBE", "StationA", "img1.jpg")))
      expect_true(file.exists(file.path(out_dir, "PBE", "StationB", "img3.jpg")))
    })
  })
  
  test_that("it stops if the output directory is not empty", {
    with_tempdir({
      out_dir <- "species_images"
      # Create a non-empty output directory beforehand
      dir.create(file.path(out_dir, "PBE"), recursive = TRUE)
      file.create(file.path(out_dir, "PBE", "existing_file.jpg"))
      
      # Expect the function to stop
      expect_error(
        getSpeciesImages(species = "PBE", recordTable = mock_record_table, outDir = out_dir),
        regexp = "target directory .* is not empty"
      )
    })
  })
})

#' Apply time shifts to JPEG image metadata
#' 
#' Change the values of digital timestamps in image metadata using ExifTool. If
#' date/time of images were set incorrectly, they can be corrected easily in
#' batch mode for further analyses. Please, always make a backup of your data
#' before using this function to avoid data loss or damage. This is because
#' ExifTool will make a copy of your images and applies the time shifts to the
#' copies. The file extension of the original images (.JPG) will be renamed to
#' ".JPG_original".
#' 
#' 
#' \code{timeShiftTable} is a data frame with columns for station ID, camera ID
#' (optional), time shift value and direction of time shift (for an example see
#' \code{\link{timeShiftTable}}). Images in \code{inDir} must be sorted into
#' station directories. If \code{hasCameraFolders = TRUE}, the function expects
#' camera subdirectories in the station directories and will only apply time
#' shifts to the camera subdirectories specified by \code{CameraCol} in
#' \code{timeShiftTable}. If \code{hasCameraFolders = FALSE}, shifts will be
#' applied to the whole station directory (including potential subdirectories).
#' 
#' The values of \code{timeShiftColumn} must adhere to the following pattern:
#' "YYYY:mm:dd HH:MM:SS" ("year:month:day hour:minute:second"). Examples:
#' "1:0:0 0:0:0" is a shift of exactly 1 year and "0:0:0 12:10:01" 12 hours and
#' 10 minutes and 1 second. Note that stating "00" may cause problems, so use
#' "0" instead if an entry is zero.
#' 
#' \code{timeShiftSignColumn} signifies the direction of the time shift. "+"
#' moves image dates into the future (i.e. the image date lagged behind the
#' actual date) and "-" moves image dates back (if the image dates were ahead
#' of actual time).
#' 
#' ExifTool stores the original images as \code{.JPG_original} files in the
#' original file location. By setting \code{undo = TRUE}, any JPG files in the
#' directories specified by \code{timeShiftTable} will be deleted and the
#' original JPEGs will be restored from the JPG_original files. Please make a
#' backup before using \code{undo}.
#' 
#' Years can have 365 or 366 days, and months 28 to 31 days. Here is how the
#' function handles these (from the exiftool help page): "The ability to shift
#' dates by Y years, M months, etc, conflicts with the design goal of
#' maintaining a constant shift for all time values when applying a batch
#' shift. This is because shifting by 1 month can be equivalent to anything
#' from 28 to 31 days, and 1 year can be 365 or 366 days, depending on the
#' starting date. The inconsistency is handled by shifting the first tag found
#' with the actual specified shift, then calculating the equivalent time
#' difference in seconds for this shift and applying this difference to
#' subsequent tags in a batch conversion."
#' 
#' \code{ignoreMinorErrors} is useful if image timestamps are not updated
#' correctly (entries in column "n_images" of the output are "... files weren't
#' updated due to errors"). This can be caused by bad MakerNotes and so far was
#' only observed in Panthera V4 and V6 cameras. In that case, set
#' \code{ignoreMinorErrors} to \code{TRUE}. This will add the "-m" option to
#' the Exiftool call, thereby ignoring minor errors and warnings and applying
#' the time shift nevertheless.
#' 
#' @param inDir character. Name of directory containing station directories
#' with images
#' @param hasCameraFolders logical. Do the station directories in \code{inDir}
#' have camera subdirectories (e.g. "inDir/StationA/Camera1")?
#' @param timeShiftTable data.frame containing information about
#' station-/camera-specific time shifts.
#' @param stationCol character. name of the column specifying Station ID in
#' \code{timeShiftTable}
#' @param cameraCol character. name of the column specifying Camera ID in
#' \code{timeShiftTable} (optional)
#' @param timeShiftColumn character. The name of the column containing time
#' shift values in \code{timeShiftTable}
#' @param timeShiftSignColumn character. The name of the column with the
#' direction of time shifts in \code{timeShiftTable}. Can only be "-" or "+".
#' @param undo logical. Undo changes and restore the original images? Please be
#' careful, this deletes any edited images if \code{TRUE}
#' @param ignoreMinorErrors logical. Ignore minor errors that would cause the
#' function to fail (set TRUE for images with bad MakerNotes, observed in
#' Panthera V4 cameras)
#' 
#' @return A \code{data.frame} containing the information about the processed
#' directories and the number of images.
#' 
#' @author Juergen Niedballa
#' 
#' @references \url{https://exiftool.org/#shift}
#' 
#' @examples
#' 
#' 
#' \dontrun{
#' 
#' # copy sample images to temporary directory (so we don't mess around in the package directory)
#' wd_images_ID <- system.file("pictures/sample_images_species_dir", package = "camtrapR")
#' file.copy(from = wd_images_ID, to = tempdir(), recursive = TRUE)
#' wd_images_ID_copy <- file.path(tempdir(), "sample_images_species_dir")
#' 
#' data(timeShiftTable)
#' 
#' 
#' timeshift_run <- timeShiftImages(inDir                = wd_images_ID_copy,
#'                                  timeShiftTable       = timeShiftTable,
#'                                  stationCol           = "Station",
#'                                  hasCameraFolders     = FALSE,
#'                                  timeShiftColumn      = "timeshift",
#'                                  timeShiftSignColumn  = "sign",
#'                                  undo                 = FALSE
#' )
#' 
#' 
#' timeshift_undo <- timeShiftImages(inDir               = wd_images_ID_copy,
#'                                   timeShiftTable      = timeShiftTable,
#'                                   stationCol          = "Station",
#'                                   hasCameraFolders    = FALSE,
#'                                   timeShiftColumn     = "timeshift",
#'                                   timeShiftSignColumn = "sign",
#'                                   undo                = TRUE
#' )
#' }
#' 
#' @export timeShiftImages
#' 
timeShiftImages <- function(inDir,
                            hasCameraFolders,
                            timeShiftTable,
                            stationCol,
                            cameraCol,
                            timeShiftColumn,
                            timeShiftSignColumn,
                            undo = FALSE,
                            ignoreMinorErrors = FALSE
)
{
  if(Sys.which("exiftool") == "") stop("cannot find ExifTool")

  timeShiftTable <- dataFrameTibbleCheck(df = timeShiftTable)
  
  # convert all columns to character
  for(i in 1:ncol(timeShiftTable)){
    timeShiftTable[,i] <- as.character(timeShiftTable[,i])
  }
  rm(i)
  
  # check column names
  checkForSpacesInColumnNames(stationCol = stationCol, timeShiftColumn = timeShiftColumn, timeShiftSignColumn = timeShiftSignColumn)
  if(!stationCol %in% colnames(timeShiftTable))           stop(paste('stationCol = "',   stationCol,     '" is not a column name in timeShiftTable', sep = ''), call. = FALSE)
  if(!timeShiftColumn %in% colnames(timeShiftTable))      stop(paste('timeShiftColumn = "',   timeShiftColumn,     '" is not a column name in timeShiftTable', sep = ''), call. = FALSE)
  if(!timeShiftSignColumn %in% colnames(timeShiftTable))  stop(paste('timeShiftSignColumn = "', timeShiftSignColumn,       '" is not a column name in timeShiftTable', sep = ''), call. = FALSE)

  
  if(isTRUE(hasCameraFolders)){
    stopifnot(cameraCol %in% colnames(timeShiftTable))
  } else {
    if(any(duplicated(timeShiftTable[,stationCol]))) stop("There are duplicates in stationCol. Check argument hasCameraFolders")
  }
   
 
  stopifnot(is.logical(hasCameraFolders))
  for(xy in 1:nrow(timeShiftTable)){
    if(length(unlist(strsplit(timeShiftTable[xy,timeShiftColumn], split = " "))) != 2) stop(paste("there is more than 1 space in your timeShiftColumn string. Only 1 space is allowed (", timeShiftTable[xy,stationCol], ")"))
    if(nchar(timeShiftTable[xy,timeShiftColumn]) - nchar(gsub(":","",timeShiftTable[xy,timeShiftColumn])) != 4) stop("there should be 4 colons in timeShiftColumn (", timeShiftTable[xy,stationCol], ")")
    if(timeShiftTable[xy,timeShiftSignColumn] %in% c("+", "-") == FALSE) stop(paste0('timeShiftSignColumn can only be "+" or "-". Found value "', timeShiftTable[xy,timeShiftSignColumn], '" at station ',
                                                                              timeShiftTable[xy,stationCol]))
    if(length(unlist(lapply(strsplit(timeShiftTable[xy,timeShiftColumn], split = " "), FUN = strsplit, split = ":"))) != 6) stop("there must be six numbers in timeShiftColumn (",
                                                                                                                                 timeShiftTable[xy,stationCol], ")")
  }

  
  if(!dir.exists(inDir)) stop("Could not find inDir:\n", inDir, call. = FALSE)
  
  if(isTRUE(hasCameraFolders)){
    shift.dirs <- file.path(inDir, timeShiftTable[,stationCol], timeShiftTable[,cameraCol])
  } else {
    shift.dirs <- file.path(inDir, timeShiftTable[,stationCol])
  }

  if(any(file.exists(shift.dirs) == FALSE)){
    stop(paste("Station directory does not exist:\n", 
               paste(shift.dirs[file.exists(shift.dirs) == FALSE], collapse = "\n"), sep = ""), call. = FALSE)
  }

  if(isTRUE(undo)){

  results_undo <- data.frame(directory = rep(NA, times = nrow(timeShiftTable)),
                             n_images_undo = rep(NA, times = nrow(timeShiftTable)))
   
  jpg2undo <- list()
  jpg2keep   <- list()
  jpg2keep_newname <- list()
  remove.tmp <- rename.tmp <- list()
  
    for(i in 1:length(shift.dirs)){
      jpg2undo[[i]]       <- list.files(shift.dirs[i], pattern = ".jpg$|.JPG$", recursive = TRUE, full.names = TRUE)
      jpg2keep[[i]]         <- list.files(shift.dirs[i], pattern = ".jpg_original$|.JPG_original$", recursive = TRUE, full.names = TRUE)
      jpg2keep_newname[[i]] <- gsub(pattern = "_original$", replacement = "", x = jpg2keep[[i]])
      
      if(length(jpg2keep[[i]]) == 0) stop(paste("found no .JPG_original files in", shift.dirs[i], "check argument 'undo'"))
      if(length(jpg2undo[[i]]) != length(jpg2keep[[i]])) stop(paste("number of jpgs in", shift.dirs[i], "is not equal to the number of JPG_original files" ))

      

	  results_undo$directory[i] <- shift.dirs[i]
	  results_undo$n_images_undo[i] <- length(jpg2undo[[i]])
    }
	
	remove.tmp <- lapply(jpg2undo, FUN = file.remove)
	for(xyz in 1:length(jpg2keep)){
		rename.tmp[[xyz]] <- file.rename(from = jpg2keep[[xyz]], to = jpg2keep_newname[[xyz]])
	}
	
	return(results_undo)
	  
  } else {

    results.tmp <- list()

    for(i in 1:nrow(timeShiftTable)){
      message(shift.dirs[i])
      command.tmp2b <- paste0('exiftool -r ', ifelse(ignoreMinorErrors, "-m ", ""), '"-DateTimeOriginal', timeShiftTable[i,timeShiftSignColumn], '=',
                             timeShiftTable[i,timeShiftColumn], '" "', shift.dirs[i], '"')
      results.tmp[[i]] <- system(command.tmp2b, intern=TRUE)
      rm(command.tmp2b)
    }

    results.tmp2 <- lapply(results.tmp, FUN = function(x){x[c(length(x) - 1, length(x))]})
    
    output <- data.frame(shift.dirs, matrix(unlist(results.tmp2),ncol = 2, byrow = 2))
    colnames(output) <- c("directory", "n_directories", "n_images")
    if(any(grepl("files weren't updated due to errors", output$n_images))){
      warning("There were problems changing the time stamps in:\n", 
              paste(output$directory[grepl("files weren't updated due to errors", output$n_images)], collapse = "\n"),
              "\n\nTry setting ignoreMinorErrors = TRUE", call. = FALSE)
    }
    
    return(output)
  }
}

#' Generate a \code{capthist} object for spatial capture-recapture analyses
#' from camera-trapping data
#' 
#' This function generates spatial detection histories of individuals of a
#' species for spatial capture-recapture analyses with package
#' \code{\link[secr:secr-package]{secr}}. Data are stored in a
#' \code{\link[secr]{capthist}} object. The \code{\link[secr]{capthist}} object
#' contains detection histories, camera-trap station location and possibly
#' individual and station-level covariates. Detection histories can have
#' adjustable occasion length and occasion start time (as in the function
#' \code{\link{detectionHistory}}).
#' 
#' The function creates a \code{\link[secr]{capthist}} object by combining
#' three different objects: 1) a record table of identified individuals of a
#' species, 2) a camera trap station table with station coordinates and 3) a
#' camera operation matrix computed with \code{\link{cameraOperation}}. The
#' record table must contain a column with individual IDs and optionally
#' individual covariates. The camera trap station table must contain station
#' coordinates and optionally station-level covariates. The camera operation
#' matrix provides the dates stations were active or not and the number of
#' active stations.
#' 
#' \code{day1} defines if each stations detection history will begin on that
#' station's setup day (\code{day1 = "station"}) or if all station's detection
#' histories have a common origin (the day the first station was set up if
#' \code{day1 = "survey"} or a fixed date if, e.g. \code{day1 = "2015-12-31"}).
#' 
#' \code{includeEffort} controls whether an effort matrix is computed or not.
#' If TRUE, effort will be used for object \code{\link[secr]{usage}}
#' information in a \code{\link[secr]{traps}}. \code{binaryEffort} makes the
#' effort information binary. \code{scaleEffort} is currently not used and must
#' be set to FALSE. The reason is that \code{\link[secr]{usage}} can only be
#' either binary, or nonnegative real values, whereas scaling effort would
#' return negative values.
#' 
#' The number of days that are aggregated is controlled by
#' \code{occasionLength}. \code{occasionStartTime} will be removed from the
#' function. It has moved to \code{\link{cameraOperation}}, to ensure daily
#' effort is computed correctly and takes the occasion start time into
#' account.%\code{occasionStartTime} can be used to make occasions begin
#' another hour than midnight (the default). This may be relevant for nocturnal
#' animals, in which 1 whole night would be considered an occasion.
#' 
#' Output can be returned as individual counts per occasion (\code{output =
#' "count"}) or as binary observation (\code{output = "binary"}).
#' 
#' Argument \code{sessionCol} can be used to a create multi-session
#' \code{\link[secr]{capthist}} object. There are two different ways in which
#' the argument is interpreted. It depends on whether a column with the name
#' you specify in argument \code{sessionCol} exists in
#' \code{recordTableIndividual} or in \code{CTtable}. If \code{sessionCol} is
#' found in \code{recordTableIndividual}, the records will be assigned to the
#' specified sessions, and it will be assumed that all camera trap station were
#' used in all sessions. Alternatively, if \code{sessionCol} is found
#' in\code{CTtable}, it will be assumed that only a subset of stations was used
#' in each session, and the records will be assigned automatically (using the
#' station IDs to identify which session they belong into). In both cases,
#' session information must be provided as a sequence of integer numbers
#' beginnign with 1, i.e., you provide the session number directly in
#' \code{sessionCol}. See \code{\link[secr]{session}} for more information
#' about sessions in \pkg{secr}.
#' 
#' \code{\link[secr]{capthist}} objects (as created by
#' \code{\link{spatialDetectionHistory}} for spatial capture-recapture
#' analyses) expect the units of coordinates (\code{Xcol} and \code{col} in
#' \code{CTtable}) to be meters. Therefore, please use a suitable coordinate
#' system (e.g. UTM).
#' 
#' \code{recordDateTimeFormat} defaults to the "YYYY-MM-DD HH:MM:SS"
#' convention, e.g. "2014-09-30 22:59:59". \code{recordDateTimeFormat} can be
#' interpreted either by base-R via \code{\link[base]{strptime}} or in
#' \pkg{lubridate} via \code{\link[lubridate]{parse_date_time}} (argument
#' "orders"). \pkg{lubridate} will be used if there are no "\%" characters in
#' \code{recordDateTimeFormat}.
#' 
#' For "YYYY-MM-DD HH:MM:SS", \code{recordDateTimeFormat} would be either
#' "\%Y-\%m-\%d \%H:\%M:\%S" or "ymd HMS". For details on how to specify date
#' and time formats in R see \code{\link[base]{strptime}} or
#' \code{\link[lubridate]{parse_date_time}}.
#' 
#' @param recordTableIndividual data.frame. the record table with individual
#' IDs created by \code{\link{recordTableIndividual}}
#' @param species character. the species for which to compute the detection
#' history
#' @param camOp The camera operability matrix as created by
#' \code{\link{cameraOperation}}
#' @param CTtable data.frame. contains station IDs and coordinates. Same as
#' used in \code{\link{cameraOperation}}.
#' @param output character. Return individual counts ("count") or binary
#' observations ("binary")?
#' @param stationCol character. name of the column specifying Station ID in
#' \code{recordTableIndividual} and \code{CTtable}
#' @param speciesCol character. name of the column specifying species in
#' \code{recordTableIndividual}
#' @param sessionCol character. name of the column specifying session IDs,
#' either in \code{recordTableIndividual} or in \code{CTtable}. See 'Details'
#' for more information. Session ID values must be a sequence of integer
#' numbers beginning with 1 (i.e., 1,2,3,...).
#' @param Xcol character. name of the column specifying x coordinates in
#' \code{CTtable}
#' @param Ycol character. name of the column specifying y coordinates in
#' \code{CTtable}
#' @param stationCovariateCols character. name of the column(s) specifying
#' station-level covariates in \code{CTtable}
#' @param individualCol character. name of the column specifying individual IDs
#' in \code{recordTableIndividual}
#' @param individualCovariateCols character. name of the column(s) specifying
#' individual covariates in \code{recordTableIndividual}
#' @param recordDateTimeCol character. name of the column specifying date and
#' time in \code{recordTableIndividual}
#' @param recordDateTimeFormat format of column \code{recordDateTimeCol} in
#' \code{recordTableIndividual}
#' @param occasionLength integer. occasion length in days
#' @param minActiveDaysPerOccasion integer. minimum number of active trap days
#' for occasions to be included (optional)
#' @param occasionStartTime (DEPRECATED) integer. time of day (the full hour)
#' at which to begin occasions. Please use argument \code{occasionStartTime} in
#' \code{\link{cameraOperation}} instead.
#' @param maxNumberDays integer. maximum number of trap days per station
#' (optional)
#' @param day1 character. When should occasions begin: station setup date
#' ("station"), first day of survey ("survey"), a specific date (e.g.
#' "2015-12-31")?
#' @param buffer integer. Makes the first occasion begin a number of days after
#' station setup. (optional)
#' @param includeEffort logical. Include trapping effort (number of active
#' camera trap days per station and occasion) as usage in
#' \code{\link[secr]{capthist}} object?
#' @param scaleEffort logical. scale and center effort matrix to mean = 0 and
#' sd = 1? Currently not used. Must be FALSE.
#' @param binaryEffort logical. Should effort be binary (1 if >1 active day per
#' occasion, 0 otherwise)?
#' @param timeZone character. Must be a value returned by
#' \code{\link[base:timezones]{OlsonNames}}
#' @param makeRMarkInput logical. If \code{FALSE}, output will be a data frame
#' for RMark. If \code{FALSE} or not specified, a secr
#' \code{\link[secr]{capthist}} object
#' 
#' @return Output depends on argument \code{makeRMarkInput}:
#' 
#' \item{list("makeRMarkInput = FALSE")}{A \code{\link[secr]{capthist}} object}
#' \item{list("makeRMarkInput = TRUE")}{A data frame for use in RMark}
#' 
#' @section Warning : Please note the section about defining argument
#' \code{timeZone} in the vignette on data extraction (accessible via
#' \code{vignette("DataExtraction")} or online
#' (\url{https://cran.r-project.org/package=camtrapR/vignettes/camtrapr3.pdf})).
#' 
#' @author Juergen Niedballa
#' 
#' @seealso \pkg{secr} \pkg{RMark}
#' 
#' @examples
#' 
#' 
#' data(recordTableIndividualSample)
#' data(camtraps)
#' 
#' # create camera operation matrix (with problems/malfunction)
#' camop_problem <- cameraOperation(CTtable      = camtraps,
#'                                  stationCol   = "Station",
#'                                  setupCol     = "Setup_date",
#'                                  retrievalCol = "Retrieval_date",
#'                                  writecsv     = FALSE,
#'                                  hasProblems  = TRUE,
#'                                  dateFormat   = "dmy"
#' )
#' 
#' sdh <- spatialDetectionHistory(recordTableIndividual = recordTableIndividualSample,
#'                                species               = "LeopardCat",
#'                                camOp                 = camop_problem,
#'                                CTtable               = camtraps,
#'                                output                = "binary",
#'                                stationCol            = "Station",
#'                                speciesCol            = "Species",
#'                                Xcol                  = "utm_x",
#'                                Ycol                  = "utm_y",
#'                                individualCol         = "Individual",
#'                                recordDateTimeCol     = "DateTimeOriginal",
#'                                recordDateTimeFormat  = "ymd HMS",
#'                                occasionLength        = 10,
#'                                day1                  = "survey",
#'                                includeEffort         = TRUE,
#'                                timeZone              = "Asia/Kuala_Lumpur"
#'   )
#' 
#' # missing space in species = "LeopardCat" was introduced by recordTableIndividual
#' # (because of CRAN package policies.
#' # In your data you can have spaces in your directory names)
#' 
#'   summary(sdh)
#'   plot(sdh, tracks = TRUE)
#' 
#'   ## multi-season capthist object
#'   # see vignette "3. Extracting Data from Camera Trapping Images, creating occupancy & secr input"
#'   
#'   data(camtrapsMultiSeason)
#'   camtrapsMultiSeason$session[camtrapsMultiSeason$session == 2009] <- 1
#'   camtrapsMultiSeason$session[camtrapsMultiSeason$session == 2010] <- 2
#' 
#'   data(recordTableIndividualSampleMultiSeason)
#' 
#'   # create camera operation matrix (with problems/malfunction)
#'   camop_session <- cameraOperation(CTtable         = camtrapsMultiSeason,
#'                                       stationCol   = "Station",
#'                                       setupCol     = "Setup_date",
#'                                       sessionCol   = "session",
#'                                       retrievalCol = "Retrieval_date",
#'                                       hasProblems  = TRUE,
#'                                       dateFormat   = "dmy"
#'   )
#' 
#' sdh_multi <- spatialDetectionHistory(recordTableIndividual = recordTableIndividualSampleMultiSeason,
#'                                species               = "LeopardCat",
#'                                output                = "binary",
#'                                camOp                 = camop_session,
#'                                CTtable               = camtrapsMultiSeason,
#'                                stationCol            = "Station",
#'                                speciesCol            = "Species",
#'                                sessionCol            = "session",
#'                                Xcol                  = "utm_x",
#'                                Ycol                  = "utm_y",
#'                                individualCol         = "Individual",
#'                                recordDateTimeCol     = "DateTimeOriginal",
#'                                recordDateTimeFormat  = "ymd HMS",
#'                                occasionLength        = 10,
#'                                day1                  = "survey",
#'                                includeEffort         = TRUE,
#'                                timeZone              = "Asia/Kuala_Lumpur",
#'                                stationCovariateCols  = "utm_y",         # example
#'                                individualCovariateCols = "Individual"   # example
#'   )
#' 
#'   summary(sdh_multi)
#'   plot(sdh_multi, tracks = TRUE)
#' 
#' @importFrom secr read.traps make.capthist traps RMarkInput usage covariates
#' @export spatialDetectionHistory
#' 
spatialDetectionHistory <- function(recordTableIndividual,
                                    species,
                                    camOp,
                                    CTtable,
                                    output = c("binary", "count"),
                                    stationCol = "Station",
                                    speciesCol = "Species",
                                    sessionCol,
                                    Xcol,
                                    Ycol,
                                    stationCovariateCols,
                                    individualCol,
                                    individualCovariateCols,
                                    recordDateTimeCol = "DateTimeOriginal",
                                    recordDateTimeFormat = "ymd HMS",
                                    occasionLength,
                                    minActiveDaysPerOccasion,
                                    occasionStartTime = "deprecated",
                                    maxNumberDays,
                                    day1,
                                    buffer,
                                    includeEffort = TRUE,
                                    scaleEffort   = FALSE,
                                    binaryEffort  = FALSE,
                                    timeZone,
                                    makeRMarkInput
)
{

  wd0 <- getwd()
  on.exit(setwd(wd0))

  #################
  # check input
  if(!hasArg(recordTableIndividual))  stop("'recordTableIndividual' is not defined", call. = FALSE)
  
  recordTableIndividual <- dataFrameTibbleCheck(df = recordTableIndividual)
  CTtable <- dataFrameTibbleCheck(df = CTtable)

  if(!hasArg(camOp))               stop("'camOp' is not defined", call. = FALSE)
  if(!inherits(camOp, "matrix"))   stop ("camOp must be a matrix", call. = FALSE)

  if(!hasArg(species))        stop("'species' is not defined", call. = FALSE)
  if(!is.character(species))  stop("species must be of class character", call. = FALSE)
  if(!length(species) == 1)   stop("species may only contain one value", call. = FALSE)

  if(hasArg(makeRMarkInput)) if(!is.logical(makeRMarkInput)) stop("makeRMarkInput must be logical", call. = FALSE)

  output <- match.arg(output)

  if(!hasArg(occasionLength)) stop("'occasionLength' is not defined", call. = FALSE)

  # check recordTableIndividual
  if(!speciesCol %in% colnames(recordTableIndividual))        stop(paste("speciesCol", speciesCol, "is not a column name in recordTableIndividual"), call. = FALSE)
  if(!recordDateTimeCol %in% colnames(recordTableIndividual)) stop(paste("recordDateTimeCol", recordDateTimeCol, "is not a column name in recordTableIndividual"), call. = FALSE)
  if(!stationCol %in% colnames(recordTableIndividual))        stop(paste("stationCol", stationCol, "is not a column name in recordTableIndividual"), call. = FALSE)
  if(!individualCol %in% colnames(recordTableIndividual))     stop(paste("individualCol", individualCol, "is not a column name in recordTableIndividual"), call. = FALSE)

  checkForSpacesInColumnNames(stationCol = stationCol, Xcol = Xcol, Ycol = Ycol, 
                                         recordDateTimeCol = recordDateTimeCol, speciesCol = speciesCol, individualCol = individualCol)
  if(!inherits(CTtable, "data.frame"))     stop("CTtable must be a data.frame", call. = FALSE)
  if(!stationCol %in% colnames(CTtable))   stop(paste('stationCol = "',   stationCol,  '" is not a column name in CTtable', sep = ''), call. = FALSE)
  if(!Xcol %in% colnames(CTtable))         stop(paste('Xcol = "',   Xcol,  '" is not a column name in CTtable', sep = ''), call. = FALSE)
  if(!Ycol %in% colnames(CTtable))         stop(paste('Ycol = "',   Ycol, '" is not a column name in CTtable', sep = ''), call. = FALSE)
  
  if(!stationCol %in% colnames(recordTableIndividual))            stop(paste('stationCol = "',   stationCol,  '" is not a column name in recordTableIndividual', sep = ''), call. = FALSE)
  if(!recordDateTimeCol %in% colnames(recordTableIndividual))     stop(paste('recordDateTimeCol = "', recordDateTimeCol,  '" is not a column name in recordTableIndividual', sep = ''), call. = FALSE)
  if(!speciesCol %in% colnames(recordTableIndividual))            stop(paste('speciesCol = "', speciesCol,  '" is not a column name in recordTableIndividual', sep = ''), call. = FALSE)
  if(!individualCol %in% colnames(recordTableIndividual))         stop(paste('individualCol = "', individualCol,  '" is not a column name in recordTableIndividual', sep = ''), call. = FALSE)
  
    
  if(!length(stationCol) == 1)  stop("stationCol may only contain one value", call. = FALSE)
  recordTableIndividual[, stationCol] <- as.character(recordTableIndividual[, stationCol])
  if(!is.character(stationCol)) stop("stationCol must be a character", call. = FALSE)

  if(!length(speciesCol) == 1)  stop("speciesCol may only contain one value", call. = FALSE)
  recordTableIndividual[, speciesCol] <- as.character(recordTableIndividual[, speciesCol])
  if(!is.character(speciesCol)) stop("speciesCol must be a character", call. = FALSE)

  if(!length(individualCol) == 1)  stop("individualCol may only contain one value", call. = FALSE)
  recordTableIndividual[, individualCol] <- as.character(recordTableIndividual[, individualCol])
  if(!is.character(individualCol)) stop("individualCol must be a character", call. = FALSE)

  if(!length(recordDateTimeCol) == 1)  stop("recordDateTimeCol may only contain one value", call. = FALSE)
  recordTableIndividual[, recordDateTimeCol] <- as.character(recordTableIndividual[, recordDateTimeCol])
  if(!is.character(recordDateTimeCol)) stop("recordDateTimeCol must be a character", call. = FALSE)
  
  
  # check CTtable
  if(any(is.na(CTtable[,Xcol]))) stop("there are NAs in Xcol", call. = FALSE)
  if(any(is.na(CTtable[,Ycol]))) stop("there are NAs in Ycol", call. = FALSE)
  if(any(is.na(CTtable[,stationCol]))) stop("there are NAs in stationCol of CTtable", call. = FALSE)

  if(hasArg(stationCovariateCols)){
  checkForSpacesInColumnNames(stationCovariateCols = stationCovariateCols)
    if(!all(stationCovariateCols %in% colnames(CTtable))) stop("did not find stationCovariateCols in column names of CTtable (",
                                                               paste(stationCovariateCols[which(stationCovariateCols %in% colnames(CTtable) == FALSE)], collapse = ", "),
                                                               ")", sep = "", call. = FALSE)
  }
  if(!all(recordTableIndividual$Station %in% CTtable$Station)) stop ("there are entries in stationCol of recordTableIndividual that are not found in stationCol of CTtable")

  #####################################################################################################################
  # bring date, time, station ids into shape
  
  recordTableIndividual$DateTime2 <- parseDateTimeObject(inputColumn = recordTableIndividual[,recordDateTimeCol],
                                                  dateTimeFormat = recordDateTimeFormat,
                                                  timeZone = timeZone,
                                                  checkNA_out = FALSE)
  
  if(any(is.na(recordTableIndividual$DateTime2))) stop(paste(sum(is.na(recordTableIndividual$DateTime2)), "out of",
                                                      nrow(recordTableIndividual),
                                                      "entries in recordDateTimeCol of recordTable could not be interpreted using recordDateTimeFormat (NA). row",
                                                      paste(rownames(recordTableIndividual)[which(is.na(recordTableIndividual$DateTime2))], collapse = ", ")))
  
  
  
  
  # check sessionCol argument
  if(hasArg(sessionCol)) {
    checkForSpacesInColumnNames(sessionCol = sessionCol)
  
    if(sessionCol %in% colnames(recordTableIndividual) &
       sessionCol %in% colnames(CTtable)) stop(paste("both CTtable and recordTableIndividual contain sessionCol '", sessionCol, "'", sep = ""), call. = FALSE)

    if(sessionCol %in% colnames(recordTableIndividual) == FALSE &
       sessionCol %in% colnames(CTtable) == FALSE) stop(paste("neither CTtable nor recordTableIndividual contain sessionCol '", sessionCol, "'", sep = ""), call. = FALSE)

    if(sessionCol %in% colnames(recordTableIndividual)) {
      if(!is.numeric(recordTableIndividual[,sessionCol])) stop("sessionCol in recordTableIndividual must be numeric", call. = FALSE)
      if(!all(sort(unique(recordTableIndividual[,sessionCol])) == seq.int(from = 1, to = max(recordTableIndividual[,sessionCol]), by = 1)))
        stop("Problem in sessionCol of recordTableIndividual: Values must come from a gapless sequence of integer numbers starting with 1", call. = FALSE)
    }

    if(sessionCol %in% colnames(CTtable)){
      if(!is.numeric(CTtable[,sessionCol])) stop("sessionCol in CTtable must be numeric", call. = FALSE)
      if(!all(sort(unique(CTtable[,sessionCol])) == seq.int(from = 1, to = max(CTtable[,sessionCol]), by = 1)))
        stop("Problem in sessionCol of CTtable: Values must come from a gapless sequence of integer numbers starting with 1", call. = FALSE)
      
      # introduce sessionCol in recordTable by matching station IDs
      # first check if there is multiple sessions per station (with identical station IDs)
      
      # 1) if there is only 1 session per station
     # recordTableIndividual <- merge(recordTableIndividual, CTtable[,c(stationCol, sessionCol)], by = stationCol)
      
      # 2) if there are multiple entries for each station 
      
      recordTableIndividual <- assignSessionIDtoRecordTable (recordTable_tmp = recordTableIndividual,
                                               camOp = camOp,
                                               dateTimeCol = "DateTime2", # recordDateTimeCol,
                                               stationCol = stationCol,
                                               sessionCol = sessionCol
      )

      
      # see if there are duplicate station#session IDs in CTtable (meaning multiple cameras per station/session)
      
      separatorSession <- "__SESS_"
      stationCol_backup <- paste(stationCol, "backup", sep = "_")
      CTtable[, stationCol_backup] <- CTtable[,stationCol]
      
      #stationCol <- paste(recordTable_tmp$stationCol, recordTable_tmp$sessionCol, sep = "_")
      CTtable[, stationCol] <- paste(CTtable[,stationCol], CTtable[,sessionCol], sep = separatorSession)
      
      
      nonDuplicatedStationSessionIDs <- which(!duplicated(CTtable[,stationCol]))
      
      
      if(length(nonDuplicatedStationSessionIDs) != nrow(camOp)) stop("mismatch in length of nonDuplicatedStationSessionIDs and nrow(camOp)") 
      
      camOp <- camOp[match(CTtable[nonDuplicatedStationSessionIDs, stationCol], rownames(camOp)),]
      
      # check that new station column 
      if(any(rownames(camOp) != CTtable[nonDuplicatedStationSessionIDs, stationCol])) stop("error in assignSessionIDtoRecordTableIndividual: station column order doesn't match camOp matrix anymore",
                                                                                           call. = FALSE)
      
      # remove redundant cameras from CTtable
      CTtable <- CTtable[nonDuplicatedStationSessionIDs,]
      
     # rm(out_tmp)
    }
  }

  
  
  # check that length of station IDs in CTtable and camop match
  if(length(unique(CTtable[,stationCol])) != length(rownames(camOp))) stop ("there is a mismatch between station IDs in CTtable and rownames in camOp (length differs)", call. = FALSE)
  if(any(unique(CTtable[,stationCol]) != rownames(camOp)))            stop ("The order of station IDs in CTtable and rownames in camOp differs [2]", call. = FALSE)

  if(hasArg(timeZone) == FALSE) {
    warning("timeZone is not specified. Assuming UTC", call. = FALSE)
    timeZone <- "UTC"
  }
  if(!is.element(timeZone , OlsonNames())){
    stop("timeZone must be an element of OlsonNames()")
  }

  if(hasArg(minActiveDaysPerOccasion)){
    if(!is.numeric(minActiveDaysPerOccasion))       stop("minActiveDaysPerOccasion must be a number", call. = FALSE)
    if(!minActiveDaysPerOccasion <= occasionLength) stop("minActiveDaysPerOccasion must be smaller than or equal to occasionLength", call. = FALSE)
  }

  if(hasArg(occasionStartTime)) warning("'occasionStartTime' is deprecated and will be removed in the next release. It is now found in cameraOperation()")
  # if(length(occasionStartTime) != 1) stop("occasionStartTime must have length 1")
  # occasionStartTime    <- as.integer(round(occasionStartTime))
  # if(occasionStartTime != 0 & !is.integer(occasionStartTime)) stop ("occasionStartTime must be between 0 and 23", call. = FALSE)
  # if(occasionStartTime < 0 | occasionStartTime >= 24)         stop ("occasionStartTime must be between 0 and 23", call. = FALSE)

  occasionLength    <- as.integer(round(occasionLength))
  if(length(occasionLength) != 1)  stop("occasionLength may only contain one value", call. = FALSE)
  if(!is.numeric(occasionLength))  stop("occasionLength must be a positive integer", call. = FALSE)
  if(occasionLength <= 0)             stop("occasionLength must be a positive integer and not 0", call. = FALSE)
  if(occasionLength > ncol(camOp)/2)  stop("occasionLength may not be greater than half the total number of days in camOp", call. = FALSE)

  if(hasArg(maxNumberDays)){
    maxNumberDays    <- as.integer(maxNumberDays)
    if(maxNumberDays > ncol(camOp))    stop("maxNumberDays must be smaller than the number of columns of camOp", call. = FALSE)
    if(maxNumberDays < occasionLength) stop("maxNumberDays must be larger than or equal to occasionLength")
  }

  if(hasArg(buffer)) {
    if(!is.numeric(buffer)) stop("buffer must be number", call. = FALSE)
    buffer <- round(buffer)
    if(!buffer >= 1)        stop("if buffer is defined, it must be 1 or higher", call. = FALSE)
  }

  # check that species is in the speciesCol
  if(!species %in% recordTableIndividual[,speciesCol]) stop("species ", species, " not found in speciesCol of recordTableIndividual")
  # check all stations in recordTableIndividual are matched in CTtable
  if(!all(recordTableIndividual[,stationCol] %in% CTtable[,stationCol])) {
    stop(paste("items of stationCol in recordTableIndividual are not matched in stationCol of CTtable: ", paste(recordTableIndividual[-which(recordTableIndividual[,stationCol] %in% CTtable[,stationCol]),stationCol], collapse = ", ")))
  }


  if(includeEffort == TRUE){
    if(hasArg(scaleEffort)){
      if(!is.logical(scaleEffort)) stop("scaleEffort must be logical (currently only FALSE is allowed)")
      if(isTRUE(scaleEffort))             stop("currently scaleEffort must be FALSE")
    }
    if(hasArg(binaryEffort)){
      if(!is.logical(binaryEffort)) stop("binaryEffort must be logical (TRUE or FALSE)")
    }
    if(binaryEffort == TRUE & scaleEffort == TRUE) stop("'scaleEffort' and 'binaryEffort' cannot both be TRUE")
  } else {
    scaleEffort  <- FALSE
    binaryEffort <- FALSE
  }

  

 # subset to species of interest
  subset_species           <- recordTableIndividual[recordTableIndividual[,speciesCol] == species,]
  
  # if sessionCol is defined and present in CTtable, check if all records are within correct session. remove if not
  if(hasArg(sessionCol)){
    if(sessionCol %in% colnames(CTtable)){

      rec_station_session <- unique(subset_species[,c(stationCol, sessionCol)])
      cam_station_session <- unique(CTtable[,c(stationCol, sessionCol)])

      rec_station_session$included_rec     <- TRUE
      cam_station_session$included_station <- TRUE
      merged_tmp <- merge(rec_station_session, cam_station_session, all=TRUE)

      if(any(is.na(merged_tmp$included_station))){
        which_to_remove_mismatch <- which(subset_species[,stationCol] == merged_tmp[is.na(merged_tmp$included_station), stationCol] &
                                            subset_species[,sessionCol] == merged_tmp[is.na(merged_tmp$included_station), sessionCol])
        if(length(which_to_remove_mismatch) >= 1){
          warning("there are records at stations which were not used during sessions. They will be ignored. \nrows: ",
                  paste(rownames(subset_species)[which_to_remove_mismatch], collapse = ", "), call. = FALSE)
          subset_species <- subset_species[-which_to_remove_mismatch,]
        }
      }
    }
  }


  # check consistency of argument day1
  if(!is.character(day1)) stop("day1 must be a character", call. = FALSE)

  if(day1 == "survey") {day1switch <- 1} else {
    if(day1 == "station") {day1switch <- 2} else {
      try(date.test <- as.Date(day1), silent = TRUE)
      if(!exists("date.test")) stop("day1 is not specified correctly. It can only be 'station', 'survey', or a date formatted as 'YYYY-MM-DD', e.g. '2016-12-31'")
      if(!is(date.test, "Date")) stop('could not interpret argument day1: can only be "station", "survey" or a specific date (e.g. "2015-12-31")')
      if(hasArg(buffer)) stop("if buffer is defined, day1 can only be 'survey' or 'station'")
      suppressWarnings(rm(date.test))
      day1switch <- 3
    }
  }


  
  # check that column names can be interpreted as day, as extract occasionStartTime if relevant
  camOp <- checkCamOpColumnNames (cameraOperationMatrix = camOp)
  if(hasArg(occasionStartTime)){
    if(occasionStartTime != attributes(camOp)$occasionStartTime){
      stop(paste("occasionStartTime", occasionStartTime, "differs from occasionStartTime in camOp", attributes(camOp)$occasionStartTime)) 
    }
  }
  occasionStartTime <- attributes(camOp)$occasionStartTime
  
  cam.op.worked0 <- as.matrix(camOp)
  
  if(all(as.character(unique(subset_species[,stationCol])) %in% rownames(cam.op.worked0)) == FALSE){
    (stop("Not all values of stationCol in recordTableIndividual are matched by rownames of camOp"))
  }


  ################################################
  # compute date range of stations and records

  arg.list0 <- list(cam.op = cam.op.worked0, subset_species_tmp = subset_species, stationCol_tmp = stationCol, day1_tmp = day1, occasionStartTime_tmp = occasionStartTime, timeZone_tmp = timeZone)

  if(hasArg(maxNumberDays))  arg.list0 <- c(arg.list0,   maxNumberDays_tmp = maxNumberDays)
  if(hasArg(buffer))   arg.list0 <- c(arg.list0, buffer_tmp =  buffer)

  date_ranges <- do.call(createDateRangeTable, arg.list0)

  rm(arg.list0)

  #######################
  # adjust camera operation matrix

  cam.op.worked <- adjustCameraOperationMatrix(cam.op = cam.op.worked0, date_ranges2 = date_ranges, timeZone_tmp = timeZone, day1_2 = day1)

  # append occasionStartTime (if != 0) to column names for output table
  if(occasionStartTime != 0){
    colnames(cam.op.worked) <- paste(colnames(cam.op.worked), "+", occasionStartTime, "h", sep = "")
  }

  ######################
  # calculate trapping effort by station and occasion

  arg.list0 <- list(cam.op = cam.op.worked, occasionLength2 = occasionLength, scaleEffort2 = scaleEffort, includeEffort2 = includeEffort)

  if(hasArg(minActiveDaysPerOccasion))  arg.list0 <- c(arg.list0, minActiveDaysPerOccasion2 = minActiveDaysPerOccasion)

  effort.tmp <- do.call(calculateTrappingEffort, arg.list0)

  rm(arg.list0)

  effort <- effort.tmp[[1]]
  if(isTRUE(scaleEffort))     scale.eff.tmp.attr <- effort.tmp[[2]]
  if(isTRUE(binaryEffort))    effort             <- ifelse(effort >= 1, 1, 0)


  ###################
  # remove records that fall into buffer period or were taken after maxNumberDays

  subset_species <- cleanSubsetSpecies(subset_species2 = subset_species, stationCol2 = stationCol, date_ranges2 = date_ranges)

  ############
  #  define the 1st day of the effective survey period.

  # for each record, what is the respective beginning of the trapping period (used to calculate the occasion), 
  if(day1 %in% c("survey")){
    time2 <- date_ranges$start_first_occasion_survey[match(subset_species[,stationCol], rownames(date_ranges))]   # difference is relative to first occasion of survey
  } else {
    time2 <- date_ranges$start_first_occasion[match(subset_species[,stationCol], rownames(date_ranges))]         # difference is relative to first occasion at each station
  }


  # calculate time difference between records and first day of detection history (the occasion each record belongs into)
  occasionCol <- "Occasion"

  subset_species[,occasionCol] <- as.numeric(ceiling((difftime(time1 = subset_species$DateTime2,
                                                               time2 =  time2,
                                                               units = "secs",
                                                               tz    = timeZone)
                                                      / (occasionLength * 86400))))


  if(max(subset_species[,occasionCol]) > ncol(effort)) {stop("encountered a bug. I'm Sorry. Please report it.")}


  # column names for effort matrix
  if(includeEffort == TRUE) {colnames(effort) <-  paste("o", seq(1,ncol(effort), by = 1), sep = "")}

  #############
  # build spatial detection history

  # get relevant columns from subset_species
  if(hasArg(individualCovariateCols)){
    columnsForCaptures <- c(individualCol, occasionCol, stationCol, individualCovariateCols)
    sdh0 <- subset_species[, columnsForCaptures]
    colnames(sdh0) <- c("ID", occasionCol, "trapID", individualCovariateCols)
  } else {
    columnsForCaptures <- c(individualCol, occasionCol, stationCol)
    sdh0 <- subset_species[, columnsForCaptures]
    colnames(sdh0) <- c("ID", occasionCol, "trapID")
  }

  # add required session column to captures
  if(hasArg(sessionCol)) {
    sdh1 <- cbind(Session = subset_species[,sessionCol], sdh0)
  } else {
    sdh1 <- cbind(Session = 1, sdh0)
  }

  # remove unidentified individuals from captures
  remove.tmp <- which(is.na(sdh1[,"ID"]))
  if(length(remove.tmp) >= 1){
    sdh1 <- sdh1[-remove.tmp,]
    warning(paste("removed", length(remove.tmp), "records because of missing individual IDs"), call. = FALSE)
  }


  # if requested,  remove duplicate records (makes capthist binary. otherwise it returns counts)
  if(output == "binary"){
    sdh2 <- unique(sdh1)                  # remove duplicate rows
  } else {
    sdh2 <- sdh1                          # keep duplicate rows
  }


  ############
  # remove records for which effort was 0 or NA
  rowindex_sdh_to_remove <- vector()

  for(rowindex_sdh in 1:nrow(sdh2)){
    if(is.na(effort[match(sdh2$trapID[rowindex_sdh], rownames(effort)), sdh2$Occasion[rowindex_sdh]]) |   # if effort = NA or
       effort[match(sdh2$trapID[rowindex_sdh], rownames(effort)), sdh2$Occasion[rowindex_sdh]] == 0){     # if effort = 0
      rowindex_sdh_to_remove <- c(rowindex_sdh_to_remove, rowindex_sdh)                                   # save row index into temporary vector
    }
  }
  
  
  
  # if there were records in occasions with effort = 0/NA, remove these records
  if(length(rowindex_sdh_to_remove) != 0){
    warning("removed ", length(rowindex_sdh_to_remove), " record(s) because of effort = 0/NA, incomplete occasions (if includeEffort = FALSE), or effort < minActiveDaysPerOccasion:\n",
            paste(capture.output(print(sdh2[rowindex_sdh_to_remove,])), collapse = "\n"),
            call. = FALSE)
    sdh2 <- sdh2[-rowindex_sdh_to_remove,]
  }
  rm(rowindex_sdh, rowindex_sdh_to_remove)


  # make sure we have no NA in effort matrix (overwrite with 0)
  # otherwise secr::verify() will crash if there are NAs in the trap usage information
  effort[which(is.na(effort))] <- 0


  ############
  # make secr traps object

  coord.ct <- CTtable[,c(Xcol, Ycol)]
  colnames(coord.ct) <- c("x", "y")
  
  if(any(duplicated(CTtable[,stationCol]))) stop("Duplicate values in stationCol of CTtable are not allowed", call. = F)
  
  rownames(coord.ct) <- CTtable[,stationCol]

  if(!all(rownames(coord.ct) == rownames(cam.op.worked))) stop("Error assigning rownames to traps data frame. Please report this bug")
  
  # set detector type according to desired output (count or binary)
  detectortype <- ifelse (output == "binary", "proximity", "count")


# make station covariate data frame

  if(hasArg(stationCovariateCols)){
    whichstationCovariateCols <- which(colnames(CTtable) %in% stationCovariateCols)
    
    stationCovsDF <- data.frame(CTtable[, whichstationCovariateCols])
    
    if(length(stationCovariateCols) == 1) {
      colnames(stationCovsDF) <- stationCovariateCols
    }
  }

  # if sessionCol is defined and it is found in CTtable, make a list of traps objects (one for each session)

  if(hasArg(sessionCol)){
    if(sessionCol %in% colnames(CTtable)){
      secr.traps <- list()

      if(any(table(CTtable[, sessionCol]) == 1)) stop ("there is a session in CTtable that has only one station.")

      for(sessionID in sort(unique(CTtable[,sessionCol]))){
        which.session.tmp <- which(CTtable[,sessionCol] == sessionID)   # index of all stations active during session
        if(includeEffort == TRUE) {
          secr.traps[[sessionID]] <- read.traps(data         = coord.ct[which.session.tmp,],
                                                detector     = detectortype,
                                                binary.usage = binaryEffort)

          effort_session_tmp <- effort[which.session.tmp,]              # subset effort to active stations only
          secr::usage(secr.traps[[sessionID]]) <- effort_session_tmp[,which(colSums(effort_session_tmp) != 0)]  # only occasions with >1 active station
          # fix occasions (if there were occasions without active cameras before an occasion)
          # if a session is empty (e.g. because day1 was after end of that session, all records will be removed from sdh2)
          # that session then needs to be deleted with a warning
          sdh2[sdh2[, "Session"] == sessionID, "Occasion"] <- sapply(sdh2[sdh2[, "Session"] == sessionID, "Occasion"],
                                                                     FUN = function(x){x - length(which(which(colSums(effort_session_tmp) == 0) <  x) == TRUE)})

          rm(effort_session_tmp)
          # assign station names as row names back
          rownames(secr.traps[[sessionID]]) <- CTtable[which.session.tmp, stationCol_backup]
          rownames(secr::usage(secr.traps[[sessionID]])) <- CTtable[which.session.tmp, stationCol_backup]
          
        } else {
          secr.traps[[sessionID]] <- read.traps(data         = coord.ct[which.session.tmp,],
                                                detector     = detectortype)
          rownames(secr.traps[[sessionID]]) <- CTtable[which.session.tmp, stationCol_backup]
        }
        if(exists("stationCovsDF")) {
          secr::covariates(secr.traps[[sessionID]]) <- stationCovsDF[which.session.tmp,]
          rownames(secr::covariates(secr.traps[[sessionID]])) <- CTtable[which.session.tmp, stationCol_backup]
        }
        rm(which.session.tmp)
      }
# remove sessions with empty effort (caused by day1 after last station retrieval)
      if(any(sapply(secr::usage(secr.traps), sum) == 0)){
        which_effort_empty <- which(sapply(secr::usage(secr.traps), sum) == 0)
        if(day1switch == 3){
          warning("will remove data from session  ", paste(which_effort_empty, collapse = ", "), "  because the effort matrix was empty.
                  This is probably due to the usage of a date outside the station operation range in argument 'day1'", call. = FALSE)
        } else {
          warning("will remove data from ", paste(which_effort_empty, collapse = ", "), " session because the effort matrix was empty.", call. = FALSE)
        }
        secr.traps <- secr.traps[-which_effort_empty]
      }
    }
    }

  # otherwise (sessionCol not in CTtable or not defined) make traps object for all stations
  if(!exists("secr.traps")){
    if(includeEffort == TRUE) {
      secr.traps <- read.traps(data         = coord.ct,
                               detector     = detectortype,
                               binary.usage = binaryEffort)
      # if there are columns in effort which have colSum = 0 (no station active), remove these from effort/usage information
      secr::usage(secr.traps) <- effort[,which(colSums(effort) != 0)]
      # consequently, occasions need to be adjusted too (if occasions were lost before a specific occasion, reduce occasion by number of removed occasions before that occasion)
      sdh2$Occasion <- sapply(sdh2$Occasion, FUN = function(x){x - length(which(which(colSums(effort) == 0) <  x) == TRUE)})

    } else {
      secr.traps <- read.traps(data     = coord.ct,
                               detector = detectortype)
    }
    if(exists("stationCovsDF")) secr::covariates(secr.traps) <- stationCovsDF
  }

  # find number of occasions
  if(includeEffort == FALSE) {  
    noccasions <- ncol(effort) 
  } else {
    if(inherits(secr.traps, "list")) {
      noccasions <- sapply(secr::usage(secr.traps), ncol)    # includeEffort = TRUE / multi-session
    } else { 
      if(is(secr::usage(secr.traps), "matrix")){
        noccasions <- ncol(secr::usage(secr.traps))          # includeEffort = TRUE / single-session
      } else {
        noccasions <- ncol(effort)                           # includeEffort = TRUE / single-session
      }
    }
  }


  ###########
  # make capthist object

  deparse_sdh2 <- deparseCamOpRownames(sdh2$trapID)
  sdh2$trapID <- deparse_sdh2$station
  
  capthist.secr <- make.capthist(captures   = sdh2,
                                 traps      = secr.traps,
                                 fmt        = "trapID",
                                 noccasions = noccasions,
                                 bysession  = TRUE)

  if(hasArg(makeRMarkInput)){
    if(isTRUE(makeRMarkInput)){
      RMarkDataframe <- RMarkInput(capthist.secr, covariates = TRUE)
      return(RMarkDataframe)
    }
  }

  return(capthist.secr)
}

#' Species Accumulation Curves for Camera Trap Data
#'
#' @description Generates species accumulation, rarefaction and extrapolation curves
#'    from camera trap data using the \pkg{iNEXT} package (Chao et al. 2014). The function
#'    creates sampling effort-based accumulation curves with sampling units being either
#'    camera trap stations or days.
#'    
#'    Note that these curves are based on observed detections only and do not account 
#'    for imperfect detection. Species may be present but not detected, leading to 
#'    underestimation of true species richness. For analyses that explicitly account 
#'    for imperfect detection, consider using occupancy-based approaches (see 
#'    \code{\link{communityModel}}).
#'
#' @details
#' The function provides three types of curves:
#' \itemize{
#'   \item Sample-size-based rarefaction/extrapolation curve
#'   \item Sample completeness curve
#'   \item Coverage-based rarefaction/extrapolation curve
#' }
#' 
#' While these curves provide useful insights into sampling completeness and species 
#' richness patterns, they should be interpreted with caution in camera trap studies. 
#' Unlike occupancy models, they do not account for:
#' \itemize{
#'   \item Imperfect detection (species present but not detected)
#'   \item Variation in detection probability among species
#'   \item Spatial variation in species occurrence
#'   \item Temporal variation in species activity
#' }
#'
#'
#' @param CTtable data.frame containing the camera trap deployment information.
#' @param recordTable data.frame containing the camera trap records.
#' @param speciesCol character. Name of the column specifying species names in recordTable
#' @param recordDateTimeCol character. Name of the column containing date and time information in recordTable
#' @param recordDateTimeFormat character. Format of column \code{recordDateTimeCol} in \code{recordTable} 
#' @param setupCol character. Name of the column containing camera setup dates in CTtable
#' @param dateFormat character. Format of column \code{setupCol} in \code{CTtable}
#' @param stationCol character. Name of the column containing station IDs in both tables
#' @param assemblageCol character. Optional. Name of column in recordTable for grouping data into separate assemblages
#' @param q numeric. The order of diversity measure. Default is 0 (species richness)
#' @param x_unit character. Whether to use "station" or "day" as sampling unit. Default is "station"
#' @param knots numeric. number of values along x axis for which values are computed
#' @param conf numeric. confidence interval
#' @param nboot numeric. number of replications
#' @param endpoint integer. Sample size used as endpoint for rarefaction/extrapolation (in \code{\link[iNEXT]{iNEXT}})
#' 
#'
#' @return An object of class "iNEXT" containing:
#' \itemize{
#'   \item DataInfo - data information
#'   \item iNextEst - diversity estimates for rarefied and extrapolated samples
#'   \item AsyEst - asymptotic diversity estimates
#' }
#'
#' @note Requires package \pkg{iNEXT}.
#'
#' @references
#' Chao, A., Gotelli, N. J., Hsieh, T. C., Sander, E. L., Ma, K. H., Colwell, R. K., 
#' & Ellison, A. M. (2014). Rarefaction and extrapolation with Hill numbers: 
#' A framework for sampling and estimation in species diversity studies. 
#' Ecological Monographs, 84(1), 45-67.
#'
#' @examples
#' \dontrun{
#' # Basic usage with stations as sampling units
#' result <- speciesAccum(
#'   CTtable = cams,
#'   recordTable = recs,
#'   speciesCol = "Species",
#'   recordDateTimeCol = "DateTime", 
#'   setupCol = "Setup_date",
#'   stationCol = "Station",
#'   q = 0,
#'   x_unit = "station"
#' )
#'
#' # Plot results
#' ggiNEXT(result, type = 1)  # Sample-size-based R/E curve
#' ggiNEXT(result, type = 2)  # Sample completeness curve
#' ggiNEXT(result, type = 3)  # Coverage-based R/E curve
#'
#' # With assemblage grouping and days as sampling units
#' result_by_assemblage <- speciesAccum(
#'   CTtable = cams,
#'   recordTable = recs,
#'   speciesCol = "Species",
#'   recordDateTimeCol = "DateTime",
#'   setupCol = "Setup_date",
#'   stationCol = "Station",
#'   assemblageCol = "Season",
#'   q = 0,
#'   x_unit = "day"
#' )
#' }
#'
#' @seealso \code{\link{surveyDashboard}} for interactive species accumulation analysis
#'
#' @author Juergen Niedballa
#'
#' @importFrom dplyr mutate arrange distinct count left_join select
#' @importFrom stats setNames
#' @importFrom lubridate parse_date_time
#' @export
#' 
speciesAccum <- function(CTtable, 
                         recordTable,
                         speciesCol,
                         recordDateTimeCol,
                         recordDateTimeFormat = "ymd HMS",
                         setupCol, 
                         dateFormat = "ymd",
                         stationCol,
                         assemblageCol = NULL,
                         q = 0,
                         x_unit = c("station", "survey_day", "station_day"),
                         knots = 40,
                         conf = 0.95, 
                         nboot = 50,
                         endpoint = NULL) {
  
  # Input validation
  x_unit <- match.arg(x_unit)
  
  if(inherits(CTtable, "sf")) CTtable <- sf::st_drop_geometry(CTtable)
  CTtable <- as.data.frame(CTtable)
  
  # Check if required columns exist
  required_cols <- c(speciesCol, recordDateTimeCol, stationCol)
  if(!all(required_cols %in% names(recordTable))) {
    stop("Missing required columns in recordTable")
  }
  if(!all(c(setupCol, stationCol) %in% names(CTtable))) {
    stop("Missing required columns in CTtable")
  }
  
  # Load required packages
  # requireNamespace("dplyr")
  
  
  # If assemblageCol is provided, create separate matrices for each assemblage
  if(!is.null(assemblageCol)) {
    if(!assemblageCol %in% names(recordTable)) {
      stop("assemblageCol not found in recordTable")
    }
    
    assemblages <- unique(recordTable[[assemblageCol]])
    incidence_matrices <- list()
    
    for(assem in assemblages) {
      records_subset <- recordTable[recordTable[[assemblageCol]] == assem,]
      stations_subset <- CTtable[CTtable[[stationCol]] %in% 
                                   unique(records_subset[[stationCol]]),]
      # species_list <- unique(records_subset[[speciesCol]])  #only species in current assemblage
      species_list <- unique(recordTable[[speciesCol]])    # all species (across assemblages)
      
      incidence_matrices[[assem]] <- create_incidence_matrix(
        records_subset = records_subset, 
        stations_subset = stations_subset,
        species_list = species_list,
        temporal = x_unit == "survey_day" | x_unit == "station_day",
        by_station = x_unit == "station_day",
        recordDateTimeCol = recordDateTimeCol,
        recordDateTimeFormat = recordDateTimeFormat,
        setupCol = setupCol, 
        dateFormat = dateFormat,
        stationCol = stationCol,
        speciesCol = speciesCol
      )
    }
  } else {  # if no assemblages defined
    # Create single matrix for all data
    species_list <- unique(recordTable[[speciesCol]])
    
    incidence_matrices <- create_incidence_matrix(
      records_subset = recordTable, 
      stations_subset = CTtable,
      species_list = species_list,
      temporal = x_unit == "survey_day" | x_unit == "station_day",
      by_station = x_unit == "station_day",
      recordDateTimeCol = recordDateTimeCol,
      recordDateTimeFormat = recordDateTimeFormat,
      setupCol = setupCol, 
      dateFormat = dateFormat,
      stationCol = stationCol,
      speciesCol = speciesCol
    )
    incidence_matrices <- list(incidence_matrices)
  }
  
  # Run iNEXT
  out <- iNEXT::iNEXT(x = incidence_matrices, 
                      datatype = "incidence_raw", 
                      q = q,
                      knots = knots,
                      conf = conf,
                      nboot = nboot,
                      endpoint = endpoint)
  
  return(out)
}



# Function to create incidence matrix
create_incidence_matrix <- function(records_subset, 
                                    stations_subset, 
                                    stationCol,
                                    setupCol,
                                    dateFormat,           # camera trap table
                                    recordDateTimeCol,        # record table
                                    recordDateTimeFormat,
                                    speciesCol,
                                    species_list, 
                                    temporal = FALSE,
                                    by_station)         # only relevant if temporal = T) 
  {

  # 1. Parse Record Date/Times (only if they aren't already POSIXt)
  if (!inherits(records_subset[[recordDateTimeCol]], "POSIXt")) {
    records_subset[[recordDateTimeCol]] <- parse_date_time(
      records_subset[[recordDateTimeCol]],
      orders = recordDateTimeFormat
    )
  }
  
  # 2. Parse Station Setup Dates (ensure they end up as Date objects)
  if (!inherits(stations_subset[[setupCol]], c("Date", "POSIXt"))) {
    stations_subset[[setupCol]] <- as.Date(parse_date_time(
      stations_subset[[setupCol]], 
      orders = dateFormat
    ))
  } else {
    stations_subset[[setupCol]] <- as.Date(stations_subset[[setupCol]])
  }
  
  
  if(temporal) {
    
    # avoid CRAN notes
    survey_date <- NULL
    station_setup_date <- NULL
    
    
    # For temporal analysis (days)
    records_subset <- records_subset %>%
      # Join with station setup dates
      left_join(stations_subset %>% 
                  select(!!sym(stationCol), !!sym(setupCol)), 
                by = stationCol) %>%
      mutate(
        survey_date = as.Date(.data[[recordDateTimeCol]]),
        station_setup_date = as.Date(.data[[setupCol]]),
        # survey day relative to first camera setup in survey
        survey_day = as.numeric(survey_date - min(as.Date(stations_subset[[setupCol]]))) + 1,
        # survey day relative to each station's setup
        survey_day_by_station = as.numeric(survey_date - station_setup_date) + 1
      )
    
    # Allow choice between the two types of survey days
    if(by_station) {
      units <- 1:max(records_subset$survey_day_by_station)
      unit_col <- "survey_day_by_station"
    } else {
      units <- 1:max(records_subset$survey_day)
      unit_col <- "survey_day"
    }
    
    # sanity check
    if(any(records_subset[, unit_col] < 0)) {
      warning(paste("Calculated negative survey days in",
                    paste(
                      unique(records_subset[records_subset[, unit_col] < 0, stationCol] ),
                      collapse = ", "
                    ), "\n"))
    }
  }
  else {
    # For spatial analysis (stations)
    units <- unique(stations_subset[[stationCol]])
    unit_col <- stationCol
  }
  
  
  
  # Create empty matrix
  incidence_matrix <- matrix(0, 
                             nrow = length(species_list), 
                             ncol = length(units),
                             dimnames = list(species_list, as.character(units)))
  
  # Fill matrix
  for(sp in species_list) {
    for(unit in units) {
      if(temporal) {
        detected <- any(records_subset[[speciesCol]] == sp & 
                          records_subset[[unit_col]] == unit)
      } else {
        detected <- any(records_subset[[speciesCol]] == sp & 
                          records_subset[[unit_col]] == unit)
      }
      if(detected) {
        incidence_matrix[sp, as.character(unit)] <- 1
      }
    }
  }
  
  return(incidence_matrix)
}




# Helper function to convert records to station-days
convert_to_station_days <- function(recordTable, 
                                    CTtable, 
                                    speciesCol, 
                                    stationCol,
                                    recordDateTimeCol, 
                                    setupCol,
                                    return = "accumulation") {
  
  return <- match.arg(return, choices = c("recordTable", "accumulation"))
  
  # Convert all times to POSIXct
  record_times <- as.POSIXct(recordTable[[recordDateTimeCol]])
  setup_times <- as.POSIXct(CTtable[[setupCol]])
  
  # Create lookup for station setup times
  station_setups <- setNames(setup_times, CTtable[[stationCol]])
  
  # prevent CRAN check notes
  datetime <- setup_time <- station_day <- species <- new_species <- station <- NULL
  
  # Calculate day number for each record
  records_df <- data.frame(
    station = recordTable[[stationCol]],
    species = recordTable[[speciesCol]],
    datetime = record_times
  ) %>%
    # Add setup time for each record's station
    mutate(
      setup_time = station_setups[station],
      # Calculate days since setup (rounded down to integer)
      station_day = as.integer(difftime(datetime, setup_time, units="days")) + 1
    )
  
  # For each station-day, get number of new species
  accumulation_df <- records_df %>%
    arrange(station_day) %>%
    # Keep first detection of each species
    distinct(species, .keep_all = TRUE) %>%
    # Count species per day
    count(station_day, name = "new_species") %>%
    # Calculate cumulative species
    mutate(cumulative_species = cumsum(new_species))
  
  if(return == "recordTable")   return(records_df)
  if(return == "accumulation") return (accumulation_df)
}

context("speciesAccum")

# Load necessary libraries
library(lubridate)
library(dplyr)

# --- Test Setup ---
# Create small, predictable mock data frames for testing.

# Mock CTtable
mock_cams <- data.frame(
  Station = c("A", "B", "C"),
  Setup_date = as.Date(c("2023-01-01", "2023-01-02", "2023-01-05"))
)

# Mock recordTable with different species, stations, and dates
mock_recs <- data.frame(
  Station = c("A", "A", "B", "B", "C", "C", "A"),
  Species = c("Fox", "Badger", "Fox", "Deer", "Badger", "Deer", "Fox"),
  DateTime = c(
    "2023-01-01 12:00:00", # Fox at A on Day 1 (station day 1)
    "2023-01-02 10:00:00", # Badger at A on Day 2 (station day 2)
    "2023-01-02 14:00:00", # Fox at B on Day 2 (station day 1)
    "2023-01-04 08:00:00", # Deer at B on Day 4 (station day 3)
    "2023-01-05 18:00:00", # Badger at C on Day 5 (station day 1)
    "2023-01-06 20:00:00", # Deer at C on Day 6 (station day 2)
    "2023-01-03 16:00:00"  # Fox at A on Day 3 (station day 3) - a second visit
  ),
  Season = c("Winter", "Winter", "Winter", "Winter", "Spring", "Spring", "Winter")
)

# --- Test Suite ---

testthat::describe("Core Functionality", {
  
  # Skip all tests if iNEXT is not installed
  skip_if_not_installed("iNEXT")
  
  # test_that("it produces a valid iNEXT object with x_unit = 'station'", {
  #   result <- speciesAccum(
  #     CTtable = mock_cams,
  #     recordTable = mock_recs,
  #     speciesCol = "Species",
  #     recordDateTimeCol = "DateTime",
  #     setupCol = "Setup_date",
  #     stationCol = "Station",
  #     x_unit = "station",
  #     nboot = 10 # Use a small nboot for speed
  #   )
  #   
  #   # 1. Check output class and structure
  #   expect_s3_class(result, "iNEXT")
  #   expect_named(result, c("DataInfo", "iNextEst", "AsyEst"))
  #   
  #   # 2. Check DataInfo for correct dimensions
  #   # 3 stations, 3 species, 5 total incidences
  #   expect_equal(result$DataInfo$T, 3) 
  #   expect_equal(result$DataInfo$S.obs, 3)
  #   expect_equal(result$DataInfo$n, 5)
  })
  
  test_that("it produces a valid iNEXT object with x_unit = 'survey_day'", {
    
    skip_if_not_installed("iNEXT")
    
    result <- speciesAccum(
      CTtable = mock_cams,
      recordTable = mock_recs,
      speciesCol = "Species",
      recordDateTimeCol = "DateTime",
      setupCol = "Setup_date",
      stationCol = "Station",
      x_unit = "survey_day", # Key argument
      nboot = 10
    )
    
    # 1. Check output class
    expect_s3_class(result, "iNEXT")
    
    # 2. Check DataInfo for correct dimensions
    # The survey runs for 6 unique days
    expect_equal(result$DataInfo$T, 6)
    expect_equal(result$DataInfo$S.obs, 3)
  })
  
#   test_that("it produces a valid iNEXT object with x_unit = 'station_day'", {
#     result <- speciesAccum(
#       CTtable = mock_cams,
#       recordTable = mock_recs,
#       speciesCol = "Species",
#       recordDateTimeCol = "DateTime",
#       setupCol = "Setup_date",
#       stationCol = "Station",
#       x_unit = "station_day", # Key argument
#       nboot = 10
#     )
#     
#     # 1. Check output class
#     expect_s3_class(result, "iNEXT")
#     
#     # 2. Check DataInfo for correct dimensions
#     # The maximum number of station_days is 3
#     expect_equal(result$DataInfo$T, 3)
#     expect_equal(result$DataInfo$S.obs, 3)
#   })
# })


# testthat::describe("Assemblage Grouping", {
#   
#   skip_if_not_installed("iNEXT")
#   
#   test_that("assemblageCol correctly splits data into a multi-assemblage iNEXT object", {
#     result_assem <- speciesAccum(
#       CTtable = mock_cams,
#       recordTable = mock_recs,
#       speciesCol = "Species",
#       recordDateTimeCol = "DateTime",
#       setupCol = "Setup_date",
#       stationCol = "Station",
#       assemblageCol = "Season", # Key argument
#       x_unit = "station",
#       nboot = 10
#     )
#     
#     # 1. Check output class and structure
#     expect_s3_class(result_assem, "iNEXT")
#     expect_true(is.list(result_assem$iNextEst)) # iNextEst is a list for multi-assemblage
#     
#     # 2. Check DataInfo for correct number of assemblages
#     expect_equal(nrow(result_assem$DataInfo), 2)
#     expect_equal(sort(result_assem$DataInfo$Assemblage), c("Spring", "Winter"))
#     
#     # 3. Check data for one of the assemblages
#     winter_data_info <- result_assem$DataInfo[result_assem$DataInfo$Assemblage == "Winter", ]
#     # Winter has 2 stations (A, B) and 3 species (Fox, Badger, Deer)
#     expect_equal(winter_data_info$T, 2)
#     expect_equal(winter_data_info$S.obs, 3)
#   })
# })


testthat::describe("Incidence Matrix Creation (Internal Logic)", {
  
  test_that("create_incidence_matrix builds a correct 'station' matrix", {
    
    # Call the internal helper directly to test its output
    mat <- create_incidence_matrix(
      records_subset = mock_recs,
      stations_subset = mock_cams,
      species_list = unique(mock_recs$Species),
      temporal = FALSE, # for x_unit = "station"
      stationCol = "Station", setupCol = "Setup_date", speciesCol = "Species",
      recordDateTimeCol = "DateTime", recordDateTimeFormat = "ymd HMS", dateFormat = "ymd"
    )
    
    # 1. Check dimensions: 3 species x 3 stations
    expect_equal(dim(mat), c(3, 3))
    expect_equal(sort(rownames(mat)), c("Badger", "Deer", "Fox"))
    expect_equal(sort(colnames(mat)), c("A", "B", "C"))
    
    # 2. Check specific cell values
    expect_equal(mat["Fox", "A"], 1)    # Fox at Station A
    expect_equal(mat["Badger", "B"], 0) # Badger NOT at Station B
    expect_equal(mat["Deer", "C"], 1)   # Deer at Station C
    expect_equal(sum(mat), 6)          # 6 total incidences
  })
  
  test_that("create_incidence_matrix builds a correct 'survey_day' matrix", {
    mat <- create_incidence_matrix(
      records_subset = mock_recs,
      stations_subset = mock_cams,
      species_list = unique(mock_recs$Species),
      temporal = TRUE, # for day-based units
      by_station = FALSE, # for "survey_day"
      stationCol = "Station", setupCol = "Setup_date", speciesCol = "Species",
      recordDateTimeCol = "DateTime", recordDateTimeFormat = "ymd HMS", dateFormat = "ymd"
    )
    
    # 1. Check dimensions: 3 species x 6 survey days
    expect_equal(dim(mat), c(3, 6))
    
    # 2. Check specific cell values
    expect_equal(mat["Fox", "1"], 1)     # Fox on Day 1
    expect_equal(mat["Badger", "2"], 1)  # Badger on Day 2
    expect_equal(mat["Deer", "2"], 0)    # Deer NOT on Day 2
    expect_equal(mat["Deer", "4"], 1)    # Deer on Day 4
  })
})


testthat::describe("Input Validation", {
  
  test_that("it stops for missing or invalid columns", {
    
    skip_if_not_installed("iNEXT")
    
    expect_error(
      speciesAccum(CTtable = mock_cams, recordTable = mock_recs, speciesCol = "WRONG_COL",
                   recordDateTimeCol = "DateTime", setupCol = "Setup_date", stationCol = "Station"),
      regexp = "Missing required columns in recordTable"
    )
    
    expect_error(
      speciesAccum(CTtable = mock_cams, recordTable = mock_recs, speciesCol = "Species",
                   recordDateTimeCol = "DateTime", setupCol = "Setup_date", stationCol = "Station",
                   assemblageCol = "WRONG_ASSEMBLAGE"),
      regexp = "assemblageCol not found in recordTable"
    )
  })
})

#' Calculate Residuals from MCMC Output of Occupancy Models
#' 
#' @description
#' Posterior Predictive Check (PPC) function that calculates Freeman-Tukey (FT) 
#' residuals, Pearson"s Chi-squared residuals, or deviance from MCMC output of 
#' occupancy models. This function compares observed data to simulated data from 
#' the posterior distribution to assess model fit.
#' 
#' 
#' @param y Observations as either a site vector or site by occasion matrix. For matrix 
#'   format, use NA for unsampled occasions.
#' @param p Array of posterior samples for detection probability (p). Dimensions should be 
#'   iterations by sites (by occasion optionally). For RN models, p should represent 
#'   individual-level detection (not conditional on local abundance).
#' @param psi Array of posterior samples for occupancy probability (psi). Dimensions should 
#'   be iterations by sites. For RN models, psi should represent expected abundance
#' @param model Character indicating model type: either "Occupancy" or "RN" 
#'   (Royle-Nichols).
#' @param type Type of residual to calculate: "FT" (Freeman Tukey), "PearChi2" 
#'   (Pearson Chi-squared), or "Deviance" (not technically a residual).
#' @param K Number of occasions as either a scalar or site vector. Calculated automatically 
#'   if y is a matrix.
#' @param z.cond Logical. If TRUE, new data is conditioned on estimated z (testing only 
#'   detection model fit). If FALSE, generates new z for each posterior sample 
#'   (testing complete model).
#' @param zhat Optional matrix with same dimensions as psi containing estimates of z from 
#'   the same model. If not provided, will be generated internally.
#' @param nmax Maximum site-level abundance (default = 20). Only used if model="RN". 
#'   Higher values increase computation time. Warning given if set too low.
#' @param return.residuals Logical. If TRUE (default), returns residuals along with 
#'   Bayesian p-value.
#' @param return.z Logical. If TRUE, returns z values conditional on y, and unconditional 
#'   z's if \code{z.cond = FALSE}. Note: if zhat is provided, the returned conditional-on-y z 
#'   values will be identical to those provided.
#'   
#'   
#'   
#' @return If return.residuals=TRUE (default), returns a list containing:
#'   \itemize{
#'     \item res.obs - residuals for observed data
#'     \item res.new - residuals for newly generated data
#'     \item BP - Bayesian p-value
#'   }
#'   If return.residuals=FALSE, returns only the Bayesian p-value.
#'   
#'   
#' @details
#' This function helps assess model fit for occupancy models using various types of residuals:
#' 
#' \subsection{Types of Residuals}{
#'   \itemize{
#'     \item Freeman-Tukey (FT):
#'           \deqn{R_j = (\sqrt{y_j} - \sqrt{E(y_j)})^2}
#'           Measures the squared difference between the square root 
#'           of observed detections and the square root of expected detections 
#'           at each site.
#'     \item Pearson Chi-squared:
#'           \deqn{R_j = \left(\frac{y_j - E(y_j)}{\sqrt{Var(y_j)}}\right)^2}
#'           Measures the squared difference between observed and 
#'           expected detections, standardized by the theoretical variance calculated 
#'           from the model parameters.
#'     \item Deviance:
#'           \deqn{R_j = -2\log[y_j|\theta_j, K_j]}
#'           Measures the contribution of each site to the overall model 
#'           likelihood, quantifying the discrepancy between observed data and model 
#'           predictions based on likelihood ratios
#'   }
#' 
#'   Where:
#'   \itemize{
#'     \item \eqn{y_j} is the number of detections of the species at site j, out of 
#'           \eqn{K_j} repeated surveys
#'     \item \eqn{E(y_j) = K_j p_j z_j}, with \eqn{p_j} = species detection probability 
#'           and \eqn{z_j} = occupancy state (1 if occupied, 0 otherwise)
#'     \item \eqn{Var(y_j) = p_j z_j (1 - p_j z_j) K_j}
#'     \item For Royle-Nichols occupancy models, the term \eqn{p_j z_j} is replaced with 
#'           \eqn{1 - (1 - r_j)^{N_j}}, where \eqn{r_j} = individual detection probability 
#'           and \eqn{N_j} = local abundance
#'     \item For Deviance, \eqn{\theta_j} is either occupancy and species detection probability 
#'           at site j (\eqn{\psi_j, p_j}) for regular occupancy models, or expected abundance 
#'           and individual detection probability (\eqn{\lambda_j, r_j}) for Royle-Nichols 
#'           occupancy models
#'   }
#' }
#' 
#' \subsection{Bayesian P-values}{
#'   The function calculates Bayesian p-values as a measure of model fit. These values:
#'   \itemize{
#'     \item Range from 0 to 1
#'     \item Values close to 0.5 suggest good model fit
#'     \item Values close to 0 or 1 suggest poor fit
#'     \item Are calculated by comparing observed residuals to residuals from 
#'           simulated data
#'   }
#' }
#' 
#' \subsection{Conditional vs Unconditional Assessment}{
#'   The z.cond parameter allows for two types of model assessment:
#'   \itemize{
#'     \item z.cond = TRUE: Tests only the detection component of the model, fixing 
#'           occupancy/abundance to estimates from the model, rather than generating them anew
#'     \item z.cond = FALSE: Tests the complete model, including both occupancy and 
#'           detection components
#'   }
#' }
#' 
#' @note 
#' FT and Chi-squared residuals have been extensively tested. Deviance calculations 
#' have undergone less testing and are only available for scenarios with constant 
#' detection probability across occasions. FT and Chi-squared residuals can handle 
#' varying detection probabilities.
#' 
#' 
#' @section Warning: This is a beta version of the function. While it has been tested extensively, 
#'   not all possible data configurations may have been captured in testing. This is 
#'   particularly true for:
#'   
#'   \itemize{
#'     \item Deviance calculations (type = "Deviance")
#'     \item Royle-Nichols models (model = "RN")
#'   }
#'   If you encounter issues with the function, please contact the package developers.
#' 
#' @author Rahel Sollmann
#' 
#' @references
#' Sollmann, Rahel. Occupancy models and the "good fit, bad prediction" dilemma. 
#' Ecology (submitted)
#' 
#' @importFrom stats dbinom rmultinom rpois
#' 
#' @export


PPC.residuals <- function(y, 
                          p, 
                          psi, 
                          model = c("Occupancy", "RN"),
                          type = c("FT", "PearChi2", "Deviance"),
                          K = NULL, 
                          z.cond = TRUE, 
                          zhat = NULL, 
                          nmax = 20,
                          return.residuals = TRUE,
                          return.z = TRUE) {
  
  #### some input data checks
  
  model <- match.arg(model)
  type  <- match.arg(type)
  
  if(model=="RN"){warning("For model=RN, matrix/array provided for
                         p must contain individual level detection probability 
                         (i.e., detection probability NOT conditional on local abundance) and
			psi must contain expected abundance.")
    ##check that nmax is sufficient
    if(dpois(nmax, max(psi))>0.01) warning("Probability of nmax at highest value 
                                            of expected abundance (psi) >0.01; 
                                            consider increasing nmax.") 
  }
  
  if(type =="Deviance" && is.matrix(y)) {stop("Deviance-based PPC currently only available for data aggregated over occasions.")}
  if (!is.null(K) && is.matrix(y)){warning("With occasion-explicit data, K is automatically derived from y (ncol); input value for K is ignored")}
  if (all(K==1)  && is.vector(y)){stop("With aggregated data, K must be >1 at least for some sites")}
  if (is.null(K)  && is.vector(y)){stop("With aggregated data, K must be supplied")}
  
  ###check that p,psi, zhat are given as arrays, not mcmc lists
  if(is.list(p) | is.list(psi) | is.list(zhat))stop("Posterior samples of p, psi (and zhat if provided) need to be in array form, not lists.")
  
  ###check number of posterior draws for p, psi, and, if applicable, zhat
  if(dim(p)[1] != dim(psi)[1]){stop("Dimensions of p and psi do not match; they need to have the same number of posterior draws (1st dimension)")}
  if(!is.null(zhat)){
    if(dim(zhat)[1] != dim(psi)[1]) stop("Dimensions of zhat do not match dimensions of psi and p; they need to have the same number of posterior draws (1st dimension)")
  }
  
  ###check spatial dimensions
  J <- ifelse(is.vector(y), length(y), nrow(y))
  if(dim(p)[2] != J |
     dim(psi)[2] != J){stop("Dimensions of p or psi do not match dimensions of data y; they need to have the same number of sites (2nd dimension)")}
  if(!is.null(zhat)){
    if(dim(zhat)[2] != dim(psi)[2]) stop("Dimensions of zhat do not match dimensions of psi and p; they need to have the same number of sites (2nd dimension)")
  }
  
  ###check temporal dimensions
  if(is.matrix(y)){
    
    # maxocc<-ncol(y)  # not used
    if(length(dim(p))<3){stop("If providing occasion specific observations, y, p must also be occasion-specific (ie, 3-dimensional array)")}
    
    ## set p for unsampled occasions to 0 - looks good
    if (sum(is.na(y))>0) {
      # message("Observations y contain NAs (missing data); detection probability
      #       for these sites and occasions will be set to 0 automatically.")   # I downgraded this to message. May not even be necessary at all?
      y.na <- matrix(FALSE, dim(y)[1], dim(y)[2])
      y.na[is.na(y)] <- TRUE
      y.arr <- aperm(array(y.na, c(dim(y), dim(p)[1])), c(3,1,2))
      p[y.arr]<-0
    }
  }
  
  if (is.vector(y) & length(dim(p))==3) stop("If p is occasion specific (3d array), y must also be occasion-specific (ie, matrix)")
  
  if(length(dim(p))==3 & is.matrix(y)){
    if(dim(p)[3]!=dim(y)[2])stop("Number of occasions in p does not match number of occasions in y; both objects need to have the same number of occasions.")
  }
  
  
  #### start calculations ##
  
  ##create var object that will be overwritten if type="PearChi2"
  var<-NULL
  
  ##identify known occupied sites
  if(is.vector(y)){
    n<-rep(as.numeric(y>0), each=nrow(psi))
    ##make vector of number of occasions
    if(length(K)==1){kvec=rep(K, length(as.vector(p)))}else{
      kvec<-rep(K, each=nrow(p))
    }
    ##set aggregated y to y
    y.agg<-y
  } 
  
  if(is.matrix(y)){
    #calculate aggregated y
    y.agg<-apply(y,1,sum, na.rm=TRUE)
    n<-rep(as.numeric(y.agg>0), each=nrow(psi))
    ##in this case, k is always 1
    kvec<-1
  }
  
  
  
  ##generate values of z for data if not supplied    ####
  if (is.null(zhat)){
    
    ## binary z for regular occu model
    if (model=="Occupancy"){
      
      if(is.matrix(y)){
        ##create vector of p(no detection)
        p0<-as.vector(apply(p, 1:2, function(x) prod(1-x)))
      } else {
        p0<-(1-as.vector(p))^kvec}
      
      ##p(presence given no detection)
      psi.eff<-(p0 * as.vector(psi))/
        ((p0 * as.vector(psi)) + (1-as.vector(psi)))
      z<-rbinom(prod(dim(psi)), 1, psi.eff)
      
      ##set to 1 for sites with observations
      z.real<-z
      z.real[which(n==1)]<-1
      
    } ## end if(occu) statement 
    
    ##same for RN model 
    if (model=="RN"){
      ##set range of possible abundance values
      N.poss<-0:nmax
      lam.arr<-array(psi, c(dim(psi), nmax+1))
      
      if(is.matrix(y)){
        ## 4d
        r.arr<-array(p, c(dim(p), nmax+1))
        N.arr<-array(rep(N.poss, each=prod(dim(p))), c(dim(p), nmax+1))
        
        # site by occ data - first repeat for posterior samples
        y1<-aperm(array(y, c(dim(y), nrow(psi)) ), c(3,1,2))
        # then repeat that for each possible N
        y.arr<-array(y1, c(dim(y1), nmax+1))
        
        #conditional-on-N p detection
        p.arr<-1-(1-r.arr)^N.arr
        ##prob (y|z) - product over occasions (na.rm bc y could have NAs)
        prob<-apply(dbinom(y.arr, 1, p.arr), c(1, 2, 4),prod, na.rm=TRUE)
        ##prob (z) - only needs one 3d slice of 4d N array
        probz<-dpois(N.arr[,,1,], lam.arr)
        
      } else { ##for vector data format
        ##make arrays
        r.arr<-array(p, c(dim(p), nmax+1))
        y.arr<-array(rep(y.agg, each=nrow(psi)), c(dim(psi), nmax+1))
        N.arr<-array(rep(N.poss, each=prod(dim(p))), c(dim(p), nmax+1))
        k.arr<-array(matrix(kvec, dim(p)), c(dim(p), nmax+1))
        
        #conditional-on-N p detection
        p.arr<-1-(1-r.arr)^N.arr
        ##prob (y|z)
        prob<-dbinom(y.arr, k.arr, p.arr)
        ##prob (z)
        probz<-dpois(N.arr, lam.arr)
      }
      
      ##generate new N conditional on y
      z.real<-apply(prob*probz, 1:2, function(x)which(rmultinom(1,1,x)>0)-1)
      z.real<-as.vector(z.real)
    } # end RN if statement
    
  } else { #end if(is.null(zhat))
    ##if provided from model, use those - works for both model types
    z.real=as.vector(zhat)}
  
  
  
  ## generate new z from model if z.cond=FALSE; otherwise set z.new=z.real   #####
  
  if (z.cond){ z.new<-z.real} else{
    if (model == "RN") {
      z.new<-rpois(prod(dim(psi)), as.vector(psi))} else {
        z.new<-rbinom(prod(dim(psi)), 1, as.vector(psi))}
  }
  
  
  
  ## generate new observations and calculate expectations for aggregated y   ####
  if(is.vector(y)){
    
    ##get p conditional on state
    if(model =="Occupancy") {p.eff<-as.vector(p)*z.real}
    if(model =="RN") {p.eff<-1-(1-as.vector(p))^z.real}
    
    # if(!expected){
    Ey <- (kvec* p.eff) 
    
    ##only calculate variance if type="Pearson" or "PearChi2"
    ##taken from McCullaugh and Nelder book (not Warton code)
    if (type %in% c("Pearson", "PearChi2")){
      var <- kvec*(p.eff)  *(1 - p.eff)
    }
    
    if(z.cond) {
      #generate new observations from p conditioned on same state
      ynew<-rbinom(prod(dim(p)), kvec, p.eff)
      Ey.new<-Ey
      var.new<-var
      p.eff.new<-p.eff
    } else {
      if(model=="Occupancy") {p.eff.new<-as.vector(p)*z.new}
      if(model=="RN") {p.eff.new<-1-(1-as.vector(p))^z.new}
      #generate new observations from p conditioned on new state
      ynew<-rbinom(prod(dim(p)), kvec, p.eff.new)
      
      Ey.new<-kvec* p.eff.new
      
      if (type %in% c("Pearson", "PearChi2")){
        var.new<-kvec*(p.eff.new)*(1-p.eff.new)
      }
    }
    
  } #end if y is vector
  
  
  if(is.matrix(y)){
    
    ##get p conditional on state
    if(model=="Occupancy") {p.eff<-as.vector(p)*rep(z.real, ncol(y))}
    if(model=="RN") {p.eff<-1-(1-as.vector(p))^rep(z.real, ncol(y))}
    
    #sum p"s over occasions 
    p.agg<-apply(array(p.eff, dim(p)), 1:2, sum)
    
    # if(!expected){
    Ey<-as.vector(p.agg)
    
    if (type %in% c("Pearson", "PearChi2")){
      pvar<-apply(array(p.eff, dim(p)), 1:2, function(x) sum(x*(1-x)))
      var<-as.vector(pvar)
    }
    
    if(z.cond) {
      ##generate new data from p conditioned on same state
      yn3d<-array(rbinom(prod(dim(p)), kvec, p.eff),
                  dim(p))
      ynew<-as.vector(apply(yn3d, 1:2, sum))
      
      Ey.new<-Ey
      var.new<-var
      p.eff.new<-p.eff
    } else{
      
      if(model=="Occupancy") {p.eff.new<-as.vector(p)*rep(z.new, ncol(y))}
      if(model=="RN") {p.eff.new<-1-(1-as.vector(p))^rep(z.new, ncol(y))}
      
      p.agg.new<-apply(array(p.eff.new, dim(p)), 1:2, sum)
      
      Ey.new<-as.vector(p.agg.new)
      
      ##generate new data from p conditioned on new state
      yn3d<-array(rbinom(prod(dim(p)), kvec, p.eff.new),
                  dim(p))
      ynew<-as.vector(apply(yn3d, 1:2, sum))
      
      if (type %in% c("Pearson", "PearChi2")){
        pvar<-apply(array(p.eff.new, dim(p)), 1:2, function(x) sum(x*(1-x)))
        var.new<-as.vector(pvar)
      }
    }
  }
  
  ## calculate residuals, deviance
  
  if(type=="FT"){
    res.new<-(sqrt(ynew) - sqrt(Ey.new) )^2
    res<-( sqrt(rep(y.agg, each=nrow(psi))) - 
             sqrt(Ey) ) ^2
  }
  
  if(type=="PearChi2"){
    res.new<-( ( ynew - Ey.new )/
                 ( sqrt(var.new) +0.0001) )^2
    res<- (( rep(y.agg, each=nrow(psi)) - Ey ) /
             ( sqrt(var) + 0.0001 ) )^2
  }
  
  
  if(type == "Deviance"){
    
    if(model=="Occupancy"){
      
      # if (is.vector(y)){ ##use dbinom to account for permutations!
      #for new data
      pd.new<-dbinom(ynew, kvec, as.vector(p)) * as.vector(psi) + #always
        (1-as.vector(psi))*(ynew==0) #only sites with no detections
      
      res.new<- -2*log(pd.new)
      
      ##for observed data 
      yarr<-as.vector(rep(y.agg, each=nrow(psi)))
      pd<-dbinom(yarr, kvec, as.vector(p)) * as.vector(psi) + #always
        (1-as.vector(psi))*(yarr==0) #only sites with no detections
      
      res<--2*log(pd)
    } #end if occu 
    
    if(model=="RN"){
      
      
      ##calculate prob, probz if not already calculated
      ## ie, if zhat was provided to function
      if (!is.null(zhat)){
        N.poss<-0:nmax
        lam.arr<-array(psi, c(dim(psi), nmax+1))
        
        ##make arrays
        r.arr<-array(p, c(dim(p), nmax+1))
        y.arr<-array(rep(y.agg, each=nrow(psi)), c(dim(psi), nmax+1))
        N.arr<-array(rep(N.poss, each=prod(dim(p))), c(dim(p), nmax+1))
        k.arr<-array(matrix(kvec, dim(p)), c(dim(p), nmax+1))
        
        #conditional-on-N p detection
        p.arr<-1-(1-r.arr)^N.arr
        ##prob (y|z)
        prob<-dbinom(y.arr, k.arr, p.arr)
        ##prob (z)
        probz<-dpois(N.arr, lam.arr)
      }
      
      ##for observed data
      #integrate out state by summing over possible states
      pd<-apply(prob*probz, 1:2, sum)
      res<--2*log(as.vector(pd))
      
      ##for new data
      y.arr.new <- array(ynew, c(dim(psi), nmax+1))
      prob.new <- dbinom(y.arr.new, k.arr, p.arr)
      pd.new <- apply(prob.new*probz, 1:2, sum)
      res.new <- -2*log(as.vector(pd.new))
      
    } #end if RN
  } #end if deviance
  
  res.out <- matrix(res, dim(psi))
  resnew.out <- matrix(res.new, dim(psi))
  
  
  pval <-  mean(apply(res.out, 1, sum) >  apply(resnew.out, 1, sum) )
  
  out <- pval
  if(return.residuals){
    out<-list(BP = pval, res.obs = res.out, res.new = resnew.out)} 
  if(return.z){
    if(z.cond){
      out<-list(BP = pval, res.obs=res.out, res.new = resnew.out,
                zhat = matrix(z.real, dim(psi)))} else {
                  out <- list(BP = pval, res.obs = res.out, res.new = resnew.out,
                              zhat = matrix(z.real, dim(psi)),
                              z.new = matrix(z.new, dim(psi)))
                }
    
  }
  return(out)
}





#' Calculate Community-Level Posterior Predictive Checks for Occupancy Models
#' 
#' @description
#' A wrapper function that applies Posterior Predictive Checks (PPC) across multiple 
#' species in a community occupancy model (from \code{\link{communityModel}}.
#' It calculates species-specific and community-level Bayesian p-values to assess 
#' model fit. The function accepts both \code{\link{predict}} output format and 
#' list format for model parameters.
#' 
#' @details
#' This function extends the single-species PPC analysis to the community level by:
#' \itemize{
#'   \item Applying residual calculations to each species in the community
#'   \item Aggregating results to assess community-level model fit
#'   \item Providing both species-specific and community-level diagnostics
#' }
#' 
#' @param p Either a 4D array [stations, species, iterations, occasions] from 
#' \code{\link{predict}} or a list of 3D arrays [iterations, stations, occasions], 
#' one per species
#' @param psi Either a 3D array [stations, species, iterations] from 
#' \code{\link{predict}} or a list of 2D arrays [iterations, stations], one per species
#' @param y List of detection histories, one matrix/vector per species
#' @param K Number of occasions as either a scalar or site vector. Calculated automatically 
#'   if y is a list of matrices.
#' @param zhat List of z estimate matrices, one per species (optional). Each matrix 
#'   should follow the format specified in \code{\link{PPC.residuals}}.
#' @param z.cond Logical. If TRUE, new data is conditioned on estimated z (testing only 
#'   detection model fit). If FALSE, generates new z for each posterior sample 
#'   (testing complete model).
#' @param model Character indicating model type ("Occupancy" or "RN")
#' @param type Character indicating residual type ("FT", "PearChi2", or "Deviance")
#' @param input_format Character indicating format of p and psi ("predict" or "lists")
#' @param show_progress Logical; whether to show a progress bar during computation 
#' (if package pbapply is available)
#' @param return.residuals Logical. If TRUE, returns species-specific residuals along 
#'   with Bayesian p-values. If FALSE, returns only the p-values.
#' @param ... Additional arguments passed to \code{\link{PPC.residuals}}.
#'   
#'   
#' @details
#' This function provides flexibility in input formats to accommodate different workflows:
#' \itemize{
#'   \item Direct output from camtrapR's predict() function (4D/3D arrays)
#'   \item Lists of species-specific arrays (for custom workflows)
#' }
#' 
#' 
#' 
#' @return If return.residuals=TRUE, returns a list containing:
#'   \itemize{
#'     \item bp_values - Data frame with species-specific and community-level 
#'           Bayesian p-values
#'     \item species_results - List containing the complete PPC.residuals output 
#'           for each species
#'   }
#'   If return.residuals=FALSE, returns only the data frame of Bayesian p-values.
#'   
#' @seealso \code{\link{PPC.residuals}} for details on the underlying single-species 
#'   calculations
#'   
#' @author Rahel Sollmann
#' 
#' @export
#' 
#'
#' @examples 
#' \dontrun{
#' 
#' # Create and fit model 
#' data("camtraps")
#' 
#' # create camera operation matrix
#' camop_no_problem <- cameraOperation(CTtable      = camtraps,
#'                                     stationCol   = "Station",
#'                                     setupCol     = "Setup_date",
#'                                     retrievalCol = "Retrieval_date",
#'                                     hasProblems  = FALSE,
#'                                     dateFormat   = "dmy"
#' )
#' 
#' data("recordTableSample")
#' 
#' # make list of detection histories
#' species_to_include <- unique(recordTableSample$Species)
#' 
#'   DetHist_list <- detectionHistory(
#'     recordTable         = recordTableSample,
#'     camOp                = camop_no_problem,
#'     stationCol           = "Station",
#'     speciesCol           = "Species",
#'     recordDateTimeCol    = "DateTimeOriginal",
#'     species              = species_to_include,
#'     occasionLength       = 7,
#'     day1                 = "station",
#'     datesAsOccasionNames = FALSE,
#'     includeEffort        = TRUE,
#'     scaleEffort          = TRUE,
#'     timeZone             = "Asia/Kuala_Lumpur"
#' )
#' 
#' 
#' # create some fake covariates for demonstration
#' sitecovs <- camtraps[, c(1:3)]
#' sitecovs$elevation <- c(300, 500, 600)   
#' 
#' # scale numeric covariates
#' sitecovs[, c(2:4)] <- scale(sitecovs[,-1])
#' 
#' 
#' # bundle input data for communityModel
#' data_list <- list(ylist = DetHist_list$detection_history,
#'                   siteCovs = sitecovs,
#'                   obsCovs = list(effort = DetHist_list$effort))
#' 
#' 
#' # create community model for JAGS
#' modelfile1 <- tempfile(fileext = ".txt")
#' mod.jags <- communityModel(data_list,
#'                            occuCovs = list(fixed = "utm_y", ranef = "elevation"),
#'                            detCovsObservation = list(fixed = "effort"),
#'                            intercepts = list(det = "ranef", occu = "ranef"),
#'                            modelFile = modelfile1)
#' 
#' summary(mod.jags)
#' 
#' # fit in JAGS
#' fit.jags <- fit(mod.jags,
#'                 n.iter = 1000,
#'                 n.burnin = 500,
#'                 chains = 3)   
#' summary(fit.jags)
#' 
#' 
#' 
#' # create predictions for p and psi
#' draws <- 100
#' 
#' p <- predict(object = mod.jags,
#'              mcmc.list = fit.jags,
#'              type = "p_array",
#'              draws = draws)
#' # output is in order [station, species, draw, occasion]
#'  
#'  psi <- predict(object = mod.jags,
#'                 mcmc.list = fit.jags,
#'                 type = "psi_array",
#'                 draws = draws)
#'  # output is in order [station, species, draw]
#' 
#' 
#' 
#' ppc_comm <- PPC.community(
#'   p = p,
#'   psi = psi,
#'   y = mod.jags@input$ylist, 
#'   model = "Occupancy",
#'   type = "FT")
#'   
#' # Bayesian p-values
#'  ppc_comm$BP
#'  
#'    
#'  str(ppc_comm$residuals)
#'  
#'  # get individual species PPC results
#'  ppc_species <- ppc_comm$residuals[[1]] # first species
#'  
#'  plot(apply(ppc_species$res.obs, 2, mean), apply(ppc_species$res.new, 2, mean),
#'  xlab = "Observed residuals",
#'  ylab = "Predicted residuals"
#'  )
#'  
#'  abline(0,1)  # diagonal line is not visible due to tiny data set
#'   
#'  }

PPC.community <- function(p, 
                          psi, 
                          y, 
                          input_format = c("predict", "lists"),
                          K = NULL, 
                          model = c('Occupancy', 'RN'),
                          zhat = NULL, 
                          z.cond = TRUE, 
                          type = c('FT', 'PearChi2', 'Deviance'),
                          return.residuals = TRUE,
                          show_progress = TRUE,
                          ...){
  
  
  input_format <- match.arg(input_format)
  model <- match.arg(model)
  type <- match.arg(type)
  
  if (!is.list(y)) stop("y must be a list of detection histories")
  n_species <- length(y)
  if (n_species == 0) stop("y list is empty")
  
  
  # Input validation and format detection
  if (input_format == "predict") {
    # Expect arrays from predict()
    if (!is.array(p) || !is.array(psi)) 
      stop("When input_format='predict', p and psi must be arrays")
    
    if (!(length(dim(p)) == 4 && length(dim(psi)) == 3)) {
      stop("When input_format='predict', expecting 4D array for p and 3D array for psi")
    }
    
    
    # check array dimensions
    p_dims <- dim(p)
    psi_dims <- dim(psi)
    
    if (length(p_dims) != 4) 
      stop("p must be 4D array [iterations, species, stations, occasions]")
    if (length(psi_dims) != 3) 
      stop("psi must be 3D array [iterations, species, stations]")
    
    # Check dimensions match
    if (p_dims[2] != n_species || psi_dims[2] != n_species)
      stop("Number of species in p/psi arrays doesn't match y list")
    if (p_dims[1] != psi_dims[1])
      stop("Number of stations in p and psi arrays don't match")
    
    
    # reorder arrays
    # p:   Reorder from [stations, species, iterations, occasions] -> [iterations, stations, species, occasions]
    p <- aperm(p, c(3, 1, 2, 4))    
    # psi: Reorder from [stations, species, iterations]  -> [iterations, stations, species]
    psi <- aperm(psi, c(3, 1, 2))   
    
    # convert arrays to lists (by species)
    p_list <- apply(p, 3, function(x) x, simplify = FALSE)
    psi_list <- apply(psi, 3, function(x) x, simplify = FALSE)
    
  } else {
    
    # Validation for list input
    if (!is.list(p) || !is.list(psi))
      stop("When input_format='lists', p and psi must be lists")
    if (length(p) != n_species || length(psi) != n_species)
      stop("Length of p and psi lists must match y list")
    
    
    # Check each element in lists
    for(i in seq_len(n_species)) {
      if (!is.array(p[[i]]) || length(dim(p[[i]])) != 3)
        stop("Each element in p list must be a 3D array [iterations, stations, occasions]")
      if (!is.array(psi[[i]]) || length(dim(psi[[i]])) != 2)
        stop("Each element in psi list must be a 2D array [iterations, stations]")
      
      # Check matching dimensions within each species
      if (dim(p[[i]])[1] != dim(psi[[i]])[1])
        stop(sprintf("Number of posterior draws don't match in p and psi for species %d", i))
      if (dim(p[[i]])[2] != dim(psi[[i]])[2])
        stop(sprintf("Number of stations don't match in p and psi for species %d", i))
    }
    
    p_list <- p
    psi_list <- psi
  }
  
  # Validate consistency
  n_species <- length(y)  # ylist length is our reference
  if (length(p_list) != n_species || length(psi_list) != n_species) {
    stop("Number of species in p, psi, and y must match")
  }
  
  
  y_list <- y
  
  # run with progress indicator if requested and if pbapply is available
  mapFun <- if (show_progress && requireNamespace("pbapply", quietly = TRUE)) {
    pbapply::pbmapply
  } else {
    base::mapply
  }
  
  
  # Create zhat_list that's either the provided list or a list of NULLs
  zhat_list <- if(is.list(zhat)) {
    zhat
  } else {
    replicate(length(y_list), NULL, simplify = FALSE)
  }
  
  # run PPC.residuals on all species
  BP.spec <- mapFun(function(p_sp, psi_sp, y_sp, z_sp) {
    PPC.residuals(p = p_sp, 
                  psi = psi_sp, 
                  y = y_sp, 
                  K = K, 
                  model = model,
                  zhat = z_sp,  # will be NULL if zhat wasn't provided
                  z.cond = z.cond, 
                  type = type,
                  return.residuals = TRUE,
                  ...)
  }, p_list, psi_list, y_list, zhat_list, SIMPLIFY = FALSE)
  
  
  # obtain overall Bayesian p-value
  # sum residuals over sites, cbind over all species, then sum across species
  obs.all<-apply(do.call(cbind, lapply(BP.spec, function(x) apply(x$res.obs, 1, sum))), 1, sum)
  new.all<-apply(do.call(cbind, lapply(BP.spec, function(x) apply(x$res.new, 1, sum))), 1, sum)
  
  BP.comm<-mean(obs.all>new.all)
  
  ## bundle into data frame
  if(is.null(names(y_list))){
    spnames<-paste0('Species', 1:length(y_list))} else {
      spnames<- names(y_list)}
  
  df.out<-data.frame(Species=c(spnames, 'Community'),
                     BP = c(sapply(BP.spec, function(x)x$BP), BP.comm))
  if(return.residuals){
    names(BP.spec)<-spnames
    return(list(BP=df.out, residuals=BP.spec))} else{
      return(df.out)
    }
}

# for function activityRadial

# the following 3 functions were adapted from package "plotrix" version 3.7-4
#   radial.plot
#   radial.grid
#   boxed.labels

# thanks to Jim Lemon et al.!


.radial.plot<-function(lengths,radial.pos=NULL,labels=NA,label.pos=NULL,
                       radlab=FALSE,start=0,clockwise=FALSE,rp.type="r",label.prop=1.1,main="",
                       xlab="",ylab="",line.col=par("fg"),lty=par("lty"),lwd=par("lwd"),
                       mar=c(2,2,3,2),show.grid=TRUE,show.grid.labels=4,show.radial.grid=TRUE,
                       rad.col="gray",grid.col="gray",grid.bg="transparent",grid.left=FALSE,
                       grid.unit=NULL,point.symbols=1,point.col=par("fg"),show.centroid=FALSE,
                       radial.lim=NULL,radial.labels=NULL,boxed.radial=FALSE,poly.col=NA,
                       add=FALSE,...) {
  
  if(is.null(radial.lim)) radial.lim<-range(lengths)
  length.dim<-dim(lengths)
  if(is.null(length.dim)) {
    npoints<-length(lengths)
    nsets<-1
    lengths<-matrix(lengths,nrow=1)
  }
  else {
    npoints<-length.dim[2]
    nsets<-length.dim[1]
    lengths<-as.matrix(lengths)
  }
  lengths<-lengths-radial.lim[1]
  lengths[lengths<0]<-NA
  if(is.null(radial.pos))
    radial.pos<-seq(0,pi*(2 - 2 * (rp.type != "l")/npoints),length.out=npoints)
  radial.pos.dim<-dim(radial.pos)
  if(is.null(radial.pos.dim))
    radial.pos<-matrix(rep(radial.pos,nsets),nrow=nsets,byrow=TRUE)
  else radial.pos<-as.matrix(radial.pos)
  if(rp.type == "l" && clockwise == TRUE) start<-pi/2
  if(clockwise) radial.pos<--radial.pos
  if(start) radial.pos<-radial.pos+start
  if(show.grid) {
    if(length(radial.lim) < 3) grid.pos<-pretty(radial.lim)
    else grid.pos<-radial.lim
    if(grid.pos[1] < radial.lim[1]) grid.pos<-grid.pos[-1]
    maxlength<-max(grid.pos-radial.lim[1])
    angles<-seq(0,1.96*pi,by=0.04*pi)
  }
  else {
    grid.pos<-NA
    maxlength<-diff(radial.lim)
  }
  oldpar<-par("xpd","mar","pty")
  if(!add) {
    par(mar=mar,pty="s")
    plot(c(-maxlength,maxlength),c(-maxlength,maxlength),type="n",axes=FALSE,
         main=main,xlab=xlab,ylab=ylab)
    if(is.null(label.pos)) {
      if(is.null(labels)) nlpos<-ifelse(npoints > 8,8,npoints)
      else {
        if(is.na(labels[1])) nlpos<-ifelse(npoints > 8,8,npoints)
        else nlpos<-length(labels)
      }
      label.pos<-seq(0,pi*(2-2/nlpos),length.out=nlpos)
    }
    if(show.grid) {
      .radial.grid(labels=labels,label.pos=label.pos,radlab=radlab,
                  radial.lim=radial.lim,start=start,clockwise=clockwise,
                  label.prop=label.prop,grid.pos=grid.pos,rad.col=rad.col,
                  grid.col=grid.col,grid.bg=grid.bg,
                  show.radial.grid=show.radial.grid)
    }
  }
  par(xpd=TRUE)
  # stretch everything out to the correct length
  if(length(line.col) < nsets) line.col<-1:nsets
  if(length(rp.type) < nsets) rp.type<-rep(rp.type,length.out=nsets)
  if(length(point.symbols) < nsets)
    point.symbols<-rep(point.symbols,length.out=nsets)
  if(length(point.col) < nsets) point.col<-rep(point.col,length.out=nsets)
  if(length(poly.col) < nsets) poly.col<-rep(poly.col,length.out=nsets)
  if(length(lty) < nsets) lty<-rep(lty,length.out=nsets)
  if(length(lwd) < nsets) lwd<-rep(lwd,length.out=nsets)
  for(i in 1:nsets) {
    if(nsets > 1) {
      linecol<-line.col[i]
      polycol<-poly.col[i]
      pointcol<-point.col[i]
      pointsymbols<-point.symbols[i]
      ltype<-lty[i]
      lwidth<-lwd[i]
    }
    else {
      linecol<-line.col
      polycol<-poly.col
      pointcol<-point.col
      pointsymbols<-point.symbols
      ltype<-lty
      lwidth<-lwd
    }
    # split up rp.type if there is a combination of displays
    rptype<-unlist(strsplit(rp.type[i],""))
    if(match("s",rptype,0)) {
      if(is.null(pointsymbols)) pointsymbols<-i
      if(is.null(pointcol)) pointcol<-i
    }
    # get the vector of x positions
    xpos<-cos(radial.pos[i,])*lengths[i,]
    # get the vector of y positions
    ypos<-sin(radial.pos[i,])*lengths[i,]
    # plot radial lines if rp.type == "r"    
    if(match("r",rptype,0))
      segments(0,0,xpos,ypos,col=linecol,lty=ltype,lwd=lwidth,...)
    if(match("p",rptype,0))
      polygon(xpos,ypos,border=linecol,col=polycol,lty=ltype,
              lwd=lwidth,...)
    if(match("s",rptype,0))
      points(xpos,ypos,pch=pointsymbols,col=pointcol,...)
    if(match("l",rptype,0))
      lines(xpos,ypos,lty=ltype,lwd=lwidth,col=linecol,...)
    if(show.centroid) {
      if(match("p",rptype,0)) {
        nvertices<-length(xpos)
        # first get the "last to first" area component
        polygonarea<-xpos[nvertices]*ypos[1] - xpos[1]*ypos[nvertices]
        for(vertex in 1:(nvertices-1))
          polygonarea<-
          polygonarea+xpos[vertex]*ypos[vertex+1]-xpos[vertex+1]*ypos[vertex]
        polygonarea<-polygonarea/2
        centroidx<-
          (xpos[nvertices]+xpos[1])*(xpos[nvertices]*ypos[1]-xpos[1]*ypos[nvertices])
        centroidy<-
          (ypos[nvertices]+ypos[1])*(xpos[nvertices]*ypos[1]-xpos[1]*ypos[nvertices])
        for(vertex in 1:(nvertices-1)) {
          centroidx<-centroidx + (xpos[vertex]+xpos[vertex+1])*
            (xpos[vertex]*ypos[vertex+1]-xpos[vertex+1]*ypos[vertex])
          centroidy<-centroidy + (ypos[vertex]+ypos[vertex+1])*
            (xpos[vertex]*ypos[vertex+1]-xpos[vertex+1]*ypos[vertex])
        }
        points(centroidx/(6*polygonarea),centroidy/(6*polygonarea),
               col=point.col[i],pch=point.symbols[i],cex=2,...)
        
      }
      else
        points(mean(xpos),mean(ypos),col=pointcol,pch=pointsymbols,
               cex=2,...)
    }
  }
  if(show.grid.labels && !add) {
    if(show.grid.labels%%2) {
      ypos<-grid.pos-radial.lim[1]
      xpos<-rep(0,length(grid.pos))
      if(show.grid.labels==1) ypos<--ypos
    }
    else {
      xpos<-grid.pos-radial.lim[1]
      ypos<-rep(0,length(grid.pos))
      if(show.grid.labels==2) xpos<--xpos
    }
    if(is.null(radial.labels)) radial.labels<-grid.pos
    if(!is.null(grid.unit))
      radial.labels[length(grid.pos)]<-
        paste(radial.labels[length(grid.pos)],grid.unit)
    if(boxed.radial)
      .boxed.labels(xpos,ypos,radial.labels,border=FALSE,
                   cex=par("cex.lab"))
    else text(xpos,ypos,radial.labels,cex=par("cex.lab"))
  }
  invisible(oldpar)
}

.radial.grid<-function(labels=NA,label.pos=NULL,radlab=FALSE,radial.lim=NULL,
                       start=0,clockwise=FALSE,label.prop=1.1,grid.pos,rad.col="gray",
                       grid.col="gray",grid.bg="transparent",show.radial.grid=TRUE,
                       start.plot=FALSE) {
  
  if(start.plot) {
    radial.span<-diff(radial.lim)
    plot(0,type="n",xlim=c(-radial.span,radial.span),
         ylim=c(-radial.span,radial.span),xlab="",ylab="",axes=FALSE)
  }
  par(xpd=TRUE)
  if(is.null(label.pos)) label.pos<-seq(0,1.8*pi,length=9)
  if(!is.null(labels)) {
    if(is.na(labels[1])) labels<-as.character(round(label.pos,2))
  }
  if(clockwise) label.pos<--label.pos
  if(start) label.pos<-label.pos+start
  # display the circumferential grid
  angles<-seq(0,1.96*pi,by=0.04* pi)
  for(i in seq(length(grid.pos),1,by=-1)) {
    xpos<-cos(angles)*(grid.pos[i]-radial.lim[1])
    ypos<-sin(angles)*(grid.pos[i]-radial.lim[1])
    polygon(xpos,ypos,border=grid.col,col=grid.bg)
  }
  maxlength<-max(grid.pos)-radial.lim[1]
  # display the radial grid
  if(show.radial.grid) {
    xpos<-cos(label.pos)*maxlength
    ypos<-sin(label.pos)*maxlength
    segments(0,0,xpos,ypos,col=rad.col)
    xpos<-cos(label.pos)*maxlength
    ypos<-sin(label.pos)*maxlength
  }
  # display the circumferential labels
  if(!is.null(labels)) {
    xpos<-cos(label.pos)*maxlength*label.prop
    ypos<-sin(label.pos)*maxlength*label.prop
    if(radlab) {
      for(label in 1:length(labels)) {
        labelsrt<-(180*label.pos[label]/pi)+180*
          (label.pos[label] > pi/2 && label.pos[label] < 3*pi/2)
        text(xpos[label],ypos[label],labels[label],
             cex=par("cex.axis"),srt=labelsrt)
      }
    }
    else
      .boxed.labels(x = xpos, y = ypos,labels,ypad=0.7,border=FALSE,cex=par("cex.axis"))
  }
  par(xpd = FALSE)
}



.boxed.labels<-function (x, y = NA, labels,
                         bg = ifelse(match(par("bg"), "transparent", 0), "white", par("bg")),
                         border = TRUE, xpad = 1.2, ypad = 1.2, 
                         srt = 0, cex = 1, adj = 0.5, xlog=FALSE, ylog=FALSE, ...) {
  
  oldpars <- par(c("cex", "xpd"))
  par(cex = cex, xpd = TRUE)
  box.adj <- adj + (xpad - 1) * cex * (0.5 - adj)
  if (srt == 90 || srt == 270) {
    bheights <- strwidth(labels)
    theights <- bheights * (1 - box.adj)
    bheights <- bheights * box.adj
    lwidths <- rwidths <- strheight(labels) * 0.5
  }
  else {
    lwidths <- strwidth(labels)
    rwidths <- lwidths * (1 - box.adj)
    lwidths <- lwidths * box.adj
    bheights <- theights <- strheight(labels) * 0.5
  }
  args <- list(x = x, y = y, labels = labels, srt = srt, adj = adj, 
               col = ifelse(colSums(col2rgb(bg) * c(1, 1.4, 0.6)) < 
                              350, "white", "black"))
  args <- modifyList(args, list(...))
  if(xlog){
    xpad<-xpad*2
    xr<-exp(log(x) - lwidths * xpad)
    xl<-exp(log(x) + lwidths * xpad)
  }
  else{
    xr<-x - lwidths * xpad
    xl<-x + lwidths * xpad
  }
  if(ylog){
    ypad<-ypad*2
    yb<-exp(log(y) - bheights * ypad)
    yt<-exp(log(y) + theights * ypad)
  }
  else{
    yb<-y - bheights * ypad
    yt<-y + theights * ypad
  }	
  rect(xr, yb, xl, yt, col = bg, border = border)
  do.call(text, args)
  par(cex = oldpars)
}

#' Show Exif metadata of JPEG images or other image or video formats
#' 
#' The function will return metadata values, metadata tag names and group names
#' of Exif metadata of JPEG images or other formats.
#' 
#' Many digital cameras record information such as ambient temperature or moon
#' phase under maker-specific tag names in Exif metadata of JPEG images. In
#' addition, many technical information are stored in Exif metadata. In order
#' to extract those information from images and add them to the record tables
#' created by the functions \code{\link{recordTable}} and
#' \code{\link{recordTableIndividual}}, the tag names must be known so they can
#' be passed to these functions via the \code{additionalMetadataTags} argument.
#' 
#' By default the function returns both metadata tag names and the metadata
#' group they belong to (via argument \code{returnTagGroup}). This is helpful
#' to unambiguously address specific metadata tags, because different groups
#' can contain tags of identical names, which may cause problems executing the
#' functions \code{\link{recordTable}} and \code{\link{recordTableIndividual}}.
#' The format is "GROUP:tag", e.g. "EXIF:Flash".
#' 
#' @param inDir character. Directory containing camera trap images sorted into
#' station subdirectories (e.g. inDir/StationA/)
#' @param whichSubDir integer or character. Either number or name of
#' subdirectory of \code{inDir} in which to look for an image
#' @param fileName character. A filename, either the file name of an image in
#' \code{inDir} or a full path with file name (in which case \code{inDir} is
#' not needed)
#' @param returnMetadata deprecated and ignored
#' @param returnTagGroup deprecated and ignored
#' 
#' @return A data frame containing three columns: metadata tag group, tag name,
#' and values.
#' 
#' @author Juergen Niedballa
#' 
#' @seealso \code{\link{recordTable}}
#' 
#' @references Phil Harvey's ExifTool \url{https://exiftool.org/} \cr
#' 
#' @examples
#' 
#' 
#' \dontrun{
#' 
#' wd_images_ID <- system.file("pictures/sample_images_species_dir", package = "camtrapR")
#' 
#' # specify directory, camtrapR will automatically take first image from first subdirectory
#' exifTagNames(inDir          = wd_images_ID)
#' 
#' # specify subdirectory by name, camtrapR will use first image
#' exifTagNames(inDir          = wd_images_ID,
#'              whichSubDir    = "StationA")
#' 
#' # specifying fileName only (line break due to R package policy)
#' exifTagNames(fileName       = file.path(wd_images_ID, "StationC", "TRA", 
#'                                         "StationC__2009-05-02__00-10-00(1).JPG"))
#'                                         
#' # specify inDir and fileName
#' exifTagNames(inDir          = wd_images_ID,
#'              fileName       = file.path("StationC", "TRA", "StationC__2009-05-02__00-10-00(1).JPG"))
#'              
#' # it also works this way
#' exifTagNames(inDir          = file.path(wd_images_ID, "StationC", "TRA"),
#'              fileName       = "StationC__2009-05-02__00-10-00(1).JPG")
#'              
#' 
#' # with tagged sample images
#' wd_images_ID_tagged <- system.file("pictures/sample_images_indiv_tag", package = "camtrapR")
#' exifTagNames(inDir          = wd_images_ID_tagged)
#' 
#' }
#' 
#' @export exifTagNames
#' 
exifTagNames <- function(inDir, 
                         whichSubDir = 1, 
                         fileName,
                         returnMetadata = "DEPRECATED", 
                         returnTagGroup = "DEPRECATED"){
  
  
  if(returnMetadata != "DEPRECATED") message("Argument returnMetadata is ignored since version 2.0.0")
  if(returnTagGroup != "DEPRECATED") message("Argument returnTagGroup is ignored since version 2.0.0")
  
  if(Sys.which("exiftool") == "") stop("cannot find ExifTool", call. = FALSE)
  
  if(hasArg(fileName)){
    
    if(!hasArg(inDir)){
      file.tmp <- fileName
    } else {
      if(!dir.exists(inDir)) stop("Could not find inDir:\n", inDir, call. = FALSE)
      file.tmp <- file.path(inDir, fileName)
    }
    if(!file.exists(file.tmp)) stop(paste(file.tmp, "not found"))
    
  } else { # if fileName is not defined
    
    if(!hasArg(inDir))     stop("fileName is not defined. Please define 'inDir'", call. = FALSE)
    if(!dir.exists(inDir)) stop("Could not find inDir:\n", inDir, call. = FALSE)
    
    if(is.numeric(whichSubDir)){
      dir.tmp0 <- list.dirs(inDir, recursive = FALSE, full.names = TRUE)
      if(length(dir.tmp0) == 0) stop("inDir has no subdirectories")
      dir.tmp <- dir.tmp0[whichSubDir]
      if(is.na(dir.tmp)) stop(paste0("Cound not find subdirectory ", whichSubDir, " in inDir: ", inDir))
    }
    if(is.character(whichSubDir)){
      dir.tmp <- file.path(inDir, whichSubDir)
    }
    if(!dir.exists(dir.tmp)) stop(paste("the specified subdirectory does not exist. Check argument 'whichSubDir.'\n", 
                                        dir.tmp), call. = FALSE)
    
    # filename of first image in specified directory
    file.tmp <- list.files(dir.tmp,
                           full.names = TRUE,
                           pattern    = ".JPG$|.jpg$",
                           recursive  = TRUE)[1]
    if(length(file.tmp) == 0) stop(paste("found no jpg in ", dir.tmp, sep = "\n"))
    
  }
  
  message(paste("Metadata of:", file.tmp, sep = "\n"))

  # generate exiftool calls
  
  # exiftool arguments: 
  # -s1: print tag  names instead of description
  # -t:  tab-delimited output
  # -G: return Group names # 0 (general location), 1 (specific location), 2 (category), 3 (document number), 4 (instance number), 5 (metadata path) or 6 (EXIF/TIFF format)
  
  groupHeadings_general  <- "-G0"    # can possibly also be -G0:1 
  #groupHeadings_specific <- "-G1"   # when using this separately from -G0, output is sorted differently (by group)
  
  
  command.tmp1  <- paste('exiftool -s1 -t ', groupHeadings_general, ' "', file.tmp, '"', sep = "")
  # run without -s1 option to extract tag descriptions instead of tag names
  #command.tmp2 <- paste('exiftool -t ', groupHeadings_general, ' "', file.tmp, '"', sep = "")
  
  # run exiftool
  metadata.tmp1 <- system(command.tmp1, intern=TRUE)
  #metadata.tmp2 <- system(command.tmp2, intern=TRUE)

  # convert output to data frames
  out1 <- read.table(text = metadata.tmp1, header = FALSE, 
                     col.names = c("tag_group", "tag_name", "value"), sep = "\t",
                     stringsAsFactors = FALSE, fill = TRUE)
  # out2 <- read.table(text = metadata.tmp2, header = FALSE, 
  #                    col.names = c("tag_group", "tag_description", "value"), sep = "\t",
  #                    stringsAsFactors = FALSE, fill = TRUE)

  # make sure out1 and out2 are ordered identically
  # if(any(out1$value != out2$value))  stop("Order of exiftool-extracted metadata differs. Please report this bug.")
  # WHY DOES THIS FAIL ON CRAN (debian & fedora, but not on solaris/windows?)
  
  # combine data frames
  out <- cbind(tag_group       = out1$tag_group,
               tag_name        = out1$tag_name,
               # tag_description = out2$tag_description,
               value           = out1$value)
  return(as.data.frame(out))
}

#' Copy and rename images based on camera trap station ID and creation date
#' 
#' The function renames and copies raw camera trap images into a new location
#' where they can be identified. Images are renamed with camera trap station
#' ID, camera ID (optional), creation date and a numeric identifier for images
#' taken within one minute of each other at a given station. Station ID and
#' camera ID are derived from the raw image directory structure. The creation
#' date is extracted from image metadata using ExifTool.
#' 
#' Setting up the correct raw image directory structure is necessary for
#' running the function successfully. \code{inDir} is the main directory that
#' contains camera trap station subdirectories (e.g. inDir/StationA). If one
#' camera was deployed per station and no camera subdirectories are used within
#' station directories, \code{hasCameraFolders} can be set to \code{FALSE}. If
#' more than one camera was deployed at stations, there must be subdirectories
#' for the individual camera traps within the station directories (e.g.
#' "inDir/StationA/CameraA1" and "inDir/StationA/CameraA2"). Even if only some
#' stations had multiple cameras, all station will need camera subdirectories.
#' The argument \code{hasCameraFolders} must be \code{TRUE}. Within the camera
#' subdirectories, the directory structure is irrelevant.
#' 
#' Renaming of images follows the following pattern: If \code{hasCameraFolders}
#' is TRUE, it is: "StationID__CameraID__Date__Time(Number).JPG", e.g.
#' "StationA__CameraA1__2015-01-31__18-59-59(1).JPG". If
#' \code{hasCameraFolders} is FALSE, it is:
#' "StationID__Date__Time(Number).JPG", e.g.
#' "StationA__2015-01-31__18-59-59(1).JPG".
#' 
#' The purpose of the number in parentheses is to prevent assigning identical
#' file names to images taken at the same station (and camera) in the same
#' second, as can happen if cameras take sequences of images. It is a
#' consecutive number given to all images taken at the same station by the same
#' camera within one minute. The double underscore "__" in the image file names
#' is for splitting and extracting information from file names in other
#' functions (e.g. for retrieving camera IDs in \code{\link{recordTable}} if
#' camera subdirectories are not preserved (\code{keepCameraSubfolders =
#' FALSE})).
#' 
#' The function finds all JPEG images (optionally, also videos) and extracts 
#' the image timestamp from the image metadata using ExifTool (digiKam database 
#' for videos) and copies the images with new file names
#' into \code{outDir}, where it will set up a directory structure based on the
#' station IDs and, if required by \code{keepCameraSubfolders = TRUE}, camera
#' IDs (e.g. outDir/StationA/ or outDir/StationA/CameraA1).
#' 
#' \code{copyImages} can be set to FALSE to simulate the renaming and check the
#' file names of the renamed images without copying. If you are handling large
#' number of images (>e.g., 100,000), the function may take some time to run.
#' 
#' Argument \code{video} is a named list 4 items 
#' (\code{file_formats}, \code{dateTimeTag}, (\code{db_directory}, 
#' \code{db_filename}). Video date/time is read from video metadata stored in
#' the digiKam database. Hence, \code{inDir} must be in your digiKam database.
#' 
#' The items of argument \code{video} are:
#' 
#' \tabular{ll}{ \code{file_formats} \tab The video formats to extract (include
#' "jpg" if you want .JPG image metadata) \cr \code{dateTimeTag} \tab the
#' metadata tag to extract date/time from (use \code{\link{exifTagNames}} to
#' find out which tag is suitable) \cr \code{db_directory} \tab The directory
#' containing digiKam database  \cr
#' \code{db_filename} \tab The digiKam database file in \code{db_directory} \cr }
#' 
#' See the examples in \code{\link{recordTable}} for for how to specify the 
#' argument \code{video}.
#' 
#' 
#' @param inDir character. Directory containing camera trap images sorted into
#' station subdirectories (e.g. inDir/StationA/)
#' @param outDir character. Directory into which the renamed images will be
#' copied
#' @param hasCameraFolders logical. Do the station directories in \code{inDir}
#' have camera subdirectories (e.g. "inDir/StationA/Camera1")?
#' @param keepCameraSubfolders logical. Should camera directories be preserved
#' as subdirectories of \code{outDir} (e.g. "outDir/StationA/CameraA1")?
#' @param createEmptyDirectories logical. If station or camera directories are
#' empty, should they be copied nevertheless (causing empty directories in
#' \code{outDir}, but preserving the whole directory structure)?
#' @param copyImages logical. Copy images to \code{outDir}?
#' @param writecsv logical. Save a data frame with a summary as a .csv? The csv
#' will be saved in \code{outDir}.
#' @param video list. Contains information on how to handle video data
#' (optional). See details.
#' 
#' @return A \code{data.frame} with original directory and file names, new
#' directory and file names and an indicator for whether images were copied
#' successfully.
#' 
#' @author Juergen Niedballa
#' 
#' @references Phil Harvey's ExifTool \url{https://exiftool.org/}
#' 
#' @examples
#' 
#' 
#' 
#'   \dontrun{
#'   
#' ### "trial" run. create a table with file names after renaming, but don't copy images.
#' 
#' # first, find sample image directory in package directory:
#' wd_images_raw <- system.file("pictures/raw_images", package = "camtrapR")
#' 
#' # because copyImages = FALSE, outDir does not need to be defined
#' renaming.table <- imageRename(inDir               = wd_images_raw,     
#'                               hasCameraFolders = FALSE,
#'                               copyImages          = FALSE,
#'                               writecsv            = FALSE
#'   )
#' 
#' 
#'   
#' ### a real example in which images are copied and renamed 
#' 
#'   # define raw image location
#' wd_images_raw <- system.file("pictures/raw_images", package = "camtrapR") 
#' 
#'   # define destination for renamed images
#' wd_images_raw_renamed <- file.path(tempdir(), "raw_images_renamed")       
#' 
#' 
#'   # now we have to define outDir because copyImages = TRUE
#' renaming.table2 <- imageRename(inDir               = wd_images_raw,
#'                                outDir              = wd_images_raw_renamed,       
#'                                hasCameraFolders    = FALSE,
#'                                copyImages          = TRUE,
#'                                writecsv            = FALSE
#'   )
#'   
#'   # show output files
#'   list.files(wd_images_raw_renamed, recursive = TRUE)
#'   
#'   # output table
#'   renaming.table2
#'   
#'   }
#' 
#' @export imageRename
#' 
imageRename <- function(inDir,
                        outDir,
                        hasCameraFolders,
                        keepCameraSubfolders,
                        createEmptyDirectories = FALSE,
                        copyImages = FALSE,
                        writecsv = FALSE,
                        video){
  
  wd0 <- getwd()
  on.exit(setwd(wd0))
  
  file.sep <- .Platform$file.sep
  
  stationCol <- "Station"
  cameraCol  <- "Camera"
  
  # check inDir / outDir
  stopifnot(length(inDir) == 1)
  if(!dir.exists(inDir)) stop("Could not find inDir:\n", inDir, call. = FALSE)
  
  if(hasArg(outDir)){
    stopifnot(length(outDir) == 1)
    if(isTRUE(all(unlist(strsplit(tolower(inDir), split = file.sep)) %in%
                  unlist(strsplit(tolower(outDir), split = file.sep))))) stop("outDir may not be identical to or a subdirectory of inDir", call. = FALSE)
  }
  if(copyImages == TRUE){
    
    if(any(c(grep("/$", inDir) == 1, grep("/$", outDir) == 1))) stop("inDir and outDir may not end with /", call. = FALSE)
    
  } else {
    if(isTRUE(grepl("/$", inDir))) stop("inDir may not end with /", call. = FALSE)
  }
  
  # Check input file path lengths
  max_length <- 255
  all_input_files <- list.files(inDir, recursive = TRUE, full.names = TRUE)
  if (any(nchar(all_input_files) > max_length)) {
    long_files <- all_input_files[nchar(all_input_files) > max_length]
    stop(paste("The following input file paths exceed", max_length, "characters:", 
               paste(long_files, collapse = ", ")), call. = FALSE)
  }
  
  
  # list of subdirectories of inDir
  dirs       <- list.dirs(inDir, full.names = TRUE,  recursive = FALSE)
  dirs_short <- list.dirs(inDir, full.names = FALSE, recursive = FALSE)
  
  # Ignore the ".dtrash" folder if present (on Mac)
  if(".dtrash" %in% dirs_short) {
    message("Ignoring .dtrash folder.")
    dirs <- dirs[dirs_short != ".dtrash"]
    dirs_short <- dirs_short[dirs_short != ".dtrash"]
  }
  
  # Ignore  the ".mysql.digikam" folder if present
  if(".mysql.digikam" %in% dirs_short) {
    message("Ignoring .mysql.digikam folder.")
    dirs <- dirs[dirs_short != ".mysql.digikam"]
    dirs_short <- dirs_short[dirs_short != ".mysql.digikam"]
  }

  # Check for Umlaute in path names
  paths_to_check <- inDir
  if (hasArg(outDir)) {
    paths_to_check <- c(paths_to_check, outDir)
  }
  
  unicode_vector <- c("\u{00E4}", "\u{00F6}", "\u{00FC}", "\u{00C4}", "\u{00D6}", "\u{00DC}")

  if (any(grepl(paste0("[", paste(unicode_vector, collapse = ""), "]"), list.dirs(paths_to_check)))) {
    warning("Umlaute (", 
            paste(unicode_vector[1:3], collapse = ", "),
    ") detected in path names. This may cause issues on certain operating systems.", call. = FALSE)
  }
  
  
  if(length(dirs) == 0) stop("inDir contains no station directories", call. = FALSE)
  
  # make sure none is empty
  fileext_pattern <- ".jpg$|.JPG$"
  names(fileext_pattern) <- "jpg"
  if(hasArg(video))  fileext_pattern <- c(fileext_pattern, 
                                          sapply(video$file_formats, FUN = function(x) paste0(tolower(x), "$|", toupper(x), "$"))
                                          )
  
  list_n_files <- lapply(dirs, list.files, recursive = TRUE)   # need to adapt this to include video formats also (in fileext_pattern)
  list_n_files <- lapply(list_n_files, FUN = function(x) sapply(fileext_pattern, grepl, x))

  
  # if(any(unlist(lapply(list_n_files, length)) == 0) && !hasArg(video)){
  #     warning("at least one station directory contains no JPEGs:  ", paste(dirs_short[which(lapply(list_n_files, length) == 0)], collapse = "; "), call. = FALSE, immediate. = TRUE)
  # }
  
  if(any(sapply(list_n_files, sum) == 0)) {
    if(hasArg(video)){
      warning("at least one station directory contains no JPEGs or ", paste(video$file_formats, collapse = ' / '), " files:  ", paste(dirs_short[which(sapply(list_n_files, sum) == 0)], collapse = "; "), call. = FALSE, immediate. = TRUE)
    } else {
      warning("at least one station directory contains no JPEGs:  ", paste(dirs_short[which(sapply(list_n_files, sum) == 0)], collapse = "; "), call. = FALSE, immediate. = TRUE)
    }
  }
  
  
  
  stopifnot(is.logical(createEmptyDirectories))
  
  # remove dirs and dirs_short if they contain no images (and if user overrides default createEmptyDirectories = FALSE)
  if(isFALSE(createEmptyDirectories)){
    stations2remove  <- which(lapply(list_n_files, length) == 0)
    if(length(stations2remove) >= 1){
      dirs        <- dirs      [-stations2remove]
      dirs_short  <- dirs_short[-stations2remove]
    }
  }
  # for pretty console messages
  max_nchar_station <- max(nchar(dirs_short))
  
  # process video argument (if present)
  if(hasArg(video)){
    # Check if the video argument is a list with the required elements
    if(!is.list(video) || !all(c("file_formats", "dateTimeTag") %in% names(video))) {
      stop("The 'video' argument must be a list with elements 'file_formats' and 'dateTimeTag'")
    }
    
    video_out <- processVideoArgument(IDfrom = "metadata",   # to ensure digikam DB is read
                                      video  = video)
    digiKam_data <- video_out$digiKam_data
    file_formats <- video_out$file_formats
  } else {
    file_formats <- "jpg"   # jpg, as the default, if video not requested
  }
  
  if(Sys.which("exiftool") == "") stop("cannot find ExifTool", call. = FALSE)
  
  # check call for consistency
  stopifnot(is.logical(copyImages))
  stopifnot(is.logical(writecsv))
  stopifnot(is.logical(hasCameraFolders))
  
  if(isTRUE(writecsv) & isFALSE(hasArg(outDir))) stop("writecsv is TRUE. Please specify outDir", call. = FALSE)
  
  if(isTRUE(hasCameraFolders)){
    if(!hasArg(keepCameraSubfolders)) stop("If hasCameraFolders is TRUE, keepCameraSubfolders must be defined")
    if(!is.logical(keepCameraSubfolders)) stop("atgument keepCameraSubfolders must be logical (TRUE or FALSE)")
  } else {
    keepCameraSubfolders <- FALSE
  }
  if(hasArg(hasCameraFolders) & hasArg(keepCameraSubfolders)){
    if(keepCameraSubfolders == TRUE & hasCameraFolders == FALSE){stop("If hasCameraFolders is FALSE, keepCameraSubfolders must be FALSE too", call. = FALSE)}
  }
  
  
  # function body
  
  copy.info.table <- data.frame()
  
  for(i in 1:length(dirs_short)){     # for all non-empty directories
    
    # check if there are any images not in camera trap subfolders
    if(hasCameraFolders == TRUE && length(list.files(dirs[i], pattern = ".jpg$|.JPG$", recursive = FALSE)) >= 1){
      stop(paste("Directory ", dirs[i], " contains images not sorted into Camera Trap subfolders. Check argument 'hasCameraFolders'"), call. = FALSE)
    }
    
    # create Exiftool call (photos only)
    # command.tmp  <- paste('exiftool -q -f -t -r -Directory -FileName -EXIF:DateTimeOriginal -ext JPG "', dirs[i], '"', sep = "")
    
    # create Exiftool call (photos and videos)
    command.tmp  <- paste('exiftool -q -f -t -r -Directory -FileName -EXIF:DateTimeOriginal', 
                         ifelse(hasArg(video),
                                 paste(" -", video$dateTimeTag, sep = ""), 
                                ""),    # if video requested, video date time tag
                          paste(" -ext", file_formats, collapse = " ", sep = " "),    # requested file extensions
                          ' "', dirs[i], '"', sep = "")
    
    colnames.tmp <- c("Directory", "FileName", "DateTimeOriginal")
    if(hasArg(video)) colnames.tmp <- c(colnames.tmp, video$dateTimeTag)
    
    # run exiftool to get image date and time
    metadata.tmp <- runExiftool(command.tmp = command.tmp, colnames.tmp = colnames.tmp)
    
    
    if(length(metadata.tmp) == 0){
      length.tmp <- length(list.files(dirs[i], pattern = ".jpg$|JPG$", ignore.case = TRUE, recursive = TRUE))
      warning(paste(dirs_short[i], "seems to contain no images;", " found", length.tmp, "jpgs"), call. = FALSE, immediate. = TRUE)    # give message if station directory contains no jpgs
    } else {
      
      message(paste(formatC(dirs_short[i], width = max_nchar_station, flag = "-"), ": ", formatC(nrow(metadata.tmp), width = 4), " images", 
                    makeProgressbar(current = i, total = length(dirs_short)), sep = ""))
      
      if(isTRUE(hasCameraFolders)){
        # sort by camera, then image name (in case cameras are not sorted alphabetically)
        metadata.tmp <- metadata.tmp[order(metadata.tmp$Directory, metadata.tmp$FileName),]
        
        filenames_by_subfolder <- lapply(list.dirs(dirs[i], full.names =TRUE, recursive = FALSE),
                                         FUN = list.files, pattern = ".jpg$|.JPG$", recursive = TRUE, ignore.case = TRUE)
        metadata.tmp$CameraID <- rep(list.dirs(dirs[i], full.names = FALSE, recursive = FALSE),
                                     times = sapply(filenames_by_subfolder, length))
        metadata.tmp$Station <- rep(dirs_short[i], times = nrow(metadata.tmp))
        colnames(metadata.tmp)[grep("CameraID", colnames(metadata.tmp))] <- cameraCol
        colnames(metadata.tmp)[grep("Station", colnames(metadata.tmp))]  <- stationCol
      } else {
        metadata.tmp$CameraID <- "NA"                                                    # "camera" name if no camera id available (only station ids)
        metadata.tmp$Station <- rep(dirs_short[i], times = nrow(metadata.tmp))
        colnames(metadata.tmp)[grep("CameraID", colnames(metadata.tmp))] <- cameraCol
        colnames(metadata.tmp)[grep("Station", colnames(metadata.tmp))]  <- stationCol
      }
      
      # make time readable
      if(hasArg(video)){
        metadata.tmp <- addVideoDateTimeOriginal(metadata.tmp = metadata.tmp, video = video)
        metadata.tmp$DateTimeOriginal <- lubridate::ymd_hms(metadata.tmp$DateTimeOriginal)
      } else {
        metadata.tmp$DateTimeOriginal <- as.POSIXct(strptime(x = as.character(metadata.tmp$DateTimeOriginal),
                                                           format = "%Y:%m:%d %H:%M:%S", tz = "UTC"))
      }
      metadata.tmp$DateTimeOriginal2 <- as.POSIXlt(metadata.tmp$DateTimeOriginal)
      
      
      # exclude images for which no DateTimeOriginal was found
      if(any(is.na(metadata.tmp$DateTimeOriginal))){
        na.date.rows <- which(is.na(metadata.tmp$DateTimeOriginal))
        warning(paste("could not read DateTimeOriginal tag of: \n",
                      paste(paste(metadata.tmp$Directory,  metadata.tmp$FileName, sep = file.sep)[na.date.rows], collapse = "\n")),
                call. = FALSE, immediate. = TRUE)
        metadata.tmp <- data.frame(metadata.tmp, DateReadable = NA)
        metadata.tmp$DateReadable[-na.date.rows] <- TRUE
        metadata.tmp$DateReadable[ na.date.rows] <- FALSE
        rm(na.date.rows)
      } else {
        metadata.tmp$DateReadable <- TRUE
      }

      
      # rearrange column order
      metadata.tmp <- metadata.tmp[,c("Directory",  "FileName", stationCol, cameraCol,
                                      "DateTimeOriginal", "DateTimeOriginal2", "DateReadable")]
      
      
      # find images taken within 1 minute of one another (to append number)
      metadata.tmp$DateTimeOriginal2$sec <- 0
      
      if(isTRUE(hasCameraFolders)){
        metadata.tmp.split <- split(x = metadata.tmp, f = list(metadata.tmp[,cameraCol],
                                                               as.POSIXct(metadata.tmp$DateTimeOriginal2)
        ), drop = TRUE)
      } else {
        metadata.tmp.split <- split(x = metadata.tmp, f = list(metadata.tmp[,stationCol],
                                                               as.POSIXct(metadata.tmp$DateTimeOriginal2)
        ), drop = TRUE)
      }
      
      metadata.tmp.split2 <- lapply(metadata.tmp.split, FUN = function(X){
        X2 <- X[with(X, order(DateTimeOriginal)), ]
        cbind(X2, minute.append = seq(from = 1, to = nrow(X2), by = 1))
      })
      
      metadata.tmp2 <- do.call("rbind", metadata.tmp.split2)    # reassemble
      if(any(metadata.tmp$DateReadable == FALSE)){
        metadata.tmp2 <- rbind(cbind(metadata.tmp[metadata.tmp$DateReadable == FALSE,], minute.append = NA), metadata.tmp2)
      }
      rm(metadata.tmp.split, metadata.tmp.split2)
      
      # convert time object to character vector and format for outfilename
      time.tmp <- gsub(pattern = ":", replacement = "-", metadata.tmp2$DateTimeOriginal)
      time.tmp2 <- gsub(pattern = " ", replacement = "__", time.tmp)
      metadata.tmp2$DateTime_for_filename <- time.tmp2
      rm(time.tmp, time.tmp2)
      
      # create outfilename
      if(hasArg(outDir)){
        if(keepCameraSubfolders == TRUE){
          metadata.tmp2$outDir <- file.path(outDir, metadata.tmp2[,stationCol], metadata.tmp2[,cameraCol])
        } else {
          metadata.tmp2$outDir <- file.path(outDir, metadata.tmp2[,stationCol])
        }
      }
      
      if(isTRUE(hasCameraFolders)){
        metadata.tmp2$filename_new <- paste(paste(metadata.tmp2[,stationCol],
                                                  metadata.tmp2[,cameraCol],
                                                  paste(metadata.tmp2$DateTime_for_filename, "(",metadata.tmp2$minute.append, ")", sep = ""),
                                                  sep = "__"),
                                            ".", tools::file_ext(metadata.tmp2$FileName), 
                                            #"JPG", 
                                            sep = "")
      } else {
        metadata.tmp2$filename_new <- paste(paste(metadata.tmp2[,stationCol],
                                                  paste(metadata.tmp2$DateTime_for_filename, "(",metadata.tmp2$minute.append, ")", sep = ""),
                                                  sep = "__"),
                                            ".", tools::file_ext(metadata.tmp2$FileName), 
                                            # ".JPG", 
                                            sep = "")
      }
      
      copy.info.table <- rbind(copy.info.table, metadata.tmp2)
      rm(metadata.tmp2)
    }
  }
  
  # error if no date/time tag is readable
  if(!any(copy.info.table$DateReadable)) stop("could not read DateTimeOriginal tag of any image. Check if the DateTimeOriginal tag is present in metadata with exifTagNames(). If not, fixDateTimeOriginal() might help.",
                                              call. = FALSE)
  # warning if at least one date/time tag is not readable
  if(!all(copy.info.table$DateReadable)) warning(paste("Some DateTimeOriginal tags are unreadable, e.g. \n", 
                                                   paste(apply(copy.info.table[!copy.info.table$DateReadable, c("Directory", "FileName")],  MARGIN = 1, FUN = paste, collapse = file.sep), collapse = "\n")),
                                              call. = FALSE)
  
  # Check output file path lengths
  if (hasArg(outDir)) {
    all_output_files <- file.path(copy.info.table$outDir, copy.info.table$filename_new)
    if (any(nchar(all_output_files) > max_length)) {
      long_files <- all_output_files[nchar(all_output_files) > max_length]
      stop(paste("The following output file paths would exceed", max_length, "characters:", 
                 paste(long_files, collapse = ", ")), call. = FALSE)
    }
  }
  
  # create directory structure in outDir
  if(isTRUE(copyImages)){
    
    if(isFALSE(keepCameraSubfolders))   dir2create <- file.path (outDir, dirs_short)    # outDir with station subdirectories
    
    if(isTRUE(keepCameraSubfolders)) {
      # create list of directories to create
      dirs_recursive <- lapply(dirs, FUN = list.dirs, recursive = TRUE, full.names = FALSE)
      names(dirs_recursive) <- dirs_short
      for(xyz in 1:length(dirs_recursive)){
        dirs_recursive[[xyz]] <- file.path(names(dirs_recursive)[[xyz]], dirs_recursive[[xyz]])
      }
      dirs_recursive <- unlist(dirs_recursive)
      
      dirs_recursive2 <- dirs_recursive[lapply(strsplit(dirs_recursive, file.sep), length) == 1 |  # find all entries with 1 items (all Stations) and
                                          lapply(strsplit(dirs_recursive, file.sep), length) == 2]   # find all entries with 2 items (Station + Camera)
      dir2create <- file.path (outDir, dirs_recursive2)                                                                         # outDir with station and camera subdirectories
    }
    
    
    sapply(dir2create, FUN = dir.create, recursive = TRUE, showWarnings = FALSE)           # create directories (recursively)
  }
  
  # check if renamed images exist already
  if(hasArg(outDir)) {
    copy.info.table$fileExistsAlready <- file.exists(file.path(copy.info.table$outDir, copy.info.table$filename_new))
  } else {
    copy.info.table$fileExistsAlready <- FALSE
  }
  
  # exit if no new images to be copied
  if(all(copy.info.table$fileExistsAlready)){
    stop("No new images to be copied. Exiting.", call. = FALSE)
  }
  
  # message if there are images in outDir already
  if(any(copy.info.table$fileExistsAlready)) {
    message(paste(sum(copy.info.table$fileExistsAlready), "out of", nrow(copy.info.table), "images exist already in outDir. They will not be copied"))
  }
  
  # copy images
  if(isTRUE(copyImages)){
    
    if(any(copy.info.table$fileExistsAlready)) {
      switch(menu(choices = c("Copy only images that are not in outDir", "Copy nothing"), title =  "outDir is not empty. What should I do?"),
             proceed <- TRUE ,
             proceed <- FALSE)
    } else {
      proceed <- TRUE
    }
    
    if(isTRUE(proceed)){
      # find items to copy
      items_to_copy <- which(copy.info.table$DateReadable == TRUE & copy.info.table$fileExistsAlready == FALSE)
      
     
      message(paste("copying", length(items_to_copy), "images to", outDir, " ... This may take some time."))
      
      copy.info.table$CopyStatus[items_to_copy] <- file.copy(from      = apply(copy.info.table[items_to_copy, c("Directory", "FileName")],  MARGIN = 1, FUN = paste, collapse = file.sep),
                                                             to        = apply(copy.info.table[items_to_copy, c("outDir", "filename_new")], MARGIN = 1, FUN = paste, collapse = file.sep),
                                                             overwrite = FALSE)
      # if not all images are copied, set copystatus of the images that were not copied to FALSE
      if(length(items_to_copy) < nrow(copy.info.table)) copy.info.table$CopyStatus[-items_to_copy] <- FALSE
    } else {
      copy.info.table$CopyStatus <- FALSE     # if copyImages = TRUE & proceed = FALSE
    }
  } else {
    copy.info.table$CopyStatus <- FALSE       # if copyImages = FALSE
  }
  
  rownames(copy.info.table) <- NULL     # remove rownames
  
  # remove unused columns
  copy.info.table <- copy.info.table[,-which(names(copy.info.table) %in%
                                               c("DateTimeOriginal2", "DateTime_for_filename", "minute.append"))]
  # save table
  if(writecsv == TRUE){
    dir.create(outDir, recursive = TRUE, showWarnings = FALSE)
    setwd(outDir)
    write.csv(copy.info.table, file =file.path(outDir, paste0("_renaming_table_", Sys.Date(), ".csv")),
              row.names = FALSE)
  }
  return(copy.info.table)
}

context("imageRename")
library(camtrapR)


wd_images_ID_species <- system.file("pictures/sample_images_species_dir", package = "camtrapR")


test_that("imageRename throws error if inDir doesn't exist", {
  expect_error(imageRename(inDir = paste(wd_images_ID_species, "blabla", sep = "")),
               "Could not find inDir")
})


test_that("imageRename throws error if inDir doesn't have subdirectories", {
  expect_error(imageRename(inDir = list.dirs(wd_images_ID_species)[3]),   # pointing to StationA/PBE, which doesn't have subdirectories
               "inDir contains no station directories")
})

# Function to create raster predictions from community occupancy model in JAGS for all species at once


# not implemented:
# - nested  / other random effects
# - allow for alpha if it only has site covariates?


setGeneric("predict", function(object, ...) {})




predictionMapsCommunity <- function(object,
                                    mcmc.list,
                                    type,
                                    draws = 1000,
                                    level = 0.95,
                                    interval = c("none", "confidence"),
                                    x = NULL,
                                    aoi = NULL,
                                    speciesSubset,
                                    batch = FALSE,
                                    seed = NULL) 
{
  
  # type <- match.arg(type, choices = c("psi_array", "psi", "richness", "pao"))
  type <- match.arg(type, choices = c("psi_array", "psi", "richness", "pao", "abundance", "lambda_array",
                                      "p_array"))
  interval <- match.arg(interval, choices = c("none", "confidence"))

  
  if(.hasSlot(object, "model")) {
    if(object@model != "RN"){
      if(type %in% c("abundance", "lambda_array")) stop(paste0("type = '", type, "' is only implemented in Royle-Nichols models. The current model is a standard occupancy model"))
    }
  } else {
      stop("This legacy model is not supported from camtrapR v3.0 and higher")
  }
  
  # subset occupancy (beta) parameters
  if(type == "p_array") {
    submodel <- "det"
  } else {
    submodel <- "state"
  }
  
  submodel <- match.arg(submodel, choices = c("det", "state"))
  
  if(submodel == "state") {
    keyword_submodel <- "^beta"
    keyword_submodel_short <- "beta"
  }
  if(submodel == "det") {
    keyword_submodel <- "^alpha"
    keyword_submodel_short <- "alpha"
  } 
  

  # get intercept information for submodel
  cov_info_intercept <- object@covariate_info[object@covariate_info$submodel == submodel & object@covariate_info$param == "intercept",]
  
  # get covariate information for submodel
  cov_info_subset <- object@covariate_info[object@covariate_info$submodel == submodel & object@covariate_info$param == "param",]
  
  

  # subset parameters of submodel
  stopifnot(all(cov_info_subset$coef %in% object@params))
  params_submodel <- object@params[grep(keyword_submodel, object@params)]
  
  
  if(!is.null(seed)){
    stopifnot(is.numeric(seed))
    set.seed(seed)
  }
  
  # define function for inverse logit (logit to probability)
  ilogit <- function(x) 1 / (1 + exp(-x))
  
  
  # subset posterior matrix to number of draws
  posterior_matrix <- as.matrix(mcmc.list)
  if(hasArg(draws)) {
    if(nrow(posterior_matrix) > draws){
      selected_rows <- sort(sample(1:nrow(posterior_matrix), draws))
      # print(selected_rows)
      posterior_matrix <- posterior_matrix[selected_rows, , drop = FALSE]
    } else {
      message(paste0("draws (", draws, ") is greater than or equal to the number of available samples. Using all samples (", nrow(posterior_matrix), ")."))
    }
  } 
  
  # subset posterior matrix to current submodel
  posterior_matrix <- posterior_matrix[, grep(keyword_submodel, colnames(posterior_matrix)), drop = F]
  
  # if(nrow(posterior_matrix) > 1000) message("More than 1000 posterior samples. Watch RAM usage")
  
  params_covariate <- cov_info_subset$covariate
  # if(length(params_covariate) == 0) stop ("No covariates found", call. = F)
  
  
  list_responses <- list()
  
  
  # if x is not defined, use the site covariates from the model input
  if(is.null(x)) x <- object@input$siteCovs
  
  # convert raster to covariate data frame
  if(inherits(x, c("SpatRaster", "RasterStack"))) {
    if(inherits(x, "RasterStack")){
    warning("x is a RasterStack. Please use SpatRaster (from the terra package) in the future. Convert via: 
rast(YourRaster)")
    }

    raster_template <- terra::rast(x, nlyrs = 1, vals = NA)
    
    values_to_predict_all <- as.data.frame(terra::values(x))
  } else {
    if(is.data.frame(x)) {
      values_to_predict_all <- x
    } else {
      stop("x must be a SpatRaster, data.frame, or NULL")
    }
  }

  
  # identify cells with values
  index_not_na <- which(apply(values_to_predict_all[, cov_info_subset$covariate[cov_info_subset$covariate_type == "siteCovs"], drop = F], 
                              1, 
                              FUN = function(x) all(!is.na(x))))
  
  if(hasArg(aoi)) {
    if(inherits(x, "data.frame")) stop("aoi can only be defined if x is a SpatRaster (not a data.frame).")
    if(inherits(aoi, "SpatRaster")) {
      
      which_not_na_aoi    <- which(!is.na(terra::values(aoi)))
      index_not_na <- intersect(index_not_na, which_not_na_aoi)
      
      aoi2 <- terra::rast(aoi)
      terra::values(aoi2) <- NA
      terra::values(aoi2) [index_not_na]  <- 1
      
    } else {
      stop("aoi must be a SpatRaster")
    }
  }
  
  # subset to cells / sites where all covariates have values
  # if submodel has site covariates
  if(nrow(cov_info_subset) >= 1 & any(cov_info_subset$covariate_type == "siteCovs")) {
    values_to_predict_subset <- values_to_predict_all[index_not_na, 
                                                      cov_info_subset$covariate [cov_info_subset$covariate_type == "siteCovs"], 
                                                      drop = F]
  } else {
    # if no covariates just keep all values (won't affect predictions)
    values_to_predict_subset <- values_to_predict_all[, , drop = F]
  }

  # store dimensions of output
  n_cells_all <- nrow(values_to_predict_subset)  # dimension 1
  n_species   <- object@data$M                     # dimension 2
  n_samples   <- nrow(posterior_matrix)            # dimension 3
  
  n_occasions <- object@data$maxocc
  n_stations  <- object@data$J
  
  
  if(type != "p_array") {
    
    
    ## occupancy / abundance, by batch  ----
    if(isTRUE(batch) | is.numeric(batch)) {
      
      # create batches of values to predict on
      split_dataframe_into_batches <- function(df, batch_size) {
        n_rows <- nrow(df)
        n_batches <- ceiling(n_rows / batch_size)
        batches <- vector("list", n_batches)
        for (i in seq_len(n_batches)) {
          start_row <- (i - 1) * batch_size + 1
          end_row <- min(i * batch_size, n_rows)
          batches[[i]] <- df[start_row:end_row, ]
        }
        return(batches)
      }
      
      values_to_predict_subset_list <- split_dataframe_into_batches(values_to_predict_subset, ifelse(isTRUE(batch), 1000, batch))
      
      
      if (!requireNamespace("abind", quietly = TRUE)) {
        stop(paste("Please install the package abind to run this function with batch =", batch))
      }
      
      # container for output of batches
      array_list <- lapply(values_to_predict_subset_list, FUN = function(values_to_predict_subset_i) {
        
        n_cells_batch <- nrow(values_to_predict_subset_i)
        
        # create 3D array for linear predictor (intercepts + covariate effects)
        lin_pred <- array(data = NA_real_, dim = c(n_cells_batch,  # raster cells (batch)
                                             n_species,      # species
                                             n_samples))     # posterior samples
        

        for(i in 1:n_species){    # species loop
          if(cov_info_intercept$ranef == TRUE | cov_info_intercept$independent == TRUE){  # random or independent intercepts
            lin_pred[,i,] <- matrix(posterior_matrix[, colnames(posterior_matrix) %in% paste0(keyword_submodel_short, "0", "[", i, "]")] , 
                                         nrow = n_cells_batch, ncol = n_samples, byrow = T)
          } else {
            lin_pred[,i,] <- matrix(posterior_matrix[, grep(paste0(keyword_submodel_short, "0$"), colnames(posterior_matrix))] , 
                                         nrow = n_cells_batch, ncol = n_samples, byrow = T)
          }
        }
        
        
        
        # estimate required memory for prediction
        ram_usage_estimate <- (prod(c(n_cells_batch,
                                      n_species,       
                                      n_samples)) *
                                 8) /   # 8 bytes per value
          1e9 *                  # to Gb
          (nrow(cov_info_subset) + 1)  # number of covariates (plus intercept)
        
        
        # ensure sufficient memory is available
        .check_memory_usage(ram_usage_estimate)
        
        
        
        # container for individual covariate effects
        out <- list()
        
        # loop over covariates 
        for(cov in 1:nrow(cov_info_subset)) {
          
          current_cov  <- cov_info_subset$covariate[cov]
          current_coef <- cov_info_subset$coef[cov]
          
          if(!current_cov %in% colnames(values_to_predict_subset_i)) {
            stop(paste("Covariate", current_cov, "not found in data for prediction (x)."), call. = FALSE)
          }
          if(!is.na(cov_info_subset$ranef_cov[cov])){
            stop(paste(current_cov, 
                       " has a random effect other than species. This is currently not supported.", call. = F))
            next
          }
          
          if(cov_info_subset$ranef_nested[cov])  {
            stop(paste(current_cov, 
                       " has a nested random effect. This is currently not supported.", call. = F))
            next
          }
          
          
          # determine data type of current covariate
          covariate_is_numeric <- cov_info_subset$data_type [cov] == "cont"
          covariate_is_factor  <- cov_info_subset$data_type [cov] == "categ"
          
          
          effect_type <- ifelse(cov_info_subset$ranef[cov], "ranef",
                                ifelse(cov_info_subset$independent[cov], "independent", "fixed"))
          
          # covariate_is_site_cov <- ifelse(cov_info_subset$covariate_type [cov] == "siteCovs", T, F) 
          
          
          # species loop
          for(i in 1:n_species){
            
            if(covariate_is_numeric) {
              
              if(effect_type == "fixed") {
                index_covariate <- grep(paste0(current_coef, "$"), colnames(posterior_matrix))
              } else {    # ranef or independent
                index_covariate <- grep(paste0(current_coef, "[", i, "]"), colnames(posterior_matrix), fixed = T)
              }
              
              if(length(index_covariate) == 0) stop(paste("Covariate", current_coef, "not found in posterior matrix"), call. = FALSE)
              if(length(index_covariate) >= 2) stop(paste("Covariate", current_coef, "has more than 2 matches in posterior matrix"), call. = FALSE)
              

              lin_pred[,i,] <-  lin_pred[,i,] + sapply(posterior_matrix[, index_covariate], FUN = function(x){
                x * values_to_predict_subset_i[, current_cov]
              })
            }
            
            if(covariate_is_factor) {
              
              if(effect_type == "fixed") index_covariate <- grep(current_coef, colnames(posterior_matrix))
              if(effect_type == "ranef") index_covariate <- grep(paste0(current_coef, "[", i, ","), colnames(posterior_matrix), fixed = T)
              
              # this assumes that the numeric values in the raster correspond to the factor levels in the covariate
              # since it uses the raster values to index the posterior matrix
              
              # for data.frames, it seems to work with the categorical column being integer (= factor level, as in as.data.frame(rasterStack)), or proper factor
              lin_pred[,i,] <- lin_pred[,i,] + t(posterior_matrix[, index_covariate[values_to_predict_subset_i[, current_cov]]])
              
            }
            suppressWarnings(rm(index_covariate))
          }   # end species loop
        }    # end covariate loop

        
          # sum up individual effects
          if(object@model == "Occupancy") {
            psi_batch <- ilogit(lin_pred)
          }
        
          if(object@model == "RN"){
            lambda_batch <- exp(lin_pred)   # lambda is expected abundance   (Poisson intensity / rate parameter)
            
            if(!type %in% c("abundance", "lambda_array")){   # convert to occupancy probability
              psi_batch <- 1-dpois(0, lambda)
            }
          }
        rm(lin_pred)

        
        if(!type %in% c("abundance", "lambda_array")) {
          return(psi_batch) 
        } else {
          return(lambda_batch)
        }
        
      })   # end lapply 
      
      if(type %in% c("abundance", "lambda_array")) {
        lambda <- abind::abind(array_list, along = 1)
      } else {
        psi <- abind::abind(array_list, along = 1)
      } 
      
      rm(array_list)
      
    }   # end (if(isTRUE(batch)))
    
    
    ## occupancy / abundance, not by batch  ----
    
    if(isFALSE(batch)) {
      
      
      # estimate required memory for prediction
      ram_usage_estimate <- (prod(c(n_cells_all,
                                    n_species,       
                                    n_samples)) *
                               8) /   # 8 bytes per value
        1e9 *                  # to Gb
        (nrow(cov_info_subset) + 1)  # number of covariates (plus intercept)
      
      
      # ensure sufficient memory is available
      .check_memory_usage(ram_usage_estimate)
      
      
      # create 3D array for linear predictor (intercepts + covariate effects)
      lin_pred <- array(data = NA_real_, dim = c(n_cells_all,    # raster cells (all)
                                                 n_species,                     # species
                                                 n_samples))           # posterior samples
      
      
      for(i in 1:n_species){    # species loop
        if(cov_info_intercept$ranef == TRUE | cov_info_intercept$independent == TRUE){    # random or independent intercepts
          lin_pred[,i,] <- matrix(posterior_matrix[, colnames(posterior_matrix) %in% paste0(keyword_submodel_short, "0", "[", i, "]")] , 
                                  nrow = n_cells_all, ncol = n_samples, byrow = T)
        } else {
          lin_pred[,i,] <- matrix(posterior_matrix[, grepl(paste0("^", keyword_submodel_short, "0$"), colnames(posterior_matrix))] , 
                                  nrow = n_cells_all, ncol = n_samples, byrow = T)
        }
      }
      
      cleanup_threshold <- 500  #Mb
      if(object.size(lin_pred) / 1e6 > cleanup_threshold)     gc()   # if intercept matrix > 500Mb, cleanup
      

      if(nrow(cov_info_subset) >= 1) {   # if there are covariates
        
        
        # loop over covariates 
        for(cov in 1:nrow(cov_info_subset)) {
          
          current_cov  <- cov_info_subset$covariate[cov]
          current_coef <- cov_info_subset$coef[cov]
          
          if(!current_cov %in% colnames(values_to_predict_subset)) {
            stop(paste("Covariate", current_cov, "not found in data for prediction (x)."), call. = FALSE)
          }
          if(!is.na(cov_info_subset$ranef_cov[cov])){
            stop(paste(current_cov, 
                       " has a random effect other than species. This is currently not supported.", call. = F))
            next
          }
          
          if(cov_info_subset$ranef_nested[cov])  {
            stop(paste(current_cov, 
                       " has a nested random effect. This is currently not supported.", call. = F))
            next
          }
          
          
          # determine data type of current covariate
          covariate_is_numeric <- cov_info_subset$data_type [cov] == "cont"
          covariate_is_factor  <- cov_info_subset$data_type [cov] == "categ"
          
          
          effect_type <- ifelse(cov_info_subset$ranef[cov], "ranef",
                                ifelse(cov_info_subset$independent[cov], "independent", "fixed"))
          
          
          # species loop
          for(i in 1:dim(lin_pred)[2]){
            
            if(covariate_is_numeric) {
              
              if(effect_type == "fixed") {
                index_covariate <- grep(paste0(current_coef, "$"), colnames(posterior_matrix))
              } else {    # ranef or independent
                index_covariate <- grep(paste0(current_coef, "[", i, "]"), colnames(posterior_matrix), fixed = T)
              }
              
              if(length(index_covariate) == 0) stop(paste("Covariate", current_coef, "not found in posterior matrix"), call. = FALSE)
              if(length(index_covariate) >= 2) stop(paste("Covariate", current_coef, "has more than 2 matches in posterior matrix"), call. = FALSE)
              
              
              lin_pred[,i,] <-  lin_pred[,i,] + sapply(posterior_matrix[, index_covariate], FUN = function(x){
                x * values_to_predict_subset[, current_cov]
              })
            }
            
            if(covariate_is_factor) {
              
              if(effect_type == "fixed") index_covariate <- grep(current_coef, colnames(posterior_matrix))
              if(effect_type == "ranef") index_covariate <- grep(paste0(current_coef, "[", i, ","), colnames(posterior_matrix), fixed = T)
              
              # this assumes that the numeric values in the raster correspond to the factor levels in the covariate
              # since it uses the raster values to index the posterior matrix
              
              # for data.frames, it seems to work with the categorical column being integer (= factor level, as in as.data.frame(rasterStack)), or proper factor
              
              lin_pred[,i,] <- lin_pred[,i,] + t(posterior_matrix[, index_covariate[values_to_predict_subset[, current_cov]]])
              
            }
            suppressWarnings(rm(index_covariate))
          }   # end species loop
        }    # end covariate loop
      } # end if there are covariates
      
      # backtransform to scale of response variable
      
      if(object@model == "Occupancy") {
        psi <- ilogit(lin_pred)
        rm(lin_pred)
      }
      
      if(object@model == "RN"){
        lambda <- exp(lin_pred)   # lambda is expected abundance   (Poisson intensity / rate parameter)
        rm(lin_pred)
        if(!type %in% c("abundance", "lambda_array")){   # convert to occupancy probability
          psi <- 1-dpois(0, lambda)
        }
      }
    }   # end  if(!batch) 
    
    
  } else {   # = if type == "p_array", do:
    
    ## detection probability
    
    if (!requireNamespace("abind", quietly = TRUE)) {
      stop(paste("Please install the package abind to run this function with type =", type))
    }
    
    
    
    # estimate required memory for prediction
    ram_usage_estimate <- (prod(c(n_cells_all,
                                  n_species,       
                                  n_samples,
                                  n_occasions
                                  )) *
                             8) /   # 8 bytes per value
      1e9 *                  # to Gb
      (nrow(cov_info_subset) + 1)  # number of covariates (plus intercept)
    
    
    # ensure sufficient memory is available
    .check_memory_usage(ram_usage_estimate)
    
    
    
    # get intercept
    
    if(cov_info_intercept$ranef == FALSE  && cov_info_intercept$independent == FALSE  ) {
      index_intercept <- rep(grep('^alpha0$', colnames(posterior_matrix)),
                             times = n_species)    # repeat for each species
    } else {
      index_intercept <- grep('alpha0[', colnames(posterior_matrix), fixed=TRUE)
    }
    
    if(length(index_intercept) == 0) stop(paste("error trying to access intercepts for submodel", submodel))
    
    a0_matrix <- posterior_matrix[, index_intercept]
    # order: [draws, species]
    
    # prepare empty list to contain parameters for covariate responses
    a1_matrix_list <- list()
    
    
    if(nrow(cov_info_subset) >= 1) {
      # loop over covariates 
      for(cov in 1:nrow(cov_info_subset)) {
        
        # collect information about current covariate
        current_cov  <- cov_info_subset$covariate[cov]
        current_coef <- cov_info_subset$coef[cov]
        
        covariate_is_numeric <- cov_info_subset$data_type [cov] == "cont"
        covariate_is_factor  <- cov_info_subset$data_type [cov] == "categ"
        
        effect_type <- ifelse(cov_info_subset$ranef[cov], "ranef",
                              ifelse(cov_info_subset$independent[cov], "independent", "fixed"))
        
        covariate_is_site_cov <- ifelse(cov_info_subset$covariate_type [cov] == "siteCovs", T, F)
        covariate_is_site_occasion_cov <- ifelse(cov_info_subset$covariate_type [cov] == "obsCovs", T, F) 
        
        
        # a few checks (should be done before covariate loop to avoid wasting time if things go wrong)
        if(!current_cov %in% colnames(values_to_predict_subset) && covariate_is_site_cov) {
          stop(paste("Covariate", current_cov, "not found in data for prediction (x)."), call. = FALSE)
        }
        
        if(!is.na(cov_info_subset$ranef_cov[cov])){
          stop(paste(current_cov, 
                     " has a random effect other than species. This is currently not supported.", call. = F))
          next
        }
        
        if(cov_info_subset$ranef_nested[cov])  {
          stop(paste(current_cov, 
                     " has a nested random effect. This is currently not supported.", call. = F))
          next
        }
        
        
        # identify covariate column in MCMC list output
        
        # first construct matching regular expression for:
        # fixed effect
        if(effect_type == "fixed") alpha_regex <- paste0("alpha.*", current_cov, "$")
        # random effect
        if(effect_type %in% c("ranef", "indep")) alpha_regex <- paste0("alpha.*", current_cov, "\\[")
        
        # find matching columns
        alpha_regex_matches <- grep(alpha_regex,  colnames(posterior_matrix))
        
        # error if no match
        if(length(alpha_regex_matches) == 0) stop(paste("Error identifying value column for covariate:", current_cov))
        
        # subset posterior matrix
        a1_matrix <- posterior_matrix[, alpha_regex_matches, drop = F]
        
        # empty list for covariate-specific effects
        p_arr_list <- list()
        
        # loop over species and occasions to fill arrays
        for (i in 1:n_species){   # species loop
          # prepare empty array (for current species)
          p_arr_list [[i]] <- array(NA_real_, c(n_samples, 
                                          n_stations, 
                                          n_occasions))
          # order: [draws, stations, occasions]
          
          for (k in 1:n_occasions){ # occasion loop
            
            # if covariate is site-occasion covariate, multiply parameter values from
            # posterior draws with covariate values (by occasion)
            if(covariate_is_site_occasion_cov){      
              if(effect_type == "fixed") {
                # a1_matrix contains only 1 column, identical for all species
                p_arr_list [[i]] [,, k]  <- outer(a1_matrix[, 1],  object@input$obsCovs[[current_cov]] [,k])
              } else { 
                # if ranef, one column per species
                p_arr_list [[i]] [,, k]  <- outer(a1_matrix[, i],  object@input$obsCovs[[current_cov]] [,k])
              }
            }
            
            # if covariate is site covariate, multiply estimates with site covariate values
            if(covariate_is_site_cov) {
              if(effect_type == "fixed") {
                # a1_matrix contains only 1 column, identical for all species
                p_arr_list [[i]] [,, k] <- outer(a1_matrix[, 1], object@input$siteCovs[, current_cov])
              } else {
                # if ranef, one column per species
                p_arr_list [[i]] [,, k] <- outer(a1_matrix[, i], object@input$siteCovs[, current_cov])
              }
            }
          }
        }
        
        # combine list of 3d arrays into 4d array
        # oder: [draws, station, occasion, species]
        p_arr_4d <- abind::abind(p_arr_list, along = 4)    
        
        # harmonize order with lin_pred
        # rearrange into order: [station, species, draw, occasion]
        p_arr_4d <- aperm(p_arr_4d,  c(2,4,1,3))   
        
        # store in list of covariate-specific arrays
        a1_matrix_list[[cov]] <- p_arr_4d
      }
      
      
      # sum up individual covariate effects
      logit.p <- Reduce('+', a1_matrix_list)  # doesn't include intercept yet
      
      
      # add a0_matrix (species-specific intercepts) to each element
      # a0_matrix is [draw, species]
      for(i in 1:dim(logit.p)[1]) {   # station
        for(k in 1:dim(logit.p)[4]) {   # occasion
          logit.p[i,,,k] <- logit.p[i,,,k] + t(a0_matrix)
        }
      }
      
    } else {
      logit.p <- array(NA_real_, 
                       dim = c(n_stations,        # stations
                               n_species,        # species
                               n_samples,
                               n_occasions   # occasions
                       ))
      for(i in 1:n_stations) {          # station
        for(k in 1:n_occasions) {   # occasion
          logit.p[i,,,k] <- t(a0_matrix)
        }
      }
    }
    
    # convert to probability
    p <- ilogit(logit.p)
    
    
    # cleanup
    rm(logit.p, a0_matrix, a1_matrix_list)
    if(exists("a1_matrix")) rm(a1_matrix)
  }
  gc()
  

  # return raw detection probabilities (4D array)
  if(type == "p_array") {
  dimnames(p) <- list(index_not_na,
                        rownames(object@data$y))
  return(p)
}
  
  
  # return raw occupancy probabilities [cell, species, posterior_draw]
  if(type == "psi_array") {
    dimnames(psi) <- list(index_not_na,
                          rownames(object@data$y))
    return(psi)
  }
  
  # return raw expected abundance [cell, species, posterior_draw]
  if(type == "lambda_array") {
    dimnames(lambda) <- list(index_not_na,
                             rownames(object@data$y))
    return(lambda)
  }
  
  # percentage of area occupied (by species)
  if(type == "pao") {
    
    # random binomial trial for each probability
    z <- array(rbinom(length(psi),prob=psi,size=1), dim = dim(psi))
    pao1 <- apply(z, MARGIN = c(2, 3), mean)  # aggregated spatially: [species, draw]
    dimnames(pao1)[1] <- dimnames(object@data$y)[1] #list(names(object@input$ylist))

    
    # summary table
    pao_summary_table <- as.data.frame(t(apply(pao1, MARGIN = 1, summary)))
    
    
    pao.lower <- as.data.frame(t(apply(pao1, MARGIN = 1, quantile, ((1-level) / 2))))
    pao.upper <- as.data.frame(t(apply(pao1, MARGIN = 1, quantile,  (1 - (1-level) / 2))))
    
    pao.lower2 <- reshape2::melt(pao.lower)
    pao.upper2 <- reshape2::melt(pao.upper)
    
    names(pao.lower2) <- c("Species", paste0("lower.ci.", ((1-level) / 2)))
    names(pao.upper2) <- c("Species", paste0("upper.ci.", (1 - (1-level) / 2)))
    

    pao_summary_table <- cbind(pao_summary_table [, "Min.", drop = F],
                               pao.lower2[,2, drop = F],
                               pao_summary_table [, c("1st Qu.", "Median", "Mean", "3rd Qu.")],
                               pao.upper2[,2, drop = F],
                               pao_summary_table [, "Max.", drop = F])
    
    pao_summary_table <- round(pao_summary_table, 3)
    
    # table with all posterior draws for all species (e.g. for ggplot2)
    pao_melt <- reshape2::melt(pao1)
    colnames(pao_melt) <- c("Species", "draw", "PAO")
    pao_melt <- pao_melt[order(pao_melt$Species, pao_melt$draw),]
    pao_melt$Species <- factor(pao_melt$Species, labels = #names(object@input$ylist))
                                 dimnames(object@data$y)[[1]])
    rownames(pao_melt) <- NULL

    
    if(hasArg(aoi)) {
      return(list(pao_summary = pao_summary_table, 
                  pao_matrix  = pao1,
                  pao_df      = pao_melt,
                  aoi         = aoi2))
    } else {
      return(list(pao_summary = pao_summary_table, 
                  pao_matrix  = pao1,
                  pao_df      = pao_melt))
    }

  }
  
  # return occupancy probabilities
  if(type == "psi") {
    
    if(hasArg(speciesSubset)) warning("speciesSubset is defined, but has no effect when type = 'psi'")
    
    # summarize estimates (across posterior samples)
    psi.mean <- apply(psi, MARGIN = c(1,2), mean)
    psi.sd   <- apply(psi, MARGIN = c(1,2), sd)
    
    # make data frame for ggplot
    psi.mean.melt <- reshape2::melt(psi.mean)
    psi.sd.melt   <- reshape2::melt(psi.sd)
    
    names(psi.mean.melt) <- c("cell_nr", "Species", "mean")
    names(psi.sd.melt)   <- c("cell_nr", "Species", "sd")
    
    if(!is.null(dimnames(object@data$y)[[1]])) {
      psi.mean.melt$Species <- dimnames(object@data$y)[[1]][psi.mean.melt$Species]
      psi.sd.melt$Species   <- dimnames(object@data$y)[[1]][psi.sd.melt$Species]
    }
    
    # fill rasters with predicted values
    if(inherits(x, "SpatRaster")) {
    
    r_pred_species <- r_pred_sd_species <- list()
    
    for(i in 1:length(unique(psi.mean.melt$Species))) {
      r_pred_species[[i]]    <- raster_template
      r_pred_sd_species[[i]] <- raster_template
      
      terra::values(r_pred_species[[i]])    [index_not_na] <- psi.mean.melt$mean[psi.mean.melt$Species == unique(psi.mean.melt$Species)[i]]
      terra::values(r_pred_sd_species[[i]]) [index_not_na] <- psi.sd.melt$sd[psi.sd.melt$Species == unique(psi.sd.melt$Species)[i]]
    }
    names(r_pred_species) <- names(r_pred_sd_species) <- unique(psi.mean.melt$Species)
    
    stack_out_mean <- terra::rast(r_pred_species)
    stack_out_sd   <- terra::rast(r_pred_sd_species)
    }

    
    if(interval == "confidence"){
      psi.lower <- apply(psi, MARGIN = c(1,2), quantile, ((1-level) / 2))                        # SLOW!!!
      psi.upper <- apply(psi, MARGIN = c(1,2), quantile, (1 - (1-level) / 2))
      
      psi.lower2 <- reshape2::melt(psi.lower)
      psi.upper2 <- reshape2::melt(psi.upper)
      
      names(psi.lower2) <- c("cell_nr", "Species", paste0("lower.ci.", level))
      names(psi.upper2) <- c("cell_nr", "Species", paste0("upper.ci.", level))
      
      if(!is.null(dimnames(object@data$y)[[1]])) {
        psi.lower2$Species <- dimnames(object@data$y)[[1]][psi.lower2$Species]
        psi.upper2$Species <- dimnames(object@data$y)[[1]][psi.upper2$Species]
      }
      
      if(inherits(x, "SpatRaster")) {
      r_pred_lower_species <- r_pred_upper_species <- list()
      
      for(i in 1:length(unique(psi.mean.melt$Species))) {
        r_pred_lower_species[[i]] <- raster_template
        r_pred_upper_species[[i]] <- raster_template
        
        terra::values(r_pred_lower_species[[i]]) [index_not_na] <- psi.lower2[psi.lower2$Species == unique(psi.lower2$Species)[i], paste0("lower.ci.", level)]
        terra::values(r_pred_upper_species[[i]]) [index_not_na] <- psi.upper2[psi.upper2$Species == unique(psi.upper2$Species)[i], paste0("upper.ci.", level)]
      }
      names(r_pred_lower_species) <- names(r_pred_upper_species) <- unique(psi.mean.melt$Species)
      
      stack_out_lower   <- terra::rast(r_pred_lower_species)
      stack_out_upper   <- terra::rast(r_pred_upper_species)
      
      return(list(mean  = stack_out_mean,
                  sd    = stack_out_sd,
                  lower = stack_out_lower,
                  upper = stack_out_upper))
      } else {
        
        return(data.frame(psi.mean.melt, 
                          sd = psi.sd.melt$sd, 
                          lower = psi.lower2[, 3], 
                          upper = psi.upper2 [,3]))
        
      }
    }
    
    if(interval == "none"){
      if(inherits(x, "SpatRaster")){
        return(list(mean = stack_out_mean,
                    sd   = stack_out_sd))
      } else {
        return(data.frame(psi.mean.melt, 
                          sd = psi.sd.melt$sd))
      }
    }
  }
  
  # return species richness
  if(type == "richness") {
    
    # generate occupancy status at each cell / species / posterior sample as random binomial trial (returns vector)
    psi_bin <- rbinom(length(psi), size = 1, prob = psi)
    
    # convert back to array
    psi_bin_array <- array(psi_bin, dim = dim(psi), dimnames = dimnames(psi))
    
    
    # subset richness estimate to certain species, if requested
    if(!hasArg(speciesSubset)) speciesSubset <- 1:dim(psi_bin_array)[2]
    
    if(is.character(speciesSubset)) {
      if(!is.null(dimnames(object@data$y)[[1]])) {
        if(any(!speciesSubset %in% dimnames(object@data$y)[[1]])) {
          stop(paste(paste(speciesSubset[!speciesSubset %in% dimnames(object@data$y)[[1]]], collapse = ", "),
                     "not found in the species names of object@data$y"))
        }
        speciesIndex <- which(dimnames(object@data$y)[[1]] %in% speciesSubset)
      } else {
        stop("speciesSubset is character, but object@data$y does not contain species names")
      }
    }
    
    if(is.numeric(speciesSubset)) speciesIndex <- speciesSubset
    
    # species richness at each pixel for each posterior sample
    psi.bin.sum <- apply(psi_bin_array[, speciesIndex,, drop = FALSE], MARGIN = c(1,3), sum)
    
    
    # Mean of richness estimate
    psi.bin.sum.mean <- apply(psi.bin.sum, 1, mean)
    # SD of richness estimate
    psi.bin.sum.sd <- apply(psi.bin.sum, 1, sd)
    
    # confidence intervals of richness estimate
    if(interval == "confidence"){
      psi.bin.sum.lower <- apply(psi.bin.sum, 1, quantile, ((1-level) / 2))
      psi.bin.sum.upper <- apply(psi.bin.sum, 1, quantile, (1 - (1-level) / 2))
    }
    
    if(inherits(x, "SpatRaster")) {
      
      r.psi.bin.sum.mean <- r.psi.bin.sum.sd <- r.psi.bin.sum.lower <- r.psi.bin.sum.upper <- raster_template
      terra::values(r.psi.bin.sum.mean) [index_not_na] <- psi.bin.sum.mean
      terra::values(r.psi.bin.sum.sd) [index_not_na] <- psi.bin.sum.sd
      
      if(interval == "none"){
        stack_out <- terra::rast(list(mean = r.psi.bin.sum.mean,
                                        sd   = r.psi.bin.sum.sd))
      }
      
      if(interval == "confidence"){
        terra::values(r.psi.bin.sum.lower) [index_not_na] <- psi.bin.sum.lower
        terra::values(r.psi.bin.sum.upper) [index_not_na] <- psi.bin.sum.upper
        
        stack_out <- terra::rast(list(mean  = r.psi.bin.sum.mean,
                                        sd    = r.psi.bin.sum.sd,
                                        lower = r.psi.bin.sum.lower,
                                        upper = r.psi.bin.sum.upper))
      }
      
      
      return(stack_out)
      
    } else {
      if(interval == "none"){
        df_out <- data.frame(mean = psi.bin.sum.mean,
                             sd   = psi.bin.sum.sd)
      }
      
      if(interval == "confidence"){
        df_out <- data.frame(mean  = psi.bin.sum.mean,
                             sd    = psi.bin.sum.sd, 
                             lower = psi.bin.sum.lower,
                             upper = psi.bin.sum.upper)
      }
      
      return(df_out)
      
    }
  }
  
  # return abundance 
  if(type == "abundance") {
    
    if(hasArg(speciesSubset)) warning("speciesSubset is defined, but has no effect when type = 'abundance'")
    
    # summarize estimates (across posterior samples)
    lambda.mean <- apply(lambda, MARGIN = c(1,2), mean)
    lambda.sd   <- apply(lambda, MARGIN = c(1,2), sd)
    
    # make data frame for ggplot
    lambda.mean.melt <- reshape2::melt(lambda.mean)
    lambda.sd.melt   <- reshape2::melt(lambda.sd)
    
    names(lambda.mean.melt) <- c("cell_nr", "Species", "mean")
    names(lambda.sd.melt)   <- c("cell_nr", "Species", "sd")
    
    if(!is.null(dimnames(object@data$y)[[1]])) {
      lambda.mean.melt$Species <- dimnames(object@data$y)[[1]][lambda.mean.melt$Species]
      lambda.sd.melt$Species   <- dimnames(object@data$y)[[1]][lambda.sd.melt$Species]
    }
    
    # fill rasters with predicted values
    if(inherits(x, "SpatRaster")) {
      r_pred_species <- r_pred_sd_species <- list()
      
      for(i in 1:length(unique(lambda.mean.melt$Species))) {
        r_pred_species[[i]]    <- raster_template
        r_pred_sd_species[[i]] <- raster_template
        
        terra::values(r_pred_species[[i]])    [index_not_na] <- lambda.mean.melt$mean[lambda.mean.melt$Species == unique(lambda.mean.melt$Species)[i]]
        terra::values(r_pred_sd_species[[i]]) [index_not_na] <- lambda.sd.melt$sd[lambda.sd.melt$Species == unique(lambda.sd.melt$Species)[i]]
      }
      names(r_pred_species) <- names(r_pred_sd_species) <- unique(lambda.mean.melt$Species)
      
      stack_out_mean <- terra::rast(r_pred_species)
      stack_out_sd   <- terra::rast(r_pred_sd_species)
    }
    
    
    if(interval == "confidence"){
      lambda.lower <- apply(lambda, MARGIN = c(1,2), quantile, ((1-level) / 2))                        # SLOW!!!
      lambda.upper <- apply(lambda, MARGIN = c(1,2), quantile, (1 - (1-level) / 2))
      
      lambda.lower2 <- reshape2::melt(lambda.lower)
      lambda.upper2 <- reshape2::melt(lambda.upper)
      
      names(lambda.lower2) <- c("cell_nr", "Species", paste0("lower.ci.", level))
      names(lambda.upper2) <- c("cell_nr", "Species", paste0("upper.ci.", level))
      
      if(!is.null(dimnames(object@data$y)[[1]])) {
        lambda.lower2$Species <- dimnames(object@data$y)[[1]][lambda.lower2$Species]
        lambda.upper2$Species <- dimnames(object@data$y)[[1]][lambda.upper2$Species]
      }
      
      if(inherits(x, "SpatRaster")) {
        r_pred_lower_species <- r_pred_upper_species <- list()
        
        for(i in 1:length(unique(lambda.mean.melt$Species))) {
          r_pred_lower_species[[i]] <- raster_template
          r_pred_upper_species[[i]] <- raster_template
          
          terra::values(r_pred_lower_species[[i]]) [index_not_na] <- lambda.lower2[lambda.lower2$Species == unique(lambda.lower2$Species)[i], paste0("lower.ci.", level)]
          terra::values(r_pred_upper_species[[i]]) [index_not_na] <- lambda.upper2[lambda.upper2$Species == unique(lambda.upper2$Species)[i], paste0("upper.ci.", level)]
        }
        names(r_pred_lower_species) <- names(r_pred_upper_species) <- unique(lambda.mean.melt$Species)
        
        stack_out_lower   <- terra::rast(r_pred_lower_species)
        stack_out_upper   <- terra::rast(r_pred_upper_species)
        
        return(list(mean  = stack_out_mean,
                    sd    = stack_out_sd,
                    lower = stack_out_lower,
                    upper = stack_out_upper))
      } else {
        
        return(data.frame(lambda.mean.melt, 
                          sd = lambda.sd.melt$sd, 
                          lower = lambda.lower2[, 3], 
                          upper = lambda.upper2 [,3]))
        
      }
    }
    
    if(interval == "none"){
      if(inherits(x, "SpatRaster")){
        return(list(mean = stack_out_mean,
                    sd   = stack_out_sd))
      } else {
        return(data.frame(lambda.mean.melt, 
                          sd = lambda.sd.melt$sd))
      }
    }
  } # close if(type == "abundance")
  
}




#' Predictions from community occupancy models
#' 
#' Create (spatial) predictions of species occupancy and species richness from community occupancy models and spatial rasters or covariate data frames.
#'
#' @param object \code{commOccu} object
#' @param mcmc.list  mcmc.list. Output of \code{fit} called on a \code{commOccu} object
#' @param type character. "psi" for species occupancy estimates, "richness" for species richness estimates, "pao" for percentage of area occupied (by species), "psi_array" for raw occupancy probabilities in an array. For Royle-Nichols models, "abundance" for species abundance, or "lambda_array" for raw species abundance estimates in an array. "p_array" for raw detection probabilities in an array.
#' @param draws  Number of draws from the posterior to use when generating the plots. If fewer than draws are available, they are all used
#' @param level  Probability mass to include in the uncertainty interval
#' @param interval  Type of interval calculation. Can be "none" or "confidence" (can be abbreviated). Calculation can be slow for type = "psi" with many cells and posterior samples.
#' @param x   SpatRaster, data.frame or NULL. Must be scaled with same parameters as site covariates used in model, and have same names. If NULL, use site covariate data frame from model input (\code{commOccu} object in parameter \code{object}) 
#' @param aoi SpatRaster with same dimensions as x (if x is a SpatRaster), indicating the area of interest (all cells with values are AOI, all NA cells are ignored). If NULL, predictions are made for all cells.
#' @param speciesSubset  species to include in richness estimates. Can be index number or species names.
#' @param batch logical or numeric. If FALSE, all raster cells / data frame rows will be processed at once (can be memory intensive). If TRUE, computation is conducted in batches of 1000. If numeric, it is the desired batch size.
#' @param seed numeric. Seed to use in \code{set.seed} for reproducible results (ensures that \code{draws} are identical).
#'
#' @details Processing can be very memory-intensive. If memory is insufficient, use the  \code{batch} parameter. This can enable processing for higher numbers of \code{draws} or very large rasters / data frames. 
#' 
#' @return A SpatRaster or data.frame, depending on \code{x} (for type = "psi", "abundance", "richness". If type = "pao", a list. If type = "psi_array" or "lambda_array", a 3D-array [site, species, draw]. If type = "p_array", a 4D-array [site, species, draw, occasion].
#' 
#' @aliases predict
#' @method predict commOccu
#' @importFrom stats rbinom dpois sd
#' @importFrom utils object.size
#' @importFrom methods .hasSlot
#' @export


setMethod("predict", signature = c(object = "commOccu"),
          predictionMapsCommunity)

#' Write a copyright tag into JPEG image metadata
#' 
#' This function writes a copyright tag into the copyright field of JPEG image
#' Exif metadata. It does so recursively, so it works both for images that are
#' sorted into subdirectories and unsorted images. Note that all images in
#' subdirectories of inDir will be tagged. It is not required to run this
#' function in the camtrapR workflow, but may be desired for data sharing or
#' publishing.
#' 
#' If askFirst = TRUE, the function will show a menu and asks the user to
#' confirm the action before execution. Type "1" to write copyright tags and
#' "2" to abort.
#' 
#' By default Exiftool creates a copy of each JPG image and preserves the
#' original images (without the copyright tag) as .JPG_original files. Note
#' that this behaviour will instantly double the number of images in inDir and
#' the disk space required. If this is not desired, set keepJPG_original =
#' FALSE.
#' 
#' \code{ignoreMinorErrors} is useful if copyright tags can't be updated
#' correctly. This can be caused by bad MakerNotes and so far was only observed
#' in Panthera V4 cameras. In that case, set \code{ignoreMinorErrors} to
#' \code{TRUE}. This will add the "-m" option to the Exiftool call, thereby
#' ignoring minor errors and warnings and assigning the copyright tag
#' regardless.
#' 
#' @param inDir character. Name of the directory containing camera trap images.
#' @param copyrightTag character. The tag to be written into the Exif Copyright
#' field
#' @param askFirst logical. Ask user to confirm before execution?
#' @param keepJPG_original logical. Keep original JPG files as .JPG_original
#' files (TRUE) or overwrite JPGs (FALSE)?
#' @param ignoreMinorErrors logical. Ignore minor errors that would cause the
#' function to fail (set TRUE for images with bad MakerNotes, observed in
#' Panthera V4 cameras)
#' 
#' @return An invisible list of Exiftool output.
#' 
#' More importantly, the specified copyright tag is written into the Copyright
#' field of the Exif metadata of all images in inDir.
#' 
#' @author Juergen Niedballa
#' 
#' @examples
#' 
#' 
#' \dontrun{
#' 
#' if (Sys.which("exiftool") != ""){        # only run this example if ExifTool is available
#' 
#' # copy sample images to temporary directory (so we don't mess around in the package directory)
#' wd_images_ID <- system.file(file.path("pictures", "sample_images_species_dir"), 
#'                             package = "camtrapR")
#' file.copy(from = wd_images_ID, to = tempdir(), recursive = TRUE)
#' wd_images_ID_copy <- file.path(tempdir(), "sample_images_species_dir")
#' 
#' # define a sample tag
#' copyrightTagToAdd <- "Your Name (Your Organisation)"
#' 
#' # add the tag to the images
#' addCopyrightTag(inDir        = wd_images_ID_copy, 
#'                 copyrightTag = copyrightTagToAdd)
#' 1     # we choose "YES", i.e., we want to add a copyright tag
#'                 
#' # you can check the outcome with function exifTagNames
#' 
#' metadat <- exifTagNames(wd_images_ID_copy)
#' metadat [metadat$tag_name == "Copyright",]
#' }
#' 
#' 
#' }
#' 
#' @export addCopyrightTag
#' 
addCopyrightTag <- function(inDir,
                            copyrightTag,
                            askFirst = TRUE,
                            keepJPG_original = TRUE,
                            ignoreMinorErrors = FALSE){

  stopifnot(is.character(inDir))
  stopifnot(file.exists(inDir))

  stopifnot(is.character(copyrightTag))
  stopifnot(length(copyrightTag) == 1)

  stopifnot(is.logical(keepJPG_original))

  if(isFALSE(keepJPG_original)) {
    overwrite_command <- " -overwrite_original "
  } else {overwrite_command <- " "}

  if(Sys.which("exiftool") == "") stop("cannot find ExifTool", call. = FALSE)

  if(isTRUE(askFirst)){
    message(paste("You are about to add the copyright tag '", copyrightTag, "' to all images in ", inDir, ". Are you sure?", sep = ""))
    switch(menu(c("YES", "NO")),
           doit <- TRUE,
           doit <- FALSE)
  } else {
    doit <- TRUE
  }

  if(isTRUE(doit)){

    command.tmp  <- paste0('exiftool -r ', ifelse(ignoreMinorErrors, "-m", ""), overwrite_command  ,'-Copyright="', copyrightTag, '" "', inDir, '"')

    output <- system(command.tmp, show.output.on.console = FALSE, intern = TRUE)
    message(paste(output[length(output) - 1],
                  output[length(output)]))
    if(grepl("files weren't updated due to errors",  output[length(output)])) {
      warning("There were problems adding the copyright tag. Try setting ignoreMinorErrors = TRUE", call. = FALSE)
    }
    return(invisible(output))
  } else {
    message("Did not write copyright tag")
    return(invisible(NULL))
  }
}

#' Create a community (multi-species) occupancy model for JAGS or Nimble
#' 
#' @description 
#' Flexibly creates complete code and input data for community occupancy models for JAGS amd Nimble (both standard occupancy models and Royle-Nichols occupancy models), and automatically sets initial values and parameters to monitor. 
#' Supports fixed and random effects of covariates on detection and occupancy probabilities, using both continuous and categorical covariates (both site and site-occasion covariates). 
#' 
#' Optionally includes data augmentation (fully open community, or up to known maximum number of species, or no data augmentation). 
#' Allows combination of all these parameters for fast and flexible customization of community occupancy models.
#' 
#' Incidentally, the function can also be used to create model code and input for single-species single-season occupancy models (it is the special case of the community model with only one species). 
#' Such a model will run slower than proper single-species model JAGS code due to the additional species loop, but it is possible.
#' 
#' The function returns several derived quantities, e.g. species richness, Bayesian p-values (overall and by species), Freeman-Tukey residuals for actual and simulated data (by station and total). If doing data augmentation, metacommunity size and number of unseen species are returned also. 
#'
#' @importFrom generics fit
#' @importFrom stats rnorm window 
#' @importFrom ggplot2 ggplot ggsave xlim geom_line facet_wrap geom_ribbon theme_bw ggtitle xlab ylab aes
#' @export
#'
#' @param data_list    list. Contains 3 slots: ylist, siteCovs, obsCovs. ylist is a list of detection histories (can be named), e.g. from \code{\link{detectionHistory}}. siteCovs is a data.frame with site covariates (optional). obsCovs is a list of site-occasion level covariates (e.g. site-occasion-specific effort, which is also returned by \code{\link{detectionHistory}}.
#' @param model character. "Occupancy" for standard occupancy model,  or "RN" for the occupancy model of Royle and Nichols (2003), which relates probability of detection of the species to the number of individuals available for detection at each station
#' @param occuCovs  list. Up to 3 items named "fixed", "independent", and/or "ranef". Specifies fixed, independent or random effects of covariates on occupancy probability (continuous or categorical covariates). Independent effects are only supported for continuous covariates.
#' @param detCovs   list. Up to 3 items named "fixed", "independent", and/or "ranef". Specifies fixed, independent or random effects of covariates on detection probability (continuous or categorical covariates). Independent effects are only supported for continuous covariates.
#' @param detCovsObservation   list. Up to 2 items named "fixed" and/or "ranef". Specifies fixed or random effects of observation-level covariates on detection probability  (continuous or categorical covariates - categorical must be coded as character matrix)
#' @param speciesSiteRandomEffect list. Two items named "det" and "occu". If TRUE, adds a random effect of species and station. Only implemented for detection probability. 
#' @param intercepts     list. Two items named "det" and "occu" for detection and occupancy probability intercepts. Values can be "fixed" (= constant across species), "independent" (= independent estimates for each species), or "ranef" (= random effect of species on intercept).
#' @param effortCov     character. Name of list item in \code{data_list$obsCovs} which contains effort. This does not include effort as a covariate on detection probability, but only uses NA / not NA information to create binary effort and ensure detection probabilities p are 0 when there was no effort (p will be 0 whereever \code{effortCov} is NA).
#' @param richnessCategories  character. Name of categorical covariate in \code{data_list$siteCovs} for which to calculate separate richness estimates (optional). Can be useful to obtain separate richness estimates for different areas.
#' @param augmentation     If NULL, no data augmentation (only use species in \code{data_list$ylist}), otherwise named list or vector with total number of (potential) species. Names: "maxknown" or "full". Example: \code{augmentation = c(maxknown = 30)} or \code{augmentation = c(full = 30)}
#' @param modelFile   character. Text file name to save model to
#' @param nimble  logical. If TRUE, model code will be for Nimble (incompatible with JAGS). If FALSE, model code is for JAGS.
#' @param keyword_quadratic  character. A suffix in covariate names in the model that indicates a covariate is a quadratic effect of another covariate which does not carry the suffix in its name (e.g. if the covariate is "elevation", the quadratic covariate would be "elevation_squared").
#'
#'
#' @details
#' For examples of implementation, see Vignette 5: Multi-species occupancy models.
#' 
#' Fixed effects of covariates are constant across species, whereas random effect covariates differ between species. Independent effect differ between species and are independent (there is no underlying hyperdistribution).
#' Fixed, independent and random effects are allowed for station-level detection and occupancy covariates (a.k.a. site covariates). Fixed and random effects are also allowed for station-occasion level covariates (a.k.a. observation covariates). 
#' Currently independent effects are only supported for continuous site covariates, not categorical site covariates or observation-level covariates.
#' 
#' By default, random effects will be by species. It is however possible to use categorical site covariates for grouping (continuous|categorical).
#' Furthermore, is is possible to use use nested random effects of species and another categorical site covariate (so that there is a random effect of species and an additional random effect of a categorical covariate within each species).
#' 
#' 
#' Derived quantities returned by the model are:
#'   
#'   \tabular{ll}{
#'     \code{Bpvalue} \tab Bayesian p-value (overall) \cr
#'     \code{Bpvalue_species} \tab Bayesian p-value (by species) \cr
#'     
#'     
#'     \code{Nspecies} \tab Species richness (only in JAGS models)\cr
#'     \code{Nspecies_station} \tab Species richness at each sampling locations (only in JAGS models) \cr
#'     \code{Nspecies_Covariate} \tab Species richness by categorical covariate (when using \code{richnessCategories}, only in JAGS models) \cr
#'     
#'     
#'     \code{R2} \tab sum of Freeman-Tukey residuals of observed data within each species \cr
#'     \code{new.R2} \tab sum of Freeman-Tukey residuals of simulated data within each species \cr
#'     \code{R3} \tab Total sum of Freeman-Tukey residuals of observed data \cr 
#'     \code{new.R3} \tab Total sum of Freeman-Tukey residuals of simulated data \cr
#'     \emph{\code{Ntotal}} \tab Total metacommunity size (= observed species + n0) \cr
#'     \emph{\code{n0}} \tab Number of unseen species in metacommunity \cr
#'     \emph{\code{omega}} \tab Data augmentation parameter \cr
#'     \emph{\code{w}} \tab Metacommunity membership indicator for each species
#'   }
#' 
#' Quantities in \emph{italic} at the bottom are only returned in full data augmentation. \code{Nspecies} and \code{Nspecies_Covariate} are only returned in JAGS models (because Nimble models don't explicitly return latent occupancy status z).
#' 
#' 
#' 
#' @return
#' 
#' \code{commOccu} object. It is an S4 class containing all information required to run the models. See  \code{\link{commOccu-class}} for details.
#'
#' @encoding UTF-8 
#' @references 
#' Kéry, M., and J. A. Royle. "Applied hierarchical modelling in ecology - Modeling distribution, abundance and species richness using R and BUGS." Volume 1: Prelude and Static Models. Elsevier/Academic Press, 2016.
#' 
#' @section Parameter naming convention: 
#' The parameter names are assembled from building blocks. The nomenclature  is as follows:    
#'   \tabular{lll}{
#'     \bold{Name}           \tab \bold{Refers to}  \tab  \bold{Description}  \cr
#'     \bold{\code{alpha}}   \tab Submodel          \tab detection submodel  \cr
#'     \bold{\code{beta}}    \tab Submodel          \tab occupancy submode \cr
#'     \bold{\code{0}}       \tab Intercept         \tab denotes the intercepts (alpha0, beta0)  \cr
#'     \bold{\code{fixed}}   \tab Effect type       \tab fixed effects (constant across species)  \cr
#'     \bold{\code{indep}}   \tab Effect type       \tab independent effects (separate for each species)  \cr
#'     \bold{\code{ranef}}   \tab Effect type       \tab random effects (of species and/or other categorical covariates)  \cr
#'     \bold{\code{cont}}    \tab Covariate type    \tab continuous covariates \cr
#'     \bold{\code{categ}}   \tab Covariate type    \tab categorical covariates \cr
#'     \bold{\code{mean}}    \tab Hyperparameter    \tab mean of random effect \cr
#'     \bold{\code{sigma}}   \tab Hyperparameter    \tab standard deviation of random effect \cr
#'     \bold{\code{tau}}     \tab Hyperparameter    \tab precision of random effect (used internally, not returned)
#'   }
#' 
#' 
#' For example, a fixed intercept of occupancy (constant across species) is \bold{\code{beta0}}, and a fixed intercept of detection probability is \bold{\code{alpha0}}.
#' 
#' 
#' An occupancy probability intercept with a random effect of species is: 
#'   
#' \bold{\code{beta0.mean}} community mean of the occupancy probability intercept
#' 
#' \bold{\code{beta0.sigma}} standard deviation of the community occupancy probability intercept.
#' 
#' \bold{\code{beta0[1]}} occupancy probability intercept of species 1 (likewise for other species). 
#' 
#' 
#' For effects of site covariates, the pattern is:
#'   
#' \code{submodel.effectType.covariateType.CovariateName.hyperparameter}
#' 
#' For example:
#'   
#' \bold{\code{beta.ranef.cont.habitat.mean}} is the mean community effect of the continuous site covariate 'habitat' on occupancy probability.
#' 
#' \bold{\code{beta.ranef.cont.habitat[1]}} is the effect of continuous site covariate 'habitat' on occupancy probability of species 1.
#' 
#' Site-occasion covariates are denoted by ".obs" after the submodel, e.g.: 
#'   
#' \bold{\code{alpha.obs.fixed.cont.effort}} is the fixed effect of the continuous observation-level covariate 'effort' on detection probability
#' 
#' 
#' @author Juergen Niedballa
#'  
#' @examples
#' 
#' \dontrun{
#' # the example below fits community occupancy models to the sample data in camtrapR
#' # models are fit both in JAGS and Nimble
#' # The data set only contains 5 species and 3 stations, so the results will be nonsense. 
#' # It is only a technical demonstration with the camtrapR workflow
#' # for more complete examples, see vignette 5
#' 
#' data("camtraps")
#' 
#' # create camera operation matrix
#' camop_no_problem <- cameraOperation(CTtable      = camtraps,
#'                                     stationCol   = "Station",
#'                                     setupCol     = "Setup_date",
#'                                     retrievalCol = "Retrieval_date",
#'                                     hasProblems  = FALSE,
#'                                     dateFormat   = "dmy"
#' )
#' 
#' data("recordTableSample")
#' 
#' # make list of detection histories
#' species_to_include <- unique(recordTableSample$Species)
#' 
#'   DetHist_list <- detectionHistory(
#'     recordTable         = recordTableSample,
#'     camOp                = camop_no_problem,
#'     stationCol           = "Station",
#'     speciesCol           = "Species",
#'     recordDateTimeCol    = "DateTimeOriginal",
#'     species              = species_to_include,
#'     occasionLength       = 7,
#'     day1                 = "station",
#'     datesAsOccasionNames = FALSE,
#'     includeEffort        = TRUE,
#'     scaleEffort          = TRUE,
#'     timeZone             = "Asia/Kuala_Lumpur"
#' )
#' 
#' 
#' # create some fake covariates for demonstration
#' sitecovs <- camtraps[, c(1:3)]
#' sitecovs$elevation <- c(300, 500, 600)   
#' 
#' # scale numeric covariates
#' sitecovs[, c(2:4)] <- scale(sitecovs[,-1])
#' 
#' 
#' # bundle input data for communityModel
#' data_list <- list(ylist = DetHist_list$detection_history,
#'                   siteCovs = sitecovs,
#'                   obsCovs = list(effort = DetHist_list$effort))
#' 
#' 
#' # create community model for JAGS
#' modelfile1 <- tempfile(fileext = ".txt")
#' mod.jags <- communityModel(data_list,
#'                            occuCovs = list(fixed = "utm_y", ranef = "elevation"),
#'                            detCovsObservation = list(fixed = "effort"),
#'                            intercepts = list(det = "ranef", occu = "ranef"),
#'                            modelFile = modelfile1)
#' 
#' summary(mod.jags)
#' 
#' # fit in JAGS
#' fit.jags <- fit(mod.jags,
#'                 n.iter = 1000,
#'                 n.burnin = 500,
#'                 chains = 3)   
#' summary(fit.jags)
#' 
#' # response curves (= marginal effect plots)
#' plot_effects(mod.jags, 
#'              fit.jags, 
#'              submodel = "state")
#' plot_effects(mod.jags, 
#'              fit.jags, 
#'              submodel = "det")
#'              
#' # effect sizes plot
#' plot_coef(mod.jags, 
#'           fit.jags, 
#'           submodel = "state")
#' plot_coef(mod.jags, 
#'           fit.jags, 
#'           submodel = "det")              
#' 
#' # create community model for Nimble
#' modelfile2 <- tempfile(fileext = ".txt")
#' mod.nimble <- communityModel(data_list,
#'                              occuCovs = list(fixed = "utm_x", ranef = "utm_y"),
#'                              detCovsObservation = list(fixed = "effort"),
#'                              intercepts = list(det = "ranef", occu = "ranef"),
#'                              modelFile = modelfile2, 
#'                              nimble = TRUE)      # set nimble = TRUE
#' 
#' # load nimbleEcology package 
#' # currently necessary to do explicitly, to avoid additional package dependencies
#' require(nimbleEcology)
#' 
#' # fit uncompiled model in Nimble
#' fit.nimble.uncomp <- fit(mod.nimble, 
#'                          n.iter = 10, 
#'                          chains = 1)
#' 
#' # fit compiled model in Nimble
#' fit.nimble.comp <- fit(mod.nimble, 
#'                        n.iter = 5000, 
#'                        n.burnin = 2500,
#'                        chains = 3, 
#'                        compile = TRUE)
#' 
#' # parameter summary statistics
#' summary(fit.nimble.comp)
#' 
#' 
#' # response curves (= marginal effect plots)
#' plot_effects(mod.nimble, 
#'              fit.nimble.comp, 
#'              submodel = "state")
#' plot_effects(mod.nimble, 
#'              fit.nimble.comp, 
#'              submodel = "det")
#' 
#' # effect sizes plot
#' plot_coef(mod.nimble, 
#'           fit.nimble.comp, 
#'           submodel = "state")
#' plot_coef(mod.nimble, 
#'           fit.nimble.comp, 
#'           submodel = "det")   
#' 
#' # traceplots
#' plot(fit.nimble.comp)
#' 
#' 
#' }
#' 

communityModel <- function(data_list,
                           model = c("Occupancy", "RN"),
                           occuCovs = list(fixed = NULL, independent = NULL, ranef = NULL),
                           detCovs = list(fixed = NULL, ranef = NULL),
                           detCovsObservation = list(fixed = NULL, ranef = NULL),
                           speciesSiteRandomEffect = list(det = FALSE, occu = FALSE),
                           intercepts = list(det = "ranef", occu = "ranef"),
                           effortCov = "effort",
                           richnessCategories = NULL,
                           augmentation = NULL,
                           modelFile = NULL,
                           nimble = FALSE,
                           keyword_quadratic = "_squared")
{   
  
  
  model <- match.arg(model, choices = c("Occupancy", "RN"))
  if(model == "RN" & nimble) stop("Royle-Nichols models are currently not supported in Nimble.")
  
  
  speciesIndex <- "i"      # k in AHMbook
  stationIndex <- "j"      # i in AHMbook
  occasionIndex <- "k"     # j in AHMbook
  
  # this function uses [species, station, occasion] indexing. 
  # AHM book uses: [station, occasion, species]
  
  # occupancy parameters = beta 
  # detection parameters = alpha
  # ...0 -> intercept, ...1-n -> covariates
  
  
    # Check parameter names ====
  # 1. Check occuCovs
  if(!is.list(occuCovs)) stop("occuCovs must be a list")
  if(!all(names(occuCovs) %in% c("fixed", "independent", "ranef"))) {
    invalid_names <- names(occuCovs)[!names(occuCovs) %in% c("fixed", "independent", "ranef")]
    stop(paste("Invalid names in occuCovs:", paste(invalid_names, collapse=", "), 
               ". Only 'fixed', 'independent', and 'ranef' are allowed."))
  }
  
  # 2. Check detCovs
  if(!is.list(detCovs)) stop("detCovs must be a list")
  if(!all(names(detCovs) %in% c("fixed", "independent", "ranef"))) {
    invalid_names <- names(detCovs)[!names(detCovs) %in% c("fixed", "independent", "ranef")]
    stop(paste("Invalid names in detCovs:", paste(invalid_names, collapse=", "), 
               ". Only 'fixed', 'independent', and 'ranef' are allowed."))
  }
  
  # 3. Check detCovsObservation
  if(!is.list(detCovsObservation)) stop("detCovsObservation must be a list")
  if(!all(names(detCovsObservation) %in% c("fixed", "ranef"))) {
    invalid_names <- names(detCovsObservation)[!names(detCovsObservation) %in% c("fixed", "ranef")]
    stop(paste("Invalid names in detCovsObservation:", paste(invalid_names, collapse=", "), 
               ". Only 'fixed' and 'ranef' are allowed."))
  }
  
  # 4. Check speciesSiteRandomEffect
  if(!is.list(speciesSiteRandomEffect)) stop("speciesSiteRandomEffect must be a list")
  if(!all(names(speciesSiteRandomEffect) %in% c("det", "occu"))) {
    invalid_names <- names(speciesSiteRandomEffect)[!names(speciesSiteRandomEffect) %in% c("det", "occu")]
    stop(paste("Invalid names in speciesSiteRandomEffect:", paste(invalid_names, collapse=", "), 
               ". Only 'det' and 'occu' are allowed."))
  }
  
  # 5. Check intercepts
  if(!is.list(intercepts)) stop("intercepts must be a list")
  if(!all(names(intercepts) %in% c("det", "occu"))) {
    invalid_names <- names(intercepts)[!names(intercepts) %in% c("det", "occu")]
    stop(paste("Invalid names in intercepts:", paste(invalid_names, collapse=", "), 
               ". Only 'det' and 'occu' are allowed."))
  }
  
  # Check covariate existence ====
  
  # 1. Check site covariates exist in data_list$siteCovs
  if(exists("siteCovs", where = data_list)) {
    
    if(inherits(data_list$siteCovs, "sf")) {
      # message("site covariates are sf object. Dropping geometry.")
      data_list$siteCovs <- st_drop_geometry(data_list$siteCovs)
    }
    
    # Collect all site covariates from parameter lists
    all_site_covs <- unique(c(
      unlist(occuCovs), 
      unlist(detCovs),
      richnessCategories
    ))
    
    # Remove any NULL values
    all_site_covs <- all_site_covs[!is.null(all_site_covs)]
    
    # Check for piped covariate names (e.g., "cov1|cov2")
    # Extract the first part (actual covariate name) for those
    has_pipe <- grepl("\\|", all_site_covs)
    if(any(has_pipe)) {
      piped_covs <- all_site_covs[has_pipe]
      first_parts <- sapply(strsplit(piped_covs, "\\|"), function(x) x[1])
      all_site_covs <- c(all_site_covs[!has_pipe], first_parts)
    }
    
    # Get rid of "+Species" suffix used for nested random effects
    all_site_covs <- gsub("\\+Species", "", all_site_covs)
    
    # Remove quadratic suffixes
    all_site_covs <- gsub(keyword_quadratic, "", all_site_covs)
    
    # Check if these covariates exist in data_list$siteCovs
    if(length(all_site_covs) > 0) {
      missing_covs <- all_site_covs[!all_site_covs %in% colnames(data_list$siteCovs)]
      if(length(missing_covs) > 0) {
        stop(paste("The following site covariates are specified but not found in data_list$siteCovs:",
                   paste(missing_covs, collapse=", ")))
      }
    }
  } else if(length(unlist(c(occuCovs, detCovs))) > 0 || !is.null(richnessCategories)) {
    stop("Site covariates are specified, but data_list$siteCovs is missing.")
  }
  
  # 2. Check observation covariates exist in data_list$obsCovs
  if(exists("obsCovs", where = data_list)) {
    # Collect all observation covariates from parameter lists
    all_obs_covs <- unique(unlist(detCovsObservation))
    
    # Remove any NULL values
    all_obs_covs <- all_obs_covs[!is.null(all_obs_covs)]
    
    # Check if these covariates exist in data_list$obsCovs
    if(length(all_obs_covs) > 0) {
      missing_covs <- all_obs_covs[!all_obs_covs %in% names(data_list$obsCovs)]
      if(length(missing_covs) > 0) {
        stop(paste("The following observation covariates are specified but not found in data_list$obsCovs:",
                   paste(missing_covs, collapse=", ")))
      }
    }
  } else if(length(unlist(detCovsObservation)) > 0) {
    stop("Observation covariates are specified, but data_list$obsCovs is missing.")
  }
  
  # 3. Make sure effortCov exists
  if(!effortCov %in% names(data_list$obsCovs)) {
    stop(paste0("effortCov '", effortCov, "' is not found in data_list$obsCovs."))
  }
  
  
  # define prior distributions to be used throughout (maybe make it an argument later)
  prior_list <- list(dnorm = "dnorm(0, 0.05)",
                     dgamma = "dgamma(0.1, 0.1)")
  
  # define inits
  inits_list <- list(
    inits_runif_mean_low = list("runif", list(1, -4.5, -1.5)),   # n_draws, min, max
    inits_runif_mean_0 = list("runif", list(1, -1, 1)), 
    inits_runif_tau  = list("runif", list(1, 0.01, 0.1)),
    inits_runif_omega = list("runif", list(1, 0, 1))
  )
  
  stopifnot(is.list(intercepts))
  stopifnot(length(intercepts) == 2)
  stopifnot(names(intercepts) %in% c("det", "occu"))
  stopifnot(intercepts$det %in% c("fixed", "ranef", "independent"))
  stopifnot(intercepts$occu %in% c("fixed", "ranef", "independent"))
  stopifnot(is.list(speciesSiteRandomEffect))
  stopifnot(is.logical(speciesSiteRandomEffect$det))
  
  if(exists("occu", speciesSiteRandomEffect)) {
    stopifnot(is.logical(speciesSiteRandomEffect$occu))
    if(speciesSiteRandomEffect$occu) stop("speciesSiteRandomEffect$occu must be FALSE")
  }
  
  
  if(!is.list(data_list)) stop("data_list must be a list")
  
  if(exists("ylist", where = data_list)) {
    if(!inherits(data_list$ylist, "list")) stop("data_list$ylist must be a list")
  } else {
    stop("data_list$ylist is missing")
  }
  
  if(exists("obsCovs", where = data_list)) {
    if(!inherits(data_list$obsCovs, "list")) stop("data_list$obsCovs must be a list")
  } else {
    stop("data_list$obsCovs is missing")
  }
  
  # data frame with information about intercepts
  covariate_info <- data.frame(param = c("intercept", "intercept"),
                           submodel = c("det", "state"),
                           covariate_type = NA,
                           covariate = NA,
                           data_type = NA,         # categ / numeric
                           is_quadratic = FALSE,
                           has_quadratic = FALSE,
                           ranef = c(ifelse(intercepts$det == "ranef", T, F), 
                                     ifelse(intercepts$occu == "ranef", T, F)), 
                           ranef_nested = FALSE,
                           ranef_cov = NA, 
                           independent = c(ifelse(intercepts$det == "independent", T, F), 
                                           ifelse(intercepts$occu == "independent", T, F)), 
                           coef = c("alpha0", "beta0"))
                           
  
  
  # keyword for nested random effects
  keyword_nested <- "+Species"
  
  # keyword for quadratic covariate effect
  # keyword_quadratic <- "_squared"
  
  # parse covariates ####
  
  ## observation level covariates  ####
  if(!effortCov %in% names(data_list$obsCovs)) stop(paste0("effortCov '", effortCov, "' is not in the list of observation covariates (data_list$obsCovs)."))
  
  
  if(!is.null(unlist(detCovsObservation))) {
    covariate_info <- rbind(covariate_info, get_cov_info (cov = detCovsObservation, 
                                                          keyword_nested, 
                                                          keyword_quadratic, 
                                                          data_list, 
                                                          type = "obs", 
                                                          submodel = "det"))
    }
  
  
  obs_covariates <- data_list$obsCovs
  obs_covariates_numeric <- names(obs_covariates)[sapply(obs_covariates, is.numeric)]
  obs_covariates_categ   <- names(obs_covariates)[sapply(obs_covariates, is.character)]
  
  if(!all(names(detCovsObservation) %in% c("fixed", "ranef"))) stop("detCovsObservation can currentlly only be 'fixed' or 'ranef'")
  
  if(!is.null(detCovsObservation$fixed)) {
    for(i in detCovsObservation$fixed) {
      if(i == effortCov) next
      if(any(is.na(data_list$obsCovs[[i]]))) warning(paste("there are NAs in obsCovs:", i))
    }
  }
  
  if(!is.null(detCovsObservation$ranef)) {
    for(i in detCovsObservation$ranef) {
      if(i == effortCov) next
      if(any(is.na(data_list$obsCovs[[i]]))) warning(paste("there are NAs in obsCovs:", i))
    }
  }
  
  detCovsObservation_categ <- list(fixed = NULL, ranef = NULL)
  
  if(!is.null(detCovsObservation$fixed)) {
    if(any(!detCovsObservation$fixed %in% obs_covariates_numeric)) {
      if(!all(detCovsObservation$fixed [which(!detCovsObservation$fixed %in% obs_covariates_numeric)] %in% obs_covariates_categ)) stop(paste("Detection covariate ", paste(detCovsObservation$fixed [which(!detCovsObservation$fixed %in% obs_covariates_numeric)], collapse = ", "), 
                                                                                                                                                                     "is not a factor"))
      if(nimble == TRUE) warning("Currently categorical observation-level covariates are only supported in JAGS.\nIn Nimble they can only be used in uncompiled models (even that is experimental). Change model structure or set 'nimble = FALSE'", call. = FALSE)
      
      detCovsObservation_categ$fixed <- detCovsObservation$fixed[detCovsObservation$fixed %in% obs_covariates_categ]
      detCovsObservation$fixed       <- detCovsObservation$fixed[detCovsObservation$fixed %in% obs_covariates_numeric]
      if(length(detCovsObservation$fixed) == 0) detCovsObservation <- list(fixed = NULL, ranef = detCovsObservation$ranef)
    }
  }
  
  if(!is.null(detCovsObservation$ranef)) {
    if(any(!detCovsObservation$ranef %in% obs_covariates_numeric)) {
      if(!all(detCovsObservation$ranef [which(!detCovsObservation$ranef %in% obs_covariates_numeric)] %in% obs_covariates_categ)) stop(paste("Detection covariate ", paste(detCovsObservation$ranef [which(!detCovsObservation$ranef %in% obs_covariates_numeric)], collapse = ", "), 
                                                                                                                                                                     "is not a factor"))
      if(nimble == TRUE) warning("Currently categorical observation-level covariates are only supported in JAGS.\nIn Nimble they can only be used in uncompiled models (even that is experimental). Change model structure or set 'nimble = FALSE'", call. = FALSE)
      
      detCovsObservation_categ$ranef <- detCovsObservation$ranef[detCovsObservation$ranef %in% obs_covariates_categ]
      detCovsObservation$ranef       <- detCovsObservation$ranef[detCovsObservation$ranef %in% obs_covariates_numeric]
      if(length(detCovsObservation$ranef) == 0) detCovsObservation <- list(fixed = detCovsObservation$fixed, ranef = NULL)
    }
  }
  
  # convert character matrices to integer
  if(length(obs_covariates_categ) >= 1) {
    for(i in 1:length(obs_covariates_categ)){
      levels_tmp <- levels(as.factor(data_list$obsCovs[[obs_covariates_categ[i]]]))
      matrix_tmp <- match(data_list$obsCovs[[obs_covariates_categ[i]]], levels_tmp)
      data_list$obsCovs <- c(data_list$obsCovs, 
                             xxx = list(matrix(matrix_tmp, 
                                               nrow = dim(data_list$obsCovs[[obs_covariates_categ[i]]])[1],
                                               ncol = dim(data_list$obsCovs[[obs_covariates_categ[i]]])[2])))
      names(data_list$obsCovs)[length(data_list$obsCovs)] <- paste0(obs_covariates_categ[i], "_integer")
      attr(data_list$obsCovs[[length(data_list$obsCovs)]], "levels") <- levels_tmp
    }
  }
  
  ## site covariates ####
  
  if(exists("siteCovs", where = data_list)) {
    if(!inherits(data_list$siteCovs, "data.frame")) stop("data_list$siteCovs must be a data.frame")
  } else {
    warning("data_list$siteCovs does not exist. No site covariates available.\n", call. = FALSE)
  }
  
  
  # get covariate information
  if(!is.null(unlist(detCovs))) {
    if(!all(c("obsCovs", "siteCovs") %in% names(data_list))) {
      stop("detCovs is defined, but data_list does not contain siteCovs or obsCovs.", call. = F)
    } else {
      covariate_info <- rbind(covariate_info, 
                              get_cov_info (detCovs, keyword_nested, keyword_quadratic, data_list, type = "site", submodel = "det"))
    }
  }
  
  if(!is.null(unlist(occuCovs))) {
    if(! "siteCovs" %in% names(data_list)) {
      stop("occuCovs is defined, but data_list does not contain siteCovs.", call. = F)
    } else {
      covariate_info <- rbind(covariate_info, 
                              get_cov_info (occuCovs, keyword_nested, keyword_quadratic, data_list, type = "site", submodel = "state")) 
    }
  }
  
  covariates_original <- data_list$siteCovs

  # identify required covariates:
  site_covs_tmp <- unlist(c(occuCovs, detCovs, richnessCategories))
  if(length(site_covs_tmp) >= 1){
    covariatenames_needed <- unique(unlist(strsplit(site_covs_tmp, split = "|", fixed = T)))
  } else { 
    covariatenames_needed <- NULL
  }
  
  
  # remove "+Species" which indicates nested random effects
  covariatenames_needed <- gsub(keyword_nested, "", covariatenames_needed, fixed = T)
  
  # subset covariates
  covariates <- covariates_original [, colnames(covariates_original) %in% covariatenames_needed, drop = FALSE]
  
  # make a note of which covariates are in model (for summary method)
  attr(data_list$siteCovs, "in_model") <- colnames(covariates_original) %in% colnames(covariates)
  
  
  
  # identify numeric and categorical covariates
  if(ncol(covariates) != 0) {
    covariates_numeric <- names(covariates)[sapply(covariates, is.numeric)]
    covariates_categ   <- names(covariates)[sapply(covariates, is.factor)]
    covariates_char    <- names(covariates)[sapply(covariates, is.character)]
  } else {
    covariates_numeric <- NULL
    covariates_categ   <- NULL
    covariates_char    <- NULL
  }
  
  # throw error for character covariates
  
  if(length(covariates_char) >= 1) {
    # browser()
    stop(paste("Covariate", covariates_char, "is character. Please convert to factor."), call. = FALSE)
    
    # attempt to convert to factor?
  }
    
  # paste cov1|cov2 (to check for random effects other than species, and nested random effects)
  covariates_numeric_categ <- unlist(sapply(covariates_numeric, paste, covariates_categ, sep = "|", simplify = F), use.names = F)
  covariates_numeric_categ <- c(covariates_numeric_categ, paste0(covariates_numeric_categ, keyword_nested))
  covariates_categ_categ   <- unlist(sapply(covariates_categ, paste, covariates_categ, sep = "|", simplify = F), use.names = F)
  

  if(!is.null(richnessCategories)) {
    if(length(richnessCategories) > 1) stop("richnessCategories can only have length 1, or be NULL")
    if(!richnessCategories %in% covariates_categ) stop(paste("richnessCategories", richnessCategories, "is not a categorical covariate in the covariate data frame"))
    if(nimble) warning("richnessCategories will be ignored because nimble = TRUE")
    occuIntercept_categ <- list(fixed = richnessCategories)
  } else {
    occuIntercept_categ <- list(fixed = NULL)
  }
  

  
  occuCovs_categ <- list(fixed = NULL, ranef = NULL)
  detCovs_categ  <- list(fixed = NULL, ranef = NULL)
  
  ## input checks on covariates  ####
  
  if(!is.null(occuCovs$fixed)) {
    if(any(!occuCovs$fixed %in% colnames(covariates))) {
      stop(paste(occuCovs$fixed [!occuCovs$fixed %in% colnames(covariates)], collapse = ", "), " is not in covariates")
    }
    if(any(!occuCovs$fixed %in% covariates_numeric)) {
      if(!all(occuCovs$fixed [which(!occuCovs$fixed %in% covariates_numeric)] %in% covariates_categ)) stop(paste("Occupancy covariate ", paste(occuCovs$fixed [which(!occuCovs$fixed %in% covariates_numeric)], collapse = ", "), 
                                                                                                                                         "is not a factor"))
      occuCovs_categ$fixed <- occuCovs$fixed[occuCovs$fixed %in% covariates_categ]
      occuCovs$fixed       <- occuCovs$fixed[occuCovs$fixed %in% covariates_numeric]
      if(length(occuCovs$fixed) == 0) occuCovs <- modifyList(occuCovs, list(fixed = NULL))
    }
  }
  
  if(!is.null(occuCovs$independent)) {
    if(any(!occuCovs$independent %in% colnames(covariates))) {
      stop(paste(occuCovs$independent [!occuCovs$independent %in% colnames(covariates)], collapse = ", "), " is not in covariates")
    }
    if(any(!occuCovs$independent %in% covariates_numeric)) {
      if(!all(occuCovs$independent [which(!occuCovs$independent %in% covariates_numeric)] %in% covariates_categ)) stop(paste("Occupancy covariate ", paste(occuCovs$independent [which(!occuCovs$independent %in% covariates_numeric)], collapse = ", "), 
                                                                                                                 "is not a factor"))
      
      # prohibit independent effects of categorical covariates
      if(any(occuCovs$independent %in% covariates_categ)) stop("independent effects of categorical site covariates are currently not supported")
      
      # occuCovs_categ$independent <- occuCovs$independent[occuCovs$independent %in% covariates_categ]
      # occuCovs$independent       <- occuCovs$independent[occuCovs$independent %in% covariates_numeric]
      # if(length(occuCovs$independent) == 0) occuCovs <- modifyList(occuCovs, list(independent = NULL))
    }
  }
  
  
  if(!is.null(occuCovs$ranef)) {
    if(any(occuCovs$ranef %in% covariates_categ_categ)) stop("random effects of categorical site covariates on other categorical site covariates are not supported yet (except for random effect of species)")
    
    if(any(!occuCovs$ranef %in% c(covariates_numeric, covariates_categ, covariates_numeric_categ))) {
      stop(paste(occuCovs$ranef [!occuCovs$ranef %in% c(covariates_numeric, covariates_numeric_categ)], collapse = ", "), " is not in covariates")
    }
    if(any(!occuCovs$ranef %in% covariates_numeric)) {  
      if(!all(occuCovs$ranef [which(!occuCovs$ranef %in% c(covariates_numeric, covariates_numeric_categ))] %in% covariates_categ)) stop(paste("Occupancy covariate ", paste(occuCovs$ranef [which(!occuCovs$ranef %in% covariates_numeric)], collapse = ", "), 
                                                                                                                                                                      "is not a factor"))
      occuCovs_categ$ranef <- c(occuCovs_categ$ranef, 
                                occuCovs$ranef[occuCovs$ranef %in% c(covariates_categ, covariates_categ_categ)])
      if(length(occuCovs_categ$ranef) == 0) occuCovs_categ$ranef <- NULL
      occuCovs$ranef       <- occuCovs$ranef[occuCovs$ranef %in% c(covariates_numeric, covariates_numeric_categ)]
      if(length(occuCovs$ranef) == 0) occuCovs <- modifyList(occuCovs, list(ranef = NULL))
    }
  }
  
  if(!is.null(detCovs$fixed)) {
    if(any(!detCovs$fixed %in% colnames(covariates))) {
      stop(paste(detCovs$fixed [!detCovs$fixed %in% colnames(covariates)], collapse = ", "), " is not in covariates")
    }
    if(any(!detCovs$fixed %in% covariates_numeric)) {
      if(!all(detCovs$fixed [which(!detCovs$fixed %in% covariates_numeric)] %in% covariates_categ)) stop(paste("Detection covariate ", paste(detCovs$fixed [which(!detCovs$fixed %in% covariates_numeric)], collapse = ", "), 
                                                                                                                                       "is not a factor"))
      
      detCovs_categ$fixed <- detCovs$fixed[detCovs$fixed %in% covariates_categ]
      detCovs$fixed       <- detCovs$fixed[detCovs$fixed %in% covariates_numeric]
      if(length(detCovs$fixed) == 0) detCovs <- modifyList(detCovs, list(fixed = NULL))
    }
  }
  
  
  if(!is.null(detCovs$independent)) {
    if(any(!detCovs$independent %in% colnames(covariates))) {
      stop(paste(detCovs$independent [!detCovs$independent %in% colnames(covariates)], collapse = ", "), " is not in covariates")
    }
    if(any(!detCovs$independent %in% covariates_numeric)) {
      if(!all(detCovs$independent [which(!detCovs$independent %in% covariates_numeric)] %in% covariates_categ)) stop(paste("Detection covariate ", paste(detCovs$independent [which(!detCovs$independent %in% covariates_numeric)], collapse = ", "), 
                                                                                                               "is not a factor"))

      # prohibit independent effects of categorical covariates
      if(any(detCovs$independent %in% covariates_categ)) stop("independent effects of categorical site covariates are currently not supported")
      
      # detCovs_categ$independent <- detCovs$independent[detCovs$independent %in% covariates_categ]
      # detCovs$independent       <- detCovs$independent[detCovs$independent %in% covariates_numeric]
      # if(length(detCovs$independent) == 0) detCovs <- modifyList(detCovs, list(independent = NULL))
    }
  }
  
  
  if(!is.null(detCovs$ranef)) {
    
    if(any(detCovs$ranef %in% covariates_categ_categ)) stop("random effects of categorical site covariates on other categorical site covariates are not supported yet (except for random effect of species)")
    
    if(any(!detCovs$ranef %in% c(colnames(covariates), covariates_categ, covariates_numeric_categ))) {
      stop(paste(detCovs$ranef [!detCovs$ranef %in% colnames(covariates)], collapse = ", "), " is not in covariates")
    }
    
    if(any(!detCovs$ranef %in% covariates_numeric)) {
      if(!all(detCovs$ranef [which(!detCovs$ranef %in% c(covariates_numeric, covariates_numeric_categ))] %in% covariates_categ)) stop(paste("Detection covariate ", paste(detCovs$ranef [which(!detCovs$ranef %in% c(covariates_numeric, covariates_numeric_categ))], collapse = ", "), 
                                                                                                                                                                    "is not a factor"))
      detCovs_categ$ranef <- c(detCovs_categ$ranef, 
                               detCovs$ranef[detCovs$ranef %in% c(covariates_categ, covariates_categ_categ)])
      if(length(detCovs_categ$ranef) == 0) detCovs_categ$ranef <- NULL
      detCovs$ranef       <- detCovs$ranef[detCovs$ranef %in% c(covariates_numeric, covariates_numeric_categ)]
      if(length(detCovs$ranef) == 0) detCovs <- modifyList(detCovs, list(ranef = NULL))
    }
  }
  
  # check if there area NAs in site covariates
  if(length(covariates_numeric) >= 1) {
    na_in_covs <- apply(data_list$siteCovs[, covariates_numeric, drop = FALSE], 2, FUN = function(x) any(is.na(x)))
    if(any(na_in_covs)) stop(paste("There are NAs in site covariate(s): ", paste(names(na_in_covs)[na_in_covs], collapse = ", ")))
  }
  
  if(length(covariates_categ) >= 1) {
    na_in_covs <- apply(data_list$siteCovs[, covariates_categ, drop = FALSE], 2, FUN = function(x) any(is.na(x)))
    if(any(na_in_covs)) stop(paste("There are NAs in site covariate(s): ", paste(names(na_in_covs)[na_in_covs], collapse = ", ")))
  }
  
  # check that dimensions / pattern of NA / not NA matches between observations and effort
  if (!all(dim(data_list$ylist[[1]]) == dim(data_list$obsCovs[[effortCov]]))) {
    stop("detection history and effort matrix must have the same dimensions")
  }
  
  all_matching <- all(!is.na(data_list$ylist[[1]]) == !is.na(data_list$obsCovs[[effortCov]]))
  if(isFALSE(all_matching)) stop("NAs in detection history and effort matrix differ")
  
  
  # check that dimensions / pattern of NA / not NA matches between observations and observation covariates
  obs_cov_tmp <- which(covariate_info$param  == "param" & covariate_info$covariate_type == "obsCovs")
  
  if(length(obs_cov_tmp) >= 1) {
    lapply(obs_cov_tmp, FUN = function(obs_cov_tmp_i) {
      obs_cov_tmp_i_name <- covariate_info$covariate[obs_cov_tmp_i]
      if (!all(dim(data_list$ylist[[1]]) == dim(data_list$obsCovs[[obs_cov_tmp_i_name]]))) {
        stop(paste("detection history and", obs_cov_tmp_i_name, "matrix must have the same dimensions"))
      }
      
      # all_matching <- all(!is.na(data_list$ylist[[1]]) == !is.na(data_list$obsCovs[[obs_cov_tmp_i_name]]))
      # browser()
      NAs_in_obs_cov_not_in_dethist <- !which(is.na(data_list$obsCovs[[obs_cov_tmp_i_name]])) %in% which(is.na(data_list$ylist[[1]]))
      if(any(NAs_in_obs_cov_not_in_dethist)) stop(paste("NAs in detection history and", obs_cov_tmp_i_name, "matrix differ"))
    })
  }
  
  # check is covariates are scaled
  if(length(covariates_numeric) >= 1) {
    if(any(abs(sapply(data_list$siteCovs[, covariates_numeric], mean)) > 1)){
      message("Site covariates may not be scaled. Scaling is strongly recommended for numeric covariates.")
    }
  }
  
                       
  
  # some more checks for dimensions and data types
  dim_dethist1 <- dim(data_list$ylist[[1]])
  if(!all(sapply(data_list$ylist,   FUN = function(x) all(dim(x) == dim_dethist1)))) stop("dimensions differ between items in data_list$ylist")
  if(!all(sapply(data_list$obsCovs, FUN = function(x) all(dim(x) == dim_dethist1)))) stop("dimensions differ between items in data_list$ylist and data_list$obsCovs")
  
  if(exists("siteCovs", where = data_list)) {
    if(inherits(data_list$siteCovs, "data.frame")) if(nrow(data_list$siteCovs) != dim_dethist1[1]) stop("nrow differs between data_list$siteCovs and the detection histories in data_list$ylist")
  }
  
  
  ## Data augmentation ####
  
  # check if all-0 detection histories present
  if(any(sapply(data_list$ylist, sum, na.rm = T) == 0)) {
    warning("There are all-0 detection histories. It is recommended to instead provide only detected species, and control data augmentation via 'augmentation'.")
  }
  
  
  if(!is.null(augmentation)) {
    if(length(augmentation) != 1) stop("'augmentation' must have length 1")
    if(is.null(names(augmentation))) stop("'augmentation' must be named")
    if(!names(augmentation) %in% c("maxknown", "full"))  stop("name of 'augmentation' must be either 'maxknown' or 'full'")
    augment <- names(augmentation)
    if(inherits(augmentation, "list")) augmentation <- as.vector(augmentation, mode = "numeric")
    
    if(augmentation <= length(data_list$ylist)) stop("'augmentation' must be larger than the number of detected species") 
    
    nspec <- length(data_list$ylist) # number of seen species
    
    nz <- augmentation - length(data_list$ylist)   # number of unseen species added
    data_list$ylist <- c(data_list$ylist, replicate(augmentation - length(data_list$ylist), data_list$ylist[[1]] * 0, simplify = FALSE))
    
  } else {
    augment = "none"
  }
  
  # set indices to use in loops
  speciesMax  <- "M"
  stationMax  <- "J"
  occasionMax <- "maxocc"
  
  speciesMax_value  <- length(data_list$ylist)
  stationMax_value  <- nrow(data_list$ylist[[1]])
  occasionMax_value <- ncol(data_list$ylist[[1]])
  
  
  
  # Priors  ####
  
  model_start <- "model{\n### PRIORS \n"
  
  
  ## Intercept - occupancy - hyperpriors   ####
  
  intercept_occu <- interceptPriors(effect = intercepts$occu, 
                                    type = "occupancy", 
                                    prior_list = prior_list, 
                                    inits_list = inits_list, 
                                    speciesIndex = speciesIndex,
                                    speciesMax = speciesMax,
                                    speciesMax_value = speciesMax_value)
  
  priors_intercept_occu <- intercept_occu$prior
  beta0_formula         <- intercept_occu$formula
  
  ## Intercept - detection - hyperpriors ####
  
  intercept_det <- interceptPriors(effect = intercepts$det, 
                                   type = "detection", 
                                   prior_list = prior_list, 
                                   inits_list = inits_list, 
                                   speciesIndex = speciesIndex,
                                   speciesMax = speciesMax,
                                   speciesMax_value = speciesMax_value)
  
  priors_intercept_det <- intercept_det$prior
  alpha0_formula       <- intercept_det$formula
  
  
  ##  data augmentation priors ####
  
  if(augment == "full") {
    priors_omega <- paste(
      "## Data augmentation parameter",
      "omega ~ dunif(0, 1)",
      "\n", sep = "\n"
    )
    attr(priors_omega, "params") <- c("omega")
    attr(priors_omega, "inits") <- list(omega = inits_list$inits_runif_omega)
  } else {
    priors_omega <- paste(
      "## Data augmentation parameter",
      "# < empty > ",
      "\n", sep = "\n"
    )
  }
  
  
  ## obs Cov - cont - fixed effect priors ####
  prior_obs_header_fixed <- paste("## Continuous observation-level covariates on detection - Fixed effects\n")
  
  if(!is.null(detCovsObservation$fixed)) {
    priors_obscovs_det_fixed <- list()
    
    for(obsCovs_det_index in 1:length(detCovsObservation$fixed)){
      current_cov <- detCovsObservation$fixed[obsCovs_det_index]
      
      priors_obscovs_det_fixed[[obsCovs_det_index]] <- paste(
        paste("# Covariate:", current_cov),
        paste0("alpha.obs.fixed.cont.", current_cov, " ~ ", prior_list$dnorm), 
        "\n", 
        sep = "\n")
      attr(priors_obscovs_det_fixed, "params") <- paste0("alpha.obs.fixed.cont.", detCovsObservation$fixed)
      
      inits_tmp <- vector(mode = "list", length = length(detCovsObservation$fixed))
      names(inits_tmp) <- c(paste0("alpha.obs.fixed.cont.", detCovsObservation$fixed))
      
      for(i in 1:length(inits_tmp)){
        inits_tmp[[i]] <- inits_list$inits_runif_mean_0      #list("runif", list(1))
      }
      attr(priors_obscovs_det_fixed, "inits") <- inits_tmp
      rm(inits_tmp)
    }
  } else {
    priors_obscovs_det_fixed <- list("# < empty > \n\n")
    attr(priors_obscovs_det_fixed, "params") <- NULL
  }
  
  ## obs Cov - cont - random effect priors ####
  prior_obs_header_ranef <- paste("## Continuous observation-level covariates on detection - with random effects\n")
  
  if(!is.null(detCovsObservation$ranef)){
    priors_obscovs_det_ranef <- list()
    alpha_draws_ranef_obs_cont <- list()
    
    for(obsCovs_det_index in 1:length(detCovsObservation$ranef)){
      current_cov <- detCovsObservation$ranef[obsCovs_det_index]
      priors_obscovs_det_ranef[[obsCovs_det_index]] <- paste(
        paste("# Observation Covariate:", current_cov),
        paste0("alpha.obs.ranef.cont.", current_cov, ".mean", " ~ ", prior_list$dnorm),
        paste0("alpha.obs.ranef.cont.", current_cov, ".tau", " ~ ", prior_list$dgamma),
        paste0("alpha.obs.ranef.cont.", current_cov, ".sigma", " <- sqrt(1 / alpha.obs.ranef.cont.", current_cov, ".tau)"),
        "\n", 
        sep = "\n")
      
      alpha_draws_ranef_obs_cont[[obsCovs_det_index]] <- paste("# continuous detection covariates with random effects:", 
                                                               paste0("alpha.obs.ranef.cont.", current_cov, "[", speciesIndex, "] ~ dnorm(alpha.obs.ranef.cont.", current_cov, ".mean", ", alpha.obs.ranef.cont.", current_cov, ".tau)"),
                                                               sep = "\n")
    }
    
    attr(priors_obscovs_det_ranef, "params") <- c(paste0("alpha.obs.ranef.cont.", detCovsObservation$ranef, ".mean"),
                                                  paste0("alpha.obs.ranef.cont.", detCovsObservation$ranef, ".sigma"))
    
    
    # make list of parameters
    inits_tmp <- vector(mode = "list", length = length(detCovsObservation$ranef) * 2)
    names(inits_tmp) <- c(paste0("alpha.obs.ranef.cont.", detCovsObservation$ranef, ".mean"),
                          paste0("alpha.obs.ranef.cont.", detCovsObservation$ranef, ".tau"))
    
    for(i in 1:length(inits_tmp)){
      if(endsWith(names(inits_tmp)[i], "mean")) inits_tmp[[i]] <- inits_list$inits_runif_mean_0
      if(endsWith(names(inits_tmp)[i], "tau"))  inits_tmp[[i]] <- inits_list$inits_runif_tau
    }
    attr(priors_obscovs_det_ranef, "inits") <- inits_tmp
    rm(inits_tmp)
    
    names(alpha_draws_ranef_obs_cont) <- paste0("alpha.obs.ranef.cont.", detCovsObservation$ranef)
    
  } else {
    priors_obscovs_det_ranef <- list("# < empty > \n\n")
    attr(priors_obscovs_det_ranef, "params") <- NULL
    alpha_draws_ranef_obs_cont <- "# continuous observation-level detection covariates: no random effect of species"
  }
  
  
  
  
  ## obs Cov - categ - fixed effect priors ####
  prior_obs_categ_header_fixed <- paste("## Categorical observation-level covariates on detection - Fixed effect\n")
  
  if(!is.null(detCovsObservation_categ$fixed)){
    priors_obscovs_det_categ_fixed <- list()
    
    attr(priors_obscovs_det_categ_fixed, "n_levels") <- rep(NA, times = length(detCovsObservation_categ$fixed))
    
    for(obsCovs_det_categ_index in 1:length(detCovsObservation_categ$fixed)){
      
      obsCovCat_tmp_fact <- factor(data_list$obsCovs[[detCovsObservation_categ$fixed[obsCovs_det_categ_index]]])
      
      current_cov <- detCovsObservation_categ$fixed[obsCovs_det_categ_index]
      current_cov_values <- obsCovCat_tmp_fact   #covariates[, current_cov]
      index_letter <- paste0("index_cat_fixed_obs_categ_", current_cov)
      
      priors_obscovs_det_categ_fixed[[obsCovs_det_categ_index]] <- paste(
        paste("# Covariate:", current_cov),
        paste0("alpha.obs.fixed.categ.", current_cov, "[1] <- 0"),
        paste0("for(", index_letter, " in 2:", length(levels(current_cov_values)), ") {"),
        paste0("alpha.obs.fixed.categ.", current_cov, "[", index_letter, "] ~ ", prior_list$dnorm),
        "}",
        "\n", 
        sep = "\n")
      attr(priors_obscovs_det_categ_fixed, "n_levels")[obsCovs_det_categ_index] <- length(levels(current_cov_values))
    }
    attr(priors_obscovs_det_categ_fixed, "params") <- paste0("alpha.obs.fixed.categ.", detCovsObservation_categ$fixed)
    attr(priors_obscovs_det_categ_fixed, "index_letters") <- paste0("index_cat_fixed_obs_categ_", detCovsObservation_categ$fixed)
    
    inits_tmp <- vector(mode = "list", length = length(detCovsObservation_categ$fixed))
    names(inits_tmp) <- paste0("alpha.obs.fixed.categ.", detCovsObservation_categ$fixed)
    for(n in 1:length(detCovsObservation_categ$fixed)) {
      inits_tmp[[n]][1] <- NA
      inits_tmp[[n]][2:attr(priors_obscovs_det_categ_fixed, "n_levels")[n]] <- rnorm(attr(priors_obscovs_det_categ_fixed, "n_levels")[n] - 1)
    }
    attr(priors_obscovs_det_categ_fixed, "inits") <- inits_tmp
    rm(inits_tmp)
    
  } else {
    priors_obscovs_det_categ_fixed <- list("# < empty > \n\n")
    attr(priors_obscovs_det_categ_fixed, "params") <- NULL
  }
  
  
  ## obs Cov - categ - random effect priors ####
  
  prior_obs_categ_header_ranef <- paste("## Categorical observation-level covariates on detection - with random effects\n")
  if(!is.null(detCovsObservation_categ$ranef)){
    priors_obscovs_det_categ_ranef <- list()
    alpha_draws_ranef_obs_categ <- list()
    
    attr(priors_obscovs_det_categ_ranef, "n_levels") <- rep(NA, times = length(detCovsObservation_categ$ranef))
    
    for(obsCovs_det_categ_index in 1:length(detCovsObservation_categ$ranef)){
      
      obsCovCat_tmp_fact <- factor(data_list$obsCovs[[detCovsObservation_categ$ranef[obsCovs_det_categ_index]]])
      
      current_cov <- detCovsObservation_categ$ranef[obsCovs_det_categ_index]
      current_cov_values <- obsCovCat_tmp_fact   #covariates[, current_cov]
      index_letter <- paste0("index_cat_ranef_obs_categ_", current_cov)
      
      
      
      priors_obscovs_det_categ_ranef[[obsCovs_det_categ_index]] <- paste(
        #paste("# Covariate:", current_cov),
        paste("# Community mean effects of", current_cov),
        paste0("alpha.obs.ranef.categ.", current_cov, ".mean[1] <- 0"),
        paste0("alpha.obs.ranef.categ.", current_cov, ".sigma[1] <- 0"),     # is included in random effect of intercept already
        paste0("alpha.obs.ranef.categ.", current_cov, ".tau[1] <- 0"),
        "",
        paste0("for(", index_letter, " in 2:", length(unique(current_cov_values)), ") {"),
        paste0("alpha.obs.ranef.categ.", current_cov, ".mean[", index_letter, "] ~ ", prior_list$dnorm),   
        paste0("alpha.obs.ranef.categ.", current_cov,  ".tau[", index_letter, "] ~ ", prior_list$dgamma),
        paste0("alpha.obs.ranef.categ.", current_cov,  ".sigma[", index_letter, "] <- sqrt(1 / ", "alpha.obs.ranef.categ.", current_cov,  ".tau[", index_letter, "]", ")"),
        "}",
        "",
        sep = "\n")
      attr(priors_obscovs_det_categ_ranef, "n_levels")[obsCovs_det_categ_index] <- length(unique(current_cov_values))
      
      
      alpha_draws_ranef_obs_categ[[obsCovs_det_categ_index]] <- paste("# categorical detection covariates with random effects:", 
                                                                      paste0("alpha.obs.ranef.categ.", current_cov, "[", speciesIndex, ", 1] <- 0"), 
                                                                      paste0("for (", index_letter, " in 2:",  length(unique(current_cov_values)), "){"),
                                                                      paste0("alpha.obs.ranef.categ.", current_cov, "[", speciesIndex, ", ", index_letter, "] ~ dnorm(alpha.obs.ranef.categ.", current_cov, ".mean[", index_letter, "], alpha.obs.ranef.categ.", current_cov, ".tau[", index_letter, "])"),
                                                                      "}", sep = "\n")
    }
    
    attr(priors_obscovs_det_categ_ranef, "params") <- c(paste0("alpha.obs.ranef.categ.", detCovsObservation_categ$ranef, ".mean"),
                                                        paste0("alpha.obs.ranef.categ.", detCovsObservation_categ$ranef, ".sigma"))
    
    attr(priors_obscovs_det_categ_ranef, "index_letters") <- paste0("index_cat_ranef_obs_", detCovsObservation_categ$ranef)
    
    inits_tmp <- vector(mode = "list", length = length(detCovsObservation_categ$ranef)  * 2)
    names(inits_tmp) <- c(paste0("alpha.obs.ranef.categ.", detCovsObservation_categ$ranef, ".mean"),
                          paste0("alpha.obs.ranef.categ.", detCovsObservation_categ$ranef, ".tau"))
    
    for(n in 1:length(detCovsObservation_categ$ranef)) {
      inits_tmp[[n]][1] <- NA
      inits_tmp[[n]][2:attr(priors_obscovs_det_categ_ranef, "n_levels")[n]] <- rnorm(attr(priors_obscovs_det_categ_ranef, "n_levels")[n] - 1)
      
      inits_tmp[[n + length(detCovsObservation_categ$ranef)]][1] <- NA
      inits_tmp[[n + length(detCovsObservation_categ$ranef)]][2:attr(priors_obscovs_det_categ_ranef, "n_levels")[n]] <- abs(rnorm(attr(priors_obscovs_det_categ_ranef, "n_levels")[n] - 1))
      
    }
    attr(priors_obscovs_det_categ_ranef, "inits") <- inits_tmp
    rm(inits_tmp)
    
    names(alpha_draws_ranef_obs_categ) <- paste0("alpha.obs.ranef.categ.", detCovsObservation_categ$ranef)
  } else {
    priors_obscovs_det_categ_ranef <- list("# < empty > \n\n")
    attr(priors_obscovs_det_categ_ranef, "params") <- NULL
    alpha_draws_ranef_obs_categ <- "# categorical observation covariates: no random effect of species"
  }
  
  
  
  
  
  ## det Cov - cont - fixed effect priors ####
  prior_det_header_fixed <- paste("## Continuous site covariates on detection - Fixed effects\n")
  
  priors_sitecovs_det_fixed <- fixedEffectPriors (effect_names = detCovs$fixed,
                                                  type = "detection", 
                                                  prior_list = prior_list, 
                                                  stationIndex = stationIndex)
  
  
  
  ## det Cov - cont - independent effect priors ####
  prior_det_header_indep <- paste("## Continuous site covariates on detection - Independent effects\n")
  
  priors_sitecovs_det_indep <- independentEffectPriors (effect_names = detCovs$independent,
                                                        type = "detection",
                                                        inits_list = inits_list,
                                                        prior_list = prior_list, 
                                                        stationIndex = stationIndex,
                                                        speciesIndex = speciesIndex,
                                                        speciesMax = speciesMax,
                                                        speciesMax_value = speciesMax_value)
  
  
  
  
  ## det Cov - cont - random effect priors ####
  prior_det_header_ranef <- paste("## Continuous site covariates on detection - with random effects\n")
  
  tmp <- randomEffectPriors(effect_names = detCovs$ranef, type = "detection", covariates = covariates,
                            prior_list = prior_list, inits_list = inits_list, speciesIndex = speciesIndex, stationIndex = stationIndex, 
                            speciesMaxNumber = speciesMax_value, speciesMax = speciesMax)
  priors_sitecovs_det_ranef  <- tmp[[1]]
  alpha_draws_ranef_det_cont <- tmp[[2]]
  rm(tmp)
  
  
  ## det Cov - categ- fixed effect priors ####
  prior_det_categ_header_fixed <- paste("## Categorical site covariates on detection - Fixed effect\n")
  
  priors_sitecovs_det_categ_fixed <- fixedEffectPriorsCateg(effect_names = detCovs_categ$fixed, 
                                                            covariates = covariates,
                                                            type = "detection",
                                                            prior_list = prior_list, 
                                                            stationIndex = stationIndex)
  
  
  
  ## det Cov - categ - random effect priors ####
  prior_det_categ_header_ranef <- paste("## Categorical site covariates on detection - with random effects\n")
  
  tmp <- randomEffectPriorsCateg(effect_names = detCovs_categ$ranef,
                                 covariates = covariates,
                                 type = "detection",
                                 prior_list = prior_list, 
                                 speciesIndex = speciesIndex, 
                                 stationIndex = stationIndex)
  priors_sitecovs_det_categ_ranef <- tmp[[1]]
  alpha_draws_ranef_det_categ <- tmp[[2]]
  rm(tmp)
  
  
  
  
  
  ## occu Cov - cont - fixed effect priors ####
  prior_occu_header_fixed <- paste("## Continuous site covariates on Occupancy - Fixed effects\n")
  
  priors_sitecovs_occu_fixed <- fixedEffectPriors (effect_names = occuCovs$fixed,
                                                   type = "occupancy",
                                                   prior_list = prior_list, 
                                                   stationIndex = stationIndex)
  
  
  ## occu Cov - cont - independent effect priors ####
  prior_occu_header_indep <- paste("## Continuous site covariates on Occupancy - Independent effects\n")
  
  priors_sitecovs_occu_indep <- independentEffectPriors (effect_names = occuCovs$independent,
                                                         type = "occupancy",
                                                         inits_list = inits_list,
                                                         prior_list = prior_list, 
                                                         stationIndex = stationIndex,
                                                         speciesIndex = speciesIndex,
                                                         speciesMax = speciesMax,
                                                         speciesMax_value = speciesMax_value)
  
  
  
  
  
  ## occu Cov - cont  - random effect priors ####
  prior_occu_header_ranef <- paste("## Continuous site covariates on occupancy - with random effects\n")
  tmp <- randomEffectPriors(effect_names = occuCovs$ranef,
                            type = "occupancy",
                            covariates = covariates,
                            prior_list = prior_list, 
                            inits_list = inits_list, 
                            speciesIndex = speciesIndex, 
                            stationIndex = stationIndex, 
                            speciesMaxNumber = speciesMax_value, 
                            speciesMax = speciesMax)
  priors_sitecovs_occu_ranef <- tmp[[1]]
  beta_draws_ranef_occu_cont <- tmp[[2]]
  rm(tmp)
  
  
  
  ## occu Cov - categ - fixed effect priors ####
  prior_occu_categ_header_fixed <- paste("## Categorical site covariates on Occupancy - Fixed effects\n")
  
  priors_sitecovs_occu_categ_fixed <- fixedEffectPriorsCateg(effect_names = occuCovs_categ$fixed,
                                                             covariates = covariates,
                                                             type = "occupancy", 
                                                             prior_list = prior_list,
                                                             stationIndex = stationIndex)
  
  
  
  
  ## occu Cov - categ - random effect priors ####
  prior_occu_categ_header_ranef <- paste("## Categorical site covariates on occupancy - with random effects\n")
  
  tmp <- randomEffectPriorsCateg(effect_names = occuCovs_categ$ranef,
                                 covariates = covariates,
                                 type = "occupancy", 
                                 prior_list = prior_list, 
                                 speciesIndex = speciesIndex, 
                                 stationIndex = stationIndex)
  priors_sitecovs_occu_categ_ranef <- tmp[[1]]
  beta_draws_ranef_occu_categ <- tmp[[2]]
  rm(tmp)
  
  
  
  
  
  
  # random effect of species and station on detection probability  ####
  prior_det_species_station_ranef <- "# Species-station random effect on detection probability\n"
  tmp <-  randomSpeciesStationEffectPriors(doit = speciesSiteRandomEffect$det,
                                           prior_list = prior_list, 
                                           inits_list = inits_list, 
                                           speciesIndex = speciesIndex, 
                                           stationIndex = stationIndex)
  priors_SpeciesStation_det_ranef <- tmp[[1]]
  alpha_draws_SpeciesStation_det_ranef <- tmp[[2]]
  rm(tmp)
  
  
  
  
  before_species_loop <- paste("## Draws of random effects other than species",
                               paste(unlist(alpha_draws_ranef_det_cont)[attr(alpha_draws_ranef_det_cont, "random_effect") == "not_species"]),
                               paste(unlist(beta_draws_ranef_occu_cont)[attr(beta_draws_ranef_occu_cont, "random_effect") == "not_species"]),
                               sep = "\n"
  )
  
  
  # Species loop ####
  loop1 <- paste("\n### MODEL LOOPS \n",
                 "# species loop", 
                 paste0("for (", speciesIndex, " in 1:", speciesMax, "){"),
                 if ("ranef" %in% c(intercepts$det, intercepts$occu)) {
                   paste("##  Draw species-specific random effect parameters from community distributions", 
                         "# intercepts:", sep = "\n")}, 
                 if(intercepts$occu == "ranef") {
                   #paste0("beta0[", speciesIndex, "] ~ dnorm(mean.beta0, tau.beta0)")},
                   paste0("beta0[", speciesIndex, "] ~ dnorm(beta0.mean, beta0.tau)")},
                 if(intercepts$det == "ranef") { 
                   #paste0("alpha0[", speciesIndex, "] ~ dnorm(mean.alpha0, tau.alpha0)")},
                   paste0("alpha0[", speciesIndex, "] ~ dnorm(alpha0.mean, alpha0.tau)")},
                 "",
                 if(augment == "full"){paste0("# Metacommunity membership indicator (data augmentation)\n", "w[", speciesIndex, "] ~ dbern(omega)\n")},
                 paste(unlist(alpha_draws_ranef_det_cont[attr(alpha_draws_ranef_det_cont, "random_effect") == "species"]), collapse = "\n"), 
                 "\n",
                 paste(unlist(alpha_draws_ranef_det_categ), collapse = "\n"),
                 "\n",         
                 paste(unlist(alpha_draws_ranef_obs_cont), collapse = "\n"),
                 "\n",
                 paste(unlist(alpha_draws_ranef_obs_categ), collapse = "\n"),
                 "\n",
                 paste(unlist(beta_draws_ranef_occu_cont[attr(beta_draws_ranef_occu_cont, "random_effect") == "species"]), collapse = "\n"), 
                 "\n", 
                 paste(unlist(beta_draws_ranef_occu_categ), collapse = "\n"),
                 "\n", sep = "\n")
  
  attr(loop1, "params") <- c("beta0", "alpha0", 
                             names(alpha_draws_ranef_det_cont), names(alpha_draws_ranef_det_categ), names(alpha_draws_ranef_obs_cont), names(alpha_draws_ranef_obs_categ), 
                             names(beta_draws_ranef_occu_cont), names(beta_draws_ranef_occu_categ))
  if(augment == "full"){
    attr(loop1, "params") <- c(attr(loop1, "params"), "w")
    attr(loop1, "inits") <- list(w = rep(1, length(data_list$ylist)))
  }
  
  
  
  
  ## construct formula for psi (ecological model - beta parameters)  ####
  
  beta_formula_occu_fixed <- attr(priors_sitecovs_occu_fixed, "formula")
  beta_formula_occu_indep <- attr(priors_sitecovs_occu_indep, "formula")
  beta_formula_occu_ranef <- attr(priors_sitecovs_occu_ranef, "formula")
  beta_formula_occu_categ_fixed <- attr(priors_sitecovs_occu_categ_fixed, "formula")
  beta_formula_occu_categ_ranef <- attr(priors_sitecovs_occu_categ_ranef, "formula")
  
  
  ## station loop ####
  # including formula for psi
  
  if(model == "Occupancy") {
  loop2 <- paste("# station loop", 
                 paste0("for (", stationIndex ," in 1:", stationMax, "){"),
                 paste("\n# Occupancy probability formula"),
                 if (nimble) {
                   paste0("logit(psi[", speciesIndex, ",", stationIndex, "]) <- ", beta0_formula, beta_formula_occu_categ_fixed, beta_formula_occu_categ_ranef, beta_formula_occu_fixed, beta_formula_occu_indep, beta_formula_occu_ranef)
                 },
                 
                 if (!nimble) {
                   paste0("logit.psi[", speciesIndex, ",", stationIndex, "] <- ", beta0_formula, beta_formula_occu_categ_fixed, beta_formula_occu_categ_ranef, beta_formula_occu_fixed, beta_formula_occu_indep, beta_formula_occu_ranef)
                 },
                 if (!nimble) {
                   paste0("psi[", speciesIndex, ",", stationIndex, "] <- exp(logit.psi[", speciesIndex, ",", stationIndex, "]) / (exp(logit.psi[", speciesIndex, ",", stationIndex, "]) + 1)")
                 },
                 
                 ifelse(!nimble, 
                        paste0("z[", speciesIndex, ",", stationIndex, "] ~ dbern(psi[", speciesIndex, ", ", stationIndex, "]", 
                               ifelse(augment == "full", paste0(" * w[", speciesIndex, "])"), ")")), 
                        ""),
                 "\n", sep = "\n"
  )
  }
  
  if(model == "RN") {
    loop2 <- paste("# station loop", 
                   paste0("for (", stationIndex ," in 1:", stationMax, "){"),
                   paste("\n# Abundance formula"),
                   if (nimble) {
                     paste0("log(lambda[", speciesIndex, ",", stationIndex, "]) <- ", beta0_formula, beta_formula_occu_categ_fixed, beta_formula_occu_categ_ranef, beta_formula_occu_fixed, beta_formula_occu_indep, beta_formula_occu_ranef)
                   },
                   
                   if (!nimble) {
                     paste0("log.lambda[", speciesIndex, ",", stationIndex, "] <- ", beta0_formula, beta_formula_occu_categ_fixed, beta_formula_occu_categ_ranef, beta_formula_occu_fixed, beta_formula_occu_indep, beta_formula_occu_ranef)
                   },
                   if (!nimble) {
                     paste0("lambda[", speciesIndex, ",", stationIndex, "] <- exp(log.lambda[", speciesIndex, ",", stationIndex, "])")
                   },
                   paste("\n# convert lambda to psi"),
                   paste0("psi[", speciesIndex, ",", stationIndex, "] <- 1 - dpois(0, lambda[", speciesIndex, ",", stationIndex, "])"),
                   
                   
                   ifelse(!nimble,
                          paste0(paste("\n# Occupancy_status matrix\n"),
                                 # "z[", speciesIndex, ",", stationIndex, "] ~ dbern(psi[", speciesIndex, ",", stationIndex, "]",
                                 "z[", speciesIndex, ",", stationIndex, "] <- ifelse(z_lambda[", speciesIndex, ",", stationIndex, "] >= 1, 1, 0",
                                 ifelse(augment == "full", paste0(" * w[", speciesIndex, "])"), ")")),
                                 
                          ""),
                   ifelse(!nimble, 
                          paste0("\n# Expected abundance matrix\n",
                                 "z_lambda[", speciesIndex, ",", stationIndex, "] ~ dpois(lambda[", speciesIndex, ",", stationIndex, "]", 
                                 ifelse(augment == "full", paste0(" * w[", speciesIndex, "])"), ")")), 
                          ""),
                   "\n", sep = "\n"
    )
  }
  
  
  attr(loop2, "params") <- NULL
  
  
  zin <- t(sapply(data_list$ylist, rowSums, na.rm = TRUE))  # species x Station matrix with n detections
  zin[zin>1] <- 1
  if(model == "RN"){
    if(!nimble) attr(loop2, "inits") <- list(z_lambda = zin)
  } else {
    if(!nimble) attr(loop2, "inits") <- list(z = zin)
  }
  if(nimble)  attr(loop2, "inits") <- NULL
  
  
  ## construct formula for p (detection model - alpha parameters)   ####
  
  alpha_formula_det_fixed <- attr(priors_sitecovs_det_fixed, "formula")
  alpha_formula_det_indep <- attr(priors_sitecovs_det_indep, "formula")
  alpha_formula_det_ranef <- attr(priors_sitecovs_det_ranef, "formula")
  alpha_formula_det_categ_fixed <- attr(priors_sitecovs_det_categ_fixed, "formula")
  alpha_formula_det_categ_ranef <- attr(priors_sitecovs_det_categ_ranef, "formula")
  
  alpha_formula_det_SpeciesStation <- attr(priors_SpeciesStation_det_ranef, "formula")
  
  
  
  if(!is.null(detCovsObservation$fixed)) {
    if(nimble) {
      alpha_formula_obs_fixed <- paste(" + alpha.obs.fixed.cont.", detCovsObservation$fixed, " * ", detCovsObservation$fixed, "[", stationIndex, ", 1:", occasionMax, "]", collapse = "", sep = "")
    } else {
      alpha_formula_obs_fixed <- paste(" + alpha.obs.fixed.cont.", detCovsObservation$fixed, " * ", detCovsObservation$fixed, "[", stationIndex, ", ", occasionIndex, "]", collapse = "", sep = "")
    }
  } else {
    alpha_formula_obs_fixed <- ""
  }
  
  
  if(!is.null(detCovsObservation$ranef)) {
    if(nimble) {
      alpha_formula_obs_ranef <- paste(" + alpha.obs.ranef.cont.", detCovsObservation$ranef, "[", speciesIndex, "] * ", detCovsObservation$ranef, "[", stationIndex, ", 1:", occasionMax, "]", collapse = "", sep = "")
    } else {
      alpha_formula_obs_ranef <- paste(" + alpha.obs.ranef.cont.", detCovsObservation$ranef, "[", speciesIndex, "] * ", detCovsObservation$ranef, "[", stationIndex, ", ", occasionIndex, "]", collapse = "", sep = "")
    }
  } else {
    alpha_formula_obs_ranef <- ""
  }
  
  
  if(!is.null(detCovsObservation_categ$fixed)) {
    if(nimble) {
      alpha_formula_obs_fixed_categ <- paste(" + alpha.obs.fixed.categ.", detCovsObservation_categ$fixed, "[index_tmp_fixed]", collapse = "", sep = "")
    } else {
      alpha_formula_obs_fixed_categ <- paste(" + alpha.obs.fixed.categ.", detCovsObservation_categ$fixed, "[", paste0(detCovsObservation_categ$fixed, "_integer"), "[", stationIndex, ", ", occasionIndex, "]", "]", collapse = "", sep = "")  
    }
  } else {
    alpha_formula_obs_fixed_categ <- ""
  }
  
  if(!is.null(detCovsObservation_categ$ranef)) {
    if(nimble){
      alpha_formula_obs_ranef_categ <- paste(" + alpha.obs.ranef.categ.", detCovsObservation_categ$ranef, "[", speciesIndex, ", ", "index_tmp_ranef", "]", collapse = "", sep = "")
    } else {
      alpha_formula_obs_ranef_categ <- paste(" + alpha.obs.ranef.categ.", detCovsObservation_categ$ranef, "[", speciesIndex, ", ", detCovsObservation_categ$ranef, "_integer", "[", stationIndex, ", ", occasionIndex, "]", "]", collapse = "", sep = "")  
    }
  } else {
    alpha_formula_obs_ranef_categ <- ""
  }
  
  
  if(speciesSiteRandomEffect$det) {
    priors_species_station_det_ranef <- paste("ran", "[", speciesIndex, ", ", stationIndex, "] ~ dnorm(0, alpha.speciesstation.ranef.tau)", collapse = "", sep = "")
    
    alpha_species_station_ranef <- paste("ran", "[", speciesIndex, ", ", stationIndex, "]", collapse = "", sep = "")
  } else {
    alpha_species_station_ranef <- ""
  }
  
  
  ### occasion loop  ####
  
  if(!nimble){
    loop3 <- paste( alpha_draws_SpeciesStation_det_ranef,
                    "# occasion loop", 
                   paste0("for (", occasionIndex, " in 1:", occasionMax, "){"), 
                   
                   paste("# Detection probability formula"),
                   
                   
                   
                   paste0("logit.p[", speciesIndex, ",", stationIndex, ",", occasionIndex, "] <- ", alpha0_formula, alpha_formula_obs_fixed, alpha_formula_obs_ranef, 
                          alpha_formula_obs_fixed_categ, alpha_formula_obs_ranef_categ,
                          alpha_formula_det_categ_fixed, alpha_formula_det_categ_ranef, 
                          alpha_formula_det_fixed, alpha_formula_det_indep, alpha_formula_det_ranef,
                          alpha_formula_det_SpeciesStation
                          ),
                   
                   paste("\n# convert p to real scale"),
                   paste0("p[", speciesIndex, ",", stationIndex, ",", occasionIndex,"] <- exp(logit.p[", speciesIndex, ",", stationIndex, ",", occasionIndex,"]) / (1+exp(logit.p[", speciesIndex, ",", stationIndex, ",", occasionIndex,"]))"),
                   
                   paste("\n# Ensure occasions without effort have p = 0"),
                   
                   if(model == "RN") {
                     paste0("p.e[", speciesIndex, ",", stationIndex, ",", occasionIndex,"] <- p[", speciesIndex, ",", stationIndex, ",", occasionIndex,"] * effort_binary[", stationIndex, ",", occasionIndex,"]")
                   },
                   if(model == "RN") {
                     paste0("p.eff[", speciesIndex, ",", stationIndex, ",", occasionIndex,"] <-  1-(1-p.e[", speciesIndex, ",", stationIndex, ",", occasionIndex, "])^z_lambda[", speciesIndex, ",", stationIndex, "]")
                     },
                   if(model == "Occupancy") {
                     paste0("p.eff[", speciesIndex, ",", stationIndex, ",", occasionIndex,"] <- z[", speciesIndex, ",", stationIndex, "] * p[", speciesIndex, ",", stationIndex, ",", occasionIndex,"] * effort_binary[", stationIndex, ",", occasionIndex,"]")
                     },
                   
                   
                   paste0("y[", speciesIndex, ",", stationIndex, ",", occasionIndex,"] ~ dbern(p.eff[", speciesIndex, ",", stationIndex, ",", occasionIndex,"])"),

                   ## remove GOF calculation temporarily until it is corrected
                   # "\n### generate new data from model under consideration",
                   # paste0("new.y[", speciesIndex, ",", stationIndex, ",", occasionIndex,"] ~ dbern(p.eff[", speciesIndex, ",", stationIndex, ",", occasionIndex,"])"),
                   "}   # close occasion loop", 
                   "\n", sep = "\n")
    attr(loop3, "params") <- NULL
  }
  
  if(nimble){
    
    loop3 <- paste(ifelse(!is.null(detCovsObservation_categ$fixed) , 
                          paste0("index_tmp_fixed <- ", paste0(detCovsObservation_categ$fixed, "_integer", "[", stationIndex, ", 1:", occasionMax, "]")),
                          ""),
                   ifelse(!is.null(detCovsObservation_categ$ranef), 
                          paste0("index_tmp_ranef <- ", paste0(detCovsObservation_categ$ranef, "_integer", "[", stationIndex, ", 1:", occasionMax, "]")),
                          ""),
                   paste("# Detection probability formula"),
                   paste0("logit(p[", speciesIndex, ",", stationIndex, ",", "1:", occasionMax, "]) <- ", alpha0_formula, alpha_formula_obs_fixed, alpha_formula_obs_ranef, 
                          alpha_formula_obs_fixed_categ, alpha_formula_obs_ranef_categ,
                          alpha_formula_det_categ_fixed, alpha_formula_det_categ_ranef, 
                          alpha_formula_det_fixed, alpha_formula_det_indep, alpha_formula_det_ranef,
                          paste0(" * all1row[1:", occasionMax, "]")),              # this is only to ensure that dimensions match if alpha is a single value
                   
                   paste("\n# Ensure occasions without effort have p = 0"),
                   paste0("p.eff[", speciesIndex, ",", stationIndex, ",", "1:", occasionMax, "] <- p[", speciesIndex, ",", stationIndex, ",", "1:", occasionMax, "]", 
                          " * effort_binary[", stationIndex, ", 1:", occasionMax, "]"),
                   
                   
                   # NOTE: which functions from nimbleEcology to use for RN-Model? 
                   # to replace: dOcc_v / rOcc_v
                   
                   
                   # "\n### calculate probability of observed data ",
                   paste0("y[", speciesIndex, ",", stationIndex, ",", "1:", occasionMax, "] ~ dOcc_v(probOcc = psi[", speciesIndex, ",", stationIndex, "], probDetect = p.eff[", speciesIndex, ",", stationIndex, ",", "1:", occasionMax, "], len = ", occasionMax, ")"),
                   
                   ## remove GOF calculation temporarily until it is corrected
                   # "\n### generate new data from model under consideration",
                   # paste0("new.y[", speciesIndex, ",", stationIndex, ",", "1:", occasionMax, "]  <- rOcc_v(n = 1, probOcc = psi[", speciesIndex, ",", stationIndex, "], probDetect = p.eff[", speciesIndex, ",", stationIndex, ",", "1:", occasionMax, "], len = ", occasionMax, ")"), 
                   
                   "\n", sep = "\n")
    attr(loop3, "params") <- NULL
    
    
  }
  
  
  
  ## derived quantities #### 
  
  if(!nimble) {    
    # remove GOF calculation  temporarily until it is corrected
    close_loop2 <-  paste(# "### calculate Freeman-Tukey residuals for real and new data",
                          # paste0("res[", speciesIndex, ",", stationIndex, "] <- (sqrt(sum(y[", speciesIndex, ",", stationIndex, ", 1:", occasionMax, "])) - sqrt(sum(p.eff[", speciesIndex, ",", stationIndex, ", 1:", occasionMax, "])))^2"),
                          # paste0("new.res[", speciesIndex, ",", stationIndex, "] <- (sqrt(sum(new.y[", speciesIndex, ",", stationIndex, ", 1:", occasionMax, "])) - sqrt(sum(p.eff[", speciesIndex, ",", stationIndex, ", 1:", occasionMax, "])))^2"),
                           "}   # close station loop",
                           "\n", sep = "\n")
    attr(close_loop2, "params") <- NULL
    
    
    
    close_loop1a1 <- paste(#"### sum residuals over stations",
                           #paste0("R2[", speciesIndex, "] <- sum(res[", speciesIndex, ", 1:", stationMax, "])"),
                           #paste0("new.R2[", speciesIndex, "] <- sum(new.res[", speciesIndex, ", 1:", stationMax, "])"),
                           
                           #"\n### species-level Bayesian p-value",
                           #paste0("Bpvalue_species[", speciesIndex, "] <- R2[", speciesIndex, "] > new.R2[", speciesIndex, "]"),
                           "# Goodness-of-fit test is now implemented in PPC.community function.",
                           "\n", sep = "\n")
    # attr(close_loop1a1, "params") <- c("R2", "new.R2", "Bpvalue_species")
  } else {
    close_loop2 <- "}   # close station loop\n"
    close_loop1a1 <- "# Goodness-of-fit test is now implemented in PPC.community function.\n"
  }
  
  
  
  if(!nimble){
    close_loop1a2 <- paste("\n### Number of occupied stations for each species",
                           paste0("NStationsOccupied[", speciesIndex, "] <- sum(z[", speciesIndex, ", 1:", stationMax, "])"),
                           "\n### species is part of community?",
                           paste0("speciesInCommunity[", speciesIndex ,"] <- 1 - equals(NStationsOccupied[", speciesIndex ,"],0)"),
                           "\n", sep = "\n")
  } else {
    close_loop1a2 <- "# Total number of occupied and community membership indicator are not returned if nimble = TRUE\n"
  }
  attr(close_loop1a2, "params") <- NULL
  
  
  
  
  
  fractionStationsOccupied <- list()
  
  if(!is.null(occuIntercept_categ$fixed)) {
    
    # get number of stations in each richness category
    stations_richness_categories <- table(covariates[,occuIntercept_categ$fixed])
    names(stations_richness_categories) <- paste0("J", 1:length(stations_richness_categories))
    
    for(s_occu in 1:length(unique(covariates[,occuIntercept_categ$fixed]))){
      
      current_s1 <- names(stations_richness_categories)[s_occu]
      current_s2a <- paste0(stationMax, 1:(s_occu - 1), collapse = " + ")
      current_s2b <- paste0(stationMax, 1:s_occu, collapse = " + ")
      current_s3  <- paste0(stationMax, s_occu)
      
      
      if(s_occu == 1) fractionStationsOccupied[[s_occu]] <- paste(paste0("### Number of stations occupied (in each level of '", richnessCategories, "')"),
                                                   paste0("fractionStationsOccupied[", speciesIndex , ", ", s_occu, "] <- sum(z[", speciesIndex ,", 1:", current_s1, "])/", current_s1), 
                                                   sep = "\n")
      
      if(s_occu > 1 & s_occu != length(unique(covariates[,occuIntercept_categ$fixed]))) fractionStationsOccupied[[s_occu]] <- paste0("fractionStationsOccupied[", speciesIndex , ", ", s_occu, "] <- sum(z[", speciesIndex ,", (", current_s2a, " + 1):(", current_s2b ,")]) / ", current_s3)
      
      if(s_occu == length(unique(covariates[,occuIntercept_categ$fixed]))) fractionStationsOccupied[[s_occu]] <- paste0("fractionStationsOccupied[", speciesIndex , ", ", s_occu, "] <- sum(z[", speciesIndex ,", (", current_s2a, " + 1):", stationMax, "]) / (", stationMax, " - (", current_s2a, "))")      
    }
  } else {
    fractionStationsOccupied[[1]] <- paste("### Fraction of stations occupied",
                            paste0("fractionStationsOccupied[", speciesIndex ,"] <- NStationsOccupied[", speciesIndex ,"] /", stationMax), sep = "\n")
  }
  
  if(!nimble) {
    close_loop1b <- paste(
      paste(unlist(fractionStationsOccupied), collapse = "\n"),
      "\n", sep = "\n")
  } else {
    close_loop1b <- "# Number of stations occupied is not returned when nimble = TRUE\n"
  }
  
  
  if(!is.null(occuIntercept_categ$fixed)) {
    occ <- paste0("occ[", speciesIndex, ", ", 1:length(unique(covariates[,occuIntercept_categ$fixed])), "]<- 1 - equals(fractionStationsOccupied[i,", 1:length(unique(covariates[,occuIntercept_categ$fixed])), "], 0)")
  } else {
    occ <- paste0("occ[", speciesIndex ,"] <- 1 - equals(fractionStationsOccupied[", speciesIndex ,"], 0)")
  }
  
  if(!nimble){
    close_loop1c <- paste(    
      paste0("### Does species occur at all (at any station)", ifelse(!is.null(richnessCategories), paste0("(within each level of '", richnessCategories, "')"), "")),
      paste(occ, collapse = "\n"),
      "}    # close species loop",
      "\n", sep = "\n")
  } else {
    close_loop1c <- paste("}    # close species loop\n\n")
  }
  
  
  # if(!nimble){    # remove GOF calculation temporarily until it it corrected
  #   finish_residuals <- paste("###sum residuals over observed species", 
  #                             paste0("R3 <- sum(R2[1:", speciesMax, "])"),
  #                             paste0("new.R3 <- sum(new.R2[1:", speciesMax, "])"),
  #                             paste0("Bpvalue <- R3 > new.R3"),
  #                             "\n", sep = "\n")
  #   attr(finish_residuals, "params") <- c("R3", "new.R3", "Bpvalue")
  # } else {
    finish_residuals <- ""
  # }
  
  
  if(augment == "full") {
    finish_nspecies <- paste("### total number of species", 
                             ifelse(nimble, 
                                    "# not returned if nimble = TRUE", 
                                    paste0("Nspecies <- sum(speciesInCommunity[1:", speciesMax , "])")),
                             "",
                             "### Total number of unobserved species",
                             paste0("n0 <- sum(w[(nspec+1):(nspec+nz)])"),
                             "",
                             "### Total metacommunity size",
                             paste0("Ntotal <- sum(w[])"),
                             "\n", sep = "\n")
    if(nimble)  attr(finish_nspecies, "params") <- c("n0", "Ntotal")
    if(!nimble) attr(finish_nspecies, "params") <- c("Nspecies", "n0", "Ntotal")
    
    
    
  } else {
    finish_nspecies <- paste("### total number of species", 
                             ifelse(nimble, 
                                    "# not returned if nimble = TRUE", 
                                    paste0("Nspecies <- sum(speciesInCommunity[1:", speciesMax , "])")),
                             "\n", sep = "\n")
    if(!nimble) attr(finish_nspecies, "params") <- "Nspecies"
    if(nimble)  attr(finish_nspecies, "params") <- NULL
  } 
  
  
  if(!is.null(occuIntercept_categ$fixed) & !nimble) {
    finish_nspecies_categ <- paste(paste0("### number of species (in each level of '", richnessCategories, "')"),
                                   paste0("for(s_occu in 1:", length(unique(covariates[,occuIntercept_categ$fixed])), "){"), 
                                   paste0("Nspecies_", richnessCategories, "[s_occu] <- sum(occ[1:", speciesMax , ", s_occu])"),
                                   "}",
                                   "", 
                                   sep = "\n")
    attr(finish_nspecies_categ, "params") <- paste0("Nspecies_", richnessCategories)
  } else {
    finish_nspecies_categ <- paste("", sep = "\n")
  }
  
  # species richness at each camera trap station
  if(nimble)  species_richness <-  paste(paste0("### Species richness at every location"),
                                         "# not returned if nimble = TRUE",
                                         "", sep = "\n")
  
  if(!nimble){ species_richness <- paste(paste0("### Species richness at every location"),
                                        paste0("for (", stationIndex, " in 1:", stationMax, "){"),
                                        paste0("Nspecies_station[", stationIndex, "] <- sum(z[1:", speciesMax, ",", stationIndex, "])"),
                                        "}",
                                        "",
                                        "", sep = "\n")
    attr(species_richness, "params") <- "Nspecies_station"
  }
  
  close_model <- paste("}",
                       "", sep = "\n")
    

  
  
  
  
  # Creating output ####
  
  ## combine model code ####
  model_text <- list(model_start, 
                     priors_intercept_occu,
                     priors_intercept_det,
                     
                     priors_omega,
                     
                     # detection covariates (site)
                     prior_det_header_fixed,
                     unlist(priors_sitecovs_det_fixed),
                     
                     prior_det_header_indep,
                     unlist(priors_sitecovs_det_indep),
                     
                     prior_det_header_ranef,
                     unlist(priors_sitecovs_det_ranef),
                     
                     prior_det_categ_header_fixed, 
                     unlist(priors_sitecovs_det_categ_fixed),
                     
                     prior_det_categ_header_ranef,
                     unlist(priors_sitecovs_det_categ_ranef),
                     
                     # detection covariates (occasion)
                     prior_obs_header_fixed,
                     unlist(priors_obscovs_det_fixed),
                     
                     prior_obs_header_ranef,
                     unlist(priors_obscovs_det_ranef),
                     
                     prior_obs_categ_header_fixed,
                     unlist(priors_obscovs_det_categ_fixed),
                     
                     prior_obs_categ_header_ranef,
                     unlist(priors_obscovs_det_categ_ranef),
                     
                     # occupancy covariates (site)
                     prior_occu_header_fixed,
                     unlist(priors_sitecovs_occu_fixed),
                     
                     prior_occu_header_indep,
                     unlist(priors_sitecovs_occu_indep),
                     
                     prior_occu_header_ranef,
                     unlist(priors_sitecovs_occu_ranef),
                     
                     prior_occu_categ_header_fixed,
                     unlist(priors_sitecovs_occu_categ_fixed),
                     
                     prior_occu_categ_header_ranef,
                     unlist(priors_sitecovs_occu_categ_ranef),
                     
                     prior_det_species_station_ranef,
                     unlist(priors_SpeciesStation_det_ranef),
                     
                     before_species_loop,
                     
                     # model
                     loop1, 
                     loop2,
                     loop3,
                     close_loop2,
                     close_loop1a1,
                     close_loop1a2,
                     close_loop1b,
                     close_loop1c,
                     finish_residuals, 
                     finish_nspecies,
                     finish_nspecies_categ,
                     species_richness,
                     close_model)
  
  out <- list()
  out$model <- unlist(model_text)
  
  
  ## combine parameters to monitor  ####
  out$params <- c(unlist(sapply(model_text, attr, "params")), 
                  attr(priors_sitecovs_det_fixed, "params"),
                  attr(priors_sitecovs_det_indep, "params"),
                  attr(priors_sitecovs_det_ranef, "params"),
                  attr(priors_sitecovs_det_categ_fixed, "params"),
                  attr(priors_sitecovs_det_categ_ranef, "params"),
                  attr(priors_obscovs_det_fixed, "params"),
                  attr(priors_obscovs_det_ranef, "params"),
                  attr(priors_obscovs_det_categ_fixed, "params"),
                  attr(priors_obscovs_det_categ_ranef, "params"),
                  attr(priors_sitecovs_occu_fixed, "params"),
                  attr(priors_sitecovs_occu_indep, "params"),
                  attr(priors_sitecovs_occu_ranef, "params"),
                  attr(priors_sitecovs_occu_categ_fixed, "params"),
                  attr(priors_sitecovs_occu_categ_ranef, "params"),
                  attr(priors_SpeciesStation_det_ranef, "params"))
  
  ## combine inits  ####
  inits_out_tmp <- c(do.call(c, lapply(model_text[sapply(model_text, 
                                                         FUN = function(x) !is.null(attr(x, "inits")))], 
                                       attr, "inits")),
                     attr(priors_sitecovs_det_fixed, "inits"),
                     attr(priors_sitecovs_det_indep, "inits"),
                     attr(priors_sitecovs_det_ranef, "inits"),
                     attr(priors_sitecovs_det_categ_fixed, "inits"),
                     attr(priors_sitecovs_det_categ_ranef, "inits"),
                     attr(priors_obscovs_det_fixed, "inits"),
                     attr(priors_obscovs_det_ranef, "inits"),
                     attr(priors_obscovs_det_categ_fixed, "inits"),
                     attr(priors_obscovs_det_categ_ranef, "inits"),
                     attr(priors_sitecovs_occu_fixed, "inits"),
                     attr(priors_sitecovs_occu_indep, "inits"),
                     attr(priors_sitecovs_occu_ranef, "inits"),
                     attr(priors_sitecovs_occu_categ_fixed, "inits"),
                     attr(priors_sitecovs_occu_categ_ranef, "inits"),
                     attr(priors_SpeciesStation_det_ranef, "inits"))
  
  
  inits_list     <- inits_out_tmp[sapply(inits_out_tmp, class) == "list"]
  inits_not_list <- inits_out_tmp[sapply(inits_out_tmp, class) != "list"]
  
  
  
  # make inits function
  inits_fun <- function(x){ 
    c(lapply(inits_list, FUN = function(x) do.call(x[[1]], x[[2]])),
      inits_not_list)
  }
  
  out$inits_fun <- inits_fun
  
  
  
  # prepare model data  ####
  
  # effort_binary = 1/0 indicator for whether station was active or not (to ensure p = 0 when no effort)
  effort_binary <- as.matrix(!is.na(data_list$obsCovs[[effortCov]])) * 1   # replace NA with 0, other values with 1
  
  
  # create data list for jags.model
  if(!"ylist" %in% names(data_list)) stop("Didn't find ylist in data_list")
  y <- simplify2array(data_list$ylist)
  y <- aperm(y,c(3,1,2))
  
  # assign names to data augmentation species
  if(!is.null(augmentation)) {
    if(any(dimnames(y)[[1]] == "")) {
      which_no_name <- which(dimnames(y)[[1]] == "")
      dimnames(y)[[1]][which_no_name] <- paste("species_DA", 1:length(which_no_name), sep = "_")
    }
  }
  
  data.list <- list(
    y = y,
    effort_binary = effort_binary
  )
  
  
  # add site-occasion covariates
  if(!is.null(data_list$obsCovs)) {
    data.list <- c(data.list, data_list$obsCovs[names(data_list$obsCovs) %in% unlist(detCovsObservation)])
    data.list <- c(data.list, data_list$obsCovs[names(data_list$obsCovs) %in% unlist(detCovsObservation_categ)])
    
    # check if any observation covariate is character
    #if(any(sapply(data_list$obsCovs, typeof) == "character")){
    if(!is.null(unlist(detCovsObservation_categ))) {
      
      # if yes, add the corresponding "..._integer" matrix
      data.list <- c(data.list, data_list$obsCovs[grep("_integer$", names(data_list$obsCovs))])    # this does not include original covariate
    }
    
    if(effortCov %in% unlist(detCovsObservation)) {
      if(!effortCov %in% names(data.list)) {
        data.list[effortCov] <- data_list$obsCovs[effortCov]
      }
      data.list[[effortCov]] [effort_binary == 0] <- 0   # replace effort = NA with 0 (doesn't affect calculation of p because of effort_binary)
    }
  }
  
  # add data dimensions
  list_design <- list(M = speciesMax_value,
                      J = stationMax_value,
                      maxocc = occasionMax_value)
  
  names(list_design) <- c(speciesMax, 
                          stationMax, 
                          occasionMax)
  
  data.list <- c(data.list, list_design)
  
  # add site covariates
  if(length(covariates_categ) >= 1)   data.list <- c(data.list, as.list(covariates[covariates_categ]))
  if(length(covariates_numeric) >= 1) data.list <- c(data.list, as.list(covariates[covariates_numeric]))
  
  if(!is.null(occuIntercept_categ$fixed)) {
    tmp <- table(covariates[,occuIntercept_categ$fixed])
    names(tmp) <- paste0("J", 1:length(tmp))
    data.list <- c(data.list, tmp)
  }
  
  # add data augmentation information
  if(augment == "full"){
    data.list <- c(data.list, 
                   list(nz = nz,          # number of unseen species added to the detection history list
                        nspec = nspec))   # number of seen species
  }
  
  
  # for nimble, convert site covariate factors to numeric
  # the values will then be used for indexing the factor levels, and will not be interpreted as continuous values
  if(nimble){
    data.list <- lapply(data.list, FUN = function(x) {
      if(is.factor(x)) {
        as.numeric(x)
      } else {
        x
      }
    })
    data.list$all1row <- rep(1, times = occasionMax_value)
  }
  
  
  out$model_data <- data.list
  
  # save model text file
  if(!is.null(modelFile)) {
    sink(modelFile)
    cat(out$model, fill = FALSE, sep = "")
    sink()
    if(file.exists(modelFile)) {
      message(paste("Wrote model to", modelFile))
    } else {
      message(paste("Model could not be written to", modelFile))
    }
  }
  
  # create output commOccu object
  out2 <- new("commOccu",   
              modelText = out$model,
              params    = out$params,
              inits_fun = out$inits_fun,
              data      = out$model_data,
              input     = data_list,
              nimble    = nimble,
              modelFile = ifelse(!is.null(modelFile), modelFile, "undefined"),
              covariate_info = covariate_info,
              model     = model
              )
  return(out2)
  
}



# Define commOccu class for output ####
#' commOccu objects
#'
#' @slot modelText JAGS model code as a character vector (made up of code chunks, use cat() to print)
#' @slot params Parameters to monitor in the model runs
#' @slot inits_fun Function to create start values for the MCMC chains. It being a function ensures different values in each chain
#' @slot data List with data needed to run the model (detection & effort matrices, site covariates, number of species / stations / occasions)
#' @slot input Input data_list (unchanged)
#' @slot nimble logical indicator for whether it is a Nimble model
#' @slot modelFile Path of the text file containing the model code
#' @slot covariate_info Data frame containing information about covariates. Only used internally in plot_* and predict methods
#' @slot model character indicating whether it is a standard "Occupancy" model or Royle-Nichols ("RN") occupancy model
#'
#' @note 
#' The \code{data} slot is a list of model input data. While the exact content depends on function input, it can be summarized as:
#' 
#' \tabular{ll}{
#' \code{y} \tab array of detection histories. Dimensions are: y[species, station, occasion] \cr
#' \code{effort_binary} \tab matrix of binary (1/0) survey effort. Only used to ensure p = 0 when effort = 0. Dimensions are: effort_binary[station, occasion] \cr
#' \code{site-occasion covariates} \tab The required content of data_list$obsCovs as named matrices with dimensions [station, occasion] \cr
#' \code{site covariates} \tab The required columns of data_list$siteCovs as named vectors (length = number of stations) \cr
#' \code{M} \tab Number of species \cr
#' \code{J} \tab Number of stations \cr
#' \code{maxocc} \tab Number of occasions \cr
#' }
#' 
#' For categorical site-occasion covariates, an addition matrix containing an integer representation of the character matrix with suffix "_integer" is stored in the data slot.
#' 
#' @return \code{commOccu} object
#' 
#' @export
#' 
setClass("commOccu", 
         slots = c(modelText      = "character",
                   params         = "character",
                   inits_fun      = "function",
                   data           = "list",
                   input          = "list",
                   nimble         = "logical",
                   modelFile      = "character", 
                   covariate_info = "data.frame",
                   model          = "character"
                   )
)




# fit method ####


fit.commOccu <- function(object,
                         n.iter = 100,
                         thin = 1,
                         n.burnin = n.iter/2,
                         n.adapt = 0,
                         chains = 3,
                         inits = NULL,
                         compile = TRUE,
                         WAIC	= FALSE,
                         quiet = FALSE,
                         ...) {
  
  if(thin == 0) stop("thin can't be 0")
  if(n.adapt != 0 & isTRUE(object@nimble)) message(paste("nimble models don't use n.adapt. It will be ignored."))
  
  if(isFALSE(object@nimble)){
    
    if(!file.exists(object@modelFile)) stop(paste("modelFile not found under", object@modelFile))
    
    if(isTRUE(WAIC)) warning("WAIC is only returned in Nimble models", immediate. = TRUE)
    
    
    
      mod <- rjags::jags.model(file = object@modelFile, 
                               data = object@data, 
                               inits = object@inits_fun(),
                               n.chain=chains, 
                               n.adapt=n.adapt,
                               quiet = quiet)
      
      out <- rjags::coda.samples(model = mod,
                                variable.names = object@params, 
                                n.iter	= n.iter, 
                                thin = thin)
      
    
    # posterior summaries
      out_mcmclist <- coda::mcmc.list(out)
    
    if(n.burnin >= 1) {
      out2 <- window(out_mcmclist, 
                     start=n.burnin+1, 
                     end = n.iter)
      return(out2)
    } else {
      return(out_mcmclist)
    }
    
  }
  
  
  if(isTRUE(object@nimble)) {
    
    mod <- nimble::readBUGSmodel(object@modelFile,
                                 data = object@data,
                                 inits = object@inits_fun())
    
    
    myMCMC <- nimble::buildMCMC(mod, 
                                monitors = object@params,
                                thin = thin,
                                enableWAIC = WAIC,
                                print = isFALSE(quiet))
    
    if(compile) {
      
      mod_comp <- nimble::compileNimble(mod)
      
      compMCMC <- nimble::compileNimble(myMCMC, project = mod)
      
      out_mcmclist <- nimble::runMCMC(compMCMC,
                                      niter = n.iter, 
                                      nburnin = n.burnin,
                                      thin = thin,
                                      nchains = chains, 
                                      samplesAsCodaMCMC = TRUE,
                                      WAIC = WAIC,
                                      progressBar = isFALSE(quiet),
                                      ...) 
    }
    
    if(!compile) {
      out_mcmclist <- nimble::runMCMC(myMCMC,
                                      niter = n.iter, 
                                      nburnin = n.burnin,
                                      thin = thin,
                                      nchains = chains, 
                                      samplesAsCodaMCMC = TRUE,
                                      WAIC = WAIC,
                                      progressBar = isFALSE(quiet),
                                      ...) 
    }
    
    return(out_mcmclist)
    
    
  }
}

setGeneric("fit", function(object, ...){})


#' Fit a community (multi-species) occupancy model
#'
#' Convenience function for fitting community occupancy models (defined in a commOccu object) in JAGS or Nimble.
#'
#' @param object \code{commOccu} object
#' @param n.iter number of iterations to monitor
#' @param thin thinning interval for monitors
#' @param n.burnin   burnin length. Defaults to half of n.iter.
#' @param n.adapt Length of adaptive phase
#' @param chains number of MCMC chains to run
#' @param inits named list. Initial values to use. If NULL (default), the values from the inits function in \code{object} are used.
#' @param compile logical. If Nimble model, compile model with \code{\link[nimble]{compileNimble}} before running model?
#' @param WAIC logical. Return WAIC (only Nimble models)
#' @param quiet if TRUE messages and progress bar will be suppressed
#' @param ...     additional arguments to pass to \code{\link[nimble]{runMCMC}} (only relevant for Nimble)
#'
#' @details 
#' Models will be fit either in JAGS or Nimble, depending on the decision made in the \code{nimble} argument in \code{\link{communityModel}}.
#' 
#' For Nimble, compilation is strongly recommended for long model runs. Uncompiled models can run extremely slow. Compilation itself can take a while also, and requires that Rtools is available on the system.
#' 
#' This is a convenience function only which hides some of the configuration options. If you require more control over model fitting, you can run all steps individually. See vignette 5 for details. 
#'  
#' @return A coda::mcmc.list
#'
#' @export
#' 
setMethod("fit", signature(object = "commOccu"), 
          fit.commOccu)



# summary method ####

summary.commOccu <- function(object, ...) {
  # Model type determination
  if(object@model == "RN")        model_type <- "Royle-Nichols community occupancy model"
  if(object@model == "Occupancy") model_type <- "community occupancy model"
  
  # Check for data augmentation by looking for unnamed elements in the input data
  data_aug <- "none"
  
  if(length(object@input$ylist) > 0) {
    species_names <- names(object@input$ylist)
    
    # Check if any species are unnamed (empty string) or if names is NULL
    has_unnamed <- is.null(species_names) || any(species_names == "")
    
    # Count augmented species (those without names)
    augmented_count <- sum(species_names == "") + sum(is.null(species_names))
    
    if(has_unnamed || augmented_count > 0) {
      total_species <- length(object@input$ylist)
      data_aug <- paste0("yes (", total_species, " potential species)")
    }
  }
  
  # MODEL SPECIFICATION section
  cat("=== Community Occupancy Model Summary ===\n\n")
  cat("MODEL SPECIFICATION\n")
  cat(sprintf("  Model type:        %s\n", model_type))
  cat(sprintf("  Platform:          %s\n", ifelse(object@nimble, "Nimble", "JAGS")))
  cat(sprintf("  Data augmentation: %s\n\n", data_aug))
  
  # DIMENSIONS section
  dims <- dim(object@data$y)
  n_species <- dims[1]
  n_stations <- dims[2]
  n_occasions <- dims[3]
  n_effort <- sum(object@data$effort_binary, na.rm = TRUE)
  
  cat("DIMENSIONS\n")
  cat(sprintf("  Species:   %d\n", n_species))
  cat(sprintf("  Stations:  %d\n", n_stations))
  cat(sprintf("  Occasions: %d\n", n_occasions))
  cat(sprintf("  Effort:    %d occasions with effort\n\n", n_effort))
  
  # DETECTIONS section
  detections_by_species <- apply(object@data$y, 1, sum, na.rm = TRUE)
  total_detections <- sum(detections_by_species)
  
  cat("DETECTIONS\n")
  cat(sprintf("  By species: %d - %d (min - max)\n", min(detections_by_species), max(detections_by_species)))
  cat(sprintf("  Total:      %d\n\n", total_detections))
  
  # COVARIATE STRUCTURE section
  cat("COVARIATE STRUCTURE\n")
  
  # Get covariate info
  cov_info <- object@covariate_info
  
  # Define a function to print a section of the covariate table
  print_covariate_table <- function(section_name, section_data) {
    if (nrow(section_data) > 0) {
      cat(sprintf("  %s:\n", section_name))
      
      # Prepare data for display
      covariate_names <- character(nrow(section_data))
      levels <- character(nrow(section_data))
      effect_types <- character(nrow(section_data))
      
      for (i in 1:nrow(section_data)) {
        # Check for independent effects first, then random effects
        if (section_data$independent[i]) {
          effect_types[i] <- "Independent"
        } else if (section_data$ranef[i]) {
          effect_types[i] <- "Random"
        } else {
          effect_types[i] <- "Fixed"
        }
        
        levels[i] <- ifelse(section_data$covariate_type[i] == "obsCovs", "Observation", "Site")
        
        # For intercepts, use [Intercept] as name and always Site level
        if (section_data$param[i] == "intercept") {
          covariate_names[i] <- "[Intercept]"
          levels[i] <- "Site"
        } else {
          covariate_names[i] <- as.character(section_data$covariate[i])
          
          # Add ^2 symbol for quadratic effects
          if (section_data$is_quadratic[i]) {
            covariate_names[i] <- paste0(covariate_names[i], "^2")
          }
        }
      }
      
      # Calculate column widths (max length + padding)
      cov_width <- max(nchar("Covariate"), max(nchar(covariate_names)) + 2)
      level_width <- max(nchar("Level"), max(nchar(levels)) + 2)
      effect_width <- max(nchar("Effect Type"), max(nchar(effect_types)) + 2)
      
      # Create format string for header and data rows
      header_fmt <- paste0("    %-", cov_width, "s %-", level_width, "s %-", effect_width, "s\n")
      
      # Print header
      cat(sprintf(header_fmt, "Covariate", "Level", "Effect Type"))
      
      # Print separator
      separator <- paste(rep("-", cov_width + level_width + effect_width + 2), collapse="")
      cat(paste0("    ", separator, "\n"))
      
      # Print data rows
      for (i in 1:length(covariate_names)) {
        cat(sprintf(header_fmt, covariate_names[i], levels[i], effect_types[i]))
      }
      cat("\n")
    }
  }
  
  # Detection covariates
  detection_covs <- cov_info[cov_info$submodel == "det", ]
  print_covariate_table("Detection Covariates", detection_covs)
  
  # Occupancy covariates 
  occupancy_covs <- cov_info[cov_info$submodel == "state", ]
  print_covariate_table("Occupancy Covariates", occupancy_covs)
  
  # Check for species-site random effects (look for alpha.speciesstation in params)
  has_species_site_ranef_det <- any(grepl("alpha.speciesstation", object@params, fixed = TRUE))
  
  if (has_species_site_ranef_det) {
    cat("SPECIES-SITE RANDOM EFFECT\n")
    cat("  Detection: TRUE\n")
    cat("  Occupancy: FALSE\n")  # Currently only implemented for detection
    cat("\n")
  }
}



#' Summarize community occupancy model
#' 
#' Gives an overview of the number of species, stations and occasions in a \code{commOccu} object. Also returns covariates. 
#' 
#' The summary method is very basic and still work in progress. 
#'
#' @param object  \code{commOccu} object
#' @param ...  currently ignored
#'
#' @return Model summary printed to console
#' @export
#'
setMethod("summary", signature(object = "commOccu"), 
          summary.commOccu)



# other helper functions

# make table with information about covariates
get_cov_info <- function(cov,
                         keyword_nested = keyword_nested,
                         keyword_quadratic = keyword_quadratic,
                         data_list = data_list,
                         type,
                         submodel) {
  
  
  if(type == "obs")  item <- "obsCovs"
  if(type == "site") item <- "siteCovs"
  
  tmp <- unlist(cov)
  
  names(tmp)[grep("fixed", names(tmp))] <- "fixed"
  names(tmp)[grep("ranef", names(tmp))] <- "ranef"
  
  if(submodel == "occu"){
    names(tmp)[grep("independent", names(tmp))] <- "independent"
  }
  
  
  
  tmp_cov1  <- sapply(strsplit(tmp, split = "|", fixed = T), FUN = function(x) x[1])
  
  if(type == "obs")  cov_type  <- sapply(data_list[[item]][tmp_cov1], typeof)
  if(type == "site") cov_type  <- sapply(data_list[[item]][,tmp_cov1, drop = FALSE], class)
  
  cov_type2 <- ifelse(cov_type %in% c("numeric", "double", "integer"), "cont", 
                      ifelse(cov_type %in% c("character", "factor"), "categ", NA))
  

  tmp2 <- data.frame(param = rep("param", times = length(tmp)),
                     submodel = rep(submodel, times = length(tmp)),
                     covariate_type = item,
                     covariate = tmp,
                     data_type = cov_type2,
                     is_quadratic = endsWith(tmp, keyword_quadratic),
                     has_quadratic = paste0(tmp, keyword_quadratic) %in% tmp,
                     ranef = ifelse(names(tmp) == "ranef", T, F), 
                     ranef_nested = grepl(keyword_nested, tmp, fixed = TRUE),
                     ranef_cov = sapply(strsplit(tmp, split = "|", fixed = TRUE), FUN = function(x) ifelse(length(x) == 2, x[2], NA)),
                     independent = ifelse(names(tmp) == "independent", T, F)
  )
  
  

  effect_type <- ifelse(tmp2$ranef, "ranef", ifelse(tmp2$independent, "indep", "fixed"))
  
  tmp2$coef <- paste0(ifelse(tmp2$submodel == "det", "alpha", "beta"), ".",
                 ifelse(tmp2$covariate_type == "obsCovs", "obs.", ""), 
                 effect_type,     #ifelse(tmp2$ranef, "ranef", "fixed"), 
                 ".",
                 tmp2$data_type, ".",
                 gsub("[|+]", "_", tmp)
                 )
  
  rownames(tmp2) <- NULL
  return(tmp2)
}



# helper functions to create priors (both detection and occupancy)  ####


# # paste index letters (not used currently)
# indices <- function(...){
# #   inargs <- list(...)
# #   species  <- ifelse("i" %in% inargs, "i", NA)
# #   station  <- ifelse("j" %in% inargs, "j", NA)
# #   occasion <- ifelse("k" %in% inargs, "k", NA)
# #  #
# # 
# #   paste0("[",
# #          paste(na.omit(c(species, station, occasion)), collapse = ","),
# #         # "]")
# paste0("[",
#        paste(unlist(strsplit(..., split = "")), collapse = ","),
#        "]")
# 
# }
## e.g. indices("ijk") gives "[i,j,k]"



# change number of random numbers created (from default = 1)
modifyInits <- function(inits, n) {
  if(!is.na(n)) {
    inits[[2]][[1]] <- as.numeric(n)
  } 
  return(inits)
}


interceptPriors <- function(effect, 
                            type,
                            prior_list,
                            inits_list, 
                            speciesIndex,
                            speciesMax,
                            speciesMax_value){
  
  
  if(type == "detection") param <- "alpha"
  if(type == "occupancy") param <- "beta"
  
  
  if(effect == "ranef"){
    
    # comenclature 2: alpha0.mean
    param_names <- list(mean  = paste0(param, "0", ".mean"), 
                        tau   = paste0(param, "0", ".tau"),
                        sigma = paste0(param, "0", ".sigma"))
    
    priors_intercept <- paste(
      paste("##", type, "intercept estimate of community (community mean)"),
      
      paste0(param_names$mean, " ~ ", prior_list$dnorm),
      paste0(param_names$tau,  " ~ ", prior_list$dgamma),
      paste0(param_names$sigma, " <- sqrt(1 / ", param_names$tau, ")"),
      
      "\n", sep = "\n")
    
    attr(priors_intercept, "params") <- c(param_names$mean, 
                                          param_names$sigma)
    
    inits_tmp <- list(mean = inits_list$inits_runif_mean_low,
                      tau  = inits_list$inits_runif_tau)
    
    names(inits_tmp) <- c(param_names$mean, 
                          param_names$tau)
    
    attr(priors_intercept, "inits") <- inits_tmp
    
    code_formula <- paste0(param, "0[", speciesIndex, "]")
  } 
  
  if(effect == "fixed") {
    priors_intercept <- paste(
      paste("##", type, "intercept estimate of community (constant across species)"),
      paste0(param, "0 ~ ", prior_list$dnorm),
      "\n", sep = "\n")
    
    inits_tmp <- list(name = inits_list$inits_runif_mean_0)
    names(inits_tmp) <- paste0(param, "0")
    attr(priors_intercept, "inits") <- inits_tmp
    code_formula <- paste0(param, "0")
  }
  
  
  if(effect == "independent") {
    priors_intercept <- paste(
      paste("##", type, "intercept estimate of community (independent between species)"),
      paste0("for(", speciesIndex, " in 1:", speciesMax, ") {"),
      paste0(param, "0[", speciesIndex, "] ~ ", prior_list$dnorm),
      "}",
      "\n", sep = "\n")
    

    # change number of random init values to create (not 1, but n_species)
    inits_tmp <- list(modifyInits(inits_list$inits_runif_mean_0, 
                                  speciesMax_value))
    
    names(inits_tmp) <- paste0(param, "0")
    
    attr(priors_intercept, "inits") <- inits_tmp
    
    code_formula <- paste0(param, "0[", speciesIndex, "]")
    
  }
  
  return(list(prior    = priors_intercept, 
              formula  = code_formula))
}




fixedEffectPriors <- function(effect_names, 
                              type,
                              prior_list, 
                              stationIndex){
  
  if(type == "detection") param <- "alpha"
  if(type == "occupancy") param <- "beta"
  
  if(!is.null(effect_names)){
    priors_list <- list()
    
    for(effect_names_index in 1:length(effect_names)){
      priors_list[[effect_names_index]] <- paste(
        paste("# Covariate:", effect_names[effect_names_index]),
        paste0(param, ".fixed.cont.", effect_names[effect_names_index], " ~ ", prior_list$dnorm), 
        "\n", 
        sep = "\n")
    }
    attr(priors_list, "params") <- paste0(param, ".fixed.cont.", effect_names)
    
    attr(priors_list, "formula") <- paste0(" + ", param, ".fixed.cont.", effect_names, " * ", effect_names, "[", stationIndex, "]", collapse = "", sep = "")
    
  } else {
    priors_list <- list("# < empty > \n\n")
    attr(priors_list, "params") <- NULL
    attr(priors_list, "formula") <- ""
  }
  
  return(priors_list)
}




independentEffectPriors <- function(effect_names, 
                                    type,
                                    inits_list,
                                    prior_list, 
                                    stationIndex,
                                    speciesIndex,
                                    speciesMax,
                                    speciesMax_value){
  
  if(type == "detection") param <- "alpha"
  if(type == "occupancy") param <- "beta"
  
  if(!is.null(effect_names)){
    priors_list <- list()
    
    for(effect_names_index in 1:length(effect_names)){
      priors_list[[effect_names_index]] <- paste(
        paste("# Covariate:", effect_names[effect_names_index]),
        paste0("for(", speciesIndex, " in 1:", speciesMax, ") {"),
        paste0(param, ".indep.cont.", effect_names[effect_names_index], "[", speciesIndex, "] ~ ", prior_list$dnorm), 
        "}",
        "\n", 
        sep = "\n")
    }
    attr(priors_list, "params") <- paste0(param, ".indep.cont.", effect_names)
    
    attr(priors_list, "formula") <- paste0(" + ", param, ".indep.cont.", effect_names,"[", speciesIndex, "]",  " * ", effect_names, "[", stationIndex, "]", collapse = "", sep = "")
    
    
    # make list of parameters
    inits_tmp <- vector(mode = "list", length = length(effect_names))
    names(inits_tmp) <- c(paste0(param, ".indep.cont.", effect_names))
    
    
    for(i in 1:length(inits_tmp)){
      inits_tmp[[i]] <- modifyInits(inits_list$inits_runif_mean_0, 
                                    speciesMax_value)
    }
    
    attr(priors_list, "inits") <- inits_tmp
    
    
  } else {
    priors_list <- list("# < empty > \n\n")
    attr(priors_list, "params") <- NULL
    attr(priors_list, "formula") <- ""
  }
  
  return(priors_list)
}


randomEffectPriors <- function(effect_names, 
                               type,
                               covariates,
                               prior_list,
                               inits_list,
                               speciesIndex, 
                               speciesMaxNumber,
                               speciesMax,
                               stationIndex, 
                               keyword_nested = "+Species"){
  
  if(type == "detection") param <- "alpha"
  if(type == "occupancy") param <- "beta"
  
  effect_names2 <- gsub("|", "_", effect_names, fixed = TRUE)
  effect_names2 <- gsub("+", "_", effect_names2, fixed = TRUE)
  
  if(!is.null(effect_names)){
    priors_list <- list()
    species_draws <- list()
    ranef_index <- vector(mode = "character")
    current_cov2 <- vector(mode = "character")
    ranef_double_all <- vector()
    
    for(effect_names_index in 1:length(effect_names)){
      current_cov <- effect_names[effect_names_index]
      if(grepl("|", current_cov, fixed = TRUE)) {
        
        # this part needs a cleanup
        
        tmp <- unlist(strsplit(current_cov, split = "|", fixed = TRUE))
        current_cov <- gsub("|", "_", current_cov, fixed = TRUE)
        current_cov2[effect_names_index] <- tmp[1]
        if(length(tmp) == 2 & endsWith(tmp[2], keyword_nested)) {
          ranef_index[effect_names_index] <- paste0(speciesIndex, ", ", gsub(keyword_nested, "", tmp[2], fixed = TRUE), "[", stationIndex, "]")  
          ranef_double <- TRUE
          ranef_double_all[effect_names_index] <- TRUE
          attr(ranef_double_all, "nlevels")[effect_names_index] <- speciesMaxNumber
          current_cov <- gsub("+", "_", current_cov, fixed = TRUE)
        } else {
          ranef_index[effect_names_index] <- paste0(tmp[2], "[", stationIndex, "]")
          ranef_double <- FALSE
          ranef_double_all[effect_names_index] <- FALSE
        }
        effect_name_print <- effect_names[effect_names_index]
      } else {
        ranef_index[effect_names_index] <- speciesIndex
        current_cov2[effect_names_index] <- current_cov
        effect_name_print <- paste(current_cov, "Species", sep = "|")
        ranef_double <- FALSE
        ranef_double_all[effect_names_index] <- FALSE
      }
      
      mean.tmp  <- paste0(param, ".ranef.cont.", current_cov, ".mean", ifelse(ranef_double, paste0("[", speciesIndex, "]"), ""))
      tau.tmp   <- paste0(param, ".ranef.cont.", current_cov, ".tau", ifelse(ranef_double, paste0("[", speciesIndex, "]"), ""))
      sigma.tmp <- paste0(param, ".ranef.cont.", current_cov, ".sigma", ifelse(ranef_double, paste0("[", speciesIndex, "]"), ""))
      
      priors_list[[effect_names_index]] <- paste(
        paste("# Covariate:", effect_name_print),
        if(ranef_double) paste0("for(", speciesIndex, " in 1:", speciesMax, "){"),
        paste0(mean.tmp, " ~ ", prior_list$dnorm),
        paste0(tau.tmp, " ~ ", prior_list$dgamma),
        paste0(sigma.tmp, " <- sqrt(1 / ", tau.tmp, ")"),
        if(ranef_double) "}",
        "\n", 
        sep = "\n")
      
      # if only species random effect
      if(ranef_index[effect_names_index] == speciesIndex) {
        species_draws[[effect_names_index]] <- paste(paste("# continuous", type, "covariate with random effects:", effect_name_print), 
                                                     paste0(param, ".ranef.cont.", current_cov, "[", ranef_index[effect_names_index], "] ~ dnorm(", mean.tmp, ", ",  tau.tmp, ")"),
                                                     sep = "\n")
      } else {   
        
        if(ranef_double) {   # two random effects (species + another)
          
          index_letter <- paste0("index_", param, "_ranef_", current_cov)
          species_draws[[effect_names_index]] <- paste(paste("# continuous", type, "covariate with random effects:", effect_name_print), 
                                                       paste0("for(", index_letter, " in 1:", length(unique(covariates[, gsub(keyword_nested, "", tmp[2], fixed = TRUE)])), ") {"),
                                                       paste0(param, ".ranef.cont.", current_cov, "[", speciesIndex, ", ", index_letter, "] ~ dnorm(", mean.tmp, ", ", tau.tmp, ")"),
                                                       "}",
                                                       sep = "\n")
          
          
        } else {    # only other random effect (not species)
          index_letter <- paste0("index_", param, "_ranef_", current_cov)
          species_draws[[effect_names_index]] <- paste(paste("# continuous", type, "covariate with random effects:", effect_name_print), 
                                                       paste0("for(", index_letter, " in 1:", length(unique(covariates[, tmp[2]])), ") {"),
                                                       paste0(param, ".ranef.cont.", current_cov, "[", index_letter, "] ~ dnorm(", mean.tmp, ", ", tau.tmp, ")"),
                                                       "}",
                                                       sep = "\n")
        }
      }
    }
    
    attr(priors_list, "params") <- c(paste0(param, ".ranef.cont.", effect_names2, ".mean"),
                                     paste0(param, ".ranef.cont.", effect_names2, ".sigma"))
    
    
    # make list of parameters
    inits_tmp <- vector(mode = "list", length = length(effect_names2) * 2)
    names(inits_tmp) <- c(paste0(param, ".ranef.cont.", effect_names2, ".mean"),
                          paste0(param, ".ranef.cont.", effect_names2, ".tau"))
    

    
    for(i in 1:length(inits_tmp)){
      if(endsWith(names(inits_tmp)[i], "mean")) if(ranef_double_all[i]){
        inits_tmp[[i]] <- lapply(attr(ranef_double_all, "nlevels"), modifyInits, inits = inits_list$inits_runif_mean_0)[[i]]
      } else {
        inits_tmp[[i]] <- inits_list$inits_runif_mean_0
      }
      
      if(endsWith(names(inits_tmp)[i], "tau"))  if(ranef_double_all[i - length(effect_names)]) {
        inits_tmp[[i]] <- lapply(attr(ranef_double_all, "nlevels"), modifyInits, inits = inits_list$inits_runif_tau)[[i - length(inits_tmp) / 2]]
      } else {
        inits_tmp[[i]] <- inits_list$inits_runif_tau
      }
    }
    attr(priors_list, "inits") <- inits_tmp
    
    attr(priors_list, "formula") <- paste(" + ", param, ".ranef.cont.", effect_names2, "[", ranef_index, "] * ", current_cov2, "[", stationIndex, "]", collapse = "", sep = "")
    
    names(species_draws) <- paste0(param, ".ranef.cont.", effect_names2)
    attr(species_draws, "random_effect") <- ifelse(ranef_index == speciesIndex | (startsWith(ranef_index, speciesIndex) & endsWith(ranef_index, paste0("[", stationIndex, "]"))),
                                                   "species", "not_species")
    
    
    ################# CHECK LINE ABOVE !!! ########
    
  } else {
    priors_list <- list("# < empty > \n\n")
    attr(priors_list, "params") <- NULL
    attr(priors_list, "formula") <- ""
    species_draws <- paste("# continuous", type, "covariates: no random effect of species")
  }
  
  
  return(list(priors_list, 
              species_draws)) 
}



randomSpeciesStationEffectPriors <- function(doit,
                                             prior_list, 
                                             inits_list,
                                             speciesIndex,
                                             stationIndex) {
  
  if(doit) {
  param <- "alpha"
    
  # mean.tmp  <- 0
  tau.tmp   <- "alpha.speciesstation.ranef.tau" #paste0(param, ".ranef.cont.", current_cov, ".tau", ifelse(ranef_double, paste0("[", speciesIndex, "]"), ""))
  sigma.tmp <- "alpha.speciesstation.ranef.sigma" #paste0(param, ".ranef.cont.", current_cov, ".sigma", ifelse(ranef_double, paste0("[", speciesIndex, "]"), ""))
  
  param_name_site_species <- "alpha.speciesstation.ranef"
  
  priors_list <- list()
  
  priors_list[[1]] <- paste(
    #paste("# Random Species-station effect"),
    # if(ranef_double) paste0("for(", speciesIndex, " in 1:", speciesMax, "){"),
    
    
    
    # paste0(mean.tmp, " <- ", 0#prior_list$dnorm),
    paste0(tau.tmp, " ~ ", prior_list$dgamma),
    paste0(sigma.tmp, " <- sqrt(1 / ", tau.tmp, ")"),
    # if(ranef_double) "}",
    "\n", 
    sep = "\n")
  

  attr(priors_list, "params") <- c(sigma.tmp, param_name_site_species)
  
  
  inits_tmp <- vector(mode = "list", length = 1)
  names(inits_tmp) <- "alpha.speciesstation.ranef.tau"
  inits_tmp[[1]] <- inits_list$inits_runif_tau
    
  attr(priors_list, "inits") <- inits_tmp
  
  attr(priors_list, "formula") <- #paste(" + ",  "[", ranef_index, "] * ", current_cov2, "[", stationIndex, "]", collapse = "", sep = "")
    paste(" + ",  param_name_site_species, "[", speciesIndex, ", ", stationIndex, "]", collapse = "", sep = "")
  
  
  species_draws <- paste("# Random effect of species and station on detection probability:" ,
                         paste(param_name_site_species, "[", speciesIndex, ", ", stationIndex, "] ~ dnorm(0, alpha.speciesstation.ranef.tau)", collapse = "", sep = ""),
                         "\n", 
                         sep = "\n")
# beta.ranef.cont.habitat[i] ~ dnorm(beta.ranef.cont.habitat.mean, beta.ranef.cont.habitat.tau)")
  
  } 
  
  
  if(!doit) {
    priors_list <- list("# < empty > \n\n")
    attr(priors_list, "params") <- NULL
    attr(priors_list, "formula") <- ""
    species_draws <- paste("# No random effect of species and station on detection probability")
  }
  
  return(list(priors_list,
              species_draws))
  }




fixedEffectPriorsCateg <- function(effect_names, 
                                   covariates, 
                                   type,
                                   prior_list, 
                                   stationIndex){
  
  if(type == "detection"){
    param <- "alpha"
    type_short <- "det"
  }
  if(type == "occupancy"){
    param <- "beta"
    type_short <- "occu"
  }
  
  
  if(!is.null(effect_names)){
    priors_list <- list()
    attr(priors_list, "n_levels") <- rep(NA, times = length(effect_names))
    
    for(effect_names_index in 1:length(effect_names)){
      
      current_cov <- effect_names[effect_names_index]
      current_cov_values <- covariates[, current_cov]
      index_letter <- paste0("index_cat_ranef_", type_short, "_", current_cov)
      
      
      if(length(unique(current_cov_values)) == 1) stop(paste(type, "covariate", effect_names[effect_names_index], "(fixed effect) has only 1 factor level."), call. = FALSE)
      
      priors_list[[effect_names_index]] <- paste(
        paste("# Covariate:", current_cov),
        paste0(param, ".fixed.categ.", current_cov, "[1] <- 0"),
        paste0("for(", index_letter, " in 2:", length(unique(current_cov_values)), ") {"),
        paste0(param, ".fixed.categ.", current_cov, "[", index_letter, "] ~ ", prior_list$dnorm),
        "}",
        "\n", 
        sep = "\n")
      attr(priors_list, "n_levels")[effect_names_index] <- length(unique(current_cov_values))
    }
    attr(priors_list, "params") <- paste0(param, ".fixed.categ.", effect_names)
    attr(priors_list, "index_letters") <- paste0("index_cat_ranef_", type_short, "_", effect_names)
    
    inits_tmp <- vector(mode = "list", length = length(effect_names))
    names(inits_tmp) <- paste0(param, ".fixed.categ.", effect_names)
    
    for(n in 1:length(effect_names)) {
      inits_tmp[[n]][1] <- NA
      inits_tmp[[n]][2:attr(priors_list, "n_levels")[n]] <- rnorm(attr(priors_list, "n_levels")[n] - 1)
    }
    attr(priors_list, "inits") <- inits_tmp
    rm(inits_tmp)
    
    attr(priors_list, "formula") <- paste(" + ", param,".fixed.categ.", effect_names, "[",   effect_names, "[", stationIndex, "]", "]", 
                                          collapse = "", sep = "")
    
  } else {
    priors_list <- list("# < empty > \n\n")
    attr(priors_list, "params") <- NULL
    attr(priors_list, "formula") <- ""
  }
  return(priors_list)
}


randomEffectPriorsCateg <- function(effect_names, 
                                    covariates, 
                                    type,
                                    prior_list,
                                    speciesIndex, 
                                    stationIndex){
  
  if(type == "detection"){
    param <- "alpha"
    type_short <- "det"
  }
  if(type == "occupancy"){
    param <- "beta"
    type_short <- "occu"
  }
  
  if(!is.null(effect_names)){
    priors_list <- list()
    species_draws <- list()
    
    attr(priors_list, "n_levels") <- rep(NA, times = length(effect_names))
    
    for(effect_names_index in 1:length(effect_names)){
      
      current_cov <- effect_names[effect_names_index]
      current_cov_values <- covariates[, current_cov]
      index_letter <- paste0("index_cat_ranef_", type_short, "_", current_cov)
      
      if(length(unique(current_cov_values)) == 1) stop(paste(type, "covariate", effect_names[effect_names_index], "(random effect) has only 1 factor level." ), call. = FALSE)
      
      priors_list[[effect_names_index]] <- paste(
        paste("# Community mean effects of", current_cov),
        paste0(param, ".ranef.categ.", current_cov, ".mean", "[1] <- 0"),
        paste0(param, ".ranef.categ.", current_cov, ".tau", "[1] <- 0"),
        paste0(param, ".ranef.categ.", current_cov, ".sigma", "[1] <- 0"),     # is included in random effect of intercept already, hence 0
        "",
        paste0("for(", index_letter, " in 2:", length(unique(current_cov_values)), ") {"),
        paste0(param, ".ranef.categ.", current_cov, ".mean",  "[", index_letter, "] ~ ", prior_list$dnorm),   
        paste0(param, ".ranef.categ.", current_cov, ".tau",   "[", index_letter, "] ~ ", prior_list$dgamma),
        paste0(param, ".ranef.categ.", current_cov, ".sigma", "[", index_letter, "] <- sqrt(1 / ", param, ".ranef.categ.", current_cov, ".tau", "[", index_letter, "]", ")"),
        "}",
        "",
        sep = "\n")
      attr(priors_list, "n_levels")[effect_names_index] <- length(unique(current_cov_values))
      
      if(attr(priors_list, "n_levels")[effect_names_index] == 1) stop(paste("covariate", current_cov, "has only 1 unique value:", unique(current_cov_values)), call. = F)
      
      
      species_draws[[effect_names_index]] <- paste(ifelse(effect_names_index == 1, paste("# categorical", type, "covariates with random effects:"), ""),
                                                   paste0("", param, ".ranef.categ.", current_cov, "[", speciesIndex, ", 1] <- 0"), 
                                                   paste0("for (", index_letter, " in 2:",  length(unique(current_cov_values)), "){"),
                                                   paste0("", param, ".ranef.categ.", current_cov, "[", speciesIndex, ", ", index_letter, "] ~ dnorm(", param, ".ranef.categ.", current_cov, ".mean", "[", index_letter, "], ", param, ".ranef.categ.", current_cov, ".tau", "[", index_letter, "])"),
                                                   "}", sep = "\n")
    }
    
    attr(priors_list, "params") <- c(paste0(param, ".ranef.categ.", effect_names, ".mean"),
                                     paste0(param, ".ranef.categ.", effect_names, ".sigma"))
    
    attr(priors_list, "index_letters") <- paste0("index_cat_ranef_", type_short, "_", effect_names)
    
    inits_tmp <- vector(mode = "list", length = length(effect_names)  * 2)
    names(inits_tmp) <- c(paste0(param, ".ranef.categ.", effect_names, ".mean"),
                          paste0(param, ".ranef.categ.", effect_names, ".tau"))
    
    for(n in 1:length(effect_names)) {
      inits_tmp[[n]][1] <- NA
      inits_tmp[[n]][2:attr(priors_list, "n_levels")[n]] <- rnorm(attr(priors_list, "n_levels")[n] - 1)
      
      inits_tmp[[n + length(effect_names)]][1] <- NA
      inits_tmp[[n + length(effect_names)]][2:attr(priors_list, "n_levels")[n]] <- abs(rnorm(attr(priors_list, "n_levels")[n] - 1))
      
    }
    attr(priors_list, "inits") <- inits_tmp
    rm(inits_tmp)
    
    names(species_draws) <- paste0("", param, ".ranef.categ.", effect_names)
    
    attr(priors_list, "formula") <- paste(" + ", param, ".ranef.categ.", effect_names, "[", speciesIndex, ", ", effect_names, "[", stationIndex, "]", "]", 
                                          collapse = "", sep = "")
  } else {
    priors_list <- list("# < empty > \n\n")
    attr(priors_list, "params") <- NULL
    species_draws <- paste("# categorical", type, "covariates: no random effect of species")
    attr(priors_list, "formula") <- ""
  }
  
  
  
  return(list(priors_list, 
              species_draws))
}

 
#' Filter species record table for temporal independence
#'
#' @param recordTable data frame as created by \code{\link{recordTable}}.
#' @param minDeltaTime integer. Time difference between records of the same
#' species at the same station to be considered independent (in minutes)
#' @param deltaTimeComparedTo character. For two records to be considered
#' independent, must the second one be at least \code{minDeltaTime} minutes
#' after the last independent record of the same species
#' (\code{"lastIndependentRecord"}), or \code{minDeltaTime} minutes after the
#' last record (\code{"lastRecord"})?
#' @param speciesCol character. name of the column specifying species in 
#' \code{recordTable}
#' @param stationCol character. Name of the camera trap station column.
#' Assuming "Station" if undefined.
#' @param cameraCol character. Name of the column specifying cameras in 
#' \code{recordTable} (optional).
#' @param camerasIndependent logical. If \code{TRUE}, species records are
#' considered to be independent between cameras at a station.
#' @param recordDateTimeCol character. Name of the column specifying date and 
#' time in \code{recordTable}.
#' @param recordDateTimeFormat character. Format of column 
#' \code{recordDateTimeCol} in \code{recordTable}
#' @param removeDuplicateRecords logical. If there are several records of the
#' same species at the same station (also same camera if cameraID is defined)
#' at exactly the same time, show only one?
#' @param exclude character. Vector of species names to be excluded from the
#' record table
#' @param timeZone character. Must be a value returned by
#' \code{\link[base:timezones]{OlsonNames}}
#' @param writecsv logical. Should the record table be saved as a .csv?
#' @param outDir character. Directory to save csv to. If NULL and
#' \code{writecsv = TRUE}, recordTable will be written to \code{inDir}.
#' @param eventSummaryColumn character. A column in the record table (e.g. from
#' a metadata tag) by to summarise non-independent records (those within
#' \code{minDeltaTime} of a given record) with a user-defined function
#' (\code{eventSummaryFunction})
#' @param eventSummaryFunction character. The function by which to summarise
#' \code{eventSummaryColumn} of non-independent records, e.g. "sum", "max"
#' (optional)
#' @param quiet logical. If TRUE, suppress printing of progress.
#'
#' @return A data frame containing species records and additional information
#' about stations, date, time, filtered for temporal independence.
#' 
#' @export
#' @importFrom lubridate ymd_hms is.POSIXt tz parse_date_time
#' 
#' @author Juergen Niedballa
#'
#' @examples
#' 
#' if (Sys.which("exiftool") != ""){        # only run example if ExifTool is available
#' 
#' # set directory with camera trap images in station directories
#' wd_images_ID_species <- system.file("pictures/sample_images_species_dir", 
#'                                     package = "camtrapR")
#'                                     
#' # create record table without temporal filtering
#'rec_table <- recordTable(inDir                  = wd_images_ID_species,
#'                         IDfrom                 = "directory",
#'                         minDeltaTime           = 0,
#'                         exclude                = "UNID",
#'                         timeZone               = "Asia/Kuala_Lumpur",
#'                         removeDuplicateRecords = TRUE
#'                         )
#'                         
#' # filter for 60 minutes temporal independence
#' rec_table_filt <- filterRecordTable(recordTable         = rec_table, 
#'                                     minDeltaTime        = 60, 
#'                                     stationCol          = "Station",
#'                                     deltaTimeComparedTo = "lastIndependentRecord")
#'  nrow(rec_table)
#'  nrow(rec_table_filt)
#'  
#'  }



filterRecordTable <- function(#inDir,
                        #IDfrom,
                        recordTable,
                        
                        minDeltaTime = 0,
                        deltaTimeComparedTo,
                        speciesCol = "Species",
                        stationCol,
                        
                        # cameraID,
                        cameraCol,
                        camerasIndependent,

                        recordDateTimeCol  = "DateTimeOriginal",
                        recordDateTimeFormat = "ymd HMS",
                        removeDuplicateRecords = TRUE,
                        
                        
                        exclude,
                        timeZone,
                        #metadataHierarchyDelimitor = "|",
                        #metadataSpeciesTag,
                        #additionalMetadataTags,
                        writecsv = FALSE,
                        outDir,
                        
                        #returnFileNamesMissingTags = FALSE,
                        eventSummaryColumn,
                        eventSummaryFunction,
                        #video
                        quiet = FALSE
)
{
  
  # wd0 <- getwd()
  # on.exit(setwd(wd0))
  
  if(!hasArg(stationCol)) stationCol <- "Station"
  if(!is.character(stationCol)) stop("stationCol must be of class 'character'")
  if(!stationCol %in% colnames(recordTable)) stop("stationCol is not a column name in recordTable")
  checkForSpacesInColumnNames(stationCol = stationCol)
  
  # speciesCol <- "Species"
  
  # if(!is.character(IDfrom)) stop("IDfrom must be of class 'character'")
  # IDfrom <- match.arg(IDfrom, choices = c("metadata", "directory"))  
  
  # if(IDfrom == "metadata"){
  #   metadataHierarchyDelimitor <- match.arg(metadataHierarchyDelimitor, choices = c("|", ":"))
  #   if(!hasArg(metadataSpeciesTag))       stop("'metadataSpeciesTag' must be defined if IDfrom = 'metadata'")
  #   if(!is.character(metadataSpeciesTag)) stop("metadataSpeciesTag must be of class 'character'")
  #   if(length(metadataSpeciesTag) != 1)   stop("metadataSpeciesTag must be of length 1")
  # }
  # 
  # multiple_tag_separator <- "_&_"
  
  # check input
  
  if(!inherits(recordTable, "data.frame")) stop("recordTable must be a data frame")
  
  # try to extract timezone (if available)
  if(is.POSIXt(recordTable[, recordDateTimeCol])){
    timeZone <- try(tz(recordTable[, recordDateTimeCol]))
    if(inherits(timeZone, "try-error")) {
      warning("recordDateTimeCol is POSIX, but no time zone information is available. Assuming UTC.") 
      recordTable[, recordDateTimeCol] <- as.character(recordTable[, recordDateTimeCol])
      timeZone <- "UTC"
    }
  } else {
    if(!hasArg(timeZone)) {
      message("timeZone is not specified. Assuming UTC")
      timeZone <- "UTC"
    }
  }
  

  
  # if(Sys.which("exiftool") == "") stop("cannot find ExifTool", call. = FALSE)
  
  if(hasArg(cameraCol)){
    # if(!is.character(cameraID))         stop("cameraID must be of class 'character'", call. = FALSE)
    # cameraID <- match.arg(cameraID, choices = c("filename", "directory"))
    if(!hasArg(camerasIndependent))     stop("camerasIndependent is not defined. It must be defined if cameraID is defined", call. = FALSE)
    if(!is.logical(camerasIndependent)) stop("camerasIndependent must be of class 'logical'", call. = FALSE)
    # stopifnot(hasArg(cameraCol))
  } else { 
    camerasIndependent <- FALSE
    cameraCol <- "Camera"
  }
  

  
  
  if(hasArg(outDir)){
    if(!is.character(outDir))         stop("outDir must be of class 'character'", call. = FALSE)
    if(isFALSE(file.exists(outDir)))  stop("outDir does not exist", call. = FALSE)
  }
  
  if(hasArg(exclude)){
    if(!is.character(exclude)) stop("exclude must be of class 'character'", call. = FALSE)
  }
  
  if(!is.logical(removeDuplicateRecords))     stop("'removeDuplicateRecords' must be logical (TRUE / FALSE)", call. = FALSE)
  # if(!is.logical(returnFileNamesMissingTags)) stop("'returnFileNamesMissingTags' must be logical (TRUE / FALSE)", call. = FALSE)
  
  
  # metadata.tagname <- "HierarchicalSubject"    # for extracting metadata assigned in tagging software
  
  # if(hasArg(additionalMetadataTags)){
  #   if(!is.character(additionalMetadataTags)) stop("additionalMetadataTags must be of class 'character'", call. = FALSE)
  #   if(any(grep(pattern = " ", x = additionalMetadataTags, fixed = TRUE))) stop("In argument additionalMetadataTags, spaces are not allowed")
  #   if("HierarchicalSubject" %in% additionalMetadataTags & IDfrom == "metadata")  {
  #     message("'HierarchicalSubject' may not be in 'additionalMetadataTags' if IDfrom = 'metadata'. It will be ignored because the function returns it anyway.", call. = FALSE)
  #     additionalMetadataTags <- additionalMetadataTags[-grep(pattern = "HierarchicalSubject", x = additionalMetadataTags)]  # remove HierarchicalSubject from additionalMetadataTags
  #   }
  # }
  
  minDeltaTime <- as.integer(minDeltaTime)
  if(!is.integer(minDeltaTime)) stop("'minDeltaTime' must be an integer", call. = FALSE)
  if(minDeltaTime < 0) stop("'minDeltaTime' must be a positive number", call. = FALSE)
  
  if(minDeltaTime != 0){
    if(isFALSE(removeDuplicateRecords)){
      warning("minDeltaTime is > 0. Therefore, removeDuplicateRecords was set to TRUE (otherwise there may be records taken at the same time)", call. = FALSE, immediate. = TRUE)
      removeDuplicateRecords <- TRUE
    }
    
    deltaTimeComparedTo < match.arg(deltaTimeComparedTo, choices = c("lastRecord", "lastIndependentRecord"))
    
    if(!hasArg(deltaTimeComparedTo)) {
      stop(paste("minDeltaTime is not 0. deltaTimeComparedTo must be defined"), call. = FALSE)
    }
  } else {
    if(hasArg(deltaTimeComparedTo)) {
      warning(paste("minDeltaTime is 0. deltaTimeComparedTo = '", deltaTimeComparedTo, "' will have no effect", sep = ""), call. = FALSE, immediate. = TRUE)
    } else {
      deltaTimeComparedTo <- "lastRecord"
    }
  }
  
  if(!is.logical(writecsv))  stop("writecsv must be logical (TRUE or FALSE)", call. = FALSE)
  # if(!is.character(inDir))   stop("inDir must be of class 'character'", call. = FALSE)
  # if(length(inDir) != 1)     stop("inDir may only consist of 1 element only", call. = FALSE)
  # if(!dir.exists(inDir))     stop("Could not find inDir:\n", inDir, call. = FALSE)
  
  if(hasArg(eventSummaryColumn)) {
    if(!is.character(eventSummaryColumn))     stop("eventSummaryColumn must be of class 'character'", call. = FALSE)
    if(!is.character(eventSummaryFunction))   stop("eventSummaryFunction must be of class 'character'", call. = FALSE)
  }
  
  
  if(!is.element(timeZone , OlsonNames())){
    stop("timeZone must be an element of OlsonNames()", call. = FALSE)
  }
  
  
  # remove existing delta.time... / Date/ Time columns
  recordTable <- recordTable[,!grepl("delta.time", colnames(recordTable))]
  recordTable <- recordTable[, !colnames(recordTable) %in% c("Date", "Time")]
  
  # find image directories
  # dirs       <- list.dirs(inDir, full.names = TRUE, recursive = FALSE)
  # dirs_short <- list.dirs(inDir, full.names = FALSE, recursive = FALSE)
  
  # if(length(dirs) == 0) stop("inDir contains no station directories", call. = FALSE)
  
  # max_nchar_station <- max(nchar(dirs_short))
  max_nchar_station <- max(nchar(recordTable[, stationCol]))
  
  
  # process video argument (if present)
  # if(hasArg(video)){
  #   video_out <- processVideoArgument(IDfrom = IDfrom,
  #                                     video  = video)
  #   digiKam_data <- video_out$digiKam_data
  #   file_formats <- video_out$file_formats
  #   # if(isFALSE("jpg" %in% file_formats)) file_formats <- c(file_formats, "jpg")
  # } else {
  #   file_formats <- "jpg"   # jpg, as the default, if video not requested
  # }
  
  # empty list for metadata output
  record.table.list <- split(recordTable, recordTable[, stationCol])
  
  # create command line calls
  # command.tmp  <- paste('exiftool -q -f -t -r -Directory -FileName -EXIF:DateTimeOriginal', 
  #                       ifelse(hasArg(video), 
  #                              paste(" -", video$dateTimeTag, sep = ""), 
  #                              ""),    # if video requested, video date time tag
  #                       ' -HierarchicalSubject',
  #                       ifelse(hasArg(additionalMetadataTags), paste(" -",additionalMetadataTags,  collapse = "", sep = ""), ""),
  #                       paste(" -ext", file_formats, collapse = " ", sep = " "),    # requested file extensions
  #                       ' "', dirs, '"', sep = "")
  
  # construct column names for metadata table
  # colnames.tmp <- c("Directory", "FileName", "DateTimeOriginal")
  # if(hasArg(video)) colnames.tmp <- c(colnames.tmp, video$dateTimeTag)
  # colnames.tmp <- c(colnames.tmp, "HierarchicalSubject")
  # if(hasArg(additionalMetadataTags)) colnames.tmp <- c(colnames.tmp, additionalMetadataTags [nchar(additionalMetadataTags) >= 2]) 
  # only add as new column if entry of additionalMetadataTag has more than 2 characters (to allow users to specify exiftool commands, e.g. -L)
  
  
  for(i in 1:length(record.table.list)){   # loop through station directories
    
    # execute exiftool
    # metadata.tmp <- runExiftool(command.tmp = command.tmp[i], colnames.tmp = colnames.tmp)
    
   # if(is.null(class(metadata.tmp))){            # omit station if no images found
      # 
      # length.tmp <- length(list.files(dirs[i], pattern = paste(".", file_formats, "$", collapse = "|", sep = ""), 
      #                                 ignore.case = TRUE, recursive = TRUE))
      # message(paste(formatC(dirs_short[i], width = max_nchar_station, flag = "-"),  ":  ",
      #               formatC(length.tmp, width = 5), " files      Skipping", sep = ""))
      # warning(paste(dirs_short[i],  ":  contains no files of interest and was omitted\n"), call. = FALSE,  immediate. = FALSE)
   # } else {
      
      # if video files extracted, add DateTimeOriginal and HierarchicalSubject
      # if(hasArg(video)){
      #   metadata.tmp <- addVideoDateTimeOriginal(metadata.tmp = metadata.tmp, video = video)
      #   
      #   # add HierachicalSubject for video files
      #   if(!is.null(digiKam_data)){
      #     digiKamVideoMetadata <- digiKamVideoHierarchicalSubject(stationDir = dirs[i],
      #                                                             digiKamTablesList = digiKam_data,    # output of accessDigiKamDatabase
      #                                                             videoFormat = file_formats[!grepl(file_formats, pattern = "jpg")])
      #     # add HierarchialSubject for video files (match by filename and path)
      #     metadata.tmp <- addVideoHierarchicalSubject (metadata.tmp = metadata.tmp,
      #                                                  video = video,
      #                                                  digiKamVideoMetadata = digiKamVideoMetadata,
      #                                                  digiKamTablesList = digiKam_data,
      #                                                  videoFormat = file_formats[!grepl(file_formats, pattern = "jpg")])
      #   }
      # }
      
    
    metadata.tmp <- record.table.list[[i]]    
      # check presence / consistency of DateTimeOriginal column, go to next station or remove records if necessary
    metadata.tmp[, recordDateTimeCol] <- as.character(metadata.tmp[, recordDateTimeCol])
    
      metadata.tmp <- checkDateTimeOriginal (intable    = metadata.tmp,
                                             # dirs_short = dirs_short,
                                             # i          = i,
                                             stationCol = stationCol,
                                             recordDateTimeCol = recordDateTimeCol
                                             )
      if(is.null(metadata.tmp)) next
      
      
      # now split HierarchicalSubject tags and add as columns to table
      # metadata.tmp <- addMetadataAsColumns (intable                    = metadata.tmp,
      #                                       metadata.tagname           = metadata.tagname,
      #                                       metadataHierarchyDelimitor = metadataHierarchyDelimitor,
      #                                       multiple_tag_separator     = multiple_tag_separator)
      
      # add species names to metadata table (from folders or metadata, otherwise NA)
      # metadata.tmp <- assignSpeciesID(intable                    = metadata.tmp,
      #                                 IDfrom                     = IDfrom,
      #                                 metadataSpeciesTag         = metadataSpeciesTag,
      #                                 speciesCol                 = speciesCol,
      #                                 dirs_short                 = dirs_short,
      #                                 i_tmp                      = i,
      #                                 multiple_tag_separator     = multiple_tag_separator,
      #                                 returnFileNamesMissingTags = returnFileNamesMissingTags)
      
      # if images in station contain no metadata species tags, skip that station
      # if(!is.data.frame(metadata.tmp)){
      #   if(metadata.tmp == "found no species tag") {
      #     # warning(paste(dirs_short[i], ":   metadataSpeciesTag '", metadataSpeciesTag, "' not found in image metadata tag 'HierarchicalSubject'. Skipping", sep = ""), call. = FALSE, immediate. = TRUE)
      #   } else {
      #     warning(paste(dirs_short[i], ":   error in species tag extraction. Skipping. Please report this bug", sep = ""), call. = FALSE, immediate. = TRUE)
      #   }
      #   next
      # }
      
      # # remove empty metadata columns (if HierarchicalSubject is all empty or if additionalMetadataTags were not found)
      # empty_cols <- which(apply(metadata.tmp, MARGIN = 2, FUN = function(X){all(X == "-")}))
      # if(length(empty_cols) >= 1){
      #   metadata.tmp <-  metadata.tmp[, -empty_cols]
      # }
      
      ## add station and camera id to metadata table
      # arg.list0 <- list(intable = metadata.tmp, dirs_short = dirs_short, stationCol = stationCol, hasStationFolders = TRUE, cameraCol = cameraCol, i = i, IDfrom = IDfrom)  # assumes station directories
      # 
      # if(!hasArg(cameraID)) metadata.tmp <- do.call(addStationCameraID, arg.list0)
      # if( hasArg(cameraID)) metadata.tmp <- do.call(addStationCameraID, c(arg.list0, cameraID = cameraID))
      
      # remove species in argument "excluded"
      if(hasArg (exclude)){
        if(any(tolower(metadata.tmp[,speciesCol]) %in% tolower(exclude))) {  # if there is anything to remove
          metadata.tmp <- metadata.tmp[-which(tolower(metadata.tmp[,speciesCol]) %in% tolower(exclude)),]
        }
      }
      
      if(nrow(metadata.tmp) >= 1){   # if anything left after excluding species, do
        
        # convert character vector extracted from images to time object and format for outfilename
        # metadata.tmp[, recordDateTimeCol] <- as.POSIXct(strptime(x = metadata.tmp[, recordDateTimeCol], 
        #                                                      # format = "%Y:%m:%d %H:%M:%S", 
        #                                                      format = recordDateTimeFormat,
        #                                                      tz = timeZone))
        
        if(inherits(metadata.tmp[, recordDateTimeCol], "character")) {
          metadata.tmp[, recordDateTimeCol] <- parse_date_time(metadata.tmp[, recordDateTimeCol], 
                                                               orders = recordDateTimeFormat,
                                                               tz = timeZone)
        }
        if(!inherits(metadata.tmp[, recordDateTimeCol], "POSIXct")) {
         stop(paste0("recordDateTimeCol '", recordDateTimeCol, "' could not be converted to a date-time object."))
        }
        
        # sort by station, (camera), species and time
        if(camerasIndependent == TRUE) {
          metadata.tmp <- metadata.tmp[order(metadata.tmp[,stationCol], metadata.tmp[,speciesCol], metadata.tmp[,cameraCol], metadata.tmp[, recordDateTimeCol]),]
        } else {
          metadata.tmp <- metadata.tmp[order(metadata.tmp[,stationCol], metadata.tmp[,speciesCol], metadata.tmp[, recordDateTimeCol]),]
        }
        
        
        
        #remove duplicate records of same species taken in same second at the same station (by the same camera, if relevant)
        metadata.tmp2 <- removeDuplicatesOfRecords(metadata.tmp          = metadata.tmp,
                                                   removeDuplicateRecords = removeDuplicateRecords,
                                                   camerasIndependent     = camerasIndependent,
                                                   stationCol             = stationCol,
                                                   speciesCol             = speciesCol,
                                                   cameraCol              = cameraCol,
                                                   recordDateTimeCol      = recordDateTimeCol,
                                                   current                = i, 
                                                   total                  = length(record.table.list),
                                                   max_nchar_station      = max_nchar_station,
                                                   quiet                  = quiet)
        
        
        # assess independence between records and calculate time differences
        args.assessTemporalIndependence <- list(intable             = metadata.tmp2,
                                                deltaTimeComparedTo = deltaTimeComparedTo,
                                                columnOfInterest    = speciesCol,
                                                cameraCol           = cameraCol,
                                                camerasIndependent  = camerasIndependent,
                                                recordDateTimeCol   = recordDateTimeCol,
                                                minDeltaTime        = minDeltaTime,
                                                stationCol          = stationCol)
        
        if(hasArg(eventSummaryColumn)) {
          args.assessTemporalIndependence <- c(args.assessTemporalIndependence,
                                               list(eventSummaryColumn   = eventSummaryColumn,
                                                    eventSummaryFunction = eventSummaryFunction))
        }
        
        d1 <- do.call(assessTemporalIndependence, args = args.assessTemporalIndependence)
        
        record.table.list[[i]] <- d1
        
        suppressWarnings(rm(d1))
      }  # end      if(nrow(metadata.tmp) >= 1){} else {...}   # i.e. not all species were excluded
    #}    # end      if(nrow(metadata.tmp) == 0){} else {...}   # i.e. directory i contained images
  }      # end      for(i in 1:length(dirs)){   # loop through station directories
  
  
  # combine all data frames from list into one data frame
  record.table <- as.data.frame(data.table::rbindlist(record.table.list, fill = TRUE, use.names = TRUE))
  
  if(nrow(record.table) == 0){
    # stop(paste("something went wrong. I looked through all those", length(dirs)  ,"folders and now your table is empty. Were date/time information unreadable? Could species tags not be extracted (if IDfrom = 'metadata')? Or did you exclude too many species?"), call. = FALSE)
    stop(paste("something went wrong. I looked through all those", length(record.table.list)  ,"folders and now your table is empty. Were date/time information unreadable? Or did you exclude too many species?"), call. = FALSE)
  }
  
  # rearrange table, add date and time as separate columns. add additional column names as needed.
  
  record.table2  <-  data.frame(record.table [, !colnames(record.table) %in% c("delta.time.secs", "delta.time.mins", 
                                                                              "delta.time.hours", "delta.time.days")],   #[,c(stationCol, speciesCol, recordDateTimeCol)],
                                Date = as.Date (record.table[, recordDateTimeCol], format = "%Y/%M/%d", tz = timeZone),
                                Time = strftime(record.table[, recordDateTimeCol], format = "%H:%M:%S", tz = timeZone),
                                record.table[,c("delta.time.secs", "delta.time.mins", "delta.time.hours", "delta.time.days")])
  
  # if("Directory" %in% colnames(record.table)) record.table2$Directory <- record.table$Directory
  # if("FileName" %in% colnames(record.table)) record.table2$FileName <- record.table$FileName
  
  # metadata_columns <- which(!colnames(record.table) %in% colnames(record.table2))
  
  # add metadata columns
  # if(length(metadata_columns) >= 1){
  #   record.table3 <- cbind(record.table2, record.table[,metadata_columns])
  #   colnames(record.table3)[(ncol(record.table2) + 1) : ncol(record.table3)] <- colnames(record.table)[metadata_columns]
# } else {
    record.table3 <- record.table2
    # }
  
  
  # add camera column (if present)
  if(camerasIndependent == TRUE){
    record.table3 <- data.frame(record.table3[,stationCol],
                                record.table[,cameraCol],
                                record.table3[,-which(colnames(record.table3) %in% c(stationCol, cameraCol))])
    colnames(record.table3)[1] <- stationCol
    colnames(record.table3)[2] <- cameraCol
  }
  
  rownames(record.table3) <- NULL
  
  
  # # warning if additionalMetadataTags were not found
  # if(hasArg(additionalMetadataTags)){
  #   #whichAdditionalMetadataTagsFound <- which(gsub(additionalMetadataTags, pattern = ":", replacement = ".") %in% colnames(record.table3))   # replace : in additionalMetadataTags (if specifying tag groups) with . as found in column names
  #   whichAdditionalMetadataTagsFound <- which(additionalMetadataTags %in% colnames(record.table3))   # replace : in additionalMetadataTags (if specifying tag groups) with . as found in column names
  #   if(length(whichAdditionalMetadataTagsFound) < length(additionalMetadataTags)){
  #     if(length(whichAdditionalMetadataTagsFound) == 0) {  # if none of the additionalMetadataTags was found
  #       warning("metadata tag(s)  not found in image metadata:  ", paste(additionalMetadataTags, collapse = ", "), call. = FALSE)
  #     } else {                                                            # if only some of the additionalMetadataTags was found
  #       warning("metadata tag(s)  not found in image metadata:  ", paste(additionalMetadataTags[-whichAdditionalMetadataTagsFound], collapse = ", "), call. = FALSE)
  #     }
  #   }
  # }
  
  # # make column "HierarchicalSubject" the last column
  # col_to_move <- which(colnames(record.table3) %in% metadata.tagname)
  # if(length(col_to_move) >= 1){
  #   record.table3 <- cbind(record.table3, record.table3[,col_to_move])
  #   record.table3 <- record.table3[,-col_to_move]
  #   colnames(record.table3)[ncol(record.table3)] <- metadata.tagname
  # }
  
  # convert to data.frame, in order to get all the column names right (: becomes .) 
  # NOTE: doesn't work unless data.frame is called. : in column names means they must be adressed as e.g. record.table3$`IPTC:Keywords`
  record.table3 <- data.frame(record.table3, stringsAsFactors = FALSE, check.names = TRUE)
  
  # save table
  if(writecsv == TRUE){
    outtable_filename <- paste("record_table_", minDeltaTime, "min_deltaT_", Sys.Date(), ".csv", sep = "")
    message("saving csv to  ", file.path(outDir, outtable_filename))
    # if(hasArg(outDir) == FALSE){
    #   setwd(inDir)
    # } else {
      setwd(outDir)
    # }
    write.csv(record.table3, file = outtable_filename)
  }
  return(record.table3)
}

#' Optical character recognition (OCR) from data fields in digital images
#' 
#' 
#' Extracts information from the data fields in camera trap images (not the
#' metadata). Many camera traps include data fields in camera trap images,
#' often including date and time of images, and sometimes other information.
#' This function extracts the information from these fields using optical
#' character recognition provided by the package \pkg{tesseract} after reading
#' images using the package \pkg{magick}.
#' 
#' Normally all these information should be in the image metadata. This
#' function is meant as a last resort if image metadata are unreadable or were
#' removed from images. OCR is not perfect and may misidentify characters, so
#' check the output carefully.
#' 
#' The output of this function can be used in
#' \code{\link{writeDateTimeOriginal}} to write date/time into the
#' DateTimeOriginal tag in image metadata, making these images available for
#' automatic processing with \code{\link{recordTable}} and other functions that
#' extract image metadata.
#' 
#' 
#' This function reads all images in inDir (including subdirectories), crops
#' them to the geometries in the "geometries" list, and performs optical
#' character recognition (OCR) on each of these fields (leveraging the magick
#' and tesseract packages).
#' 
#' Geometries are defined with \code{geometry_area} from \pkg{magick}. See
#' \code{\link[magick]{geometry}} for details on how to specify geometries with
#' \code{geometry_area}. The format is: "widthxheight+x_off+y_off", where:
#' 
#' \describe{ \item{width}{width of the area of interest} \item{height}{height
#' of the area of interest} \item{x_off}{offset from the left side of the
#' image} \item{y_off}{offset from the top of the image} }
#' 
#' Units are pixels for all fields. digiKam can help in identifying the correct
#' specification for geometries. Open the Image Editor, left-click and draw a
#' box around the data field of interest. Ensure the entire text field is
#' included inside the box, but nothing else. Now note two pairs of numbers at
#' the bottom of the window, showing the offsets and box size as e.g.:
#' 
#' "(400, 1800) (300 x 60)"
#' 
#' This corresponds to the geometry values as follows:
#' 
#' "(x_off, y_off) (width x height)"
#' 
#' Using these values, you'd run:
#' 
#' \code{geometry_area(x_off = 400, y_off = 1800, width = 300, height = 60)}
#' 
#' and receive
#' 
#' "300x60+400+1800"
#' 
#' as your geometry.
#' 
#' OCR in tesseract has problems with white font on black background. If that
#' is the case in your images, set \code{invert} to \code{TRUE} to invert the
#' image and ensure OCR uses black text on white background.
#' 
#' Even then, output will not be perfect. Error rates in OCR depend on multiple
#' factors, including the text size and font type used. We don't have control
#' over these, so check the output carefully and edit as required.
#' 
#' @param inDir character. Directory containing camera trap images (or
#' subdirectories containing images)
#' @param geometries list. A (possibly named) list of geometry strings defining
#' the image area(s) to extract.
#' @param invert logical. Invert colors in the image? Set to TRUE if text in
#' data field is white on black background. Leave if FALSE if text is black in
#' white background.
#' 
#' @return A \code{data.frame} with original directory and file names, and
#' additional columns for the OCR data of each extracted geometry.
#' 
#' @author Juergen Niedballa
#' 
#' @seealso \code{\link{writeDateTimeOriginal}}
#' 
#' @examples
#' 
#' 
#' \dontrun{
#' # dontrun is to avoid forcing users to install additional dependencies
#' 
#' wd_images_OCR <- system.file("pictures/full_size_for_ocr", package = "camtrapR")
#' 
#' library(magick)
#' 
#' # define geometries
#' geometry1 <- geometry_area(x_off = 0, y_off = 0, width = 183, height = 37)
#' geometry2 <- geometry_area(x_off = 196, y_off = 0, width = 200, height = 17)
#' geometry3 <- geometry_area(x_off = 447, y_off = 0, width = 63, height = 17)
#' geometry4 <- geometry_area(x_off = 984, y_off = 0, width = 47, height = 17)
#' geometry5 <- geometry_area(x_off = 0, y_off = 793, width = 320, height = 17)
#' 
#' # combine geometries into list
#' geometries <- list(date = geometry1, 
#'                    time = geometry2, 
#'                    sequence_id = geometry3,
#'                    temperature = geometry4,
#'                    camera_model = geometry5)
#' 
#' df_image_data  <- OCRdataFields(inDir = wd_images_OCR,
#'                                 geometries = geometries, 
#'                                 invert = TRUE)       
#' df_image_data
#' 
#' # note the mistake in "camera_model"
#' # it should be "PC850", not "PC8S0O"
#' # date and time are correct though
#'                    
#' }
#' 
#' 
#' 
#' @export OCRdataFields
#' 
OCRdataFields <- function(inDir,
                          geometries,
                          invert = FALSE){
  
  if (!requireNamespace("magick", quietly = TRUE)) {
    stop("Please install the package magick to run this function")
  }
  if (!requireNamespace("tesseract", quietly = TRUE)) {
    stop("Please install the package tesseract to run this function")
  }
  
  if(!is.character(inDir)) stop("inDir must be a character string")
  if(!is.list(geometries)) stop("geometries must be a list")
  if(!is.logical(invert)) stop("invert must be logical (TRUE/FALSE)")
  
  lf_full  <- list.files(inDir, recursive = TRUE, full.names = TRUE, pattern = ".JPG$|.jpg$|jpeg$")
  lf_short <- list.files(inDir, recursive = TRUE, full.names = FALSE, pattern = ".JPG$|.jpg$|jpeg$")
  
  if(length(lf_short) == 0) stop("Did not find any images")
  
  img <- magick::image_read(lf_full)
  
  out_list <- lapply(geometries, function (x) {
    # crop image to data field only
    img_crop <- magick::image_crop(img, 
                           geometry = x)  
    
    if(invert) {
      img_crop <- magick::image_negate(img_crop)
    }
    
    # add a white margin to help OCR
    img_crop <- magick::image_border(img_crop, color = "white")
    
    
    ocr_list <- lapply(img_crop, magick::image_ocr)
    
    # remove line break at end of string
    ocr_unlist <- gsub("\n$", "", unlist(ocr_list))
    
    return(ocr_unlist)
  })
  
  if(is.null(names(geometries))){
    names(out_list) <- paste(geometries)
  }
  
  out_df1 <- as.data.frame(out_list)
  
  out <- data.frame(filename_full = lf_full,
                    # filename_short = lf_short,
                    out_df1)
  
  return(out)
}


# NOTE: it would be good to offer the option to write those tags to the DateTimeOriginal field in image metadata
# maybe not immediately so users can first check everything is in order. 
# exiftool syntax see Q5: https://exiftool.org/faq.html

#' Overview of the functions in the camtrapR package
#' 
#' This package provides a streamlined workflow for processing data generated
#' in camera trap-based wildlife studies and prepares input for further
#' analyses, particularly in occupancy and spatial capture-recapture
#' frameworks. It suggests a simple data structure and provides functions for
#' managing digital camera trap photographs (and videos), generating record
#' tables, maps of species richness and species detections and species activity
#' diagrams. It further helps prepare subsequent analyses by creating
#' detection/non-detection matrices for occupancy analyses, e.g. in the
#' \pkg{unmarked} or \pkg{ubms} packages, and \code{capthist} objects for
#' spatial capture-recapture analyses in the \pkg{secr} package. In addition,
#' basic survey statistics are computed. The functions build on one another in
#' a logical sequence. The only manual input needed is species (and individual)
#' identification, which is achieved by moving images into species directories
#' or by tagging images in image management software. Besides, a table holding
#' basic information about camera trap station IDs, locations and trapping
#' periods must be created in spreadsheet software.
#' 
#' Image metadata (such as date and time or user-assigned tags) are extracted
#' from the images using Phil Harvey's ExifTool (available from
#' \url{https://exiftool.org/}) and the information is stored in a record
#' table. An adjustable criterion for temporal independence of records can be
#' applied. Maps of species presence and species richness can be generated.
#' Several functions are available for plotting single- and two-species
#' activity patterns. Information about the camera-specific trapping periods
#' (and periods of malfunction) are summarized into a matrix about camera trap
#' operability. These, together with the record table, are used to generate
#' species detection histories for occupancy and spatial capture-recapture
#' analyses. The user has considerable freedom in generating the detection
#' histories; sampling occasion length, beginning date and and occasion start
#' times are adjustable. In addition, trapping effort (i.e. active trap nights
#' per station and occasion) can be computed for use as a covariate / offset on
#' detection probability.
#' 
#' @name camtrapR-package
#' @aliases camtrapR-package camtrapR
#' @docType package
#' 
#' @section User support: The camtrapR Google group is an online support and
#' help forum for camtrapR users. You can find it here:
#' \url{https://groups.google.com/forum/#!forum/camtrapr}.
#' 
#' @author Juergen Niedballa
#' 
#' Maintainer:Juergen Niedballa <camtrapr@@gmail.com>
#' 
#' @seealso \pkg{overlap} \pkg{unmarked} \pkg{ubms} \pkg{secr} \pkg{wiqid}
#' 
#' @references
#' 
#' Niedballa, J., Sollmann, R., Courtiol, A., Wilting, A. (2016): camtrapR: an
#' R package for efficient camera trap data management. Methods in Ecology and
#' Evolution, 7(12).
#' \url{https://besjournals.onlinelibrary.wiley.com/doi/full/10.1111/2041-210X.12600}
#' \cr
#' 
#' camtrapR Google Group \url{https://groups.google.com/forum/#!forum/camtrapr}
#' \cr
#' 
#' Lemon, J. (2006) Plotrix: a package in the red light district of R. R-News,
#' 6(4): 8-12. \cr
#' 
#' Mike Meredith and Martin Ridout (2018). overlap: Estimates of coefficient of
#' overlapping for animal activity patterns. R package version 0.3.2.
#' \url{https://CRAN.R-project.org/package=overlap} \cr
#' 
#' Phil Harvey's ExifTool \url{https://exiftool.org/} \cr
#' @keywords internal
"_PACKAGE"





#' Sample camera trap station information
#' 
#' Example camera trap station information table
#' 
#' This is a general example of how information about camera trap stations are
#' arranged in camtrapR. It contains setup and retrieval dates and coordinates.
#' If more than 1 camera was set up at a station (e.g. 2 cameras facing each
#' other), a camera ID column must be added, with camera-specific information
#' instead of station-specific information. If cameras malfunctioned
#' repeatedly, additional pairs of problem columns can be added, e.g.
#' "Problem2_from" and "Problem2_to" etc..
#' 
#' The variables are as follows:
#' 
#' \itemize{ \item Station.  Camera trap station ID \item utm_y.  y coordinate
#' of station (northing) \item utm_x.  x coordinate of station (easting) \item
#' Setup_date.  camera trap setup date \item Retrieval_date.  camera trap
#' retrieval date \item Problem1_from.  first day of camera malfunction \item
#' Problem1_to.  last day of camera malfunction }
#' 
#' @name camtraps
#' @docType data
#' @format A data frame with 3 rows and 7 variables
#' @note The coordinates can be in the units of any coordinate system. UTM was
#' chosen as an example, but it could be latlong or anything else, too.
#' \code{\link[secr]{capthist}} objects (as created by
#' \code{\link{spatialDetectionHistory}} for spatial capture-recapture
#' analyses) expect the unit to be meters.
#' @keywords datasets
NULL





#' Sample multi-season camera trap station information
#' 
#' Example multi-season camera trap station information table
#' 
#' This is a general example of how information about camera trap stations from
#' multiple seasons are arranged in camtrapR. It contains setup and retrieval
#' dates, coordinates and a season identifier. If more than 1 camera was set up
#' at a station (e.g. 2 cameras facing each other), a camera ID column must be
#' added, with camera-specific information instead of station-specific
#' information. If cameras malfunctioned repeatedly, additional pairs of
#' problem columns can be added, e.g. "Problem2_from" and "Problem2_to" etc..
#' 
#' Note that season 2010 has an additional station (StationD). This is to
#' simulate a situation where a station was not set up during an entire season.
#' 
#' The variables are as follows:
#' 
#' \itemize{ \item Station.  Camera trap station ID \item utm_y.  y coordinate
#' of station (northing) \item utm_x.  x coordinate of station (easting) \item
#' Setup_date.  camera trap setup date \item Retrieval_date.  camera trap
#' retrieval date \item Problem1_from.  first day of camera malfunction \item
#' Problem1_to.  last day of camera malfunction \item session.  Identified for
#' trapping session / season }
#' 
#' @name camtrapsMultiSeason
#' @docType data
#' @format A data frame with 7 rows and 8 variables
#' @note The coordinates can be in the units of any coordinate system. UTM was
#' chosen as an example, but it could be latlong or anything else, too.
#' \code{\link[secr]{capthist}} objects (as created by
#' \code{\link{spatialDetectionHistory}} for spatial capture-recapture
#' analyses) expect the unit to be meters. \code{\link[secr]{capthist}} alse
#' require session information as integer numbers starting with 1.
#' 
#' "Season" and "session" are used synonymously here. \pkg{secr} nomenclature
#' is "session", in \pkg{unmarked} it is "season".
#' @keywords datasets
#' @examples
#' 
#' # data were created with the following code:
#' data(camtraps)
#' 	
#' camtraps_season2 <- camtraps
#' 
#' # change 2009 to 2010
#' camtraps_season2[, "Setup_date"]     <- gsub("2009", "2010", camtraps_season2[, "Setup_date"])
#' camtraps_season2[, "Retrieval_date"] <- gsub("2009", "2010", camtraps_season2[, "Retrieval_date"])
#' camtraps_season2[, "Problem1_from"]  <- gsub("2009", "2010", camtraps_season2[, "Problem1_from"])
#' camtraps_season2[, "Problem1_to"]    <- gsub("2009", "2010", camtraps_season2[, "Problem1_to"])
#' 
#' # add an extra station with different dates in session 2010
#' camtraps_season2 <- rbind(camtraps_season2, NA)
#' camtraps_season2$Station[4] <- "StationD"
#' camtraps_season2$utm_y[4]  <- 607050
#' camtraps_season2$utm_x[4]  <- 525000
#' camtraps_season2$Setup_date[4]      <- "04/04/2010"
#' camtraps_season2$Retrieval_date[4]  <- "17/06/2010"
#' camtraps_season2$Problem1_from[4]   <- "20/05/2010"
#' camtraps_season2$Problem1_to[4]     <- "30/05/2010"
#' 
#' # add season column
#' camtraps$session         <- 2009
#' camtraps_season2$session <- 2010
#' 
#' # combine the tables for 2 seasons
#' camtrapsMultiSeason <- rbind(camtraps, camtraps_season2)
#' 
#' 
NULL





#' Sample single-species record table with custom metadata from camera trap
#' images
#' 
#' Sample single-species record table with individual IDs from the tagged
#' sample images in the package. Generated with function
#' \code{\link{recordTableIndividual}}.
#' 
#' 
#' The variables are as follows:
#' 
#' \itemize{ \item Station.  Camera trap station ID \item Species.  Species ID
#' \item Individual.  Individual ID \item DateTimeOriginal.  Date and time as
#' extracted from image \item Date.  record date \item Time.  record time of
#' day \item delta.time.secs.  time difference to first species record at a
#' station (seconds) \item delta.time.mins.  time difference to first species
#' record at a station (minutes) \item delta.time.hours.  time difference to
#' first species record at a station (hours) \item delta.time.days.  time
#' difference to first species record at a station (days) \item Directory.
#' Image directory \item FileName.  image filename \item HierarchicalSubject.
#' content of the HierarchicalSubject image metadata tag \item Model.  camera
#' model extracted from image metadata \item Make.  camera make extracted from
#' image metadata \item metadata_Species.  content of custom image metadata tag
#' "Species" (see HierarchicalSubject) \item metadata_individual.  content of
#' custom image metadata tag "individual" (see HierarchicalSubject) }
#' 
#' @name recordTableIndividualSample
#' @docType data
#' @format A data frame with 21 rows and 17 variables
#' @keywords datasets
NULL





#' Sample single-species multi-season record table with custom metadata from
#' camera trap images
#' 
#' Sample single-species multi-season record table with individual IDs from the
#' tagged sample images in the package. Generated with function
#' \code{\link{recordTableIndividual}}, then duplicated to simulate a second
#' year.
#' 
#' 
#' The variables are as follows:
#' 
#' \itemize{ \item Station.  Camera trap station ID \item Species.  Species ID
#' \item Individual.  Individual ID \item DateTimeOriginal.  Date and time as
#' extracted from image \item Date.  record date \item Time.  record time of
#' day \item delta.time.secs.  time difference to first species record at a
#' station (seconds) \item delta.time.mins.  time difference to first species
#' record at a station (minutes) \item delta.time.hours.  time difference to
#' first species record at a station (hours) \item delta.time.days.  time
#' difference to first species record at a station (days) \item Directory.
#' Image directory \item FileName.  image filename \item HierarchicalSubject.
#' content of the HierarchicalSubject image metadata tag \item Model.  camera
#' model extracted from image metadata \item Make.  camera make extracted from
#' image metadata \item metadata_Species.  content of custom image metadata tag
#' "Species" (see HierarchicalSubject) \item metadata_individual.  content of
#' custom image metadata tag "individual" (see HierarchicalSubject) }
#' 
#' @name recordTableIndividualSampleMultiSeason
#' @docType data
#' @format A data frame with 31 rows and 17 variables
#' @keywords datasets
#' @examples
#' 
#' # example data were created as follows:
#' data(recordTableIndividualSample)
#' 
#' recordTableIndividualSample_season2 <- recordTableIndividualSample[1:10,]
#' recordTableIndividualSample_season2$DateTimeOriginal <- gsub("2009", "2010", 
#'       recordTableIndividualSample_season2$DateTimeOriginal)
#' recordTableIndividualSampleMultiSeason <- rbind(recordTableIndividualSample, 
#'       recordTableIndividualSample_season2)
#' 
NULL





#' Sample species record table from camera trap images
#' 
#' Sample species record table from camera trap images generated from the
#' sample images in the package with the function \code{\link{recordTable}} .
#' 
#' 
#' The variables are as follows:
#' 
#' \itemize{ \item Station.  Camera trap station ID \item Species.  Species ID
#' \item DateTimeOriginal.  Date and time as extracted from image \item Date.
#' record date \item Time.  record time of day \item delta.time.secs.  time
#' difference to first species record at a station (seconds) \item
#' delta.time.mins.  time difference to first species record at a station
#' (minutes) \item delta.time.hours.  time difference to first species record
#' at a station (hours) \item delta.time.days.  time difference to first
#' species record at a station (days) \item Directory.  Image directory \item
#' FileName.  image filename }
#' 
#' @name recordTableSample
#' @docType data
#' @format A data frame with 39 rows and 11 variables
#' @keywords datasets
NULL





#' Sample multi-season species record table from camera trap images
#' 
#' Sample multi-season species record table from camera trap images generated
#' from the sample images in the package with the function
#' \code{\link{recordTable}}. Season 2009 is the same as
#' \code{\link{recordTableSample}}, season 2010 was simulated by adding 1 year
#' to these records.
#' 
#' 
#' The variables are as follows:
#' 
#' \itemize{ \item Station.  Camera trap station ID \item Species.  Species ID
#' \item DateTimeOriginal.  Date and time as extracted from image \item Date.
#' record date \item Time.  record time of day \item delta.time.secs.  time
#' difference to first species record at a station (seconds) \item
#' delta.time.mins.  time difference to first species record at a station
#' (minutes) \item delta.time.hours.  time difference to first species record
#' at a station (hours) \item delta.time.days.  time difference to first
#' species record at a station (days) \item Directory.  Image directory \item
#' FileName.  image filename }
#' 
#' @name recordTableSampleMultiSeason
#' @docType data
#' @format A data frame with 78 rows and 11 variables
#' @keywords datasets
#' @examples
#' 
#' 	# data were created with the following code:
#' 
#' 	data(recordTableSample)
#' 	recordTableSample_season2 <- recordTableSample
#' 	
#' 	# substitute 2009 with 2010
#' 	recordTableSample_season2$DateTimeOriginal <- gsub("2009", "2010", 
#' 	     recordTableSample_season2$DateTimeOriginal) 
#'   # combine with season 2009
#' 	recordTableSampleMultiSeason <- rbind(recordTableSample, recordTableSample_season2)    
#' 
NULL





#' Sample camera trap time shift table
#' 
#' Sample camera trap time shift table
#' 
#' If image Exif metadata timestamps are wrong systematically (e.g. because
#' camera system time was not set after changing batteries), it can be
#' corrected using a \code{data.frame} in the following format using function
#' \code{\link{timeShiftImages}}. For details on data format, please see
#' \code{\link{timeShiftImages}}.
#' 
#' The variables are as follows:
#' 
#' \itemize{ \item Station.  Camera trap station ID \item camera.  Camera trap
#' ID (optional) \item timeshift.  time shift amount to be applied \item sign.
#' direction of time shift
#' 
#' }
#' 
#' @name timeShiftTable
#' @docType data
#' @format A data frame with 2 rows and 4 variables
#' @keywords datasets
NULL

#' Add a directory to PATH temporarily
#' 
#' Temporarily adds a directory to the environmental variable PATH for system
#' calls from within R. This allows Windows users to store exiftool.exe
#' anywhere on their hard drive and is useful if they cannot store the file 
#' in system directories. It is not needed on Linux or MacOS machines.
#' 
#' Several functions within this package depend on ExifTool. Under Windows,
#' exiftool.exe cannot be used if it is not in a directory path specified in
#' PATH. This can be solved by adding the directory containing exiftool.exe for
#' temporary use within the running R process. It can also be useful in other 
#' contexts besides Exiftool.
#' 
#' @param directory character. the directory in the file system to add to PATH 
#' (e.g. the directory containing exiftool.exe).
#' 
#' @return invisible logical indicating whether \code{directory} was added to
#' PATH successfully (in the running R process).
#' 
#' @note The directories in PATH can be queried by \code{Sys.getenv("PATH")}.
#' 
#' @author Juergen Niedballa
#' 
#' @examples
#' 
#' exiftool_dir <- "C:/Path/To/Exiftool"
#' addToPath(directory = exiftool_dir)
#' 
#' # check if it has been added to PATH
#' grepl(exiftool_dir,  Sys.getenv("PATH"))
#' 
#' @export addToPath
#' 
addToPath <- function(directory){
  WPATH <- Sys.getenv("PATH")
  WPATH1 <- paste(directory, WPATH, sep=";")
  Sys.setenv(PATH=WPATH1)
  return(invisible(grepl(directory,  Sys.getenv("PATH"))))
}

#' Create species directories for species identification
#' 
#' This function creates species subdirectories within station directories.
#' They can be used for species identification by manually moving images into
#' the respective species directories. The function can also delete empty
#' species directories (if species were not detected at sites). It is not
#' necessary to run this function if animals will be identified by metadata
#' tagging.
#' 
#' This function should be run after \code{\link{imageRename}}. Empty
#' directories can be created as containers for species identification if
#' images are identified with the drag & drop method. After species
#' identification is complete, empty species directories can be deleted using
#' \code{removeFolders = TRUE}. The function will delete only directories which
#' are specified in \code{species}. If \code{hasCameraFolders} was set to
#' \code{TRUE} in function \code{\link{imageRename}}, \code{hasCameraFolders}
#' must be set to \code{TRUE} here too. Species directories will then be
#' created within each camera subdirectory of each station directory. if the
#' user wishes to identify species by metadata tagging, running this function
#' is not needed.
#' 
#' @param inDir character. Directory containing camera trap images sorted into
#' station subdirectories (e.g. inDir/StationA/)
#' @param hasCameraFolders logical. Do the station directories in \code{inDir}
#' have camera-subdirectories (e.g. inDir/StationA/CameraA1;
#' inDir/StationA/CameraA2)?
#' @param species character. names of species directories to be created in
#' every station (or station/camera) subdirectory of \code{inDir}
#' @param removeFolders logical. Indicating whether to create (TRUE) or remove
#' (FALSE) species directories .
#' @return A \code{data.frame} with directory names and an indicator for
#' whether directories were created or deleted.
#' 
#' @author Juergen Niedballa
#' 
#' @examples
#' 
#' \dontrun{
#' 
#' # create dummy directories for tests
#' # (normally, you'd use directory containing renamed, unsorted images)
#' 
#' # this will be used as inDir
#' wd_createDirTest <- file.path(getwd(), "createSpeciesFoldersTest")
#' 
#' # now we create 2 station subdirectories
#' dirs_to_create <- file.path(wd_createDirTest, c("StationA", "StationB"))
#' sapply(dirs_to_create, FUN = dir.create, recursive = TRUE)
#' 
#' # species names for which we want to create subdirectories
#' species <- c("Sambar Deer", "Bay Cat")
#' 
#' # create species subdirectories
#' SpecFolderCreate1 <- createSpeciesFolders (inDir               = wd_createDirTest,
#'                                            species             = species,
#'                                            hasCameraFolders = FALSE,
#'                                            removeFolders       = FALSE)
#'   
#' SpecFolderCreate1
#' 
#' # check if directories were created
#' list.dirs(wd_createDirTest)
#' 
#' # delete empty species directories
#' SpecFolderCreate2 <- createSpeciesFolders (inDir               = wd_createDirTest,
#'                                            species             = species,
#'                                            hasCameraFolders = FALSE,
#'                                            removeFolders       = TRUE)
#' 
#' SpecFolderCreate2
#' 
#' # check if species directories were deleted
#' list.dirs(wd_createDirTest)
#' 
#' }
#' 
#' @export createSpeciesFolders
#' 
createSpeciesFolders <- function(inDir,
                                 hasCameraFolders,
                                 species,
                                 removeFolders = FALSE){

  stopifnot(is.character(species))
  stopifnot(is.logical(removeFolders))
  stopifnot(is.logical(hasCameraFolders))
  
  if(!dir.exists(inDir)) stop("Could not find inDir:\n", inDir, call. = FALSE)

  dirs       <- list.dirs(inDir, full.names = TRUE, recursive = FALSE)
  dirs_short <- list.dirs(inDir, full.names = FALSE , recursive = FALSE)

  if(length(dirs) == 0) stop("inDir has no (station) subdirectories")

  if(hasCameraFolders == TRUE){
    dirs.to.create <- file.path(rep(list.dirs(dirs, recursive = FALSE), each = length(species)), species)
  } else {
    dirs.to.create <- file.path(rep(dirs, each = length(species)), species)
  }

  if(removeFolders == FALSE){
    tmp.create <- suppressWarnings(sapply(dirs.to.create, FUN = dir.create, showWarnings = TRUE, recursive = FALSE))
    dat.out1 <- data.frame(directory = dirs.to.create,
                           created = tmp.create,
                           exists = file.exists(dirs.to.create))
    rownames(dat.out1) <- NULL

    message(paste("created", sum(tmp.create == TRUE), "directories"))
    if(sum(tmp.create == FALSE) != 0){
      message(paste(sum(tmp.create == FALSE & file.exists(dirs.to.create)), "directories already existed"))
    }
    return(dat.out1)

  } else {

    dirs.to.delete <- dirs.to.create[file.exists(dirs.to.create)]

    if(length(dirs.to.delete) == 0) stop("nothing to delete: the specified species subdirectories do not exist")

    n.images.per.folder <- sapply(sapply(dirs.to.delete, FUN = list.files, recursive = TRUE), FUN = length)

    which.to.delete <- which(n.images.per.folder == 0)
    which.not.to.delete <- which(n.images.per.folder != 0)

    unlink(dirs.to.delete[which.to.delete], recursive=TRUE)

    Sys.sleep(0.2)     # because a network drive may sync too slowly and thus can still show a deleted dir as existing

    dat.out2 <- data.frame(directory = dirs.to.delete,
                           still.exists = file.exists(dirs.to.delete))

    if(length(which.to.delete)!= 0){
      message(paste("deleted", sum(dat.out2$still.exists == FALSE), "empty directories"))
    }

    if(length(which.not.to.delete)!= 0){
      cat("\n")
      message(paste("could not delete", sum(dat.out2$still.exists == TRUE), "non-empty directories"))
    }
    return(dat.out2)
  }
}

#' Extract covariate values from spatial rasters and prepare rasters for spatial predictions
#'
#' @description
#'  This function extracts covariate values from spatial raster data 
#'  (e.g. for use in modelling) and prepares these covariates for use in 
#'  spatial predictions. 
#'  
#'  It accepts a camera trap table containing spatial information, 
#'  along with either a directory containing covariate raster files, a character 
#'  vector specifying the file paths of these covariate rasters, or direct
#'  SpatRaster objects from the terra package.
#'  
#'  Additionally, users can provide parameters to control how covariates are
#'  extracted, and how they are aggregated to prediction rasters.
#'  
#'  The function can also download elevation data and calculate terrain metrics
#'  if requested.
#'  
#'  The function generates prediction rasters based on a provided template or 
#'  creates one automatically if no template is provided.
#'  
#'  The function returns a list containing the camera trap dataset with 
#'  extracted covariate values (e.g for use in occupancy modelling), and 
#'  prediction rasters ready for spatial modeling.
#'  
#' @param CTtable sf object as defined by the \code{sf} package. Essentially a camera trap data frame with spatial information.
#' @param directory character. The directory containing the covariate rasters.
#' @param filenames character (optionally named). A vector of file paths of covariate rasters. If it is named the covariates will be named according to the names. If unnamed the file names will be used as covariate names.
#' @param rasters SpatRaster object or list of SpatRaster objects. Direct input of rasters from the terra package instead of reading from disk.
#' @param buffer_ct numeric. A value (in meters) by which to buffer the point locations in \code{CTtable} for extraction of covariate values. 
#' @param bilinear logical. If \code{TRUE}, extract covariate values with bilinear interpolation (nearest 4 raster cells). If \code{FALSE}, extract value at the cell the point falls in. Only relevant if \code{buffer_ct} is 0.
#' @param buffer_aoi numeric. A value (in meters) by which to buffer the overall camera trapping grid to ensure that prediction rasters are larger than the camera trapping grid.
#' @param raster_template SpatRaster. A SpatRaster (as defined in the \code{terra} package) to use as template for the creation of prediction rasters.
#' @param resolution numeric. Spatial resolution of prediction rasters in the units of the coordinate reference system. Ignored if \code{raster_template} is provided.
#' @param append logical. If \code{TRUE}, add the extracted covariates to the existing \code{CTtable}. If \code{FALSE}, return only the extracted covariate values without the existing \code{CTtable}.
#' @param formats character. Possible file formats for raster data (must include the dot). Defaults to \code{.tif} files.
#' @param recursive logical. If \code{TRUE}, search for raster files recursively in subdirectories when a directory is provided. Defaults to \code{FALSE}.
#' @param download_elevation logical. If \code{TRUE}, download elevation data from AWS. Defaults to \code{FALSE}.
#' @param elevation_zoom numeric. Zoom level for elevation data download (6-12). Higher values provide more detail but longer download times. Zoom 12 corresponds to ~20m pixel resolution, 11 to ~40m, 10 to ~80m, and so on (resolution halves with each decrease in zoom level). Defaults to 9.
#' @param terrain_measures character. Vector of terrain metrics to calculate from elevation data. Options include "slope" (slope in degrees), "aspect" (compass direction in degrees), "TRI" (Terrain Ruggedness Index, measuring elevation difference between adjacent cells), "TPI" (Topographic Position Index, comparing cell elevation to mean of surrounding cells), and "roughness" (difference between max and min of surrounding cells). Defaults to NULL (no terrain metrics).
#' @param standardize_na logical. Logical. If \code{TRUE}, ensures all layers in the prediction raster have identical NA patterns by setting a cell to NA in all bands if it's NA in any band. This creates consistency for spatial predictions across covariates but may lose data in covariates.
#' @param scale_covariates logical. If \code{TRUE}, scale numeric covariates and return both original and scaled versions of data and prediction rasters. Scaling is performed using R's \code{scale} function. Defaults to \code{FALSE}.
#'
#' @details
#' 
#' The input camera trap table must be an \code{\link[sf]{sf}} object (a data 
#' frame with a geometry column specifying the spatial information). For details 
#' on how to convert an exising camera trap table to sf, see the examples below.
#' The input rasters can be in different coordinate systems. During covariate 
#' extraction the CTtable is projected to each raster's coordinate system 
#' individually. For the prediction raster all input rasters are either 
#' resampled or reprojected to a consistent coordinate system.
#' 
#' When \code{recursive = TRUE} and a directory is provided, the function will search
#' for raster files in all subdirectories. In this case, the subdirectory names are
#' used as covariate names, and only one raster file per subdirectory is allowed.
#'
#' When \code{download_elevation = TRUE}, the function will download elevation data
#' from AWS using the \code{elevatr} package. The \code{elevation_zoom}
#' parameter controls the level of detail, with values between 6 and 12.
#' Higher zoom levels provide finer resolution but require longer download times and
#' may consume significant memory. Approximate resolutions: zoom 12 = ~20m, 
#' 11 = ~40m, 10 = ~80m, etc.
#' 
#' If \code{terrain_measures} is specified, the function calculates the requested 
#' terrain metrics from the elevation data using \code{terra::terrain()} with the 
#' default 3x3 neighborhood. Available terrain metrics include "slope", "aspect", 
#' "TRI" (Terrain Ruggedness Index), "TPI" (Topographic Position Index), and 
#' "roughness".
#' 
#' When using \code{scale_covariates = TRUE}, the function returns both original and
#' scaled versions of the data and prediction rasters. #' The function uses R's 
#' \code{scale()} function to perform centering and scaling, and
#' includes the scaling parameters in the returned metadata.
#' 
#' \subsection{Warning about Categorical Covariates}{
#' This function does not explicitly handle categorical rasters. All raster values are
#' treated as numeric, which can be problematic when scaling is applied. The function
#' attempts to identify "likely categorical" variables (numeric variables with few unique
#' integer values) and will provide warnings, but it cannot automatically handle them
#' correctly for scaling.
#' 
#' When using scaled covariates with categorical variables in models:
#' \itemize{
#'   \item Use \code{CTtable_scaled} for numeric predictors
#'   \item Use \code{CTtable} (original) for categorical predictors
#'   \item Similarly, use \code{predictionRaster_scaled} for numeric predictors in spatial predictions
#'   \item Use \code{predictionRaster} for categorical predictors in spatial predictions
#' }
#' 
#' Future versions may implement proper categorical raster handling with RAT (Raster
#' Attribute Table) support.
#' }
#' 
#' @return
#' \describe{
#' When \code{scale_covariates = FALSE}, a list containing three elements:
#'   \item{CTtable}{An \code{sf} object representing the camera trap data frame with extracted covariate values.}
#'   \item{predictionRaster}{A \code{SpatRaster} object containing covariate raster layers}
#'   \item{originalRaster}{A list of the original input rasters}
#' 
#' When \code{scale_covariates = TRUE}, a list containing six elements:
#'   \item{CTtable}{The original \code{sf} object with unscaled covariate values}
#'   \item{CTtable_scaled}{The \code{sf} object with scaled numeric covariate values}
#'   \item{predictionRaster}{The original unscaled prediction raster}
#'   \item{predictionRaster_scaled}{The prediction raster with scaled numeric layers}
#'   \item{originalRaster}{A list of the original input rasters}
#'   \item{scaling_params}{A list containing center and scale information of numeric covariates}
#' }
#' @export
#'
#' @examples
#' 
#' \dontrun{
#' # load camera trap table
#' data(camtraps)
#' 
#' # create sf object
#' camtraps_sf <- st_as_sf(camtraps, 
#'                         coords = c("utm_x", "utm_y"), 
#'                         crs = 32650)
#' 
#' # extract covariates (with 100m buffer around cameras)
#' # doesn't run because 'directory' is only a placeholder
#' 
#' covariates <- createCovariates(camtraps_sf, 
#' "path/to/covariate_rasters", 
#' buffer_ct = 100, 
#' buffer_aoi = 1000,
#' resolution = 100)
#' 
#' 
#' # extract covariates with elevation data (this code runs)
#' 
#' covariates_elev <- createCovariates(camtraps_sf,
#' buffer_ct = 100,
#' buffer_aoi = 1000,
#' resolution = 100,
#' download_elevation = TRUE,
#' elevation_zoom = 11,
#' terrain_measures = c("slope", "aspect", "TRI"))
#' 
#' # Note that if local rasters are available they can be extracted alongside
#' # elevation data in a single function call
#' 
#' # camera trap table with extracted covariate values
#' camtraps_sf_cov <- covariates_elev$CTtable
#' 
#' # covariate raster layer
#' r_cov <- covariates_elev$predictionRaster
#' plot(r_cov)
#' 
#' # Use SpatRaster objects directly as input
#' r1 <- rast("elevation.tif")
#' r2 <- rast("landcover.tif")
#' raster_list <- list(elevation = r1, landcover = r2)
#' 
#' covariates_direct <- createCovariates(camtraps_sf,
#'                                       rasters = raster_list,
#'                                       buffer_ct = 100,
#'                                       resolution = 100)
#'
#' # Scale numeric covariates for modeling
#' covariates_scaled <- createCovariates(camtraps_sf,
#'                                       rasters = raster_list,
#'                                       buffer_ct = 100,
#'                                       resolution = 100,
#'                                       scale_covariates = TRUE)
#'
#' # Use scaled data with categorical variables
#' # Mix and match from original and scaled outputs for tabular data
#' model_data <- covariates_scaled$CTtable_scaled  # Use scaled numeric covariates
#' model_data$landcover <- covariates_direct$CTtable$landcover  # Use original categorical covariate
#' 
#' # Mix and match for prediction rasters
#' # Create combined prediction raster (scaled numeric variables & original categorical variables)
#' # Extract scaled elevation layer
#' elev_scaled <- covariates_scaled$predictionRaster_scaled$elevation
#' 
#' # Extract original landcover layer (categorical)
#' landcover_orig <- covariates_direct$predictionRaster$landcover
#' 
#' # Combine into a new SpatRaster for predictions
#' prediction_raster <- c(elev_scaled, landcover_orig)
#' names(prediction_raster) <- c("elevation", "landcover")
#' 
#' # Use this combined raster for spatial predictions
#' plot(prediction_raster)
#' }
#' 

createCovariates <- function(CTtable,
                             directory,
                             filenames,
                             rasters,
                             buffer_ct = 0,
                             bilinear = FALSE,
                             buffer_aoi = 1000,
                             raster_template = NULL,
                             resolution = NULL,
                             append = TRUE, 
                             formats = ".tif",
                             recursive = FALSE,
                             download_elevation = FALSE,
                             elevation_zoom = 10,
                             terrain_measures = NULL,
                             standardize_na = FALSE,
                             scale_covariates = FALSE)
{
  if(!inherits(CTtable, "sf")){ 
    stop("CTtable is not an sf object.")
  }
  if(is.na(sf::st_crs(CTtable))) stop("Coordinate reference system of CTtable is not defined. See: sf::st_crs(CTtable).")
  
  if (nrow(CTtable) == 0) {
    stop("The 'CTtable' is empty.")
  }
  
  # Count how many input sources are provided
  input_sources <- sum(
    hasArg(directory),
    hasArg(filenames),
    hasArg(rasters),
    download_elevation
  )
  
  # Check if any input source is provided
  if (input_sources == 0) {
    stop("Please provide one of 'directory', 'filenames', 'rasters', or set 'download_elevation = TRUE'.")
  }
  
  # Check that only one of directory, filenames, or rasters is provided
  if (sum(hasArg(directory), hasArg(filenames), hasArg(rasters)) > 1) {
    stop("Provide only one of 'directory', 'filenames', or 'rasters'.")
  }
  
  # Validate formats
  if(!all(grepl("^\\.", formats))) {
    stop("All values in 'formats' must start with a dot (e.g., '.tif')")
  }
  
  # Validate terrain_measures if provided
  if (!is.null(terrain_measures)) {
    valid_measures <- c("slope", "aspect", "TRI", "TPI", "roughness")
    terrain_measures <- match.arg(terrain_measures, valid_measures, several.ok = TRUE)
  }
  
  # ensure we have internet connection for download
  if (download_elevation) {
    if(isFALSE(has_internet_socket())) stop("Your machine appears to be offline. Please connect to the internet to download elevation data, or set download_elevation = FALSE", call. = FALSE)
    }
  
  
  # Validate elevation_zoom
  if (download_elevation) {
    if (!is.numeric(elevation_zoom) || elevation_zoom < 6 || elevation_zoom > 12) {
      stop("'elevation_zoom' must be a numeric value between 6 and 12.")
    }
  }
  

  
  # Check buffer sizes
  if (!is.numeric(buffer_ct) || buffer_ct < 0) {
    stop("Buffer size for 'buffer_ct' must be numeric and positive.")
  }
  if (!is.numeric(buffer_aoi) || buffer_aoi < 0) {
    stop("Buffer size for 'buffer_aoi' must be numeric and positive.")
  }
  
  # Throw warning if buffer sizes exceed thresholds
  if (buffer_ct > 2000) {
    warning("Buffer size for 'buffer_ct' exceeds 2km. Ensure this is intended.", call. = F, immediate. = T)
  }
  if (buffer_aoi > 50000) {
    warning("Buffer size for 'buffer_aoi' exceeds 50km. Ensure this is intended.", call. = F, immediate. = T)
  }
  
  if (is.null(raster_template) && is.null(resolution)) {
    warning("Neither 'raster_template' or 'resolution' are defined. No prediction rasters will be created.", call. = F, immediate. = T)
    create_raster <- FALSE
  } else {
    create_raster <- TRUE
  }
  
  # Initialize list for all rasters
  list_r_cov <- list()
  
  # Create raster template if needed
  if(is.null(raster_template) && !is.null(resolution)) {
    if (resolution <= 0) {
      stop("Invalid spatial resolution. Please provide a positive value.")
    }
    
    if(buffer_aoi > 0) {
      extent_aoi <- terra::ext(terra::buffer(terra::vect(CTtable), width = buffer_aoi))
    } else {
      extent_aoi <- terra::ext(CTtable)
    }
    
    # Get the CRS of the input data
    input_crs <- sf::st_crs(CTtable)
    
    # Check if the CRS is geographic (like WGS84)
    if(input_crs$proj4string %>% stringr::str_detect("\\+proj=longlat")) {
      # If geographic, create a temporary UTM-based CRS for the center of the study area
      center_point <- sf::st_coordinates(sf::st_centroid(sf::st_as_sfc(sf::st_bbox(CTtable))))
      utm_zone <- (floor((center_point[1] + 180)/6) %% 60) + 1
      temp_crs <- sprintf("+proj=utm +zone=%d +datum=WGS84", utm_zone)
      
      # Transform extent to the temporary UTM CRS
      extent_aoi_utm <- sf::st_transform(CTtable, temp_crs) %>% 
        terra::vect() %>% 
        terra::buffer(width = buffer_aoi) %>%
        terra::ext()
      
      # Create raster template in UTM
      raster_template_utm <- terra::rast(extent_aoi_utm, 
                                         resolution = resolution, 
                                         crs = temp_crs)
      
      # Project back to original CRS
      raster_template <- terra::project(raster_template_utm, input_crs$proj4string)
    } else {
      # If not geographic, create raster template directly
      raster_template <- terra::rast(extent_aoi, 
                                     resolution = resolution, 
                                     crs = input_crs$wkt)
    }
    
    # After creating raster_template
    n_cells <- terra::ncell(raster_template)
    if(n_cells < 100) {
      warning("The resulting raster has very few cells (", n_cells, "). Consider using a smaller resolution value.")
    } else if(n_cells > 1e8) {
      warning("The resulting raster has a very large number of cells (", n_cells, "). This may cause performance issues. Consider using a larger resolution value.")
    }
  }
  
  # Process local raster files if directory or filenames are provided
  if (hasArg(directory) || hasArg(filenames)) {
    
    if (hasArg(directory)) {
      if(length(directory) != 1) stop("The 'directory' parameter should be a single directory path.")
      
      # Get file paths from directory
      lf_covariates <- list.files(directory, recursive = recursive, pattern = paste0(formats, "$"), full.names = TRUE)
      if(length(lf_covariates) == 0) stop(paste("No files with suffix", paste(formats, collapse = "/"), "found in", directory))
      
      # Always use the file names (not directory names) as the base name
      covariate_names <- tools::file_path_sans_ext(basename(lf_covariates))
      
      if(recursive) {
        # Check for subdirectories with depth > 1
        dir_depth <- sapply(lf_covariates, function(x) length(strsplit(x, .Platform$file.sep)[[1]]))
        if(any(dir_depth > length(strsplit(directory, .Platform$file.sep)[[1]]) + 2)) {
          warning("Subdirectories with depth > 1 found. Only using rasters in the first level of subdirectories.", call. = F, immediate. = T)
          valid_depth <- dir_depth <= length(strsplit(directory, .Platform$file.sep)[[1]]) + 2
          lf_covariates <- lf_covariates[valid_depth]
          covariate_names <- covariate_names[valid_depth]
        }
        
        # Check for multiple files per covariate
        files_per_cov <- table(covariate_names)
        if(any(files_per_cov > 1)) {
          multi_file_covs <- names(files_per_cov[files_per_cov > 1])
          stop(paste("Multiple raster files found for covariate(s):", 
                     paste(multi_file_covs, collapse = ", "), 
                     ". Ensure only one raster file per covariate directory."))
        }
      }
      
      local_r_cov <- lapply(lf_covariates, terra::rast)
      names(local_r_cov) <- covariate_names
    } else {
      if (!is.vector(filenames) && !is.character(filenames)) {
        stop("Invalid 'filenames'. Please provide a (named) vector.")
      }
      
      # ensure files exist
      file_exists <- file.exists(filenames)
      if(!all(file_exists)) stop(paste("file in filenames don't exist:\n", 
                                       paste(filenames[!file_exists], collapse = "\n")))
      
      # Extract covariate names and file paths from parameter 'filenames'
      if(!is.null(names(filenames))) {
        covariate_names <- names(filenames)  # use names if available
      } else {
        covariate_names <- tools::file_path_sans_ext(basename(filenames))
      }
      local_r_cov <- lapply(filenames, function(file) terra::rast(file))
      names(local_r_cov) <- covariate_names
    }
    
    # Add local rasters to the master list
    list_r_cov <- c(list_r_cov, local_r_cov)
  }
  
  # Process provided SpatRaster objects
  if (hasArg(rasters)) {
    if (inherits(rasters, "SpatRaster")) {
      # Single SpatRaster object
      if (terra::nlyr(rasters) == 1) {
        # Single layer raster
        if (is.null(names(rasters)) || names(rasters) == "") {
          # No name provided, assign default name
          warning("No name provided for input raster. Using 'layer1' as name.", call. = F, immediate. = T)
          names(rasters) <- "layer1"
        }
        spat_r_cov <- list(rasters)
        names(spat_r_cov) <- names(rasters)
      } else {
        # Multi-layer raster - handle as a single multi-band raster
        spat_r_cov <- list(rasters)
        names(spat_r_cov) <- "raster"
      }
    } else if (is.list(rasters)) {
      # List of SpatRaster objects
      if (!all(sapply(rasters, inherits, "SpatRaster"))) {
        stop("All elements in 'rasters' must be SpatRaster objects.")
      }
      
      # Check for and assign names to rasters
      if (is.null(names(rasters))) {
        # No names provided, assign default names
        warning("No names provided for input rasters. Using 'layer1', 'layer2', ... as names.", call. = F, immediate. = T)
        names(rasters) <- paste0("layer", 1:length(rasters))
      }
      
      spat_r_cov <- rasters
    } else {
      stop("'rasters' must be a SpatRaster object or a list of SpatRaster objects.")
    }
    
    # Add spatial rasters to the master list
    list_r_cov <- c(list_r_cov, spat_r_cov)
  }
  
  # Download elevation data and calculate terrain metrics if requested
  if (download_elevation) {
    # Create buffered polygon from points for elevation download
    buffered_sf <- sf::st_buffer(CTtable, dist = buffer_aoi)
    
    # Transform to EPSG:4326 if needed for elevatr
    if (sf::st_crs(buffered_sf) != sf::st_crs(4326)) {
      buffered_sf_4326 <- sf::st_transform(buffered_sf, 4326)
    } else {
      buffered_sf_4326 <- buffered_sf
    }
    
    # Download elevation data from AWS
    elevation_rast <- elevatr::get_elev_raster(
      locations = buffered_sf_4326,
      z = elevation_zoom,
      source = "aws",
      clip = "bbox"
    )
    
    # Convert to terra raster if necessary
    if(inherits(elevation_rast, "RasterLayer")) {
      elevation_rast <- terra::rast(elevation_rast)
    }
    
    # Initialize terrain rasters list with elevation
    terrain_rasts <- list(elevation = elevation_rast)
    
    # Calculate requested terrain metrics using terra::terrain with default 3x3 neighborhood
    if (!is.null(terrain_measures)) {
      if ("slope" %in% terrain_measures) {
        terrain_rasts$slope <- terra::terrain(elevation_rast, "slope", unit = "degrees")
      }
      if ("aspect" %in% terrain_measures) {
        terrain_rasts$aspect <- terra::terrain(elevation_rast, "aspect", unit = "degrees")
      }
      if ("TRI" %in% terrain_measures) {
        terrain_rasts$TRI <- terra::terrain(elevation_rast, "TRI")
      }
      if ("TPI" %in% terrain_measures) {
        terrain_rasts$TPI <- terra::terrain(elevation_rast, "TPI")
      }
      if ("roughness" %in% terrain_measures) {
        terrain_rasts$roughness <- terra::terrain(elevation_rast, "roughness")
      }
    }
    
    # Add terrain rasters to the master list
    list_r_cov <- c(list_r_cov, terrain_rasts)
  }
  
  if(length(list_r_cov) == 0) {
    stop("No covariate raster loaded. Please check your inputs.")
  }
  
  # Set extraction method based on bilinear parameter
  extract_method <- if(bilinear) "bilinear" else "simple"
  
  
  # Extract values from all rasters
  cov_value_list <- lapply(1:length(list_r_cov), FUN = function(i) {
    current_crs <- sf::st_crs(list_r_cov[[i]])$wkt
    CTtable_i <- sf::st_transform(CTtable, current_crs)
    CTtable_i_v <- terra::vect(CTtable_i)
    
    # extract at exact camera trap location or with buffer
    if(buffer_ct == 0) {
      extracted <- terra::extract(list_r_cov[[i]], 
                                  CTtable_i_v, 
                                  method = extract_method)
    } else {
      CTtable_i_buff <- terra::buffer(CTtable_i_v, 
                                      width = buffer_ct)
      extracted <- terra::extract(list_r_cov[[i]], 
                                  CTtable_i_buff, 
                                  fun = mean, 
                                  na.rm = TRUE)
    }
    
    # Handle multi-band rasters
    if(terra::nlyr(list_r_cov[[i]]) > 1) {
      layer_names <- names(list_r_cov[[i]])
      base_name <- names(list_r_cov)[i]
      
      # Apply make.names to ensure column names match R's sanitization rules
      colnames(extracted)[-1] <- make.names(paste(base_name, layer_names, sep = "_"), unique = FALSE)
    } else {
      colnames(extracted)[-1] <- make.names(names(list_r_cov)[i], unique = FALSE)
    }
    
    return(extracted[, -1, drop = FALSE])
  })
  
  df_covariates <- do.call(cbind, cov_value_list)
  
  # Check for NAs in extracted covariates
  na_counts <- sapply(df_covariates, function(x) sum(is.na(x)))
  if(any(na_counts > 0)) {
    na_cols <- names(na_counts[na_counts > 0])
    na_percent <- na_counts[na_counts > 0] / nrow(df_covariates) * 100
    
    # Format the warning message
    warning_msg <- "NAs found in extracted covariates:\n"
    for(i in seq_along(na_cols)) {
      warning_msg <- paste0(warning_msg, "  ", na_cols[i], ": ", na_counts[na_cols[i]],
                            " NAs (", round(na_percent[i], 1), "%)\n")
    }
    warning(warning_msg, call. = FALSE, immediate. = TRUE)
  }
  
  # cbind to input camera trap table, if requested
  if(append) {
    df_covariates_out <- cbind(CTtable, df_covariates)
  } else {
    df_covariates_out <- df_covariates
  }
  
  # Identify which columns are covariates (needed for scaling later)
  if(append) {
    # Get the names of the columns that were added as covariates
    if(inherits(df_covariates_out, "sf")) {
      geom_col <- attr(df_covariates_out, "sf_column")
      original_cols <- setdiff(names(CTtable), geom_col)
    } else {
      original_cols <- names(CTtable)
    }
    covariate_cols <- setdiff(names(df_covariates_out), original_cols)
  } else {
    # All columns are covariates if append=FALSE
    covariate_cols <- names(df_covariates_out)
  }
  
  # clip rasters to desired extent (in their original CRS)
  list_r_cov_c <- lapply(1:length(list_r_cov), 
                         FUN = function(i){
                           # get raster CRS
                           current_crs <- sf::st_crs(list_r_cov[[i]])
                           # project camera trap table to that CRS
                           CTtable_i <- sf::st_transform(CTtable, crs = current_crs)
                           CTtable_i_v <- terra::vect(CTtable_i)
                           
                           # get extent of AOI (buffered, if requested)
                           if(buffer_aoi > 0) {
                             extent_aoi <- terra::ext(terra::buffer(CTtable_i_v, buffer_aoi))
                           } else {
                             extent_aoi <- terra::ext(CTtable_i)
                           }
                           
                           # clip raster
                           out <- try(terra::crop(list_r_cov[[i]], 
                                                  y = extent_aoi)) 
                           
                           if(inherits(out, "try-error")) {
                             warning(paste("Failed to crop raster '", names(list_r_cov)[i], "'. Skipping this covariate."), call. = F, immediate. = T)
                             return(NULL)
                           } else {
                             return(out)
                           }
                         })
  names(list_r_cov_c) <- names(list_r_cov)
  
  
  if(create_raster){
    # resample rasters according to template
    r_cov_list <- lapply(1:length(list_r_cov_c),
                         FUN = function(i){
                           if(is.null(list_r_cov_c[[i]])) return(NULL)
                           current_crs <- sf::st_crs(list_r_cov_c[[i]])
                           
                           crs_template <- sf::st_crs(raster_template)
                           
                           # Process the raster
                           if(current_crs == crs_template) {
                             out <- terra::resample(list_r_cov_c[[i]], raster_template)
                           } else {
                             out <- terra::project(list_r_cov_c[[i]], raster_template)
                           }
                           
                           # Ensure consistent naming for multi-band rasters
                           if(terra::nlyr(list_r_cov_c[[i]]) > 1) {
                             base_name <- names(list_r_cov)[i]
                             layer_names <- names(list_r_cov_c[[i]])
                             # Make the raster names match the sanitized column names
                             names(out) <- make.names(paste(base_name, layer_names, sep = "_"), unique = FALSE)
                             
                           } else {
                             names(out) <- make.names(names(list_r_cov)[i], unique = FALSE)
                           }
                           
                           return(out)
                         })
    names(r_cov_list) <- names(list_r_cov)
    
    r_cov <- terra::rast(r_cov_list)
  } else {
    r_cov <- NULL
  }
  
  # ensure the output has correct layer names (rast doesn't preserve names of multi-band raster bands)
  if (!is.null(r_cov)) {
    names(r_cov) <- unlist(sapply(r_cov_list, names))
  }
  
  # Standardize NA patterns across all layers if requested
  if(standardize_na && !is.null(r_cov) && terra::nlyr(r_cov) > 1) {
    r_cov <- terra::mask(x = r_cov,                  # update this layer
                         mask = any(is.na(r_cov)),   # where any band is NA
                         maskvalues = 1,             # (coded as "1" in any(is.na()))
                         updatevalue = NA)           # replace cells where any band is 1 in the mask with NA
  }
  
  # Scale covariates if requested
  if(scale_covariates && length(covariate_cols) > 0) {
    # Check which columns are numeric and can be scaled
    is_numeric <- sapply(df_covariates_out[covariate_cols], is.numeric)
    numeric_covariate_cols <- covariate_cols[is_numeric]
    
    if(length(numeric_covariate_cols) > 0) {
      # Handle sf objects for scaling
      is_sf <- inherits(df_covariates_out, "sf")
      if(is_sf) {
        df_temp <- sf::st_drop_geometry(df_covariates_out)
      } else {
        df_temp <- df_covariates_out
      }
      
      # Scale numeric covariates
      scaled_data <- scale(df_temp[numeric_covariate_cols])
      center_values <- attr(scaled_data, "scaled:center")
      scale_values <- attr(scaled_data, "scaled:scale")
      names(center_values) <- numeric_covariate_cols
      names(scale_values) <- numeric_covariate_cols
      
      # Create scaled dataframe
      df_scaled <- df_temp
      df_scaled[numeric_covariate_cols] <- scaled_data
      
      # Restore geometry if needed
      if(is_sf) {
        geom_col <- attr(df_covariates_out, "sf_column")
        geometry <- sf::st_geometry(df_covariates_out)
        df_scaled <- sf::st_sf(df_scaled, geometry = geometry)
      }
      
      # Create scaled raster if prediction raster exists
      r_cov_scaled <- NULL
      if(!is.null(r_cov) && terra::nlyr(r_cov) > 0) {
        r_cov_scaled <- r_cov
        
        # Scale applicable layers
        for(col in numeric_covariate_cols) {
          if(col %in% names(r_cov_scaled)) {
            r_cov_scaled[[col]] <- (r_cov_scaled[[col]] - center_values[col]) / scale_values[col]
          } else {
            # warning if covariate in df_scaled but not in prediction stack
            warning(paste(col, "is not in prediction raster"), call. = F, immediate. = T)
          }
        }
      }
      
      # Print summary message
      message(paste0("Scaled ", length(numeric_covariate_cols), " numeric covariates."))
      
      # Prepare full output with both original and scaled data
      result <- list(
        CTtable = df_covariates_out,
        CTtable_scaled = df_scaled,
        predictionRaster = r_cov,
        predictionRaster_scaled = r_cov_scaled,
        originalRaster = list_r_cov,
        scaling_params = list(center = center_values, scale = scale_values)
      )
    } else {
      warning("No numeric covariates found to scale.", call. = F, immediate. = T)
      # Return unscaled data if no numeric covariates
      result <- list(
        CTtable = df_covariates_out,
        predictionRaster = r_cov,
        originalRaster = list_r_cov
      )
    }
  } else {
    # Return standard output without scaling
    result <- list(
      CTtable = df_covariates_out,
      predictionRaster = r_cov,
      originalRaster = list_r_cov
    )
  }
  
  return(result)
}

context("createCovariates")

library(testthat)
library(sf)
library(terra)
library(withr)

# --- Test Setup ---

# 1. Create a mock sf object from the 'camtraps' dataset
data(camtraps)
camtraps_sf <- st_as_sf(camtraps,
                        coords = c("utm_x", "utm_y"),
                        crs = 32650) # UTM Zone 50N

# 2. Create mock SpatRaster objects for reproducible testing
# Raster 1: Same CRS as camtraps_sf, predictable values
r1 <- rast(crs = "EPSG:32650", extent = ext(c(522500 , 526300 , 603700 , 607300)), resolution = 100)
values(r1) <- 1:ncell(r1)
names(r1) <- "cov_utm"

# Raster 2: Different CRS (WGS84), tests reprojection
r2 <- rast(crs = "EPSG:4326", extent = ext(c(117.2, 117.24, 5.46, 5.5)), resolution = 0.01)
values(r2) <- (1:ncell(r2)) * 10
names(r2) <- "cov_wgs"


# --- Test Suite ---

testthat::describe("Input Validation and Error Handling", {
  
  test_that("it stops for invalid CTtable inputs", {
    expect_error(createCovariates(camtraps), "CTtable is not an sf object.")
    
    
    expect_error({
      camtraps_sf_no_crs <- st_as_sf(camtraps, coords = c("utm_x", "utm_y"))
      createCovariates(camtraps_sf_no_crs)
    }, "Coordinate reference system of CTtable is not defined.")
    
    expect_error(createCovariates(camtraps_sf[0,]), "The 'CTtable' is empty.")
  })
  
  test_that("it stops when no data source is provided", {
    expect_error(createCovariates(camtraps_sf), 
                 "Please provide one of 'directory', 'filenames', 'rasters', or set 'download_elevation = TRUE'.")
  })
  
  test_that("it stops when multiple data sources are provided", {
    expect_error(createCovariates(camtraps_sf, directory = "a", rasters = r1),
                 "Provide only one of 'directory', 'filenames', or 'rasters'.")
  })
  
  test_that("it stops for invalid buffer or resolution values", {
    expect_error(createCovariates(camtraps_sf, rasters = r1, buffer_ct = -10),
                 "Buffer size for 'buffer_ct' must be numeric and positive.")
    expect_error(createCovariates(camtraps_sf, rasters = r1, resolution = -50),
                 "Invalid spatial resolution. Please provide a positive value.")
  })
})


testthat::describe("Core Functionality from SpatRaster Objects", {
  # Core Functionality from SpatRaster Objects
  
  test_that("it extracts values correctly from a list of SpatRasters", {
    raster_list <- list(my_utm_ras = r1, my_wgs_ras = r2)
    result <- createCovariates(camtraps_sf, rasters = raster_list, resolution = 200)
    
    # Check output structure
    expect_s3_class(result$CTtable, "sf")
    expect_s4_class(result$predictionRaster, "SpatRaster")
    expect_named(result, c("CTtable", "predictionRaster", "originalRaster"))
    
    # Check that new columns were added
    expect_true(all(c("my_utm_ras", "my_wgs_ras") %in% names(result$CTtable)))
    expect_equal(ncol(result$CTtable), ncol(camtraps_sf) + 2)
    
    # Check that prediction raster has the right layers and CRS
    expect_equal(names(result$predictionRaster), c("my_utm_ras", "my_wgs_ras"))
    expect_equal(crs(result$predictionRaster, proj = TRUE), crs(camtraps_sf, proj = TRUE))
  })
  
  test_that("buffering and interpolation options work", {
    # Test with a 200m buffer
    result_buffer <- createCovariates(camtraps_sf, rasters = r1, buffer_ct = 200, resolution = 200)
    # Test without buffer (simple extraction)
    result_simple <- createCovariates(camtraps_sf, rasters = r1, buffer_ct = 0, bilinear = FALSE, resolution = 200)
    
    # Values should be different due to averaging in the buffer
    expect_false(all(result_buffer$CTtable$cov_utm == result_simple$CTtable$cov_utm))
  })
  
  test_that("append = FALSE works correctly", {
    result <- createCovariates(camtraps_sf, rasters = r1, append = FALSE, resolution = 200)
    
    # Output should be a data.frame, not an sf object
    expect_s3_class(result$CTtable, "data.frame")
    expect_false(inherits(result$CTtable, "sf"))
    expect_equal(ncol(result$CTtable), 1) # Only the one covariate column
    expect_named(result$CTtable, "cov_utm")
  })
})


testthat::describe("Core Functionality from Files", {
  
  # Use a temporary directory for file-based tests
  with_tempdir({
    
    # Save our mock rasters to the temp directory
    writeRaster(r1, "cov_utm.tif")
    writeRaster(r2, "cov_wgs.tif")
    dir.create("subdir")
    writeRaster(r1 * 2, "subdir/cov_sub.tif", names = "cov_sub")
    
    test_that("it extracts values from a directory", {
      result <- createCovariates(camtraps_sf, directory = ".", resolution = 200)
      
      expect_true(all(c("cov_utm", "cov_wgs") %in% names(result$CTtable)))
      expect_equal(nlyr(result$predictionRaster), 2)
    })
    
    test_that("it extracts values recursively from a directory", {
      result <- createCovariates(camtraps_sf, directory = ".", recursive = TRUE, resolution = 200)
      
      # Should find all 3 rasters
      expect_true(all(c("cov_utm", "cov_wgs", "cov_sub") %in% names(result$CTtable)))
      expect_equal(nlyr(result$predictionRaster), 3)
    })
    
    test_that("it extracts values from a named character vector of filenames", {
      file_paths <- c(ras1 = "cov_utm.tif", ras2 = "cov_wgs.tif")
      result <- createCovariates(camtraps_sf, filenames = file_paths, resolution = 200)
      
      # Check that the provided names ("ras1", "ras2") are used
      expect_true(all(c("ras1", "ras2") %in% names(result$CTtable)))
      expect_equal(names(result$predictionRaster), c("ras1", "ras2"))
    })
  })
})


testthat::describe("Elevation Download and Terrain Measures", {
  
  # skip_if_not_installed("mockery")
  skip_if_not_installed("elevatr")
  
  test_that("it downloads elevation and calculates terrain measures", {
    # Create a mock elevation raster to be returned by the mocked function
    mock_elev_rast <- rast(crs = "EPSG:4326", extent = ext(c(117.2, 117.24, 5.46, 5.5)), resolution = 0.01)
    values(mock_elev_rast) <- 1000 + (1:ncell(mock_elev_rast)) / 10
    names(mock_elev_rast) <- "mock_elevation"
    
    # Mock the elevatr::get_elev_raster function to avoid actual download.
    # It will now return our predictable mock raster instead.
    m <- mockery::mock(mock_elev_rast)
    mockery::stub(where = createCovariates, what = "elevatr::get_elev_raster", how = m)
    
    # Run the function
    result <- expect_warning(
      createCovariates(camtraps_sf,
                       download_elevation = TRUE,
                       terrain_measures = c("slope", "TRI"),
                       resolution = 200),
      regexp = "NAs found in extracted covariates"
    )
    
    # Check that the mock function was called exactly once
    mockery::expect_called(m, 1)
    
    # Check for correct columns and layers
    expected_names <- c("elevation", "slope", "TRI")
    expect_true(all(expected_names %in% names(result$CTtable)))
    expect_true(all(expected_names %in% names(result$predictionRaster)))
  })
})


testthat::describe("Scaling and Output Structure", {
  
  test_that("scale_covariates = TRUE produces correct output structure and values", {
    # Use a list of rasters for this test
    raster_list <- list(elevation = r1, other_metric = r2)
    
    result <- createCovariates(camtraps_sf, 
                               rasters = raster_list, 
                               resolution = 200,
                               scale_covariates = TRUE)
    
    # 1. Check the output list structure
    expected_names <- c("CTtable", "CTtable_scaled", "predictionRaster", 
                        "predictionRaster_scaled", "originalRaster", "scaling_params")
    expect_named(result, expected_names)
    
    # 2. Check the scaled table
    ct_scaled <- result$CTtable_scaled
    expect_s3_class(ct_scaled, "sf")
    
    # Scaled numeric columns should have mean ~ 0 and sd ~ 1
    expect_equal(mean(ct_scaled$elevation), 0, tolerance = 1e-9)
    expect_equal(sd(ct_scaled$elevation), 1, tolerance = 1e-9)
    expect_equal(mean(ct_scaled$other_metric), 0, tolerance = 1e-9)
    expect_equal(sd(ct_scaled$other_metric), 1, tolerance = 1e-9)
    
    # 3. Check the scaled raster
    ras_scaled <- result$predictionRaster_scaled
    expect_s4_class(ras_scaled, "SpatRaster")
    
    # The global mean of the scaled raster layer should be close to 0
    global_mean <- global(ras_scaled, "mean", na.rm = TRUE)$mean
    names(global_mean) <- names(ras_scaled)
    expect_lt(abs(global_mean["elevation"]), 0.2)
    
    # 4. Check scaling parameters
    params <- result$scaling_params
    expect_named(params, c("center", "scale"))
    expect_named(params$center, c("elevation", "other_metric"))
    expect_true(is.numeric(params$scale["elevation"]))
  })
})

#' Plot overlapping kernel densities of two-species activities
#' 
#' This function plots kernel density estimates of two species' diel activity
#' data by calling the function \code{\link[overlap]{overlapPlot}} from package
#' \pkg{overlap}. It further computes the overlap coefficient Dhat1 by calling
#' \code{\link[overlap]{overlapEst}}.
#' 
#' \code{...} can be graphical parameters passed on to function
#' \code{\link[overlap]{overlapPlot}}, e.g. \code{linetype}, \code{linewidth},
#' \code{linecol} (see example below).
#' 
#' \code{recordDateTimeFormat} defaults to the "YYYY-MM-DD HH:MM:SS"
#' convention, e.g. "2014-09-30 22:59:59". \code{recordDateTimeFormat} can be
#' interpreted either by base-R via \code{\link[base]{strptime}} or in
#' \pkg{lubridate} via \code{\link[lubridate]{parse_date_time}} (argument
#' "orders"). \pkg{lubridate} will be used if there are no "\%" characters in
#' \code{recordDateTimeFormat}.
#' 
#' For "YYYY-MM-DD HH:MM:SS", \code{recordDateTimeFormat} would be either
#' "\%Y-\%m-\%d \%H:\%M:\%S" or "ymd HMS". For details on how to specify date
#' and time formats in R see \code{\link[base]{strptime}} or
#' \code{\link[lubridate]{parse_date_time}}.
#' 
#' @param recordTable data.frame. the record table created by
#' \code{\link{recordTable}}
#' @param speciesA Name of species 1 (as found in \code{speciesCol} of
#' recordTable)
#' @param speciesB Name of species 2 (as found in \code{speciesCol} of
#' recordTable)
#' @param speciesCol character. name of the column specifying species names in
#' \code{recordTable}
#' @param recordDateTimeCol character. name of the column specifying date and
#' time in \code{recordTable}
#' @param recordDateTimeFormat character. format of column
#' \code{recordDateTimeCol} in \code{recordTable}
#' @param plotR logical. Show plots in R graphics device?
#' @param writePNG logical. Create pngs of the plots?
#' @param addLegend logical. Add a legend to the plots?
#' @param legendPosition character. Position of the legend (keyword)
#' @param plotDirectory character. Directory in which to create png plots if
#' \code{writePNG = TRUE}
#' @param createDir logical. Create \code{plotDirectory}?
#' @param pngMaxPix integer. image size of png (pixels along x-axis)
#' @param add.rug logical. add a rug to the plot?
#' @param overlapEstimator character. Which overlap estimator to return (passed
#' on to argument \code{type} in \code{\link[overlap]{overlapEst}})
#' @param \dots additional arguments to be passed to function
#' \code{\link[overlap]{overlapPlot}}
#' 
#' @return Returns invisibly the \code{data.frame} with plot coordinates
#' returned by \code{\link[overlap]{overlapPlot}}.
#' 
#' @note Please be aware that the function (like the other activity... function
#' of this package) use clock time, not solar time. If your survey was long
#' enough to see changes in sunrise and sunset times, this may result in biased
#' representations of species activity.
#' 
#' @author Juergen Niedballa
#' 
#' @seealso \code{\link{activityDensity}} \cr
#' \url{https://www.kent.ac.uk/smsas/personal/msr/overlap.html}
#' 
#' @references Mike Meredith and Martin Ridout (2018). overlap: Estimates of
#' coefficient of overlapping for animal activity patterns. R package version
#' 0.3.2. \url{https://CRAN.R-project.org/package=overlap} \cr Ridout, M.S. and
#' Linkie, M. (2009) Estimating overlap of daily activity patterns from camera
#' trap data. Journal of Agricultural, Biological and Environmental Statistics,
#' 14, 322-337.
#' 
#' @examples
#' 
#' if(requireNamespace("overlap")) {
#' 
#' # load record table
#' data(recordTableSample)
#' 
#' # define species of interest
#' speciesA_for_activity <- "VTA"    # = Viverra tangalunga, Malay Civet
#' speciesB_for_activity <- "PBE"    # = Prionailurus bengalensis, Leopard Cat
#' 
#' # create activity overlap plot (basic)
#' activityOverlap (recordTable = recordTableSample,
#'                  speciesA    = "VTA",    # = Viverra tangalunga, Malay Civet
#'                  speciesB    = "PBE",    # = Prionailurus bengalensis, Leopard Cat
#'                  writePNG    = FALSE,
#'                  plotR       = TRUE
#' )
#' 
#' 
#'                                      
#' # create activity overlap plot (prettier and with some overlapPlot arguments set)
#' 
#' activityOverlap (recordTable = recordTableSample,
#'                  speciesA    = speciesA_for_activity,
#'                  speciesB    = speciesB_for_activity,
#'                  writePNG    = FALSE,
#'                  plotR       = TRUE,
#'                  createDir   = FALSE,
#'                  pngMaxPix   = 1000,
#'                  linecol     = c("black", "blue"),
#'                  linewidth   = c(5,3),
#'                  linetype    = c(1, 2),
#'                  olapcol     = "darkgrey",
#'                  add.rug     = TRUE,
#'                  extend      = "lightgrey",
#'                  ylim        = c(0, 0.25),
#'                  main        = paste("Activity overlap between ", 
#'                                      speciesA_for_activity, "and", 
#'                                      speciesB_for_activity)
#' )
#' }
#' 
#' @export activityOverlap
#' 
activityOverlap <- function(recordTable,
                            speciesA,
                            speciesB,
                            speciesCol = "Species",
                            recordDateTimeCol = "DateTimeOriginal",
                            recordDateTimeFormat = "ymd HMS",
                            plotR = TRUE,
                            writePNG = FALSE,
                            addLegend = TRUE,
                            legendPosition = "topleft",
                            plotDirectory,
                            createDir = FALSE,
                            pngMaxPix = 1000,
                            add.rug = TRUE,
                            overlapEstimator = c("Dhat1", "Dhat4", "Dhat5"),
                            ...){

  if (!requireNamespace("overlap", quietly = TRUE)) {
    stop("Please install the package overlap to run this function")
  }

  wd0  <- getwd()
  mar0 <- par()$mar
  on.exit(setwd(wd0))
  on.exit(par(mar = mar0), add = TRUE)

  overlapEstimator <- match.arg(overlapEstimator)
  
  checkForSpacesInColumnNames(speciesCol = speciesCol, recordDateTimeCol = recordDateTimeCol)
  recordTable <- dataFrameTibbleCheck(df = recordTable)
  
  if(!speciesCol %in% colnames(recordTable))        stop(paste('speciesCol = "', speciesCol, '" is not a column name in recordTable', sep = ''), call. = FALSE)
  if(!recordDateTimeCol %in% colnames(recordTable)) stop(paste('recordDateTimeCol = "', recordDateTimeCol, '" is not a column name in recordTable', sep = ''), call. = FALSE)
  
  stopifnot(is.logical(c(writePNG, plotR, createDir)))
  stopifnot(hasArg(speciesA))
  stopifnot(hasArg(speciesB))
  stopifnot(all(c(speciesA, speciesB) %in% recordTable[,speciesCol]))
  
  timeZone <- "UTC"

  recordTable$DateTime2 <- parseDateTimeObject(inputColumn = recordTable[,recordDateTimeCol],
                                               dateTimeFormat = recordDateTimeFormat,
                                               timeZone = timeZone)

  recordTable$Time2 <-   format(recordTable$DateTime2, format = "%H:%M:%S", usetz = FALSE)

  # convert time to radians
  recordTable$Time.rad <- (as.numeric(as.POSIXct(strptime(recordTable$Time2, format = "%H:%M:%S", tz = timeZone))) -
                             as.numeric(as.POSIXct(strptime("0", format = "%S", tz = timeZone)))) / 3600 * (pi/12)

  subset_speciesA <- recordTable[recordTable[,speciesCol] == speciesA,]
  subset_speciesB <- recordTable[recordTable[,speciesCol] == speciesB,]
  
  if(nrow(subset_speciesA) == 1) stop("speciesA has only 1 record. Cannot estimate density.")
  if(nrow(subset_speciesB) == 1) stop("speciesB has only 1 record. Cannot estimate density.")
  
  if(any(is.na( c(subset_speciesA$Time.rad, subset_speciesB$Time.rad))))stop("NAs produced in converting to radial time. Sorry, that's a bug. Please report it.")

  n_record_string <- paste("number of records:", paste(nrow(subset_speciesA), nrow(subset_speciesB), sep = " / "))
  
   
  # set graphics  parameters and out directory
  overlapEst.tmp <- overlap::overlapEst(A = subset_speciesA$Time.rad, B = subset_speciesB$Time.rad, 
                                        type = overlapEstimator)

  dhat.tmp <- paste(names(overlapEst.tmp),
                    round(overlapEst.tmp, digits = 2), sep = "=")
  cex.sub      <- 0.75
  pngWidth     <- pngMaxPix
  pngHeight    <- round(pngMaxPix * 0.8)
  mar.tmp      <-  c(7.1, 4.1, 4.1, 2.1)
  y_usr_factor <- 0.22

  # extract line type, color and width for use in legend
  dots <- list(...)
  if(!is.null(dots[['linetype']])){ lty.tmp <- dots[['linetype']]}  else {lty.tmp <- c(1, 2)}
  if(!is.null(dots[['linewidth']])){lwd.tmp <- dots[['linewidth']]} else {lwd.tmp <- c(1, 1)}
  if(!is.null(dots[['linecol']])){  col.tmp <- dots[['linecol']]}   else {col.tmp <- c("black", "blue")}
  if(!is.null(dots[['main']])){     main.tmp <- dots[['main']]}   else {main.tmp <- paste("Activity overlap: ", 
                                                                                          substitute(speciesA), "and", 
                                                                                          substitute(speciesB))}

# check that plot directory exists, create it if needed and desired, and set working directory
  if(isTRUE(writePNG)){
    if(is.null(plotDirectory) == FALSE){
      if(isTRUE(createDir)){
        dir.create(plotDirectory, recursive = TRUE, showWarnings = FALSE)
        setwd(plotDirectory)
      } else {
        stopifnot(file.exists(plotDirectory))
        setwd(plotDirectory)
      }
    } else {stop("writePNG is TRUE. Please set plotDirectory")}
  }

  if(isTRUE(writePNG)){

    png(filename = paste("activity_overlap_", speciesA, "-", speciesB, "_", Sys.Date(), ".png", sep = ""),
        width    = pngWidth, 
        height   = pngHeight, 
        units    = "px", 
        res      = 96, 
        type     = "cairo")

    par(mar = mar.tmp)
    overlap::overlapPlot(A    = subset_speciesA$Time.rad, 
                B    = subset_speciesB$Time.rad,
                rug  = add.rug,
                main = main.tmp,
                ...)
    
    legend( x      = "top",
            legend = dhat.tmp,
            bty    = "n",
            cex    = cex.sub)

    if(isTRUE(addLegend)){
      legend(x      = legendPosition,
             legend = c(speciesA, speciesB),
             lty    = lty.tmp,
             lwd    = lwd.tmp,
             col    = col.tmp,
             bg     = "white"
      )
    }
    mtext(n_record_string, side = 3, line = 0)            
    dev.off()
  }

  if(isTRUE(plotR)){

    par(mar = mar.tmp)
    plot.values <- overlap::overlapPlot(A = subset_speciesA$Time.rad, 
                               B = subset_speciesB$Time.rad,
                               rug = add.rug,
                               main = main.tmp,
                               ...)

    legend(x      = "top",
           legend = dhat.tmp,
           bty    = "n",
           cex    = cex.sub)

    if(isTRUE(addLegend)){
      legend(x      = legendPosition,
             legend = c(speciesA, speciesB),
             lty    = lty.tmp,
             lwd    = lwd.tmp,
             col    = col.tmp,
             bg     = "white"
      )
    }
    mtext(n_record_string, side = 3, line = 0)            
    
    return(invisible(plot.values))
  }
}

data(recordTableSample)

# --- Tests ---
context("Activity Plotting Functions")

#############################################
# Tests for activityRadial()
#############################################
testthat::describe("activityRadial()", {
  
  test_that("Input validation works correctly", {
    expect_error(activityRadial(recordTableSample, species = "PBE", speciesCol = "WRONG_COL"),
                 "is not a column name in recordTable")
    expect_error(activityRadial(recordTableSample, species = "PBE", recordDateTimeCol = "WRONG_COL"),
                 "is not a column name in recordTable")
    expect_error(activityRadial(recordTableSample, species = "UNKNOWN_SPECIES"),
                 "species %in% recordTable\\[, speciesCol\\] is not TRUE")
    expect_error(activityRadial(recordTableSample, species = "PBE", writePNG = TRUE), # missing plotDirectory
                 "writePNG is TRUE. Please set plotDirectory")
  })
  
  test_that("Output for a single species is correct", {
    output <- activityRadial(recordTableSample, species = "PBE", plotR = FALSE)
    expect_s3_class(output, "data.frame")
    expect_equal(nrow(output), 24)
    expect_named(output, c("hour", "n", "perc", "radial.pos", "length4plot"))
    expect_equal(sum(output$n), 18) # 18 records for PBE
    expect_equal(sum(output$perc), 1)
    
    # Check byNumber = FALSE (default): length4plot should be proportions
    expect_equal(output$length4plot, output$perc)
    
    # Check byNumber = TRUE: length4plot should be counts
    output_by_number <- activityRadial(recordTableSample, species = "PBE", byNumber = TRUE, plotR = FALSE)
    expect_equal(output_by_number$length4plot, output_by_number$n)
  })
  
  test_that("Output for all species is correct", {
    output_list <- activityRadial(recordTableSample, allSpecies = TRUE, plotR = FALSE)
    species_in_data <- unique(recordTableSample$Species)
    
    expect_type(output_list, "list")
    expect_named(output_list, species_in_data)
    expect_length(output_list, length(species_in_data))
    
    # Check structure and counts of an element in the list
    expect_s3_class(output_list$VTA, "data.frame")
    expect_equal(nrow(output_list$VTA), 24)
    expect_equal(sum(output_list$VTA$n), 5) # 12 records for VTA
  })
  
  test_that("File writing works as expected", {
    # Using a temporary directory for safe file operations
    withr::with_tempdir({
      # Test 1: createDir = TRUE successfully creates directory
      plot_dir <- "radial_plots"
      expect_false(dir.exists(plot_dir))
      activityRadial(recordTableSample, allSpecies = TRUE, writePNG = TRUE, plotDirectory = plot_dir, createDir = TRUE, plotR = FALSE)
      expect_true(dir.exists(plot_dir))
      expect_equal(length(list.files(plot_dir, pattern = "\\.png$")), 5) # One plot per species, plut richness
      
      # Test 2: Errors when directory doesn't exist and createDir = FALSE
      plot_dir_2 <- "non_existent_dir"
      expect_error(
        activityRadial(recordTableSample, species = "PBE", writePNG = TRUE, plotDirectory = plot_dir_2, createDir = FALSE),
        "plotDirectory does not exist"
      )
    })
  })
})


#############################################
# Tests for activityHistogram()
#############################################
testthat::describe("activityHistogram()", {
  
  test_that("Input validation works correctly", {
    expect_error(activityHistogram(recordTableSample, species = "VTA", speciesCol = "WRONG_COL"),
                 "is not a column name in recordTable")
    expect_error(activityHistogram(recordTableSample, species = "VTA", recordDateTimeCol = "WRONG_COL"),
                 "is not a column name in recordTable")
    expect_error(activityHistogram(recordTableSample, species = "UNKNOWN_SPECIES"),
                 "species %in% recordTable\\[, speciesCol\\] is not TRUE")
    expect_error(activityHistogram(recordTableSample, species = "VTA", writePNG = TRUE), # missing plotDirectory
                 "writePNG is TRUE. Please set plotDirectory")
  })
  
  test_that("Output for a single species is correct", {
    output <- activityHistogram(recordTableSample, species = "VTA", plotR = FALSE)
    expect_s3_class(output, "POSIXct")
    expect_length(output, 5) # 12 records for VTA
  })
  
  test_that("Output for all species is correct", {
    output_list <- activityHistogram(recordTableSample, allSpecies = TRUE, plotR = FALSE)
    species_in_data <- unique(recordTableSample$Species)
    
    expect_type(output_list, "list")
    expect_named(output_list, species_in_data)
    expect_length(output_list, length(species_in_data))
    
    # Check structure and length of one element
    expect_s3_class(output_list$PBE, "POSIXct")
    expect_length(output_list$PBE, 18)
  })
})


#############################################
# Tests for activityDensity()
#############################################
testthat::describe("activityDensity()", {
  
  skip_if_not_installed("overlap")
  
  test_that("Input validation works correctly", {
    expect_error(activityDensity(recordTableSample, species = "PBE", speciesCol = "WRONG_COL"),
                 "is not a column name in recordTable")
    expect_error(activityDensity(recordTableSample, species = "UNKNOWN_SPECIES"),
                 "species %in% recordTable\\[, speciesCol\\] is not TRUE")
    # Should fail for species with only 1 record
    expect_error(activityDensity(recordTableSample[1,], species = "PBE", plotR = FALSE),
                 "had only 1 record. Cannot estimate density.")
    expect_error(activityDensity(recordTableSample, species = "PBE", writePNG = TRUE), # missing plotDirectory
                 "writePNG is TRUE. Please set plotDirectory")
  })
  
  test_that("Output for a single species is correct (times in radians)", {
    output <- activityDensity(recordTableSample, species = "PBE", plotR = FALSE)
    expect_type(output, "double")
    expect_length(output, 18)
    expect_true(all(output >= 0 & output <= 2 * pi))
  })
  
  test_that("Output for all species is correct (with warnings for skipped species)", {
    # It should warn about the species with 1 record and skip it
    expect_warning(
      output_list <- activityDensity(recordTableSample[-which(recordTableSample$Species == "MNE")[2],], allSpecies = TRUE, plotR = FALSE),
      "MNE: It had only 1 record. Cannot estimate density.   - SKIPPED"
    )
    
    # The returned list contains NULL for the skipped species, so we filter it out
    output_list_clean <- output_list[!sapply(output_list, is.null)]
    species_with_enough_data <- c("PBE", "VTA", "EGY", "TRA")
    
    expect_type(output_list, "list")
    expect_named(output_list_clean, species_with_enough_data)
    expect_length(output_list_clean, 4)
    
    # Check structure of one element
    expect_type(output_list_clean$VTA, "double")
    expect_length(output_list_clean$VTA, 5)
  })
})


#############################################
# Tests for activityOverlap()
#############################################
testthat::describe("activityOverlap()", {
  
  skip_if_not_installed("overlap")
  
  test_that("Input validation works correctly", {
    expect_error(activityOverlap(recordTableSample, speciesA = "PBE", speciesB = "VTA", speciesCol = "WRONG_COL"),
                 "is not a column name in recordTable")
    expect_error(activityOverlap(recordTableSample, speciesA = "UNKNOWN", speciesB = "VTA"),
                 "all\\(c\\(speciesA, speciesB\\) %in% recordTable\\[, speciesCol\\]\\) is not TRUE")
    expect_error(activityOverlap(recordTableSample[-which(recordTableSample$Species == "MNE")[2],], speciesA = "PBE", speciesB = "MNE"),
                 "speciesB has only 1 record. Cannot estimate density.")
    
    # Check for invalid overlap estimator
    expect_error(activityOverlap(recordTableSample, speciesA = "PBE", speciesB = "VTA", overlapEstimator = "Dhat_WRONG"),
                 "'arg' should be one of")
  })
  
  test_that("Output is a data.frame with correct structure when plotR = TRUE", {
    # Set plotR = TRUE to ensure overlapPlot is called and returns its data
    plot_data <- activityOverlap(recordTableSample, speciesA = "PBE", speciesB = "VTA", plotR = TRUE)
    
    expect_s3_class(plot_data, "data.frame")
    expect_named(plot_data, c("x", "densityA", "densityB"))
    expect_gt(nrow(plot_data), 0)
  })
  
  test_that("Function returns NULL when plotR is FALSE and writePNG is FALSE", {
    # The return value is inside an `if(plotR)` block, so it should be NULL otherwise.
    output <- activityOverlap(recordTableSample, speciesA = "PBE", speciesB = "VTA", plotR = FALSE, writePNG = FALSE)
    expect_null(output)
  })
  
  test_that("File writing works as expected", {
    withr::with_tempdir({
      plot_dir <- "overlap_plots"
      activityOverlap(recordTableSample, 
                      speciesA = "PBE", 
                      speciesB = "VTA", 
                      writePNG = TRUE, 
                      plotDirectory = plot_dir, 
                      createDir = TRUE, 
                      plotR = FALSE) # Set plotR to FALSE as we only test writing
      
      expect_true(dir.exists(plot_dir))
      files <- list.files(plot_dir, pattern = "\\.png$")
      expect_equal(length(files), 1)
      expect_true(grepl("PBE-VTA", files[1]))
    })
  })
})

#' Fix DateTimeOriginal Exif metadata tag in Reconyx Hyperfire cameras
#' 
#' Some camera models don't store the date/time information in the standard
#' Exif metadata tag. Consequently, camtrapR cannot find that information. This
#' function uses Exiftool to update the DateTimeOriginal metadata tag in all
#' images within a directory to make them readable with camtrapR (and other
#' software).
#' 
#' Some Reconyx Hyperfire cameras (e.g. HC500) are known to show this problem.
#' 
#' @param inDir character. Name of the directory containing images to be fixed
#' @param recursive logical. Recursively find images in subdirectories of
#' \code{inDir}?
#' 
#' @return Returns invisibly the messages returned by the Exiftool call
#' (warnings etc.).
#' 
#' @section Warning : Please make a backup of your images before running this
#' function.
#' 
#' @author Juergen Niedballa
#' 
#' @references This function uses the code from: \cr Tobler, Mathias (2015).
#' Camera Base Version 1.7 User Guide
#' 
#' @examples
#' 
#' \dontrun{
#' # a hypothetical example
#' 
#' wd_images_hyperfire <- "C:/Some/Directory"
#' 
#' fixDateTimeOriginal(inDir     = wd_images_hyperfire,
#'                     recursive = TRUE)
#' }
#' 
#' @export fixDateTimeOriginal
#' 
fixDateTimeOriginal <- function(inDir,
                                recursive = TRUE){

  if(Sys.which("exiftool") == "") stop("cannot find ExifTool", call. = FALSE)
  stopifnot(is.logical(recursive))
  if(!dir.exists(inDir)) stop("Could not find inDir:\n", inDir, call. = FALSE)

  command.tmp  <- paste('exiftool "-DateTimeOriginal>DateTimeOriginal" ',  ifelse(recursive, '-r ', '') , '-overwrite_original -ext JPG "', inDir, '"', sep = '')

  tmp <- system(command.tmp,  intern=TRUE)

  message(tmp[c(length(tmp) - 1, length(tmp))])

  return(invisible(tmp))
}

.onAttach <- function(...){
  # adapted from http://thecoatlessprofessor.com/programming/automatically-check-if-r-package-is-the-latest-version-on-package-load/. Thank you!

  # Avoid running if in batch job / user not present
  if (!interactive()) return()
  
  # Obtain the installed package information
  local_version = utils::packageDescription('camtrapR')
  
  # Grab the package information from CRAN
  cran_version = .pkgVersionCRAN("camtrapR")     # function defined in "variousOtherHelperFunctions.R"
  
  # Verify we have package information
  if(!is.null(cran_version) && length(cran_version) != 0L){
    latest_version = utils::compareVersion(cran_version, local_version$Version)
    
    d = if(latest_version == 0){
      'CURRENT'
    }else if(latest_version == 1){
      'OUT OF DATE'
    }else{
      'DEVELOPMENT'
    }
  
  }else{ # Gracefully fail.
   d2 = "cannot connect to CRAN for package version check"
   latest_version = 0
  }
  
  # Use packageStartUpMessages() so that folks can suppress package messages with 
  # suppressPackageStartupMessages(library(pkg))
  
  if(exists("d"))  packageStartupMessage('This is camtrapR ', local_version$Version, ' (', d,')')  # built on ', local_version$Date)
  if(exists("d2")) packageStartupMessage('This is camtrapR ', local_version$Version, '    (', d2,')')  # built on ', local_version$Date)
  
  
  if(latest_version == 1){
    packageStartupMessage('\n NEW VERSION AVAILABLE ONLINE: ', cran_version , ' !')
    packageStartupMessage('Download it from CRAN via:    install.packages("camtrapR")\n')
  }
}

#' Plot histogram of single-species activity
#' 
#' The function generates a histogram of species diel activity in 1-hour
#' intervals.
#' 
#' Activity is calculated from the time of day of records. The date is ignored.
#' 
#' \code{recordDateTimeFormat} defaults to the "YYYY-MM-DD HH:MM:SS"
#' convention, e.g. "2014-09-30 22:59:59". \code{recordDateTimeFormat} can be
#' interpreted either by base-R via \code{\link[base]{strptime}} or in
#' \pkg{lubridate} via \code{\link[lubridate]{parse_date_time}} (argument
#' "orders"). \pkg{lubridate} will be used if there are no "\%" characters in
#' \code{recordDateTimeFormat}.
#' 
#' For "YYYY-MM-DD HH:MM:SS", \code{recordDateTimeFormat} would be either
#' "\%Y-\%m-\%d \%H:\%M:\%S" or "ymd HMS". For details on how to specify date
#' and time formats in R see \code{\link[base]{strptime}} or
#' \code{\link[lubridate]{parse_date_time}}.
#' 
#' @param recordTable data.frame. the record table created by
#' \code{\link{recordTable}}
#' @param species Name of the single species for which to create a histogram of
#' activity
#' @param allSpecies logical. Create plots for all species in \code{speciesCol}
#' of \code{recordTable}? Overrides argument \code{species}
#' @param speciesCol character. name of the column specifying species names in
#' \code{recordTable}
#' @param recordDateTimeCol character. name of the column specifying date and
#' time in \code{recordTable}
#' @param recordDateTimeFormat character. format of column
#' \code{recordDateTimeCol} in \code{recordTable}
#' @param plotR logical. Show plots in R graphics device?
#' @param writePNG logical. Create pngs of the plots?
#' @param plotDirectory character. Directory in which to create png plots if
#' \code{writePNG = TRUE}
#' @param createDir logical. Create \code{plotDirectory}?
#' @param pngMaxPix integer. image size of png (pixels along x-axis)
#' @param \dots additional arguments to be passed to function
#' \code{\link[graphics]{hist}}
#' @return It returns invisibly a vector of species record date and time in
#' \code{POSIXlt} format. If \code{allSpecies == TRUE}, all species' vectors
#' are returned in an invisible named list.
#' 
#' @note If you have a sufficiently large number of records you may wish to
#' consider using \code{\link{activityDensity}} instead. Please be aware that
#' this function (like the other activity... function of this package) use
#' clock time. If your survey was long enough to see changes in sunrise and
#' sunset times, this may result in biased representations of species activity.
#' 
#' @author Juergen Niedballa
#' 
#' @seealso \code{\link{activityDensity}}, \code{\link{activityRadial}},
#' \code{\link{activityOverlap}}
#' 
#' @examples
#' 
#' 
#' # load record table
#' data(recordTableSample)
#' 
#' # generate activity histogram
#' species4activity <- "VTA"    # = Viverra tangalunga, Malay Civet
#' 
#' activityHistogram (recordTable = recordTableSample,
#'                    species     = species4activity,
#'                    allSpecies = FALSE)
#' 
#' @export activityHistogram
#' 
activityHistogram <- function(recordTable,
                              species,
                              allSpecies = FALSE,
                              speciesCol = "Species",
                              recordDateTimeCol = "DateTimeOriginal",
                              recordDateTimeFormat = "ymd HMS",
                              plotR = TRUE,
                              writePNG = FALSE,
                              plotDirectory,
                              createDir = FALSE,
                              pngMaxPix = 1000,
                              ...){

  wd0 <- getwd()
  mar0 <- par()$mar
  on.exit(setwd(wd0))
  on.exit(par(mar = mar0), add = TRUE)

  checkForSpacesInColumnNames(speciesCol = speciesCol, recordDateTimeCol = recordDateTimeCol)
  recordTable <- dataFrameTibbleCheck(df = recordTable)
  
  if(!speciesCol %in% colnames(recordTable))           stop(paste('speciesCol = "', speciesCol, '" is not a column name in recordTable', sep = ''), call. = FALSE)
  if(!recordDateTimeCol %in% colnames(recordTable)) stop(paste('recordDateTimeCol = "', recordDateTimeCol, '" is not a column name in recordTable', sep = ''), call. = FALSE)
  
  stopifnot(is.logical(c(allSpecies, writePNG, plotR, createDir)))
  if(allSpecies == FALSE){
    stopifnot(hasArg(species))
    stopifnot(species %in% recordTable[,speciesCol])
  }

  timeZone <- "UTC"
  
  recordTable$DateTime2 <- parseDateTimeObject(inputColumn = recordTable[,recordDateTimeCol],
                                               dateTimeFormat = recordDateTimeFormat,
                                               timeZone = timeZone)

  recordTable$Hour <- as.POSIXlt(recordTable$DateTime2)$hour

  # set graphics  parameters and out directory
  col_bars <- "gray"
  hist_breaks <-  seq(0,24,1)  # 24
  xlab.tmp = "Time of Day [h]"

  pngWidth <- pngMaxPix
  pngHeight <- round(pngMaxPix * 0.8)

  if(isTRUE(writePNG)){
    if(hasArg(plotDirectory)){
      if(isTRUE(createDir)){
        dir.create(plotDirectory, recursive = TRUE, showWarnings = FALSE)
        setwd(plotDirectory)
      } else {
        stopifnot(file.exists(plotDirectory))
        setwd(plotDirectory)
      }
    } else {stop("writePNG is TRUE. Please set plotDirectory")}
  }

  if(allSpecies == FALSE){
    subset_species <- recordTable[recordTable[,speciesCol] == species,]
    
    subset_species$Hour <- subset_species$Hour + 0.1   # otherwise both 0 and 1 will be in histogram class 0

    if(isTRUE(writePNG)){
      png(filename = paste("activity_histogram_", species, "_", Sys.Date(), ".png", sep = ""),
          width = pngWidth, height = pngHeight, units = "px", res = 96, type = "cairo")
      hist(subset_species$Hour, breaks = hist_breaks,
           col  = col_bars,
           main = paste("Activity of", species),
           xlab = xlab.tmp,
           axes = FALSE,
           ...)
      axis(1, at = seq(0,24, by = 3))
      axis(2)
      box()
      mtext(paste("number of records:", length(subset_species$Hour)), side = 3, line = 0)
      dev.off()
    }

    if(isTRUE(plotR)){
      hist(subset_species$Hour,
           breaks = hist_breaks,
           col = col_bars,
           freq = TRUE,
           main = paste("Activity of", species),
           xlab = xlab.tmp,
           axes = FALSE,
           ...)
      axis(1, at = seq(0,24, by = 3))
      axis(2)
      box()
      mtext(paste("number of records:", length(subset_species$Hour)), side = 3, line = 0)
    }


  } else {

    subset_species_list <- list()

    for(i in 1:length(unique(recordTable[,speciesCol]))){

      spec.tmp <- unique(recordTable[,speciesCol])[i]
      subset_species <- subset(recordTable, recordTable[,speciesCol] == spec.tmp)

      if(isTRUE(writePNG)){
        png(filename = paste("activity_histogram_", spec.tmp, "_", Sys.Date(), ".png", sep = ""),
            width = pngWidth, height = pngHeight, units = "px", res = 96, type = "cairo")
        hist(subset_species$Hour,
             breaks = hist_breaks,
             col  = col_bars,
             main = paste("Activity of", spec.tmp),
             xlab = xlab.tmp,
             axes = FALSE,
             ...)
        axis(1, at = seq(0,24, by = 3))
        axis(2)
        box()
        mtext(paste("number of records:", length(subset_species$Hour)), side = 3, line = 0)
        dev.off()
      }

      if(isTRUE(plotR)){
        hist(subset_species$Hour,
             breaks = hist_breaks,
             col  = col_bars,
             main = paste("Activity of", spec.tmp),
             xlab = xlab.tmp,
             axes = FALSE,
             ...)
        axis(1, at = seq(0,24, by = 3))
        axis(2)
        box()
        mtext(paste("number of records:", length(subset_species$Hour)), side = 3, line = 0)
      }
      subset_species_list[[i]] <- subset_species$DateTime2
      names(subset_species_list)[i] <- spec.tmp
    }
  }

  if(allSpecies == FALSE){
    return(invisible(subset_species$DateTime2))
  } else {
    return(invisible(subset_species_list))
  }
}

#' Plot kernel density estimation of single-species activity
#' 
#' The function plots a kernel density estimation of species diel activity
#' using function \code{\link[overlap]{densityPlot}} from package
#' \pkg{overlap}.
#' 
#' \code{species} must be in the \code{speciesCol} of \code{recordTable}.
#' 
#' \code{recordDateTimeFormat} defaults to the "YYYY-MM-DD HH:MM:SS"
#' convention, e.g. "2014-09-30 22:59:59". \code{recordDateTimeFormat} can be
#' interpreted either by base-R via \code{\link[base]{strptime}} or in
#' \pkg{lubridate} via \code{\link[lubridate]{parse_date_time}} (argument
#' "orders"). \pkg{lubridate} will be used if there are no "\%" characters in
#' \code{recordDateTimeFormat}.
#' 
#' For "YYYY-MM-DD HH:MM:SS", \code{recordDateTimeFormat} would be either
#' "\%Y-\%m-\%d \%H:\%M:\%S" or "ymd HMS". For details on how to specify date
#' and time formats in R see \code{\link[base]{strptime}} or
#' \code{\link[lubridate]{parse_date_time}}.
#' 
#' @param recordTable data.frame. the record table created by
#' \code{\link{recordTable}}
#' @param species Name of the species for which to create an kernel density
#' plot of activity
#' @param allSpecies logical. Create plots for all species in \code{speciesCol}
#' of \code{recordTable}? Overrides argument \code{species}
#' @param speciesCol character. name of the column specifying species names in
#' \code{recordTable}
#' @param recordDateTimeCol character. name of the column specifying date and
#' time in \code{recordTable}
#' @param recordDateTimeFormat character. format of column
#' \code{recordDateTimeCol} in \code{recordTable}
#' @param plotR logical. Show plots in R graphics device?
#' @param writePNG logical. Create pngs of the plots?
#' @param plotDirectory character. Directory in which to create png plots if
#' \code{writePNG = TRUE}
#' @param createDir logical. Create \code{plotDirectory} if \code{writePNG =
#' TRUE}?
#' @param pngMaxPix integer. image size of png (pixels along x-axis)
#' @param add.rug logical. add a rug to the plot?
#' @param \dots additional arguments to be passed to function
#' \code{densityPlot}
#' 
#' @return Returns invisibly a vector of species record observation times in
#' radians, i.e. scaled to [\eqn{0, 2\pi}]. If \code{allSpecies == TRUE}, all
#' species' vectors are returned in an invisible named list.
#' 
#' @author Juergen Niedballa
#' 
#' @seealso \code{\link{activityHistogram}}, \code{\link{activityRadial}},
#' \code{\link{activityOverlap}}
#' \url{https://www.kent.ac.uk/smsas/personal/msr/overlap.html}
#' 
#' @references Martin Ridout and Matthew Linkie (2009). Estimating overlap of
#' daily activity patterns from camera trap data. Journal of Agricultural,
#' Biological and Environmental Statistics, 14(3), 322-337 \cr Mike Meredith
#' and Martin Ridout (2018). overlap: Estimates of coefficient of overlapping
#' for animal activity patterns. R package version 0.3.2.
#' \url{https://CRAN.R-project.org/package=overlap}
#' 
#' @examples
#' 
#' if(requireNamespace("overlap")) {
#' # load record table
#' data(recordTableSample)
#' 
#' species4activity <- "VTA"    # = Viverra tangalunga, Malay Civet
#' 
#' activityDensity(recordTable = recordTableSample,
#'                 species     = species4activity)
#' 
#' 
#' # all species at once
#' 
#' activityDensity(recordTable = recordTableSample,
#'                 allSpecies  = TRUE,
#'                 writePNG    = FALSE,
#'                 plotR       = TRUE,
#'                 add.rug     = TRUE)
#'  }
#' 
#' @export activityDensity
#' 
activityDensity <- function(recordTable,
                            species,
                            allSpecies = FALSE,
                            speciesCol = "Species",
                            recordDateTimeCol = "DateTimeOriginal",
                            recordDateTimeFormat = "ymd HMS",
                            plotR = TRUE,
                            writePNG = FALSE,
                            plotDirectory,
                            createDir = FALSE,
                            pngMaxPix = 1000,
                            add.rug = TRUE,
                            ...
){
  
  if (!requireNamespace("overlap", quietly = TRUE)) {
    stop("Please install the package overlap to run this function")
  }
  
  wd0 <- getwd()
  mar0 <- par()$mar
  on.exit(setwd(wd0))
  on.exit(par(mar = mar0), add = TRUE)

  recordTable <- dataFrameTibbleCheck(df = recordTable)
  
  timeZone <- "UTC"
  
  checkForSpacesInColumnNames(speciesCol = speciesCol, recordDateTimeCol = recordDateTimeCol)
  if(!is.data.frame(recordTable)) stop("recordTable must be a data frame", call. = FALSE)
  if(!speciesCol %in% colnames(recordTable))           stop(paste('speciesCol = "', speciesCol, '" is not a column name in recordTable', sep = ''), call. = FALSE)
  if(!recordDateTimeCol %in% colnames(recordTable)) stop(paste('recordDateTimeCol = "', recordDateTimeCol, '" is not a column name in recordTable', sep = ''), call. = FALSE)
  
  stopifnot(is.logical(c(allSpecies, writePNG, plotR, createDir)))
  if(allSpecies == FALSE) {
    stopifnot(species %in% recordTable[,speciesCol])
    stopifnot(hasArg(species))
  }
  
  
  recordTable$DateTime2 <- parseDateTimeObject(inputColumn = recordTable[,recordDateTimeCol],
                                               dateTimeFormat = recordDateTimeFormat,
                                               timeZone = timeZone)
  recordTable$Time2 <- format(recordTable$DateTime2, format = "%H:%M:%S", usetz = FALSE)
  
  
  # radians time
  recordTable$Time.rad <- (as.numeric(as.POSIXct(strptime(recordTable$Time2, format = "%H:%M:%S", tz = timeZone))) -
                             as.numeric(as.POSIXct(strptime("0", format = "%S", tz = timeZone)))) / 3600 * (pi/12)
  
  if(isTRUE(writePNG)){
    if(hasArg(plotDirectory)){
      if(isTRUE(createDir)){
        dir.create(plotDirectory, recursive = TRUE, showWarnings = FALSE)
        setwd(plotDirectory)
      } else {
        stopifnot(file.exists(plotDirectory))
        setwd(plotDirectory)
      }
    } else {
      stop("writePNG is TRUE. Please set plotDirectory", call. = FALSE)}
  }
  
  pngWidth <- pngMaxPix
  pngHeight <- round(pngMaxPix * 0.8)
  
  if(allSpecies == FALSE){
    
    subset_species <- recordTable[recordTable[,speciesCol] == species,]

    if(nrow(subset_species) == 1)  stop(paste(species, "had only 1 record. Cannot estimate density."), call. = FALSE)
    
    try_error_tmp <- try({
      if(isTRUE(writePNG))  png(filename = paste("activity_density_", species, "_", Sys.Date(), ".png", sep = ""),
                                width = pngWidth, height = pngHeight, units = "px", res = 96, type = "cairo")
      if(isTRUE(writePNG) | isTRUE(plotR)){ overlap::densityPlot(subset_species$Time.rad,
                                                        main = paste("Activity of", species),
                                                        rug  = add.rug,
                                                        ...)
        mtext(paste("number of records:", nrow(subset_species)), side = 3, line = 0)}
      if(isTRUE(writePNG))  dev.off()
    }, silent = TRUE)
    if(inherits(try_error_tmp, "try-error")) warning(paste(toupper(species), ": ", try_error_tmp[1], "    - SKIPPED", sep = ""), call. = FALSE)
    
  } else {
    
    subset_species_list <- list()
    
    for(i in 1:length(unique(recordTable[,speciesCol]))){
      
      spec.tmp <- unique(recordTable[,speciesCol])[i]
      subset_species <- subset(recordTable, recordTable[,speciesCol] == spec.tmp)
      plot_main_title <- paste("Activity of", spec.tmp)
      
      if(nrow(subset_species) == 1){
        warning(paste(toupper(spec.tmp), ": It had only 1 record. Cannot estimate density.   - SKIPPED", sep = ""), call. = FALSE)
        next
      } else {
        
        try_error_tmp <- try({
          if(isTRUE(writePNG))  png(filename = paste("activity_density_", spec.tmp, "_", Sys.Date(), ".png", sep = ""),
                                    width = pngWidth, height = pngHeight, units = "px", res = 96, type = "cairo")
          if(isTRUE(writePNG) | isTRUE(plotR)){ overlap::densityPlot(subset_species$Time.rad,
                                                            main = plot_main_title,
                                                            rug  = add.rug,
                                                            ...)
            mtext(paste("number of records:", nrow(subset_species)), side = 3, line = 0)
          }
          if(isTRUE(writePNG))  dev.off()
        }, silent = TRUE)
        if(inherits(try_error_tmp, "try-error")) warning(paste(toupper(spec.tmp), ": ", try_error_tmp[1], "    - SKIPPED", sep = ""), call. = FALSE)
      }
      
      
      subset_species_list[[i]] <- subset_species$Time.rad
      names(subset_species_list)[i] <- spec.tmp
    }
  }
  
  if(allSpecies == FALSE){
    return(invisible(subset_species$Time.rad))
  } else {
    return(invisible(subset_species_list))
  }
}

#' Generate maps of observed species richness and species presences by station
#' 
#' Generates maps of observed species richness and species presence by species
#' and station. Output can be R graphics, PNG graphics or a shapefile for use
#' in GIS software.
#' 
#' The column name \code{stationCol} must be identical in \code{CTtable} and
#' \code{recordTable} and station IDs must match.
#' 
#' Shapefile creation depends on the packages \pkg{sf}.
#' Argument \code{shapefileProjection} must be a valid argument of
#' \code{\link[sf]{st_crs}} (one of (i) character: a string accepted by GDAL, 
#' (ii) integer, a valid EPSG value (numeric), or (iii) an object of class crs.
#' If \code{shapefileProjection} is undefined,
#' the resulting shapefile will lack a coordinate reference system.
#' 
#' @param CTtable data.frame. contains station IDs and coordinates
#' @param Xcol character. name of the column specifying x coordinates in
#' \code{CTtable}
#' @param Ycol character. name of the column specifying y coordinates in
#' \code{CTtable}
#' @param backgroundPolygon SpatialPolygons or SpatialPolygonsDataFrame.
#' Polygon to be plotted in the background of the map (e.g. project area
#' boundary)
#' @param stationCol character. name of the column specifying station ID in
#' \code{CTtable} and \code{recordTable}
#' @param recordTable data.frame. the record table created by
#' \code{\link{recordTable}}
#' @param speciesCol character. name of the column specifying species in
#' \code{recordTable}
#' @param speciesToShow character. Species to include in the maps. If missing,
#' all species in \code{recordTable} will be included.
#' @param writePNG logical. Create PNGs of the plots?
#' @param plotR logical. Create plots in R graphics device?
#' @param plotDirectory character. Directory in which to save the PNGs
#' @param createPlotDir logical. Create \code{plotDirectory}?
#' @param richnessPlot logical. Generate a species richness plot?
#' @param speciesPlots logical. Generate plots of all species number of
#' independent events?
#' @param printLabels logical. Add station labels to the plots?
#' @param smallPoints numeric. Number by which to decrease point sizes in plots
#' (optional).
#' @param addLegend logical. Add legends to the plots?
#' @param pngMaxPix integer. number of pixels in pngs on the longer side
#' @param writeShapefile logical. Create a shapefile from the output?
#' @param shapefileName character. Name of the shapefile to be saved. If empty,
#' a name will be generated automatically.
#' @param shapefileDirectory character. Directory in which to save the
#' shapefile.
#' @param shapefileProjection character. A character string of projection
#' arguments to use in the shapefile.
#' 
#' @return An invisible \code{data.frame} with station coordinates, numbers of
#' events by species at each station and total species number by station. In
#' addition and optionally, R graphics or png image files.
#' 
#' @author Juergen Niedballa
#' 
#' @references A great resource for coordinate system information
#' is \url{https://spatialreference.org/}. Use the Proj4 string as
#' \code{shapefileProjection} argument.
#' 
#' @examples
#' 
#' 
#' # load station information
#' data(camtraps)
#' 
#' # load record table
#' data(recordTableSample)
#' 
#' 
#' # create maps
#' Mapstest <- detectionMaps(CTtable           = camtraps,
#'                           recordTable       = recordTableSample,
#'                           Xcol              = "utm_x",
#'                           Ycol              = "utm_y",
#'                           stationCol        = "Station",
#'                           speciesCol        = "Species",
#'                           writePNG          = FALSE,
#'                           plotR             = TRUE,
#'                           printLabels       = TRUE,
#'                           richnessPlot      = TRUE,
#'                           addLegend         = TRUE
#' )
#' 
#' 
#' 
#' # with a polygon in the background, and for one species only
#' 
#' # make a dummy polygon for the background
#' library(sf)
#' 
#' Sr1 = st_polygon(list(cbind(c(521500,526500,527000, 521500, 521500),
#'                             c(607500, 608000, 603500, 603500, 607500))))
#' aoi <- data.frame(name = "My AOI")
#' st_geometry(aoi) <- st_geometry(Sr1)
#' st_crs(aoi) <- 32650  # assign CRS: UTM50N
#' 
#' Mapstest2 <- detectionMaps(CTtable           = camtraps,
#'                            recordTable       = recordTableSample,
#'                            Xcol              = "utm_x",
#'                            Ycol              = "utm_y",
#'                            backgroundPolygon = aoi,               # this was added
#'                            speciesToShow     = c("PBE", "VTA"),   # this was added
#'                            stationCol        = "Station",
#'                            speciesCol        = "Species",
#'                            writePNG          = FALSE,
#'                            plotR             = TRUE,
#'                            printLabels       = TRUE,
#'                            richnessPlot      = TRUE,
#'                            addLegend         = TRUE
#' )
#' 
#' 
#' 
#' 
#' @importFrom sf st_as_sf st_set_crs st_write st_coordinates
#' @importFrom grDevices col2rgb dev.off extendrange heat.colors png rgb
#' @importFrom graphics abline axis box grconvertX grconvertY hist image legend lines mtext par plot plot.default  points polygon rect segments strheight strwidth text
#' @export detectionMaps
#' 
detectionMaps <- function(CTtable,
                          recordTable,
                          Xcol,
                          Ycol,
                          backgroundPolygon,
                          stationCol = "Station",
                          speciesCol = "Species",
                          speciesToShow,
                          richnessPlot = TRUE,
                          speciesPlots = TRUE,
                          addLegend = TRUE,
                          printLabels = FALSE,
                          smallPoints,
                          plotR = TRUE,
                          writePNG = FALSE,
                          plotDirectory,
                          createPlotDir = FALSE,
                          pngMaxPix = 1000,
                          writeShapefile = FALSE,
                          shapefileName,
                          shapefileDirectory,
                          shapefileProjection){

  wd0 <- getwd()
  opar <- par(no.readonly = TRUE)
  on.exit(setwd(wd0))
  on.exit(par(opar), add = TRUE)

  # check column names
  checkForSpacesInColumnNames(stationCol = stationCol, speciesCol = speciesCol, Xcol = Xcol, Ycol = Ycol)
  
  CTtable     <- dataFrameTibbleCheck(df = CTtable)
  recordTable <- dataFrameTibbleCheck(df = recordTable)

  if(!stationCol %in% colnames(CTtable))      stop(paste('stationCol = "', stationCol,     '" is not a column name in CTtable', sep = ''), call. = FALSE)
  if(!stationCol %in% colnames(recordTable))  stop(paste('stationCol = "', stationCol,     '" is not a column name in recordTable', sep = ''), call. = FALSE)
  if(!speciesCol %in% colnames(recordTable))  stop(paste('speciesCol = "', speciesCol,     '" is not a column name in recordTable', sep = ''), call. = FALSE)

  if(!Xcol %in% colnames(CTtable))            stop(paste('Xcol = "',   Xcol, '" is not a column name in CTtable', sep = ''), call. = FALSE)
  if(!Ycol %in% colnames(CTtable))            stop(paste('Ycol = "',   Ycol, '" is not a column name in CTtable', sep = ''), call. = FALSE)
  
  if(!is.numeric(CTtable[,Xcol]))              stop(paste('The values of Xcol "',   Xcol, '" must be numeric', sep = ''), call. = FALSE)
  if(!is.numeric(CTtable[,Ycol]))              stop(paste('The values of Ycol "',   Ycol, '" must be numeric', sep = ''), call. = FALSE)
  

  CTtable[,stationCol] <- as.character(CTtable[,stationCol])
  recordTable[,stationCol] <- as.character(recordTable[,stationCol])

  # check all stations in recordTable are matched in CTtable
  if(all(recordTable[,stationCol] %in% CTtable[,stationCol]) == FALSE) {
    stop(paste("items of stationCol in recordTable are not matched in stationCol of CTtable: ", paste(unique(recordTable[-which(recordTable[,stationCol] %in% CTtable[,stationCol]),stationCol]), collapse = ", ")))
  }

  recordTable[,speciesCol] <- as.character(recordTable[,speciesCol])

  if(any(is.na(CTtable[,Xcol])))stop("there are NAs in Xcol")
  if(any(is.na(CTtable[,Ycol])))stop("there are NAs in Ycol")
  if(any(is.na(CTtable[,stationCol])))stop("there are NAs in stationCol")

  if(writeShapefile == TRUE) {
    if (!requireNamespace("sf", quietly = TRUE)) {
      stop("the package 'sf' is needed for writing shapefiles.")
    } 
    
    stopifnot(hasArg(shapefileDirectory))
    stopifnot(file.exists(shapefileDirectory))
  }

  if(hasArg(backgroundPolygon)){
    if (!requireNamespace("sf", quietly = TRUE)) stop("package 'sf' is required to plot backgroundPolygon")
    stopifnot(inherits(backgroundPolygon, c("sf", "sfg")))
  }

  if(hasArg(speciesToShow)){
    stopifnot(is.character(speciesToShow))
    if(any(!speciesToShow %in% recordTable[,speciesCol])) stop(" these speciesToShow are not in speciesCol of recordTable:   ", paste(speciesToShow[!speciesToShow %in% recordTable[,speciesCol]], collapse = ", "), call. = FALSE)
    recordTable <- recordTable[recordTable[,speciesCol] %in% speciesToShow,]
  }

  # data preparation
  dat2 <- aggregate(CTtable[, c(Xcol, Ycol)], 
                    by = list(CTtable[,stationCol]), 
                    FUN = mean)    # get coordinates
  colnames(dat2)[1] <- stationCol

  # ensure that stations without records are not dropped (resulting in NAs)
  recordTable[, stationCol] <- factor(recordTable[, stationCol], levels = unique(CTtable[,stationCol]))
  
  # number of records of each species at each station
  t3 <- data.frame(rbind(table(recordTable[, stationCol], 
                                 recordTable[, speciesCol])), 
             check.names = FALSE)
  t3 <- data.frame(rownames(t3), t3, row.names=NULL, check.names = FALSE)
  colnames(t3)[1] <- stationCol
  
  t3 <- t3[match(toupper(as.character(dat2[,stationCol])), 
                     toupper(as.character(t3[,stationCol]))),]
  cex.t3 <- data.frame(t3[,1], apply(data.frame(t3[,-1]),
                                     MARGIN = 2,
                                     FUN = function(x){x/max(x, na.rm = TRUE)}))
  colnames(cex.t3)[1] <- stationCol
  
  # number of species detected per station
  t4 <- as.data.frame(tapply(X = recordTable[, speciesCol],
                             INDEX = list(recordTable[, stationCol]),
                             FUN = function(x) length(unique(x))))
  
  colnames(t4) <- "n_species"
  t4[,stationCol] <- rownames(t4)
  
  # replace NA (no species at station) with 0
  t4$n_species [is.na(t4$n_species)] <- 0
  
  t4 <- t4[,c(2,1)]
  t4 <- t4[match(toupper(as.character(dat2[,stationCol])), 
                 toupper(as.character(t4[,stationCol]))),]
  
  cex.t4 <- data.frame(t4[,1],
                       n_species_scaled = t4$n_species / max(t4$n_species, na.rm = TRUE))
  
  # set graphics  parameters and out directory
  if(hasArg(smallPoints) == FALSE) {smallPoints <- 0} else {
    if(smallPoints >  3 | smallPoints < 0 | !is.numeric(smallPoints)) stop("smallPoints must be a number between 0 and 3")
  }
  cex_pt1 <- 4.5 - smallPoints
  cex_pt2.max <- cex_pt1 - 1.5 + smallPoints/3
  cex_pt1_gg <- 9
  col_pt1_fill <- rgb(1,1,1, 0)
  col_pt1_border <- rgb(0,0,0, 0.5)
  col_pt2 <- "black"
  cex.labels <- 1.1
  cex.legend <- 1.5
  x_intersp <- 2
  y_intersp <- 1.3
  grconvertX.val <- 0.87
  grconvertY.val <- 0.9
  par.mar.tmp <- c(5.1, 4.1, 4.1, 8.1)
  pch1 <- 21
  pch2 <- 16
  col_polygon_border <- rgb(0.5,0.5,0.5,1)
  lwd_polygon_border <- 1.5
  lty_polygon_border <- 1

  range.expand.factor <- 0.04

  if(hasArg(backgroundPolygon)){
    
    # for sp
    # x.range.orig <- range(c(coordinates(backgroundPolygon@bbox)[1,], dat2[,Xcol]))
    # y.range.orig <- range(c(coordinates(backgroundPolygon@bbox)[2,], dat2[,Ycol]))
    
    # for sf
    x.range.orig <- range(c(st_coordinates(backgroundPolygon )[,1], dat2[,Xcol]))
    y.range.orig <- range(c(st_coordinates(backgroundPolygon )[,2], dat2[,Ycol]))
    
    x.range <- extendrange(r = x.range.orig, f = range.expand.factor)
    y.range <- extendrange(r = y.range.orig, f = range.expand.factor)

    X.tmp <- pngMaxPix / diff(x.range.orig)
    Y.tmp <- pngMaxPix / diff(y.range.orig)

  } else {

    x.range <- extendrange(r = range(range(dat2[,Xcol])), f = range.expand.factor)
    y.range <- extendrange(r = range(range(dat2[,Ycol])), f = range.expand.factor)

    X.tmp <- pngMaxPix / diff(range(dat2[Xcol]))
    Y.tmp <- pngMaxPix / diff(range(dat2[Ycol]))
    }

  par(mar = par.mar.tmp,
      xpd = TRUE,
      xaxs = "i")



  if(X.tmp > Y.tmp){
    pngWidth <-  pngMaxPix
    pngHeight <- round(pngMaxPix / (X.tmp /  Y.tmp))
  } else {
    pngWidth <-  round(pngMaxPix / (X.tmp /  Y.tmp))
    pngHeight <-  pngMaxPix
  }
  # if any of these is NA or NaN, set to pngMaxPix (may happen if only 1 station and x or y range is 0, bcause of division by 0)
  if(is.na(pngWidth))  pngWidth <- pngMaxPix
  if(is.na(pngHeight))  pngHeight <- pngMaxPix

  if(isTRUE(writePNG)){
    if(hasArg(plotDirectory)){
      if(isTRUE(createPlotDir)){
        dir.create(plotDirectory, recursive = TRUE, showWarnings = FALSE)
        setwd(plotDirectory)
      } else {
        stopifnot(file.exists(plotDirectory))
        setwd(plotDirectory)
      }
    } else {stop("please set plotDirectory")}
  }

  # species richness plot
  if(isTRUE(richnessPlot)){

    legend.label.richness <- sort(unique(t4$n_species))
    legend.cex.richness <- sort(unique((cex.t4$n_species_scaled * (cex_pt2.max))))

    if(isTRUE(writePNG)){
      png(filename = paste("n_Species_", Sys.Date(), ".png", sep = ""),
          width = pngWidth, height = pngHeight, units = "px", res = 96, type = "cairo")
      par(mar = par.mar.tmp, xpd=TRUE, xaxs = "i")

      plot(x = 0, type = "n", 
           main = ifelse(hasArg(speciesToShow), 
                         paste("Species Richness (out of subset of ", length(speciesToShow) , " species)", sep = ""), 
                         "Species Richness"),
           ylab = Ycol, xlab = Xcol,
           xlim = x.range, ylim = y.range,
           asp = 1,
           xaxs = "i", yaxs = "i")

      if(hasArg(backgroundPolygon)){plot(backgroundPolygon, add = TRUE, 
                                         border = col_polygon_border,
                                         lty = lty_polygon_border, lwd = lwd_polygon_border)}

      # station points
      points(y = dat2[, Ycol], x = dat2[, Xcol], pch = pch1,  bg  = col_pt1_fill, col = col_pt1_border,
           cex = cex_pt1)
      # detection points (scaled)
      points(y = dat2[, Ycol], x = dat2[, Xcol],  pch = pch2,
             col = col_pt2,
             cex = cex.t4$n_species_scaled * (cex_pt2.max))

      if(isTRUE(printLabels)){
        text(y = dat2[, Ycol], x = dat2[, Xcol], labels = dat2[,stationCol], cex = cex.labels, pos = 1, col = "red")
      }
      if(isTRUE(addLegend)){
        legend(
          x =  grconvertX(grconvertX.val, from = "ndc", to = "user"),
          y = grconvertY(grconvertY.val, from = "ndc", to = "user"),
          legend = legend.label.richness,
          pch = pch2,
          col = col_pt2,
          cex = cex.legend,
          pt.cex = legend.cex.richness,
          x.intersp = x_intersp,
          y.intersp = y_intersp,
          bty = "n")
      }
      dev.off()
    }

    if(isTRUE(plotR)){

      # this works. migrate everything to ggplot?

            # ggplot(dat2, aes(dat2[,Xcol], dat2[,Ycol])) +
            #   geom_point(cex =  cex_pt1_gg , fill = "white", colour = "black", pch = 21) +
            #   geom_point(aes(size = factor(t4$n_species)), colour = "black") +
            #   theme_bw() +
            #   theme(legend.position = "right",
            #         legend.box = "vertical",
            #         legend.key = element_blank()) +
            #   labs(x = Xcol,
            #        y = Ycol,
            #        title = "Species Richness")  +
            #   coord_fixed(ratio = 1) +
            #   scale_size_discrete(name = "n species")


      plot(x = 0, type = "n", 
           main = ifelse(hasArg(speciesToShow), 
                         paste("Species Richness (out of subset of ", length(speciesToShow) , " species)", sep = ""), 
                         "Species Richness"),
           ylab = Ycol, 
           xlab = Xcol,
           xlim = x.range, 
           ylim = y.range,
           asp = 1,
           xaxs = "i", yaxs = "i")

      if(hasArg(backgroundPolygon)){plot(backgroundPolygon, add = TRUE, border = col_polygon_border, lty = lty_polygon_border, lwd = lwd_polygon_border)}

      # station points
      points(y = dat2[, Ycol], x = dat2[, Xcol], pch = pch1,  bg  = col_pt1_fill, col = col_pt1_border,
             cex = cex_pt1)
      # detection points (scaled)
      points(y = dat2[, Ycol], x = dat2[, Xcol],  pch = pch2,
             col = col_pt2,
             cex = cex.t4$n_species_scaled * (cex_pt2.max))

      if(isTRUE(printLabels)){
        text(y = dat2[, Ycol], x = dat2[, Xcol], labels = dat2[,stationCol], cex = cex.labels, pos = 1, col = "red")
      }
      if(isTRUE(addLegend)){
        legend(
          x =  grconvertX(grconvertX.val, from = "ndc", to = "user"),
          y = grconvertY(grconvertY.val, from = "ndc", to = "user"),
          legend = legend.label.richness,
          pch = pch2,
          col = col_pt2,
          cex = cex.legend,
          pt.cex = legend.cex.richness,
          x.intersp = x_intersp,
          y.intersp = y_intersp,
          bty = "n")
      }
    }
  }

  if(isTRUE(speciesPlots)){
    #  species presence plots
    for(i in 2:ncol(t3)){
      i2 <- gsub(pattern = ".", replacement = "_",
                 x = colnames(t3)[i], fixed = TRUE)
      i2a <- gsub(pattern = ".", replacement = " ",
                  x = colnames(t3)[i], fixed = TRUE)

      legend.label <- sort(unique(t3[which(as.integer(t3[,i]) != 0),i]))
      legend.cex <- sort(unique((cex.t3[,i]^(1/2) * (cex_pt2.max))[which(as.integer(t3[,i]) != 0)]))

      if(isTRUE(writePNG)){
        png(filename = paste("Presence_", i2, "_", Sys.Date(), ".png", sep = ""),
            width = pngWidth, height = pngHeight, units = "px", res = 96, type = "cairo")
        par(mar = par.mar.tmp, xpd=TRUE, xaxs = "i")

        plot(x = 0, type = "n", main = i2a, ylab = Ycol, xlab = Xcol,
             xlim = x.range, ylim = y.range,
             asp = 1,
             xaxs = "i", yaxs = "i")

        if(hasArg(backgroundPolygon)){plot(backgroundPolygon, add = TRUE, border = col_polygon_border, lty = lty_polygon_border, lwd = lwd_polygon_border)}

        # station points
        points(y = dat2[, Ycol], x = dat2[, Xcol], pch = pch1,  bg  = col_pt1_fill, col = col_pt1_border,
               cex = cex_pt1)
        # detection points (scaled)
        points(y = dat2[, Ycol], x = dat2[, Xcol],  pch = pch2,
               col = col_pt2,
               cex = cex.t3[,i]^(1/2) * (cex_pt2.max))

        if(isTRUE(printLabels)){
          text(y = dat2[, Ycol], x = dat2[, Xcol], labels = dat2[,stationCol], cex = cex.labels, pos = 1, col = "red")
        }
        if(isTRUE(addLegend)){
          legend(
            x =  grconvertX(grconvertX.val, from = "ndc", to = "user"),
            y = grconvertY(grconvertY.val, from = "ndc", to = "user"),
            legend = legend.label,
            pch = pch2,
            col = col_pt2,
            cex = cex.legend,
            pt.cex = legend.cex,
            x.intersp = x_intersp,
            y.intersp = y_intersp,
            bty = "n")
        }
        dev.off()
      }

      if(isTRUE(plotR)){
        plot(x = 0, type = "n", main = i2a, ylab = Ycol, xlab = Xcol,
             xlim = x.range, ylim = y.range,
             asp = 1,
             xaxs = "i", yaxs = "i")

        if(hasArg(backgroundPolygon)){plot(backgroundPolygon, add = TRUE, border = col_polygon_border, lty = lty_polygon_border, lwd = lwd_polygon_border)}

        # station points
        points(y = dat2[, Ycol], x = dat2[, Xcol], pch = pch1,  bg  = col_pt1_fill, col = col_pt1_border,
               cex = cex_pt1)
        # detection points (scaled)
        points(y = dat2[, Ycol], x = dat2[, Xcol],  pch = pch2,
               col = col_pt2,
               cex = cex.t3[,i]^(1/2) * (cex_pt2.max))

        if(isTRUE(printLabels)){
          text(y = dat2[, Ycol], x = dat2[, Xcol], labels = dat2[,stationCol], cex = cex.labels, pos = 1, col = "red")
        }
        if(isTRUE(addLegend)){
          legend(
            x =  grconvertX(grconvertX.val, from = "ndc", to = "user"),
            y = grconvertY(grconvertY.val, from = "ndc", to = "user"),
            legend = legend.label,
            pch = pch2,
            col = col_pt2,
            cex = cex.legend,
            pt.cex = legend.cex,
            x.intersp = x_intersp,
            y.intersp = y_intersp,
            bty = "n")
        }
      }
    }
  }
  
  outtable <- data.frame(dat2, t3[,-1], n_species = t4[,-1])
  # if only 1 species, add column name 
  if(ncol(t3) == 2 & hasArg(speciesToShow)) colnames(outtable)[ncol(outtable) - 1] <- speciesToShow
  
  rownames(outtable) <- NULL
  # write Shapefile
  
  if(writeShapefile == TRUE){
    #if(hasArg(shapefileProjection)){proj.tmp <- shapefileProjection } else {proj.tmp <- NA}
    if(hasArg(shapefileName)){
      layer.tmp <- shapefileName 
    } else {
      layer.tmp <- paste("species_detection_", Sys.Date(), sep = "")
    }

    
    outtable_sf <- st_as_sf(outtable, 
                            coords = c(Xcol, Ycol))
    
    if(hasArg(shapefileProjection)) outtable_sf <- st_set_crs(outtable_sf, shapefileProjection)
    
    st_write(obj = outtable_sf,
             dsn = shapefileDirectory,
             layer = layer.tmp,
             driver = "ESRI Shapefile")
    
    
    return(invisible(outtable_sf))
    
  } else {
    return(invisible(outtable))
  }
}

# test-detectionMaps.R

context("detectionMaps")

# Load necessary libraries for testing
library(testthat)
library(sf)
library(withr)

# --- Test Setup ---
# This block creates shared, self-contained data for all tests.

# 1. Load sample data from the package
data(camtraps)
data(recordTableSample)

# 2. Create the sample Area of Interest (AOI) polygon from the example
aoi_poly <- st_polygon(list(cbind(c(521500, 526500, 527000, 521500, 521500),
                                  c(607500, 608000, 603500, 603500, 607500))))
aoi <- st_sf(name = "My AOI", geometry = st_sfc(aoi_poly), crs = 32650)


# --- Test Suite ---

testthat::describe("Core Functionality & Return Value", {
  
  test_that("it returns a correct data.frame with expected calculations", {
    # Run with plotting and file writing disabled to test only the data output
    maps_df <- detectionMaps(
      CTtable     = camtraps,
      recordTable = recordTableSample,
      Xcol        = "utm_x",
      Ycol        = "utm_y",
      stationCol  = "Station",
      speciesCol  = "Species",
      plotR       = FALSE, # Disable plotting for this test
      writePNG    = FALSE,
      writeShapefile = FALSE
    )
    
    # 1. Check the output type and dimensions
    expect_s3_class(maps_df, "data.frame")
    expect_equal(nrow(maps_df), nrow(camtraps)) # Should return a row for every station
    
    # 2. Check for expected columns (station, coords, species, richness)
    expected_species <- sort(unique(recordTableSample$Species))
    expected_cols <- c("Station", "utm_x", "utm_y", expected_species, "n_species")
    expect_named(maps_df, expected_cols)
    
    # 3. Verify specific calculated values based on sample data
    stationA_data <- maps_df[maps_df$Station == "StationA", ]
    expect_equal(stationA_data$PBE, 4)
    expect_equal(stationA_data$VTA, 2)
    expect_equal(stationA_data$n_species, 2) # PBE, VTA were detected here
    
    stationC_data <- maps_df[maps_df$Station == "StationC", ]
    expect_equal(stationC_data$EGY, 6)
    expect_equal(stationC_data$PBE, 6)
    expect_equal(stationC_data$MNE, 0)
    expect_equal(stationC_data$n_species, 4)
  })
})


testthat::describe("Filtering and Subset Options", {
  
  plot(1)  # avoid warning: calling par(new=TRUE) with no plot
  
  test_that("speciesToShow argument correctly filters the output", {
    species_subset <- c("PBE", "VTA")
    maps_df_filtered <- detectionMaps(
      CTtable       = camtraps,
      recordTable   = recordTableSample,
      speciesToShow = species_subset,
      Xcol = "utm_x", Ycol = "utm_y", plotR = FALSE
    )
    
    # 1. Check that only the specified species columns exist
    expected_cols <- c("Station", "utm_x", "utm_y", sort(species_subset), "n_species")
    expect_named(maps_df_filtered, expected_cols)
    
    # 2. Check that richness is recalculated based on the subset
    stationB_data <- maps_df_filtered[maps_df_filtered$Station == "StationB", ]
    # StationB had MNE, PBE VTA. After filtering, it should only count 2 species.
    expect_equal(stationB_data$n_species, 2)
  })
})


testthat::describe("File Generation (Side Effects)", {
  
  # Use a temporary directory to avoid creating files in the project
  with_tempdir({
    
    test_that("writePNG = TRUE creates the expected number of PNG files", {
      # Suppress plotting to the R device to keep the test console clean
      pdf(NULL)
      detectionMaps(
        CTtable       = camtraps,
        recordTable   = recordTableSample,
        Xcol = "utm_x", Ycol = "utm_y",
        writePNG      = TRUE,
        plotR         = FALSE, # Don't plot to R device
        plotDirectory = ".",   # Write to the temp directory
        createPlotDir = FALSE
      )
      dev.off() # Close the null PDF device
      
      png_files <- list.files(pattern = "\\.png$")
      
      # Expecting 7 PNGs: 1 for richness + 6 for the unique species in the sample data
      n_species <- length(unique(recordTableSample$Species))
      expect_equal(length(png_files), 1 + n_species) 
      expect_true(any(grepl("n_Species_", png_files)))
      expect_true(any(grepl("Presence_PBE", png_files)))
    })
    
    test_that("writeShapefile = TRUE creates a valid shapefile", {
      skip_if_not_installed("sf")
      
      # Generate a unique name to avoid conflicts
      shapefile_name <- "test_detections"
      
      maps_sf_obj <- detectionMaps(
        CTtable            = camtraps,
        recordTable        = recordTableSample,
        Xcol = "utm_x", Ycol = "utm_y", plotR = FALSE,
        writeShapefile     = TRUE,
        shapefileName      = shapefile_name,
        shapefileDirectory = ".", # Write to the temp directory
        shapefileProjection= 4326 # Use a known CRS (WGS84)
      )
      
      # 1. Check that the shapefile components were created
      expect_true(file.exists(paste0(shapefile_name, ".shp")))
      expect_true(file.exists(paste0(shapefile_name, ".dbf")))
      
      # 2. Check that the returned object is an sf object
      expect_s3_class(maps_sf_obj, "sf")
      
      # 3. Read the shapefile back in and verify its properties
      sf_read <- st_read(dsn = ".", layer = shapefile_name, quiet = TRUE)
      expect_s3_class(sf_read, "sf")
      expect_equal(nrow(sf_read), nrow(camtraps))
      expect_true(st_crs(sf_read) == st_crs(4326)) # Verify projection was set
    })
  })
})


testthat::describe("Spatial Features and Plotting Options", {
  
  test_that("backgroundPolygon argument runs without error", {
    # The main test is that the function doesn't crash when trying to plot the polygon.
    # We can't easily test the visual output, so we just check for successful execution.
    pdf(NULL) # Suppress plot output
    expect_no_error(
      detectionMaps(
        CTtable           = camtraps,
        recordTable       = recordTableSample,
        backgroundPolygon = aoi,
        Xcol = "utm_x", Ycol = "utm_y",
        plotR             = TRUE,
        writePNG          = FALSE
      )
    )
    dev.off()
  })
})


testthat::describe("Input Validation", {
  
  plot(1)  # avoid warning: calling par(new=TRUE) with no plot
  
  test_that("it stops for missing or invalid column names", {
    expect_error(
      detectionMaps(CTtable = camtraps, recordTable = recordTableSample, Xcol = "WRONG_X", Ycol = "utm_y"),
      regexp = "Xcol = \"WRONG_X\" is not a column name in CTtable"
    )
    
    # Mismatched station IDs
    recordTable_bad_station <- recordTableSample
    recordTable_bad_station$Station[1] <- "UNKNOWN_STATION"
    expect_error(
      detectionMaps(CTtable = camtraps, recordTable = recordTable_bad_station, Xcol = "utm_x", Ycol = "utm_y"),
      regexp = "items of stationCol in recordTable are not matched"
    )
  })
  
  test_that("it stops if writeShapefile = TRUE and directory is missing", {
    expect_error(
      detectionMaps(CTtable = camtraps, recordTable = recordTableSample, Xcol = "utm_x", Ycol = "utm_y",
                    writeShapefile = TRUE, plotR = FALSE),
      fixed = "hasArg(shapefileDirectory) is not TRUE"
    )
  })
})

#' Generate a single-species record table with individual identification from
#' camera trap images or videos
#' 
#' The function generates a single-species record table containing individual
#' IDs, e.g. for (spatial) capture-recapture analyses. It prepares input for
#' the function \code{\link{spatialDetectionHistory}}.
#' 
#' The function can handle a number of different ways of storing images and
#' videos. In every case, images need to be stored in a species directory first
#' (e.g. using function \code{\link{getSpeciesImages}}). Station subdirectories
#' are optional. Camera subdirectories are not supported. This directory
#' structure can be created easily with function
#' \code{\link{getSpeciesImages}}.
#' 
#' As with species identification, individuals can be identified in 2 different
#' ways: by moving images into individual directories
#' ("Species/Station/Individual/XY.JPG" or "Species/Individual/XY.JPG") or by
#' metadata tagging (without the need for individual directories:
#' "Species/XY.JPG" or "Species/Station/XY.JPG").
#' 
#' \code{minDeltaTime} is a criterion for temporal independence of records of
#' an individual at the same station/location. Setting it to 0 will make the
#' function return all records. \code{camerasIndependent} defines if the
#' cameras at a station are to be considered independent (e.g. \code{FALSE} if
#' both cameras face each other and possibly \code{TRUE} if they face different
#' trails). \code{stationCol} is the station column name to be used in the
#' resulting table. Station IDs are read from the station directory names if
#' \code{hasStationFolders = TRUE}. Otherwise, the function will try to extract
#' station IDs from the image filenames (requires images renamed with
#' \code{\link{imageRename}}.
#' 
#' If individual IDs were assigned with image metadata tags,
#' \code{metadataIDTag} must be set to the name of the metadata tag group used
#' for individual identification. \code{metadataHierarchyDelimitor} is "|" for
#' images tagged in DigiKam and images tagged in Adobe Bridge/ Lightroom with
#' the default settings. Manufacturer-specific Exif metadata tags such as
#' "AmbientTemperature" or "MoonPhase" can be extracted if specified in
#' \code{additionalMetadataTags}. Multiple names can be specified as a
#' character vector as: \code{c(Tag1, Tag2, ...)}. Because they are not
#' standardized, function \code{\link{exifTagNames}} provides a vector of all
#' available tag names. The metadata tags thus extracted may be used as
#' individual covariates in spatial capture-recapture models.
#' 
#' \code{eventSummaryColumn} and \code{eventSummaryFunction} can be used to
#' extract summary statistics for independent sampling events. For example, you
#' assigned a "count" tag to your images, indicating the number of individuals
#' in a picture. In a sequence of pictures taken within 1 minute, most pictures
#' show one individual, but one image shows two individuals. You tagged the
#' images accordingly (count = 1 or count = 2) and run \code{recordTable}. Set
#' \code{eventSummaryColumn = "count"} and \code{eventSummaryFunction = "max"}
#' to obtain the maximum number of \code{count} in all images within
#' \code{minDeltaTime} minutes of a given record. The results is in a new
#' column, in this example \code{count_max}. You can also calculate several
#' statistics at the same time, by supplying vectors of values, e.g.
#' \code{eventSummaryColumn = c("count", "count", "camera")} and
#' \code{eventSummaryFunction = c("min", "max", "unique")} to get minimum and
#' maximum count and all unique camera IDs for that event. Note that
#' \code{eventSummaryColumn} and \code{eventSummaryFunction} must be of same
#' length.
#' 
#' Argument \code{video} is analogous to \code{\link{recordTable}}, a named
#' list with 2 or 4 items. 2 items (\code{file_formats}, \code{dateTimeTag})
#' are always required, and are sufficent if \code{IDfrom = "directory"}. In
#' that case, no digiKam tags will be returned.  To return digiKam tags, two
#' additional items are required (\code{db_directory}, \code{db_filename}).
#' This is essential when using \code{IDfrom = "metadata"}. When using
#' \code{IDfrom = "directory"}, it is optional, but allows to extract metadata
#' tags assigned to videos in digiKam. This workaround is necessary because
#' digiKam tags are not written into video metadata, but are only saved in the
#' digiKam database. So in contrast to JPG images, they can not be extracted
#' with ExifTool. It also requires that \code{inDir} is in your digiKam
#' database.
#' 
#' The items of argument \code{video} are:
#' 
#' \tabular{ll}{ \code{file_formats} \tab The video formats to extract (include
#' "jpg" if you want .JPG image metadata) \cr \code{dateTimeTag} \tab the
#' metadata tag to extract date/time from (use \code{\link{exifTagNames}} to
#' find out which tag is suitable) \cr \code{db_directory} \tab The directory
#' containing digiKam database (optional if \code{IDfrom = "directory"}) \cr
#' \code{db_filename} \tab The digiKam database file in \code{db_directory}
#' (optional if \code{IDfrom = "directory"}) \cr }
#' 
#' See the example below for for how to specify the argument \code{video}.
#' 
#' @param inDir character. Directory containing images of individuals. Must end
#' with species name (e.g. ".../speciesImages/Clouded Leopard")
#' @param hasStationFolders logical. Does \code{inDir} have station
#' subdirectories? If \code{TRUE}, station IDs will be taken from directory
#' names. If \code{FALSE}, they will be taken from image filenames (requires
#' images renamed with \code{\link{imageRename}}).
#' @param IDfrom character. Read individual ID from image metadata ("metadata")
#' of from directory names ("directory")?
#' @param cameraID character. Should the function look for camera IDs in the
#' image file names? If so, set to 'filename'. Requires images renamed with
#' \code{\link{imageRename}}. If missing, no camera ID will be assigned and it
#' will be assumed there was 1 camera only per station.
#' @param camerasIndependent logical. If \code{TRUE}, cameras at a station are
#' assumed to record individuals independently. If \code{FALSE}, cameras are
#' assumed to be non-independent (e.g. in pairs). Takes effect only if there
#' was more than 1 camera per station and cameraID = "filename".
#' @param minDeltaTime numeric. time difference between observation of the same
#' individual at the same station/camera to be considered independent (in
#' minutes)
#' @param deltaTimeComparedTo character. For two records to be considered
#' independent, must the second one be at least \code{minDeltaTime} minutes
#' after the last independent record of the same individual
#' (\code{"lastIndependentRecord"}), or \code{minDeltaTime} minutes after the
#' last record (\code{"lastRecord"})?
#' @param timeZone character. Must be a value returned by
#' \code{\link[base:timezones]{OlsonNames}}
#' @param stationCol character. Name of the camera trap station column in the
#' output table.
#' @param writecsv logical. Should the individual record table be saved as a
#' .csv file?
#' @param outDir character. Directory to save csv file to. If NULL and
#' \code{writecsv = TRUE}, the output csv will be written to \code{inDir}.
#' @param metadataHierarchyDelimitor character. The character delimiting
#' hierarchy levels in image metadata tags in field "HierarchicalSubject".
#' Either "|" or ":".
#' @param metadataIDTag character. In custom image metadata, the individual ID
#' tag name.
#' @param additionalMetadataTags character. additional camera model-specific
#' metadata tags to be extracted. (If possible specify tag groups as returned
#' by \code{\link{exifTagNames}})
#' @param removeDuplicateRecords logical. If there are several records of the
#' same individual at the same station (also same camera if cameraID is
#' defined) at exactly the same time, show only one?
#' @param returnFileNamesMissingTags logical. If species are assigned with
#' metadata and images are not tagged, return a few file names of these images
#' as a message?
#' @param eventSummaryColumn character. A column in the record table (e.g. from
#' a metadata tag) by to summarise non-independent records (those within
#' \code{minDeltaTime} of a given record) with a user-defined function
#' (\code{eventSummaryFunction})
#' @param eventSummaryFunction character. The function by which to summarise
#' \code{eventSummaryColumn} of non-independent records, e.g. "sum", "max"
#' (optional)
#' @param video list. Contains information on how to handle video data
#' (optional). See details.
#' 
#' @return A data frame containing species records with individual IDs and
#' additional information about stations, date, time and (optionally) further
#' metadata.
#' 
#' @section Warning : Be sure to read the section on individual identification
#' in the package vignette
#' (\url{https://CRAN.R-project.org/package=camtrapR/vignettes/camtrapr2.pdf}).
#' 
#' Af you use image metadata tags for identification, the tags must be written
#' to the image metadata. The function cannot read tags from .xmp sidecar
#' files. Make sure you set the preferences of your image management software
#' accordingly. In DigiKam, go to Settings/Configure digiKam/Metadata. There,
#' make sure "Write to sidecar files" is unchecked.
#' 
#' Please note the section about defining argument \code{timeZone} in the
#' vignette on data extraction (accessible via
#' \code{vignette("DataExtraction")} or online
#' (\url{https://cran.r-project.org/package=camtrapR/vignettes/camtrapr3.pdf})).
#' 
#' @author Juergen Niedballa
#' 
#' @references Phil Harvey's ExifTool \url{https://exiftool.org/}
#' 
#' @examples
#' 
#' 
#' \dontrun{   # the examples run too long to pass CRAN tests
#' 
#'  wd_images_ID_individual <- system.file("pictures/sample_images_indiv_tag/LeopardCat", 
#'                                         package = "camtrapR")
#'  # missing space in species = "LeopardCat" is because of CRAN package policies
#'  # note argument additionalMetadataTags: contains tag names as returned by function exifTagNames
#' 
#'  if (Sys.which("exiftool") != ""){        # only run these examples if ExifTool is available
#' 
#'  rec_table_pbe <- recordTableIndividual(inDir                  = wd_images_ID_individual,
#'                                         minDeltaTime           = 60,
#'                                         deltaTimeComparedTo    = "lastRecord",
#'                                         hasStationFolders      = FALSE,
#'                                         IDfrom                 = "metadata",
#'                                         camerasIndependent     = FALSE,
#'                                         writecsv               = FALSE,
#'                                         metadataIDTag          = "individual",
#'                                         additionalMetadataTags = c("EXIF:Model", "EXIF:Make"),
#'                                         timeZone               = "Asia/Kuala_Lumpur"
#'  )
#'  
#'  
#'  # extracting some example summary stats too 
#'  # a nonsensical example, get all unique cameras with which the event was photographed
#'  
#'   rec_table_pbe2 <- recordTableIndividual(inDir                = wd_images_ID_individual,
#'                                         minDeltaTime           = 60,
#'                                         deltaTimeComparedTo    = "lastRecord",
#'                                         hasStationFolders      = FALSE,
#'                                         IDfrom                 = "metadata",
#'                                         camerasIndependent     = FALSE,
#'                                         writecsv               = FALSE,
#'                                         metadataIDTag          = "individual",
#'                                         additionalMetadataTags = c("EXIF:Model", "EXIF:Make"),
#'                                         timeZone               = "Asia/Kuala_Lumpur",
#'                                         eventSummaryColumn     = "EXIF:Make",
#'                                         eventSummaryFunction   = "unique"
#'  )
#'  
#'  ### Video example (the sample data don't contain a video, this is hypothetical)
#'  # with JPG, video mp4, avi, mov, ID = metadata
#'  
#' rec_table_ind_video <- recordTableIndividual(inDir = wd_images_ID_individual,
#'                           hasStationFolder  = FALSE,
#'                           IDfrom            = "metadata", 
#'                           metadataIDTag     = "individual",
#'                           video = list(file_formats = c("jpg", "mp4", "avi", "mov"),
#'                                        dateTimeTag = "QuickTime:CreateDate",
#'                                        db_directory = "C:/Users/YourName/Pictures",
#'                                        db_filename = "digikam4.db")
#' )
#' 
#'   
#' } else {  
#' # show function output if ExifTool is not available
#' message("ExifTool is not available. Cannot test function. Loading recordTableSample instead")
#' data(recordTableSample)
#' }
#' }
#' 
#' 
#' @export recordTableIndividual
#' 
recordTableIndividual <- function(inDir,
                                  hasStationFolders,
                                  IDfrom,
                                  cameraID,
                                  camerasIndependent,
                                  minDeltaTime = 0,
                                  deltaTimeComparedTo,
                                  timeZone,
                                  stationCol,
                                  writecsv = FALSE,
                                  outDir,
                                  metadataHierarchyDelimitor = "|",
                                  metadataIDTag,
                                  additionalMetadataTags,
                                  removeDuplicateRecords = TRUE,
                                  returnFileNamesMissingTags = FALSE,
                                  eventSummaryColumn,
                                  eventSummaryFunction,
                                  video

)
{
  wd0 <- getwd()
  on.exit(setwd(wd0))

  if(!hasArg(stationCol)) stationCol <- "Station"
  if(!is.character(stationCol))     stop("stationCol must be of class 'character'", call. = FALSE)
  individualCol <- "Individual"
  speciesCol    <- "Species"

  checkForSpacesInColumnNames(stationCol = stationCol)

  # check input
  if(hasArg(timeZone) == FALSE) {
    message("timeZone is not specified. Assuming UTC")
    timeZone <- "UTC"
  }
  if(!is.element(timeZone , OlsonNames())){
    stop("timeZone must be an element of OlsonNames()", call. = FALSE)
  }
  
  if(Sys.which("exiftool") == "") stop("cannot find ExifTool", call. = FALSE)

  
  if(!is.character(inDir))  stop("inDir must be of class 'character'", call. = FALSE)
  if(length(inDir) != 1)    stop("inDir may only consist of 1 element only", call. = FALSE)
  if(!dir.exists(inDir))    stop("Could not find inDir:\n", inDir, call. = FALSE)
  
  IDfrom <- match.arg(IDfrom, choices = c("metadata", "directory") )
            
 if(IDfrom == "metadata"){
   metadataHierarchyDelimitor <- match.arg(metadataHierarchyDelimitor, choices = c("|", ":"))
    if(!hasArg(metadataIDTag))       stop('"metadataIDTag" must be defined if IDfrom = "metadata"')
    if(!is.character(metadataIDTag)) stop("metadataIDTag must be of class 'character'")
    if(length(metadataIDTag) != 1)   stop("metadataIDTag must be of length 1")
  }

  multiple_tag_separator <- "_&_"

  if(hasArg(cameraID)){
    if(!is.character(cameraID))              stop("cameraID must be of class 'character'", call. = FALSE)
    if(cameraID %in% c("filename") == FALSE) stop("cameraID can only be 'filename' or missing", call. = FALSE)
    if(!hasArg(camerasIndependent))          stop("camerasIndependent is not defined. It must be defined if cameraID is defined", call. = FALSE)
    if(!is.logical(camerasIndependent))      stop("camerasIndependent must be of class 'logical'", call. = FALSE)
  } else {
    camerasIndependent <- FALSE
    }

  cameraCol <- "Camera"


  if(hasArg(outDir)){
    if(!is.character(outDir))        stop("outDir must be of class 'character'", call. = FALSE)
    if(isFALSE(file.exists(outDir))) stop("outDir does not exist", call. = FALSE)
  }

  if(hasArg(additionalMetadataTags)){
    if(!is.character(additionalMetadataTags)) stop("additionalMetadataTags must be of class 'character'", call. = FALSE)
    if(any(grep(pattern = " ", x = additionalMetadataTags, fixed = TRUE))) stop("In argument additionalMetadataTags, spaces are not allowed")
    if("HierarchicalSubject" %in% additionalMetadataTags & IDfrom == "metadata")  {
      warning("'HierarchicalSubject' may not be in 'additionalMetadataTags' if IDfrom = 'metadata'. It will be ignored because the function returns it anyway.", call. = FALSE)
      additionalMetadataTags <- additionalMetadataTags[-grep(pattern = "HierarchicalSubject", x = additionalMetadataTags)]  # remove it
    }
  }

  if(!is.logical(removeDuplicateRecords))     stop("'removeDuplicateRecords' must be logical (TRUE / FALSE)", call. = FALSE)
  if(!is.logical(returnFileNamesMissingTags)) stop("'returnFileNamesMissingTags' must be logical (TRUE / FALSE)", call. = FALSE)

  metadata.tagname <- "HierarchicalSubject"    # for extracting metadata assigned in tagging software

  minDeltaTime <- as.integer(minDeltaTime)
  if(!is.integer(minDeltaTime)) stop("'minDeltaTime' must be an integer", call. = FALSE)

  if(minDeltaTime != 0){
    if(removeDuplicateRecords == FALSE){
      warning("minDeltaTime is > 0. Therefore, removeDuplicateRecords was set to TRUE (otherwise there may be records taken at the same time)", call. = FALSE, immediate. = TRUE)
      removeDuplicateRecords <- TRUE
    }
    deltaTimeComparedTo < match.arg(deltaTimeComparedTo, choices = c("lastRecord", "lastIndependentRecord"))
    if(!hasArg(deltaTimeComparedTo)) stop(paste("minDeltaTime is not 0. deltaTimeComparedTo must be defined"), call. = FALSE)
  } else {
    if(hasArg(deltaTimeComparedTo)) {
      warning(paste("minDeltaTime is 0. deltaTimeComparedTo = '", deltaTimeComparedTo, "' will have no effect", sep = ""), call. = FALSE, immediate. = TRUE)
    } else {
      deltaTimeComparedTo <- "lastRecord"
    }
  }

  if(!is.logical(writecsv)) stop("writecsv must be logical")


  # find image directories

  
  if(!is.logical(hasStationFolders))    stop("hasStationFolders must be of class 'logical'", call. = FALSE)
  
  
  if(hasStationFolders == TRUE){
    dirs       <- list.dirs(inDir, full.names = TRUE,  recursive = FALSE)
    dirs_short <- list.dirs(inDir, full.names = FALSE, recursive = FALSE)
    if(length(dirs) == 0) stop("inDir contains no station directories", call. = FALSE)
  } else {
    dirs <- dirs_short <- inDir
  }

  max_nchar_station <- max(nchar(dirs_short))

  # Ignore the ".dtrash" folder if present (on Mac)
  if(".dtrash" %in% dirs_short) {
    # message("Ignoring .dtrash folder.")
    dirs <- dirs[dirs_short != ".dtrash"]
    dirs_short <- dirs_short[dirs_short != ".dtrash"]
  }
  
  # Ignore  the ".mysql.digikam" folder if present
  if(".mysql.digikam" %in% dirs_short) {
    # message("Ignoring .mysql.digikam folder.")
    dirs <- dirs[dirs_short != ".mysql.digikam"]
    dirs_short <- dirs_short[dirs_short != ".mysql.digikam"]
  }
  
  
  # process video argument (if present)
  if(hasArg(video)){
    video_out <- processVideoArgument(IDfrom = IDfrom,
                                      video  = video)
    digiKam_data <- video_out$digiKam_data
    file_formats <- video_out$file_formats
    # if(isFALSE("jpg" %in% file_formats)) file_formats <- c(file_formats, "jpg")
  } else {
    file_formats <- "jpg"   # jpg, as the default, if video not requested
  }
  
  record.table.list <- list()

    # # create command line and execute exiftool
    #   if(hasArg(additionalMetadataTags)){
    #     command.tmp  <- paste('exiftool -q -f -t -r -Directory -FileName -EXIF:DateTimeOriginal -HierarchicalSubject', paste(" -",additionalMetadataTags,  collapse = "", sep = ""), ' -ext JPG "', dirs, '"', sep = "")
    #     colnames.tmp <- c("Directory", "FileName", "DateTimeOriginal", "HierarchicalSubject", additionalMetadataTags)
    #   } else {
    #     command.tmp  <- paste('exiftool -q -f -t -r -Directory -FileName -EXIF:DateTimeOriginal -HierarchicalSubject -ext JPG "',dirs, '"', sep = "")
    #     colnames.tmp <- c("Directory", "FileName", "DateTimeOriginal", "HierarchicalSubject")
    #   }
  
  command.tmp  <- paste('exiftool -q -f -t -r -Directory -FileName -EXIF:DateTimeOriginal', 
                        ifelse(hasArg(video), paste(" -", video$dateTimeTag, sep = ""), ""),    # if video requested, video date time tag
                        ' -HierarchicalSubject',
                        ifelse(hasArg(additionalMetadataTags), paste(" -",additionalMetadataTags,  collapse = "", sep = ""), ""),
                        paste(" -ext", file_formats, collapse = " ", sep = " "),    # requested file extensions
                        ' "', dirs, '"', sep = "")
  
  # construct column names for metadata table
  colnames.tmp <- c("Directory", "FileName", "DateTimeOriginal")
  if(hasArg(video)) colnames.tmp <- c(colnames.tmp, video$dateTimeTag)
  colnames.tmp <- c(colnames.tmp, "HierarchicalSubject")
  if(hasArg(additionalMetadataTags)) colnames.tmp <- c(colnames.tmp, additionalMetadataTags [nchar(additionalMetadataTags) >= 2]) 
  # only add as new column if entry of additionalMetadataTag has more than 2 characters (to allow users to specify exiftool commands, e.g. -L)
  
  
  for(i in 1:length(dirs)){   # loop through station directories

    # execute exiftool
    metadata.tmp <- runExiftool(command.tmp = command.tmp[i], colnames.tmp = colnames.tmp)


    if(is.null(class(metadata.tmp))){            # omit station if no images found

      length.tmp <- length(list.files(dirs[i], pattern = paste(".", file_formats, "$", collapse = "|", sep = ""), 
                                      ignore.case = TRUE, recursive = TRUE))
      message(paste(formatC(dirs_short[i], width = max_nchar_station, flag = "-"),  ":  ",
                    formatC(length.tmp, width = 5), " images      Skipping", sep = ""))
      warning(paste(dirs_short[i],  ":  contains no images and was omitted\n"), call. = FALSE,  immediate. = FALSE)

    } else {

      # if video files extracted, add DateTimeOriginal and HierarchicalSubject
      if(hasArg(video)){
        metadata.tmp <- addVideoDateTimeOriginal(metadata.tmp = metadata.tmp, video = video)
        
        # add HierachicalSubject for video files
        if(!is.null(digiKam_data)){
          digiKamVideoMetadata <- digiKamVideoHierarchicalSubject(stationDir = dirs[i],
                                                                    digiKamTablesList = digiKam_data,    # output of accessDigiKamDatabase
                                                                    videoFormat = file_formats[!grepl(file_formats, pattern = "jpg")])
          # add HierarchialSubject for video files (match by filename, must be unique)
          metadata.tmp <- addVideoHierarchicalSubject (metadata.tmp = metadata.tmp,
                                                       video = video,
                                                       digiKamVideoMetadata = digiKamVideoMetadata)
        }
      }
      
      # check presence / consistency of DateTimeOriginal column, go to next station or remove records if necessary
      metadata.tmp <- checkDateTimeOriginal (intable    = metadata.tmp,
                                             dirs_short = dirs_short,
                                             i          = i)
      if(is.null(metadata.tmp)) next

      # now split HierarchicalSubject tags and add as columns to table
      metadata.tmp <- addMetadataAsColumns (intable                    = metadata.tmp,
                                            metadata.tagname           = metadata.tagname,
                                            metadataHierarchyDelimitor = metadataHierarchyDelimitor,
                                            multiple_tag_separator     = multiple_tag_separator)


      # add individual ID to metadata table (from folders or metadata, otherwise NA) - function is called assignSpeciesID but works on individual here

      metadata.tmp <- assignSpeciesID (intable                = metadata.tmp,            # also works for individual IDs assuming that there is 1 species only
                                       IDfrom                 = IDfrom,
                                       metadataSpeciesTag     = metadataIDTag,
                                       speciesCol             = individualCol,           # individuals are treated as species here (so multiple individuals tagged in the same picture will be returned)
                                       dirs_short             = dirs_short,
                                       i_tmp                  = i,
                                       multiple_tag_separator = multiple_tag_separator,
                                       returnFileNamesMissingTags = returnFileNamesMissingTags,
                                       parent                 = "recordTableIndividual"
      )


      # if images in station contain no metadata individual ID tag or are not tagged, skip that station
      if(!is.data.frame(metadata.tmp)){
        if(metadata.tmp == "found no species tag") {
          # warning(paste(dirs_short[i], ":   metadataIDTag '", metadataIDTag, "' not found in image metadata tag 'HierarchicalSubject'. Skipping", sep = ""), call. = FALSE, immediate. = TRUE)
        } else {
          warning(paste(dirs_short[i], ":   error in individual tag extraction. Skipping. Please report", sep = ""), call. = FALSE, immediate. = TRUE)
        }
        next
      }


      # remove empty metadata columns (if HierarchicalSubject is all empty or if additionalMetadataTags were not found)
      empty_cols <- which(apply(metadata.tmp, MARGIN = 2, FUN = function(X){all(X == "-")}))
      if(length(empty_cols) >= 1){
        metadata.tmp <-  metadata.tmp[, -empty_cols]
      }

      # add station and camera id to metadata table
      arg.list0 <- list(intable = metadata.tmp, dirs_short = dirs_short, stationCol = stationCol, hasStationFolders = hasStationFolders, cameraCol = cameraCol, i = i, IDfrom = IDfrom)

      if(!hasArg(cameraID)) metadata.tmp <- do.call(addStationCameraID, arg.list0)
      if( hasArg(cameraID)) metadata.tmp <- do.call(addStationCameraID, c(arg.list0, cameraID = cameraID))   # if cameraID is defined, it will be extracted from file names

      # add species name to table (from last part of inDir)
      inDir.split <- unlist(strsplit(inDir, split = .Platform$file.sep, fixed = TRUE))
      metadata.tmp[,speciesCol] <- inDir.split[length(inDir.split)]

      if(nrow(metadata.tmp) >= 1){   # if anything left, do

        # convert character vector extracted from images to time object and format for outfilename
        metadata.tmp$DateTimeOriginal <- as.POSIXct(strptime(x = metadata.tmp$DateTimeOriginal, format = "%Y:%m:%d %H:%M:%S", tz = timeZone))

        # sort by station, (camera), individual and time
        if(camerasIndependent == TRUE) {
          if(all(is.na(metadata.tmp[, cameraCol]))) stop(paste("Failed to extract camera ID from", cameraID))
          metadata.tmp <- metadata.tmp[order(metadata.tmp[,stationCol], metadata.tmp[,individualCol], metadata.tmp[,cameraCol], metadata.tmp$DateTimeOriginal),]
        } else {
          metadata.tmp <- metadata.tmp[order(metadata.tmp[,stationCol], metadata.tmp[,individualCol], metadata.tmp$DateTimeOriginal),]
        }

        #remove duplicate records of same individual taken in same second at the same station (by the same camera, if relevant)
        metadata.tmp2 <- removeDuplicatesOfRecords(metadata.tmp           = metadata.tmp,
                                                   removeDuplicateRecords = removeDuplicateRecords,
                                                   camerasIndependent     = camerasIndependent,
                                                   stationCol             = stationCol,
                                                   speciesCol             = individualCol,    # meaning, there will be no duplicate records of the same individual at the same second and station
                                                   cameraCol              = cameraCol,
                                                   current                = i, 
                                                   total                  = length(dirs),
                                                   max_nchar_station      = max_nchar_station)
        
        # assess independence between records and calculate time differences (and possibly summarise a column)
        args.assessTemporalIndependence <- list(intable             = metadata.tmp2,
                                                deltaTimeComparedTo = deltaTimeComparedTo,
                                                columnOfInterest    = individualCol,
                                                cameraCol           = cameraCol,
                                                camerasIndependent  = camerasIndependent,
                                                minDeltaTime        = minDeltaTime,
                                                stationCol          = stationCol)
        
        if(hasArg(eventSummaryColumn)) {
          if(!is.character(eventSummaryColumn))     stop("eventSummaryColumn must be of class 'character'", call. = FALSE)
          if(!is.character(eventSummaryFunction))   stop("eventSummaryFunction must be of class 'character'", call. = FALSE)
          args.assessTemporalIndependence <- c(args.assessTemporalIndependence,
                                               eventSummaryColumn   = eventSummaryColumn,
                                               eventSummaryFunction = eventSummaryFunction)
        }
        
        d1 <- do.call(assessTemporalIndependence, args = args.assessTemporalIndependence)
        
        # save station table to list
        record.table.list[[i]] <- d1
        
       # # add potential new columns to global record.table
       #  d2 <- addNewColumnsToGlobalTable (intable      = d1,
       #                                    i            = i,
       #                                    record.table = record.table)
       # 
       # 
       # 
       #  # append table of station i's images metadata to global record table
       #  record.table <- rbind(d2[[2]], d2[[1]])

        suppressWarnings(rm(d1))
      }
    }
  }       # end loop through station directories
  
  
  # combine all data frames from list into one data frame
  record.table <- as.data.frame(data.table::rbindlist(record.table.list, fill = TRUE, use.names = TRUE))

  if(nrow(record.table) == 0){
    stop(paste("something went wrong. I looked through all those", length(dirs)  ,"folders and now your table is empty"), call. = FALSE)
  }

  # rearrange table, add date and time as separate columns. add additional column names as needed.

  record.table2  <-  data.frame(record.table[,c(stationCol, speciesCol, individualCol, "DateTimeOriginal")],
                                Date = as.Date (record.table$DateTimeOriginal, format = "%Y/%M/%d", tz = timeZone),
                                Time = strftime(record.table$DateTimeOriginal, format = "%H:%M:%S", tz = timeZone),
                                record.table[,c("delta.time.secs", "delta.time.mins", "delta.time.hours", "delta.time.days",
                                                "Directory", "FileName")])

  metadata_columns <- which(colnames(record.table) %in% colnames(record.table2) == FALSE)

  # add metadata columns
  if(length(metadata_columns) >= 1){
    record.table3 <- cbind(record.table2, record.table[,metadata_columns])
    colnames(record.table3)[(ncol(record.table2) + 1) : ncol(record.table3)] <- colnames(record.table)[metadata_columns]
  } else {record.table3 <- record.table2}


  # add camera column (if present
  if(hasArg(cameraID)){
    record.table3 <- data.frame(record.table3[,stationCol],
                                record.table[,cameraCol],
                                record.table3[,-which(colnames(record.table3) %in% c(stationCol, cameraCol))])
    colnames(record.table3)[1] <- stationCol
    colnames(record.table3)[2] <- cameraCol
  }

  record.table3 <- record.table3[with(record.table3, order(record.table3[,stationCol], record.table3[,individualCol], DateTimeOriginal)), ]
  rownames(record.table3) <- NULL

   # warning if additionalMetadataTags were not found
  if(hasArg(additionalMetadataTags)){
    #whichAdditionalMetadataTagsFound <- which(gsub(additionalMetadataTags, pattern = ":", replacement = ".") %in% colnames(record.table3))   # replace : in additionalMetadataTags (if specifying tag groups) with . as found in column names
    whichAdditionalMetadataTagsFound <- which(additionalMetadataTags %in% colnames(record.table3))
  if(length(whichAdditionalMetadataTagsFound) < length(additionalMetadataTags)){
      if(length(whichAdditionalMetadataTagsFound) == 0) {  # if none of the additionalMetadataTags was found
        warning("metadata tag(s)  not found in image metadata:  ", paste(additionalMetadataTags, collapse = ", "), call. = FALSE)
        } else {                                                            # if only some of the additionalMetadataTags was found
        warning("metadata tag(s)  not found in image metadata:  ", paste(additionalMetadataTags[-whichAdditionalMetadataTagsFound], collapse = ", "), call. = FALSE)
      }
    }
  }


  # make column "HierarchicalSubject" the last column
  col_to_move <- which(colnames(record.table3) %in% metadata.tagname)
  if(length(col_to_move) >= 1){
	  record.table3 <- cbind(record.table3, record.table3[,col_to_move])
	  record.table3 <- record.table3[,-col_to_move]
    colnames(record.table3)[ncol(record.table3)] <- metadata.tagname
  }

  # convert to data.frame, in order to get all the column names right (: becomes .)
  record.table3 <- data.frame(record.table3, stringsAsFactors = FALSE, check.names = TRUE)
  
  # save table
  if(length(unique(record.table3[,speciesCol])) > 1){
    warning("there was more than 1 species in your images. Cannot save the record table")
    return(record.table3)
  }

  # save table
  if(writecsv == TRUE){
    outtable_filename <- paste("record_table_individuals", minDeltaTime, "min_deltaT_", Sys.Date(), ".csv", sep = "")
    if(hasArg(outDir) == FALSE){
      setwd(inDir)
    } else {
      setwd(outDir)
    }
  write.csv(record.table3, file = outtable_filename)
  }
  return(record.table3)
}

context("recordTableIndividual")

# Load necessary libraries
library(testthat)
library(withr)

# --- Test Setup ---
# Define paths to sample data once to be used in multiple tests.
wd_images_ID_individual <- system.file("pictures/sample_images_indiv_tag/LeopardCat",
                                       package = "camtrapR")

# --- Test Suite ---

testthat::describe("Core Functionality: ID from Metadata", {
  
  # All tests in this block require ExifTool
  skip_if(Sys.which("exiftool") == "", message = "ExifTool not found, skipping tests.")
  
  test_that("it replicates the main example correctly", {
    rec_table_pbe <- recordTableIndividual(
      inDir                  = wd_images_ID_individual,
      minDeltaTime           = 60,
      deltaTimeComparedTo    = "lastRecord",
      hasStationFolders      = FALSE,
      IDfrom                 = "metadata",
      metadataIDTag          = "individual",
      additionalMetadataTags = c("EXIF:Model", "EXIF:Make"),
      timeZone               = "Asia/Kuala_Lumpur"
    )
    
    # 1. Check object structure and dimensions
    expect_s3_class(rec_table_pbe, "data.frame")
    # Based on the sample data and a 60-min filter, 18 independent records are expected
    expect_equal(nrow(rec_table_pbe), 18)
    
    # 2. Check for expected columns
    expected_cols <- c("Station", "Species", "Individual", "DateTimeOriginal", "Date", "Time", "EXIF.Model", "EXIF.Make")
    expect_true(all(expected_cols %in% names(rec_table_pbe)))
    
    # 3. Check content
    expect_equal(unique(rec_table_pbe$Species), "LeopardCat")
    expect_equal(sort(unique(rec_table_pbe$Individual)), c("1", "2", "3"))
    # Check that station was correctly parsed from filename
    expect_true(all(grepl("StationA|StationB|StationC", rec_table_pbe$Station)))
  })
  
  test_that("eventSummary arguments work correctly", {
    rec_table_pbe2 <- recordTableIndividual(
      inDir                = wd_images_ID_individual,
      minDeltaTime         = 60,
      deltaTimeComparedTo  = "lastRecord",
      hasStationFolders    = FALSE,
      IDfrom               = "metadata",
      metadataIDTag        = "individual",
      timeZone             = "Asia/Kuala_Lumpur",
      eventSummaryColumn   = "metadata_individual",
      eventSummaryFunction = "max"
    )
    
    # 1. Check for the new summary column with the correct name convention
    expect_true("metadata_individual_max" %in% names(rec_table_pbe2))

  })
})


testthat::describe("Core Functionality: ID from Directory", {
  
  skip_if(Sys.which("exiftool") == "", message = "ExifTool not found, skipping tests.")
  
  test_that("it correctly identifies individuals from directory structure", {
    
    # We need to create a temporary directory structure and copy the sample files into it
    with_tempdir({
      # 1. Create the required directory structure
      species_dir <- "MyLeopardCat"
      station_dir <- file.path(species_dir, "StationX")
      indiv1_dir <- file.path(station_dir, "Cat1")
      indiv2_dir <- file.path(station_dir, "Cat2")
      dir.create(indiv1_dir, recursive = TRUE)
      dir.create(indiv2_dir, recursive = TRUE)
      
      # 2. Copy some sample images into these new directories
      # Get a list of the original sample images
      original_files <- list.files(wd_images_ID_individual, full.names = TRUE, pattern = "\\.JPG$")
      # Copy 2 images for Leo1 and 1 for Leo2
      file.copy(original_files[1:2], indiv1_dir)
      file.copy(original_files[3], indiv2_dir)
      
      # 3. Call the function on our temporary directory structure
      rec_table_dir <- recordTableIndividual(
        inDir             = species_dir,
        hasStationFolders = TRUE,
        IDfrom            = "directory", # Key argument
        minDeltaTime      = 0, # Get all records
        timeZone          = "Asia/Kuala_Lumpur"
      )
      
      # 4. Assert the output is correct
      expect_equal(nrow(rec_table_dir), 3) # We copied 3 files
      expect_equal(unique(rec_table_dir$Station), "StationX")
      expect_equal(sort(unique(rec_table_dir$Individual)), c("Cat1", "Cat2"))
      expect_equal(sum(rec_table_dir$Individual == "Cat1"), 2)
      expect_equal(sum(rec_table_dir$Individual == "Cat2"), 1)
    })
  })
})


testthat::describe("Argument Handling and Side Effects", {
  
  skip_if(Sys.which("exiftool") == "", message = "ExifTool not found, skipping tests.")
  
  test_that("cameraID and camerasIndependent arguments work", {
    # The sample filenames do not contain camera IDs, so this test is conceptual.
    # To test properly, you'd need files named e.g. "StationA__...__C1.JPG"
    
    # We test that the function doesn't error and that the camera column is/isn't created.
    rec_table_no_cam <- recordTableIndividual(inDir = wd_images_ID_individual, 
                                              IDfrom = "metadata", 
                                              metadataIDTag = "individual", 
                                              minDeltaTime = 0,
                                              hasStationFolders = F)
    expect_false("Camera" %in% names(rec_table_no_cam))
    
    # When cameraID is specified, expect a 'Camera' column, even if it's all NA
    # because the filenames don't match the required pattern.
    rec_table_with_cam <- recordTableIndividual(inDir = wd_images_ID_individual, 
                                                IDfrom = "metadata", 
                                                metadataIDTag = "individual", 
                                                minDeltaTime = 0,
                                                cameraID = "filename", 
                                                camerasIndependent = FALSE,
                                                hasStationFolders = F)
    expect_true("Camera" %in% names(rec_table_with_cam))
    expect_true(is.character(rec_table_with_cam$Camera))
  })
  
  test_that("writecsv = TRUE creates a CSV file", {
    with_tempdir({
      rec_table_to_write <- recordTableIndividual(
        inDir = wd_images_ID_individual,
        IDfrom = "metadata",
        metadataIDTag = "individual",
        minDeltaTime = 0,
        writecsv = TRUE,
        outDir = ".",  # Write to the temp directory
        hasStationFolders = F
      )
      
      # Check if a CSV file was created
      csv_files <- list.files(pattern = "\\.csv$")
      expect_equal(length(csv_files), 1)
      expect_true(grepl("record_table_individuals", csv_files[1]))
      
      # Check that the written file can be read and has the right number of rows
      written_data <- read.csv(csv_files[1])
      expect_equal(nrow(written_data), nrow(rec_table_to_write))
    })
  })
})


testthat::describe("Input Validation", {
  
  # These tests do not require ExifTool as they check arguments before it is called.
  test_that("it stops for invalid or missing arguments", {
    expect_error(recordTableIndividual(inDir = "non_existent_dir"), "Could not find inDir")
    
    expect_error(recordTableIndividual(inDir = wd_images_ID_individual, IDfrom = "metadata"),
                 regexp = '"metadataIDTag" must be defined if IDfrom = "metadata"')
    
    expect_error(recordTableIndividual(inDir = wd_images_ID_individual, IDfrom = "directory",
                                       minDeltaTime = 60),
                 regexp = "argument \"deltaTimeComparedTo\" is missing, with no default")

  })
  
  test_that("it warns when deltaTimeComparedTo is given with minDeltaTime = 0", {
    expect_warning(
      recordTableIndividual(inDir = wd_images_ID_individual, IDfrom = "metadata", metadataIDTag = "individual",
                            minDeltaTime = 0, deltaTimeComparedTo = "lastRecord", hasStationFolders = F),
      regexp = "minDeltaTime is 0. deltaTimeComparedTo = 'lastRecord' will have no effect"
    )
  })
})

#' Add or remove species names from JPEG image filenames
#' 
#' Add or remove species names from JPEG image filenames. It makes it easier to
#' find images of a species.
#' 
#' Species names can be appended or removed from image filenames. Before
#' running the function, you may want to run
#' \code{\link{checkSpeciesIdentification}} to detect possible
#' misidentifications. As an example, the function would change an image file
#' name from "StationA__2015-05-41__20-59-59(1).JPG" to
#' "StationA__2015-05-41__20-59-59(1)__Species Name.JPG". If species names were
#' appended several times by accident, they can all be removed by running the
#' function with \code{removeNames = TRUE}
#' 
#' @param inDir character. Directory containing camera trap images sorted into
#' station subdirectories (e.g. inDir/StationA/)
#' @param IDfrom character. Read species ID from image metadata ("metadata") of
#' from species directory names ("directory")?
#' @param hasCameraFolders logical. Do the station subdirectories of
#' \code{inDir} have camera-subdirectories (e.g. inDir/StationA/CameraA1;
#' inDir/StationA/CameraA2)?
#' @param metadataSpeciesTag character. The species ID tag name in image
#' metadata (if IDfrom = "metadata").
#' @param metadataHierarchyDelimitor character. The character delimiting
#' hierarchy levels in image metadata tags in field "HierarchicalSubject".
#' Either "|" or ":".
#' @param removeNames logical. remove appended species names?
#' @param writecsv logical. write csv table containing old and new file names
#' into \code{inDir}?
#' 
#' @return A \code{data.frame} containing the old and new file names and
#' directories.
#' 
#' @author Juergen Niedballa
#' 
#' @examples
#' 
#' \dontrun{
#' 
#' # copy sample images to another location (so we don't mess around in the package directory)
#' wd_images_ID <- system.file("pictures/sample_images_species_dir", package = "camtrapR")
#' file.copy(from = wd_images_ID, to = getwd(), recursive = TRUE)       
#' wd_images_ID_copy <- file.path(getwd(), "sample_images_species_dir")
#' 
#' # append species names
#' SpecNameAppend1 <- appendSpeciesNames(inDir            = wd_images_ID_copy,
#'                                       IDfrom           = "directory",
#'                                       hasCameraFolders = FALSE,
#'                                       removeNames      = FALSE,
#'                                       writecsv         = FALSE)
#'   
#' SpecNameAppend1
#' 
#' # remove species names
#' SpecNameRemove1 <- appendSpeciesNames(inDir            = wd_images_ID_copy,
#'                                       IDfrom           = "directory",
#'                                       hasCameraFolders = FALSE,
#'                                       removeNames      = TRUE,
#'                                       writecsv         = FALSE)
#'   
#' SpecNameRemove1
#' }
#' 
#' @export appendSpeciesNames
#' 
appendSpeciesNames <- function(inDir,
                               IDfrom,
                               hasCameraFolders,
                               metadataSpeciesTag,
                               metadataHierarchyDelimitor = "|",
                               removeNames = FALSE,
                               writecsv = FALSE
)
{
  wd0 <- getwd()
  on.exit(setwd(wd0))
  # check inpuit
  stopifnot(is.logical(removeNames))
  stopifnot(is.logical(writecsv))
  stopifnot(is.logical(hasCameraFolders))

  file.sep <- .Platform$file.sep

  if(!is.character(IDfrom)){stop("IDfrom must be of class 'character'")}
  if(IDfrom %in% c("metadata", "directory") == FALSE) stop("'IDfrom' must be 'metadata' or 'directory'")

  if(IDfrom == "metadata"){
    if(metadataHierarchyDelimitor %in% c("|", ":") == FALSE) stop("'metadataHierarchyDelimitor' must be '|' or ':'")
    metadata.tagname <- "HierarchicalSubject"

    if(!hasArg(metadataSpeciesTag))       stop("'metadataSpeciesTag' must be defined if IDfrom = 'metadata'")
    if(!is.character(metadataSpeciesTag)) stop("metadataSpeciesTag must be of class 'character'")
    if(length(metadataSpeciesTag) != 1)   stop("metadataSpeciesTag must be of length 1")
    if(Sys.which("exiftool") == "")       stop("cannot find ExifTool")
  }

multiple_tag_separator = "__"

  ## find station directories
  dirs       <- list.dirs(inDir, full.names = TRUE,  recursive = FALSE)
  dirs_short <- list.dirs(inDir, full.names = FALSE, recursive = FALSE)

  if(length(dirs) == 0) stop("inDir contains no station directories", call. = FALSE)

  renaming.table <- data.frame()
  nrow.metadata.tmp <- vector()


  # loop over all station directories
  for(i in 1:length(dirs)){

    if(IDfrom == "directory"){     # ID via species directories

      dirlist.i <- list.dirs(dirs[i], full.names = TRUE, recursive = FALSE)      # list directories in station directory (camera or species)


      if(hasCameraFolders == FALSE){
        filenames.by.folder <- lapply(dirlist.i,                 # list species directories
                                      FUN         = list.files,
                                      pattern     = ".jpg$|.JPG$",
                                      recursive   = TRUE,
                                      ignore.case = TRUE)
        names(filenames.by.folder) <- dirlist.i

      } else {
        dirlist.k <- list.dirs(dirlist.i, full.names = TRUE, recursive = FALSE)

        filenames.by.folder <- lapply(dirlist.k,                   # list species directories in camera directories
                                      FUN         = list.files,
                                      pattern     = ".jpg$|.JPG$",
                                      recursive   = TRUE,
                                      ignore.case = TRUE)
        names(filenames.by.folder) <- dirlist.k
        rm(dirlist.k)
      }

      rm(dirlist.i)

      folders.tmp <- rep(names(filenames.by.folder),
                         times = lapply(filenames.by.folder, FUN = length))

      species.tmp <- rep(unlist(lapply(strsplit(names(filenames.by.folder),
                                                split = file.sep,
                                                fixed = TRUE),
                                       FUN = function(X){X[length(X)]})),
                         times = lapply(filenames.by.folder, length))

      if(hasCameraFolders == TRUE){
        camera.tmp <- rep(unlist(lapply(strsplit(names(filenames.by.folder),
                                                 split = file.sep,
                                                 fixed = TRUE),
                                        FUN = function(X){X[length(X) - 1]})),
                          times = lapply(filenames.by.folder, length))
      }

      if(hasCameraFolders == FALSE){
        renaming.table <- rbind(renaming.table, data.frame(directory    = folders.tmp,
                                                           filename_old = unlist(filenames.by.folder),
                                                           species      = species.tmp))
      } else {
        renaming.table <- rbind(renaming.table, data.frame(directory    = folders.tmp,
                                                           filename_old = unlist(filenames.by.folder),
                                                           species      = species.tmp,
                                                           camera       = camera.tmp))
      }
      rownames(renaming.table) <- NULL


    } else {     #  if ID via metadata


      command.tmp  <- paste('exiftool -t -q -r -f -Directory -FileName -HierarchicalSubject -ext JPG "', dirs[i], '"', sep = "")
      colnames.tmp <- c("Directory", "FileName", "HierarchicalSubject")


      # run exiftool and make data frame
      metadata.tmp <- runExiftool(command.tmp = command.tmp, colnames.tmp = colnames.tmp)


      if(is.data.frame(metadata.tmp)){
        # if(IDfrom == "directory"){
          # message(paste(dirs_short[i], ": ", nrow(metadata.tmp), "images"))
        # }

        # store number of images
        nrow.metadata.tmp[[i]] <- nrow(metadata.tmp)

        # add metadata
        metadata.tmp <- addMetadataAsColumns (intable                    = metadata.tmp,
                                              metadata.tagname           = metadata.tagname,
                                              metadataHierarchyDelimitor = metadataHierarchyDelimitor,
                                              multiple_tag_separator     = multiple_tag_separator)

        # assign species ID
        metadata.tmp <- assignSpeciesID (intable                = metadata.tmp,
                                         IDfrom                 = "metadata",
                                         metadataSpeciesTag     = metadataSpeciesTag,
                                         speciesCol             = "species",
                                         dirs_short             = dirs_short,
                                         i_tmp                  = i,
                                         multiple_tag_separator = "_&_",   # this is different from the one defined above to prevent (!) separating multiple entries in the same image: Species will be something like "Leopard Cat__Malay Badger"
                                         returnFileNamesMissingTags = FALSE)

      # if images in station contain not metadata species tags, skip that station
      if(!is.data.frame(metadata.tmp)){
        if(metadata.tmp == "found no species tag") {
          # warning(paste(dirs_short[i], ":   metadataSpeciesTag '", metadataSpeciesTag, "' not found in image metadata tag 'HierarchicalSubject'. Skipping", sep = ""), call. = FALSE, immediate. = TRUE)
        } else {
          warning(paste(dirs_short[i], ":   error in species tag extraction. Skipping. Please report this bug", sep = ""), call. = FALSE, immediate. = TRUE)
        }
        next
      }
        # assign camera ID
        if(hasCameraFolders == TRUE){
          if(IDfrom == "directory"){
            metadata.tmp$camera  <- sapply(strsplit(metadata.tmp$Directory, split = file.sep, fixed = TRUE), FUN = function(X){X[length(X) - 1]})
          } else {
            metadata.tmp$camera  <- sapply(strsplit(metadata.tmp$Directory, split = file.sep, fixed = TRUE), FUN = function(X){X[length(X)]})
          }
        }

        # add station ID and assemble table
        metadata.tmp <- cbind(station = rep(dirs_short[i], times = nrow(metadata.tmp)),
                              metadata.tmp)



        if(hasCameraFolders == TRUE){
          renaming.table <- rbind(renaming.table, data.frame(directory    = metadata.tmp$Directory,
                                                             filename_old = metadata.tmp$FileName,
                                                             species      = metadata.tmp$species,
                                                             camera       = metadata.tmp$camera))
        } else {
          renaming.table <- rbind(renaming.table, data.frame(directory    = metadata.tmp$Directory,
                                                             filename_old = metadata.tmp$FileName,
                                                             species      = metadata.tmp$species))
        }
        rownames(renaming.table) <- NULL

      } # end if(is.data.frame(metadata.tmp)){
    }   # end ID via metadata
  }     # end for (i...)

      # construct new filename
      if(removeNames == FALSE){
        renaming.table$filename_new <- paste(paste(gsub(pattern     = ".jpg|.JPG$",
                                                        replacement = "",
                                                        x           = renaming.table$filename_old,
                                                        ignore.case = TRUE),
                                                   renaming.table$species,
                                                   sep   = "__"),
                                             ".JPG", sep = "")
      }
      if(removeNames == TRUE){
        # find all the species names ("__SpeciesName") and remove
        # it not  work if you changed species ID after assigning species names
        renaming.table$filename_new <- gsub(pattern     = paste(paste("__", unique(renaming.table$species),
                                                                      sep      = ""),
                                                                      collapse = "|"),
                                            replacement = "",
                                            x           = renaming.table$filename_old,
                                            ignore.case = TRUE)
      }


  # rename
  file.rename(from = file.path(renaming.table$directory, renaming.table$filename_old),
              to   = file.path(renaming.table$directory, renaming.table$filename_new))

  renaming.table$renamed <- renaming.table$filename_old != renaming.table$filename_new

  # write outtable
  if(writecsv == TRUE){
    if(removeNames == TRUE){
      filename.tmp <- paste("renaming_table_UNDO_", Sys.Date(), ".csv", sep = "")
    } else {
      filename.tmp <- paste("renaming_table_", Sys.Date(), ".csv", sep = "")
    }
    setwd(inDir)
    write.csv(renaming.table, file = filename.tmp, row.names = FALSE)
  }



  if(IDfrom == "directory"){
    message(paste("renamed", sum(renaming.table$renamed), "out of", nrow(renaming.table), "images in", inDir))
  } else {
    message(paste("renamed", sum(renaming.table$renamed),
      "out of", nrow(renaming.table), "images with species ID out of", sum(unlist(nrow.metadata.tmp), na.rm = TRUE), "images in", inDir))
  }

  return(renaming.table)

}

#' Consistency check on species image identification
#' 
#' This function serves 2 purposes: 1) it assesses possible misidentification
#' of species and 2) compares double observer species identification (only if
#' metadata tagging was used for species identification).
#' 
#' Within each station, it assesses whether there are images of a species taken
#' within a given time interval of another species. Often, it is unlikely that
#' different species are encountered within a very short time intervals at the
#' same location. This type of misidentification can arise easily if some
#' images belonging to a sequence of images were accidentally moved into
#' different species directories or tagged incorrectly.
#' 
#' Double observer identification may be desirable to increase reliability of
#' species identification. The function returns conflicts in species
#' identification between 2 observers. These conflicts can then be corrected.
#' 
#' Images may accidentally be misidentified by assigning wrong species tags or
#' by moving them into wrong species directories. Imagine your cameras take
#' sequences of images each time they are triggered and one image of the
#' sequence is misidentified. The time difference between these images (that
#' have different species assigned to them) will be very small, usually a few
#' seconds. This function will return all these images for you to check if they
#' were identified correctly.
#' 
#' If multiple observers identify images independently using metadata tagging,
#' their identifications can be compared by setting
#' \code{metadataSpeciesTagToCompare}. Conflicting or missing identifications
#' will be reported. This feature is only available if images were identified
#' by metadata tagging.
#' 
#' Species like "blank" or "team" can be ignored using \code{excludeSpecies}.
#' If only specific stations are to be checked, \code{stationsToCheck} can be
#' set.
#' 
#' @param inDir character. Directory containing identified camera trap images
#' sorted into station subdirectories (e.g. inDir/StationA/)
#' @param IDfrom character. Read species ID from image metadata ("metadata") of
#' from species directory names ("directory")?
#' @param hasCameraFolders logical. Do the station directories in \code{inDir}
#' have camera subdirectories (e.g. "inDir/StationA/Camera1" or
#' "inDir/StationA/Camera1/Species1")?
#' @param metadataSpeciesTag character. The species ID tag name in image
#' metadata (if IDfrom = "metadata").
#' @param metadataSpeciesTagToCompare character. A second species ID tag name
#' in image metadata (if IDfrom = "metadata"). For comparing double observer
#' species identification.
#' @param metadataHierarchyDelimitor character. The character delimiting
#' hierarchy levels in image metadata tags in field "HierarchicalSubject".
#' Either "|" or ":"
#' @param maxDeltaTime numeric. Maximum time interval between images to be
#' returned (in seconds)
#' @param excludeSpecies character. vector of species to exclude from checks
#' @param stationsToCheck character. vector of stations to be checked
#' (optionally)
#' @param writecsv logical. Should the resulting data.frame be saved as a .csv?
#' 
#' @return A \code{list} containing 2 data frames. The first contains a data
#' frame with images file names, directories, time stamp and species ID that
#' were taken within \code{maxDeltaTime} seconds of another species image at a
#' particular station. The second data frame contains images with conflicting
#' species IDs (if \code{IDfrom = "metadata"} and
#' \code{metadataSpeciesTagToCompare} is defined)
#' 
#' @note The function will not be able to find "isolated" images, i.e. images
#' that were misidentified, but were not part of a sequence of images.
#' Likewise, if all images of a sequence were misidentified, they cannot be
#' found either. From version 0.99.0, the function can also handle images
#' identied with metadata tags.
#' 
#' @author Juergen Niedballa
#' 
#' @examples
#' 
#' 
#' \donttest{
#' wd_images_ID <- system.file("pictures/sample_images_species_dir", package = "camtrapR")
#' 
#' if (Sys.which("exiftool") != ""){        # only run this example if ExifTool is available
#' check.folders <- checkSpeciesIdentification(inDir             = wd_images_ID,
#'                                             IDfrom            = "directory",
#'                                             hasCameraFolders  = FALSE,
#'                                             maxDeltaTime      = 120,
#'                                             writecsv          = FALSE)
#'                                      
#' check.folders   # In the example, 2 different species were photographed within 2 minutes. 
#' }
#' }
#' 
#' \dontrun{
#' # now exclude one of these 2 species 
#' check.folders2 <- checkSpeciesIdentification(inDir             = wd_images_ID,
#'                                              IDfrom            = "directory",
#'                                              hasCameraFolders  = FALSE,
#'                                              maxDeltaTime      = 120,
#'                                              excludeSpecies    = "EGY",
#'                                              writecsv          = FALSE)           
#'   
#' check.folders2   # the data frame is empty
#' 
#' # now we check only one station
#' check.folders3 <- checkSpeciesIdentification(inDir             = wd_images_ID,
#'                                              IDfrom            = "directory",
#'                                              hasCameraFolders  = FALSE,
#'                                              maxDeltaTime      = 120,
#'                                              stationsToCheck   = "StationB",
#'                                              writecsv          = FALSE)
#' check.folders3   # the data frame is empty
#' }
#' 
#' 
#' @export checkSpeciesIdentification
#' 
checkSpeciesIdentification <- function(inDir,
                                       IDfrom,
                                       hasCameraFolders,
                                       metadataSpeciesTag,
                                       metadataSpeciesTagToCompare,
                                       metadataHierarchyDelimitor = "|",
                                       maxDeltaTime,
                                       excludeSpecies,
                                       stationsToCheck,
                                       writecsv = FALSE
)
{
  wd0 <- getwd()
  on.exit(setwd(wd0))

  if(Sys.which("exiftool") == "") stop("cannot find ExifTool")
  if(hasArg(excludeSpecies)){
    if(!is.character(excludeSpecies)) stop("excludeSpecies must be of class 'character'")
  }
  if(hasArg(stationsToCheck)){
    if(!is.character(stationsToCheck)) stop("stationsToCheck must be of class 'character'")
  }
  stopifnot(is.logical(hasCameraFolders))

  stopifnot(is.numeric(maxDeltaTime))

  file.sep <- .Platform$file.sep

 if(!is.character(IDfrom)){stop("IDfrom must be of class 'character'")}
 if(IDfrom %in% c("metadata", "directory") == FALSE) stop("'IDfrom' must be 'metadata' or 'directory'")

 if(IDfrom == "metadata"){
    if(metadataHierarchyDelimitor %in% c("|", ":") == FALSE) stop("'metadataHierarchyDelimitor' must be '|' or ':'")
    metadata.tagname <- "HierarchicalSubject"

    if(!hasArg(metadataSpeciesTag)) {stop("'metadataSpeciesTag' must be defined if IDfrom = 'metadata'")}
    if(!is.character(metadataSpeciesTag)){stop("metadataSpeciesTag must be of class 'character'")}
    if(length(metadataSpeciesTag) != 1){stop("metadataSpeciesTag must be of length 1")}

    if(hasArg(metadataSpeciesTagToCompare)) {
      if(!is.character(metadataSpeciesTagToCompare)){stop("metadataSpeciesTagToCompare must be of class 'character'")}
      if(length(metadataSpeciesTagToCompare) != 1){stop("metadataSpeciesTagToCompare must be of length 1")}
    }
  }

  multiple_tag_separator <- "_&_"

  if(!dir.exists(inDir)) stop("Could not find inDir:\n", inDir, call. = FALSE)
  # find station directories
  dirs       <- list.dirs(inDir, full.names = TRUE, recursive = FALSE)
  dirs_short <- list.dirs(inDir, full.names = FALSE, recursive = FALSE)

  if(length(dirs) == 0) stop("inDir contains no station directories", call. = FALSE)

  check_table <- conflict_ID_table <-  data.frame(stringsAsFactors = FALSE)

  # if only checking certain station, subset dirs/dirs_short
  if(hasArg(stationsToCheck)){
    whichStationToCheck <- which(dirs_short %in% stationsToCheck)
    if(length(whichStationToCheck) == 0) {stop("found no directories of names specified in stationsToCheck")} else {
      dirs       <- dirs      [whichStationToCheck]
      dirs_short <- dirs_short[whichStationToCheck]
    }
  }


  for(i in 1:length(dirs)){

    if(IDfrom == "directory"){
      dirs.to.check.sho <- list.dirs(dirs[i], full.names = FALSE)[-1]
      dirs.to.check     <- list.dirs(dirs[i], full.names = TRUE)[-1]
      if(hasArg(excludeSpecies)){
        dirs.to.check     <- dirs.to.check    [!dirs.to.check.sho %in% excludeSpecies]
        dirs.to.check.sho <- dirs.to.check.sho[!dirs.to.check.sho %in% excludeSpecies]
      }
    }

    # remove empty species directories
#     empty_dirs <- sapply(dirs.to.check, FUN = function(X){length(list.files(X)) == 0})
#     if(any(empty_dirs)){
#       dirs.to.check <- dirs.to.check[-empty_dirs]
#       dirs.to.check.sho <- dirs.to.check.sho[-empty_dirs]
#     }

    # create command line for exiftool execution

    if(IDfrom == "directory"){
      if(hasArg(excludeSpecies)) {   # under some rare circumstances, this caused an error if directories were empty
        command.tmp <- paste('exiftool -t -q -r -f -Directory -FileName -EXIF:DateTimeOriginal -HierarchicalSubject -ext JPG "', paste(dirs.to.check, collapse = '" "'), '"', sep = "")
      } else {
        command.tmp <- paste('exiftool -t -q -r -f -Directory -FileName -EXIF:DateTimeOriginal -HierarchicalSubject -ext JPG "', dirs[i], '"', sep = "")
      }
    } else {
      command.tmp <- paste('exiftool -t -q -r -f -Directory -FileName -EXIF:DateTimeOriginal -HierarchicalSubject -ext JPG "', dirs[i], '"', sep = "")
    }
    colnames.tmp <- c("Directory", "FileName", "DateTimeOriginal", "HierarchicalSubject")

    # run exiftool and make data frame
    metadata.tmp <- runExiftool(command.tmp = command.tmp, colnames.tmp = colnames.tmp)


    if(inherits(metadata.tmp, "data.frame")){
		if(IDfrom == "directory"){
      message(paste(dirs_short[i], ": checking", nrow(metadata.tmp), "images in", length(dirs.to.check.sho), "directories"))
		}
      # write metadata from HierarchicalSubject field to individual columns
      if(IDfrom == "metadata"){
        message(paste(dirs_short[i], ": ", formatC(nrow(metadata.tmp), width = 4), " images", 
                      makeProgressbar(current = i, total = length(dirs_short)), sep = ""))
        
        metadata.tmp <- addMetadataAsColumns (intable                    = metadata.tmp,
                                              metadata.tagname           = metadata.tagname,
                                              metadataHierarchyDelimitor = metadataHierarchyDelimitor,
                                              multiple_tag_separator     = multiple_tag_separator)
      }


      # assign species ID
      metadata.tmp <- assignSpeciesID (intable                = metadata.tmp,
                                       IDfrom                 = IDfrom,
                                       metadataSpeciesTag     = metadataSpeciesTag,
                                       speciesCol             = "species",
                                       dirs_short             = dirs_short,
                                       i_tmp                  = i,
                                       multiple_tag_separator = multiple_tag_separator,
                                       returnFileNamesMissingTags = FALSE
      )

      # if images in station contain no metadata species tags, skip that station
      if(!inherits(metadata.tmp, "data.frame")){
        if(metadata.tmp == "found no species tag") {
          warning(paste(dirs_short[i], ":   metadataSpeciesTag '", metadataSpeciesTag, "' not found in image metadata tag 'HierarchicalSubject'. Skipping", sep = ""), call. = FALSE, immediate. = TRUE)
        } else {
          warning(paste(dirs_short[i], ":   error in species tag extraction. Skipping. Please report", sep = ""), call. = FALSE, immediate. = TRUE)
        }
        next
      }

      # exclude species if using metadata tags (if using IDfrom = "directory", they were removed above already)
      if(IDfrom == "metadata"){
        if(hasArg(excludeSpecies)){
          metadata.tmp <- metadata.tmp[!metadata.tmp$species %in% excludeSpecies,]
        }
      }

      # assign camera ID
      if(IDfrom == "directory" & hasCameraFolders == TRUE){
        metadata.tmp$camera  <- sapply(strsplit(metadata.tmp$Directory, split = file.sep, fixed = TRUE), FUN = function(X){X[length(X) - 1]})
      }
      if(IDfrom == "metadata" & hasCameraFolders == TRUE){
        metadata.tmp$camera  <- sapply(strsplit(metadata.tmp$Directory, split = file.sep, fixed = TRUE), FUN = function(X){X[length(X)]})
      }


      # make date/time R-readable
      metadata.tmp$DateTimeOriginal <- as.POSIXct(strptime(x = metadata.tmp$DateTimeOriginal, format = "%Y:%m:%d %H:%M:%S"))

      # add station ID and assemble table
      metadata.tmp <- cbind(station = rep(dirs_short[i], times = nrow(metadata.tmp)),
                            metadata.tmp)


      # compare ID between different observers
      if(hasArg(metadataSpeciesTagToCompare)){
        metadataSpeciesTag2 <- paste("metadata", metadataSpeciesTag, sep = "_")
        metadataSpeciesTagToCompare2 <- paste("metadata", metadataSpeciesTagToCompare, sep = "_")
        if(metadataSpeciesTagToCompare2 %in% colnames(metadata.tmp)){
          metadata.tmp.conflict <- metadata.tmp[metadata.tmp[,metadataSpeciesTag2] != metadata.tmp[,metadataSpeciesTagToCompare2] |
                                                is.na(metadata.tmp[,metadataSpeciesTag2] != metadata.tmp[,metadataSpeciesTagToCompare2]) ,]
          metadata.tmp.conflict <- metadata.tmp.conflict[,which(colnames(metadata.tmp.conflict) %in% c("station", "Directory", "FileName", metadataSpeciesTag2, metadataSpeciesTagToCompare2))]
          # if anything to report, append to main table
          if(nrow(metadata.tmp.conflict) >= 1){
            conflict_ID_table <- rbind(conflict_ID_table, metadata.tmp.conflict)
          }
        } else {warning(paste("metadata tag '", metadataSpeciesTagToCompare, "' was not found in image metadata in Station ", dirs_short[i], sep = ""), call. = FALSE, immediate. = TRUE)}
        suppressWarnings(rm(metadataSpeciesTag2, metadataSpeciesTagToCompare2, metadata.tmp.conflict))
      }

      # calculate minimum delta time between image and all images in other species folders at station i
      if(length(unique(metadata.tmp$species)) >= 2){

        for(rowindex in 1:nrow(metadata.tmp)){

          if(hasCameraFolders == TRUE){
            # only compare within a camera folder if there was >1 camera per station
            which.tmp1 <- which(metadata.tmp$species != metadata.tmp$species[rowindex] &
                                metadata.tmp$camera  == metadata.tmp$camera[rowindex])
            if(length(which.tmp1) >= 1){
              metadata.tmp$min.delta.time[rowindex] <- round(min(abs(difftime(time1 = metadata.tmp$DateTimeOriginal[rowindex],
                                                                              time2 = metadata.tmp$DateTimeOriginal[which.tmp1],
                                                                              units = "secs"))))
            } else {
              metadata.tmp$min.delta.time[rowindex] <- NA
            }
            rm(which.tmp1)
          } else {         # if no camera subfolders
            # compare to other species
            which.tmp2 <- which(metadata.tmp$species != metadata.tmp$species[rowindex])
            if(length(which.tmp2) >= 1){
              metadata.tmp$min.delta.time[rowindex] <- round(min(abs(difftime(time1 = metadata.tmp$DateTimeOriginal[rowindex],
                                                                              time2 = metadata.tmp$DateTimeOriginal[which.tmp2],
                                                                              units = "secs"))))
            } else {
              metadata.tmp$min.delta.time[rowindex] <- NA
            }
            rm(which.tmp2)
          } # end ifelse hasCameraFolders
        }   # end for

        if(hasCameraFolders == TRUE){
          check_table_tmp <- metadata.tmp[metadata.tmp$min.delta.time <= maxDeltaTime & !is.na(metadata.tmp$min.delta.time), c("station", "Directory", "FileName", "species", "DateTimeOriginal", "camera")]
        } else {
          check_table_tmp <- metadata.tmp[metadata.tmp$min.delta.time <= maxDeltaTime & !is.na(metadata.tmp$min.delta.time), c("station", "Directory", "FileName", "species", "DateTimeOriginal")]
        }
        # order output
        check_table_tmp <- check_table_tmp[order(check_table_tmp$DateTimeOriginal),]

        # if anything to report, append to main table
        if(nrow(check_table_tmp) >= 1){
          check_table <- rbind(check_table, check_table_tmp)
        }
        suppressWarnings(rm(metadata.tmp, check_table_tmp))

      }  # end  if(length(unique(metadata.tmp$species)) >= 2){
    }    # end  if(inherits(metadata.tmp, "data.frame")){
  }      # end for (i ...)

  if(writecsv == TRUE){
    check_table_filename <- paste("species_ID_check_", Sys.Date(), ".csv", sep = "")
    conflict_table_filename <- paste("species_ID_conflicts_", Sys.Date(), ".csv", sep = "")
    setwd(inDir)
    write.csv(check_table, file = check_table_filename)
    write.csv(conflict_ID_table, file = conflict_table_filename)
  }

  # make output list
    outlist <- list(check_table, conflict_ID_table)
    names(outlist) <- c("temporalIndependenceCheck", "IDconflictCheck")

  return(outlist)

}

test_that("Input argument validation works", {
  
  testthat::skip_if_not_installed("mockery")
  
  #  Add cycle = TRUE because the function is called 3 times in this block
  m_sys_which <- mockery::mock("/usr/bin/exiftool", cycle = TRUE)
  mockery::stub(checkSpeciesIdentification, "Sys.which", m_sys_which)
  
  # Create a dummy directory to pass dir.exists check
  temp_dir <- tempdir()
  
  # 1. Invalid IDfrom
  expect_error(checkSpeciesIdentification(inDir = temp_dir, 
                                          IDfrom = "wrong_option", 
                                          hasCameraFolders = FALSE, 
                                          maxDeltaTime = 120),
               "'IDfrom' must be 'metadata' or 'directory'")
  
  # 2. Missing metadataSpeciesTag when IDfrom = "metadata"
  expect_error(checkSpeciesIdentification(inDir = temp_dir, 
                                          IDfrom = "metadata", 
                                          hasCameraFolders = FALSE, 
                                          maxDeltaTime = 120),
               "'metadataSpeciesTag' must be defined if IDfrom = 'metadata'")
  
  # 3. Invalid types
  expect_error(checkSpeciesIdentification(inDir = temp_dir, 
                                          IDfrom = "directory", 
                                          hasCameraFolders = FALSE, 
                                          maxDeltaTime = "Not a number"),
               "is.numeric")
})

test_that("Logic: ID from Directory - Temporal Independence Check", {
  testthat::skip_if_not_installed("mockery")
  # --- Setup ---
  temp_dir <- file.path(tempdir(), "test_station_dir")
  dir.create(temp_dir, showWarnings = FALSE)
  dir.create(file.path(temp_dir, "StationA"), showWarnings = FALSE)
  # Create dummy species folders so list.dirs works
  dir.create(file.path(temp_dir, "StationA", "SpeciesA"), showWarnings = FALSE)
  dir.create(file.path(temp_dir, "StationA", "SpeciesB"), showWarnings = FALSE)
  on.exit(unlink(temp_dir, recursive = TRUE))
  
  # --- Mocks ---
  m_sys_which <- mockery::mock("found")
  
  # Mock runExiftool return data
  mock_exif_data <- data.frame(
    Directory = c(file.path(temp_dir, "StationA", "SpeciesA"), 
                  file.path(temp_dir, "StationA", "SpeciesB")),
    FileName = c("img1.jpg", "img2.jpg"),
    DateTimeOriginal = c("2023:01:01 12:00:00", "2023:01:01 12:00:10"),
    HierarchicalSubject = c("", ""),
    stringsAsFactors = FALSE
  )
  m_runExiftool <- mockery::mock(mock_exif_data)
  
  # Mock assignSpeciesID
  fake_assign_species <- function(intable, ...) {
    intable$species <- basename(intable$Directory)
    return(intable)
  }
  
  # --- Stubbing ---
  mockery::stub(checkSpeciesIdentification, "Sys.which", m_sys_which)
  mockery::stub(checkSpeciesIdentification, "runExiftool", m_runExiftool)
  mockery::stub(checkSpeciesIdentification, "assignSpeciesID", fake_assign_species)
  
  # --- Execution ---
  result <- checkSpeciesIdentification(
    inDir = temp_dir,
    IDfrom = "directory",
    hasCameraFolders = FALSE,
    maxDeltaTime = 60 
  )
  
  # --- Assertions ---
  expect_type(result, "list")
  expect_named(result, c("temporalIndependenceCheck", "IDconflictCheck"))
  
  check_table <- result$temporalIndependenceCheck
  expect_equal(nrow(check_table), 2)
  expect_equal(check_table$species, c("SpeciesA", "SpeciesB"))
  expect_equal(check_table$station, c("StationA", "StationA"))
})

test_that("Logic: ID from Metadata - Double Observer Conflict Check", {
  
  testthat::skip_if_not_installed("mockery")
  
  # --- Setup ---
  temp_dir <- file.path(tempdir(), "test_station_meta")
  dir.create(temp_dir, showWarnings = FALSE)
  dir.create(file.path(temp_dir, "StationB"), showWarnings = FALSE)
  on.exit(unlink(temp_dir, recursive = TRUE))
  
  # --- Mocks ---
  m_sys_which   <- mockery::mock("found")
  m_progressbar <- mockery::mock(NULL)
  
  mock_exif_data <- data.frame(
    Directory = file.path(temp_dir, "StationB"),
    FileName = "img_conflict.jpg",
    DateTimeOriginal = "2023:01:01 12:00:00",
    HierarchicalSubject = "Species|Fox_&_Compare|Wolf", 
    stringsAsFactors = FALSE
  )
  m_runExiftool <- mockery::mock(mock_exif_data)
  
  fake_add_meta <- function(intable, ...) {
    intable$metadata_SpeciesTag <- "Fox"
    intable$metadata_CompareTag <- "Wolf"
    return(intable)
  }
  
  fake_assign_species <- function(intable, ...) {
    intable$species <- intable$metadata_SpeciesTag
    return(intable)
  }
  
  # --- Stubbing ---
  mockery::stub(checkSpeciesIdentification, "Sys.which", m_sys_which)
  mockery::stub(checkSpeciesIdentification, "runExiftool", m_runExiftool)
  mockery::stub(checkSpeciesIdentification, "makeProgressbar", m_progressbar)
  mockery::stub(checkSpeciesIdentification, "addMetadataAsColumns", fake_add_meta)
  mockery::stub(checkSpeciesIdentification, "assignSpeciesID", fake_assign_species)
  
  # --- Execution ---
  result <- checkSpeciesIdentification(
    inDir = temp_dir,
    IDfrom = "metadata",
    hasCameraFolders = FALSE,
    metadataSpeciesTag = "SpeciesTag",
    metadataSpeciesTagToCompare = "CompareTag",
    maxDeltaTime = 60
  )
  
  # --- Assertions ---
  conflict_table <- result$IDconflictCheck
  expect_equal(nrow(conflict_table), 1)
  expect_true("metadata_SpeciesTag" %in% colnames(conflict_table))
  expect_equal(conflict_table$metadata_SpeciesTag, "Fox")
  expect_equal(conflict_table$metadata_CompareTag, "Wolf")
  expect_equal(nrow(result$temporalIndependenceCheck), 0)
})

test_that("Logic: Exclude Species functionality", {
  
  testthat::skip_if_not_installed("mockery")
  
  # --- Setup ---
  temp_dir <- file.path(tempdir(), "test_exclude")
  dir.create(temp_dir, showWarnings = FALSE)
  dir.create(file.path(temp_dir, "StationC"), showWarnings = FALSE)
  dir.create(file.path(temp_dir, "StationC", "Human"), showWarnings = FALSE) 
  dir.create(file.path(temp_dir, "StationC", "Deer"), showWarnings = FALSE)
  on.exit(unlink(temp_dir, recursive = TRUE))
  
  # --- Mocks ---
  m_sys_which <- mockery::mock("found")
  
  mock_exif_data <- data.frame(
    Directory = file.path(temp_dir, "StationC", "Deer"),
    FileName = "deer.jpg",
    DateTimeOriginal = "2023:01:01 12:00:00",
    HierarchicalSubject = "",
    stringsAsFactors = FALSE
  )
  m_runExiftool <- mockery::mock(mock_exif_data)
  
  fake_assign_species <- function(intable, ...) {
    intable$species <- "Deer"
    return(intable)
  }
  
  mockery::stub(checkSpeciesIdentification, "Sys.which", m_sys_which)
  mockery::stub(checkSpeciesIdentification, "runExiftool", m_runExiftool)
  mockery::stub(checkSpeciesIdentification, "assignSpeciesID", fake_assign_species)
  
  # --- Execution ---
  result <- checkSpeciesIdentification(
    inDir = temp_dir,
    IDfrom = "directory",
    hasCameraFolders = FALSE,
    maxDeltaTime = 60,
    excludeSpecies = "Human"
  )
  
  # --- Assertions ---
  expect_equal(nrow(result$temporalIndependenceCheck), 0)
  
  # Verify runExiftool command string excluded Human
  args_passed <- mockery::mock_args(m_runExiftool)[[1]]
  command_str <- args_passed$command.tmp
  expect_false(grepl("Human", command_str))
  expect_true(grepl("Deer", command_str))
})

test_that("Logic: hasCameraFolders = TRUE logic", {
  
  testthat::skip_if_not_installed("mockery")
  
  # --- Setup ---
  temp_dir <- file.path(tempdir(), "test_cam_folders")
  dir.create(temp_dir, showWarnings = FALSE)
  dir.create(file.path(temp_dir, "StationD"), showWarnings = FALSE)
  on.exit(unlink(temp_dir, recursive = TRUE))
  
  m_sys_which <- mockery::mock("found")
  
  mock_exif_data <- data.frame(
    Directory = c(file.path(temp_dir, "StationD", "Cam1", "Fox"),
                  file.path(temp_dir, "StationD", "Cam1", "Deer"),
                  file.path(temp_dir, "StationD", "Cam2", "Deer")),
    FileName = c("fox1.jpg", "deer1.jpg", "deer2.jpg"),
    DateTimeOriginal = c("2023:01:01 12:00:00", 
                         "2023:01:01 12:00:05", 
                         "2023:01:01 12:00:05"),
    HierarchicalSubject = "",
    stringsAsFactors = FALSE
  )
  m_runExiftool <- mockery::mock(mock_exif_data)
  
  fake_assign_species <- function(intable, ...) {
    intable$species <- basename(intable$Directory)
    return(intable)
  }
  
  mockery::stub(checkSpeciesIdentification, "Sys.which", m_sys_which)
  mockery::stub(checkSpeciesIdentification, "runExiftool", m_runExiftool)
  mockery::stub(checkSpeciesIdentification, "assignSpeciesID", fake_assign_species)
  
  # --- Execution ---
  result <- checkSpeciesIdentification(
    inDir = temp_dir,
    IDfrom = "directory",
    hasCameraFolders = TRUE,
    maxDeltaTime = 60
  )
  
  # --- Assertions ---
  check_table <- result$temporalIndependenceCheck
  expect_true("camera" %in% colnames(check_table))
  expect_equal(nrow(check_table), 2)
  expect_true(all(check_table$camera == "Cam1"))
  expect_false("Cam2" %in% check_table$camera)
})

#' Check species names against the ITIS taxonomic database
#' 
#' The function checks species names (common or scientific names) provided by
#' the user with the ITIS taxonomic database (\url{https://www.itis.gov/}) via
#' functions from the package \pkg{taxize}. It returns both common and
#' scientific names, the taxon authors, taxon rank name and status, the TSN
#' (taxonomic serial numbers) and ITIS urls.
#' 
#' Arguments \code{searchtype}, \code{accepted} and \code{ask} are passed on to
#' \code{\link[taxize]{get_tsn}}.
#' 
#' @param speciesNames character. Vector of species names to check. Either
#' common names or scientific names.
#' @param searchtype character. Type of names specified in \code{speciesNames}.
#' One of 'scientific' or 'common'.
#' @param accepted logical. Return only accepted valid names? If TRUE, invalid
#' names are returned as NA. Set to FALSE to return both accepted and
#' unaccepted names.
#' @param ask logical. Should the function be run in interactive mode? If TRUE
#' and more than one TSN is found for a species, the user is asked to choose
#' one. If FALSE, NA is returned for multiple matches.
#' @return A \code{data.frame} with the names supplied by the user, matching
#' common and scientific names, taxon author and year, taxonomic rank, status,
#' TSNs (taxonomic serial numbers) and ITIS urls.
#' 
#' @author Juergen Niedballa
#' 
#' @references \url{https://www.itis.gov/}
#' 
#' @examples
#' 
#' 
#' \dontrun{
#' 
#' 
#' species_common <- c("Leopard Cat", "moonrat")
#' 
#' # ask = TRUE. Multiple matches for leopard cat will cause menu to pop up asking user input.
#' 
#' species.names.check1 <- checkSpeciesNames(speciesNames = species_common,
#'                                           searchtype   = "common",
#'                                           accepted     = TRUE,
#'                                           ask          = TRUE)
#' 2   # we choose entry 2
#' species.names.check1
#' 
#' 
#' # ask = FALSE. Multiple matches for leopard cat will cause NA.
#' 
#' species.names.check2 <- checkSpeciesNames(speciesNames = species_common,
#'                                           searchtype   = "common",
#'                                           accepted     = TRUE,
#'                                           ask          = FALSE)
#' species.names.check2
#' 
#' 
#' 
#' # search for scientific names
#' 
#' species_scientific <- c("Tragulus", "Prionailurus bengalensis")
#' 
#' species.names.check3 <- checkSpeciesNames(speciesNames = species_scientific,
#'                                           searchtype   = "scientific",
#'                                           accepted     = TRUE,
#'                                           ask          = TRUE)
#' species.names.check3
#' }
#' 
#' 
#' @export checkSpeciesNames
#' 
checkSpeciesNames <- function(speciesNames,
                              searchtype,
                              accepted = TRUE,
                              ask = TRUE
){

  if (!requireNamespace("taxize", quietly = TRUE)) {
    stop("Please install the package taxize to run this function")
  }
  if (!requireNamespace("ritis", quietly = TRUE)) {
    stop("Please install the package ritis to run this function")
  }
  # check input
  searchtype <- match.arg(searchtype, choices =  c("scientific", "common"))
  stopifnot(is.logical(accepted))
  stopifnot(is.character(speciesNames) | is.factor(speciesNames))
  speciesNames <- unique(as.character(speciesNames))

  file.sep <- .Platform$file.sep

  # query ITIS TSN (taxnonomic serial number)
  tsns <- try(
    taxize::get_tsn(sci_com    = speciesNames,
                    searchtype = searchtype,
                    accepted   = accepted,
                    ask        = ask,
                    messages   = FALSE)
  )
  
  # # NOTE: possible alternative
  # if(searchtype == "common") {
  #   search_common(speciesNames)
  # }
  # if(searchtype == "scientific") {
  #   search_common(search_scientific)
  # }
  
  
  # if try returned an error, end function
  if(inherits(tsns, "try-error")) {
    message(paste("error in get_tsn. Exiting without results:\n", 
                  tsns, sep = ""))
    return(invisible(NULL))
  }
  

  tsns <- taxize::as.tsn(unique(tsns), check = FALSE)    # remove duplicates

  # warning if a name was not found
  if(any(is.na(tsns))){
    not.matched <- which(is.na(tsns))
    warning(paste("found no matches for", length(not.matched), "name(s):\n",  paste(speciesNames[not.matched], collapse = ", ")), immediate. = TRUE, call. = FALSE)
    tsns_worked <- taxize::as.tsn(tsns[-not.matched], check = FALSE)
  } else {
    tsns_worked <- tsns
  }

  if(length(tsns_worked) >= 1){

    scientific <- common <- author <- rankname  <- taxon_status <- data.frame(matrix(NA,
                                                                                     nrow = length(tsns_worked),
                                                                                     ncol = 2),
                                                                              stringsAsFactors = FALSE)

    colnames(scientific)   <- c("tsn", "combinedname")
    colnames(common)       <- c("tsn", "commonName")
    colnames(author)       <- c("tsn", "authorship")
    colnames(rankname)     <- c("tsn", "rankname")
    colnames(taxon_status) <- c("tsn", "taxonUsageRating")


    for(i in 1:length(tsns_worked)){

      scientific_tmp   <- ritis::scientific_name  (tsns_worked[i])   # get scientific names
      common_tmp       <- ritis::common_names     (tsns_worked[i])   # get common names
      author_tmp       <- ritis::taxon_authorship (tsns_worked[i])   # get author
      rankname_tmp     <- ritis::rank_name        (tsns_worked[i])   # get rank name

      if("tsn" %in% colnames(scientific_tmp)) {
        scientific[i,] <- scientific_tmp [c("tsn", "combinedname")]
      }

      if("tsn" %in% colnames(common_tmp)) {
        # if more than 1 common name, condense
        if(table(common_tmp$tsn) > 1) {
          common2 <- tapply(common_tmp$commonName, INDEX = common_tmp$tsn, FUN = paste, collapse = file.sep)
          common_tmp <- data.frame(commonName = common2,
                                   tsn        = rownames(common2),
                                   stringsAsFactors = FALSE)
        }
        common[i,] <- common_tmp [,c("tsn", "commonName")]
      }

      if("tsn" %in% colnames(author_tmp)) {
        author[i,] <- author_tmp [c("tsn", "authorship")]
      }

      if("tsn" %in% colnames(rankname_tmp)) {
        rankname[i,] <- rankname_tmp [c("tsn", "rankname")]
      }

      if(accepted == FALSE){
        taxon_status_tmp <- ritis::core_metadata (tsns_worked[i])  # get taxonomic status

        if("tsn" %in% colnames(taxon_status_tmp)) {
          taxon_status[i,] <- taxon_status_tmp [c("tsn", "taxonUsageRating")]
        }
      }
    }


    # make outtable
    dat.out <- data.frame(user_name       = speciesNames,
                          tsn             = as.numeric(tsns))


    dat.out <- merge(x = dat.out, y = scientific, by = "tsn", all.x = TRUE, sort = FALSE)
    dat.out <- merge(x = dat.out, y = common,     by = "tsn", all.x = TRUE, sort = FALSE)
    dat.out <- merge(x = dat.out, y = author,     by = "tsn", all.x = TRUE, sort = FALSE)
    dat.out <- merge(x = dat.out, y = rankname,   by = "tsn", all.x = TRUE, sort = FALSE)

    dat.out$itis_url <- NA
    dat.out$itis_url [match(tsns_worked, dat.out$tsn)]       <- attributes(tsns_worked)$uri

    colnames(dat.out)[colnames(dat.out) == "combinedname"] <- "scientificName"

    if(accepted == FALSE){
      dat.out <- merge(x = dat.out, y = taxon_status, by = "tsn", all.x = TRUE, sort = FALSE)
    } else {
      dat.out$taxon_status[!is.na(dat.out$tsn)]  <- "valid"
    }

    dat.out.sorted <- dat.out[match(dat.out$user_name, speciesNames),]
    
    return(dat.out.sorted)

  } else {stop("found no TSNs for speciesNames", call. = FALSE)}   # error if no tsns found
}

# --- Helper to create mock TSN objects ---
# Mimics taxize::as.tsn behavior:
# 1. Preserves input length (including 0 for empty vectors)
# 2. Sets class to "tsn"
# 3. Sets "uri" attribute (required by checkSpeciesNames)
fake_as_tsn_impl <- function(x, check = FALSE) {
  if (length(x) == 0) {
    res <- character(0)
    class(res) <- "tsn"
    attr(res, "uri") <- character(0)
    return(res)
  }
  res <- x
  class(res) <- "tsn"
  # Provide a dummy URI for every element
  attr(res, "uri") <- rep("http://itis.gov/mock", length(x))
  return(res)
}

test_that("Input validation works", {
  
  testthat::skip_if_not_installed("taxize")
  
  expect_error(checkSpeciesNames(speciesNames = "Cat", searchtype = "invalid"),
               "should be one of")
  
  expect_error(checkSpeciesNames(speciesNames = 123, searchtype = "common"),
               "is.character")
  expect_error(checkSpeciesNames(speciesNames = "Cat", searchtype = "common", accepted = "yes"),
               "is.logical")
})

test_that("Happy Path: Valid Scientific Name found", {
  
  testthat::skip_if_not_installed("mockery")
  testthat::skip_if_not_installed("taxize")
  
  # --- Mocks ---
  # get_tsn returns a TSN object
  m_get_tsn <- mockery::mock(fake_as_tsn_impl("123456"))
  
  m_sci_name <- mockery::mock(data.frame(tsn = "123456", combinedname = "Panthera leo", stringsAsFactors = FALSE))
  m_com_name <- mockery::mock(data.frame(tsn = "123456", commonName = "Lion", stringsAsFactors = FALSE))
  m_author   <- mockery::mock(data.frame(tsn = "123456", authorship = "Linnaeus", stringsAsFactors = FALSE))
  m_rank     <- mockery::mock(data.frame(tsn = "123456", rankname = "Species", stringsAsFactors = FALSE))
  
  # --- Stubs ---
  mockery::stub(checkSpeciesNames, "taxize::get_tsn", m_get_tsn)
  mockery::stub(checkSpeciesNames, "taxize::as.tsn", fake_as_tsn_impl)
  mockery::stub(checkSpeciesNames, "ritis::scientific_name", m_sci_name)
  mockery::stub(checkSpeciesNames, "ritis::common_names", m_com_name)
  mockery::stub(checkSpeciesNames, "ritis::taxon_authorship", m_author)
  mockery::stub(checkSpeciesNames, "ritis::rank_name", m_rank)
  
  # --- Execution ---
  result <- checkSpeciesNames(speciesNames = "Panthera leo", 
                              searchtype = "scientific", 
                              accepted = TRUE)
  
  # --- Assertions ---
  expect_equal(nrow(result), 1)
  expect_equal(result$tsn, 123456)
  expect_equal(result$scientificName, "Panthera leo")
  expect_equal(result$commonName, "Lion")
  expect_equal(result$taxon_status, "valid") 
  expect_false(is.na(result$itis_url))
})

test_that("Logic: Multiple Common Names are collapsed", {
  
  testthat::skip_if_not_installed("mockery")
  testthat::skip_if_not_installed("taxize")
  
  # --- Mocks ---
  m_get_tsn <- mockery::mock(fake_as_tsn_impl("123"))
  
  m_sci_name <- mockery::mock(data.frame(tsn = "123", combinedname = "Simba", stringsAsFactors = FALSE))
  m_com_name <- mockery::mock(data.frame(tsn = c("123", "123"), 
                                commonName = c("NameA", "NameB"), 
                                stringsAsFactors = FALSE))
  m_author   <- mockery::mock(data.frame(tsn = "123", authorship = "Disney", stringsAsFactors = FALSE))
  m_rank     <- mockery::mock(data.frame(tsn = "123", rankname = "Species", stringsAsFactors = FALSE))
  
  # --- Stubs ---
  mockery::stub(checkSpeciesNames, "taxize::get_tsn", m_get_tsn)
  mockery::stub(checkSpeciesNames, "taxize::as.tsn", fake_as_tsn_impl)
  mockery::stub(checkSpeciesNames, "ritis::scientific_name", m_sci_name)
  mockery::stub(checkSpeciesNames, "ritis::common_names", m_com_name)
  mockery::stub(checkSpeciesNames, "ritis::taxon_authorship", m_author)
  mockery::stub(checkSpeciesNames, "ritis::rank_name", m_rank)
  
  # --- Execution ---
  result <- checkSpeciesNames("Simba", "common")
  
  # --- Assertions ---
  sep <- .Platform$file.sep
  expected_common <- paste("NameA", "NameB", sep = sep)
  expect_equal(result$commonName, expected_common)
})

test_that("Error Handling: No Matches Found (All NA)", {
  
  testthat::skip_if_not_installed("mockery")
  testthat::skip_if_not_installed("taxize")
  
  # --- Mocks ---
  # taxize::get_tsn returns NA
  m_get_tsn <- mockery::mock(NA)
  
  # Important: When checkSpeciesNames gets NA, it filters the list.
  # The resulting list is empty. fake_as_tsn_impl correctly returns length 0.
  # This prevents the loop from running and calling ritis functions with bad data.
  
  mockery::stub(checkSpeciesNames, "taxize::get_tsn", m_get_tsn)
  mockery::stub(checkSpeciesNames, "taxize::as.tsn", fake_as_tsn_impl)
  
  # --- Execution & Assertion ---
  # Since the filtered list is empty, the function hits the `else { stop(...) }` block
  expect_error(
    suppressWarnings(checkSpeciesNames("Unicorn", "common")),
    "found no TSNs for speciesNames"
  )
})

test_that("Partial Success: One match, One mismatch", {
  
  testthat::skip_if_not_installed("mockery")
  testthat::skip_if_not_installed("taxize")
  
  input_names <- c("Lion", "Unicorn")
  
  # --- Mocks ---
  # taxize::get_tsn returns vector: "123" for Lion, NA for Unicorn
  tsn_raw <- c("123", NA)
  m_get_tsn <- mockery::mock(tsn_raw)
  
  # Mock ritis functions (only called for valid TSN "123")
  m_sci_name <- mockery::mock(data.frame(tsn = "123", combinedname = "Leo", stringsAsFactors = FALSE))
  m_com_name <- mockery::mock(data.frame(tsn = "123", commonName = "Lion", stringsAsFactors = FALSE))
  m_author   <- mockery::mock(data.frame(tsn = "123", authorship = "L.", stringsAsFactors = FALSE))
  m_rank     <- mockery::mock(data.frame(tsn = "123", rankname = "Spp", stringsAsFactors = FALSE))
  
  mockery::stub(checkSpeciesNames, "taxize::get_tsn", m_get_tsn)
  mockery::stub(checkSpeciesNames, "taxize::as.tsn", fake_as_tsn_impl)
  mockery::stub(checkSpeciesNames, "ritis::scientific_name", m_sci_name)
  mockery::stub(checkSpeciesNames, "ritis::common_names", m_com_name)
  mockery::stub(checkSpeciesNames, "ritis::taxon_authorship", m_author)
  mockery::stub(checkSpeciesNames, "ritis::rank_name", m_rank)
  
  # --- Execution ---
  expect_warning(
    result <- checkSpeciesNames(input_names, "common"),
    "found no matches for 1 name"
  )
  
  # --- Assertions ---
  expect_equal(nrow(result), 2)
  
  # Check Lion row
  lion_row <- result[result$user_name == "Lion", ]
  expect_equal(lion_row$tsn, 123)
  
  # Check Unicorn row
  unicorn_row <- result[result$user_name == "Unicorn", ]
  expect_true(is.na(unicorn_row$tsn))
})

test_that("Error Handling: taxize network error", {
  
  testthat::skip_if_not_installed("mockery")
  testthat::skip_if_not_installed("taxize")
  
  # --- Mock ---
  err <- structure("Network Error", class = "try-error")
  m_get_tsn <- mockery::mock(err)
  
  mockery::stub(checkSpeciesNames, "taxize::get_tsn", m_get_tsn)
  
  # --- Execution ---
  expect_message(
    result <- checkSpeciesNames("Lion", "common"),
    "error in get_tsn"
  )
  
  expect_null(result)
})

test_that("Logic: Accepted = FALSE fetches usage rating", {
  
  testthat::skip_if_not_installed("mockery")
  testthat::skip_if_not_installed("taxize")
  
  # --- Mocks ---
  m_get_tsn <- mockery::mock(fake_as_tsn_impl("999"))
  
  m_sci_name <- mockery::mock(data.frame(tsn = "999", combinedname = "OldName", stringsAsFactors = FALSE))
  m_com_name <- mockery::mock(data.frame()) 
  m_author   <- mockery::mock(data.frame()) 
  m_rank     <- mockery::mock(data.frame())
  m_core_meta <- mockery::mock(data.frame(tsn = "999", taxonUsageRating = "invalid", stringsAsFactors = FALSE))
  
  mockery::stub(checkSpeciesNames, "taxize::get_tsn", m_get_tsn)
  mockery::stub(checkSpeciesNames, "taxize::as.tsn", fake_as_tsn_impl)
  mockery::stub(checkSpeciesNames, "ritis::scientific_name", m_sci_name)
  mockery::stub(checkSpeciesNames, "ritis::common_names", m_com_name)
  mockery::stub(checkSpeciesNames, "ritis::taxon_authorship", m_author)
  mockery::stub(checkSpeciesNames, "ritis::rank_name", m_rank)
  mockery::stub(checkSpeciesNames, "ritis::core_metadata", m_core_meta)
  
  # --- Execution ---
  result <- checkSpeciesNames("OldName", "scientific", accepted = FALSE)
  
  # --- Assertions ---
  expect_equal(result$tsn, 999)
  expect_equal(result$taxonUsageRating, "invalid")
})

#' Radial plots of single-species activity
#' 
#' The function generates a radial plot of species diel activity using an
#' adapted version of function \code{\link[plotrix]{radial.plot}} from package
#' \pkg{plotrix} (without the need to install the package). Records are
#' aggregated by hour. The number of independent events is used as input, which
#' in turn is based on the argument \code{minDeltaTime} in
#' \code{\link{recordTable}}.
#' 
#' 
#' \code{radial.plot} was adjusted to show a clockwise 24-hour clock face. It
#' is recommended to set argument \code{lwd} to a value >= 2. You may also wish
#' to add argument \code{rp.type="p"} to show a polygon instead of bars.
#' 
#' \code{recordDateTimeFormat} defaults to the "YYYY-MM-DD HH:MM:SS"
#' convention, e.g. "2014-09-30 22:59:59". \code{recordDateTimeFormat} can be
#' interpreted either by base-R via \code{\link[base]{strptime}} or in
#' \pkg{lubridate} via \code{\link[lubridate]{parse_date_time}} (argument
#' "orders"). \pkg{lubridate} will be used if there are no "\%" characters in
#' \code{recordDateTimeFormat}.
#' 
#' For "YYYY-MM-DD HH:MM:SS", \code{recordDateTimeFormat} would be either
#' "\%Y-\%m-\%d \%H:\%M:\%S" or "ymd HMS". For details on how to specify date
#' and time formats in R see \code{\link[base]{strptime}} or
#' \code{\link[lubridate]{parse_date_time}}.
#' 
#' @param recordTable data.frame. the record table created by
#' \code{\link{recordTable}}
#' @param species Name of the species for which to create an kernel density
#' plot of activity
#' @param allSpecies logical. Create plots for all species in \code{speciesCol}
#' of \code{recordTable}? Overrides argument \code{species}
#' @param speciesCol character. name of the column specifying species names in
#' \code{recordTable}
#' @param recordDateTimeCol character. name of the column specifying date and
#' time in \code{recordTable}
#' @param recordDateTimeFormat character. format of column
#' \code{recordDateTimeCol} in \code{recordTable}
#' @param byNumber logical. If FALSE, plot proportion of records. If TRUE, plot
#' number of records
#' @param plotR logical. Show plots in R graphics device?
#' @param writePNG logical. Create pngs of the plots?
#' @param plotDirectory character. Directory in which to create png plots if
#' \code{writePNG = TRUE}
#' @param createDir logical. Create \code{plotDirectory}?
#' @param pngMaxPix integer. image size of png (pixels along x-axis)
#' @param \dots additional arguments to be passed to function
#' \code{\link[plotrix]{radial.plot}}
#' 
#' @return Returns invisibly a data.frame containing all information needed to
#' create the plot: radial position, lengths, hour (for labels). If
#' \code{allSpecies == TRUE}, all species' data frames are returned in an
#' invisible named list.
#' 
#' @author Juergen Niedballa
#' 
#' @seealso \code{\link{activityDensity}}, \code{\link{activityHistogram}},
#' \code{\link{activityOverlap}}
#' 
#' @references Lemon, J. (2006) Plotrix: a package in the red light district of
#' R. R-News, 6(4): 8-12. \cr \url{https://CRAN.R-project.org/package=plotrix }
#' 
#' @examples
#' 
#' 
#' # load record table
#' data(recordTableSample)
#' 
#' species4activity <- "PBE"    # = Prionailurus bengalensis, Leopard Cat
#' 
#' activityRadial(recordTable       = recordTableSample,
#'                species           = species4activity,
#'                allSpecies        = FALSE,
#'                speciesCol        = "Species",
#'                recordDateTimeCol = "DateTimeOriginal",
#'                plotR             = TRUE,
#'                writePNG          = FALSE,
#'                lwd               = 5
#' )
#' 
#' # plot type = polygon
#' 
#' activityRadial(recordTable       = recordTableSample,
#'                species           = species4activity,
#'                allSpecies        = FALSE,
#'                speciesCol        = "Species",
#'                recordDateTimeCol = "DateTimeOriginal",
#'                plotR             = TRUE,
#'                writePNG          = FALSE,
#'                lwd               = 5,
#'                rp.type           = "p"      
#' )
#' 
#' 
#' @export activityRadial
#' 
activityRadial <- function(recordTable,
                           species,
                           allSpecies = FALSE,
                           speciesCol = "Species",
                           recordDateTimeCol = "DateTimeOriginal",
                           recordDateTimeFormat = "ymd HMS",
                           byNumber = FALSE,
                           plotR = TRUE,
                           writePNG = FALSE,
                           plotDirectory,
                           createDir = FALSE,
                           pngMaxPix = 1000,
                           ...){

  wd0 <- getwd()
  mar0 <- par()$mar
  on.exit(setwd(wd0))
  on.exit(par(mar = mar0), add = TRUE)
						                  
  checkForSpacesInColumnNames(speciesCol = speciesCol, recordDateTimeCol = recordDateTimeCol)
  
  recordTable <- dataFrameTibbleCheck(df = recordTable)

  if(!speciesCol %in% colnames(recordTable))           stop(paste('speciesCol = "', speciesCol, '" is not a column name in recordTable', sep = ''), call. = FALSE)
  if(!recordDateTimeCol %in% colnames(recordTable)) stop(paste('recordDateTimeCol = "', recordDateTimeCol, '" is not a column name in recordTable', sep = ''), call. = FALSE)

  stopifnot(is.logical(c(allSpecies, writePNG, plotR, createDir, byNumber)))
  if(allSpecies == FALSE) {
    stopifnot(hasArg(species))
    stopifnot(species %in% recordTable[,speciesCol])
  }
  
  timeZone <- "UTC"

  recordTable$DateTime2 <- parseDateTimeObject(inputColumn = recordTable[,recordDateTimeCol],
                                               dateTimeFormat = recordDateTimeFormat,
                                               timeZone = timeZone)

  recordTable$Time2 <- as.POSIXlt(recordTable$DateTime2)$hour
  
  recordTable[,speciesCol] <- as.character(recordTable[,speciesCol])



  if(isTRUE(writePNG)){
    if(hasArg(plotDirectory)){
      if(isTRUE(createDir)){
        dir.create(plotDirectory, recursive = TRUE, showWarnings = FALSE)
        setwd(plotDirectory)
      } else {
        if(file.exists(plotDirectory) == FALSE) stop("plotDirectory does not exist.")
        setwd(plotDirectory)
      }
    } else {stop("writePNG is TRUE. Please set plotDirectory")}
  }

  pngWidth <- pngMaxPix
  pngHeight <- pngMaxPix


  if(allSpecies == FALSE){

    subset_species <- recordTable[recordTable[,speciesCol] == species,]
    lengths.tmp <- table(subset_species$Time2)

    seq.tmp <- data.frame(hour = seq(0,23, length.out = 24),
                          n = 0)
    seq.tmp$n[match(as.numeric(names(lengths.tmp)), seq.tmp$hour)] <- lengths.tmp
    seq.tmp$perc <- seq.tmp$n / sum(seq.tmp$n)
    seq.tmp$radial.pos <- seq.tmp$hour/(24/(2*pi))
    if(isTRUE(byNumber)){
      seq.tmp$length4plot <- seq.tmp$n
    } else {
      seq.tmp$length4plot <- seq.tmp$perc
    }

    if(isTRUE(writePNG)){
      png(filename = paste("activity_radial_", species, "_", Sys.Date(), ".png", sep = ""),
          width = pngWidth, height = pngHeight, units = "px", res = 96, type = "cairo")
      .radial.plot(lengths = seq.tmp$length4plot, radial.pos = seq.tmp$radial.pos,
                   clockwise = TRUE,
                   start = (pi/2),
                   labels = paste(formatC(seq.tmp$hour, width = 2,  flag = 0), "00", sep = ""),
                   main = species,
                   boxed.radial = FALSE,
                   ...)
      #title(main = paste("Activity of", species), line = 3)
      #mtext(paste("number of records:", nrow(subset_species)), side = 3, line = 0)
      dev.off()
    }
    if(isTRUE(plotR)){
      .radial.plot(lengths = seq.tmp$length4plot, radial.pos = seq.tmp$radial.pos,
                   clockwise = TRUE,
                   start = (pi/2),
                   labels = paste(formatC(seq.tmp$hour, width = 2,  flag = 0), "00", sep = ""),
                   main = species,
                   boxed.radial = FALSE,
                   ...)
      #title(main = paste("Activity of", species), line = 3)
      #mtext(paste("number of records:", nrow(subset_species)), side = 3, line = 0)
    }

  } else {

    subset_species_list <- list()

    for(i in 1:length(unique(recordTable[,speciesCol]))){

      spec.tmp <- unique(recordTable[,speciesCol])[i]
      subset_species <- subset(recordTable, recordTable[,speciesCol] == spec.tmp)

      lengths.tmp <- table(subset_species$Time2)

      seq.tmp <- data.frame(hour = seq(0,23, length.out = 24),
                            n = 0)
      seq.tmp$n[match(as.numeric(names(lengths.tmp)), seq.tmp$hour)] <- lengths.tmp
      seq.tmp$perc <- seq.tmp$n / sum(seq.tmp$n)
      seq.tmp$radial.pos <- seq.tmp$hour/(24/(2*pi))
      if(isTRUE(byNumber)){
        seq.tmp$length4plot <- seq.tmp$n
      } else {
        seq.tmp$length4plot <- seq.tmp$perc
      }

      if(isTRUE(writePNG)){
        png(filename = paste("activity_radial_", spec.tmp, "_", Sys.Date(), ".png", sep = ""),
            width = pngWidth, height = pngHeight, units = "px", res = 96, type = "cairo")
        .radial.plot(lengths = seq.tmp$length4plot, radial.pos = seq.tmp$radial.pos,
                     clockwise = TRUE,
                     start = (pi/2),
                     labels = paste(formatC(seq.tmp$hour, width = 2,  flag = 0), "00", sep = ""),
                     main = spec.tmp,
                     boxed.radial = FALSE,
                     ...)
        #title(main = paste("Activity of", spec.tmp), line = 3)             
        #mtext(paste("number of records:", nrow(subset_species)), side = 3, line = 0)
        dev.off()
      }

      if(isTRUE(plotR)){
        .radial.plot(lengths = seq.tmp$length4plot, radial.pos = seq.tmp$radial.pos,
                     clockwise = TRUE,
                     start = (pi/2),
                     labels = paste(formatC(seq.tmp$hour, width = 2,  flag = 0), "00", sep = ""),
                     main = spec.tmp,
                     boxed.radial = FALSE,
                     ...)
        #title(main = paste("Activity of", spec.tmp), line = 3)             
        #mtext(paste("number of records:", nrow(subset_species)), side = 3, line = 0)
      }
      subset_species_list[[i]] <- seq.tmp
      names(subset_species_list)[i] <- spec.tmp
    }
  }
  if(allSpecies == FALSE){
    return(invisible(seq.tmp))
  } else {
    return(invisible(subset_species_list))
  }
}

#' Convert Camtrap DP format data to camtrapR format
#'
#' @description
#' This function converts camera trap data from the Camtrap DP standard format to the format
#' used by camtrapR. Camtrap DP is an open standard for the FAIR exchange and archiving of 
#' camera trap data, structured in a simple yet flexible data model consisting of three tables: 
#' Deployments, Media, and Observations.
#'
#' @details
#' Camtrap DP is a standardized format developed under the umbrella of the Biodiversity Information
#' Standards (TDWG) that facilitates data exchange between different camera trap platforms and systems.
#' It supports a wide range of camera deployment designs, classification techniques, and analytical use cases.
#' 
#' While the 'camtrapdp' package (available on CRAN) provides general functionality for reading, 
#' manipulating, and transforming Camtrap DP data, this function specifically converts Camtrap DP 
#' data directly into the format required by camtrapR, producing the CTtable and recordTable objects 
#' that camtrapR functions expect.
#' 
#' This function reads the three primary tables from a Camtrap DP dataset:
#' \itemize{
#'   \item Deployments: Information about camera trap placements
#'   \item Observations: Classifications derived from the media files
#'   \item Media: Information about the media files (images/videos) recorded
#' }
#'
#' The Media table is only read if the \code{add_file_path} parameter is set to \code{TRUE}. 
#'
#' The function converts these into two primary camtrapR data structures:
#' \itemize{
#'   \item CTtable: Contains deployment information including station ID, setup/retrieval dates, 
#'         camera operation problems, camera setup, and more
#'   \item recordTable: Contains observation records with taxonomic and temporal information
#' }
#'
#' Additional features include:
#' \itemize{
#'   \item Parsing of deploymentTags and deploymentGroups (in deployments) and observationTags (in observations)
#'   \item Extraction of taxonomic information from metadata 
#'   \item Handling of deployment intervals and gaps
#' }
#'
#' @param deployments_file character. Path to deployments.csv file
#' @param media_file character. Path to media.csv file
#' @param observations_file character. Path to observations.csv file
#' @param datapackage_file character. Path to datapackage.json file
#' @param min_gap_hours numeric. Minimum gap in hours to consider as a camera interruption (default: 24)
#' @param removeNA logical. Whether to remove columns with only NA values
#' @param removeEmpty logical. Whether to remove columns with only empty values
#' @param remove_bbox logical. Whether to remove bounding box columns in the observation table
#' @param add_file_path logical. Whether to add file path from media table to the observation table
#' @param filter_observations Controls which observation types to include. \code{NULL} or \code{FALSE} keeps all observations (default), \code{TRUE} keeps only animal observations, or provide a character vector of specific observation types to include.
#' 
#' @return List containing three elements:
#' \itemize{
#'   \item CTtable: Data frame with camera trap deployment information in camtrapR format
#'   \item recordTable: Data frame with species records in camtrapR format
#'   \item metadata: List containing project metadata extracted from datapackage.json
#' }
#' 
#' @note
#' Camtrap DP was developed as a consensus of a long, in-depth consultation process with 
#' camera trap data management platforms and major players in the field of camera trapping.
#' It builds upon the earlier Camera Trap Metadata Standard (CTMS) but addresses its limitations.
#' 
#' The Camtrap DP standard structures data in a way that supports both media-based observations
#' (using a single media file as source) and event-based observations (considering an event with
#' a specified duration as source). For the purpose of this function, both are treated to be equivalent. 
#' Event-based observations are converted to a single timestamp in the recordTable (using the event start time).
#'
#' Consider using the 'camtrapdp' package (Desmet et al., 2024) for more general operations on 
#' Camtrap DP data before converting to camtrapR format with this function.
#'
#' @references
#' Bubnicki, J.W., Norton, B., Baskauf, S.J., et al. (2023). Camtrap DP: an open standard for the 
#' FAIR exchange and archiving of camera trap data. Remote Sensing in Ecology and Conservation, 10(3), 283-295.
#' 
#' Desmet, P., Govaert, S., Huybrechts, P., Oldoni, D. (2024). camtrapdp: Read and Manipulate Camera 
#' Trap Data Packages. R package version 0.3.1. https://CRAN.R-project.org/package=camtrapdp
#' 
#' @examples
#' \dontrun{
#' # Read a Camtrap DP dataset
#' camtrap_data <- readcamtrapDP(
#'   deployments_file = "path/to/deployments.csv",
#'   media_file = "path/to/media.csv",
#'   observations_file = "path/to/observations.csv",
#'   datapackage_file = "path/to/datapackage.json"
#' )
#' 
#' # alternatively, set the working directory only
#' setwd("path/to/camtrapdp_data")
#' camtrap_data <- readcamtrapDP()   # uses default file names
#' 
#' 
#' # Extract components
#' ct_table <- camtrap_data$CTtable
#' record_table <- camtrap_data$recordTable
#' metadata <- camtrap_data$metadata
#' 
#' }
#' 
#' @seealso 
#' \code{\link[camtrapdp]{read_camtrapdp}} in the 'camtrapdp' package for reading Camtrap DP data
#' \code{\link[camtrapdp]{check_camtrapdp}} in the 'camtrapdp' package for validating Camtrap DP data
#' 
#' @export
#' @importFrom dplyr bind_rows
#' 
readcamtrapDP <- function(
    deployments_file = "deployments.csv", 
    media_file = "media.csv", 
    observations_file = "observations.csv",
    datapackage_file = "datapackage.json",
    min_gap_hours = 24,
    removeNA = FALSE,
    removeEmpty = FALSE,
    remove_bbox = TRUE,
    add_file_path = FALSE,
    filter_observations = NULL 
) {
  
  # Read input files
  deployments <- read.csv(deployments_file, stringsAsFactors = FALSE)
  observations <- read.csv(observations_file, stringsAsFactors = FALSE)
  if(add_file_path) media <- read.csv(media_file, stringsAsFactors = FALSE)
  
  # Load metadata from datapackage.json
  metadata <- process_datapackage_json(datapackage_file)
  
  
  # Define column names
  station_id_col = "locationID"
  camera_id_col = "cameraID"
  coord_x = "longitude"
  coord_y = "latitude"
  
  # Ensure date fields are properly formatted 
  deployments$deploymentStart <- ymd_hms(deployments$deploymentStart)
  deployments$deploymentEnd   <- ymd_hms(deployments$deploymentEnd)
  if(add_file_path) media$timestamp  <- ymd_hms(media$timestamp)
  
  # Convert logical columns to proper logical type
  deployments  <- convert_logical_columns(deployments)
  observations <- convert_logical_columns(observations)
  if(add_file_path) media <- convert_logical_columns(media)
  
  # Parse deploymentTags and deploymentGroups
  if ("deploymentTags" %in% colnames(deployments)) {
    tag_columns <- parse_tags(deployments, "deploymentTags")
    if (ncol(tag_columns) > 0) {
      deployments <- cbind(deployments, tag_columns)
    }
  }
  
  if ("deploymentGroups" %in% colnames(deployments)) {
    group_columns <- parse_tags(deployments, "deploymentGroups")
    if (ncol(group_columns) > 0) {
      deployments <- cbind(deployments, group_columns)
    }
  }
  
  # Parse observationTags in observations table
  if ("observationTags" %in% colnames(observations)) {
    obs_tag_columns <- parse_tags(observations, "observationTags")
    if (ncol(obs_tag_columns) > 0) {
      observations <- cbind(observations, obs_tag_columns)
    }
  }
  
  # Get unique stations
  stations <- unique(deployments[[station_id_col]])
  
  # Create data frame to hold all rows
  ctTable <- data.frame(stringsAsFactors = FALSE)
  
  for (station in stations) {
    # Get all deployments for this station
    station_deployments <- deployments[deployments[[station_id_col]] == station, ]
    station_deployments <- station_deployments[order(station_deployments$deploymentStart), ]
    
    # Create basic row with overall deployment period (as character)
    row <- data.frame(
      Station = as.character(station),
      Setup_date = as.character(format(min(station_deployments$deploymentStart), "%Y-%m-%d %H:%M:%S")),
      Retrieval_date = as.character(format(max(station_deployments$deploymentEnd), "%Y-%m-%d %H:%M:%S")),
      stringsAsFactors = FALSE
    )
    
    # Handle problems (gaps between deployments)
    if (nrow(station_deployments) > 1) {
      problem_count <- 0
      
      for (i in 2:nrow(station_deployments)) {
        current_start <- station_deployments$deploymentStart[i]
        previous_end <- station_deployments$deploymentEnd[i-1]
        
        # Calculate gap
        gap_hours <- as.numeric(difftime(current_start, previous_end, units = "hours"))
        
        # If gap exceeds threshold, add as a problem
        if (gap_hours >= min_gap_hours) {
          problem_count <- problem_count + 1
          row[[paste0("Problem", problem_count, "_from")]] <- as.character(format(previous_end, "%Y-%m-%d %H:%M:%S"))
          row[[paste0("Problem", problem_count, "_to")]] <- as.character(format(current_start, "%Y-%m-%d %H:%M:%S"))
        }
      }
    }
    
    # Add all columns from deployments table as character
    for (col in colnames(station_deployments)) {
      # Skip deploymentID, deploymentStart, deploymentEnd which we've already used
      if (col %in% c("deploymentID", "deploymentStart", "deploymentEnd")) {
        next
      }
      
      # Get unique values for this column across all deployments for this station
      values <- unique(station_deployments[[col]])
      values <- values[!is.na(values)] # Remove NA values
      
      if (length(values) > 0) {
        # Convert all values to character before concatenating
        values_char <- as.character(values)
        row[[col]] <- paste(values_char, collapse = "; ")
      } else {
        row[[col]] <- NA_character_  # Use NA_character_ for consistency
      }
    }
    
    # If this is the first row, initialize the ctTable with the same columns
    if (nrow(ctTable) == 0) {
      empty_df <- as.data.frame(lapply(row, function(x) character(0)))
      ctTable <- empty_df
    }
    
    # Add to CTtable
    ctTable <- bind_rows(ctTable, row)
  }
  
  # Convert setup and retrieval dates to character with appropriate format
  # ctTable$Setup_date     <- format(ctTable$Setup_date, "%Y-%m-%d %H:%M:%S")
  # ctTable$Retrieval_date <- format(ctTable$Retrieval_date, "%Y-%m-%d %H:%M:%S")
  
  
  # Convert numeric columns to appropriate types
  numeric_cols <- c("latitude", "longitude", "coordinateUncertainty", 
                    "cameraDelay", "cameraHeight", "cameraTilt", 
                    "cameraHeading", "detectionDistance")
  
  for (col in numeric_cols) {
    if (col %in% colnames(ctTable)) {
      # Only convert if the column doesn't contain semicolons (which would indicate multiple values)
      has_multiple <- grepl(";", ctTable[[col]])
      if (!any(has_multiple, na.rm = TRUE)) {
        ctTable[[col]] <- as.numeric(ctTable[[col]])
      }
    }
  }
  
  # Convert logical columns properly
  logical_cols <- c("timestampIssues", "baitUse")
  for (col in logical_cols) {
    if (col %in% colnames(ctTable)) {
      has_multiple <- grepl(";", ctTable[[col]])
      if (!any(has_multiple, na.rm = TRUE)) {
        ctTable[[col]] <- as.logical(ctTable[[col]])
      }
    }
  }
  
  
  # Process recordTable
  recordTable <- observations
  
  # Add station and camera information from deployments
  selected_deployment_cols <- c("deploymentID", station_id_col, camera_id_col, "locationName")
  selected_deployment_cols <- selected_deployment_cols[selected_deployment_cols %in% colnames(deployments)]
  
  recordTable <- merge(
    recordTable,
    deployments[, selected_deployment_cols],
    by = "deploymentID",
    all.x = TRUE
  )
  
  
  # Add file path if requested
  if (add_file_path && "mediaID" %in% colnames(recordTable)) {
    recordTable <- merge(
      recordTable,
      media[, c("mediaID", "filePath")],
      by = "mediaID",
      all.x = TRUE
    )
  }
  
  # Format DateTimeOriginal from eventStart
  if ("eventStart" %in% colnames(recordTable)) {
    recordTable$DateTimeOriginal <- ymd_hms(recordTable$eventStart)
  }
  
  # Remove bounding box columns if not wanted
  if (remove_bbox) {
    bbox_cols <- c("bboxX", "bboxY", "bboxWidth", "bboxHeight")
    bbox_cols <- bbox_cols[bbox_cols %in% colnames(recordTable)]
    if (length(bbox_cols) > 0) {
      recordTable <- recordTable[, !colnames(recordTable) %in% bbox_cols]
    }
  }
  
  # Remove columns with all NA if requested
  if (removeNA) {
    ctTable <- remove_na_columns(ctTable)
    recordTable <- remove_na_columns(recordTable)
  }
  
  # Remove columns with all empty values if requested
  if (removeEmpty) {
    ctTable <- remove_empty_columns(ctTable)
    recordTable <- remove_empty_columns(recordTable)
  }
  
  
  # Filter observations by type if requested
  if (!is.null(filter_observations)) {
    if (isTRUE(filter_observations)) {
      # Keep only animal observations if TRUE
      recordTable <- recordTable[recordTable$observationType == "animal", ]
    } else if (is.character(filter_observations)) {
      # Keep specified observation types if character vector
      recordTable <- recordTable[recordTable$observationType %in% filter_observations, ]
    }
  }
  
  
  # Add taxonomic information from datapackage.json if available
  if (!is.null(metadata) && !is.null(metadata$taxonomic)) {
    # Create taxonomic lookup
    tax_data <- metadata$taxonomic
    
    # Map scientific names to vernacular names and rank
    if ("scientificName" %in% colnames(recordTable)) {
      # Add taxon rank if available
      if ("taxonRank" %in% names(tax_data)) {
        taxon_rank_map <- setNames(
          as.character(tax_data$taxonRank),
          tax_data$scientificName
        )
        recordTable$taxonRank <- taxon_rank_map[recordTable$scientificName]
      }
      
      # Add vernacular names in all available languages
      if ("vernacularNames" %in% names(tax_data)) {
        vern_names <- tax_data$vernacularNames
        if (is.data.frame(vern_names)) {
          for (lang in colnames(vern_names)) {
            lang_col <- paste0("vernacularName_", lang)
            name_map <- setNames(
              as.character(vern_names[[lang]]),
              tax_data$scientificName
            )
            
            # Map scientific names to vernacular names
            recordTable[[lang_col]] <- name_map[recordTable$scientificName]
            
            # Use observationType as fallback for empty vernacular names
            empty_names <- is.na(recordTable[[lang_col]]) | recordTable[[lang_col]] == ""
            has_obs_type <- !is.na(recordTable$observationType) & recordTable$observationType != ""
            
            # Replace NA/empty vernacular names with observationType where possible
            recordTable[[lang_col]][empty_names & has_obs_type] <- 
              recordTable$observationType[empty_names & has_obs_type]
          }
        }
      } else {
        # # If no vernacular names in metadata, create at least one column with observationType fallback
        # recordTable$vernacularName <- recordTable$scientificName
        # empty_names <- is.na(recordTable$vernacularName) | recordTable$vernacularName == ""
        # has_obs_type <- !is.na(recordTable$observationType) & recordTable$observationType != ""
        # recordTable$vernacularName[empty_names & has_obs_type] <- 
        #   recordTable$observationType[empty_names & has_obs_type]
        
        stop("No vernacularName in metadata")
      }
    }
  } else {
    # # Even without metadata, create a vernacular name column with observationType fallback
    # recordTable$vernacularName <- recordTable$scientificName
    # empty_names <- is.na(recordTable$vernacularName) | recordTable$vernacularName == ""
    # has_obs_type <- !is.na(recordTable$observationType) & recordTable$observationType != ""
    # recordTable$vernacularName[empty_names & has_obs_type] <- 
    #   recordTable$observationType[empty_names & has_obs_type]
    
    stop("No metadata or not taxonomic information in metadata")
  }
  
  
  # Reorder columns of both tables
  ctTable <- reorder_deployment_columns(ctTable)
  recordTable <- reorder_observation_columns(recordTable)
  
  
  return(list(
    CTtable = ctTable,
    recordTable = recordTable,
    metadata = metadata
  ))
}




# #' Reorder columns in the deployment table for better readability
# #'
# #' @param df Deployment data frame
# #' @return Data frame with reordered columns
# #'
reorder_deployment_columns <- function(df) {
  # Get all column names
  all_cols <- colnames(df)
  
  # Define column groups in order of priority
  location_cols <- c("locationName", "Station", "locationID")
  temporal_cols <- c("Setup_date", "Retrieval_date")
  problem_cols <- grep("^Problem[0-9]+_(from|to)$", all_cols, value = TRUE)
  spatial_cols <- c("latitude", "longitude", "coordinateUncertainty")
  
  # Extract tag and group columns (from parsing)
  tag_cols <- grep("^deploymentTags_", all_cols, value = TRUE)
  group_cols <- grep("^deploymentGroups_", all_cols, value = TRUE)
  
  # Original tag/group columns
  original_tag_cols <- c("deploymentTags", "deploymentGroups")
  
  # Environment and habitat info
  env_cols <- c("habitat", "featureType")
  
  # Camera setup info
  camera_cols <- c("cameraID", "cameraModel", "cameraDelay", "cameraHeight", 
                   "cameraTilt", "cameraHeading", "detectionDistance", 
                   "timestampIssues", "baitUse", "setupBy")
  
  # Comment columns
  comment_cols <- c("deploymentComments")
  
  # Combine the priority columns
  priority_cols <- c(
    intersect(location_cols, all_cols),
    intersect(temporal_cols, all_cols),
    intersect(problem_cols, all_cols),
    intersect(spatial_cols, all_cols),
    intersect(tag_cols, all_cols),
    intersect(group_cols, all_cols),
    intersect(original_tag_cols, all_cols),
    intersect(env_cols, all_cols),
    intersect(camera_cols, all_cols),
    intersect(comment_cols, all_cols)
  )
  
  # Any remaining columns go at the end
  remaining_cols <- setdiff(all_cols, priority_cols)
  
  # Final column order
  new_order <- c(priority_cols, remaining_cols)
  
  # Check for missing columns and use only those that exist
  existing_cols <- intersect(new_order, all_cols)
  
  # Return reordered data frame
  return(df[, existing_cols, drop = FALSE])
}

# #' Reorder columns in the observation table for better readability
# #'
# #' @param df Observation data frame
# #' @return Data frame with reordered columns


# Update column reordering to handle taxonomic columns
reorder_observation_columns <- function(df) {
  # Get all column names
  all_cols <- colnames(df)
  
  # Define column groups in order of priority
  location_cols <- c("locationName", "Station", "locationID", "cameraID")
  
  # Expanded taxonomic columns
  species_cols <- c("Species", "scientificName")
  vernacular_cols <- grep("^vernacularName_", all_cols, value = TRUE)
  taxon_cols <- c("taxonRank")
  
  temporal_cols <- c("DateTimeOriginal", "eventStart", "eventEnd")
  
  # Individual info
  individual_cols <- c("count", "sex", "lifeStage", "individualID", "behavior")
  
  # Extract tag columns (from parsing)
  tag_cols <- grep("^observationTags_", all_cols, value = TRUE)
  
  # Original tags column
  original_tag_cols <- c("observationTags")
  
  # Observation metadata
  obs_meta_cols <- c("observationType", "observationLevel", "cameraSetupType")
  
  # Classification info
  classification_cols <- c("classificationMethod", "classifiedBy", 
                           "classificationTimestamp", "classificationProbability")
  
  # Bounding box info
  bbox_cols <- c("bboxX", "bboxY", "bboxWidth", "bboxHeight")
  
  # File and media info
  file_cols <- c("filePath", "mediaID")
  
  # Comment columns
  comment_cols <- c("observationComments")
  
  # ID columns placed at the end
  id_cols <- c("observationID", "eventID", "deploymentID")
  
  # Combine the priority columns
  priority_cols <- c(
    intersect(location_cols, all_cols),
    intersect(species_cols, all_cols),
    intersect(vernacular_cols, all_cols),
    intersect(taxon_cols, all_cols),
    intersect(temporal_cols, all_cols),
    intersect(individual_cols, all_cols),
    intersect(tag_cols, all_cols),
    intersect(original_tag_cols, all_cols),
    intersect(obs_meta_cols, all_cols),
    intersect(classification_cols, all_cols),
    intersect(bbox_cols, all_cols),
    intersect(file_cols, all_cols),
    intersect(comment_cols, all_cols),
    intersect(id_cols, all_cols)
  )
  
  # Any remaining columns go at the end
  remaining_cols <- setdiff(all_cols, priority_cols)
  
  # Final column order
  new_order <- c(priority_cols, remaining_cols)
  
  # Check for missing columns and use only those that exist
  existing_cols <- intersect(new_order, all_cols)
  
  # Return reordered data frame
  return(df[, existing_cols, drop = FALSE])
}

# #' Parse pipe-separated tag strings into separate columns
# #'
# #' @param df Data frame containing the tag column
# #' @param tag_column Name of the column containing tags
# #' @return Data frame with new columns for each tag key
# #'
parse_tags <- function(df, tag_column) {
  # Return empty data frame if tag column doesn't exist or all entries are NA
  if (!tag_column %in% colnames(df) || all(is.na(df[[tag_column]]))) {
    return(data.frame())
  }
  
  # Initialize empty data frame for tag columns
  tag_df <- data.frame(row.names = rownames(df))
  
  # Process each row
  for (i in 1:nrow(df)) {
    if (is.na(df[[tag_column]][i]) || df[[tag_column]][i] == "") {
      next
    }
    
    # Split tags by pipe
    tags <- strsplit(df[[tag_column]][i], "\\|")[[1]]
    
    # Process each tag
    for (tag in tags) {
      tag <- trimws(tag)
      
      # Skip empty tags
      if (tag == "") {
        next
      }
      
      # Check if tag has key:value format
      if (grepl(":", tag)) {
        # Split into key and value
        key_value <- strsplit(tag, ":")[[1]]
        key <- trimws(key_value[1])
        value <- trimws(key_value[2])
        
        # Use tag_column as prefix to avoid column name conflicts
        column_name <- paste0(tag_column, "_", key)
        
        # Convert value to appropriate type if possible
        if (tolower(value) %in% c("true", "false")) {
          value <- as.logical(value)
        } else if (grepl("^[0-9]+$", value)) {
          value <- as.integer(value)
        } else if (grepl("^[0-9]*\\.[0-9]+$", value)) {
          value <- as.numeric(value)
        }
        
        # Add or update column
        if (!column_name %in% colnames(tag_df)) {
          tag_df[[column_name]] <- NA
        }
        tag_df[i, column_name] <- value
      } else {
        # Tag is a flag (no value)
        column_name <- paste0(tag_column, "_", tag)
        
        # Add or update column
        if (!column_name %in% colnames(tag_df)) {
          tag_df[[column_name]] <- FALSE
        }
        tag_df[i, column_name] <- TRUE
      }
    }
  }
  
  return(tag_df)
}

# #' Convert columns with "true"/"false" values to logical type
# #'
# #' @param df Data frame to process
# #' @return Data frame with logical columns converted
# #'
convert_logical_columns <- function(df) {
  for (col in colnames(df)) {
    # Check if the column is character type
    if (is.character(df[[col]])) {
      # Remove NA values for checking
      values <- df[[col]][!is.na(df[[col]])]
      
      # Skip empty columns
      if (length(values) == 0) {
        next
      }
      
      # Trim whitespace and convert to lowercase for comparison
      trimmed_lower_values <- tolower(trimws(values))
      
      # Check if all values are "true" or "false" (case insensitive)
      if (all(trimmed_lower_values %in% c("true", "false"))) {
        # Convert to logical, handling case insensitivity and whitespace
        df[[col]] <- tolower(trimws(df[[col]])) == "true"
      }
    }
  }
  return(df)
}

# #' Remove columns containing only NA values
# #'
# #' @param df Data frame to process
# #' @return Data frame with NA-only columns removed
# #'
remove_na_columns <- function(df) {
  # Calculate the number of NA values in each column
  na_counts <- sapply(df, function(x) sum(is.na(x)))
  
  # Keep columns where there's at least one non-NA value
  keep_cols <- names(na_counts[na_counts < nrow(df)])
  
  # Return the filtered data frame
  if (length(keep_cols) > 0) {
    return(df[, keep_cols, drop = FALSE])
  } else {
    return(df)  # Return original if all columns would be removed
  }
}

# #' Remove columns containing only empty strings
# #'
# #' @param df Data frame to process
# #' @return Data frame with empty-string-only columns removed
# #'
remove_empty_columns <- function(df) {
  # Find columns to keep (those with at least one non-empty value)
  keep_cols <- sapply(df, function(x) {
    # Skip non-character columns
    if (!is.character(x)) {
      return(TRUE)
    }
    
    # Count number of empty strings (ignoring NAs)
    empty_count <- sum(x == "", na.rm = TRUE)
    non_na_count <- sum(!is.na(x))
    
    # Keep if at least one non-empty value exists
    return(empty_count < non_na_count)
  })
  
  # Return filtered data frame
  if (any(keep_cols)) {
    return(df[, keep_cols, drop = FALSE])
  } else {
    return(df)  # Return original if all columns would be removed
  }
}

# Function to load and process datapackage.json
process_datapackage_json <- function(datapackage_file) {
  if (!file.exists(datapackage_file)) {
    warning("datapackage.json file not found")
    return(NULL)
  }
  
  # Try built-in R JSON parser first
  tryCatch({
    metadata <- jsonlite::fromJSON(datapackage_file, simplifyVector = TRUE)
    return(metadata)
  }, error = function(e) {
    # Fallback to jsonify if available
    if (requireNamespace("jsonify", quietly = TRUE)) {
      tryCatch({
        metadata <- jsonify::from_json(datapackage_file)
        return(metadata)
      }, error = function(e2) {
        warning("Failed to parse datapackage.json: ", e2$message)
        return(NULL)
      })
    } else {
      warning("Failed to parse datapackage.json: ", e$message)
      return(NULL)
    }
  })
}

context("readcamtrapDP")

# Load necessary libraries for testing
library(testthat)
library(withr)
library(jsonlite)

# --- Mock Data Creation ---
# For most tests, we'll create data programmatically to ensure the tests
# are self-contained and predictable.

# Mock deployments.csv data
mock_deployments <- data.frame(
  deploymentID = c("dep1", "dep2", "dep3"),
  locationID = c("StationA", "StationA", "StationB"),
  locationName = c("Forest Edge", "Forest Edge", "River Bank"),
  deploymentStart = c("2022-01-01T12:00:00Z", "2022-01-15T12:00:00Z", "2022-01-05T10:00:00Z"),
  deploymentEnd = c("2022-01-10T12:00:00Z", "2022-01-25T12:00:00Z", "2022-01-20T10:00:00Z"),
  deploymentTags = c("habitat:forest|research_project:X", "research_project:X", NA),
  stringsAsFactors = FALSE
)
# The gap between dep1 and dep2 is 5 days (120 hours), which should trigger the problem logic.

# Mock observations.csv data
mock_observations <- data.frame(
  observationID = 1:4,
  deploymentID = c("dep1", "dep1", "dep2", "dep3"),
  observationType = c("animal", "human", "animal", "unidentified"),
  scientificName = c("Panthera pardus", NA, "Sus scrofa", NA),
  count = c(1, 2, 1, 1),
  eventStart = c("2022-01-02T14:00:00Z", "2022-01-03T10:00:00Z", "2022-01-16T22:00:00Z", "2022-01-08T05:00:00Z"),
  bboxX = c(10, NA, 30, NA),
  stringsAsFactors = FALSE
)

# Mock media.csv data
mock_media <- data.frame(
  mediaID = 1:4,
  deploymentID = c("dep1", "dep1", "dep2", "dep3"),
  filePath = c("imgs/dep1/img1.jpg", "imgs/dep1/img2.jpg", "imgs/dep2/img3.jpg", "imgs/dep3/img4.jpg"),
  stringsAsFactors = FALSE
)
# Note: In a real CamtrapDP, mediaIDs should link to observationIDs. We simplify here for the test.

# Mock datapackage.json content
mock_datapackage <- toJSON(list(
  name = "mock-project",
  taxonomic = list(
    scientificName = c("Panthera pardus", "Sus scrofa"),
    taxonRank = c("species", "species"),
    vernacularNames = list(
      eng = c("Leopard", "Wild Boar"),
      fra = c("Léopard", "Sanglier")
    )
  )
), auto_unbox = TRUE, pretty = TRUE)


# --- Test Suite ---

testthat::describe("Core Functionality: Reading Data", {
  
  test_that("it reads correctly from individual CSV files in a temp directory", {
    with_tempdir({
      # Create mock files in the temporary directory
      write.csv(mock_deployments, "deps.csv", row.names = FALSE)
      write.csv(mock_observations, "obs.csv", row.names = FALSE)
      write.csv(mock_media, "med.csv", row.names = FALSE)
      writeLines(mock_datapackage, "dp.json")
      
      # Call the function with explicit file paths
      result <- readcamtrapDP(
        deployments_file = "deps.csv",
        observations_file = "obs.csv",
        media_file = "med.csv", # needed for add_file_path later
        datapackage_file = "dp.json"
      )
      
      # 1. Check output structure
      expect_equal(class(result), "list")
      expect_named(result, c("CTtable", "recordTable", "metadata"))
      
      # 2. Check CTtable
      ct_table <- result$CTtable
      expect_s3_class(ct_table, "data.frame")
      expect_equal(nrow(ct_table), 2) # StationA and StationB
      
      # 3. Check recordTable
      rec_table <- result$recordTable
      expect_s3_class(rec_table, "data.frame")
      expect_equal(nrow(rec_table), 4) # 4 observations
      
      # 4. Check metadata
      expect_true("taxonomic" %in% names(result$metadata))
    })
  })
  
  test_that("it reads correctly when only a directory is set (default filenames)", {
    with_tempdir({
      # Create files with default names
      write.csv(mock_deployments, "deployments.csv", row.names = FALSE)
      write.csv(mock_observations, "observations.csv", row.names = FALSE)
      writeLines(mock_datapackage, "datapackage.json")
      
      # Temporarily change working directory to the temp dir
      with_dir(".", {
        # Call with no arguments
        result <- readcamtrapDP()
      })
      
      # Check that the output is valid
      expect_equal(nrow(result$CTtable), 2)
      expect_equal(nrow(result$recordTable), 4)
    })
  })
  
  test_that("it reads correctly from the package's fixture data (real files)", {
    # This test uses the unzipped data you placed in tests/testthat/fixtures/
    fixture_path <- test_path("fixtures", "sample_camtrap_dp_data")
    
    # Skip if the directory doesn't exist, so the test is robust
    skip_if_not(dir.exists(fixture_path), "Sample fixture data not found.")
    
    # Use with_dir to temporarily set the working directory
    with_dir(fixture_path, {
      result <- readcamtrapDP()
    })
    
    # Perform some basic checks on the real data
    expect_equal(class(result), "list")
    expect_gt(nrow(result$CTtable), 0)
    expect_gt(nrow(result$recordTable), 0)
  })
})


testthat::describe("Data Processing Logic", {
  
  test_that("it correctly detects and records deployment gaps as problems", {
    with_tempdir({
      write.csv(mock_deployments, "deployments.csv", row.names = FALSE)
      write.csv(mock_observations, "observations.csv", row.names = FALSE)
      writeLines(mock_datapackage, "datapackage.json")
      
      result <- readcamtrapDP(deployments_file = "deployments.csv", 
                              observations_file = "observations.csv",
                              datapackage_file = "datapackage.json",
                              min_gap_hours = 24) # Default
      
      ct_table <- result$CTtable
      station_a_data <- ct_table[ct_table$Station == "StationA", ]
      
      # StationA had a 5-day gap, so Problem columns should exist
      expect_true("Problem1_from" %in% names(station_a_data))
      expect_true("Problem1_to" %in% names(station_a_data))
      expect_equal(station_a_data$Problem1_from, "2022-01-10 12:00:00")
      expect_equal(station_a_data$Problem1_to, "2022-01-15 12:00:00")
    })
  })
  
  test_that("it correctly parses tags and adds taxonomic info", {
      
    fixture_path <- test_path("fixtures", "sample_camtrap_dp_data")
    
      result <- readcamtrapDP(deployments_file = file.path(fixture_path, "deployments.csv"), 
                              observations_file = file.path(fixture_path, "observations.csv"),
                              # media_file = file.path(fixture_path, "media.csv"),
                              datapackage_file = file.path(fixture_path, "datapackage.json"),)
      

      # 1. Check taxonomic info
      rec_table <- result$recordTable
      expect_true("vernacularName_eng" %in% names(rec_table))
      species_test_row <- rec_table[rec_table$scientificName == "Anas strepera", ]
      expect_equal(species_test_row$vernacularName_eng[1], "gadwall")
      expect_equal(species_test_row$vernacularName_nld[1], "krakeend")
    # })
  })
})


testthat::describe("Argument Handling", {
  
  test_that("filter_observations correctly subsets the recordTable", {
    with_tempdir({
      write.csv(mock_deployments, "deployments.csv", row.names = FALSE)
      write.csv(mock_observations, "observations.csv", row.names = FALSE)
      writeLines(mock_datapackage, "datapackage.json")
      
      # Filter to keep only "animal"
      result_animal <- readcamtrapDP(deployments_file = "deployments.csv", 
                                     observations_file = "observations.csv",
                                     datapackage_file = "datapackage.json",
                                     filter_observations = TRUE)
      expect_equal(nrow(result_animal$recordTable), 2)
      expect_true(all(result_animal$recordTable$observationType == "animal"))
      
      # Filter to keep specific types
      result_specific <- readcamtrapDP(deployments_file = "deployments.csv", 
                                       observations_file = "observations.csv",
                                       datapackage_file = "datapackage.json",
                                       filter_observations = c("human", "unidentified"))
      expect_equal(nrow(result_specific$recordTable), 2)
    })
  })
  
  test_that("add_file_path and remove_bbox work correctly", {

    fixture_path <- test_path("fixtures", "sample_camtrap_dp_data")
    
      # Test with file paths added and bbox removed (defaults)
      result1 <- readcamtrapDP(deployments_file = file.path(fixture_path, "deployments.csv"), 
                               observations_file = file.path(fixture_path, "observations.csv"),
                               media_file = file.path(fixture_path, "media.csv"),
                               datapackage_file = file.path(fixture_path, "datapackage.json"),
                               add_file_path = TRUE,
                               remove_bbox = TRUE)
      expect_true("filePath" %in% names(result1$recordTable))
      expect_false("bboxX" %in% names(result1$recordTable))
      
      # Test with file paths removed and bbox kept
      result2 <- readcamtrapDP(deployments_file = file.path(fixture_path, "deployments.csv"), 
                               observations_file = file.path(fixture_path, "observations.csv"),
                               media_file = file.path(fixture_path, "media.csv"),
                               datapackage_file = file.path(fixture_path, "datapackage.json"),
                               add_file_path = FALSE,
                               remove_bbox = FALSE)
      expect_false("filePath" %in% names(result2$recordTable))
      expect_true("bboxX" %in% names(result2$recordTable))
  })
})

plot_effects_commOccu <- function(object,
                                  mcmc.list,
                                  submodel = "state",
                                  response = "occupancy",
                                  speciesSubset,
                                  draws = 1000,
                                  outdir,
                                  level = 0.95,
                                  keyword_quadratic = "_squared",
                                  ...)
{
  
  
  submodel <- match.arg(submodel, choices = c("det", "state"))
  
  if(submodel == "state") {
    keyword_submodel <- "^beta"
    keyword_submodel_short <- "beta"
  }
  if(submodel == "det") {
    keyword_submodel <- "^alpha"
    keyword_submodel_short <- "alpha"
  } 
  
  response <- match.arg(response, choices = c("occupancy", "abundance"))
  if(response == "abundance" & object@model != "RN") stop("response = 'abundance' is only available for Royle-Nichols models.")
  if(response == "abundance" & submodel == "det") warning("response = 'abundance' is not available for the detection submodel.")
  
  # get covariate information for submodel
  cov_info_subset <- object@covariate_info[object@covariate_info$submodel == submodel & object@covariate_info$param == "param",]
  
  if(nrow(cov_info_subset) == 0) stop(paste("No covariates in submodel", submodel), call. = F)
  
  # get intercept information for submodel
  cov_info_intercept <- object@covariate_info[object@covariate_info$submodel == submodel & object@covariate_info$param == "intercept",]

  # subset parameters of submodel
  stopifnot(all(cov_info_subset$coef %in% object@params))
  params_submodel <- object@params[grep(keyword_submodel, object@params)]
  

  # subset posterior matrix to number of draws
  posterior_matrix <- as.matrix(mcmc.list)
  if(hasArg(draws)) {
    if(nrow(posterior_matrix) > draws)     posterior_matrix <- posterior_matrix[sample(1:nrow(posterior_matrix), draws),]
  } 
  
  # subset posterior matrix to current submodel
  posterior_matrix <- posterior_matrix[, grep(keyword_submodel, colnames(posterior_matrix))]
  
  params_covariate <- cov_info_subset$covariate
  if(length(params_covariate) == 0) stop ("No covariates found", call. = F)
  
  
  list_responses <- list()
  
  
  
  
  # loop over covariates
  for(cov in 1:nrow(cov_info_subset)) {
    
    current_cov <- cov_info_subset$covariate[cov]
    current_coef <- cov_info_subset$coef[cov]
    
    
    is_squared <- cov_info_subset$is_quadratic[cov]
    
    if(is_squared) {
      attr(params_covariate, "include") [cov] <- FALSE
      if(gsub(keyword_quadratic, "", current_cov) %in% params_covariate) next
    } 
    attr(params_covariate, "include") [cov] <- TRUE
    

    if(!is.na(cov_info_subset$ranef_cov[cov])){
      warning(paste(current_cov, 
                    " has a random effect other than species. This is currently not supported. Skipping", call. = F))
      next
      }
    
    if(cov_info_subset$ranef_nested[cov])  {
      warning(paste(current_cov, 
                    " has a nested random effect. This is currently not supported. Skipping", call. = F))
      next
      }
    
    
    
    # check if there is a squared version of the current covariate
    
    has_squared <- cov_info_subset$has_quadratic[cov]
    if(paste0(current_cov, keyword_quadratic) %in% params_covariate){
      squared_cov <- paste0(current_cov, keyword_quadratic)
    } 
    
    # determine data type of current covariate
    covariate_is_numeric <- cov_info_subset$data_type [cov] == "cont"
    covariate_is_factor  <- cov_info_subset$data_type [cov] == "categ"
    
    effect_type <- ifelse(cov_info_subset$ranef[cov], "ranef",
                          ifelse(cov_info_subset$independent[cov], "independent", "fixed"))
    
    covariate_is_site_cov <- ifelse(cov_info_subset$covariate_type [cov] == "siteCovs", T, F) 
    
    
    # create values to predict to
    if(covariate_is_factor) {
      if(covariate_is_site_cov){
        values_to_predict <- seq(1,
                                 length(levels(object@data[[current_cov]])))
      } else {
        values_to_predict <- attr(object@data[[paste0(current_cov, "_integer")]], "levels")
      }
    }
    
    
    if(covariate_is_numeric) {  
      values_to_predict <- seq(min(object@data[[current_cov]]),
                               max(object@data[[current_cov]]), 
                               length.out = 100)
      }
    
    
    # empty matrix for predicted values
    out <- array(data = NA, dim = c(length(values_to_predict),    # number of values to predict
                                    object@data$M,                # number of species
                                    nrow(posterior_matrix)))      # number of posterior draws
    # likewise for intercept
    out_intercept <- out
    
    
    if(has_squared){
      values_to_predict_sq <- values_to_predict ^ 2
      
      out_sq <- array(data = NA, dim = c(length(values_to_predict_sq),
                                         object@data$M, 
                                         nrow(posterior_matrix)))
    }
    
    
    # species loop
    for(i in 1:dim(out)[2]){
      
      if(cov_info_intercept$ranef == TRUE | cov_info_intercept$independent == TRUE){  # random or independent intercepts
        out_intercept[,i,] <- matrix(posterior_matrix[, colnames(posterior_matrix) %in% paste0(keyword_submodel_short, "0", "[", i, "]")] , 
                 nrow = dim(out_intercept)[1], ncol = dim(out_intercept)[3], byrow = T)
      } else {
        out_intercept[,i,] <- matrix(posterior_matrix[, grepl(paste0(keyword_submodel_short, "0$"), colnames(posterior_matrix))] , 
                 nrow = dim(out_intercept)[1], ncol = dim(out_intercept)[3], byrow = T)
      }
      
        
        
      if(covariate_is_numeric) {
        
        if(effect_type == "fixed") {
          index_covariate <- grep(paste0(current_coef, "$"), colnames(posterior_matrix))
        } else {    # ranef or independent
          index_covariate <- grep(paste0(current_coef, "[", i, "]"), colnames(posterior_matrix), fixed = T)
        }
        if(length(index_covariate) == 0) stop(paste("Covariate", current_coef, "not found in posterior matrix"), call. = FALSE)
        if(length(index_covariate) >= 2) stop(paste("Covariate", current_coef, "has more than 2 matches in posterior matrix"), call. = FALSE)
        
        
        out[,i,] <-  sapply(posterior_matrix[, index_covariate], FUN = function(x){
          x * values_to_predict
        })
        
        
        if(has_squared){
          out_sq[,i,] <-  sapply(posterior_matrix[, grep(paste0(squared_cov, "[", i, "]"), colnames(posterior_matrix), fixed = TRUE)], FUN = function(x){
            x * values_to_predict_sq
          })
        }
      }
      
      if(covariate_is_factor) {
        
        if(effect_type == "fixed") index_covariate <- grep(current_coef, colnames(posterior_matrix))
        if(effect_type == "ranef") index_covariate <- grep(paste0(current_coef, "[", i, ","), colnames(posterior_matrix), fixed = T)
        
        
        for(j in 1:length(index_covariate)){
          out[j,i,] <- posterior_matrix[, index_covariate [j]]
        }
      }
      
      suppressWarnings(rm(index_covariate))
      
    }   # end species loop
    
    
    # intercept + linear covariate effect
    if(!has_squared){
      out_comb <- out_intercept + out
    }
    
    # intercept + linear + quadratic covariate effect
    if(has_squared){
      out_comb <- out_intercept + out + out_sq
    }
    
    if(!.hasSlot(object, "model")) {  # legacy support (when no RN model available)
      val <- exp(out_comb) / (exp(out_comb) + 1)    # prediction for each species / habitat value (from mean estimates)
    } else {
      if(object@model == "RN" & submodel == "state"){
        lambda <- exp(out_comb)    # val is lambda. out_comb is log(lambda)
        psi <- 1 - dpois(0, lambda)  
        
        if(response == "abundance") val <- lambda
        if(response == "occupancy") val <- psi
        
      } else {
        val <- exp(out_comb) / (exp(out_comb) + 1) 
      }
    }
    
    # summarize estimates (across posterior samples)
    val.mean  <- apply(val, MARGIN = c(1,2), mean)
    val.lower <- apply(val, MARGIN = c(1,2), quantile, (1-level) / 2, na.rm = FALSE)
    val.upper <- apply(val, MARGIN = c(1,2), quantile, (1 - (1-level) / 2), na.rm = FALSE)
    
    # make data frame for ggplot
    val.mean2  <- reshape2::melt(val.mean)
    val.lower2 <- reshape2::melt(val.lower)
    val.upper2 <- reshape2::melt(val.upper)
    
    names(val.mean2)  <- c("Index", "Species", "mean")
    names(val.lower2) <- c("Index", "Species", "lower")
    names(val.upper2) <- c("Index", "Species", "upper")
    
    vals <- cbind(val.mean2, lower = val.lower2$lower, upper = val.upper2$upper)
    vals <- cbind(cov = values_to_predict,
                   vals)
    colnames(vals) [1] <- current_cov
    
    
    # assign species names (if available)
    if(!is.null(dimnames(object@data$y)[[1]])) {
      vals$Species <- dimnames(object@data$y)[[1]][vals$Species]
    }
    
    vals <- vals[order(vals$Species, vals[, 1]),]
    
    
    if(submodel == "det")  {
      label_filename <- "detection"
      ylabel <- "Detection probability p"
    }
    if(submodel == "state" & object@model == "Occupancy"){
      label_filename <- "occupancy"
      ylabel <- expr(paste("Occupancy probability  ", psi))
    }
    if(submodel == "state" & object@model == "RN")  {
      if(response == "occupancy"){
        label_filename <- "occupancy"
        ylabel <- expr(paste("Occupancy probability  ", psi))
      }
      
      if(response == "abundance"){
        label_filename <- "abundance"
        ylabel <- expr(paste("Local abundance"))
      }
    } 
    
    main <- paste0(ifelse(covariate_is_site_cov, "Site", "Observation"), " covariate: ", current_cov)
    
    
    subtitle <- paste0(ifelse(effect_type == "ranef", "Random effect", ifelse(effect_type == "independent", "Independent effects", "Fixed effect")),
                       ifelse(has_squared, " (with quadratic term)", ""),
                       ifelse(is_squared, " quadratic term (no linear term)", ""))
    

    # for squared covariates which have no unsquared version, sqrt-transform covariate and add expand covariate range to negative values (by mirr)
    if(is_squared & !has_squared) {
      vals[,1] <- sqrt(vals[,1])
      
      vals2 <- vals
      vals2[,1] <- -vals2[,1]
      vals <- rbind(vals, vals2)
      
    }
    
    
    # subset species, if requested
    if(hasArg(speciesSubset)) {
      if(any(!speciesSubset %in% unique(vals$Species))) {
        stop(paste("Species", 
                   paste(speciesSubset[!speciesSubset %in% unique(vals$Species)], collapse = ", "), 
                   "not found in model."))
      }
      vals <- vals[vals$Species %in% speciesSubset, ]
    }
    
    
    # plot
    
    combine <- FALSE
    
    if(covariate_is_numeric){
      
		p <- ggplot(vals, aes(x = .data[[params_covariate[[cov]]]], y = .data[["mean"]], group = .data[["Species"]])) + 
        geom_line() +
        theme_bw() +
        ggtitle(label = main,
                subtitle = subtitle) +
        xlab (ifelse(is_squared, gsub(keyword_quadratic, "", current_cov), current_cov)) +
        ylab(ylabel) +
        xlim(range(vals[, 1])) +
        theme(panel.grid.minor = element_blank())
      
      if(.hasSlot(object, "model")) if(object@model == "Occupancy") p <- p + ylim(0, 1) 
      if(!.hasSlot(object, "model")) p <- p + ylim(0, 1) 
      
      # note for later, can optionally plot all species in  one plot if combine = TRUE (= ggplot code to this point)
      if(!combine){
        p <- p + facet_wrap(~Species) +
          geom_ribbon(aes(ymin = .data[["lower"]], ymax = .data[["upper"]]), alpha = 0.2)
      }
    }
    
    
    if(covariate_is_factor){
      
      # create x axis labels for factors
      if(covariate_is_site_cov){
        vals[,1] <- factor(levels(object@data[[current_cov]]) [vals[,1]],
                           levels = levels(object@data[[current_cov]]))
      } 
      
      p <- ggplot(vals, aes(x = .data[[params_covariate[[cov]]]], y = .data[["mean"]], group = .data[["Species"]])) + 
        geom_col() +
        facet_wrap(~Species) +
        geom_linerange(aes(ymin = lower, ymax = upper)) +
        theme_bw() +
        ggtitle(label = main,
                subtitle = subtitle) +
        xlab (current_cov) +
        ylab(ylabel) +
        theme(panel.grid.minor = element_blank())

      if(.hasSlot(object, "model")) if(object@model == "Occupancy") p <- p + ylim(0, 1) 
      if(!.hasSlot(object, "model")) p <- p + ylim(0, 1) 
      
      # don't know yet how to combine species on one plot. 
      
    }
    
    
    if(hasArg(outdir)) {
      ggsave(filename = file.path(outdir, paste0("response_curves_", label_filename, "_", current_cov, "_", Sys.Date(),  ".png")),
             plot = p,
             ...)
    }
    
    list_responses [[cov]] <- p
    
  }
  
  names(list_responses) <- params_covariate[attr(params_covariate, "include")]
  
  return(list_responses)
}


  setGeneric("plot_effects", def = function(object, ...) standardGeneric("plot_effects"))
  
  
  #' Plot Marginal Effects of Covariates
  #' 
  #' Plot marginal effect plots (= response curves if covariates are continuous) for all species in a community (multi-species) occupancy model. Takes into account species-specific intercepts (if any). Currently only supports continuous covariates, not categorical covariates.
  #'
  #' @aliases plot_effects
  #' @param object \code{commOccu} object
  #' @param mcmc.list  mcmc.list. Output of \code{fit} called on a \code{commOccu} object
  #' @param submodel  character. Submodel to get plots for. Can be \code{"det"} (detection submodel) or \code{"state"} (occupancy submodel)
  #' @param response character. response type on y axis. Only relevant for \code{submodel = "state"}. Default is \code{"occupancy"}, can be set to \code{"abundance"} for Royle-Nichols models
  #' @param speciesSubset  character. Species to include in effect plots.
  #' @param draws  integer. Number of draws from the posterior to use when generating the plots. If fewer posterior samples than specified in \code{draws} are available, all posterior samples are used.
  #' @param outdir character. Directory to save plots to (optional)
  #' @param level  numeric. Probability mass to include in the uncertainty interval.
  #' @param keyword_quadratic  character. A suffix in covariate names in the model that indicates a covariate is a quadratic effect of another covariate which does not carry the suffix in its name (e.g. if the covariate is "elevation", the quadratic covariate would be "elevation_squared").
  #' @param ...  additional arguments for \code{\link[ggplot2]{ggsave}} - only relevant if \code{outdir} is defined
  #'
  #' @details
  #' Users who wish to create their own visualizations can use the data stored in the ggplot output. It is accessed via e.g. \code{output$covariate_name$data}
  #' 
  #'
  #' @return A list of ggplot objects (one list item per covariate).
  #' @export
  #' @importFrom ggplot2 geom_vline geom_linerange geom_pointrange element_blank theme labs expr
  #' @importFrom ggplot2 scale_color_manual scale_y_discrete aes vars facet_grid facet_wrap ylim geom_col
  # @import coda
  #'
  setMethod("plot_effects", 
            signature = c(object = "commOccu"), 
            plot_effects_commOccu)
  
  
  
  

plot_coef_commOccu <- function(object, 
                               mcmc.list,
                               submodel = "state",
                               speciesSubset,
                               ordered = TRUE,
                               combine = FALSE,
                               outdir,
                               level = c(outer = 0.95, inner = 0.75),
                               colorby = "significance",
                               scales = "free_y",
                               community_lines = FALSE,
                               ...) {
  
  
  submodel <- match.arg(submodel, choices = c("det", "state"))
  colorby  <- match.arg(colorby, choices = c("significance", "Bayesian p-value"))
  if(colorby == "Bayesian p-value") stop(paste("colorby = 'Bayesian p-value' is currently not supported."))
  
  scales <- match.arg(scales, choices = c("free", "free_y"))
  
  if(submodel == "state") {
    keyword_submodel <- "^beta"
    keyword_submodel_short <- "beta"
  }
  if(submodel == "det") {
    keyword_submodel <- "^alpha"
    keyword_submodel_short <- "alpha"
  } 
  
  
  stopifnot(is.logical(ordered))
  stopifnot(is.logical(combine))
  
  if(combine & ordered) {
    message("'combine' and 'ordered' can't both be TRUE. Setting 'ordered = FALSE'")
    ordered <- FALSE
  }
  
  
  # get covariate information for submodel
  cov_info_subset <- object@covariate_info[object@covariate_info$submodel == submodel & object@covariate_info$param == "param",]
  
  if(nrow(cov_info_subset) == 0) stop(paste("No covariates in submodel", submodel), call. = F)
  
  
  list_responses <- list()
  
  
  # posterior summaries
  stopifnot(length(level) == 2)
  
  posteriorSummary  <- summary(mcmc.list, quantiles = c((1-level[1]) / 2,         # lower outer
                                                        (1-level[2]) / 2,         # lower inner      
                                                        0.5,                      # median
                                                        1-((1-level[2]) / 2),     # upper inner
                                                        1-((1-level[1]) / 2)))    # upper outer
  df_quantiles <- data.frame(posteriorSummary$quantiles)
  # 
  
  # all estimates model parameters
  params_all <- rownames(df_quantiles)
  
  # container for output plots
  p_list <- list()
  df_quantiles_list <- list()
  
  
  # get Bayesian p-values
  df_statistics <- posteriorSummary$statistics
  df_statistics_Bayes_pvals_overall <- as.data.frame(df_statistics[grep("Bpvalue$", rownames(df_statistics)), , drop = F])
  df_statistics_Bayes_pvals_species <- df_statistics[grep("Bpvalue_species", rownames(df_statistics)), ]
  
  # loop over covariates
  for(cov in 1:nrow(cov_info_subset)) {
    
    current_cov  <- cov_info_subset$covariate[cov]
    current_coef <- cov_info_subset$coef[cov]
    
    #if(covariate %in% skip) next
    if(!is.na(cov_info_subset$ranef_cov[cov])){
      warning(paste(current_cov, 
                    " has a random effect other than species. This is currently not supported. Skipping", call. = F))
      next
    }
    
    if(cov_info_subset$ranef_nested[cov])  {
      warning(paste(current_cov, 
                    " has a nested random effect. This is currently not supported. Skipping", call. = F))
      next
    }
    
    
    # determine data type of current covariate
    covariate_is_numeric <- cov_info_subset$data_type [cov] == "cont"
    covariate_is_factor  <- cov_info_subset$data_type [cov] == "categ"
    
    
    
    effect_type <- ifelse(cov_info_subset$ranef[cov], "ranef",
                          ifelse(cov_info_subset$independent[cov], "independent", "fixed"))
    
    
    covariate_is_site_cov <- ifelse(cov_info_subset$covariate_type [cov] == "siteCovs", T, F) 
    

    
    if(covariate_is_numeric){
      
      if(effect_type == "fixed")       index_covariate <- grep(paste0(current_coef, "$"), rownames(df_quantiles))
      if(effect_type == "ranef")       index_covariate <- grep(paste0(current_coef, "[" ), rownames(df_quantiles), fixed = T)
      if(effect_type == "independent") index_covariate <- grep(paste0(current_coef, "[" ), rownames(df_quantiles), fixed = T)
      
      
      df_quantiles_i  <- df_quantiles[index_covariate, ]
      
      if(effect_type == "fixed")   df_quantiles_i$type  <- c("mean")
      if(effect_type == "ranef") {
        
        # get community mean
        index_covariate_mean_ranef <- grep(paste0(current_coef, ".mean$"), rownames(df_quantiles))
        
        df_quantiles_i  <- rbind(df_quantiles[index_covariate_mean_ranef, ], df_quantiles_i)
        
        df_quantiles_i$type  <- c("mean", rep("species", times = length(index_covariate)))
      }  
      
      if(effect_type == "independent"){
        df_quantiles_i$type  <- c("species")
      }
      
    }
    
    
    if(covariate_is_factor){
      
      if(covariate_is_site_cov){
        levels_tmp <- levels(object@input$siteCovs[, current_cov])
        nlev <- length(levels_tmp)
      }
      if(!covariate_is_site_cov){
        levels_tmp <- attr(object@data[[paste0(current_cov, "_integer")]], "levels")
        nlev <- length(levels_tmp)
      }
      
      
      if(effect_type == "fixed")  index_covariate <- grep(paste0(current_coef, "["),  rownames(df_quantiles), fixed = T)
      if(effect_type == "ranef")  index_covariate <- grep(paste0(current_coef, "[" ), rownames(df_quantiles), fixed = T)
      
      
      df_quantiles_i  <- df_quantiles[index_covariate, ]  
      
      if(effect_type == "fixed") df_quantiles_i$type  <- c("mean")
      
      if(effect_type == "ranef") {
        
        # add community mean
        index_covariate_mean_ranef <- grep(paste0(current_coef, ".mean"), rownames(df_quantiles), fixed = T)   # does this affect categ fixed effects?
        
        df_quantiles_i  <- rbind(df_quantiles[index_covariate_mean_ranef, ], df_quantiles_i)
        
        df_quantiles_i$type  <- c(rep("mean", times = length(index_covariate_mean_ranef)), rep("species", times = length(index_covariate)))
      }
    }
    
    
    
    colnames(df_quantiles_i)[1:5] <- c("lower_outer", "lower_inner", "median", "upper_inner", "upper_outer")
    
    
    # get significance levels
    significance <- rep("no", times = nrow(df_quantiles_i))
    significance[which(df_quantiles_i$lower_inner < 0 & df_quantiles_i$upper_inner < 0 | 
                         df_quantiles_i$lower_inner > 0 & df_quantiles_i$upper_inner > 0)] <- "inner"
    significance[which(df_quantiles_i$lower_outer < 0 & df_quantiles_i$upper_outer < 0 | 
                         df_quantiles_i$lower_outer > 0 & df_quantiles_i$upper_outer > 0)] <- "outer"
    
    df_quantiles_i$significance  <- significance
    
    # add Bayesian p-values
    
    # if colorby == "Bayesian p-value":
    # slight inconsistency: values in "level" will still be used for error bar width
    # but will affect colors only via Bayesian p-values of the species
    # not via the actual parameter estimates. 
    # So a parameter estimate can be highly significant but grey bc bayesian p-value of the the species is ok
    # it's mostly for model checking, so I guess it's fine
    if(colorby == "Bayesian p-value"){
      if(covariate_is_numeric){
        if(effect_type == "fixed") df_pval <- df_statistics_Bayes_pvals_overall
        if(effect_type == "ranef") df_pval <- rbind(df_statistics_Bayes_pvals_overall, 
                                                    df_statistics_Bayes_pvals_species)
      }
      if(covariate_is_factor){
        if(effect_type == "fixed") df_pval <- df_statistics_Bayes_pvals_overall [rep(1, times = nlev),]
        if(effect_type == "ranef") df_pval <- rbind(df_statistics_Bayes_pvals_overall [rep(1, times = nlev),], 
                                                    df_statistics_Bayes_pvals_species [rep(1:nrow(df_statistics_Bayes_pvals_species), 
                                                                                           times = nlev),])
      }
      
      stopifnot(nrow(df_pval) == nrow(df_quantiles_i))
      
      
      significance2 <- rep("no", times = nrow(df_pval))
      significance2[which(df_pval$Mean < (1-level[2]) / 2 | df_pval$Mean > (1 - (1-level[2]) / 2))] <- "inner"
      significance2[which(df_pval$Mean < (1-level[1]) / 2 | df_pval$Mean > (1 - (1-level[1]) / 2))] <- "outer"
      
      df_quantiles_i$significance2  <- significance2
      
    }
    
    # assign species names
    if(!is.null(dimnames(object@data$y)[[1]])) speciesnames <- dimnames(object@data$y)[[1]]
    if( is.null(dimnames(object@data$y)[[1]])) speciesnames <- seq_len(dim(object@data$y)[1])
    
  
    
    if(effect_type == "ranef")  {
      if(covariate_is_numeric) df_quantiles_i$species <- c("community", speciesnames)
      if(covariate_is_factor) {
        df_quantiles_i$species <- c(rep("community", times = length(index_covariate_mean_ranef)), rep(speciesnames, times = nlev))
      }
    }
    
    if(effect_type == "fixed")     df_quantiles_i$species  <- "community"
    if(effect_type == "independent") df_quantiles_i$species  <- speciesnames
    
    
    
        # subset species, if requested
    if(hasArg(speciesSubset)) {
      if(any(!speciesSubset %in% speciesnames)) {
        stop(paste("Species", 
                   paste(speciesSubset[!speciesSubset %in% speciesnames], collapse = ", "), 
                   "not found in model."))
      }
      df_quantiles_i <- df_quantiles_i[c(1, which(df_quantiles_i$species %in% speciesSubset)), ]
    }
    
    
    
    
    
    # add covariate name as column
    if(covariate_is_numeric){
      df_quantiles_i$covariate  <- current_cov
    }
    
    if(covariate_is_factor){
      if(effect_type == "fixed"){
        if(covariate_is_site_cov)  df_quantiles_i$covariate  <- factor(paste0(current_cov, "_", levels_tmp),
                                                                       labels = levels_tmp)
        if(!covariate_is_site_cov) df_quantiles_i$covariate  <- factor(paste0(current_cov, "_",  levels_tmp),
                                                                       labels = levels_tmp)
      }
      
      if(effect_type == "ranef"){
        if(covariate_is_site_cov)  df_quantiles_i$covariate  <- factor(paste0(current_cov, "_", c(levels_tmp,
                                                                                                  rep(levels_tmp, each = object@data$M))),
                                                                       levels = paste0(current_cov, "_", levels_tmp),
                                                                       labels = levels_tmp)
        
        if(!covariate_is_site_cov) df_quantiles_i$covariate  <- factor(paste0(current_cov, "_", c(levels_tmp,
                                                                                                  rep(levels_tmp, each = object@data$M))),
                                                                       levels = paste0(current_cov, "_", levels_tmp),
                                                                       labels = levels_tmp)
        
      }
    }
    
    
    # sort species (either by median effect size or by names)
    if(ordered) {
      # this currently does not sort categorical covariates with random effects correctly. 
      # Ideally if combine = FALSE they should be sorted in descending order (at least for the second factor level)
      if(covariate_is_numeric){
        df_quantiles_i$species <- factor(df_quantiles_i$species, 
                                         levels = unique(df_quantiles_i$species[order(df_quantiles_i$median)]))
      }
      
      if(covariate_is_factor){
        if(effect_type == "fixed") {
          df_quantiles_i$species <- factor(df_quantiles_i$species, 
                                           levels = unique(df_quantiles_i$species[order(df_quantiles_i$median)]))  
        }
        
        if(effect_type == "ranef"){
          subset_level2 <- df_quantiles_i[df_quantiles_i$covariate == #paste0(current_cov, "_", 
                                            levels_tmp[2]#)
                                          ,]
          
          df_quantiles_i$species <- factor(df_quantiles_i$species, 
                                           levels = subset_level2$species[order(subset_level2$median)])
        }
      }
      
    } else {
      df_quantiles_i$species <- factor(df_quantiles_i$species, levels = unique(rev(df_quantiles_i$species)))  
    }
    
    df_quantiles_list[[cov]]  <- df_quantiles_i
    
    
    # plot
    
    type <- NULL   # just for CRAN checks
    covariate <- NULL
    lower_outer <- NULL
    upper_outer <- NULL
    median <- NULL
    
    if(colorby == "significance")      color_by <- "significance"
    if(colorby == "Bayesian p-value")  color_by <- "significance2"
    
    alpha_community <- ifelse(effect_type == "ranef", 0.3, 0)
    alpha_zero <- 0.3
    color_community <- "blue"
    
    if(!combine){
      
      
      p_list[[cov]] <- ggplot (df_quantiles_i, aes(y = .data[["species"]], x = .data[["median"]], color = color_by))
        
        if(community_lines) {
        # community effect
          p_list[[cov]] <- p_list[[cov]] + 
            geom_vline(data = df_quantiles_i[df_quantiles_i$type == "mean", -which(colnames(df_quantiles_i) == "type")],
                       aes(xintercept = median), col = color_community, linetype = 1, alpha = alpha_community) +
            geom_vline(data = df_quantiles_i[df_quantiles_i$type == "mean", -which(colnames(df_quantiles_i) == "type")],
                       aes(xintercept = lower_outer), col = color_community, linetype = 2, alpha = alpha_community) +
            geom_vline(data = df_quantiles_i[df_quantiles_i$type == "mean", -which(colnames(df_quantiles_i) == "type")],
                       aes(xintercept = upper_outer), col = color_community, linetype = 2, alpha = alpha_community)
        }
      
      p_list[[cov]] <- p_list[[cov]] +
        geom_vline(xintercept = 0, alpha = alpha_zero) +
        
        # species effects
        geom_pointrange(aes(xmin = .data[["lower_outer"]], xmax = .data[["upper_outer"]])) + 
        geom_linerange( aes(xmin = .data[["lower_inner"]], xmax = .data[["upper_inner"]]), linewidth = 1) +
        facet_grid(rows = vars(type),
                   cols = vars(covariate),
                   scales = scales,
                   space = "free_y"
        ) +
        xlab ("Effect size") +  ylab(element_blank()) +
        theme_bw() +
        theme(panel.grid.minor = element_blank(),
              panel.grid.major.y = element_blank(),
              strip.background.y = element_blank(),
              strip.text.y = element_blank()) + 
        scale_color_manual(breaks = c("outer", "inner", "no"),
                           values=c("firebrick", "black", "grey50"),
                           guide = "none") +
        ggtitle(paste("Effect sizes:", current_cov))
      
      
      if(!covariate_is_factor) {
        p_list[[cov]] <- p_list[[cov]] + theme(strip.background.x = element_blank(),
                                               strip.text.x = element_blank())
      }

      if(color_by == "significance2"){
        p_list[[cov]] <- p_list[[cov]] + labs(subtitle = "colors indicate Bayesian p-values of species")
      }
      
      
      if(hasArg(outdir)) {
        ggsave(filename = file.path(outdir, paste0("effect_sizes_", submodel, "_", covariate, "_", Sys.Date(),  ".png")),
               plot = p_list[[cov]],
               ...)
      }
    }   
    
  }   # end covariate loop
  
  
  if(combine){
    
    df_quantiles_all  <- do.call(rbind, df_quantiles_list)
    df_quantiles_all$species <- factor(df_quantiles_all$species, 
                                       levels = rev(sort(unique(as.character(df_quantiles_all$species)))))
    
    p <- ggplot (df_quantiles_all, aes(y = .data[["species"]], x = .data[["median"]], color = color_by))
      
    if(community_lines)  {
      # community effect
     p <- p + 
       geom_vline(data = df_quantiles_all[df_quantiles_all$type == "mean", -which(colnames(df_quantiles_all) == "type")],    
                  aes(xintercept = median), col = color_community, linetype = 1, alpha = alpha_community) +
       geom_vline(data = df_quantiles_all[df_quantiles_all$type == "mean", -which(colnames(df_quantiles_all) == "type")],
                  aes(xintercept = lower_outer), col = color_community, linetype = 2, alpha = alpha_community) +
       geom_vline(data = df_quantiles_all[df_quantiles_all$type == "mean", -which(colnames(df_quantiles_all) == "type")],
                  aes(xintercept = upper_outer), col = color_community, linetype = 2, alpha = alpha_community) 
    }
    p <- p + geom_vline(xintercept = 0, alpha = alpha_zero) +
      # species effects
      geom_pointrange(aes(xmin = .data[["lower_outer"]], xmax = .data[["upper_outer"]])) + 
      geom_linerange (aes(xmin = .data[["lower_inner"]], xmax = .data[["upper_inner"]]), linewidth = 1) +
      facet_grid(rows = vars(type), 
                 cols = vars(covariate),
                 scales = scales, 
                 space = "free_y") +
      xlab ("Effect size") +  ylab(element_blank()) +
      theme_bw() +
      theme(panel.grid.minor = element_blank(),
            panel.grid.major.y = element_blank(),
            strip.background.y = element_blank(),
            strip.text.y = element_blank()) + 
      scale_color_manual(breaks = c("outer", "inner", "no"),
                         values=c("firebrick", "black", "grey50"),
                         guide = "none")
    
    if(hasArg(outdir)) {
      ggsave(filename = file.path(outdir, paste0("effect_sizes_", submodel, "_", paste(cov_info_subset$covariate, collapse = "_"), "_", 
                                                 ifelse(!colorby == "significance", "Bayesian_pval_", ""), Sys.Date(),  ".png")),
             plot = p,
             ...)
    }
    return(p)
  }
  
  
  if(!combine){
    names(p_list) <-  cov_info_subset$covariate
    return(p_list)
  }
}



  setGeneric("plot_coef", def = function(object, ...) standardGeneric("plot_coef"))
  
  
  #' Plot effect sizes of covariates in community occupancy model
  #' 
  #' Plot effect sizes for all species in a community (multi-species) occupancy model. Currently only supports continuous covariates, not categorical covariates.
  #'
  #' @aliases plot_coef
  #' @param object \code{commOccu} object
  #' @param mcmc.list  mcmc.list. Output of \code{fit} called on a \code{commOccu} object
  #' @param submodel  character. Submodel to get plots for. Can be \code{"det"} or \code{"state"}
  #' @param speciesSubset  character. Species to include in coefficient plots.
  #' @param ordered logical. Order species in plot by median effect (TRUE) or by species name (FALSE)
  #' @param combine logical. Combine multiple plots into one plot (via facets)?
  #' @param outdir character. Directory to save plots to (optional)
  #' @param level  numeric. Probability mass to include in the uncertainty interval (two values named "outer" and "inner", in that order). Second value (= inner interval) will be plotted thicker.
  #' @param colorby character. Whether to color estimates by \code{"significance"} (of the effect estimates), or \code{"Bayesian p-value"} (of the species). Currently allows only \code{"significance"}.
  #' @param scales character. Passed to \code{\link[ggplot2]{facet_grid}}. Can be \code{"free"} to scale x axes of effect estimates independently, or "free_y" to scale all x axes identically.
  #' @param community_lines logical. Add faint vertical lines to the plot indicating median community effect (solid line) and its confidence interval (first value from \code{level}).
  #' @param ...  additional arguments for \code{\link[ggplot2]{ggsave}} - only relevant if \code{outdir} is defined.
  #'
  #' @return A list of ggplot objects (one list item per covariate).
  #' 
  #' @details
  #' Users who wish to create their own visualizations can use the data stored in the ggplot output. It is accessed via e.g. \code{output$covariate_name$data}
  #' 
  #' @export
  #' 
  #'
  setMethod("plot_coef", 
            signature = c(object = "commOccu"), 
            plot_coef_commOccu)
  
  

#' Create camera trap station directories for raw camera trap images
#' 
#' This function creates camera trap station directories, if needed with camera
#' subdirectories. They can be used as an initial directory structure for
#' storing raw camera trap images.
#' 
#' 
#' The empty directories serve as containers for saving raw camera trap images.
#' If more than 1 camera was set up at a station, specifying \code{cameras} is
#' required in order to keep images from different cameras separate. Otherwise,
#' generic filenames (e.g., IMG0001.JPG) from different cameras may lead to
#' accidental overwriting of images if images from these cameras are saved in
#' one station directory.
#' 
#' @param inDir character. Directory in which station directories are to be
#' created
#' @param stations character. Station IDs to be used as directory names within
#' \code{inDir}
#' @param cameras character. Camera trap IDs to be used as subdirectory names
#' in each station directory (optionally)
#' @param createinDir logical. If inDir does not exist, create it?
#' 
#' @return A \code{data.frame} with station (and possibly camera) directory
#' names and an indicator for whether they were created successfully.
#' 
#' @author Juergen Niedballa
#' 
#' @examples
#' 
#' \dontrun{
#' 
#' # create dummy directory for tests (this will be used as inDir)
#' # (normally, you'd set up an empty directory, e.g. .../myStudy/rawImages)
#' wd_createStationDir <- file.path(tempdir(), "createStationFoldersTest")
#' 
#' # now we load the sample camera trap station data frame
#' data(camtraps)
#' 
#' # create station directories in wd_createStationDir
#' StationFolderCreate1 <- createStationFolders (inDir       = wd_createStationDir,
#'                                               stations    = as.character(camtraps$Station),
#'                                               createinDir = TRUE)
#'   
#' StationFolderCreate1
#' 
#' # check if directories were created
#' list.dirs(wd_createStationDir)
#' 
#' }
#' 
#' @export createStationFolders
#' 
createStationFolders <- function(inDir,
                                 stations,
                                 cameras,
                                 createinDir
                                 ){

# check input

if(hasArg(createinDir) == FALSE) createinDir <- FALSE
stopifnot(is.logical(createinDir))

if(createinDir == FALSE & !dir.exists(inDir))  stop("Could not find inDir:\n", inDir, call. = FALSE)
if(createinDir == TRUE  & !dir.exists(inDir))  dir.create(inDir, recursive = TRUE)

  stopifnot(is.character(stations))
  if(hasArg(cameras)){
  stopifnot(is.character(cameras))
  stopifnot(length(stations) == length(cameras))
  }
 
 # create directories
   if(hasArg(cameras)){
  dirs.to.create <- file.path(inDir, stations, cameras)
  
  tmp.create <- suppressWarnings(sapply(dirs.to.create, FUN = dir.create, showWarnings = TRUE, recursive = TRUE))
    dat.out1 <- data.frame(station = stations, 
                           camera = cameras,
                           directory = dirs.to.create,
                           created = tmp.create,
                           exists = file.exists(dirs.to.create))
    rownames(dat.out1) <- NULL

    message(paste("created", sum(tmp.create == TRUE), "directories"))
    if(sum(tmp.create == FALSE) != 0){
      message(paste(sum(tmp.create == FALSE & file.exists(dirs.to.create)), "directories already existed"))
    }
    return(dat.out1)

 } else {
 
 if(any(duplicated(stations))) stop("duplicates in stations are not allowed if cameras is not defined")
  dirs.to.create <- file.path(inDir, stations)
  
  
    tmp.create <- suppressWarnings(sapply(dirs.to.create, FUN = dir.create, showWarnings = TRUE, recursive = TRUE))
    dat.out2 <- data.frame(station = stations, 
                            directory = dirs.to.create,
                           created = tmp.create,
                           exists = file.exists(dirs.to.create))
    rownames(dat.out2) <- NULL

    message(paste("created", sum(tmp.create == TRUE), "directories"))
    if(sum(tmp.create == FALSE) != 0){
      message(paste(sum(tmp.create == FALSE & file.exists(dirs.to.create)), "directories already existed"))
    }
    return(dat.out2)
}
}

#' Import Wildlife Insights data to camtrapR
#'
#' @description 
#' This function imports camera trap data from Wildlife Insights into a format compatible with camtrapR.
#' It can read data from a directory containing CSV files, from a ZIP file, or from individual CSV files.
#' 
#' @param directory character. Path to folder containing CSV files exported from Wildlife Insights.
#' @param zipfile character. Path to a ZIP file exported from Wildlife Insights.
#' @param deployment_file character. Path to the deployments CSV file.
#' @param image_file character. Path to the images CSV file.
#'
#' @return A list containing three elements:
#' \describe{
#'   \item{CTtable}{The full camera trap table, based on deployments.csv}
#'   \item{CTtable_aggregated}{An aggregated version of the camera trap table, with one row per station}
#'   \item{recordTable}{The record table, based on images.csv with additional columns from deployments.csv}
#' }
#' @export
#'
#' @examples
#' \dontrun{
#' # Reading from a directory
#' wi_data <- readWildlifeInsights(directory = "path/to/csv/files")
#'
#' # Reading from a ZIP file
#' wi_data <- readWildlifeInsights(zipfile = "path/to/wildlife_insights_export.zip")
#'
#' # Reading from separate CSV files
#' wi_data <- readWildlifeInsights(deployment_file = "path/to/deployments.csv",
#'                                 image_file = "path/to/images.csv")
#' }
readWildlifeInsights <- function(directory = NULL, 
                                 zipfile = NULL, 
                                 deployment_file = NULL, 
                                 image_file = NULL)
{
  # Check consistency of input parameters
  if (sum(!is.null(directory), !is.null(zipfile), 
          (!is.null(deployment_file) & !is.null(image_file))) != 1) {
    stop("Please provide either 'directory', 'zipfile', or both 'deployment_file' and 'image_file'.")
  }
  
  # Define required columns for validation
  required_deployment_cols <- c("deployment_id", "placename", "camera_id", "start_date", "end_date")
  required_image_cols <- c("deployment_id", "timestamp")
  
  # Load data from ZIP file
  if (!is.null(zipfile)) {
    if (!endsWith(zipfile, ".zip")) {
      stop("The provided zipfile does not have a .zip extension.")
    }
    
    if (!file.exists(zipfile)) {
      stop(paste("The specified ZIP file does not exist:", zipfile))
    }
    
    unzip_dir <- file.path(tempdir(), gsub(":", "_", Sys.time()))
    dir.create(unzip_dir)
    
    # Add error handling for unzip operation
    tryCatch({
      unzip(zipfile, exdir = unzip_dir)
    }, error = function(e) {
      stop(paste("Failed to unzip the file:", zipfile, "\nError:", e$message))
    })
    
    lf <- list.files(unzip_dir, recursive = TRUE)
    
    # Check for existence and uniqueness of required files
    image_files <- lf[grep("^images", basename(lf))]
    deployment_files <- lf[grep("^deployments", basename(lf))]
    
    if (length(image_files) == 0) {
      stop("No 'images' file found in the ZIP archive.")
    } else if (length(image_files) > 1) {
      stop("Multiple 'images' files found in the ZIP archive. Expected only one.")
    }
    
    if (length(deployment_files) == 0) {
      stop("No 'deployments' file found in the ZIP archive.")
    } else if (length(deployment_files) > 1) {
      stop("Multiple 'deployments' files found in the ZIP archive. Expected only one.")
    }
    
    # Read files with error handling
    tryCatch({
      images <- read.csv(file.path(unzip_dir, image_files))
    }, error = function(e) {
      stop(paste("Failed to read images file from ZIP archive:", e$message))
    })
    
    tryCatch({
      deployments <- read.csv(file.path(unzip_dir, deployment_files))
    }, error = function(e) {
      stop(paste("Failed to read deployments file from ZIP archive:", e$message))
    })
    
  } else if (!is.null(directory)) {
    if (!dir.exists(directory)) {
      stop("The specified directory does not exist.")
    }
    
    lf <- list.files(directory, recursive = TRUE)
    
    # Check for required files
    image_files <- lf[grep("^images", basename(lf))]
    deployment_files <- lf[grep("^deployments", basename(lf))]
    
    if (length(image_files) == 0) {
      stop("No 'images' file found in the directory.")
    } else if (length(image_files) > 1) {
      stop("Multiple 'images' files found in the directory. Expected only one.")
    }
    
    if (length(deployment_files) == 0) {
      stop("No 'deployments' file found in the directory.")
    } else if (length(deployment_files) > 1) {
      stop("Multiple 'deployments' files found in the directory. Expected only one.")
    }
    
    # Read files with error handling
    tryCatch({
      images <- read.csv(file.path(directory, image_files))
    }, error = function(e) {
      stop(paste("Failed to read images file from directory:", e$message))
    })
    
    tryCatch({
      deployments <- read.csv(file.path(directory, deployment_files))
    }, error = function(e) {
      stop(paste("Failed to read deployments file from directory:", e$message))
    })
    
  } else if (!is.null(deployment_file) && !is.null(image_file)) {
    if (!file.exists(deployment_file)) {
      stop(paste("The specified deployment file does not exist:", deployment_file))
    }
    
    if (!file.exists(image_file)) {
      stop(paste("The specified image file does not exist:", image_file))
    }
    
    # Read files with error handling
    tryCatch({
      deployments <- read.csv(deployment_file)
    }, error = function(e) {
      stop(paste("Failed to read deployments file:", e$message))
    })
    
    tryCatch({
      images <- read.csv(image_file)
    }, error = function(e) {
      stop(paste("Failed to read images file:", e$message))
    })
  }
  
  # Check if data frames are empty
  if (nrow(deployments) == 0) {
    stop("The deployments data is empty. Please check your input files.")
  }
  
  if (nrow(images) == 0) {
    stop("The images data is empty. Please check your input files.")
  }
  
  # Check for required columns in deployments
  missing_deployment_cols <- required_deployment_cols[!required_deployment_cols %in% colnames(deployments)]
  if (length(missing_deployment_cols) > 0) {
    stop(paste("Missing required columns in deployments data:", 
               paste(missing_deployment_cols, collapse = ", "),
               "\nAvailable columns are:", paste(colnames(deployments), collapse = ", ")))
  }
  
  # Check for required columns in images
  missing_image_cols <- required_image_cols[!required_image_cols %in% colnames(images)]
  if (length(missing_image_cols) > 0) {
    stop(paste("Missing required columns in images data:", 
               paste(missing_image_cols, collapse = ", "),
               "\nAvailable columns are:", paste(colnames(images), collapse = ", ")))
  }
  
  # Column references
  deploymentCol <- "deployment_id"
  stationCol <- "placename"
  cameraCol  <- "camera_id"
  date_formats <- c("ymd!*HMS!z", 
                    "ymd HMS", 
                    "mdy HMS",  
                    "dmy HMS"            
  )
  
  # Check for duplicate deployment IDs
  if(any(duplicated(deployments[, deploymentCol]))) {
    dup_deployments <- deployments[, deploymentCol][duplicated(deployments[, deploymentCol])]
    stop(paste("Duplicate deployment_id found:", paste0(dup_deployments, collapse = ", "),
               "\nEach deployment must have a unique identifier."))
  }
  
  # Check for missing values in critical columns
  for (col in c(deploymentCol, stationCol, cameraCol)) {
    if (any(is.na(deployments[, col])) || any(deployments[, col] == "")) {
      warning(paste("Found", sum(is.na(deployments[, col]) | deployments[, col] == ""), 
                    "missing or empty values in column", col, "of deployments data."))
    }
  }
  
  if (any(is.na(images[, deploymentCol])) || any(images[, deploymentCol] == "")) {
    warning(paste("Found", sum(is.na(images[, deploymentCol]) | images[, deploymentCol] == ""), 
                  "missing or empty values in deployment_id column of images data."))
  }
  
  if (any(is.na(images$timestamp)) || any(images$timestamp == "")) {
    warning(paste("Found", sum(is.na(images$timestamp) | images$timestamp == ""), 
                  "missing or empty values in timestamp column of images data."))
  }
  
  # Handle deployments with identical start and end dates
  remove_these <- deployments$start_date == deployments$end_date
  
  if(any(remove_these)) {
    remove_these_deployments_id <- deployments[remove_these, deploymentCol]
    warning(paste(sum(remove_these), "deployments with identical start and end date removed:\n", 
                  paste(remove_these_deployments_id, collapse = "\n")),
            call. = FALSE)
    deployments <- deployments[!remove_these,]
    
    if(any(remove_these_deployments_id %in% images[, deploymentCol])) {
      warning(sum(images[, deploymentCol] %in% remove_these_deployments_id), 
              " images removed because they were from deployments with identical start and end dates.")
      images <- images[!images[, deploymentCol] %in% remove_these_deployments_id,]
    }
  }
  
  # Save original date strings for error reporting
  deployments_start_date_original <- deployments$start_date
  deployments_end_date_original <- deployments$end_date
  
  # Validate that date columns contain string data before parsing
  if (!is.character(deployments$start_date)) {
    stop("The 'start_date' column must contain character (text) data. Check if your CSV file was imported correctly.")
  }
  
  if (!is.character(deployments$end_date)) {
    stop("The 'end_date' column must contain character (text) data. Check if your CSV file was imported correctly.")
  }
  
  # Parse dates
  deployments$start_date <- lubridate::parse_date_time(deployments$start_date, 
                                                       orders = date_formats,
                                                       truncated = 1)
  deployments$end_date   <- lubridate::parse_date_time(deployments$end_date, 
                                                       orders = date_formats,
                                                       truncated = 1)
  
  # Check for NA in start_date
  if(any(is.na(deployments$start_date))) {
    problem_indices <- which(is.na(deployments$start_date))
    example_strings <- head(as.character(deployments_start_date_original[problem_indices]), 3)
    deployment_ids <- head(deployments[problem_indices, deploymentCol], 3)
    
    stop(paste0(
      "Failed to parse ", sum(is.na(deployments$start_date)), " out of ", nrow(deployments), " start date(s) in Wildlife Insights data.\n",
      "Examples of unparseable dates from deployments: ", 
      paste(paste0(deployment_ids, ": '", example_strings, "'"), collapse = ", "), "\n",
      "Current supported formats: ", paste(date_formats, collapse = ", "), ".\n",
      "Check if date format in your data matches one of the supported formats.\n",
      "Date parsing may fail if data was edited in Excel or exported with non-default settings."
    ))
  }
  
  # Check for NA in end_date
  if(any(is.na(deployments$end_date))) {
    problem_indices <- which(is.na(deployments$end_date))
    example_strings <- head(as.character(deployments_end_date_original[problem_indices]), 3)
    deployment_ids <- head(deployments[problem_indices, deploymentCol], 3)
    
    stop(paste0(
      "Failed to parse ", sum(is.na(deployments$end_date)), " out of ", nrow(deployments), " end date(s) in Wildlife Insights data.\n",
      "Examples of unparseable dates from deployments: ", 
      paste(paste0(deployment_ids, ": '", example_strings, "'"), collapse = ", "), "\n",
      "Current supported formats: ", paste(date_formats, collapse = ", "), ".\n",
      "Check if date format in your data matches one of the supported formats.\n",
      "Date parsing may fail if data was edited in Excel or exported with non-default settings."
    ))
  }
  
  # Check that end dates are after start dates
  invalid_deployments <- which(deployments$end_date <= deployments$start_date)
  
  if(length(invalid_deployments) > 0) {
    problem_deployments <- deployments[invalid_deployments, ]
    examples <- head(problem_deployments, 3)
    
    # Format dates for cleaner display
    formatted_examples <- paste0(
      examples[, deploymentCol], ": ",
      "start_date = ", as.character(examples$start_date), ", ",
      "end_date = ", as.character(examples$end_date)
    )
    
    stop(paste0(
      "Found ", length(invalid_deployments), " deployment(s) where end date is not after start date.\n\n",
      "Examples of invalid deployments:\n",
      paste(formatted_examples, collapse = "\n"), 
      if(length(invalid_deployments) > 3) paste0("\n(and ", length(invalid_deployments) - 3, " more)") else "",
      "\n\nPlease check your Wildlife Insights data for date inconsistencies (especially in date/time format)."
    ))
  }
  
  # Aggregate table (by placename)
  col_types <- sapply(deployments, typeof)
  cols_char <- which(col_types %in% "character")
  cols_not_char <- which(!col_types %in% "character")
  
  # Handle character columns
  if (length(cols_char) > 0) {
    deployments1 <- aggregate(deployments[, cols_char, drop = FALSE], 
                              by = list(deployments[, stationCol]),
                              FUN = function(x) paste0(unique(x), collapse = " / "))
  } else {
    deployments1 <- data.frame(Group.1 = unique(deployments[, stationCol]))
  }
  
  # Handle non-character columns
  if (length(cols_not_char) > 0) {
    deployments2 <- aggregate(deployments[, cols_not_char, drop = FALSE], 
                              by = list(deployments[, stationCol]),
                              FUN = mean)
  } else {
    deployments2 <- data.frame(Group.1 = unique(deployments[, stationCol]))
  }
  
  # Combine aggregated data with error handling
  if (ncol(deployments1) > 1 && ncol(deployments2) > 1) {
    deployments_agg <- cbind(deployments1[, -1, drop = FALSE], deployments2[, -1, drop = FALSE])
    deployments_agg <- cbind(deployments1[, 1, drop = FALSE], deployments_agg)
    colnames(deployments_agg)[1] <- stationCol
  } else if (ncol(deployments1) > 1) {
    deployments_agg <- deployments1
    colnames(deployments_agg)[1] <- stationCol
  } else if (ncol(deployments2) > 1) {
    deployments_agg <- deployments2
    colnames(deployments_agg)[1] <- stationCol
  } else {
    # Fallback for aggregation
    stop("Aggregation failed")
    # deployments_agg <- data.frame(deployments[, c(stationCol, deploymentCol, cameraCol)])
    # deployments_agg <- deployments_agg[!duplicated(deployments_agg[, stationCol]), ]
  }
  
  # Remove bounding box column if present
  if("bounding_boxes" %in% colnames(images)) {
    images <- images[, colnames(images) != "bounding_boxes"]
  }
  
  # Check if all deployment_ids in images exist in deployments
  unknown_deployments <- unique(images[, deploymentCol][!images[, deploymentCol] %in% deployments[, deploymentCol]])
  if (length(unknown_deployments) > 0) {
    warning(paste0(
      "Found ", length(unknown_deployments), " deployment_id(s) in images that don't exist in deployments table.\n",
      "Example unknown deployment_ids: ", paste(head(unknown_deployments, 3), collapse = ", "),
      if(length(unknown_deployments) > 3) paste0(" (and ", length(unknown_deployments) - 3, " more)"),
      "\nThese images will be excluded from the merged output."
    ))
  }
  
  # Merge images with deployment data
  images_merged <- merge(images, 
                         deployments[, c(deploymentCol, stationCol, cameraCol)])
  
  # Check if merge reduced the number of records
  if (nrow(images_merged) < nrow(images)) {
    warning(paste0(
      "Merging reduced the number of image records from ", nrow(images), " to ", nrow(images_merged), " (",
      round((1 - nrow(images_merged) / nrow(images)) * 100, 1), "% reduction).\n",
      "This happens because some deployment_ids in the images table don't match any in the deployments table."
    ))
  }
  
  # Ensure record date/time column is formatted correctly
  record_datetime_format <- c("ymd HMS",    # standard format
                              "mdy HM")     # if user modified content of zip in Excel
  
  # Validate timestamp column
  if (!is.character(images_merged$timestamp)) {
    stop("The 'timestamp' column in images must contain character (text) data. Check if your CSV file is formatted correctly.")
  }
  
  # Parse timestamps with error handling
  timestamps_parsed <- lubridate::parse_date_time(images_merged$timestamp, orders = record_datetime_format)
  
  if(any(is.na(timestamps_parsed))) {
    problem_indices <- which(is.na(timestamps_parsed))
    example_strings <- head(images_merged$timestamp[problem_indices], 3)
    
    warning(paste0(
      "Failed to parse ", sum(is.na(timestamps_parsed)), " out of ", nrow(images_merged), " timestamps in images data.\n",
      "Examples of unparseable timestamps: ", paste(paste0("'", example_strings, "'"), collapse = ", "), ".\n",
      "Supported formats are: ", paste(record_datetime_format, collapse = ", "), ".\n",
      "These records will have NA timestamps in the output."
    ))
  }
  
  images_merged$timestamp <- as.character(timestamps_parsed)
  
  # Final checks for empty output tables
  if (nrow(deployments) == 0) {
    warning("The final deployments table is empty after processing.")
  }
  
  if (nrow(deployments_agg) == 0) {
    warning("The aggregated deployments table is empty after processing.")
  }
  
  if (nrow(images_merged) == 0) {
    warning("The merged images table is empty after processing. Check for data compatibility issues.")
  }
  
  return(list(CTtable = deployments, 
              CTtable_aggregated = deployments_agg,
              recordTable = images_merged))
}

library(testthat)
library(lubridate)

# --- Helper: Create Mock Data ---
create_mock_wi_data <- function(type = "csv", 
                                bad_dates = FALSE, 
                                dup_ids = FALSE, 
                                bad_range = FALSE,
                                identical_dates = FALSE,
                                orphan_data = FALSE) { # Added flag to control orphans
  
  # Create a temporary directory
  temp_dir <- file.path(tempdir(), paste0("wi_test_", as.integer(Sys.time())))
  dir.create(temp_dir, showWarnings = FALSE)
  
  # 1. Mock Deployments Data
  deps <- data.frame(
    project_id = 123,
    deployment_id = c("Dep1", "Dep2", "Dep3"),
    placename = c("SiteA", "SiteB", "SiteA"), 
    camera_id = c("Cam1", "Cam2", "Cam3"),
    start_date = c("2023-01-01 12:00:00", "2023-01-02 12:00:00", "2023-01-03 10:00:00"),
    end_date = c("2023-02-01 12:00:00", "2023-02-02 12:00:00", "2023-02-03 10:00:00"),
    latitude = c(10, 20, 30),
    longitude = c(10, 20, 30),
    stringsAsFactors = FALSE
  )
  
  # Inject Errors based on flags
  if (dup_ids) {
    deps$deployment_id[2] <- "Dep1"
  }
  if (bad_dates) {
    deps$start_date[1] <- "Not a date"
  }
  if (bad_range) {
    deps$end_date[1] <- "2022-01-01 12:00:00" 
  }
  if (identical_dates) {
    deps <- rbind(deps, data.frame(
      project_id = 123,
      deployment_id = "Dep_BadTime",
      placename = "SiteC",
      camera_id = "Cam4",
      start_date = "2023-03-01 12:00:00",
      end_date = "2023-03-01 12:00:00",
      latitude = 40, longitude = 40
    ))
  }
  
  # 2. Mock Images Data
  imgs <- data.frame(
    project_id = 123,
    deployment_id = c("Dep1", "Dep1", "Dep2"), # No orphan by default
    image_id = c("Img1", "Img2", "Img3"),
    timestamp = c("2023-01-15 12:00:00", "2023-01-15 12:05:00", "2023-01-15 13:00:00"),
    common_name = c("Deer", "Deer", "Boar"),
    species = c("species1", "species1", "species2"),
    bounding_boxes = c("box1", "box2", "box3"), 
    stringsAsFactors = FALSE
  )
  
  # Inject Orphan Image if requested
  if (orphan_data) {
    imgs <- rbind(imgs, data.frame(
      project_id = 123,
      deployment_id = "Dep_Unknown",
      image_id = "Img4",
      timestamp = "2023-01-15 14:00:00",
      common_name = "Bird",
      species = "species3",
      bounding_boxes = "box4"
    ))
  }
  
  if (identical_dates) {
    imgs <- rbind(imgs, data.frame(
      project_id = 123,
      deployment_id = "Dep_BadTime",
      image_id = "Img5",
      timestamp = "2023-03-01 12:00:00",
      common_name = "Ghost", species = "Ghost", bounding_boxes = NA
    ))
  }
  
  dep_path <- file.path(temp_dir, "deployments.csv")
  img_path <- file.path(temp_dir, "images.csv")
  
  write.csv(deps, dep_path, row.names = FALSE)
  write.csv(imgs, img_path, row.names = FALSE)
  
  if (type == "zip") {
    zip_path <- file.path(temp_dir, "export.zip")
    old_wd <- getwd()
    setwd(temp_dir)
    zip(zipfile = "export.zip", files = c("deployments.csv", "images.csv"))
    setwd(old_wd)
    return(list(zip = zip_path, dir = temp_dir))
  }
  
  return(list(dep = dep_path, img = img_path, dir = temp_dir))
}

# --- UNIT TESTS ---

test_that("Input parameter validation works", {
  paths <- create_mock_wi_data()
  on.exit(unlink(paths$dir, recursive = TRUE))
  
  expect_error(readWildlifeInsights(), 
               "Please provide either 'directory', 'zipfile', or both")
  
  expect_error(readWildlifeInsights(directory = paths$dir, zipfile = "fake.zip"), 
               "Please provide either 'directory', 'zipfile', or both")
  
  expect_error(readWildlifeInsights(deployment_file = "missing.csv", image_file = paths$img), 
               "specified deployment file does not exist")
})

test_that("Reading from individual CSV files works (Happy Path)", {
  # Default orphan_data=FALSE, so no warnings expected
  paths <- create_mock_wi_data(type = "csv", orphan_data = FALSE) 
  on.exit(unlink(paths$dir, recursive = TRUE))
  
  result <- readWildlifeInsights(deployment_file = paths$dep, image_file = paths$img)
  
  expect_type(result, "list")
  expect_named(result, c("CTtable", "CTtable_aggregated", "recordTable"))
  expect_false("bounding_boxes" %in% colnames(result$recordTable))
  expect_equal(nrow(result$CTtable_aggregated), 2) 
})

test_that("Reading from Directory works", {
  # Default orphan_data=FALSE
  paths <- create_mock_wi_data(type = "csv", orphan_data = FALSE)
  on.exit(unlink(paths$dir, recursive = TRUE))
  
  result <- readWildlifeInsights(directory = paths$dir)
  
  expect_equal(nrow(result$CTtable), 3)
  expect_true("timestamp" %in% colnames(result$recordTable))
})

test_that("Reading from Zip file works", {
  skip_if_not(nzchar(Sys.which("zip")))
  
  # Default orphan_data=FALSE
  paths <- create_mock_wi_data(type = "zip", orphan_data = FALSE)
  on.exit(unlink(paths$dir, recursive = TRUE))
  
  result <- readWildlifeInsights(zipfile = paths$zip)
  
  expect_equal(nrow(result$CTtable), 3)
  expect_equal(nrow(result$recordTable), 3) 
})

test_that("Data Integrity: Duplicate Deployment IDs triggers error", {
  paths <- create_mock_wi_data(dup_ids = TRUE)
  on.exit(unlink(paths$dir, recursive = TRUE))
  
  expect_error(readWildlifeInsights(directory = paths$dir), 
               "Duplicate deployment_id found")
})

test_that("Data Integrity: Unparseable Dates triggers error", {
  paths <- create_mock_wi_data(bad_dates = TRUE)
  on.exit(unlink(paths$dir, recursive = TRUE))
  
  # lubridate warns about parsing failures before the function errors out.
  # We suppress warnings to cleanly catch the error we defined.
  expect_error(suppressWarnings(readWildlifeInsights(directory = paths$dir)), 
               "Failed to parse.*start date")
})

test_that("Data Integrity: End date before Start date triggers error", {
  paths <- create_mock_wi_data(bad_range = TRUE)
  on.exit(unlink(paths$dir, recursive = TRUE))
  
  expect_error(readWildlifeInsights(directory = paths$dir), 
               "end date is not after start date")
})

test_that("Logic: Deployments with identical start/end dates are removed", {
  paths <- create_mock_wi_data(identical_dates = TRUE)
  on.exit(unlink(paths$dir, recursive = TRUE))
  
  # Expect warning about removed deployments
  expect_warning(
    result <- readWildlifeInsights(directory = paths$dir),
    "deployments with identical start and end date removed"
  )
  
  expect_false("Dep_BadTime" %in% result$CTtable$deployment_id)
  expect_false("Img5" %in% result$recordTable$image_id)
})



test_that("Logic: Orphan images (no matching deployment) warnings", {
  # Explicitly request orphan data for this test
  paths <- create_mock_wi_data(orphan_data = TRUE) 
  on.exit(unlink(paths$dir, recursive = TRUE))
  
  # 1. Capture all warnings to verify specific messages match without nesting issues
  warn_msgs <- capture_warnings(readWildlifeInsights(directory = paths$dir))
  
  # Verify "Found X deployment_id" warning
  expect_match(warn_msgs, "Found 1 deployment_id", all = FALSE)
  
  # Verify "Merging reduced" warning
  expect_match(warn_msgs, "Merging reduced", all = FALSE)
  
  # 2. Get result for data validation (suppress warnings to prevent console clutter)
  result <- suppressWarnings(readWildlifeInsights(directory = paths$dir))
  
  # Ensure the merged table only has records where deployment existed
  # 4 images created in mock data, 1 was orphan -> 3 remain
  expect_equal(nrow(result$recordTable), 3) 
})

test_that("Logic: Required columns check", {
  paths <- create_mock_wi_data()
  on.exit(unlink(paths$dir, recursive = TRUE))
  
  dep_data <- read.csv(paths$dep)
  dep_data$camera_id <- NULL 
  write.csv(dep_data, paths$dep, row.names = FALSE)
  
  expect_error(readWildlifeInsights(directory = paths$dir),
               "Missing required columns in deployments data")
})

#' @title Aggregate Camera Trap Table to Station Level
#'
#' @description
#' Aggregates a camera trap table from a station-camera (location-deployment) level to a station (location) level.
#' This function is useful when modeling or analysis is conducted at the station level,
#' when multiple cameras were deployed at a single station or a single camera had multiple deployments.
#'
#'
#' @param CTtable A data frame or `sf` object representing the camera trap table.
#'           Each row typically represents a unique camera deployment.
#' @param stationCol A character string specifying the name of the column in \code{CTtable}
#'                   that contains the unique station identifiers.
#' @param cameraCol A character string specifying the name of the column in \code{CTtable}
#'                   that contains the unique camera (deployment) identifiers.
#' @param setupCol character. name of the column containing camera setup dates
#'                  in \code{CTtable}
#' @param retrievalCol character. name of the column containing camera
#'                  retrieval dates in \code{CTtable}
#' @param dateFormat character. The format of columns \code{setupCol} and
#'                  \code{retrievalCol} (and potential problem columns) in \code{CTtable}. Must
#'                  be interpretable by either \code{as.Date} or the "orders" argument
#'                  \code{\link[lubridate]{parse_date_time}} in \pkg{lubridate}. Can be a date
#'                  or a date-time.
#' @param quiet logical. If TRUE, print message if no aggregation is needed.
#'
#' @details
#' The aggregation logic handles different column types as follows:
#' \itemize{
#'   \item \strong{Station ID (specified by `stationCol`):} Remains unique for each row in the output.
#'   \item \strong{Camera ID (if present):} A character vector (comma-separated)
#'         of unique camera IDs deployed at that station.
#'   \item \strong{Setup/Retrieval Dates:} The earliest setup date and the latest retrieval date
#'         for all deployments within a station will be retained.
#'   \item \strong{Numeric Columns:} The `mean` of all values for that station will be calculated.
#'   \item \strong{Logical Columns:} The `mean` (which effectively calculates the proportion of `TRUE`s)
#'         of all values for that station will be calculated.
#'   \item \strong{Character Columns (other than Camera ID):} A character vector of unique values
#'         (comma-separated) for that station will be created.
#'   \item \strong{Factor Columns:} Similar to character columns, unique levels will be combined.
#'   \item \strong{Geometry Column (`sf` objects):} Centroid (midpoint) of all unique points at a 
#'          station is calculated.
#' }
#' 
#' It is recommended to inspect the aggregated output carefully, especially for columns
#' with mixed data types or specific aggregation requirements not covered by the defaults.
#'
#' @note The function is mainly intended for aggregating covariates to station level. 
#'          It does currently not handle camera malfunction (via columns `ProblemX_from` / `ProblemX_to`)
#'          and does not provide proper handling of problem columns.
#'          Use \code{\link{cameraOperation}} to aggregate camera trap tables to station level for 
#'          analyses while accounting for camera malfunction / problem periods.
#' 
#' @return A data frame with one row per unique station, containing aggregated information.
#'          Spatial information (for `sf` objects) is preserved.
#'         
#' 
#' @export
#' 
#' @importFrom dplyr summarise
#' @importFrom lubridate parse_date_time as_date
#'
# #' @examples
#' 
#' 
aggregateStations <- function(CTtable, 
                              stationCol,
                              cameraCol = NULL,
                              setupCol = NULL,
                              retrievalCol = NULL,
                              dateFormat = NULL,
                              quiet = FALSE) {
  
  
  # --- Handle sf objects ---
  is_sf_input <- inherits(CTtable, "sf")
  
  if (is_sf_input) {
    # message("Input 'CTtable' is an sf object. Geometry column will be dropped during aggregation.")
    df <- sf::st_drop_geometry(CTtable)
    df_geometry <- CTtable[, stationCol]
  } else {
    df <- as.data.frame(CTtable)  
  }
  
  
  
  # --- Input Validation ---
  required_cols <- c(stationCol)
  if (!is.null(cameraCol)) required_cols <- c(required_cols, cameraCol)
  if (!is.null(setupCol)) required_cols <- c(required_cols, setupCol)
  if (!is.null(retrievalCol)) required_cols <- c(required_cols, retrievalCol)
  
  missing_cols <- required_cols[!required_cols %in% names(df)]
  if (length(missing_cols) > 0) {
    stop(paste("Error in aggregateStations: The following required columns were not found in 'CTtable':",
               paste(missing_cols, collapse = ", "),
               ". Available columns are:", paste(names(df), collapse = ", ")))
  }
  
  # Validate dateFormat if date columns are provided and are characters
  if ((!is.null(setupCol) && is.character(df[[setupCol]])) ||
      (!is.null(retrievalCol) && is.character(df[[retrievalCol]]))) {
    if (is.null(dateFormat)) {
      stop("Error in aggregateStations: 'dateFormat' must be provided if 'setupCol' or 'retrievalCol' are character strings.")
    }
  }
  
  
  # --- Early exit if no aggregation needed ---
  if (all(table(df[[stationCol]]) == 1)) {
    if(!is.logical(quiet)) stop("'quiet' must be TRUE or FALSE")
    if(isFALSE(quiet)) message("No aggregation needed: All 'stationCol' values are already unique.")
    return(df)
  }
  
  
  
  # --- Coerce date/datetime columns to appropriate types using lubridate ---
  # Determine if date format is likely date-only or datetime
  is_datetime_format <- !is.null(dateFormat) && grepl("[HhMmsS]", dateFormat)
  
  if (!is.null(setupCol) && is.character(df[[setupCol]])) {
    if (is_datetime_format) {
      df[[setupCol]] <- lubridate::parse_date_time(df[[setupCol]], orders = dateFormat)
    } else {
      df[[setupCol]] <- lubridate::as_date(df[[setupCol]], format = dateFormat)
    }
    if (anyNA(df[[setupCol]]) && any(is.na(df[[setupCol]][!is.na(CTtable[[setupCol]])]))) {
      stop(paste0("Error in aggregateStations: Failed to convert '", setupCol, "' to date/datetime using format '", dateFormat, "'. Check 'dateFormat' or data consistency."))
    }
  }
  if (!is.null(retrievalCol) && is.character(df[[retrievalCol]])) {
    if (is_datetime_format) {
      df[[retrievalCol]] <- lubridate::parse_date_time(df[[retrievalCol]], orders = dateFormat)
    } else {
      df[[retrievalCol]] <- lubridate::as_date(df[[retrievalCol]], format = dateFormat)
    }
    if (anyNA(df[[retrievalCol]]) && any(is.na(df[[retrievalCol]][!is.na(CTtable[[retrievalCol]])]))) {
      stop(paste0("Error in aggregateStations: Failed to convert '", retrievalCol, "' to date/datetime using format '", dateFormat, "'. Check 'dateFormat' or data consistency."))
    }
  }
  
  
  
  # --- Build the aggregation list for dplyr::summarise ---
  agg_list <- list()
  
  for (col_name in names(df)) {
    if (col_name == stationCol) {
      next # This is the grouping variable
    } else if (!is.null(cameraCol) && col_name == cameraCol) {
      # Use expr() and .data[[col_name]]
      agg_list[[col_name]] <- rlang::expr(paste(sort(unique(.data[[!!sym(col_name)]])), collapse = ", "))
    } else if (!is.null(setupCol) && col_name == setupCol) {
      agg_list[[col_name]] <- rlang::expr(min(.data[[!!sym(col_name)]], na.rm = TRUE))
    } else if (!is.null(retrievalCol) && col_name == retrievalCol) {
      agg_list[[col_name]] <- rlang::expr(max(.data[[!!sym(col_name)]], na.rm = TRUE))
    } else if (is.numeric(df[[col_name]])) {
      agg_list[[col_name]] <- rlang::expr(mean(.data[[!!sym(col_name)]], na.rm = TRUE))
    } else if (is.logical(df[[col_name]])) {
      agg_list[[col_name]] <- rlang::expr(mean(as.integer(.data[[!!sym(col_name)]]), na.rm = TRUE))
    } else if (is.character(df[[col_name]]) || is.factor(df[[col_name]])) {
      agg_list[[col_name]] <- rlang::expr(paste(sort(unique(.data[[!!sym(col_name)]])), collapse = ", "))
    } else if (lubridate::is.Date(df[[col_name]]) || lubridate::is.POSIXct(df[[col_name]]) || lubridate::is.POSIXlt(df[[col_name]])) {
      warning(paste0("Column '", col_name, "' is a date/datetime type not specified as 'setupCol' or 'retrievalCol'. Aggregating by pasting unique values. Consider handling this column manually if this is not desired."))
      agg_list[[col_name]] <- rlang::expr(paste(sort(unique(.data[[!!sym(col_name)]])), collapse = ", "))
    } else {
      warning(paste0("Column '", col_name, "' has an unhandled data type and will be aggregated by pasting unique values. Consider converting it to a supported type if this is not desired."))
      agg_list[[col_name]] <- rlang::expr(paste(sort(unique(.data[[!!sym(col_name)]])), collapse = ", "))
    }
  }

  
  # Perform aggregation using dplyr
  df_agg <- df |>
    dplyr::group_by(!!dplyr::sym(stationCol)) |>
    dplyr::summarise(!!!agg_list, .groups = "drop")
  
  
  if(is_sf_input){
    df_agg_geometry <- df_geometry |> 
      group_by(!!dplyr::sym(stationCol)) |>
      dplyr::summarise(geometry = sf::st_centroid(sf::st_union(.data$geometry)),
                       .groups = "drop")
    
    df_agg <- dplyr::left_join(df_agg, df_agg_geometry, by = stationCol) |>
      sf::st_as_sf()   # Convert back to sf
  }
  
  return(df_agg)
}

#' Create a report about a camera trapping survey and species detections
#' 
#' This function creates a report about a camera trapping survey and species
#' records. It uses a camera trap station information table and a record table
#' (generated with \code{\link{recordTable}}) as input. Output tables can be
#' saved and a zip file for simple data sharing can be created easily.
#' 
#' 
#' \code{dateFormat} defaults to "YYYY-MM-DD", e.g. "2014-10-31". It can be
#' specified either in the format required by \code{\link[base]{strptime}} or
#' the 'orders' argument in \code{\link[lubridate]{parse_date_time}} in
#' \pkg{lubridate}. In the example above, "YYYY-MM-DD" would be specified as
#' "\%Y-\%m-\%d" or "ymd".
#' 
#' \code{recordDateTimeFormat} defaults to the "YYYY-MM-DD HH:MM:SS"
#' convention, e.g. "2014-09-30 22:59:59". \code{recordDateTimeFormat} can be
#' interpreted either by base-R via \code{\link[base]{strptime}} or in
#' \pkg{lubridate} via \code{\link[lubridate]{parse_date_time}} (argument
#' "orders"). \pkg{lubridate} will be used if there are no "\%" characters in
#' \code{recordDateTimeFormat}.
#' 
#' For "YYYY-MM-DD HH:MM:SS", \code{recordDateTimeFormat} would be either
#' "\%Y-\%m-\%d \%H:\%M:\%S" or "ymd HMS". For details on how to specify date
#' and time formats in R see \code{\link[base]{strptime}} or
#' \code{\link[lubridate]{parse_date_time}}.
#' 
#' Note: as of version 2.1, argument \code{CTHasProblems} is deprecated and
#' defunct. Please use \code{camOp} instead to provide information about
#' periods of camera activity and malfunction. If \code{camOp} is not provided
#' the legacy version of surveyReport (from camtrapR 2.0.3) will be run with a
#' warning.
#' 
#' @param recordTable data.frame containing a species record table as given by
#' \code{\link{recordTable}}
#' @param CTtable data.frame containing information about location and trapping
#' period of camera trap stations (equivalent to \code{\link{camtraps}}
#' @param camOp camera operation matrix created with
#' \code{\link{cameraOperation}}
#' @param speciesCol character. name of the column specifying Species ID in
#' \code{recordTable}
#' @param stationCol character. name of the column specifying Station ID in
#' \code{CTtable} and \code{recordTable}
#' @param cameraCol character. name of the column specifying Camera ID in
#' \code{CTtable} and \code{recordTable}
#' @param setupCol character. name of the column containing camera setup dates
#' in \code{CTtable}
#' @param retrievalCol character. name of the column containing camera
#' retrieval dates in \code{CTtable}
#' @param CTDateFormat character. The format of columns \code{setupCol} and
#' \code{retrievalCol} (and potential problem columns) in \code{CTtable}. Must
#' be interpretable by either \code{as.Date} or the "orders" argument
#' \code{\link[lubridate]{parse_date_time}} in \pkg{lubridate}.
#' @param CTHasProblems deprecated (since version 2.1)
#' @param recordDateTimeCol character. The name of the column containing date
#' and time of records in \code{recordTable}
#' @param recordDateTimeFormat character. The date/time format of column
#' \code{recordDateTimeCol} in \code{recordTable}.
#' @param Xcol character. name of the column specifying x coordinates in
#' \code{CTtable}. Used to create detection maps if \code{makezip} is TRUE.
#' (optional)
#' @param Ycol character. name of the column specifying y coordinates in
#' \code{CTtable}. Used to create detection maps if \code{makezip} is TRUE.
#' (optional)
#' @param sinkpath character. The directory into which the survey report is
#' saved (optional)
#' @param makezip logical. Create a zip file containing tables, plots and maps
#' in \code{sinkpath}?
#' 
#' @return An invisible list containing 5 \code{data.frames}.
#' 
#' \item{survey_dates}{station and image date ranges, number of total and
#' active trap days (calendar days and taking into account independent effort
#' of multiple cameras, if applicable), number of cameras per station}
#' \item{species_by_station}{species numbers by station}
#' \item{events_by_species}{number of events and stations by species}
#' \item{events_by_station}{number of events for every species by station (only
#' species that were recorded)} \item{events_by_station2}{number of events for
#' all species at all stations (including species that were not recorded)}
#' 
#' The output will be saved to a .txt file if \code{sinkpath} is defined.
#' 
#' If \code{makezip} is TRUE, a zip file will be created in \code{sinkpath}. It
#' contains single-species activity plots, detection maps (if \code{Xcol} and
#' \code{Ycol} are defined), the survey report tables, the record table and the
#' camera trap station table, and an example R script.
#' 
#' @author Juergen Niedballa
#' 
#' @seealso \code{\link{recordTable}}
#' 
#' @examples
#' 
#' data(camtraps)
#' data(recordTableSample)
#' 
#' # since version 2.1, camera operation matrix is required as input
#' 
#' camop_no_problem <- cameraOperation(CTtable      = camtraps,
#'                                     stationCol   = "Station",
#'                                     setupCol     = "Setup_date",
#'                                     retrievalCol = "Retrieval_date",
#'                                     writecsv     = FALSE,
#'                                     hasProblems  = FALSE,
#'                                     dateFormat   = "dmy"
#' )
#' 
#' reportTest <- surveyReport (recordTable          = recordTableSample,
#'                             CTtable              = camtraps,
#'                             camOp                = camop_no_problem,
#'                             speciesCol           = "Species",
#'                             stationCol           = "Station",
#'                             setupCol             = "Setup_date",
#'                             retrievalCol         = "Retrieval_date",
#'                             CTDateFormat         = "dmy", 
#'                             recordDateTimeCol    = "DateTimeOriginal",
#'                             recordDateTimeFormat = "ymd HMS")
#' 
#' class(reportTest)  # a list with
#' length(reportTest) # 5 elements
#' 
#' reportTest[[1]]    # camera trap operation times and image date ranges
#' reportTest[[2]]    # number of species by station
#' reportTest[[3]]    # number of events and number of stations by species
#' reportTest[[4]]    # number of species events by station
#' reportTest[[5]]    # number of species events by station including 0s (non-observed species)
#' 
#' # with camera problems
#' 
#' camop_problem <- cameraOperation(CTtable      = camtraps,
#'                                  stationCol   = "Station",
#'                                  setupCol     = "Setup_date",
#'                                  retrievalCol = "Retrieval_date",
#'                                  writecsv     = FALSE,
#'                                  hasProblems  = TRUE,
#'                                  dateFormat   = "dmy"
#' )
#' 
#' reportTest_problem <- surveyReport (recordTable          = recordTableSample,
#'                                     CTtable              = camtraps,
#'                                     camOp                = camop_problem,
#'                                     speciesCol           = "Species",
#'                                     stationCol           = "Station",
#'                                     setupCol             = "Setup_date",
#'                                     retrievalCol         = "Retrieval_date",
#'                                     CTDateFormat         = "dmy", 
#'                                     recordDateTimeCol    = "DateTimeOriginal",
#'                                     recordDateTimeFormat = "ymd HMS")
#' 
#' reportTest_problem$survey_dates
#' 
#' 
#' 
#' 
#' \dontrun{
#' # run again with sinkpath defined
#' reportTest <- surveyReport (recordTable          = recordTableSample,
#'                             CTtable              = camtraps,
#'                             camOp                = camop_no_problem,
#'                             speciesCol           = "Species",
#'                             stationCol           = "Station",
#'                             setupCol             = "Setup_date",
#'                             retrievalCol         = "Retrieval_date",
#'                             CTDateFormat         = "dmy",, 
#'                             recordDateTimeCol    = "DateTimeOriginal",
#'                             recordDateTimeFormat = "ymd HMS",
#'                             sinkpath             = getwd())
#' 
#' # have a look at the text file
#' readLines(list.files(getwd(), pattern = paste("survey_report_", Sys.Date(), ".txt", sep = ""), 
#' 					 full.names = TRUE))
#' }
#' 
#' @export surveyReport
#' 
surveyReport <- function(recordTable,
                         CTtable,
                         camOp,
                         speciesCol = "Species",
                         stationCol = "Station",
                         cameraCol,
                         setupCol,
                         retrievalCol,
                         CTDateFormat = "ymd",
                         CTHasProblems = "deprecated",
                         recordDateTimeCol = "DateTimeOriginal",
                         recordDateTimeFormat = "ymd HMS",
                         Xcol,
                         Ycol,
                         sinkpath,
                         makezip
){
  
  # To do: 
  # - setup/retrieval with hour (seems to work, when using base R date-time string is converted to date.)
  # - sessionCol? As long as camOp and CTtable match it should be fine. But all aggregate commands need a session column (dummy if only one session).
  # - explanation in vignette / help file
  
  
  
  # for users running old (<= 2.0.3) code: error out if camop not defined
  
  if(!hasArg(camOp)){
    stop("Argument 'camOp' is missing (expected since camtrapR 2.1). The legacy version of surveyReport which did not require camOp is no longer supported.",
            immediate. = TRUE)
  }
  
  
  # check column names
  checkForSpacesInColumnNames(stationCol = stationCol, 
                              setupCol = setupCol, 
                              retrievalCol = retrievalCol,
                              recordDateTimeCol = recordDateTimeCol, 
                              speciesCol = speciesCol)
  
  # check input
  if(!hasArg(recordTable))         stop("'recordTable' is not defined", call. = FALSE)
  if(!is.data.frame(recordTable))  stop("'recordTable' must be a data.frame", call. = FALSE)
  if(hasArg(CTHasProblems))        message("Since version 2.1, argument CTHasProblems is deprecated. It was replaced with the new argument camOp. Periods when cameras malfunctioned are now assinged via camOp.")
  if(!hasArg(camOp))               stop("'camOp' is not defined", call. = FALSE)
  if(!inherits(camOp, "matrix"))   stop("camOp must be a matrix", call. = FALSE)
  if(any(camOp < 0, na.rm = TRUE)) stop("camOp may only contain positive values")
  
  camop.info.df <- deparseCamOpRownames(camOp)   # only works if camOp is station-specific
  if("camera" %in% colnames(camop.info.df))  stop("camera operation matrix is camera-specific. Please provide a camera operation matrix that is by station (byCamera = FALSE).\nThis is necessary for correctly calculating the number of active days (taking into account whether cameras were independent or not).", call. = FALSE)
  if("session" %in% colnames(camop.info.df)) stop("camera operation matrix contains sessions. This is currently not supported. Please run the survey report separately for the different sessions.", call. = FALSE)
  colnames(camop.info.df) <- stationCol
  
  
  CTtable     <- dataFrameTibbleCheck(df = CTtable)
  recordTable <- dataFrameTibbleCheck(df = recordTable)
  
  if(!stationCol %in% colnames(CTtable))    stop(paste('stationCol = "',   stationCol,   '" is not a column name in CTtable', sep = ''), call. = FALSE)
  if(!setupCol %in% colnames(CTtable))      stop(paste('setupCol = "',     setupCol,     '" is not a column name in CTtable', sep = ''), call. = FALSE)
  if(!retrievalCol %in% colnames(CTtable))  stop(paste('retrievalCol = "', retrievalCol, '" is not a column name in CTtable', sep = ''), call. = FALSE)
  
  if(!stationCol %in% colnames(recordTable))         stop(paste('stationCol = "',   stationCol,  '" is not a column name in recordTable', sep = ''), call. = FALSE)
  if(!recordDateTimeCol %in% colnames(recordTable))  stop(paste('recordDateTimeCol = "', recordDateTimeCol,  '" is not a column name in recordTable', sep = ''), call. = FALSE)
  if(!speciesCol %in% colnames(recordTable))         stop(paste('speciesCol = "', speciesCol,  '" is not a column name in recordTable', sep = ''), call. = FALSE)
  
  
  # make columns character
  recordTable[,speciesCol]        <- as.character(recordTable[,speciesCol])
  recordTable[,stationCol]        <- as.character(recordTable[,stationCol])
  recordTable[,recordDateTimeCol] <- as.character(recordTable[,recordDateTimeCol])
  
  CTtable[,stationCol]    <- as.character(CTtable[,stationCol])
  CTtable[,setupCol]      <- as.character(CTtable[,setupCol])
  CTtable[,retrievalCol]  <- as.character(CTtable[,retrievalCol])
  
  
  
  if(hasArg(makezip)){
    stopifnot(is.logical(makezip))
  } else {
    makezip <- FALSE
  }
  
  if(isTRUE(makezip)){
    if(hasArg(sinkpath) == FALSE) stop("if makezip is TRUE, please define sinkpath")
    if(!dir.exists(sinkpath)) stop("Could not find sinkpath:\n", sinkpath, call. = FALSE)
    
    if(Sys.getenv("R_ZIPCMD") == "" & 
       Sys.getenv("zip") == "") {
      if(isFALSE(requireNamespace("zip", quietly = TRUE))) {
        stop('cannot create zip file. Try installing the package "zip" via:  install.packages("zip")', call. = FALSE)
      } else {
        requireNamespace("zip")
        usePackageZip <- TRUE
      }
    } else {
      usePackageZip <- FALSE
    }
  }
  
  if(hasArg(cameraCol)){
    if(cameraCol %in% colnames(CTtable) == FALSE) stop(paste(cameraCol, "is not a column of CTtable"))
  } else {
    if(any(table(CTtable[,stationCol]) > 1)){
      stop("at least 1 station has more than 1 item in CTtable. Please specify 'cameraCol'")
    }
  }
  
  if(hasArg(Xcol)){
    stopifnot(hasArg(Ycol))
    stopifnot(c(Xcol, Ycol) %in% colnames(CTtable))
    CTtable[,Xcol] <- as.numeric(as.character(CTtable[,Xcol]))
    CTtable[,Ycol] <- as.numeric(as.character(CTtable[,Ycol]))
  } 
  
  
  recordTable$DateTime2 <- parseDateTimeObject(inputColumn = recordTable[,recordDateTimeCol],
                                               dateTimeFormat = recordDateTimeFormat,
                                               timeZone = "UTC")
  recordTable$Date2 <- as.Date(recordTable$DateTime2, tz = "UTC")
  
  
  # if(all(as.character(unique(recordTable[,stationCol])) %in% CTtable[,stationCol]) == FALSE){
  #   (stop("Not all values of stationCol in recordTable are matched by values of stationCol in CTtable"))
  # }
  if(!all(as.character(unique(recordTable[,stationCol])) %in% CTtable[,stationCol])){
    stop(paste("Not all values of stationCol in recordTable are matched by values of stationCol in CTtable:\n",
               paste(as.character(unique(recordTable[,stationCol])) [!as.character(unique(recordTable[,stationCol])) %in% CTtable[,stationCol]], collapse = ", ")
    ))
  }
  
  
  # check date columns and format
  CTtable[,setupCol]     <- parseDateObject(inputColumn = CTtable[,setupCol],     CTDateFormat, checkNA = TRUE, checkEmpty = TRUE)
  CTtable[,retrievalCol] <- parseDateObject(inputColumn = CTtable[,retrievalCol], CTDateFormat, checkNA = TRUE, checkEmpty = TRUE)
  
  
  # station and image date ranges
  station.tmp1 <- aggregate(CTtable[,setupCol],
                            list(CTtable[,stationCol]),
                            FUN = min)
  station.tmp2 <- aggregate(CTtable[,retrievalCol],
                            list(CTtable[,stationCol]),
                            FUN = max)
  image.tmp1   <- aggregate(recordTable$Date2,
                            list(recordTable[,stationCol]),
                            FUN = min)
  image.tmp2   <- aggregate(recordTable$Date2,
                            list(recordTable[,stationCol]),
                            FUN = max)
  n_cameras.tmp   <- aggregate(CTtable[,stationCol],  # is this correct when sessionCol is present?
                               list(CTtable[,stationCol]),
                               FUN = length)

  colnames(n_cameras.tmp) <- c(stationCol, "n_cameras")

  
  df1  <- data.frame(station.tmp1, station.tmp2[,2])
  colnames(df1) <- c(stationCol, "setup", "retrieval")
  df2  <- data.frame(image.tmp1, image.tmp2[,2])
  colnames(df2) <- c(stationCol, "image_first", "image_last")
  
  station_dates <- merge(merge(df1, 
                               df2, 
                               by = stationCol, 
                               all = T), 
                         n_cameras.tmp)
  
  # station_dates <- data.frame(station     = station.tmp1[,1], 
  #                             setup       = station.tmp1[,2],
  #                             image_first = image.tmp1[,2],
  #                             image_last  = image.tmp2[,2],
  #                             retrieval   = station.tmp2[,2],
  #                             n_cameras   = n_cameras.tmp[,2])
  
  n_calendar_days_total <- rowSums(camOp >= 0, na.rm =TRUE)
  
  # Calendar days on which cameras were active, by row.
  n_calendar_days_active <- rowSums(camOp > 0, na.rm =TRUE)
  
  # sum of inactive calendar days (by row)
  n_calendar_days_inactive <- rowSums(camOp == 0, na.rm =TRUE)
  
  # trap-nights (counting independent cameras separately)
  # per station, number of active cameras * number of active days
  n_trap_nights_active <- rowSums(camOp, na.rm =TRUE) 
  
  
  
  #######################
  # # Legacy calculation of active days - for the sake of continuity - full old code
  
  CTHasProblems <- TRUE
  
  # check that problems are arranged in order 1,2,3,...
  cols.prob.from <- grep(colnames(CTtable), pattern = "Problem\\d\\Sfrom")
  cols.prob.to   <- grep(colnames(CTtable), pattern = "Problem\\d\\Sto")
  
  if(length(cols.prob.from) == 0 | length(cols.prob.to) == 0) CTHasProblems <- FALSE 
  if(!all(order(colnames(CTtable)[cols.prob.from]) == seq(1:length(cols.prob.from)))) CTHasProblems <- FALSE 
  if(!all(order(colnames(CTtable)[cols.prob.to])   == seq(1:length(cols.prob.to)))) CTHasProblems <- FALSE 
  if(length(cols.prob.from) != length(cols.prob.to)) CTHasProblems <- FALSE 
  if(all(is.na(CTtable[,cols.prob.from])) | all(is.na(CTtable[,cols.prob.to]))) CTHasProblems <- FALSE 
  
  if(isTRUE(CTHasProblems)){    # camera problem columns
    
    # convert problem column entries to character
    for(problem_col_index in c(cols.prob.from, cols.prob.to)){
      CTtable[, problem_col_index] <- as.character(CTtable[, problem_col_index])
    }
    
    n_days_inactive_legacy <- data.frame(matrix(NA,
                                         ncol = length(cols.prob.from),
                                         nrow = nrow(CTtable)))
    
    for(xy in 1:length(cols.prob.from)){
      
      CTtable[, cols.prob.from[xy]] <- parseDateObject(inputColumn = CTtable[, cols.prob.from[xy]], CTDateFormat, checkNA = FALSE, checkEmpty = FALSE)
      CTtable[, cols.prob.to[xy]]   <- parseDateObject(inputColumn = CTtable[, cols.prob.to[xy]],   CTDateFormat, checkNA = FALSE, checkEmpty = FALSE)
      
      n_days_inactive_legacy[,xy] <- CTtable[cols.prob.to[xy]] - CTtable[cols.prob.from[xy]]       # compute number of inactive trap nights
      # EDIT 2021-01-17: the line above computes time difference between start and end of problem. So yesterday and end today = 1 day, not 2 days as in the new version
      # basically, that's assuming problems began / ended at noon (for the sake of computing the effort)
      # hence, the number of inactive days was 1 less in the old version compared to the new version. 
      # if problem_to = retrieval, difference will only be 0.5 per camera (because in new version, retrieval is assumed to be 12 noon)
      
      n_days_inactive_legacy[,xy] <- as.integer(n_days_inactive_legacy[,xy])
    }

    n_days_inactive_rowsum_legacy <- rowSums(n_days_inactive_legacy, na.rm = TRUE)
  } else {
    n_days_inactive_rowsum_legacy <- rep(0, times = nrow(CTtable))
  }
  stopifnot(length(n_days_inactive_rowsum_legacy) == nrow(CTtable))
  
  n_days_inactive_rowsum_legacy <- aggregate(n_days_inactive_rowsum_legacy,
                                             by    = list(CTtable[,stationCol]),
                                             FUN   = sum,
                                             na.rm = TRUE)

  n_nights_total_legacy      <- as.integer(CTtable[,retrievalCol] - CTtable[,setupCol])
  n_nights_total_agg_legacy  <- aggregate(n_nights_total_legacy,
                                          by  = list(CTtable[,stationCol]),
                                          FUN = sum)
  # order properly
  n_nights_total_agg_legacy <- n_nights_total_agg_legacy[match(camop.info.df[, stationCol], n_nights_total_agg_legacy$Group.1),]
  n_days_inactive_rowsum_legacy <- n_days_inactive_rowsum_legacy[match(camop.info.df[, stationCol], n_days_inactive_rowsum_legacy$Group.1),]
  
  
  n_nights_active_legacy    <- n_nights_total_agg_legacy[,2] - n_days_inactive_rowsum_legacy[,2]

  
  ############################
  # make data frame with the number of active / inactive days
  if(!all(station_dates$station == camop.info.df$station)) stop("Mismatch between station column in CTtable and camOp")
  
  camop.info.df2 <- data.frame(camop.info.df,
                               n_calendar_days_total    = n_calendar_days_total,    # this is 1 for setup/retrieval on consecutive days
                               n_calendar_days_active   = n_calendar_days_active,   # this is 2 for setup/retrieval on consecutive days
                               n_calendar_days_inactive = n_calendar_days_inactive, # number of days stations was inactive
                               n_trap_nights_active     = n_trap_nights_active      # this is precisely how long the cameras were active ( 1 for setup/retrieval on consecutive days - for each camera).
                               # n_nights_active_legacy   = n_nights_active_legacy,
                               # n_nights_total_legacy    = n_nights_total_agg_legacy$x
                               )       
  
  
  
  station.info.combined <- merge(station_dates, 
                                 camop.info.df2, 
                                 by = stationCol)
  
  # adjust options for printing results
  options.tmp <- options()
  on.exit(options(options.tmp))
  options(max.print=1e6)
  options(width = 1000)
  
  
  # sink/print output
  if(hasArg(sinkpath)){
    sinkfile <- file.path(sinkpath, paste("survey_report_", Sys.Date(), ".txt", sep = ""))
    sink(file = sinkfile)
    print(paste("Survey Report generated", Sys.Date() ))
  }
  cat("\n-------------------------------------------------------\n")
  print(paste("Total number of stations: ", length(unique(CTtable[,stationCol]))))
  
  cat("\n-------------------------------------------------------\n")
  print(paste("Number of operational stations: ", sum(station.info.combined$n_trap_nights_active > 0)))
  
  if(hasArg(cameraCol)){
    cat("\n-------------------------------------------------------\n")
    print(paste("Total number of cameras: ", length(unique(paste(CTtable[,stationCol], CTtable[,cameraCol], sep = "_")))))
  }
  
  cat("\n-------------------------------------------------------\n")
  print(paste("Trap nights (number of active 24 hour cycles completed by independent cameras): ",
              round(sum(station.info.combined$n_trap_nights_active), 1)
  ))
  
  # cat("\n-------------------------------------------------------\n")
  # print(paste("n nights with cameras set up and active (trap nights - LECAGY CALCULATION - WHOLE DAYS): ",
  #             sum(station.info.combined$n_nights_active_legacy)
  # ))
  
  # cat("\n-------------------------------------------------------\n")
  # print(paste("n nights with cameras set up (LECAGY CALCULATION - WHOLE DAYS): ",
  #             sum(station.info.combined$n_nights_total_legacy)
  # ))
  
  cat("\n-------------------------------------------------------\n")
  print(paste("Calendar days with cameras set up (operational or not): ",
              sum(station.info.combined$n_calendar_days_total)
  ))
  
  cat("\n-------------------------------------------------------\n")
  print(paste("Calendar days with cameras set up and active: ",
              sum(station.info.combined$n_calendar_days_active)
  ))
  
  cat("\n-------------------------------------------------------\n")
  print(paste("Calendar days with cameras set up but inactive: ",
              sum(station.info.combined$n_calendar_days_inactive)
  ))
  
  cat("\n-------------------------------------------------------\n")
  print(paste("total trapping period: ", paste(min(station.info.combined$setup), max(station.info.combined$retrieval), sep = " - ")))
  
  
  
  # total number of independent records by species
  species_record_table <- data.frame(species    = rep(NA, times = length(unique(recordTable[, speciesCol]))),
                                     n_events   = rep(NA, times = length(unique(recordTable[, speciesCol]))), 
                                     n_stations = rep(NA, times = length(unique(recordTable[, speciesCol]))))
  
  for(i in 1:length(unique(recordTable[, speciesCol]))){
    
    tmp                           <- unique(recordTable[, speciesCol])[i]
    subset.tmp                    <- subset(recordTable, recordTable[, speciesCol] == tmp)
    species_record_table[i, ]     <- c(tmp, nrow(subset.tmp), length(unique(subset.tmp[,stationCol])))
    rm(subset.tmp, tmp)
  }
  species_record_table2           <- species_record_table[order(species_record_table$species),]
  rownames(species_record_table2) <- NULL
  
  # total number of independent records by station
  
  # only species that were recorded
  station_record_table1           <- aggregate(recordTable[,1], by = list(recordTable[,stationCol],recordTable[,speciesCol]), FUN = length)
  colnames(station_record_table1) <- c(stationCol, speciesCol, "n_events")
  station_record_table1           <- station_record_table1[order(station_record_table1[,stationCol], station_record_table1[,speciesCol]),]
  rownames(station_record_table1) <- NULL
  
  #including all species and 0s
  station_record_table           <- expand.grid(sort(unique(recordTable[,stationCol])), sort(unique(recordTable[,speciesCol])))
  station_record_table           <- data.frame(station_record_table, n_events = 0)
  colnames(station_record_table) <- c(stationCol, speciesCol, "n_events")
  rownames(station_record_table) <- NULL
  # species lists by station
  
  n_spec_by_station           <- aggregate(station_record_table1[,speciesCol], by = list(station_record_table1[,stationCol]), FUN = length)
  colnames(n_spec_by_station) <- c(stationCol, "n_species")
  rownames(n_spec_by_station) <- NULL
  
  for(j in 1:length(unique(recordTable[, stationCol]))){
    
    tmp                      <- unique(recordTable[, stationCol])[j]
    subset.tmp               <- table(subset(recordTable, recordTable[, stationCol] == tmp)[,speciesCol] )
    
    station_record_table.tmp <- station_record_table[station_record_table[, stationCol] == tmp,]
    station_record_table.tmp$n_events[match(names(subset.tmp), station_record_table.tmp[,speciesCol])] <- subset.tmp
    
    station_record_table[station_record_table[, stationCol] == tmp,] <- station_record_table.tmp
    rm(station_record_table.tmp)
  }
  station_record_table2 <-  station_record_table[order(station_record_table[,stationCol], station_record_table[,speciesCol]),]
  rownames(station_record_table2) <- NULL
  
  if(hasArg(sinkpath)){
    cat("\n\n-------------------------------------------------------\n\n")
    print(" survey station and image date ranges")
    print(station.info.combined) #print(date_range_combined)
    cat("\n\n-------------------------------------------------------\n\n")
    print(" number of species by station")
    print(n_spec_by_station)
    cat("\n\n-------------------------------------------------------\n\n")
    print(" number of events and station by species")
    print(species_record_table2)
    cat("\n\n-------------------------------------------------------\n\n")
    print(" number of events and species by station (only species that were recorded at stations)")
    print(station_record_table1)
    cat("\n\n-------------------------------------------------------\n\n")
    print(" number of events and species by station (all species, all stations, including species that were not recorded)")
    print(station_record_table2)
    sink()
    message("saved output to file \n",
            paste(sinkfile, "\n\n"))
  }
  
  output <- list(station.info.combined, #date_range_combined, 
                 n_spec_by_station, species_record_table2, station_record_table1, station_record_table2)
  names(output) <- c("survey_dates", "species_by_station", "events_by_species",
                     "events_by_station", "events_by_station2")
  
  # make zip file
  if(isTRUE(makezip)){
    
    arglist_zip <-  list(output               = output,
                         recordTable          = recordTable,
                         CTtable              = CTtable,
                         #camOp                = camOp,
                         speciesCol           = speciesCol,
                         stationCol           = stationCol,
                         setupCol             = setupCol,
                         retrievalCol         = retrievalCol,
                         CTDateFormat         = CTDateFormat,
                         CTHasProblems        = CTHasProblems,
                         recordDateTimeCol    = recordDateTimeCol,
                         recordDateTimeFormat = recordDateTimeFormat,
                         sinkpath             = sinkpath,
                         usePackageZip        = usePackageZip)
    
    if(hasArg(cameraCol)) arglist_zip <- c(arglist_zip,  cameraCol = cameraCol)
    if(hasArg(Xcol) & hasArg(Ycol)) arglist_zip <- c(arglist_zip,  Xcol = Xcol, Ycol = Ycol)
    
    do.call(makeSurveyZip, arglist_zip)
    
  }
  
  return(invisible(output))
}

context("surveyDashboard")

library(camtrapR)
library(shiny)


# Basic usage with minimal parameters

data("camtraps")
data("recordTableSample")



test_that("surveyDashboard can be launched without parameters without error", {
  
  testthat::skip_if_not_installed("mapview")
  testthat::skip_if_not_installed("plotly")
  testthat::skip_if_not_installed("corrplot")
  testthat::skip_if_not_installed("unmarked")
  testthat::skip_if_not_installed("ubms")
  
  testServer(app = surveyDashboard(), {
    
    # If code inside this block is reached means the server initialized successfully.
    expect_true(TRUE) 
  })
})


test_that("surveyDashboard server logic initializes correctly with data", {
  
  testthat::skip_if_not_installed("mapview")
  testthat::skip_if_not_installed("plotly")
  testthat::skip_if_not_installed("corrplot")
  testthat::skip_if_not_installed("unmarked")
  testthat::skip_if_not_installed("ubms")
  
  # testServer runs the server function in the current R process,
  # making its internal state (like reactives) accessible for testing.
  
  testServer(
    
    app = surveyDashboard(
      CTtable = camtraps,
      recordTable = recordTableSample,
      xcol = "utm_x",
      ycol = "utm_y",
      crs = "epsg:32650",      # = UTM50N
      stationCol = "Station",
      setupCol = "Setup_date",
      retrievalCol = "Retrieval_date",
      CTdateFormat = "dmy"
    ),
    {
      # Inside this block, you can access server-side objects like
      # input, output, session, and reactive expressions.
      
      
      expect_equal(num_stations(), 3)
    }
  )
})

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/getSpeciesImages.R
\name{getSpeciesImages}
\alias{getSpeciesImages}
\title{Collect all images of a species}
\usage{
getSpeciesImages(
  species,
  recordTable,
  speciesCol = "Species",
  stationCol = "Station",
  inDir,
  outDir,
  createStationSubfolders = FALSE,
  IDfrom,
  metadataSpeciesTag,
  metadataHierarchyDelimitor = "|"
)
}
\arguments{
\item{species}{character. Species whose images are to be fetched}

\item{recordTable}{data frame. A data frame as returned by function
\code{\link{recordTable}}. If you specify this argument, do not specify
\code{inDir}}

\item{speciesCol}{character. Name of the column specifying species ID in
\code{recordTable}. Only required if \code{recordTable} is defined}

\item{stationCol}{character. Name of the column specifying station ID in
\code{recordTable}. Only required if \code{recordTable} is defined}

\item{inDir}{character. Directory containing identified (species level)
camera trap images sorted into station subdirectories (e.g.
inDir/StationA/). If you specify this argument, do not specify
\code{recordTable}.}

\item{outDir}{character. Directory in which to save species images. A
species subdirectory will be created in \code{outDir} automatically.}

\item{createStationSubfolders}{logical. Save images in station directories
within the newly created species directory in \code{outDir}?}

\item{IDfrom}{character. Read species ID from image metadata ("metadata") of
from species directory names ("directory")? Only required if \code{inDir} is
defined.}

\item{metadataSpeciesTag}{character. The species ID tag name in image
metadata (if IDfrom = "metadata"). Only required if \code{inDir} is defined.}

\item{metadataHierarchyDelimitor}{character. The character delimiting
hierarchy levels in image metadata tags in field "HierarchicalSubject".
Either "|" or ":" (if IDfrom = "metadata"). Only required if \code{inDir} is
defined and \code{IDfrom = "metadata"}.}
}
\value{
A \code{data.frame} with old and new directories and file names and
the copy status (\code{copy_ok}; TRUE if copying was successful, FALSE if
not).
}
\description{
This function will fetch all images of a particular species from all camera
trap stations and copies these images to a new location. The images which
are to be copied are found in one of 2 possible ways, 1) by providing an
existing record table (created with \code{\link{recordTable}}) or 2) by
reading species IDs from species directories or from metadata (calling
ExifTool). Earlier in the workflow, i.e., before running this function,
images should have been renamed (with \code{\link{imageRename}}) to give
images unique file names based on station ID and date/time.
}
\details{
The function finds the images to be copied by either consulting a record
table created with \code{\link{recordTable}} or by reading species IDs from
images. The former is considerable faster because ExifTool is not called,
but requires images to be in precisely the location given by the columns
\code{Directory} and \code{FileName} in \code{recordTable}. To use this
feature, provide the function with a record table in argument
\code{recordTable}.

If you'd rather read species IDs from images within the function (to make
sure all file paths are correct), images need to be in the directory
structure required by the package, e.g.

> inDir/Station/Species

or

> inDir/Station/Camera/Species

if using species directories for species IDs, and

> inDir/Station

or

> inDir/Station/Camera

if reading IDs from species metadata tags. In the latter case, only station
directories are needed. In any case, the argument \code{species} must match
species IDs (either the \code{speciesCol} in \code{recordTable}, species
directory names or species metadata tags).

Before running the function, first rename the images using function
\code{\link{imageRename}} to provide unique file names and prevent several
images from having the same name (if generic names like "IMGP0001.jpg" are
used). The function will not copy images if there are duplicate filenames to
prevent overwriting images unintentionally.
}
\examples{

\dontrun{
# define image directory
wd_images_ID <- system.file("pictures/sample_images_species_dir", package = "camtrapR")
wd_images_ID_copy <- file.path(tempdir(), "sample_images_species_dir")

species_to_copy <- "VTA"    # = Viverra tangalunga, Malay Civet

specImagecopy <- getSpeciesImages(species                 = species_to_copy,
                                  inDir                   = wd_images_ID,
                                  outDir                  = wd_images_ID_copy,
                                  createStationSubfolders = FALSE,
                                  IDfrom                  = "directory"
                                  )
}

}
\author{
Juergen Niedballa
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/communityModel.R
\name{fit,commOccu-method}
\alias{fit,commOccu-method}
\title{Fit a community (multi-species) occupancy model}
\usage{
\S4method{fit}{commOccu}(
  object,
  n.iter = 100,
  thin = 1,
  n.burnin = n.iter/2,
  n.adapt = 0,
  chains = 3,
  inits = NULL,
  compile = TRUE,
  WAIC = FALSE,
  quiet = FALSE,
  ...
)
}
\arguments{
\item{object}{\code{commOccu} object}

\item{n.iter}{number of iterations to monitor}

\item{thin}{thinning interval for monitors}

\item{n.burnin}{burnin length. Defaults to half of n.iter.}

\item{n.adapt}{Length of adaptive phase}

\item{chains}{number of MCMC chains to run}

\item{inits}{named list. Initial values to use. If NULL (default), the values from the inits function in \code{object} are used.}

\item{compile}{logical. If Nimble model, compile model with \code{\link[nimble]{compileNimble}} before running model?}

\item{WAIC}{logical. Return WAIC (only Nimble models)}

\item{quiet}{if TRUE messages and progress bar will be suppressed}

\item{...}{additional arguments to pass to \code{\link[nimble]{runMCMC}} (only relevant for Nimble)}
}
\value{
A coda::mcmc.list
}
\description{
Convenience function for fitting community occupancy models (defined in a commOccu object) in JAGS or Nimble.
}
\details{
Models will be fit either in JAGS or Nimble, depending on the decision made in the \code{nimble} argument in \code{\link{communityModel}}.

For Nimble, compilation is strongly recommended for long model runs. Uncompiled models can run extremely slow. Compilation itself can take a while also, and requires that Rtools is available on the system.

This is a convenience function only which hides some of the configuration options. If you require more control over model fitting, you can run all steps individually. See vignette 5 for details.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/imageRename.R
\name{imageRename}
\alias{imageRename}
\title{Copy and rename images based on camera trap station ID and creation date}
\usage{
imageRename(
  inDir,
  outDir,
  hasCameraFolders,
  keepCameraSubfolders,
  createEmptyDirectories = FALSE,
  copyImages = FALSE,
  writecsv = FALSE,
  video
)
}
\arguments{
\item{inDir}{character. Directory containing camera trap images sorted into
station subdirectories (e.g. inDir/StationA/)}

\item{outDir}{character. Directory into which the renamed images will be
copied}

\item{hasCameraFolders}{logical. Do the station directories in \code{inDir}
have camera subdirectories (e.g. "inDir/StationA/Camera1")?}

\item{keepCameraSubfolders}{logical. Should camera directories be preserved
as subdirectories of \code{outDir} (e.g. "outDir/StationA/CameraA1")?}

\item{createEmptyDirectories}{logical. If station or camera directories are
empty, should they be copied nevertheless (causing empty directories in
\code{outDir}, but preserving the whole directory structure)?}

\item{copyImages}{logical. Copy images to \code{outDir}?}

\item{writecsv}{logical. Save a data frame with a summary as a .csv? The csv
will be saved in \code{outDir}.}

\item{video}{list. Contains information on how to handle video data
(optional). See details.}
}
\value{
A \code{data.frame} with original directory and file names, new
directory and file names and an indicator for whether images were copied
successfully.
}
\description{
The function renames and copies raw camera trap images into a new location
where they can be identified. Images are renamed with camera trap station
ID, camera ID (optional), creation date and a numeric identifier for images
taken within one minute of each other at a given station. Station ID and
camera ID are derived from the raw image directory structure. The creation
date is extracted from image metadata using ExifTool.
}
\details{
Setting up the correct raw image directory structure is necessary for
running the function successfully. \code{inDir} is the main directory that
contains camera trap station subdirectories (e.g. inDir/StationA). If one
camera was deployed per station and no camera subdirectories are used within
station directories, \code{hasCameraFolders} can be set to \code{FALSE}. If
more than one camera was deployed at stations, there must be subdirectories
for the individual camera traps within the station directories (e.g.
"inDir/StationA/CameraA1" and "inDir/StationA/CameraA2"). Even if only some
stations had multiple cameras, all station will need camera subdirectories.
The argument \code{hasCameraFolders} must be \code{TRUE}. Within the camera
subdirectories, the directory structure is irrelevant.

Renaming of images follows the following pattern: If \code{hasCameraFolders}
is TRUE, it is: "StationID__CameraID__Date__Time(Number).JPG", e.g.
"StationA__CameraA1__2015-01-31__18-59-59(1).JPG". If
\code{hasCameraFolders} is FALSE, it is:
"StationID__Date__Time(Number).JPG", e.g.
"StationA__2015-01-31__18-59-59(1).JPG".

The purpose of the number in parentheses is to prevent assigning identical
file names to images taken at the same station (and camera) in the same
second, as can happen if cameras take sequences of images. It is a
consecutive number given to all images taken at the same station by the same
camera within one minute. The double underscore "__" in the image file names
is for splitting and extracting information from file names in other
functions (e.g. for retrieving camera IDs in \code{\link{recordTable}} if
camera subdirectories are not preserved (\code{keepCameraSubfolders =
FALSE})).

The function finds all JPEG images (optionally, also videos) and extracts 
the image timestamp from the image metadata using ExifTool (digiKam database 
for videos) and copies the images with new file names
into \code{outDir}, where it will set up a directory structure based on the
station IDs and, if required by \code{keepCameraSubfolders = TRUE}, camera
IDs (e.g. outDir/StationA/ or outDir/StationA/CameraA1).

\code{copyImages} can be set to FALSE to simulate the renaming and check the
file names of the renamed images without copying. If you are handling large
number of images (>e.g., 100,000), the function may take some time to run.

Argument \code{video} is a named list 4 items 
(\code{file_formats}, \code{dateTimeTag}, (\code{db_directory}, 
\code{db_filename}). Video date/time is read from video metadata stored in
the digiKam database. Hence, \code{inDir} must be in your digiKam database.

The items of argument \code{video} are:

\tabular{ll}{ \code{file_formats} \tab The video formats to extract (include
"jpg" if you want .JPG image metadata) \cr \code{dateTimeTag} \tab the
metadata tag to extract date/time from (use \code{\link{exifTagNames}} to
find out which tag is suitable) \cr \code{db_directory} \tab The directory
containing digiKam database  \cr
\code{db_filename} \tab The digiKam database file in \code{db_directory} \cr }

See the examples in \code{\link{recordTable}} for for how to specify the 
argument \code{video}.
}
\examples{



  \dontrun{
  
### "trial" run. create a table with file names after renaming, but don't copy images.

# first, find sample image directory in package directory:
wd_images_raw <- system.file("pictures/raw_images", package = "camtrapR")

# because copyImages = FALSE, outDir does not need to be defined
renaming.table <- imageRename(inDir               = wd_images_raw,     
                              hasCameraFolders = FALSE,
                              copyImages          = FALSE,
                              writecsv            = FALSE
  )


  
### a real example in which images are copied and renamed 

  # define raw image location
wd_images_raw <- system.file("pictures/raw_images", package = "camtrapR") 

  # define destination for renamed images
wd_images_raw_renamed <- file.path(tempdir(), "raw_images_renamed")       


  # now we have to define outDir because copyImages = TRUE
renaming.table2 <- imageRename(inDir               = wd_images_raw,
                               outDir              = wd_images_raw_renamed,       
                               hasCameraFolders    = FALSE,
                               copyImages          = TRUE,
                               writecsv            = FALSE
  )
  
  # show output files
  list.files(wd_images_raw_renamed, recursive = TRUE)
  
  # output table
  renaming.table2
  
  }

}
\references{
Phil Harvey's ExifTool \url{https://exiftool.org/}
}
\author{
Juergen Niedballa
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/exifTagNames.R
\name{exifTagNames}
\alias{exifTagNames}
\title{Show Exif metadata of JPEG images or other image or video formats}
\usage{
exifTagNames(
  inDir,
  whichSubDir = 1,
  fileName,
  returnMetadata = "DEPRECATED",
  returnTagGroup = "DEPRECATED"
)
}
\arguments{
\item{inDir}{character. Directory containing camera trap images sorted into
station subdirectories (e.g. inDir/StationA/)}

\item{whichSubDir}{integer or character. Either number or name of
subdirectory of \code{inDir} in which to look for an image}

\item{fileName}{character. A filename, either the file name of an image in
\code{inDir} or a full path with file name (in which case \code{inDir} is
not needed)}

\item{returnMetadata}{deprecated and ignored}

\item{returnTagGroup}{deprecated and ignored}
}
\value{
A data frame containing three columns: metadata tag group, tag name,
and values.
}
\description{
The function will return metadata values, metadata tag names and group names
of Exif metadata of JPEG images or other formats.
}
\details{
Many digital cameras record information such as ambient temperature or moon
phase under maker-specific tag names in Exif metadata of JPEG images. In
addition, many technical information are stored in Exif metadata. In order
to extract those information from images and add them to the record tables
created by the functions \code{\link{recordTable}} and
\code{\link{recordTableIndividual}}, the tag names must be known so they can
be passed to these functions via the \code{additionalMetadataTags} argument.

By default the function returns both metadata tag names and the metadata
group they belong to (via argument \code{returnTagGroup}). This is helpful
to unambiguously address specific metadata tags, because different groups
can contain tags of identical names, which may cause problems executing the
functions \code{\link{recordTable}} and \code{\link{recordTableIndividual}}.
The format is "GROUP:tag", e.g. "EXIF:Flash".
}
\examples{


\dontrun{

wd_images_ID <- system.file("pictures/sample_images_species_dir", package = "camtrapR")

# specify directory, camtrapR will automatically take first image from first subdirectory
exifTagNames(inDir          = wd_images_ID)

# specify subdirectory by name, camtrapR will use first image
exifTagNames(inDir          = wd_images_ID,
             whichSubDir    = "StationA")

# specifying fileName only (line break due to R package policy)
exifTagNames(fileName       = file.path(wd_images_ID, "StationC", "TRA", 
                                        "StationC__2009-05-02__00-10-00(1).JPG"))
                                        
# specify inDir and fileName
exifTagNames(inDir          = wd_images_ID,
             fileName       = file.path("StationC", "TRA", "StationC__2009-05-02__00-10-00(1).JPG"))
             
# it also works this way
exifTagNames(inDir          = file.path(wd_images_ID, "StationC", "TRA"),
             fileName       = "StationC__2009-05-02__00-10-00(1).JPG")
             

# with tagged sample images
wd_images_ID_tagged <- system.file("pictures/sample_images_indiv_tag", package = "camtrapR")
exifTagNames(inDir          = wd_images_ID_tagged)

}

}
\references{
Phil Harvey's ExifTool \url{https://exiftool.org/} \cr
}
\seealso{
\code{\link{recordTable}}
}
\author{
Juergen Niedballa
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/predict.R
\name{predict,commOccu-method}
\alias{predict,commOccu-method}
\alias{predict}
\title{Predictions from community occupancy models}
\usage{
\S4method{predict}{commOccu}(
  object,
  mcmc.list,
  type,
  draws = 1000,
  level = 0.95,
  interval = c("none", "confidence"),
  x = NULL,
  aoi = NULL,
  speciesSubset,
  batch = FALSE,
  seed = NULL
)
}
\arguments{
\item{object}{\code{commOccu} object}

\item{mcmc.list}{mcmc.list. Output of \code{fit} called on a \code{commOccu} object}

\item{type}{character. "psi" for species occupancy estimates, "richness" for species richness estimates, "pao" for percentage of area occupied (by species), "psi_array" for raw occupancy probabilities in an array. For Royle-Nichols models, "abundance" for species abundance, or "lambda_array" for raw species abundance estimates in an array. "p_array" for raw detection probabilities in an array.}

\item{draws}{Number of draws from the posterior to use when generating the plots. If fewer than draws are available, they are all used}

\item{level}{Probability mass to include in the uncertainty interval}

\item{interval}{Type of interval calculation. Can be "none" or "confidence" (can be abbreviated). Calculation can be slow for type = "psi" with many cells and posterior samples.}

\item{x}{SpatRaster, data.frame or NULL. Must be scaled with same parameters as site covariates used in model, and have same names. If NULL, use site covariate data frame from model input (\code{commOccu} object in parameter \code{object})}

\item{aoi}{SpatRaster with same dimensions as x (if x is a SpatRaster), indicating the area of interest (all cells with values are AOI, all NA cells are ignored). If NULL, predictions are made for all cells.}

\item{speciesSubset}{species to include in richness estimates. Can be index number or species names.}

\item{batch}{logical or numeric. If FALSE, all raster cells / data frame rows will be processed at once (can be memory intensive). If TRUE, computation is conducted in batches of 1000. If numeric, it is the desired batch size.}

\item{seed}{numeric. Seed to use in \code{set.seed} for reproducible results (ensures that \code{draws} are identical).}
}
\value{
A SpatRaster or data.frame, depending on \code{x} (for type = "psi", "abundance", "richness". If type = "pao", a list. If type = "psi_array" or "lambda_array", a 3D-array [site, species, draw]. If type = "p_array", a 4D-array [site, species, draw, occasion].
}
\description{
Create (spatial) predictions of species occupancy and species richness from community occupancy models and spatial rasters or covariate data frames.
}
\details{
Processing can be very memory-intensive. If memory is insufficient, use the  \code{batch} parameter. This can enable processing for higher numbers of \code{draws} or very large rasters / data frames.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/spatialDetectionHistory.R
\name{spatialDetectionHistory}
\alias{spatialDetectionHistory}
\title{Generate a \code{capthist} object for spatial capture-recapture analyses
from camera-trapping data}
\usage{
spatialDetectionHistory(
  recordTableIndividual,
  species,
  camOp,
  CTtable,
  output = c("binary", "count"),
  stationCol = "Station",
  speciesCol = "Species",
  sessionCol,
  Xcol,
  Ycol,
  stationCovariateCols,
  individualCol,
  individualCovariateCols,
  recordDateTimeCol = "DateTimeOriginal",
  recordDateTimeFormat = "ymd HMS",
  occasionLength,
  minActiveDaysPerOccasion,
  occasionStartTime = "deprecated",
  maxNumberDays,
  day1,
  buffer,
  includeEffort = TRUE,
  scaleEffort = FALSE,
  binaryEffort = FALSE,
  timeZone,
  makeRMarkInput
)
}
\arguments{
\item{recordTableIndividual}{data.frame. the record table with individual
IDs created by \code{\link{recordTableIndividual}}}

\item{species}{character. the species for which to compute the detection
history}

\item{camOp}{The camera operability matrix as created by
\code{\link{cameraOperation}}}

\item{CTtable}{data.frame. contains station IDs and coordinates. Same as
used in \code{\link{cameraOperation}}.}

\item{output}{character. Return individual counts ("count") or binary
observations ("binary")?}

\item{stationCol}{character. name of the column specifying Station ID in
\code{recordTableIndividual} and \code{CTtable}}

\item{speciesCol}{character. name of the column specifying species in
\code{recordTableIndividual}}

\item{sessionCol}{character. name of the column specifying session IDs,
either in \code{recordTableIndividual} or in \code{CTtable}. See 'Details'
for more information. Session ID values must be a sequence of integer
numbers beginning with 1 (i.e., 1,2,3,...).}

\item{Xcol}{character. name of the column specifying x coordinates in
\code{CTtable}}

\item{Ycol}{character. name of the column specifying y coordinates in
\code{CTtable}}

\item{stationCovariateCols}{character. name of the column(s) specifying
station-level covariates in \code{CTtable}}

\item{individualCol}{character. name of the column specifying individual IDs
in \code{recordTableIndividual}}

\item{individualCovariateCols}{character. name of the column(s) specifying
individual covariates in \code{recordTableIndividual}}

\item{recordDateTimeCol}{character. name of the column specifying date and
time in \code{recordTableIndividual}}

\item{recordDateTimeFormat}{format of column \code{recordDateTimeCol} in
\code{recordTableIndividual}}

\item{occasionLength}{integer. occasion length in days}

\item{minActiveDaysPerOccasion}{integer. minimum number of active trap days
for occasions to be included (optional)}

\item{occasionStartTime}{(DEPRECATED) integer. time of day (the full hour)
at which to begin occasions. Please use argument \code{occasionStartTime} in
\code{\link{cameraOperation}} instead.}

\item{maxNumberDays}{integer. maximum number of trap days per station
(optional)}

\item{day1}{character. When should occasions begin: station setup date
("station"), first day of survey ("survey"), a specific date (e.g.
"2015-12-31")?}

\item{buffer}{integer. Makes the first occasion begin a number of days after
station setup. (optional)}

\item{includeEffort}{logical. Include trapping effort (number of active
camera trap days per station and occasion) as usage in
\code{\link[secr]{capthist}} object?}

\item{scaleEffort}{logical. scale and center effort matrix to mean = 0 and
sd = 1? Currently not used. Must be FALSE.}

\item{binaryEffort}{logical. Should effort be binary (1 if >1 active day per
occasion, 0 otherwise)?}

\item{timeZone}{character. Must be a value returned by
\code{\link[base:timezones]{OlsonNames}}}

\item{makeRMarkInput}{logical. If \code{FALSE}, output will be a data frame
for RMark. If \code{FALSE} or not specified, a secr
\code{\link[secr]{capthist}} object}
}
\value{
Output depends on argument \code{makeRMarkInput}:

\item{list("makeRMarkInput = FALSE")}{A \code{\link[secr]{capthist}} object}
\item{list("makeRMarkInput = TRUE")}{A data frame for use in RMark}
}
\description{
This function generates spatial detection histories of individuals of a
species for spatial capture-recapture analyses with package
\code{\link[secr:secr-package]{secr}}. Data are stored in a
\code{\link[secr]{capthist}} object. The \code{\link[secr]{capthist}} object
contains detection histories, camera-trap station location and possibly
individual and station-level covariates. Detection histories can have
adjustable occasion length and occasion start time (as in the function
\code{\link{detectionHistory}}).
}
\details{
The function creates a \code{\link[secr]{capthist}} object by combining
three different objects: 1) a record table of identified individuals of a
species, 2) a camera trap station table with station coordinates and 3) a
camera operation matrix computed with \code{\link{cameraOperation}}. The
record table must contain a column with individual IDs and optionally
individual covariates. The camera trap station table must contain station
coordinates and optionally station-level covariates. The camera operation
matrix provides the dates stations were active or not and the number of
active stations.

\code{day1} defines if each stations detection history will begin on that
station's setup day (\code{day1 = "station"}) or if all station's detection
histories have a common origin (the day the first station was set up if
\code{day1 = "survey"} or a fixed date if, e.g. \code{day1 = "2015-12-31"}).

\code{includeEffort} controls whether an effort matrix is computed or not.
If TRUE, effort will be used for object \code{\link[secr]{usage}}
information in a \code{\link[secr]{traps}}. \code{binaryEffort} makes the
effort information binary. \code{scaleEffort} is currently not used and must
be set to FALSE. The reason is that \code{\link[secr]{usage}} can only be
either binary, or nonnegative real values, whereas scaling effort would
return negative values.

The number of days that are aggregated is controlled by
\code{occasionLength}. \code{occasionStartTime} will be removed from the
function. It has moved to \code{\link{cameraOperation}}, to ensure daily
effort is computed correctly and takes the occasion start time into
account.%\code{occasionStartTime} can be used to make occasions begin
another hour than midnight (the default). This may be relevant for nocturnal
animals, in which 1 whole night would be considered an occasion.

Output can be returned as individual counts per occasion (\code{output =
"count"}) or as binary observation (\code{output = "binary"}).

Argument \code{sessionCol} can be used to a create multi-session
\code{\link[secr]{capthist}} object. There are two different ways in which
the argument is interpreted. It depends on whether a column with the name
you specify in argument \code{sessionCol} exists in
\code{recordTableIndividual} or in \code{CTtable}. If \code{sessionCol} is
found in \code{recordTableIndividual}, the records will be assigned to the
specified sessions, and it will be assumed that all camera trap station were
used in all sessions. Alternatively, if \code{sessionCol} is found
in\code{CTtable}, it will be assumed that only a subset of stations was used
in each session, and the records will be assigned automatically (using the
station IDs to identify which session they belong into). In both cases,
session information must be provided as a sequence of integer numbers
beginnign with 1, i.e., you provide the session number directly in
\code{sessionCol}. See \code{\link[secr]{session}} for more information
about sessions in \pkg{secr}.

\code{\link[secr]{capthist}} objects (as created by
\code{\link{spatialDetectionHistory}} for spatial capture-recapture
analyses) expect the units of coordinates (\code{Xcol} and \code{col} in
\code{CTtable}) to be meters. Therefore, please use a suitable coordinate
system (e.g. UTM).

\code{recordDateTimeFormat} defaults to the "YYYY-MM-DD HH:MM:SS"
convention, e.g. "2014-09-30 22:59:59". \code{recordDateTimeFormat} can be
interpreted either by base-R via \code{\link[base]{strptime}} or in
\pkg{lubridate} via \code{\link[lubridate]{parse_date_time}} (argument
"orders"). \pkg{lubridate} will be used if there are no "\%" characters in
\code{recordDateTimeFormat}.

For "YYYY-MM-DD HH:MM:SS", \code{recordDateTimeFormat} would be either
"\%Y-\%m-\%d \%H:\%M:\%S" or "ymd HMS". For details on how to specify date
and time formats in R see \code{\link[base]{strptime}} or
\code{\link[lubridate]{parse_date_time}}.
}
\section{Warning }{
 Please note the section about defining argument
\code{timeZone} in the vignette on data extraction (accessible via
\code{vignette("DataExtraction")} or online
(\url{https://cran.r-project.org/package=camtrapR/vignettes/camtrapr3.pdf})).
}

\examples{


data(recordTableIndividualSample)
data(camtraps)

# create camera operation matrix (with problems/malfunction)
camop_problem <- cameraOperation(CTtable      = camtraps,
                                 stationCol   = "Station",
                                 setupCol     = "Setup_date",
                                 retrievalCol = "Retrieval_date",
                                 writecsv     = FALSE,
                                 hasProblems  = TRUE,
                                 dateFormat   = "dmy"
)

sdh <- spatialDetectionHistory(recordTableIndividual = recordTableIndividualSample,
                               species               = "LeopardCat",
                               camOp                 = camop_problem,
                               CTtable               = camtraps,
                               output                = "binary",
                               stationCol            = "Station",
                               speciesCol            = "Species",
                               Xcol                  = "utm_x",
                               Ycol                  = "utm_y",
                               individualCol         = "Individual",
                               recordDateTimeCol     = "DateTimeOriginal",
                               recordDateTimeFormat  = "ymd HMS",
                               occasionLength        = 10,
                               day1                  = "survey",
                               includeEffort         = TRUE,
                               timeZone              = "Asia/Kuala_Lumpur"
  )

# missing space in species = "LeopardCat" was introduced by recordTableIndividual
# (because of CRAN package policies.
# In your data you can have spaces in your directory names)

  summary(sdh)
  plot(sdh, tracks = TRUE)

  ## multi-season capthist object
  # see vignette "3. Extracting Data from Camera Trapping Images, creating occupancy & secr input"
  
  data(camtrapsMultiSeason)
  camtrapsMultiSeason$session[camtrapsMultiSeason$session == 2009] <- 1
  camtrapsMultiSeason$session[camtrapsMultiSeason$session == 2010] <- 2

  data(recordTableIndividualSampleMultiSeason)

  # create camera operation matrix (with problems/malfunction)
  camop_session <- cameraOperation(CTtable         = camtrapsMultiSeason,
                                      stationCol   = "Station",
                                      setupCol     = "Setup_date",
                                      sessionCol   = "session",
                                      retrievalCol = "Retrieval_date",
                                      hasProblems  = TRUE,
                                      dateFormat   = "dmy"
  )

sdh_multi <- spatialDetectionHistory(recordTableIndividual = recordTableIndividualSampleMultiSeason,
                               species               = "LeopardCat",
                               output                = "binary",
                               camOp                 = camop_session,
                               CTtable               = camtrapsMultiSeason,
                               stationCol            = "Station",
                               speciesCol            = "Species",
                               sessionCol            = "session",
                               Xcol                  = "utm_x",
                               Ycol                  = "utm_y",
                               individualCol         = "Individual",
                               recordDateTimeCol     = "DateTimeOriginal",
                               recordDateTimeFormat  = "ymd HMS",
                               occasionLength        = 10,
                               day1                  = "survey",
                               includeEffort         = TRUE,
                               timeZone              = "Asia/Kuala_Lumpur",
                               stationCovariateCols  = "utm_y",         # example
                               individualCovariateCols = "Individual"   # example
  )

  summary(sdh_multi)
  plot(sdh_multi, tracks = TRUE)

}
\seealso{
\pkg{secr} \pkg{RMark}
}
\author{
Juergen Niedballa
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/fixDateTimeOriginal.R
\name{fixDateTimeOriginal}
\alias{fixDateTimeOriginal}
\title{Fix DateTimeOriginal Exif metadata tag in Reconyx Hyperfire cameras}
\usage{
fixDateTimeOriginal(inDir, recursive = TRUE)
}
\arguments{
\item{inDir}{character. Name of the directory containing images to be fixed}

\item{recursive}{logical. Recursively find images in subdirectories of
\code{inDir}?}
}
\value{
Returns invisibly the messages returned by the Exiftool call
(warnings etc.).
}
\description{
Some camera models don't store the date/time information in the standard
Exif metadata tag. Consequently, camtrapR cannot find that information. This
function uses Exiftool to update the DateTimeOriginal metadata tag in all
images within a directory to make them readable with camtrapR (and other
software).
}
\details{
Some Reconyx Hyperfire cameras (e.g. HC500) are known to show this problem.
}
\section{Warning }{
 Please make a backup of your images before running this
function.
}

\examples{

\dontrun{
# a hypothetical example

wd_images_hyperfire <- "C:/Some/Directory"

fixDateTimeOriginal(inDir     = wd_images_hyperfire,
                    recursive = TRUE)
}

}
\references{
This function uses the code from: \cr Tobler, Mathias (2015).
Camera Base Version 1.7 User Guide
}
\author{
Juergen Niedballa
}

\docType{data}
\name{camtrapsMultiSeason}
\alias{camtrapsMultiSeason}
\title{Sample multi-season camera trap station information}
\format{A data frame with 7 rows and 8 variables}
\usage{
data(camtrapsMultiSeason)
}
\description{
Example multi-season camera trap station information table
}
\details{
This is a general example of how information about camera trap stations from multiple seasons are arranged in camtrapR. It contains setup and retrieval dates, coordinates and a season identifier. If more than 1 camera was set up at a station (e.g. 2 cameras facing each other), a camera ID column must be added, with camera-specific information instead of station-specific information.
If cameras malfunctioned repeatedly, additional pairs of problem columns can be added, e.g. "Problem2_from" and "Problem2_to" etc..

Note that season 2010 has an additional station (StationD). This is to simulate a situation where a station was not set up during an entire season.

The variables are as follows:

\tabular{ll}{
\code{Station} \tab Camera trap station ID \cr
\code{utm_y} \tab y coordinate of station (northing) \cr
\code{utm_x} \tab x coordinate of station (easting) \cr
\code{Setup_date} \tab camera trap setup date \cr
\code{Retrieval_date} \tab camera trap retrieval date \cr
\code{Problem1_from} \tab first day of camera malfunction \cr
\code{Problem1_to} \tab  last day of camera malfunction \cr
\code{session} \tab  Identified for trapping session / season \cr
}

}

\note{
The coordinates can be in the units of any coordinate system. UTM was chosen as an example, but it could be latlong or anything else, too. \code{\link[secr]{capthist}} objects (as created by \code{\link{spatialDetectionHistory}} for spatial capture-recapture analyses) expect the unit to be meters. \code{\link[secr]{capthist}} alse require session information as integer numbers starting with 1.

"Season" and "session" are used synonymously here. \pkg{secr} nomenclature is "session", in \pkg{unmarked} it is "season". 
}

\examples{
# data were created with the following code:
data(camtraps)

camtraps_season2 <- camtraps

# change 2009 to 2010
camtraps_season2[, "Setup_date"] <- gsub("2009", "2010", camtraps_season2[,
    "Setup_date"])
camtraps_season2[, "Retrieval_date"] <- gsub("2009", "2010", camtraps_season2[,
    "Retrieval_date"])
camtraps_season2[, "Problem1_from"] <- gsub("2009", "2010", camtraps_season2[,
    "Problem1_from"])
camtraps_season2[, "Problem1_to"] <- gsub("2009", "2010", camtraps_season2[,
    "Problem1_to"])

# add an extra station with different dates in session 2010
camtraps_season2 <- rbind(camtraps_season2, NA)
camtraps_season2$Station[4] <- "StationD"
camtraps_season2$utm_y[4] <- 607050
camtraps_season2$utm_x[4] <- 525000
camtraps_season2$Setup_date[4] <- "04/04/2010"
camtraps_season2$Retrieval_date[4] <- "17/06/2010"
camtraps_season2$Problem1_from[4] <- "20/05/2010"
camtraps_season2$Problem1_to[4] <- "30/05/2010"

# add season column
camtraps$session <- 2009
camtraps_season2$session <- 2010

# combine the tables for 2 seasons
camtrapsMultiSeason <- rbind(camtraps, camtraps_season2)
}
\keyword{datasets}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/timeShiftImages.R
\name{timeShiftImages}
\alias{timeShiftImages}
\title{Apply time shifts to JPEG image metadata}
\usage{
timeShiftImages(
  inDir,
  hasCameraFolders,
  timeShiftTable,
  stationCol,
  cameraCol,
  timeShiftColumn,
  timeShiftSignColumn,
  undo = FALSE,
  ignoreMinorErrors = FALSE
)
}
\arguments{
\item{inDir}{character. Name of directory containing station directories
with images}

\item{hasCameraFolders}{logical. Do the station directories in \code{inDir}
have camera subdirectories (e.g. "inDir/StationA/Camera1")?}

\item{timeShiftTable}{data.frame containing information about
station-/camera-specific time shifts.}

\item{stationCol}{character. name of the column specifying Station ID in
\code{timeShiftTable}}

\item{cameraCol}{character. name of the column specifying Camera ID in
\code{timeShiftTable} (optional)}

\item{timeShiftColumn}{character. The name of the column containing time
shift values in \code{timeShiftTable}}

\item{timeShiftSignColumn}{character. The name of the column with the
direction of time shifts in \code{timeShiftTable}. Can only be "-" or "+".}

\item{undo}{logical. Undo changes and restore the original images? Please be
careful, this deletes any edited images if \code{TRUE}}

\item{ignoreMinorErrors}{logical. Ignore minor errors that would cause the
function to fail (set TRUE for images with bad MakerNotes, observed in
Panthera V4 cameras)}
}
\value{
A \code{data.frame} containing the information about the processed
directories and the number of images.
}
\description{
Change the values of digital timestamps in image metadata using ExifTool. If
date/time of images were set incorrectly, they can be corrected easily in
batch mode for further analyses. Please, always make a backup of your data
before using this function to avoid data loss or damage. This is because
ExifTool will make a copy of your images and applies the time shifts to the
copies. The file extension of the original images (.JPG) will be renamed to
".JPG_original".
}
\details{
\code{timeShiftTable} is a data frame with columns for station ID, camera ID
(optional), time shift value and direction of time shift (for an example see
\code{\link{timeShiftTable}}). Images in \code{inDir} must be sorted into
station directories. If \code{hasCameraFolders = TRUE}, the function expects
camera subdirectories in the station directories and will only apply time
shifts to the camera subdirectories specified by \code{CameraCol} in
\code{timeShiftTable}. If \code{hasCameraFolders = FALSE}, shifts will be
applied to the whole station directory (including potential subdirectories).

The values of \code{timeShiftColumn} must adhere to the following pattern:
"YYYY:mm:dd HH:MM:SS" ("year:month:day hour:minute:second"). Examples:
"1:0:0 0:0:0" is a shift of exactly 1 year and "0:0:0 12:10:01" 12 hours and
10 minutes and 1 second. Note that stating "00" may cause problems, so use
"0" instead if an entry is zero.

\code{timeShiftSignColumn} signifies the direction of the time shift. "+"
moves image dates into the future (i.e. the image date lagged behind the
actual date) and "-" moves image dates back (if the image dates were ahead
of actual time).

ExifTool stores the original images as \code{.JPG_original} files in the
original file location. By setting \code{undo = TRUE}, any JPG files in the
directories specified by \code{timeShiftTable} will be deleted and the
original JPEGs will be restored from the JPG_original files. Please make a
backup before using \code{undo}.

Years can have 365 or 366 days, and months 28 to 31 days. Here is how the
function handles these (from the exiftool help page): "The ability to shift
dates by Y years, M months, etc, conflicts with the design goal of
maintaining a constant shift for all time values when applying a batch
shift. This is because shifting by 1 month can be equivalent to anything
from 28 to 31 days, and 1 year can be 365 or 366 days, depending on the
starting date. The inconsistency is handled by shifting the first tag found
with the actual specified shift, then calculating the equivalent time
difference in seconds for this shift and applying this difference to
subsequent tags in a batch conversion."

\code{ignoreMinorErrors} is useful if image timestamps are not updated
correctly (entries in column "n_images" of the output are "... files weren't
updated due to errors"). This can be caused by bad MakerNotes and so far was
only observed in Panthera V4 and V6 cameras. In that case, set
\code{ignoreMinorErrors} to \code{TRUE}. This will add the "-m" option to
the Exiftool call, thereby ignoring minor errors and warnings and applying
the time shift nevertheless.
}
\examples{


\dontrun{

# copy sample images to temporary directory (so we don't mess around in the package directory)
wd_images_ID <- system.file("pictures/sample_images_species_dir", package = "camtrapR")
file.copy(from = wd_images_ID, to = tempdir(), recursive = TRUE)
wd_images_ID_copy <- file.path(tempdir(), "sample_images_species_dir")

data(timeShiftTable)


timeshift_run <- timeShiftImages(inDir                = wd_images_ID_copy,
                                 timeShiftTable       = timeShiftTable,
                                 stationCol           = "Station",
                                 hasCameraFolders     = FALSE,
                                 timeShiftColumn      = "timeshift",
                                 timeShiftSignColumn  = "sign",
                                 undo                 = FALSE
)


timeshift_undo <- timeShiftImages(inDir               = wd_images_ID_copy,
                                  timeShiftTable      = timeShiftTable,
                                  stationCol          = "Station",
                                  hasCameraFolders    = FALSE,
                                  timeShiftColumn     = "timeshift",
                                  timeShiftSignColumn = "sign",
                                  undo                = TRUE
)
}

}
\references{
\url{https://exiftool.org/#shift}
}
\author{
Juergen Niedballa
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/checkSpeciesNames.R
\name{checkSpeciesNames}
\alias{checkSpeciesNames}
\title{Check species names against the ITIS taxonomic database}
\usage{
checkSpeciesNames(speciesNames, searchtype, accepted = TRUE, ask = TRUE)
}
\arguments{
\item{speciesNames}{character. Vector of species names to check. Either
common names or scientific names.}

\item{searchtype}{character. Type of names specified in \code{speciesNames}.
One of 'scientific' or 'common'.}

\item{accepted}{logical. Return only accepted valid names? If TRUE, invalid
names are returned as NA. Set to FALSE to return both accepted and
unaccepted names.}

\item{ask}{logical. Should the function be run in interactive mode? If TRUE
and more than one TSN is found for a species, the user is asked to choose
one. If FALSE, NA is returned for multiple matches.}
}
\value{
A \code{data.frame} with the names supplied by the user, matching
common and scientific names, taxon author and year, taxonomic rank, status,
TSNs (taxonomic serial numbers) and ITIS urls.
}
\description{
The function checks species names (common or scientific names) provided by
the user with the ITIS taxonomic database (\url{https://www.itis.gov/}) via
functions from the package \pkg{taxize}. It returns both common and
scientific names, the taxon authors, taxon rank name and status, the TSN
(taxonomic serial numbers) and ITIS urls.
}
\details{
Arguments \code{searchtype}, \code{accepted} and \code{ask} are passed on to
\code{\link[taxize]{get_tsn}}.
}
\examples{


\dontrun{


species_common <- c("Leopard Cat", "moonrat")

# ask = TRUE. Multiple matches for leopard cat will cause menu to pop up asking user input.

species.names.check1 <- checkSpeciesNames(speciesNames = species_common,
                                          searchtype   = "common",
                                          accepted     = TRUE,
                                          ask          = TRUE)
2   # we choose entry 2
species.names.check1


# ask = FALSE. Multiple matches for leopard cat will cause NA.

species.names.check2 <- checkSpeciesNames(speciesNames = species_common,
                                          searchtype   = "common",
                                          accepted     = TRUE,
                                          ask          = FALSE)
species.names.check2



# search for scientific names

species_scientific <- c("Tragulus", "Prionailurus bengalensis")

species.names.check3 <- checkSpeciesNames(speciesNames = species_scientific,
                                          searchtype   = "scientific",
                                          accepted     = TRUE,
                                          ask          = TRUE)
species.names.check3
}


}
\references{
\url{https://www.itis.gov/}
}
\author{
Juergen Niedballa
}

\docType{data}
\name{camtraps}
\alias{camtraps}
\title{Sample camera trap station information}
\format{A data frame with 3 rows and 7 variables}
\usage{
data(camtraps)
}
\description{
Example camera trap station information table
}
\details{
This is a general example of how information about camera trap stations are arranged in camtrapR. It contains setup and retrieval dates and coordinates. If more than 1 camera was set up at a station (e.g. 2 cameras facing each other), a camera ID column must be added, with camera-specific information instead of station-specific information.
If cameras malfunctioned repeatedly, additional pairs of problem columns can be added, e.g. "Problem2_from" and "Problem2_to" etc..

The variables are as follows:

 
\tabular{ll}{
\code{Station} \tab Camera trap station ID \cr
\code{utm_y} \tab y coordinate of station (northing) \cr
\code{utm_x} \tab x coordinate of station (easting) \cr
\code{Setup_date} \tab camera trap setup date \cr
\code{Retrieval_date} \tab camera trap retrieval date \cr
\code{Problem1_from} \tab first day of camera malfunction \cr
\code{Problem1_to} \tab  last day of camera malfunction \cr
}
 
}

\note{
The coordinates can be in the units of any coordinate system. UTM was chosen as an example, but it could be latlong or anything else, too. \code{\link[secr]{capthist}} objects (as created by \code{\link{spatialDetectionHistory}} for spatial capture-recapture analyses) expect the unit to be meters.
}
\keyword{datasets}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/readcamtrapDP.R
\name{readcamtrapDP}
\alias{readcamtrapDP}
\title{Convert Camtrap DP format data to camtrapR format}
\usage{
readcamtrapDP(
  deployments_file = "deployments.csv",
  media_file = "media.csv",
  observations_file = "observations.csv",
  datapackage_file = "datapackage.json",
  min_gap_hours = 24,
  removeNA = FALSE,
  removeEmpty = FALSE,
  remove_bbox = TRUE,
  add_file_path = FALSE,
  filter_observations = NULL
)
}
\arguments{
\item{deployments_file}{character. Path to deployments.csv file}

\item{media_file}{character. Path to media.csv file}

\item{observations_file}{character. Path to observations.csv file}

\item{datapackage_file}{character. Path to datapackage.json file}

\item{min_gap_hours}{numeric. Minimum gap in hours to consider as a camera interruption (default: 24)}

\item{removeNA}{logical. Whether to remove columns with only NA values}

\item{removeEmpty}{logical. Whether to remove columns with only empty values}

\item{remove_bbox}{logical. Whether to remove bounding box columns in the observation table}

\item{add_file_path}{logical. Whether to add file path from media table to the observation table}

\item{filter_observations}{Controls which observation types to include. \code{NULL} or \code{FALSE} keeps all observations (default), \code{TRUE} keeps only animal observations, or provide a character vector of specific observation types to include.}
}
\value{
List containing three elements:
\itemize{
  \item CTtable: Data frame with camera trap deployment information in camtrapR format
  \item recordTable: Data frame with species records in camtrapR format
  \item metadata: List containing project metadata extracted from datapackage.json
}
}
\description{
This function converts camera trap data from the Camtrap DP standard format to the format
used by camtrapR. Camtrap DP is an open standard for the FAIR exchange and archiving of 
camera trap data, structured in a simple yet flexible data model consisting of three tables: 
Deployments, Media, and Observations.
}
\details{
Camtrap DP is a standardized format developed under the umbrella of the Biodiversity Information
Standards (TDWG) that facilitates data exchange between different camera trap platforms and systems.
It supports a wide range of camera deployment designs, classification techniques, and analytical use cases.

While the 'camtrapdp' package (available on CRAN) provides general functionality for reading, 
manipulating, and transforming Camtrap DP data, this function specifically converts Camtrap DP 
data directly into the format required by camtrapR, producing the CTtable and recordTable objects 
that camtrapR functions expect.

This function reads the three primary tables from a Camtrap DP dataset:
\itemize{
  \item Deployments: Information about camera trap placements
  \item Observations: Classifications derived from the media files
  \item Media: Information about the media files (images/videos) recorded
}

The Media table is only read if the \code{add_file_path} parameter is set to \code{TRUE}. 

The function converts these into two primary camtrapR data structures:
\itemize{
  \item CTtable: Contains deployment information including station ID, setup/retrieval dates, 
        camera operation problems, camera setup, and more
  \item recordTable: Contains observation records with taxonomic and temporal information
}

Additional features include:
\itemize{
  \item Parsing of deploymentTags and deploymentGroups (in deployments) and observationTags (in observations)
  \item Extraction of taxonomic information from metadata 
  \item Handling of deployment intervals and gaps
}
}
\note{
Camtrap DP was developed as a consensus of a long, in-depth consultation process with 
camera trap data management platforms and major players in the field of camera trapping.
It builds upon the earlier Camera Trap Metadata Standard (CTMS) but addresses its limitations.

The Camtrap DP standard structures data in a way that supports both media-based observations
(using a single media file as source) and event-based observations (considering an event with
a specified duration as source). For the purpose of this function, both are treated to be equivalent. 
Event-based observations are converted to a single timestamp in the recordTable (using the event start time).

Consider using the 'camtrapdp' package (Desmet et al., 2024) for more general operations on 
Camtrap DP data before converting to camtrapR format with this function.
}
\examples{
\dontrun{
# Read a Camtrap DP dataset
camtrap_data <- readcamtrapDP(
  deployments_file = "path/to/deployments.csv",
  media_file = "path/to/media.csv",
  observations_file = "path/to/observations.csv",
  datapackage_file = "path/to/datapackage.json"
)

# alternatively, set the working directory only
setwd("path/to/camtrapdp_data")
camtrap_data <- readcamtrapDP()   # uses default file names


# Extract components
ct_table <- camtrap_data$CTtable
record_table <- camtrap_data$recordTable
metadata <- camtrap_data$metadata

}

}
\references{
Bubnicki, J.W., Norton, B., Baskauf, S.J., et al. (2023). Camtrap DP: an open standard for the 
FAIR exchange and archiving of camera trap data. Remote Sensing in Ecology and Conservation, 10(3), 283-295.

Desmet, P., Govaert, S., Huybrechts, P., Oldoni, D. (2024). camtrapdp: Read and Manipulate Camera 
Trap Data Packages. R package version 0.3.1. https://CRAN.R-project.org/package=camtrapdp
}
\seealso{
\code{\link[camtrapdp]{read_camtrapdp}} in the 'camtrapdp' package for reading Camtrap DP data
\code{\link[camtrapdp]{check_camtrapdp}} in the 'camtrapdp' package for validating Camtrap DP data
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/createCovariates.R
\name{createCovariates}
\alias{createCovariates}
\title{Extract covariate values from spatial rasters and prepare rasters for spatial predictions}
\usage{
createCovariates(
  CTtable,
  directory,
  filenames,
  rasters,
  buffer_ct = 0,
  bilinear = FALSE,
  buffer_aoi = 1000,
  raster_template = NULL,
  resolution = NULL,
  append = TRUE,
  formats = ".tif",
  recursive = FALSE,
  download_elevation = FALSE,
  elevation_zoom = 10,
  terrain_measures = NULL,
  standardize_na = FALSE,
  scale_covariates = FALSE
)
}
\arguments{
\item{CTtable}{sf object as defined by the \code{sf} package. Essentially a camera trap data frame with spatial information.}

\item{directory}{character. The directory containing the covariate rasters.}

\item{filenames}{character (optionally named). A vector of file paths of covariate rasters. If it is named the covariates will be named according to the names. If unnamed the file names will be used as covariate names.}

\item{rasters}{SpatRaster object or list of SpatRaster objects. Direct input of rasters from the terra package instead of reading from disk.}

\item{buffer_ct}{numeric. A value (in meters) by which to buffer the point locations in \code{CTtable} for extraction of covariate values.}

\item{bilinear}{logical. If \code{TRUE}, extract covariate values with bilinear interpolation (nearest 4 raster cells). If \code{FALSE}, extract value at the cell the point falls in. Only relevant if \code{buffer_ct} is 0.}

\item{buffer_aoi}{numeric. A value (in meters) by which to buffer the overall camera trapping grid to ensure that prediction rasters are larger than the camera trapping grid.}

\item{raster_template}{SpatRaster. A SpatRaster (as defined in the \code{terra} package) to use as template for the creation of prediction rasters.}

\item{resolution}{numeric. Spatial resolution of prediction rasters in the units of the coordinate reference system. Ignored if \code{raster_template} is provided.}

\item{append}{logical. If \code{TRUE}, add the extracted covariates to the existing \code{CTtable}. If \code{FALSE}, return only the extracted covariate values without the existing \code{CTtable}.}

\item{formats}{character. Possible file formats for raster data (must include the dot). Defaults to \code{.tif} files.}

\item{recursive}{logical. If \code{TRUE}, search for raster files recursively in subdirectories when a directory is provided. Defaults to \code{FALSE}.}

\item{download_elevation}{logical. If \code{TRUE}, download elevation data from AWS. Defaults to \code{FALSE}.}

\item{elevation_zoom}{numeric. Zoom level for elevation data download (6-12). Higher values provide more detail but longer download times. Zoom 12 corresponds to ~20m pixel resolution, 11 to ~40m, 10 to ~80m, and so on (resolution halves with each decrease in zoom level). Defaults to 9.}

\item{terrain_measures}{character. Vector of terrain metrics to calculate from elevation data. Options include "slope" (slope in degrees), "aspect" (compass direction in degrees), "TRI" (Terrain Ruggedness Index, measuring elevation difference between adjacent cells), "TPI" (Topographic Position Index, comparing cell elevation to mean of surrounding cells), and "roughness" (difference between max and min of surrounding cells). Defaults to NULL (no terrain metrics).}

\item{standardize_na}{logical. Logical. If \code{TRUE}, ensures all layers in the prediction raster have identical NA patterns by setting a cell to NA in all bands if it's NA in any band. This creates consistency for spatial predictions across covariates but may lose data in covariates.}

\item{scale_covariates}{logical. If \code{TRUE}, scale numeric covariates and return both original and scaled versions of data and prediction rasters. Scaling is performed using R's \code{scale} function. Defaults to \code{FALSE}.}
}
\value{
\describe{
When \code{scale_covariates = FALSE}, a list containing three elements:
  \item{CTtable}{An \code{sf} object representing the camera trap data frame with extracted covariate values.}
  \item{predictionRaster}{A \code{SpatRaster} object containing covariate raster layers}
  \item{originalRaster}{A list of the original input rasters}

When \code{scale_covariates = TRUE}, a list containing six elements:
  \item{CTtable}{The original \code{sf} object with unscaled covariate values}
  \item{CTtable_scaled}{The \code{sf} object with scaled numeric covariate values}
  \item{predictionRaster}{The original unscaled prediction raster}
  \item{predictionRaster_scaled}{The prediction raster with scaled numeric layers}
  \item{originalRaster}{A list of the original input rasters}
  \item{scaling_params}{A list containing center and scale information of numeric covariates}
}
}
\description{
This function extracts covariate values from spatial raster data 
 (e.g. for use in modelling) and prepares these covariates for use in 
 spatial predictions. 
 
 It accepts a camera trap table containing spatial information, 
 along with either a directory containing covariate raster files, a character 
 vector specifying the file paths of these covariate rasters, or direct
 SpatRaster objects from the terra package.
 
 Additionally, users can provide parameters to control how covariates are
 extracted, and how they are aggregated to prediction rasters.
 
 The function can also download elevation data and calculate terrain metrics
 if requested.
 
 The function generates prediction rasters based on a provided template or 
 creates one automatically if no template is provided.
 
 The function returns a list containing the camera trap dataset with 
 extracted covariate values (e.g for use in occupancy modelling), and 
 prediction rasters ready for spatial modeling.
}
\details{
The input camera trap table must be an \code{\link[sf]{sf}} object (a data 
frame with a geometry column specifying the spatial information). For details 
on how to convert an exising camera trap table to sf, see the examples below.
The input rasters can be in different coordinate systems. During covariate 
extraction the CTtable is projected to each raster's coordinate system 
individually. For the prediction raster all input rasters are either 
resampled or reprojected to a consistent coordinate system.

When \code{recursive = TRUE} and a directory is provided, the function will search
for raster files in all subdirectories. In this case, the subdirectory names are
used as covariate names, and only one raster file per subdirectory is allowed.

When \code{download_elevation = TRUE}, the function will download elevation data
from AWS using the \code{elevatr} package. The \code{elevation_zoom}
parameter controls the level of detail, with values between 6 and 12.
Higher zoom levels provide finer resolution but require longer download times and
may consume significant memory. Approximate resolutions: zoom 12 = ~20m, 
11 = ~40m, 10 = ~80m, etc.

If \code{terrain_measures} is specified, the function calculates the requested 
terrain metrics from the elevation data using \code{terra::terrain()} with the 
default 3x3 neighborhood. Available terrain metrics include "slope", "aspect", 
"TRI" (Terrain Ruggedness Index), "TPI" (Topographic Position Index), and 
"roughness".

When using \code{scale_covariates = TRUE}, the function returns both original and
scaled versions of the data and prediction rasters. #' The function uses R's 
\code{scale()} function to perform centering and scaling, and
includes the scaling parameters in the returned metadata.

\subsection{Warning about Categorical Covariates}{
This function does not explicitly handle categorical rasters. All raster values are
treated as numeric, which can be problematic when scaling is applied. The function
attempts to identify "likely categorical" variables (numeric variables with few unique
integer values) and will provide warnings, but it cannot automatically handle them
correctly for scaling.

When using scaled covariates with categorical variables in models:
\itemize{
  \item Use \code{CTtable_scaled} for numeric predictors
  \item Use \code{CTtable} (original) for categorical predictors
  \item Similarly, use \code{predictionRaster_scaled} for numeric predictors in spatial predictions
  \item Use \code{predictionRaster} for categorical predictors in spatial predictions
}

Future versions may implement proper categorical raster handling with RAT (Raster
Attribute Table) support.
}
}
\examples{

\dontrun{
# load camera trap table
data(camtraps)

# create sf object
camtraps_sf <- st_as_sf(camtraps, 
                        coords = c("utm_x", "utm_y"), 
                        crs = 32650)

# extract covariates (with 100m buffer around cameras)
# doesn't run because 'directory' is only a placeholder

covariates <- createCovariates(camtraps_sf, 
"path/to/covariate_rasters", 
buffer_ct = 100, 
buffer_aoi = 1000,
resolution = 100)


# extract covariates with elevation data (this code runs)

covariates_elev <- createCovariates(camtraps_sf,
buffer_ct = 100,
buffer_aoi = 1000,
resolution = 100,
download_elevation = TRUE,
elevation_zoom = 11,
terrain_measures = c("slope", "aspect", "TRI"))

# Note that if local rasters are available they can be extracted alongside
# elevation data in a single function call

# camera trap table with extracted covariate values
camtraps_sf_cov <- covariates_elev$CTtable

# covariate raster layer
r_cov <- covariates_elev$predictionRaster
plot(r_cov)

# Use SpatRaster objects directly as input
r1 <- rast("elevation.tif")
r2 <- rast("landcover.tif")
raster_list <- list(elevation = r1, landcover = r2)

covariates_direct <- createCovariates(camtraps_sf,
                                      rasters = raster_list,
                                      buffer_ct = 100,
                                      resolution = 100)

# Scale numeric covariates for modeling
covariates_scaled <- createCovariates(camtraps_sf,
                                      rasters = raster_list,
                                      buffer_ct = 100,
                                      resolution = 100,
                                      scale_covariates = TRUE)

# Use scaled data with categorical variables
# Mix and match from original and scaled outputs for tabular data
model_data <- covariates_scaled$CTtable_scaled  # Use scaled numeric covariates
model_data$landcover <- covariates_direct$CTtable$landcover  # Use original categorical covariate

# Mix and match for prediction rasters
# Create combined prediction raster (scaled numeric variables & original categorical variables)
# Extract scaled elevation layer
elev_scaled <- covariates_scaled$predictionRaster_scaled$elevation

# Extract original landcover layer (categorical)
landcover_orig <- covariates_direct$predictionRaster$landcover

# Combine into a new SpatRaster for predictions
prediction_raster <- c(elev_scaled, landcover_orig)
names(prediction_raster) <- c("elevation", "landcover")

# Use this combined raster for spatial predictions
plot(prediction_raster)
}

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/plot.R
\name{plot_effects,commOccu-method}
\alias{plot_effects,commOccu-method}
\alias{plot_effects}
\title{Plot Marginal Effects of Covariates}
\usage{
\S4method{plot_effects}{commOccu}(
  object,
  mcmc.list,
  submodel = "state",
  response = "occupancy",
  speciesSubset,
  draws = 1000,
  outdir,
  level = 0.95,
  keyword_quadratic = "_squared",
  ...
)
}
\arguments{
\item{object}{\code{commOccu} object}

\item{mcmc.list}{mcmc.list. Output of \code{fit} called on a \code{commOccu} object}

\item{submodel}{character. Submodel to get plots for. Can be \code{"det"} (detection submodel) or \code{"state"} (occupancy submodel)}

\item{response}{character. response type on y axis. Only relevant for \code{submodel = "state"}. Default is \code{"occupancy"}, can be set to \code{"abundance"} for Royle-Nichols models}

\item{speciesSubset}{character. Species to include in effect plots.}

\item{draws}{integer. Number of draws from the posterior to use when generating the plots. If fewer posterior samples than specified in \code{draws} are available, all posterior samples are used.}

\item{outdir}{character. Directory to save plots to (optional)}

\item{level}{numeric. Probability mass to include in the uncertainty interval.}

\item{keyword_quadratic}{character. A suffix in covariate names in the model that indicates a covariate is a quadratic effect of another covariate which does not carry the suffix in its name (e.g. if the covariate is "elevation", the quadratic covariate would be "elevation_squared").}

\item{...}{additional arguments for \code{\link[ggplot2]{ggsave}} - only relevant if \code{outdir} is defined}
}
\value{
A list of ggplot objects (one list item per covariate).
}
\description{
Plot marginal effect plots (= response curves if covariates are continuous) for all species in a community (multi-species) occupancy model. Takes into account species-specific intercepts (if any). Currently only supports continuous covariates, not categorical covariates.
}
\details{
Users who wish to create their own visualizations can use the data stored in the ggplot output. It is accessed via e.g. \code{output$covariate_name$data}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/communityModel.R
\name{summary,commOccu-method}
\alias{summary,commOccu-method}
\title{Summarize community occupancy model}
\usage{
\S4method{summary}{commOccu}(object, ...)
}
\arguments{
\item{object}{\code{commOccu} object}

\item{...}{currently ignored}
}
\value{
Model summary printed to console
}
\description{
Gives an overview of the number of species, stations and occasions in a \code{commOccu} object. Also returns covariates.
}
\details{
The summary method is very basic and still work in progress.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/writeDateTimeOriginal.R
\name{writeDateTimeOriginal}
\alias{writeDateTimeOriginal}
\title{Write values to DateTimeOriginal tag in image metadata}
\usage{
writeDateTimeOriginal(DateTimeOriginal, fileNames, parallel, overwrite = FALSE)
}
\arguments{
\item{DateTimeOriginal}{character. DateTimeOriginal values to write into
EXIF:DateTimeOriginal tag of the files in fileNames.}

\item{fileNames}{character. Full file names (including directories) of
images to process.}

\item{parallel}{A parallel cluster object. Specify if you wish to run this
function in parallel. Provide a cluster object (output of makeCluster()) -
optional.}

\item{overwrite}{logical. Overwrite existing files (TRUE) or create new
files while saving the original data as jpg_original files as a backup
(FALSE)?}
}
\value{
Invisible NULL. The actual output is the JPG images which now have a
DateTimeOriginal tag.
}
\description{
This function assigns values to the DateTimeOriginal tag in image's EXIF
metadata using Exiftool. It can be used when the original DateTimeOriginal
values in the metadata were lost for whatever reason. In order to first read
the Date/Time values from the data fields in the images, see the function
\code{\link{OCRdataFields}}. After running \code{OCRdataFields} and checking
its output you can run \code{writeDateTimeOriginal}.
}
\details{
The first value in \code{DateTimeOriginal} will be assigned to the first
image in \code{fileNames}, and so on. Both \code{DateTimeOriginal} and
\code{fileNames} can be obtained from the output of
\code{\link{OCRdataFields}}.  \code{DateTimeOriginal} uses the standard
"YYYY-MM-SS HH:MM:SS"" notation. If the values extracted via
\code{\link{OCRdataFields}} are in a different format you'll need to
reformat them first. Please provide them as character. Also, before using
this function, make sure that the date/time values read by
\code{\link{OCRdataFields}} are correct (sometimes OCR misreads values, so
check carefully).

Parallel processing is advised since the function is rather slow (due to
calling Exiftool separately on every, so about 1 second per image). If you
know how to batch-assign DateTimeOriginal values in one Exiftool call,
please let me know.

The function only works on JPG images, not video files.
}
\examples{


\dontrun{
# dontrun is to avoid forcing users to install additional dependencies

wd_images_OCR <- system.file("pictures/full_size_for_ocr", package = "camtrapR")

library(magick)

# define geometries
geometry1 <- geometry_area(x_off = 0, y_off = 0, width = 183, height = 37)
geometry2 <- geometry_area(x_off = 196, y_off = 0, width = 200, height = 17)
geometry3 <- geometry_area(x_off = 447, y_off = 0, width = 63, height = 17)
geometry4 <- geometry_area(x_off = 984, y_off = 0, width = 47, height = 17)
geometry5 <- geometry_area(x_off = 0, y_off = 793, width = 320, height = 17)

# combine geometries into list
geometries <- list(date = geometry1, 
                   time = geometry2, 
                   sequence_id = geometry3,
                   temperature = geometry4,
                   camera_model = geometry5)

df_image_data  <- OCRdataFields(inDir = wd_images_OCR,
                                geometries = geometries, 
                                invert = TRUE)       
df_image_data



library(parallel)
library(lubridate)

# prepare DateTimeOriginal column (ymd_hms() automatically respects the PM indicator)
df_image_data$DateTimeOriginal <- paste(df_image_data$date, df_image_data$time)
df_image_data$DateTimeOriginal <- as.character(ymd_hms(df_image_data$DateTimeOriginal))

# create cluster (3 cores)
cl <- makeCluster(3)

# assign new DateTimeOriginal
writeDateTimeOriginal(DateTimeOriginal = df_image_data$DateTimeOriginal,
                      fileNames = df_image_data$filename_full,
                      parallel = cl)
                      
stopCluster(cl)
                   
}



}
\seealso{
\code{\link{OCRdataFields}}
}
\author{
Juergen Niedballa
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/PPC.residuals.R
\name{PPC.community}
\alias{PPC.community}
\title{Calculate Community-Level Posterior Predictive Checks for Occupancy Models}
\usage{
PPC.community(
  p,
  psi,
  y,
  input_format = c("predict", "lists"),
  K = NULL,
  model = c("Occupancy", "RN"),
  zhat = NULL,
  z.cond = TRUE,
  type = c("FT", "PearChi2", "Deviance"),
  return.residuals = TRUE,
  show_progress = TRUE,
  ...
)
}
\arguments{
\item{p}{Either a 4D array [stations, species, iterations, occasions] from 
\code{\link{predict}} or a list of 3D arrays [iterations, stations, occasions], 
one per species}

\item{psi}{Either a 3D array [stations, species, iterations] from 
\code{\link{predict}} or a list of 2D arrays [iterations, stations], one per species}

\item{y}{List of detection histories, one matrix/vector per species}

\item{input_format}{Character indicating format of p and psi ("predict" or "lists")}

\item{K}{Number of occasions as either a scalar or site vector. Calculated automatically 
if y is a list of matrices.}

\item{model}{Character indicating model type ("Occupancy" or "RN")}

\item{zhat}{List of z estimate matrices, one per species (optional). Each matrix 
should follow the format specified in \code{\link{PPC.residuals}}.}

\item{z.cond}{Logical. If TRUE, new data is conditioned on estimated z (testing only 
detection model fit). If FALSE, generates new z for each posterior sample 
(testing complete model).}

\item{type}{Character indicating residual type ("FT", "PearChi2", or "Deviance")}

\item{return.residuals}{Logical. If TRUE, returns species-specific residuals along 
with Bayesian p-values. If FALSE, returns only the p-values.}

\item{show_progress}{Logical; whether to show a progress bar during computation 
(if package pbapply is available)}

\item{...}{Additional arguments passed to \code{\link{PPC.residuals}}.}
}
\value{
If return.residuals=TRUE, returns a list containing:
  \itemize{
    \item bp_values - Data frame with species-specific and community-level 
          Bayesian p-values
    \item species_results - List containing the complete PPC.residuals output 
          for each species
  }
  If return.residuals=FALSE, returns only the data frame of Bayesian p-values.
}
\description{
A wrapper function that applies Posterior Predictive Checks (PPC) across multiple 
species in a community occupancy model (from \code{\link{communityModel}}.
It calculates species-specific and community-level Bayesian p-values to assess 
model fit. The function accepts both \code{\link{predict}} output format and 
list format for model parameters.
}
\details{
This function extends the single-species PPC analysis to the community level by:
\itemize{
  \item Applying residual calculations to each species in the community
  \item Aggregating results to assess community-level model fit
  \item Providing both species-specific and community-level diagnostics
}

This function provides flexibility in input formats to accommodate different workflows:
\itemize{
  \item Direct output from camtrapR's predict() function (4D/3D arrays)
  \item Lists of species-specific arrays (for custom workflows)
}
}
\examples{
\dontrun{

# Create and fit model 
data("camtraps")

# create camera operation matrix
camop_no_problem <- cameraOperation(CTtable      = camtraps,
                                    stationCol   = "Station",
                                    setupCol     = "Setup_date",
                                    retrievalCol = "Retrieval_date",
                                    hasProblems  = FALSE,
                                    dateFormat   = "dmy"
)

data("recordTableSample")

# make list of detection histories
species_to_include <- unique(recordTableSample$Species)

  DetHist_list <- detectionHistory(
    recordTable         = recordTableSample,
    camOp                = camop_no_problem,
    stationCol           = "Station",
    speciesCol           = "Species",
    recordDateTimeCol    = "DateTimeOriginal",
    species              = species_to_include,
    occasionLength       = 7,
    day1                 = "station",
    datesAsOccasionNames = FALSE,
    includeEffort        = TRUE,
    scaleEffort          = TRUE,
    timeZone             = "Asia/Kuala_Lumpur"
)


# create some fake covariates for demonstration
sitecovs <- camtraps[, c(1:3)]
sitecovs$elevation <- c(300, 500, 600)   

# scale numeric covariates
sitecovs[, c(2:4)] <- scale(sitecovs[,-1])


# bundle input data for communityModel
data_list <- list(ylist = DetHist_list$detection_history,
                  siteCovs = sitecovs,
                  obsCovs = list(effort = DetHist_list$effort))


# create community model for JAGS
modelfile1 <- tempfile(fileext = ".txt")
mod.jags <- communityModel(data_list,
                           occuCovs = list(fixed = "utm_y", ranef = "elevation"),
                           detCovsObservation = list(fixed = "effort"),
                           intercepts = list(det = "ranef", occu = "ranef"),
                           modelFile = modelfile1)

summary(mod.jags)

# fit in JAGS
fit.jags <- fit(mod.jags,
                n.iter = 1000,
                n.burnin = 500,
                chains = 3)   
summary(fit.jags)



# create predictions for p and psi
draws <- 100

p <- predict(object = mod.jags,
             mcmc.list = fit.jags,
             type = "p_array",
             draws = draws)
# output is in order [station, species, draw, occasion]
 
 psi <- predict(object = mod.jags,
                mcmc.list = fit.jags,
                type = "psi_array",
                draws = draws)
 # output is in order [station, species, draw]



ppc_comm <- PPC.community(
  p = p,
  psi = psi,
  y = mod.jags@input$ylist, 
  model = "Occupancy",
  type = "FT")
  
# Bayesian p-values
 ppc_comm$BP
 
   
 str(ppc_comm$residuals)
 
 # get individual species PPC results
 ppc_species <- ppc_comm$residuals[[1]] # first species
 
 plot(apply(ppc_species$res.obs, 2, mean), apply(ppc_species$res.new, 2, mean),
 xlab = "Observed residuals",
 ylab = "Predicted residuals"
 )
 
 abline(0,1)  # diagonal line is not visible due to tiny data set
  
 }
}
\seealso{
\code{\link{PPC.residuals}} for details on the underlying single-species 
  calculations
}
\author{
Rahel Sollmann
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/PPC.residuals.R
\name{PPC.residuals}
\alias{PPC.residuals}
\title{Calculate Residuals from MCMC Output of Occupancy Models}
\usage{
PPC.residuals(
  y,
  p,
  psi,
  model = c("Occupancy", "RN"),
  type = c("FT", "PearChi2", "Deviance"),
  K = NULL,
  z.cond = TRUE,
  zhat = NULL,
  nmax = 20,
  return.residuals = TRUE,
  return.z = TRUE
)
}
\arguments{
\item{y}{Observations as either a site vector or site by occasion matrix. For matrix 
format, use NA for unsampled occasions.}

\item{p}{Array of posterior samples for detection probability (p). Dimensions should be 
iterations by sites (by occasion optionally). For RN models, p should represent 
individual-level detection (not conditional on local abundance).}

\item{psi}{Array of posterior samples for occupancy probability (psi). Dimensions should 
be iterations by sites. For RN models, psi should represent expected abundance}

\item{model}{Character indicating model type: either "Occupancy" or "RN" 
(Royle-Nichols).}

\item{type}{Type of residual to calculate: "FT" (Freeman Tukey), "PearChi2" 
(Pearson Chi-squared), or "Deviance" (not technically a residual).}

\item{K}{Number of occasions as either a scalar or site vector. Calculated automatically 
if y is a matrix.}

\item{z.cond}{Logical. If TRUE, new data is conditioned on estimated z (testing only 
detection model fit). If FALSE, generates new z for each posterior sample 
(testing complete model).}

\item{zhat}{Optional matrix with same dimensions as psi containing estimates of z from 
the same model. If not provided, will be generated internally.}

\item{nmax}{Maximum site-level abundance (default = 20). Only used if model="RN". 
Higher values increase computation time. Warning given if set too low.}

\item{return.residuals}{Logical. If TRUE (default), returns residuals along with 
Bayesian p-value.}

\item{return.z}{Logical. If TRUE, returns z values conditional on y, and unconditional 
z's if \code{z.cond = FALSE}. Note: if zhat is provided, the returned conditional-on-y z 
values will be identical to those provided.}
}
\value{
If return.residuals=TRUE (default), returns a list containing:
  \itemize{
    \item res.obs - residuals for observed data
    \item res.new - residuals for newly generated data
    \item BP - Bayesian p-value
  }
  If return.residuals=FALSE, returns only the Bayesian p-value.
}
\description{
Posterior Predictive Check (PPC) function that calculates Freeman-Tukey (FT) 
residuals, Pearson"s Chi-squared residuals, or deviance from MCMC output of 
occupancy models. This function compares observed data to simulated data from 
the posterior distribution to assess model fit.
}
\details{
This function helps assess model fit for occupancy models using various types of residuals:

\subsection{Types of Residuals}{
  \itemize{
    \item Freeman-Tukey (FT):
          \deqn{R_j = (\sqrt{y_j} - \sqrt{E(y_j)})^2}
          Measures the squared difference between the square root 
          of observed detections and the square root of expected detections 
          at each site.
    \item Pearson Chi-squared:
          \deqn{R_j = \left(\frac{y_j - E(y_j)}{\sqrt{Var(y_j)}}\right)^2}
          Measures the squared difference between observed and 
          expected detections, standardized by the theoretical variance calculated 
          from the model parameters.
    \item Deviance:
          \deqn{R_j = -2\log[y_j|\theta_j, K_j]}
          Measures the contribution of each site to the overall model 
          likelihood, quantifying the discrepancy between observed data and model 
          predictions based on likelihood ratios
  }

  Where:
  \itemize{
    \item \eqn{y_j} is the number of detections of the species at site j, out of 
          \eqn{K_j} repeated surveys
    \item \eqn{E(y_j) = K_j p_j z_j}, with \eqn{p_j} = species detection probability 
          and \eqn{z_j} = occupancy state (1 if occupied, 0 otherwise)
    \item \eqn{Var(y_j) = p_j z_j (1 - p_j z_j) K_j}
    \item For Royle-Nichols occupancy models, the term \eqn{p_j z_j} is replaced with 
          \eqn{1 - (1 - r_j)^{N_j}}, where \eqn{r_j} = individual detection probability 
          and \eqn{N_j} = local abundance
    \item For Deviance, \eqn{\theta_j} is either occupancy and species detection probability 
          at site j (\eqn{\psi_j, p_j}) for regular occupancy models, or expected abundance 
          and individual detection probability (\eqn{\lambda_j, r_j}) for Royle-Nichols 
          occupancy models
  }
}

\subsection{Bayesian P-values}{
  The function calculates Bayesian p-values as a measure of model fit. These values:
  \itemize{
    \item Range from 0 to 1
    \item Values close to 0.5 suggest good model fit
    \item Values close to 0 or 1 suggest poor fit
    \item Are calculated by comparing observed residuals to residuals from 
          simulated data
  }
}

\subsection{Conditional vs Unconditional Assessment}{
  The z.cond parameter allows for two types of model assessment:
  \itemize{
    \item z.cond = TRUE: Tests only the detection component of the model, fixing 
          occupancy/abundance to estimates from the model, rather than generating them anew
    \item z.cond = FALSE: Tests the complete model, including both occupancy and 
          detection components
  }
}
}
\note{
FT and Chi-squared residuals have been extensively tested. Deviance calculations 
have undergone less testing and are only available for scenarios with constant 
detection probability across occasions. FT and Chi-squared residuals can handle 
varying detection probabilities.
}
\section{Warning}{
 This is a beta version of the function. While it has been tested extensively, 
  not all possible data configurations may have been captured in testing. This is 
  particularly true for:
  
  \itemize{
    \item Deviance calculations (type = "Deviance")
    \item Royle-Nichols models (model = "RN")
  }
  If you encounter issues with the function, please contact the package developers.
}

\references{
Sollmann, Rahel. Occupancy models and the "good fit, bad prediction" dilemma. 
Ecology (submitted)
}
\author{
Rahel Sollmann
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/recordTableIndividual.R
\name{recordTableIndividual}
\alias{recordTableIndividual}
\title{Generate a single-species record table with individual identification from
camera trap images or videos}
\usage{
recordTableIndividual(
  inDir,
  hasStationFolders,
  IDfrom,
  cameraID,
  camerasIndependent,
  minDeltaTime = 0,
  deltaTimeComparedTo,
  timeZone,
  stationCol,
  writecsv = FALSE,
  outDir,
  metadataHierarchyDelimitor = "|",
  metadataIDTag,
  additionalMetadataTags,
  removeDuplicateRecords = TRUE,
  returnFileNamesMissingTags = FALSE,
  eventSummaryColumn,
  eventSummaryFunction,
  video
)
}
\arguments{
\item{inDir}{character. Directory containing images of individuals. Must end
with species name (e.g. ".../speciesImages/Clouded Leopard")}

\item{hasStationFolders}{logical. Does \code{inDir} have station
subdirectories? If \code{TRUE}, station IDs will be taken from directory
names. If \code{FALSE}, they will be taken from image filenames (requires
images renamed with \code{\link{imageRename}}).}

\item{IDfrom}{character. Read individual ID from image metadata ("metadata")
of from directory names ("directory")?}

\item{cameraID}{character. Should the function look for camera IDs in the
image file names? If so, set to 'filename'. Requires images renamed with
\code{\link{imageRename}}. If missing, no camera ID will be assigned and it
will be assumed there was 1 camera only per station.}

\item{camerasIndependent}{logical. If \code{TRUE}, cameras at a station are
assumed to record individuals independently. If \code{FALSE}, cameras are
assumed to be non-independent (e.g. in pairs). Takes effect only if there
was more than 1 camera per station and cameraID = "filename".}

\item{minDeltaTime}{numeric. time difference between observation of the same
individual at the same station/camera to be considered independent (in
minutes)}

\item{deltaTimeComparedTo}{character. For two records to be considered
independent, must the second one be at least \code{minDeltaTime} minutes
after the last independent record of the same individual
(\code{"lastIndependentRecord"}), or \code{minDeltaTime} minutes after the
last record (\code{"lastRecord"})?}

\item{timeZone}{character. Must be a value returned by
\code{\link[base:timezones]{OlsonNames}}}

\item{stationCol}{character. Name of the camera trap station column in the
output table.}

\item{writecsv}{logical. Should the individual record table be saved as a
.csv file?}

\item{outDir}{character. Directory to save csv file to. If NULL and
\code{writecsv = TRUE}, the output csv will be written to \code{inDir}.}

\item{metadataHierarchyDelimitor}{character. The character delimiting
hierarchy levels in image metadata tags in field "HierarchicalSubject".
Either "|" or ":".}

\item{metadataIDTag}{character. In custom image metadata, the individual ID
tag name.}

\item{additionalMetadataTags}{character. additional camera model-specific
metadata tags to be extracted. (If possible specify tag groups as returned
by \code{\link{exifTagNames}})}

\item{removeDuplicateRecords}{logical. If there are several records of the
same individual at the same station (also same camera if cameraID is
defined) at exactly the same time, show only one?}

\item{returnFileNamesMissingTags}{logical. If species are assigned with
metadata and images are not tagged, return a few file names of these images
as a message?}

\item{eventSummaryColumn}{character. A column in the record table (e.g. from
a metadata tag) by to summarise non-independent records (those within
\code{minDeltaTime} of a given record) with a user-defined function
(\code{eventSummaryFunction})}

\item{eventSummaryFunction}{character. The function by which to summarise
\code{eventSummaryColumn} of non-independent records, e.g. "sum", "max"
(optional)}

\item{video}{list. Contains information on how to handle video data
(optional). See details.}
}
\value{
A data frame containing species records with individual IDs and
additional information about stations, date, time and (optionally) further
metadata.
}
\description{
The function generates a single-species record table containing individual
IDs, e.g. for (spatial) capture-recapture analyses. It prepares input for
the function \code{\link{spatialDetectionHistory}}.
}
\details{
The function can handle a number of different ways of storing images and
videos. In every case, images need to be stored in a species directory first
(e.g. using function \code{\link{getSpeciesImages}}). Station subdirectories
are optional. Camera subdirectories are not supported. This directory
structure can be created easily with function
\code{\link{getSpeciesImages}}.

As with species identification, individuals can be identified in 2 different
ways: by moving images into individual directories
("Species/Station/Individual/XY.JPG" or "Species/Individual/XY.JPG") or by
metadata tagging (without the need for individual directories:
"Species/XY.JPG" or "Species/Station/XY.JPG").

\code{minDeltaTime} is a criterion for temporal independence of records of
an individual at the same station/location. Setting it to 0 will make the
function return all records. \code{camerasIndependent} defines if the
cameras at a station are to be considered independent (e.g. \code{FALSE} if
both cameras face each other and possibly \code{TRUE} if they face different
trails). \code{stationCol} is the station column name to be used in the
resulting table. Station IDs are read from the station directory names if
\code{hasStationFolders = TRUE}. Otherwise, the function will try to extract
station IDs from the image filenames (requires images renamed with
\code{\link{imageRename}}.

If individual IDs were assigned with image metadata tags,
\code{metadataIDTag} must be set to the name of the metadata tag group used
for individual identification. \code{metadataHierarchyDelimitor} is "|" for
images tagged in DigiKam and images tagged in Adobe Bridge/ Lightroom with
the default settings. Manufacturer-specific Exif metadata tags such as
"AmbientTemperature" or "MoonPhase" can be extracted if specified in
\code{additionalMetadataTags}. Multiple names can be specified as a
character vector as: \code{c(Tag1, Tag2, ...)}. Because they are not
standardized, function \code{\link{exifTagNames}} provides a vector of all
available tag names. The metadata tags thus extracted may be used as
individual covariates in spatial capture-recapture models.

\code{eventSummaryColumn} and \code{eventSummaryFunction} can be used to
extract summary statistics for independent sampling events. For example, you
assigned a "count" tag to your images, indicating the number of individuals
in a picture. In a sequence of pictures taken within 1 minute, most pictures
show one individual, but one image shows two individuals. You tagged the
images accordingly (count = 1 or count = 2) and run \code{recordTable}. Set
\code{eventSummaryColumn = "count"} and \code{eventSummaryFunction = "max"}
to obtain the maximum number of \code{count} in all images within
\code{minDeltaTime} minutes of a given record. The results is in a new
column, in this example \code{count_max}. You can also calculate several
statistics at the same time, by supplying vectors of values, e.g.
\code{eventSummaryColumn = c("count", "count", "camera")} and
\code{eventSummaryFunction = c("min", "max", "unique")} to get minimum and
maximum count and all unique camera IDs for that event. Note that
\code{eventSummaryColumn} and \code{eventSummaryFunction} must be of same
length.

Argument \code{video} is analogous to \code{\link{recordTable}}, a named
list with 2 or 4 items. 2 items (\code{file_formats}, \code{dateTimeTag})
are always required, and are sufficent if \code{IDfrom = "directory"}. In
that case, no digiKam tags will be returned.  To return digiKam tags, two
additional items are required (\code{db_directory}, \code{db_filename}).
This is essential when using \code{IDfrom = "metadata"}. When using
\code{IDfrom = "directory"}, it is optional, but allows to extract metadata
tags assigned to videos in digiKam. This workaround is necessary because
digiKam tags are not written into video metadata, but are only saved in the
digiKam database. So in contrast to JPG images, they can not be extracted
with ExifTool. It also requires that \code{inDir} is in your digiKam
database.

The items of argument \code{video} are:

\tabular{ll}{ \code{file_formats} \tab The video formats to extract (include
"jpg" if you want .JPG image metadata) \cr \code{dateTimeTag} \tab the
metadata tag to extract date/time from (use \code{\link{exifTagNames}} to
find out which tag is suitable) \cr \code{db_directory} \tab The directory
containing digiKam database (optional if \code{IDfrom = "directory"}) \cr
\code{db_filename} \tab The digiKam database file in \code{db_directory}
(optional if \code{IDfrom = "directory"}) \cr }

See the example below for for how to specify the argument \code{video}.
}
\section{Warning }{
 Be sure to read the section on individual identification
in the package vignette
(\url{https://CRAN.R-project.org/package=camtrapR/vignettes/camtrapr2.pdf}).

Af you use image metadata tags for identification, the tags must be written
to the image metadata. The function cannot read tags from .xmp sidecar
files. Make sure you set the preferences of your image management software
accordingly. In DigiKam, go to Settings/Configure digiKam/Metadata. There,
make sure "Write to sidecar files" is unchecked.

Please note the section about defining argument \code{timeZone} in the
vignette on data extraction (accessible via
\code{vignette("DataExtraction")} or online
(\url{https://cran.r-project.org/package=camtrapR/vignettes/camtrapr3.pdf})).
}

\examples{


\dontrun{   # the examples run too long to pass CRAN tests

 wd_images_ID_individual <- system.file("pictures/sample_images_indiv_tag/LeopardCat", 
                                        package = "camtrapR")
 # missing space in species = "LeopardCat" is because of CRAN package policies
 # note argument additionalMetadataTags: contains tag names as returned by function exifTagNames

 if (Sys.which("exiftool") != ""){        # only run these examples if ExifTool is available

 rec_table_pbe <- recordTableIndividual(inDir                  = wd_images_ID_individual,
                                        minDeltaTime           = 60,
                                        deltaTimeComparedTo    = "lastRecord",
                                        hasStationFolders      = FALSE,
                                        IDfrom                 = "metadata",
                                        camerasIndependent     = FALSE,
                                        writecsv               = FALSE,
                                        metadataIDTag          = "individual",
                                        additionalMetadataTags = c("EXIF:Model", "EXIF:Make"),
                                        timeZone               = "Asia/Kuala_Lumpur"
 )
 
 
 # extracting some example summary stats too 
 # a nonsensical example, get all unique cameras with which the event was photographed
 
  rec_table_pbe2 <- recordTableIndividual(inDir                = wd_images_ID_individual,
                                        minDeltaTime           = 60,
                                        deltaTimeComparedTo    = "lastRecord",
                                        hasStationFolders      = FALSE,
                                        IDfrom                 = "metadata",
                                        camerasIndependent     = FALSE,
                                        writecsv               = FALSE,
                                        metadataIDTag          = "individual",
                                        additionalMetadataTags = c("EXIF:Model", "EXIF:Make"),
                                        timeZone               = "Asia/Kuala_Lumpur",
                                        eventSummaryColumn     = "EXIF:Make",
                                        eventSummaryFunction   = "unique"
 )
 
 ### Video example (the sample data don't contain a video, this is hypothetical)
 # with JPG, video mp4, avi, mov, ID = metadata
 
rec_table_ind_video <- recordTableIndividual(inDir = wd_images_ID_individual,
                          hasStationFolder  = FALSE,
                          IDfrom            = "metadata", 
                          metadataIDTag     = "individual",
                          video = list(file_formats = c("jpg", "mp4", "avi", "mov"),
                                       dateTimeTag = "QuickTime:CreateDate",
                                       db_directory = "C:/Users/YourName/Pictures",
                                       db_filename = "digikam4.db")
)

  
} else {  
# show function output if ExifTool is not available
message("ExifTool is not available. Cannot test function. Loading recordTableSample instead")
data(recordTableSample)
}
}


}
\references{
Phil Harvey's ExifTool \url{https://exiftool.org/}
}
\author{
Juergen Niedballa
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/activityRadial.R
\name{activityRadial}
\alias{activityRadial}
\title{Radial plots of single-species activity}
\usage{
activityRadial(
  recordTable,
  species,
  allSpecies = FALSE,
  speciesCol = "Species",
  recordDateTimeCol = "DateTimeOriginal",
  recordDateTimeFormat = "ymd HMS",
  byNumber = FALSE,
  plotR = TRUE,
  writePNG = FALSE,
  plotDirectory,
  createDir = FALSE,
  pngMaxPix = 1000,
  ...
)
}
\arguments{
\item{recordTable}{data.frame. the record table created by
\code{\link{recordTable}}}

\item{species}{Name of the species for which to create an kernel density
plot of activity}

\item{allSpecies}{logical. Create plots for all species in \code{speciesCol}
of \code{recordTable}? Overrides argument \code{species}}

\item{speciesCol}{character. name of the column specifying species names in
\code{recordTable}}

\item{recordDateTimeCol}{character. name of the column specifying date and
time in \code{recordTable}}

\item{recordDateTimeFormat}{character. format of column
\code{recordDateTimeCol} in \code{recordTable}}

\item{byNumber}{logical. If FALSE, plot proportion of records. If TRUE, plot
number of records}

\item{plotR}{logical. Show plots in R graphics device?}

\item{writePNG}{logical. Create pngs of the plots?}

\item{plotDirectory}{character. Directory in which to create png plots if
\code{writePNG = TRUE}}

\item{createDir}{logical. Create \code{plotDirectory}?}

\item{pngMaxPix}{integer. image size of png (pixels along x-axis)}

\item{\dots}{additional arguments to be passed to function
\code{\link[plotrix]{radial.plot}}}
}
\value{
Returns invisibly a data.frame containing all information needed to
create the plot: radial position, lengths, hour (for labels). If
\code{allSpecies == TRUE}, all species' data frames are returned in an
invisible named list.
}
\description{
The function generates a radial plot of species diel activity using an
adapted version of function \code{\link[plotrix]{radial.plot}} from package
\pkg{plotrix} (without the need to install the package). Records are
aggregated by hour. The number of independent events is used as input, which
in turn is based on the argument \code{minDeltaTime} in
\code{\link{recordTable}}.
}
\details{
\code{radial.plot} was adjusted to show a clockwise 24-hour clock face. It
is recommended to set argument \code{lwd} to a value >= 2. You may also wish
to add argument \code{rp.type="p"} to show a polygon instead of bars.

\code{recordDateTimeFormat} defaults to the "YYYY-MM-DD HH:MM:SS"
convention, e.g. "2014-09-30 22:59:59". \code{recordDateTimeFormat} can be
interpreted either by base-R via \code{\link[base]{strptime}} or in
\pkg{lubridate} via \code{\link[lubridate]{parse_date_time}} (argument
"orders"). \pkg{lubridate} will be used if there are no "\%" characters in
\code{recordDateTimeFormat}.

For "YYYY-MM-DD HH:MM:SS", \code{recordDateTimeFormat} would be either
"\%Y-\%m-\%d \%H:\%M:\%S" or "ymd HMS". For details on how to specify date
and time formats in R see \code{\link[base]{strptime}} or
\code{\link[lubridate]{parse_date_time}}.
}
\examples{


# load record table
data(recordTableSample)

species4activity <- "PBE"    # = Prionailurus bengalensis, Leopard Cat

activityRadial(recordTable       = recordTableSample,
               species           = species4activity,
               allSpecies        = FALSE,
               speciesCol        = "Species",
               recordDateTimeCol = "DateTimeOriginal",
               plotR             = TRUE,
               writePNG          = FALSE,
               lwd               = 5
)

# plot type = polygon

activityRadial(recordTable       = recordTableSample,
               species           = species4activity,
               allSpecies        = FALSE,
               speciesCol        = "Species",
               recordDateTimeCol = "DateTimeOriginal",
               plotR             = TRUE,
               writePNG          = FALSE,
               lwd               = 5,
               rp.type           = "p"      
)


}
\references{
Lemon, J. (2006) Plotrix: a package in the red light district of
R. R-News, 6(4): 8-12. \cr \url{https://CRAN.R-project.org/package=plotrix }
}
\seealso{
\code{\link{activityDensity}}, \code{\link{activityHistogram}},
\code{\link{activityOverlap}}
}
\author{
Juergen Niedballa
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/cameraOperation.R
\name{cameraOperation}
\alias{cameraOperation}
\title{Create a camera trap station operation matrix}
\usage{
cameraOperation(
  CTtable,
  stationCol = "Station",
  cameraCol,
  sessionCol,
  setupCol,
  retrievalCol,
  hasProblems = FALSE,
  byCamera,
  allCamsOn,
  camerasIndependent,
  dateFormat = "ymd",
  occasionStartTime = 0,
  writecsv = FALSE,
  outDir
)
}
\arguments{
\item{CTtable}{data.frame containing information about location and trapping
period of camera trap stations}

\item{stationCol}{character. name of the column specifying Station ID in
\code{CTtable}}

\item{cameraCol}{character. name of the column specifying Camera ID in
\code{CTtable} (optional). If empty, 1 camera per station is assumed.}

\item{sessionCol}{character. name of the column specifying session ID in
\code{CTtable} (optional). Use it for creating multi-session / multi-season
detection histories (unmarked: \code{\link[unmarked]{unmarkedMultFrame}};
secr: \code{\link[secr]{capthist}} )}

\item{setupCol}{character. name of the column containing camera setup dates
in \code{CTtable}}

\item{retrievalCol}{character. name of the column containing camera
retrieval dates in \code{CTtable}}

\item{hasProblems}{logical. If TRUE, function will look for columns
specifying malfunction periods in \code{CTtable} (naming convention:
\code{ProblemX_from} and \code{ProblemX_to}, where X is a number)}

\item{byCamera}{logical. If TRUE, camera operation matrix is computed by
camera, not by station (requires \code{cameraCol})}

\item{allCamsOn}{logical. Takes effect only if \code{cameraCol} is defined
and if \code{byCamera} is FALSE. If \code{allCamsOn = TRUE}, all cameras at
a station need to be operational for the station to be operational (e.g. 1
camera out of 2 malfunctioning renders the station inoperational). Output
values can be 1/0/NA only (all cameras at a station operational/ at least 1
camera not operational/ no camera set up). If \code{allCamsOn = FALSE}, at
least 1 active camera makes a station operational.}

\item{camerasIndependent}{logical. Return number of active camera traps by
station? Only if \code{byCamera} is FALSE and \code{allCamsOn} is FALSE. If
\code{camerasIndependent} is TRUE, output values will be the number of
operational cameras at a station. If \code{camerasIndependent} is FALSE, the
value is 1 if at least 1 camera was operational, otherwise 0. In both cases,
values are NA if no camera was set up.}

\item{dateFormat}{character. The format of columns \code{setupCol} and
\code{retrievalCol} (and potential problem columns) in \code{CTtable}. Must
be interpretable by either \code{as.Date} or the "orders" argument
\code{\link[lubridate]{parse_date_time}} in \pkg{lubridate}. Can be a date
or (since version 2.1) a date-time.}

\item{occasionStartTime}{integer. time of day (the full hour) at which to
begin occasions. Replaces \code{occasionStartTime} from
\code{\link{detectionHistory}} and \code{\link{spatialDetectionHistory}}.}

\item{writecsv}{logical. Should the camera operation matrix be saved as a
.csv?}

\item{outDir}{character. Directory into which csv is saved}
}
\value{
A matrix. Row names always indicate Station IDs. If
\code{sessionCol} and/or \code{cameraCol} are defined, they are contained in
the row names also (camera ID only if \code{byCamera = TRUE}). Column names
are dates. \cr Legend: NA: camera(s) not set up, 0: camera(s) not
operational, 1 (or higher): number of operational camera(s) or an indicator
for whether the station was operational (depending on
\code{camerasIndependent} and \code{allCamsOn})
}
\description{
Construct a matrix of daily camera trap station operation for use in
\code{\link{detectionHistory}} and \code{\link{spatialDetectionHistory}},
where it is needed for calculating trapping effort per occasion. It is also
used in \code{\link{surveyReport}} to calculate the number of trap nights
durig a survey. If several cameras were deployed per station, the matrix can
contain camera- or station-specific trap operation information, or
information about sessions during repeated surveys.
}
\details{
\code{cameraCol} is NULL by default, meaning the function assumes there was
1 camera per station in \code{CTtable}. If more than 1 camera was deployed
per station, \code{cameraCol} needs to be specified to identify individual
cameras within a station. Likewise, \code{sessionCol} can be used to if
camera trap stations were operated during multiple sessions / trapping
seasons.

\code{dateFormat} defaults to "YYYY-MM-DD", e.g. "2014-10-31", but can be
any other date format or date-time also. It can be specified either in the
format required by \code{\link[base]{strptime}} or the 'orders' argument in
\code{\link[lubridate]{parse_date_time}} in \pkg{lubridate}. In the example
above, "YYYY-MM-DD" would be specified as "\%Y-\%m-\%d" in base R or "ymd"
in \pkg{lubridate}.

Since version 2.1, dateFormat can be a date-time. That makes it possible to
specify the exact time cameras were set up / retrieved / malfunctioned /
worked again. This information is used to calculate the daily trapping
effort more precisely on days with incomplete effort.

Previously, setup and retrival day were counted as 1, indicating a whole day
of effort on those days. Since version 2.1, setup and retrieval are assumed
to have happened at 12 noon (resulting in daily effort of 0.5 instead of 1).
Users can also specify the exact time cameras were set up (by providing a
date-time in the setup / retrieval / problem columns). See vignette 3 for
more details.

If \code{hasProblems} is TRUE, the function tries to find columns
\code{ProblemX_from} and \code{ProblemX_to} in \code{CTtable}. \code{X} is a
consecutive number from 1 to n, specifying periods in which a camera or
station was not operational. If \code{hasProblems} is FALSE, cameras are
assumed to have been operational uninterruptedly from setup to retrieval
(see \code{\link{camtraps}} for details).

\code{allCamsOn} only has an effect if there was more than 1 camera at a
station. If TRUE, for the station to be considered operational, all cameras
at a station need to be operational. If FALSE, at least 1 active camera
renders the station operational. Argument \code{camerasIndependent} defines
if cameras record animals independently (it thus only has an effect if there
was more than 1 camera at a station). This is the case if an observation at
one camera does not increase the probability for detection at another camera
(cameras face different trails at a distance of one another).
Non-independence occurs if an animal is likely to trigger both camers (as
would be the case with 2 cameras facing each other).

If \code{camerasIndependent} is TRUE, 2 active cameras at a station will
result in a station operation value of 2 in the resulting matrix, i.e., 2
independent trap days at 1 station and day. If \code{camerasIndependent} is
FALSE, 2 active cameras will return value 1, i.e., 1 trap night at 1 station
per day.

Row names depend on the input arguments and contain the station name and
potentially session and camera names (if \code{sessionCol} and/or
\code{cameraCol} are defined).

Naming convention is (since version 1.2) \bold{Bold} information are from
the columns \code{stationCol}, \code{sessionCol} and \code{cameraCol} in
\code{CTtable}:

\bold{Station} \cr \bold{Station}__SESS_\bold{SessionID} \cr
\bold{Station}__CAM_\bold{CameraID} \cr
\bold{Station}__SESS_\bold{SessionID}__CAM_\bold{CameraID}

Session are designated with prefix "__SESS_", cameras with prefix "__CAM_".
Therefore, these are reserved words and may not be part of station, session
or camera names. Here's what it may look like in real life:

\bold{Station1} \cr \bold{Station1}__SESS_\bold{2019} \cr
\bold{Station1}__CAM_\bold{1024152} \cr
\bold{Station1}__SESS_\bold{2019}__CAM_\bold{1024152}

Functions detectionHistory and spatialDetectionHistory recognize these and
use the information accordingly.
}
\note{
Setting \code{camerasIndependent} according to the sampling situation
is important for the functions \code{\link{detectionHistory}} and
\code{\link{spatialDetectionHistory}}, if sampling effort (the number of
active trap nights in a occasion) is to be computed and returned.
}
\examples{


data(camtraps)

# no problems/malfunction
camop_no_problem <- cameraOperation(CTtable      = camtraps,
                                    stationCol   = "Station",
                                    setupCol     = "Setup_date",
                                    retrievalCol = "Retrieval_date",
                                    writecsv     = FALSE,
                                    hasProblems  = FALSE,
                                    dateFormat   = "dmy"
)

# with problems/malfunction
camop_problem <- cameraOperation(CTtable      = camtraps,
                                 stationCol   = "Station",
                                 setupCol     = "Setup_date",
                                 retrievalCol = "Retrieval_date",
                                 writecsv     = FALSE,
                                 hasProblems  = TRUE,
                                 dateFormat   = "dmy"
)

# The examples above specified dateFormat using lubdridate package, which is more intuitive. 
# Alternatively one can used in base-R date conversions (strptime) as below:

# with problems/malfunction / dateFormat in strptime format
camop_problem_oldformat <- cameraOperation(CTtable      = camtraps,
                                           stationCol   = "Station",
                                           setupCol     = "Setup_date",
                                           retrievalCol = "Retrieval_date",
                                           writecsv     = FALSE,
                                           hasProblems  = TRUE,
                                           dateFormat   = "\%d/\%m/\%Y"
)

camop_no_problem
camop_problem
camop_problem_oldformat

}
\author{
Juergen Niedballa
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/activityDensity.R
\name{activityDensity}
\alias{activityDensity}
\title{Plot kernel density estimation of single-species activity}
\usage{
activityDensity(
  recordTable,
  species,
  allSpecies = FALSE,
  speciesCol = "Species",
  recordDateTimeCol = "DateTimeOriginal",
  recordDateTimeFormat = "ymd HMS",
  plotR = TRUE,
  writePNG = FALSE,
  plotDirectory,
  createDir = FALSE,
  pngMaxPix = 1000,
  add.rug = TRUE,
  ...
)
}
\arguments{
\item{recordTable}{data.frame. the record table created by
\code{\link{recordTable}}}

\item{species}{Name of the species for which to create an kernel density
plot of activity}

\item{allSpecies}{logical. Create plots for all species in \code{speciesCol}
of \code{recordTable}? Overrides argument \code{species}}

\item{speciesCol}{character. name of the column specifying species names in
\code{recordTable}}

\item{recordDateTimeCol}{character. name of the column specifying date and
time in \code{recordTable}}

\item{recordDateTimeFormat}{character. format of column
\code{recordDateTimeCol} in \code{recordTable}}

\item{plotR}{logical. Show plots in R graphics device?}

\item{writePNG}{logical. Create pngs of the plots?}

\item{plotDirectory}{character. Directory in which to create png plots if
\code{writePNG = TRUE}}

\item{createDir}{logical. Create \code{plotDirectory} if \code{writePNG =
TRUE}?}

\item{pngMaxPix}{integer. image size of png (pixels along x-axis)}

\item{add.rug}{logical. add a rug to the plot?}

\item{\dots}{additional arguments to be passed to function
\code{densityPlot}}
}
\value{
Returns invisibly a vector of species record observation times in
radians, i.e. scaled to [\eqn{0, 2\pi}]. If \code{allSpecies == TRUE}, all
species' vectors are returned in an invisible named list.
}
\description{
The function plots a kernel density estimation of species diel activity
using function \code{\link[overlap]{densityPlot}} from package
\pkg{overlap}.
}
\details{
\code{species} must be in the \code{speciesCol} of \code{recordTable}.

\code{recordDateTimeFormat} defaults to the "YYYY-MM-DD HH:MM:SS"
convention, e.g. "2014-09-30 22:59:59". \code{recordDateTimeFormat} can be
interpreted either by base-R via \code{\link[base]{strptime}} or in
\pkg{lubridate} via \code{\link[lubridate]{parse_date_time}} (argument
"orders"). \pkg{lubridate} will be used if there are no "\%" characters in
\code{recordDateTimeFormat}.

For "YYYY-MM-DD HH:MM:SS", \code{recordDateTimeFormat} would be either
"\%Y-\%m-\%d \%H:\%M:\%S" or "ymd HMS". For details on how to specify date
and time formats in R see \code{\link[base]{strptime}} or
\code{\link[lubridate]{parse_date_time}}.
}
\examples{

if(requireNamespace("overlap")) {
# load record table
data(recordTableSample)

species4activity <- "VTA"    # = Viverra tangalunga, Malay Civet

activityDensity(recordTable = recordTableSample,
                species     = species4activity)


# all species at once

activityDensity(recordTable = recordTableSample,
                allSpecies  = TRUE,
                writePNG    = FALSE,
                plotR       = TRUE,
                add.rug     = TRUE)
 }

}
\references{
Martin Ridout and Matthew Linkie (2009). Estimating overlap of
daily activity patterns from camera trap data. Journal of Agricultural,
Biological and Environmental Statistics, 14(3), 322-337 \cr Mike Meredith
and Martin Ridout (2018). overlap: Estimates of coefficient of overlapping
for animal activity patterns. R package version 0.3.2.
\url{https://CRAN.R-project.org/package=overlap}
}
\seealso{
\code{\link{activityHistogram}}, \code{\link{activityRadial}},
\code{\link{activityOverlap}}
\url{https://www.kent.ac.uk/smsas/personal/msr/overlap.html}
}
\author{
Juergen Niedballa
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/addToPath.R
\name{addToPath}
\alias{addToPath}
\title{Add a directory to PATH temporarily}
\usage{
addToPath(directory)
}
\arguments{
\item{directory}{character. the directory in the file system to add to PATH 
(e.g. the directory containing exiftool.exe).}
}
\value{
invisible logical indicating whether \code{directory} was added to
PATH successfully (in the running R process).
}
\description{
Temporarily adds a directory to the environmental variable PATH for system
calls from within R. This allows Windows users to store exiftool.exe
anywhere on their hard drive and is useful if they cannot store the file 
in system directories. It is not needed on Linux or MacOS machines.
}
\details{
Several functions within this package depend on ExifTool. Under Windows,
exiftool.exe cannot be used if it is not in a directory path specified in
PATH. This can be solved by adding the directory containing exiftool.exe for
temporary use within the running R process. It can also be useful in other 
contexts besides Exiftool.
}
\note{
The directories in PATH can be queried by \code{Sys.getenv("PATH")}.
}
\examples{

exiftool_dir <- "C:/Path/To/Exiftool"
addToPath(directory = exiftool_dir)

# check if it has been added to PATH
grepl(exiftool_dir,  Sys.getenv("PATH"))

}
\author{
Juergen Niedballa
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/createStationFolders.R
\name{createStationFolders}
\alias{createStationFolders}
\title{Create camera trap station directories for raw camera trap images}
\usage{
createStationFolders(inDir, stations, cameras, createinDir)
}
\arguments{
\item{inDir}{character. Directory in which station directories are to be
created}

\item{stations}{character. Station IDs to be used as directory names within
\code{inDir}}

\item{cameras}{character. Camera trap IDs to be used as subdirectory names
in each station directory (optionally)}

\item{createinDir}{logical. If inDir does not exist, create it?}
}
\value{
A \code{data.frame} with station (and possibly camera) directory
names and an indicator for whether they were created successfully.
}
\description{
This function creates camera trap station directories, if needed with camera
subdirectories. They can be used as an initial directory structure for
storing raw camera trap images.
}
\details{
The empty directories serve as containers for saving raw camera trap images.
If more than 1 camera was set up at a station, specifying \code{cameras} is
required in order to keep images from different cameras separate. Otherwise,
generic filenames (e.g., IMG0001.JPG) from different cameras may lead to
accidental overwriting of images if images from these cameras are saved in
one station directory.
}
\examples{

\dontrun{

# create dummy directory for tests (this will be used as inDir)
# (normally, you'd set up an empty directory, e.g. .../myStudy/rawImages)
wd_createStationDir <- file.path(tempdir(), "createStationFoldersTest")

# now we load the sample camera trap station data frame
data(camtraps)

# create station directories in wd_createStationDir
StationFolderCreate1 <- createStationFolders (inDir       = wd_createStationDir,
                                              stations    = as.character(camtraps$Station),
                                              createinDir = TRUE)
  
StationFolderCreate1

# check if directories were created
list.dirs(wd_createStationDir)

}

}
\author{
Juergen Niedballa
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/OCRdataFields.R
\name{OCRdataFields}
\alias{OCRdataFields}
\title{Optical character recognition (OCR) from data fields in digital images}
\usage{
OCRdataFields(inDir, geometries, invert = FALSE)
}
\arguments{
\item{inDir}{character. Directory containing camera trap images (or
subdirectories containing images)}

\item{geometries}{list. A (possibly named) list of geometry strings defining
the image area(s) to extract.}

\item{invert}{logical. Invert colors in the image? Set to TRUE if text in
data field is white on black background. Leave if FALSE if text is black in
white background.}
}
\value{
A \code{data.frame} with original directory and file names, and
additional columns for the OCR data of each extracted geometry.
}
\description{
Extracts information from the data fields in camera trap images (not the
metadata). Many camera traps include data fields in camera trap images,
often including date and time of images, and sometimes other information.
This function extracts the information from these fields using optical
character recognition provided by the package \pkg{tesseract} after reading
images using the package \pkg{magick}.
}
\details{
Normally all these information should be in the image metadata. This
function is meant as a last resort if image metadata are unreadable or were
removed from images. OCR is not perfect and may misidentify characters, so
check the output carefully.

The output of this function can be used in
\code{\link{writeDateTimeOriginal}} to write date/time into the
DateTimeOriginal tag in image metadata, making these images available for
automatic processing with \code{\link{recordTable}} and other functions that
extract image metadata.


This function reads all images in inDir (including subdirectories), crops
them to the geometries in the "geometries" list, and performs optical
character recognition (OCR) on each of these fields (leveraging the magick
and tesseract packages).

Geometries are defined with \code{geometry_area} from \pkg{magick}. See
\code{\link[magick]{geometry}} for details on how to specify geometries with
\code{geometry_area}. The format is: "widthxheight+x_off+y_off", where:

\describe{ \item{width}{width of the area of interest} \item{height}{height
of the area of interest} \item{x_off}{offset from the left side of the
image} \item{y_off}{offset from the top of the image} }

Units are pixels for all fields. digiKam can help in identifying the correct
specification for geometries. Open the Image Editor, left-click and draw a
box around the data field of interest. Ensure the entire text field is
included inside the box, but nothing else. Now note two pairs of numbers at
the bottom of the window, showing the offsets and box size as e.g.:

"(400, 1800) (300 x 60)"

This corresponds to the geometry values as follows:

"(x_off, y_off) (width x height)"

Using these values, you'd run:

\code{geometry_area(x_off = 400, y_off = 1800, width = 300, height = 60)}

and receive

"300x60+400+1800"

as your geometry.

OCR in tesseract has problems with white font on black background. If that
is the case in your images, set \code{invert} to \code{TRUE} to invert the
image and ensure OCR uses black text on white background.

Even then, output will not be perfect. Error rates in OCR depend on multiple
factors, including the text size and font type used. We don't have control
over these, so check the output carefully and edit as required.
}
\examples{


\dontrun{
# dontrun is to avoid forcing users to install additional dependencies

wd_images_OCR <- system.file("pictures/full_size_for_ocr", package = "camtrapR")

library(magick)

# define geometries
geometry1 <- geometry_area(x_off = 0, y_off = 0, width = 183, height = 37)
geometry2 <- geometry_area(x_off = 196, y_off = 0, width = 200, height = 17)
geometry3 <- geometry_area(x_off = 447, y_off = 0, width = 63, height = 17)
geometry4 <- geometry_area(x_off = 984, y_off = 0, width = 47, height = 17)
geometry5 <- geometry_area(x_off = 0, y_off = 793, width = 320, height = 17)

# combine geometries into list
geometries <- list(date = geometry1, 
                   time = geometry2, 
                   sequence_id = geometry3,
                   temperature = geometry4,
                   camera_model = geometry5)

df_image_data  <- OCRdataFields(inDir = wd_images_OCR,
                                geometries = geometries, 
                                invert = TRUE)       
df_image_data

# note the mistake in "camera_model"
# it should be "PC850", not "PC8S0O"
# date and time are correct though
                   
}



}
\seealso{
\code{\link{writeDateTimeOriginal}}
}
\author{
Juergen Niedballa
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/surveyDashboard.R
\name{surveyDashboard}
\alias{surveyDashboard}
\title{Survey Dashboard for Camera Trap Data Analysis}
\usage{
surveyDashboard(
  CTtable = NULL,
  recordTable = NULL,
  stationCol = NULL,
  cameraCol = NULL,
  xcol = NULL,
  ycol = NULL,
  crs = NULL,
  setupCol = NULL,
  retrievalCol = NULL,
  hasProblems = FALSE,
  CTdateFormat = "ymd",
  camerasIndependent = NULL,
  speciesCol = "Species",
  recordDateTimeCol = "DateTimeOriginal",
  recordDateTimeFormat = "ymd HMS",
  timeZone = "UTC",
  exclude = NULL
)
}
\arguments{
\item{CTtable}{A data.frame containing the camera trap deployment information.}

\item{recordTable}{A data.frame containing the camera trap records.}

\item{stationCol}{The column name containing the camera trap station ID}

\item{cameraCol}{The column name containing the camera trap IDs (optional, only if 2 or more cameras per station)}

\item{xcol}{The column name containing the X coordinate of the camera trap station.}

\item{ycol}{The column name containing the Y coordinate of the camera trap station.}

\item{crs}{The coordinate reference system (CRS) of the camera trap data. Must be a valid argument to \code{\link[sf]{st_crs}}}

\item{setupCol}{The column name containing the camera trap deployment date (and time).}

\item{retrievalCol}{The column name containing the camera trap retrieval date (optionally date-time).}

\item{hasProblems}{A logical indicating whether there are periods of cameras malfunctioning}

\item{CTdateFormat}{The date format of the camera trap deployment and retrieval date and time (default: "ymd").}

\item{camerasIndependent}{logical. If multiple camera per station, are they independent?}

\item{speciesCol}{The column name containing the species names}

\item{recordDateTimeCol}{The column name containing the record date and time}

\item{recordDateTimeFormat}{The date/time format of \code{recordDateTimeCol}}

\item{timeZone}{Time zone of records in recordTable}

\item{exclude}{Species to be excluded from the data set}
}
\value{
A Shiny dashboard application for camera trap survey data analysis
}
\description{
A comprehensive Shiny dashboard for analyzing camera trap survey data. The dashboard provides 
interactive visualization, data exploration, and analysis tools including:

\itemize{
  \item Data import from CSV files, Wildlife Insights exports, or camtrapDP format
  \item Interactive maps for camera locations and species detections
  \item Species activity pattern analysis
  \item Covariate extraction and analysis tools
  \item Single-species and community occupancy modeling
  \item Spatial prediction capabilities
}
}
\details{
The dashboard includes several major components:

\strong{Data Import & Management:}
\itemize{
  \item CSV file import with column mapping
  \item Wildlife Insights data import (zip, CSV, or directory)
  \item camtrap DP data import
  \item Study area import from shapefile
  \item Save/restore functionality for app state
  \item Export functionality to save data from dashboard
}

\strong{Data Processing:}
\itemize{
  \item Flexible station filtering with multiple criteria
  \item Temporal record filtering with independence criteria
  \item Filtering species records by species name
  \item Automated covariate extraction from local rasters or online 
  elevation data
  \item Covariate correlation analysis with visualization
  \item Species accumulation curves
}

\strong{Basic Analysis:}
\itemize{
  \item Basic summary statistics
  \item Interactive overview and species detection maps
  \item Activity pattern analysis (single species and two-species overlap)
  \item Camera operation visualization
}

\strong{Occupancy Modeling:}
\itemize{
  \item Basic workflow for simple model specification (linear effects)
  \item Support for both unmarked and ubms packages
  \item Automated detection history creation
  \item Model comparison and selection
  \item Response curves and spatial predictions
}

\strong{Community Occupancy Modeling:}
\itemize{
  \item Flexible species selection with filtering
  \item Support for fixed, random, and independent effects
  \item Species-site random effects
  \item Effort handling on detection
  \item MCMC diagnostics and convergence assessment
  \item Species occupancy, richness and PAO predictions
}
}
\note{
\itemize{
  \item Interactive maps with multiple basemap options
  \item Covariate scaling is performed automatically if requested (includes
   automatic scaling of prediction rasters)
  \item The app state can be saved and restored
}

Current limitations include:

- supports only single-season data
- no support for spatial capture-recapture models (or anything related to individual IDs)
}
\examples{

\dontrun{

# Start the dashboard without parameters
# This opens the application with a welcome screen where data can be imported
surveyDashboard()

# Basic usage with minimal parameters

data("camtraps")
data("recordTableSample")

  surveyDashboard(
    CTtable = camtraps,
    recordTable = recordTableSample,
    xcol = "utm_x",
    ycol = "utm_y",
    crs = "epsg:32650",      # = UTM50N
    stationCol = "Station",
    setupCol = "Setup_date",
    retrievalCol = "Retrieval_date",
    CTdateFormat = "dmy"
  )
  }
  
  

}
\author{
Juergen Niedballa
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/addCopyrightTag.R
\name{addCopyrightTag}
\alias{addCopyrightTag}
\title{Write a copyright tag into JPEG image metadata}
\usage{
addCopyrightTag(
  inDir,
  copyrightTag,
  askFirst = TRUE,
  keepJPG_original = TRUE,
  ignoreMinorErrors = FALSE
)
}
\arguments{
\item{inDir}{character. Name of the directory containing camera trap images.}

\item{copyrightTag}{character. The tag to be written into the Exif Copyright
field}

\item{askFirst}{logical. Ask user to confirm before execution?}

\item{keepJPG_original}{logical. Keep original JPG files as .JPG_original
files (TRUE) or overwrite JPGs (FALSE)?}

\item{ignoreMinorErrors}{logical. Ignore minor errors that would cause the
function to fail (set TRUE for images with bad MakerNotes, observed in
Panthera V4 cameras)}
}
\value{
An invisible list of Exiftool output.

More importantly, the specified copyright tag is written into the Copyright
field of the Exif metadata of all images in inDir.
}
\description{
This function writes a copyright tag into the copyright field of JPEG image
Exif metadata. It does so recursively, so it works both for images that are
sorted into subdirectories and unsorted images. Note that all images in
subdirectories of inDir will be tagged. It is not required to run this
function in the camtrapR workflow, but may be desired for data sharing or
publishing.
}
\details{
If askFirst = TRUE, the function will show a menu and asks the user to
confirm the action before execution. Type "1" to write copyright tags and
"2" to abort.

By default Exiftool creates a copy of each JPG image and preserves the
original images (without the copyright tag) as .JPG_original files. Note
that this behaviour will instantly double the number of images in inDir and
the disk space required. If this is not desired, set keepJPG_original =
FALSE.

\code{ignoreMinorErrors} is useful if copyright tags can't be updated
correctly. This can be caused by bad MakerNotes and so far was only observed
in Panthera V4 cameras. In that case, set \code{ignoreMinorErrors} to
\code{TRUE}. This will add the "-m" option to the Exiftool call, thereby
ignoring minor errors and warnings and assigning the copyright tag
regardless.
}
\examples{


\dontrun{

if (Sys.which("exiftool") != ""){        # only run this example if ExifTool is available

# copy sample images to temporary directory (so we don't mess around in the package directory)
wd_images_ID <- system.file(file.path("pictures", "sample_images_species_dir"), 
                            package = "camtrapR")
file.copy(from = wd_images_ID, to = tempdir(), recursive = TRUE)
wd_images_ID_copy <- file.path(tempdir(), "sample_images_species_dir")

# define a sample tag
copyrightTagToAdd <- "Your Name (Your Organisation)"

# add the tag to the images
addCopyrightTag(inDir        = wd_images_ID_copy, 
                copyrightTag = copyrightTagToAdd)
1     # we choose "YES", i.e., we want to add a copyright tag
                
# you can check the outcome with function exifTagNames

metadat <- exifTagNames(wd_images_ID_copy)
metadat [metadat$tag_name == "Copyright",]
}


}

}
\author{
Juergen Niedballa
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/checkSpeciesIdentification.R
\name{checkSpeciesIdentification}
\alias{checkSpeciesIdentification}
\title{Consistency check on species image identification}
\usage{
checkSpeciesIdentification(
  inDir,
  IDfrom,
  hasCameraFolders,
  metadataSpeciesTag,
  metadataSpeciesTagToCompare,
  metadataHierarchyDelimitor = "|",
  maxDeltaTime,
  excludeSpecies,
  stationsToCheck,
  writecsv = FALSE
)
}
\arguments{
\item{inDir}{character. Directory containing identified camera trap images
sorted into station subdirectories (e.g. inDir/StationA/)}

\item{IDfrom}{character. Read species ID from image metadata ("metadata") of
from species directory names ("directory")?}

\item{hasCameraFolders}{logical. Do the station directories in \code{inDir}
have camera subdirectories (e.g. "inDir/StationA/Camera1" or
"inDir/StationA/Camera1/Species1")?}

\item{metadataSpeciesTag}{character. The species ID tag name in image
metadata (if IDfrom = "metadata").}

\item{metadataSpeciesTagToCompare}{character. A second species ID tag name
in image metadata (if IDfrom = "metadata"). For comparing double observer
species identification.}

\item{metadataHierarchyDelimitor}{character. The character delimiting
hierarchy levels in image metadata tags in field "HierarchicalSubject".
Either "|" or ":"}

\item{maxDeltaTime}{numeric. Maximum time interval between images to be
returned (in seconds)}

\item{excludeSpecies}{character. vector of species to exclude from checks}

\item{stationsToCheck}{character. vector of stations to be checked
(optionally)}

\item{writecsv}{logical. Should the resulting data.frame be saved as a .csv?}
}
\value{
A \code{list} containing 2 data frames. The first contains a data
frame with images file names, directories, time stamp and species ID that
were taken within \code{maxDeltaTime} seconds of another species image at a
particular station. The second data frame contains images with conflicting
species IDs (if \code{IDfrom = "metadata"} and
\code{metadataSpeciesTagToCompare} is defined)
}
\description{
This function serves 2 purposes: 1) it assesses possible misidentification
of species and 2) compares double observer species identification (only if
metadata tagging was used for species identification).
}
\details{
Within each station, it assesses whether there are images of a species taken
within a given time interval of another species. Often, it is unlikely that
different species are encountered within a very short time intervals at the
same location. This type of misidentification can arise easily if some
images belonging to a sequence of images were accidentally moved into
different species directories or tagged incorrectly.

Double observer identification may be desirable to increase reliability of
species identification. The function returns conflicts in species
identification between 2 observers. These conflicts can then be corrected.

Images may accidentally be misidentified by assigning wrong species tags or
by moving them into wrong species directories. Imagine your cameras take
sequences of images each time they are triggered and one image of the
sequence is misidentified. The time difference between these images (that
have different species assigned to them) will be very small, usually a few
seconds. This function will return all these images for you to check if they
were identified correctly.

If multiple observers identify images independently using metadata tagging,
their identifications can be compared by setting
\code{metadataSpeciesTagToCompare}. Conflicting or missing identifications
will be reported. This feature is only available if images were identified
by metadata tagging.

Species like "blank" or "team" can be ignored using \code{excludeSpecies}.
If only specific stations are to be checked, \code{stationsToCheck} can be
set.
}
\note{
The function will not be able to find "isolated" images, i.e. images
that were misidentified, but were not part of a sequence of images.
Likewise, if all images of a sequence were misidentified, they cannot be
found either. From version 0.99.0, the function can also handle images
identied with metadata tags.
}
\examples{


\donttest{
wd_images_ID <- system.file("pictures/sample_images_species_dir", package = "camtrapR")

if (Sys.which("exiftool") != ""){        # only run this example if ExifTool is available
check.folders <- checkSpeciesIdentification(inDir             = wd_images_ID,
                                            IDfrom            = "directory",
                                            hasCameraFolders  = FALSE,
                                            maxDeltaTime      = 120,
                                            writecsv          = FALSE)
                                     
check.folders   # In the example, 2 different species were photographed within 2 minutes. 
}
}

\dontrun{
# now exclude one of these 2 species 
check.folders2 <- checkSpeciesIdentification(inDir             = wd_images_ID,
                                             IDfrom            = "directory",
                                             hasCameraFolders  = FALSE,
                                             maxDeltaTime      = 120,
                                             excludeSpecies    = "EGY",
                                             writecsv          = FALSE)           
  
check.folders2   # the data frame is empty

# now we check only one station
check.folders3 <- checkSpeciesIdentification(inDir             = wd_images_ID,
                                             IDfrom            = "directory",
                                             hasCameraFolders  = FALSE,
                                             maxDeltaTime      = 120,
                                             stationsToCheck   = "StationB",
                                             writecsv          = FALSE)
check.folders3   # the data frame is empty
}


}
\author{
Juergen Niedballa
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/activityOverlap.R
\name{activityOverlap}
\alias{activityOverlap}
\title{Plot overlapping kernel densities of two-species activities}
\usage{
activityOverlap(
  recordTable,
  speciesA,
  speciesB,
  speciesCol = "Species",
  recordDateTimeCol = "DateTimeOriginal",
  recordDateTimeFormat = "ymd HMS",
  plotR = TRUE,
  writePNG = FALSE,
  addLegend = TRUE,
  legendPosition = "topleft",
  plotDirectory,
  createDir = FALSE,
  pngMaxPix = 1000,
  add.rug = TRUE,
  overlapEstimator = c("Dhat1", "Dhat4", "Dhat5"),
  ...
)
}
\arguments{
\item{recordTable}{data.frame. the record table created by
\code{\link{recordTable}}}

\item{speciesA}{Name of species 1 (as found in \code{speciesCol} of
recordTable)}

\item{speciesB}{Name of species 2 (as found in \code{speciesCol} of
recordTable)}

\item{speciesCol}{character. name of the column specifying species names in
\code{recordTable}}

\item{recordDateTimeCol}{character. name of the column specifying date and
time in \code{recordTable}}

\item{recordDateTimeFormat}{character. format of column
\code{recordDateTimeCol} in \code{recordTable}}

\item{plotR}{logical. Show plots in R graphics device?}

\item{writePNG}{logical. Create pngs of the plots?}

\item{addLegend}{logical. Add a legend to the plots?}

\item{legendPosition}{character. Position of the legend (keyword)}

\item{plotDirectory}{character. Directory in which to create png plots if
\code{writePNG = TRUE}}

\item{createDir}{logical. Create \code{plotDirectory}?}

\item{pngMaxPix}{integer. image size of png (pixels along x-axis)}

\item{add.rug}{logical. add a rug to the plot?}

\item{overlapEstimator}{character. Which overlap estimator to return (passed
on to argument \code{type} in \code{\link[overlap]{overlapEst}})}

\item{\dots}{additional arguments to be passed to function
\code{\link[overlap]{overlapPlot}}}
}
\value{
Returns invisibly the \code{data.frame} with plot coordinates
returned by \code{\link[overlap]{overlapPlot}}.
}
\description{
This function plots kernel density estimates of two species' diel activity
data by calling the function \code{\link[overlap]{overlapPlot}} from package
\pkg{overlap}. It further computes the overlap coefficient Dhat1 by calling
\code{\link[overlap]{overlapEst}}.
}
\details{
\code{...} can be graphical parameters passed on to function
\code{\link[overlap]{overlapPlot}}, e.g. \code{linetype}, \code{linewidth},
\code{linecol} (see example below).

\code{recordDateTimeFormat} defaults to the "YYYY-MM-DD HH:MM:SS"
convention, e.g. "2014-09-30 22:59:59". \code{recordDateTimeFormat} can be
interpreted either by base-R via \code{\link[base]{strptime}} or in
\pkg{lubridate} via \code{\link[lubridate]{parse_date_time}} (argument
"orders"). \pkg{lubridate} will be used if there are no "\%" characters in
\code{recordDateTimeFormat}.

For "YYYY-MM-DD HH:MM:SS", \code{recordDateTimeFormat} would be either
"\%Y-\%m-\%d \%H:\%M:\%S" or "ymd HMS". For details on how to specify date
and time formats in R see \code{\link[base]{strptime}} or
\code{\link[lubridate]{parse_date_time}}.
}
\note{
Please be aware that the function (like the other activity... function
of this package) use clock time, not solar time. If your survey was long
enough to see changes in sunrise and sunset times, this may result in biased
representations of species activity.
}
\examples{

if(requireNamespace("overlap")) {

# load record table
data(recordTableSample)

# define species of interest
speciesA_for_activity <- "VTA"    # = Viverra tangalunga, Malay Civet
speciesB_for_activity <- "PBE"    # = Prionailurus bengalensis, Leopard Cat

# create activity overlap plot (basic)
activityOverlap (recordTable = recordTableSample,
                 speciesA    = "VTA",    # = Viverra tangalunga, Malay Civet
                 speciesB    = "PBE",    # = Prionailurus bengalensis, Leopard Cat
                 writePNG    = FALSE,
                 plotR       = TRUE
)


                                     
# create activity overlap plot (prettier and with some overlapPlot arguments set)

activityOverlap (recordTable = recordTableSample,
                 speciesA    = speciesA_for_activity,
                 speciesB    = speciesB_for_activity,
                 writePNG    = FALSE,
                 plotR       = TRUE,
                 createDir   = FALSE,
                 pngMaxPix   = 1000,
                 linecol     = c("black", "blue"),
                 linewidth   = c(5,3),
                 linetype    = c(1, 2),
                 olapcol     = "darkgrey",
                 add.rug     = TRUE,
                 extend      = "lightgrey",
                 ylim        = c(0, 0.25),
                 main        = paste("Activity overlap between ", 
                                     speciesA_for_activity, "and", 
                                     speciesB_for_activity)
)
}

}
\references{
Mike Meredith and Martin Ridout (2018). overlap: Estimates of
coefficient of overlapping for animal activity patterns. R package version
0.3.2. \url{https://CRAN.R-project.org/package=overlap} \cr Ridout, M.S. and
Linkie, M. (2009) Estimating overlap of daily activity patterns from camera
trap data. Journal of Agricultural, Biological and Environmental Statistics,
14, 322-337.
}
\seealso{
\code{\link{activityDensity}} \cr
\url{https://www.kent.ac.uk/smsas/personal/msr/overlap.html}
}
\author{
Juergen Niedballa
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/createSpeciesFolders.R
\name{createSpeciesFolders}
\alias{createSpeciesFolders}
\title{Create species directories for species identification}
\usage{
createSpeciesFolders(inDir, hasCameraFolders, species, removeFolders = FALSE)
}
\arguments{
\item{inDir}{character. Directory containing camera trap images sorted into
station subdirectories (e.g. inDir/StationA/)}

\item{hasCameraFolders}{logical. Do the station directories in \code{inDir}
have camera-subdirectories (e.g. inDir/StationA/CameraA1;
inDir/StationA/CameraA2)?}

\item{species}{character. names of species directories to be created in
every station (or station/camera) subdirectory of \code{inDir}}

\item{removeFolders}{logical. Indicating whether to create (TRUE) or remove
(FALSE) species directories .}
}
\value{
A \code{data.frame} with directory names and an indicator for
whether directories were created or deleted.
}
\description{
This function creates species subdirectories within station directories.
They can be used for species identification by manually moving images into
the respective species directories. The function can also delete empty
species directories (if species were not detected at sites). It is not
necessary to run this function if animals will be identified by metadata
tagging.
}
\details{
This function should be run after \code{\link{imageRename}}. Empty
directories can be created as containers for species identification if
images are identified with the drag & drop method. After species
identification is complete, empty species directories can be deleted using
\code{removeFolders = TRUE}. The function will delete only directories which
are specified in \code{species}. If \code{hasCameraFolders} was set to
\code{TRUE} in function \code{\link{imageRename}}, \code{hasCameraFolders}
must be set to \code{TRUE} here too. Species directories will then be
created within each camera subdirectory of each station directory. if the
user wishes to identify species by metadata tagging, running this function
is not needed.
}
\examples{

\dontrun{

# create dummy directories for tests
# (normally, you'd use directory containing renamed, unsorted images)

# this will be used as inDir
wd_createDirTest <- file.path(getwd(), "createSpeciesFoldersTest")

# now we create 2 station subdirectories
dirs_to_create <- file.path(wd_createDirTest, c("StationA", "StationB"))
sapply(dirs_to_create, FUN = dir.create, recursive = TRUE)

# species names for which we want to create subdirectories
species <- c("Sambar Deer", "Bay Cat")

# create species subdirectories
SpecFolderCreate1 <- createSpeciesFolders (inDir               = wd_createDirTest,
                                           species             = species,
                                           hasCameraFolders = FALSE,
                                           removeFolders       = FALSE)
  
SpecFolderCreate1

# check if directories were created
list.dirs(wd_createDirTest)

# delete empty species directories
SpecFolderCreate2 <- createSpeciesFolders (inDir               = wd_createDirTest,
                                           species             = species,
                                           hasCameraFolders = FALSE,
                                           removeFolders       = TRUE)

SpecFolderCreate2

# check if species directories were deleted
list.dirs(wd_createDirTest)

}

}
\author{
Juergen Niedballa
}

\docType{data}
\name{timeShiftTable}
\alias{timeShiftTable}
\title{Sample camera trap time shift table}
\format{A data frame with 2 rows and 4 variables}
\usage{
data(timeShiftTable)
}
\description{
Sample camera trap time shift table
}
\details{
If image Exif metadata timestamps are wrong systematically (e.g. because camera system time was not set after changing batteries), it can be corrected using a \code{data.frame} in the following format using function \code{\link{timeShiftImages}}. For details on data format, please see \code{\link{timeShiftImages}}.

The variables are as follows:

\tabular{ll}{
\code{Station} \tab Camera trap station ID \cr
\code{camera} \tab Camera trap ID (optional) \cr
\code{timeshift} \tab time shift amount to be applied \cr
\code{sign} \tab direction of time shift \cr
}
}
\keyword{datasets}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/surveyReport.R
\name{surveyReport}
\alias{surveyReport}
\title{Create a report about a camera trapping survey and species detections}
\usage{
surveyReport(
  recordTable,
  CTtable,
  camOp,
  speciesCol = "Species",
  stationCol = "Station",
  cameraCol,
  setupCol,
  retrievalCol,
  CTDateFormat = "ymd",
  CTHasProblems = "deprecated",
  recordDateTimeCol = "DateTimeOriginal",
  recordDateTimeFormat = "ymd HMS",
  Xcol,
  Ycol,
  sinkpath,
  makezip
)
}
\arguments{
\item{recordTable}{data.frame containing a species record table as given by
\code{\link{recordTable}}}

\item{CTtable}{data.frame containing information about location and trapping
period of camera trap stations (equivalent to \code{\link{camtraps}}}

\item{camOp}{camera operation matrix created with
\code{\link{cameraOperation}}}

\item{speciesCol}{character. name of the column specifying Species ID in
\code{recordTable}}

\item{stationCol}{character. name of the column specifying Station ID in
\code{CTtable} and \code{recordTable}}

\item{cameraCol}{character. name of the column specifying Camera ID in
\code{CTtable} and \code{recordTable}}

\item{setupCol}{character. name of the column containing camera setup dates
in \code{CTtable}}

\item{retrievalCol}{character. name of the column containing camera
retrieval dates in \code{CTtable}}

\item{CTDateFormat}{character. The format of columns \code{setupCol} and
\code{retrievalCol} (and potential problem columns) in \code{CTtable}. Must
be interpretable by either \code{as.Date} or the "orders" argument
\code{\link[lubridate]{parse_date_time}} in \pkg{lubridate}.}

\item{CTHasProblems}{deprecated (since version 2.1)}

\item{recordDateTimeCol}{character. The name of the column containing date
and time of records in \code{recordTable}}

\item{recordDateTimeFormat}{character. The date/time format of column
\code{recordDateTimeCol} in \code{recordTable}.}

\item{Xcol}{character. name of the column specifying x coordinates in
\code{CTtable}. Used to create detection maps if \code{makezip} is TRUE.
(optional)}

\item{Ycol}{character. name of the column specifying y coordinates in
\code{CTtable}. Used to create detection maps if \code{makezip} is TRUE.
(optional)}

\item{sinkpath}{character. The directory into which the survey report is
saved (optional)}

\item{makezip}{logical. Create a zip file containing tables, plots and maps
in \code{sinkpath}?}
}
\value{
An invisible list containing 5 \code{data.frames}.

\item{survey_dates}{station and image date ranges, number of total and
active trap days (calendar days and taking into account independent effort
of multiple cameras, if applicable), number of cameras per station}
\item{species_by_station}{species numbers by station}
\item{events_by_species}{number of events and stations by species}
\item{events_by_station}{number of events for every species by station (only
species that were recorded)} \item{events_by_station2}{number of events for
all species at all stations (including species that were not recorded)}

The output will be saved to a .txt file if \code{sinkpath} is defined.

If \code{makezip} is TRUE, a zip file will be created in \code{sinkpath}. It
contains single-species activity plots, detection maps (if \code{Xcol} and
\code{Ycol} are defined), the survey report tables, the record table and the
camera trap station table, and an example R script.
}
\description{
This function creates a report about a camera trapping survey and species
records. It uses a camera trap station information table and a record table
(generated with \code{\link{recordTable}}) as input. Output tables can be
saved and a zip file for simple data sharing can be created easily.
}
\details{
\code{dateFormat} defaults to "YYYY-MM-DD", e.g. "2014-10-31". It can be
specified either in the format required by \code{\link[base]{strptime}} or
the 'orders' argument in \code{\link[lubridate]{parse_date_time}} in
\pkg{lubridate}. In the example above, "YYYY-MM-DD" would be specified as
"\%Y-\%m-\%d" or "ymd".

\code{recordDateTimeFormat} defaults to the "YYYY-MM-DD HH:MM:SS"
convention, e.g. "2014-09-30 22:59:59". \code{recordDateTimeFormat} can be
interpreted either by base-R via \code{\link[base]{strptime}} or in
\pkg{lubridate} via \code{\link[lubridate]{parse_date_time}} (argument
"orders"). \pkg{lubridate} will be used if there are no "\%" characters in
\code{recordDateTimeFormat}.

For "YYYY-MM-DD HH:MM:SS", \code{recordDateTimeFormat} would be either
"\%Y-\%m-\%d \%H:\%M:\%S" or "ymd HMS". For details on how to specify date
and time formats in R see \code{\link[base]{strptime}} or
\code{\link[lubridate]{parse_date_time}}.

Note: as of version 2.1, argument \code{CTHasProblems} is deprecated and
defunct. Please use \code{camOp} instead to provide information about
periods of camera activity and malfunction. If \code{camOp} is not provided
the legacy version of surveyReport (from camtrapR 2.0.3) will be run with a
warning.
}
\examples{

data(camtraps)
data(recordTableSample)

# since version 2.1, camera operation matrix is required as input

camop_no_problem <- cameraOperation(CTtable      = camtraps,
                                    stationCol   = "Station",
                                    setupCol     = "Setup_date",
                                    retrievalCol = "Retrieval_date",
                                    writecsv     = FALSE,
                                    hasProblems  = FALSE,
                                    dateFormat   = "dmy"
)

reportTest <- surveyReport (recordTable          = recordTableSample,
                            CTtable              = camtraps,
                            camOp                = camop_no_problem,
                            speciesCol           = "Species",
                            stationCol           = "Station",
                            setupCol             = "Setup_date",
                            retrievalCol         = "Retrieval_date",
                            CTDateFormat         = "dmy", 
                            recordDateTimeCol    = "DateTimeOriginal",
                            recordDateTimeFormat = "ymd HMS")

class(reportTest)  # a list with
length(reportTest) # 5 elements

reportTest[[1]]    # camera trap operation times and image date ranges
reportTest[[2]]    # number of species by station
reportTest[[3]]    # number of events and number of stations by species
reportTest[[4]]    # number of species events by station
reportTest[[5]]    # number of species events by station including 0s (non-observed species)

# with camera problems

camop_problem <- cameraOperation(CTtable      = camtraps,
                                 stationCol   = "Station",
                                 setupCol     = "Setup_date",
                                 retrievalCol = "Retrieval_date",
                                 writecsv     = FALSE,
                                 hasProblems  = TRUE,
                                 dateFormat   = "dmy"
)

reportTest_problem <- surveyReport (recordTable          = recordTableSample,
                                    CTtable              = camtraps,
                                    camOp                = camop_problem,
                                    speciesCol           = "Species",
                                    stationCol           = "Station",
                                    setupCol             = "Setup_date",
                                    retrievalCol         = "Retrieval_date",
                                    CTDateFormat         = "dmy", 
                                    recordDateTimeCol    = "DateTimeOriginal",
                                    recordDateTimeFormat = "ymd HMS")

reportTest_problem$survey_dates




\dontrun{
# run again with sinkpath defined
reportTest <- surveyReport (recordTable          = recordTableSample,
                            CTtable              = camtraps,
                            camOp                = camop_no_problem,
                            speciesCol           = "Species",
                            stationCol           = "Station",
                            setupCol             = "Setup_date",
                            retrievalCol         = "Retrieval_date",
                            CTDateFormat         = "dmy",, 
                            recordDateTimeCol    = "DateTimeOriginal",
                            recordDateTimeFormat = "ymd HMS",
                            sinkpath             = getwd())

# have a look at the text file
readLines(list.files(getwd(), pattern = paste("survey_report_", Sys.Date(), ".txt", sep = ""), 
					 full.names = TRUE))
}

}
\seealso{
\code{\link{recordTable}}
}
\author{
Juergen Niedballa
}

\name{camtrapR-package}
\alias{camtrapR-package}
\alias{camtrapR}

\docType{package}

\title{
Overview of the functions in the camtrapR package
}
\description{
This package provides a streamlined workflow for processing data generated in camera trap-based wildlife studies and prepares input for further analyses, particularly in occupancy and spatial capture-recapture frameworks.
It suggests a simple data structure and provides functions for managing digital camera trap photographs (and videos), generating record tables, maps of species richness and species detections and species activity diagrams. It further helps prepare subsequent analyses by creating detection/non-detection matrices for occupancy analyses, e.g. in the \pkg{unmarked} or \pkg{ubms} packages, and \code{capthist} objects for spatial capture-recapture analyses in the \pkg{secr} package. In addition, basic survey statistics are computed. The functions build on one another in a logical sequence.
The only manual input needed is species (and individual) identification, which is achieved by moving images into species directories or by tagging images in image management software. Besides, a table holding basic information about camera trap station IDs, locations and trapping periods must be created in spreadsheet software.
}

\section{User support}{
The camtrapR Google group is an online support and help forum for camtrapR users. You can find it here: \url{https://groups.google.com/forum/#!forum/camtrapr}.
}

\details{
Image metadata (such as date and time or user-assigned tags) are extracted from the images using Phil Harvey's ExifTool (available from \url{https://exiftool.org/}) and the information is stored in a record table. An adjustable criterion for temporal independence of records can be applied. Maps of species presence and species richness can be generated. Several functions are available for plotting single- and two-species activity patterns.
Information about the camera-specific trapping periods (and periods of malfunction) are summarized into a matrix about camera trap operability. These, together with the record table, are used to generate species detection histories for occupancy and spatial capture-recapture analyses. The user has considerable freedom in generating the detection histories; sampling occasion length, beginning date and and occasion start times are adjustable. In addition, trapping effort (i.e. active trap nights per station and occasion) can be computed for use as a covariate / offset on detection probability.
}

\section{Image organisation and management}{

The functions in this section set up a directory structure for storing camera trap images and identifying species and individuals from images. They build on one another and can be run in sequential order as needed.
\tabular{ll}{
\code{\link{createStationFolders}} \tab Create camera trap station directories for raw images \cr
\code{\link{fixDateTimeOriginal}}  \tab Fix DateTimeOriginal Exif metadata tag in Reconyx Hyperfire cameras \cr
\code{\link{OCRdataFields}}        \tab Optical character recognition (OCR) from data fields in images \cr
\code{\link{writeDateTimeOriginal}} \tab Write values to DateTimeOriginal tag in image metadata \cr
\code{\link{timeShiftImages}}      \tab Apply time shifts to JPEG images \cr
\code{\link{imageRename}}          \tab Copy and rename images based on station ID and image creation date \cr
\code{\link{addCopyrightTag}}      \tab Write a copyright tag into JPEG image metadata \cr
\code{\link{appendSpeciesNames}}   \tab Add or remove species names from image filenames \cr
}
}

\section{Species / individual identification}{

These functions assist in species identification and prepare individual identification of animals.
\tabular{ll}{
\code{\link{checkSpeciesNames}}          \tab Check species names against the ITIS taxonomic database \cr
\code{\link{createSpeciesFolders}}       \tab Create directories for species identification\cr
\code{\link{checkSpeciesIdentification}} \tab Consistency check on species image identification \cr
\code{\link{getSpeciesImages}}           \tab Gather all images of a species in a new directory \cr
}
}


\section{Image data extraction}{

These function use the directory structure built above (Section 'Image management workflow') and a table containing basic information about camera traps and/or stations (IDs, location, trapping period).
\tabular{ll}{
\code{\link{recordTable}}           \tab Create a species record table from camera trap images and videos \cr
\code{\link{recordTableIndividual}} \tab Create a single-species record table from camera trap images and videos with individual IDs \cr
\code{\link{exifTagNames}}          \tab Return Exif metadata tags and tag names from JPEG images \cr
\code{\link{addToPath}}             \tab Add the directory containing exiftool.exe to PATH temporarily (Windows only) \cr
\code{\link{filterRecordTable}}     \tab Filter existing record table for temporal independence. \cr 
}
}

\section{Data exploration and visualisation}{

These plots are generated from the record table and the camera trap table.
\tabular{ll}{
\code{\link{detectionMaps}}     \tab Generate maps of species richness and species presence by station, export shapefiles \cr
\code{\link{activityHistogram}} \tab Single-species diel activity histograms \cr
\code{\link{activityDensity}}   \tab Single-species diel activity kernel density estimation plots \cr
\code{\link{activityRadial}}    \tab Single-species diel activity radial plot \cr
\code{\link{activityOverlap}}   \tab Two-species diel activity overlap plots and estimates \cr
}
}


\section{Prepare occupancy and spatial capture-recapture analyses, and summarise surveys}{

\tabular{ll}{
\code{\link{createCovariates}}        \tab Extract covariate values from spatial rasters and prepare rasters for spatial predictions \cr
\code{\link{cameraOperation}}         \tab Create a camera operation matrix \cr
\code{\link{detectionHistory}}        \tab Species detection histories for occupancy analyses (single and multi-season) \cr
\code{\link{spatialDetectionHistory}} \tab Detection histories of individuals for spatial capture-recapture analyses  \cr
\code{\link{surveyReport}}            \tab Create a report about camera trap surveys and species detections \cr
\code{\link{surveyDashboard}}         \tab Shiny dashboard for summarizing and analyzing camera trap survey data \cr
}
}


\section{Community (multi-species) occupancy models}{

\tabular{ll}{
\code{\link{communityModel}}        \tab Create a community (multi-species) occupancy model for JAGS or Nimble \cr
\code{\link{commOccu-class}}        \tab commOccu objects \cr
\code{\link{fit,commOccu-method}}   \tab Fit a community (multi-species) occupancy model \cr
\code{\link{predict,commOccu-method}}   \tab Predictions from community occupancy models \cr
\code{\link{summary,commOccu-method}}   \tab Summarize community occupancy model \cr
\code{\link{plot_coef}}             \tab Plot effect sizes of covariates in community occupancy model \cr
\code{\link{plot_effects}}          \tab Plot Marginal Effects of Covariates \cr
\code{\link{PPC.residuals}}         \tab Calculate residuals from MCMC output of occupancy models \cr
\code{\link{PPC.community}}         \tab Calculate community-level posterior predictive checks for occupancy models \cr
}
}


\section{Sample data}{

\tabular{ll}{
\code{\link{camtraps}}          \tab  Sample camera trap station information table \cr
\code{\link{recordTableSample}} \tab Sample species record table \cr
\code{\link{recordTableIndividualSample}}  \tab Single-species record table with individual IDs\cr
\code{\link{camtrapsMultiSeason}}          \tab  Sample multi season camera trap station information table \cr
\code{\link{recordTableSampleMultiSeason}} \tab Sample  multi season species record table \cr
\code{\link{recordTableIndividualSampleMultiSeason}}  \tab Single-species multi season record table with individual IDs\cr
\code{\link{timeShiftTable}}    \tab Sample camera trap time shift information \cr
}
}

\section{Vignettes}{

\tabular{ll}{

\href{https://CRAN.R-project.org/package=camtrapR/vignettes/camtrapr1.pdf}{1. Organising raw camera trap images}  \cr
\href{https://CRAN.R-project.org/package=camtrapR/vignettes/camtrapr2.pdf}{2. Identifying species and individuals}  \cr
\href{https://CRAN.R-project.org/package=camtrapR/vignettes/camtrapr3.pdf}{3. Extracting Data from Camera Trapping Images and Videos}  \cr
\href{https://CRAN.R-project.org/package=camtrapR/vignettes/camtrapr4.pdf}{4. Data exploration and visualisation}  \cr
\href{https://CRAN.R-project.org/package=camtrapR/vignettes/camtrapr5.html}{5. Multi-species occupancy models}  \cr
}
}



\author{
Juergen Niedballa

Maintainer:Juergen Niedballa <camtrapr@gmail.com>
}

\references{

Niedballa, J., Sollmann, R., Courtiol, A., Wilting, A. (2016): camtrapR: an R package for efficient camera trap data management. Methods in Ecology and Evolution, 7(12). \url{https://besjournals.onlinelibrary.wiley.com/doi/full/10.1111/2041-210X.12600} \cr

camtrapR Google Group \url{https://groups.google.com/forum/#!forum/camtrapr} \cr

Phil Harvey's ExifTool \url{https://exiftool.org/} \cr
}

%~~ Optionally other standard keywords, one per line, from file KEYWORDS in the R documentation directory ~~
\keyword{package}
\seealso{
\pkg{overlap}
\pkg{unmarked}
\pkg{ubms}
\pkg{secr}
\pkg{wiqid}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/detectionMaps.R
\name{detectionMaps}
\alias{detectionMaps}
\title{Generate maps of observed species richness and species presences by station}
\usage{
detectionMaps(
  CTtable,
  recordTable,
  Xcol,
  Ycol,
  backgroundPolygon,
  stationCol = "Station",
  speciesCol = "Species",
  speciesToShow,
  richnessPlot = TRUE,
  speciesPlots = TRUE,
  addLegend = TRUE,
  printLabels = FALSE,
  smallPoints,
  plotR = TRUE,
  writePNG = FALSE,
  plotDirectory,
  createPlotDir = FALSE,
  pngMaxPix = 1000,
  writeShapefile = FALSE,
  shapefileName,
  shapefileDirectory,
  shapefileProjection
)
}
\arguments{
\item{CTtable}{data.frame. contains station IDs and coordinates}

\item{recordTable}{data.frame. the record table created by
\code{\link{recordTable}}}

\item{Xcol}{character. name of the column specifying x coordinates in
\code{CTtable}}

\item{Ycol}{character. name of the column specifying y coordinates in
\code{CTtable}}

\item{backgroundPolygon}{SpatialPolygons or SpatialPolygonsDataFrame.
Polygon to be plotted in the background of the map (e.g. project area
boundary)}

\item{stationCol}{character. name of the column specifying station ID in
\code{CTtable} and \code{recordTable}}

\item{speciesCol}{character. name of the column specifying species in
\code{recordTable}}

\item{speciesToShow}{character. Species to include in the maps. If missing,
all species in \code{recordTable} will be included.}

\item{richnessPlot}{logical. Generate a species richness plot?}

\item{speciesPlots}{logical. Generate plots of all species number of
independent events?}

\item{addLegend}{logical. Add legends to the plots?}

\item{printLabels}{logical. Add station labels to the plots?}

\item{smallPoints}{numeric. Number by which to decrease point sizes in plots
(optional).}

\item{plotR}{logical. Create plots in R graphics device?}

\item{writePNG}{logical. Create PNGs of the plots?}

\item{plotDirectory}{character. Directory in which to save the PNGs}

\item{createPlotDir}{logical. Create \code{plotDirectory}?}

\item{pngMaxPix}{integer. number of pixels in pngs on the longer side}

\item{writeShapefile}{logical. Create a shapefile from the output?}

\item{shapefileName}{character. Name of the shapefile to be saved. If empty,
a name will be generated automatically.}

\item{shapefileDirectory}{character. Directory in which to save the
shapefile.}

\item{shapefileProjection}{character. A character string of projection
arguments to use in the shapefile.}
}
\value{
An invisible \code{data.frame} with station coordinates, numbers of
events by species at each station and total species number by station. In
addition and optionally, R graphics or png image files.
}
\description{
Generates maps of observed species richness and species presence by species
and station. Output can be R graphics, PNG graphics or a shapefile for use
in GIS software.
}
\details{
The column name \code{stationCol} must be identical in \code{CTtable} and
\code{recordTable} and station IDs must match.

Shapefile creation depends on the packages \pkg{sf}.
Argument \code{shapefileProjection} must be a valid argument of
\code{\link[sf]{st_crs}} (one of (i) character: a string accepted by GDAL, 
(ii) integer, a valid EPSG value (numeric), or (iii) an object of class crs.
If \code{shapefileProjection} is undefined,
the resulting shapefile will lack a coordinate reference system.
}
\examples{


# load station information
data(camtraps)

# load record table
data(recordTableSample)


# create maps
Mapstest <- detectionMaps(CTtable           = camtraps,
                          recordTable       = recordTableSample,
                          Xcol              = "utm_x",
                          Ycol              = "utm_y",
                          stationCol        = "Station",
                          speciesCol        = "Species",
                          writePNG          = FALSE,
                          plotR             = TRUE,
                          printLabels       = TRUE,
                          richnessPlot      = TRUE,
                          addLegend         = TRUE
)



# with a polygon in the background, and for one species only

# make a dummy polygon for the background
library(sf)

Sr1 = st_polygon(list(cbind(c(521500,526500,527000, 521500, 521500),
                            c(607500, 608000, 603500, 603500, 607500))))
aoi <- data.frame(name = "My AOI")
st_geometry(aoi) <- st_geometry(Sr1)
st_crs(aoi) <- 32650  # assign CRS: UTM50N

Mapstest2 <- detectionMaps(CTtable           = camtraps,
                           recordTable       = recordTableSample,
                           Xcol              = "utm_x",
                           Ycol              = "utm_y",
                           backgroundPolygon = aoi,               # this was added
                           speciesToShow     = c("PBE", "VTA"),   # this was added
                           stationCol        = "Station",
                           speciesCol        = "Species",
                           writePNG          = FALSE,
                           plotR             = TRUE,
                           printLabels       = TRUE,
                           richnessPlot      = TRUE,
                           addLegend         = TRUE
)




}
\references{
A great resource for coordinate system information
is \url{https://spatialreference.org/}. Use the Proj4 string as
\code{shapefileProjection} argument.
}
\author{
Juergen Niedballa
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/communityModel.R
\encoding{UTF-8}
\name{communityModel}
\alias{communityModel}
\title{Create a community (multi-species) occupancy model for JAGS or Nimble}
\usage{
communityModel(
  data_list,
  model = c("Occupancy", "RN"),
  occuCovs = list(fixed = NULL, independent = NULL, ranef = NULL),
  detCovs = list(fixed = NULL, ranef = NULL),
  detCovsObservation = list(fixed = NULL, ranef = NULL),
  speciesSiteRandomEffect = list(det = FALSE, occu = FALSE),
  intercepts = list(det = "ranef", occu = "ranef"),
  effortCov = "effort",
  richnessCategories = NULL,
  augmentation = NULL,
  modelFile = NULL,
  nimble = FALSE,
  keyword_quadratic = "_squared"
)
}
\arguments{
\item{data_list}{list. Contains 3 slots: ylist, siteCovs, obsCovs. ylist is a list of detection histories (can be named), e.g. from \code{\link{detectionHistory}}. siteCovs is a data.frame with site covariates (optional). obsCovs is a list of site-occasion level covariates (e.g. site-occasion-specific effort, which is also returned by \code{\link{detectionHistory}}.}

\item{model}{character. "Occupancy" for standard occupancy model,  or "RN" for the occupancy model of Royle and Nichols (2003), which relates probability of detection of the species to the number of individuals available for detection at each station}

\item{occuCovs}{list. Up to 3 items named "fixed", "independent", and/or "ranef". Specifies fixed, independent or random effects of covariates on occupancy probability (continuous or categorical covariates). Independent effects are only supported for continuous covariates.}

\item{detCovs}{list. Up to 3 items named "fixed", "independent", and/or "ranef". Specifies fixed, independent or random effects of covariates on detection probability (continuous or categorical covariates). Independent effects are only supported for continuous covariates.}

\item{detCovsObservation}{list. Up to 2 items named "fixed" and/or "ranef". Specifies fixed or random effects of observation-level covariates on detection probability  (continuous or categorical covariates - categorical must be coded as character matrix)}

\item{speciesSiteRandomEffect}{list. Two items named "det" and "occu". If TRUE, adds a random effect of species and station. Only implemented for detection probability.}

\item{intercepts}{list. Two items named "det" and "occu" for detection and occupancy probability intercepts. Values can be "fixed" (= constant across species), "independent" (= independent estimates for each species), or "ranef" (= random effect of species on intercept).}

\item{effortCov}{character. Name of list item in \code{data_list$obsCovs} which contains effort. This does not include effort as a covariate on detection probability, but only uses NA / not NA information to create binary effort and ensure detection probabilities p are 0 when there was no effort (p will be 0 whereever \code{effortCov} is NA).}

\item{richnessCategories}{character. Name of categorical covariate in \code{data_list$siteCovs} for which to calculate separate richness estimates (optional). Can be useful to obtain separate richness estimates for different areas.}

\item{augmentation}{If NULL, no data augmentation (only use species in \code{data_list$ylist}), otherwise named list or vector with total number of (potential) species. Names: "maxknown" or "full". Example: \code{augmentation = c(maxknown = 30)} or \code{augmentation = c(full = 30)}}

\item{modelFile}{character. Text file name to save model to}

\item{nimble}{logical. If TRUE, model code will be for Nimble (incompatible with JAGS). If FALSE, model code is for JAGS.}

\item{keyword_quadratic}{character. A suffix in covariate names in the model that indicates a covariate is a quadratic effect of another covariate which does not carry the suffix in its name (e.g. if the covariate is "elevation", the quadratic covariate would be "elevation_squared").}
}
\value{
\code{commOccu} object. It is an S4 class containing all information required to run the models. See  \code{\link{commOccu-class}} for details.
}
\description{
Flexibly creates complete code and input data for community occupancy models for JAGS amd Nimble (both standard occupancy models and Royle-Nichols occupancy models), and automatically sets initial values and parameters to monitor. 
Supports fixed and random effects of covariates on detection and occupancy probabilities, using both continuous and categorical covariates (both site and site-occasion covariates). 

Optionally includes data augmentation (fully open community, or up to known maximum number of species, or no data augmentation). 
Allows combination of all these parameters for fast and flexible customization of community occupancy models.

Incidentally, the function can also be used to create model code and input for single-species single-season occupancy models (it is the special case of the community model with only one species). 
Such a model will run slower than proper single-species model JAGS code due to the additional species loop, but it is possible.

The function returns several derived quantities, e.g. species richness, Bayesian p-values (overall and by species), Freeman-Tukey residuals for actual and simulated data (by station and total). If doing data augmentation, metacommunity size and number of unseen species are returned also.
}
\details{
For examples of implementation, see Vignette 5: Multi-species occupancy models.

Fixed effects of covariates are constant across species, whereas random effect covariates differ between species. Independent effect differ between species and are independent (there is no underlying hyperdistribution).
Fixed, independent and random effects are allowed for station-level detection and occupancy covariates (a.k.a. site covariates). Fixed and random effects are also allowed for station-occasion level covariates (a.k.a. observation covariates). 
Currently independent effects are only supported for continuous site covariates, not categorical site covariates or observation-level covariates.

By default, random effects will be by species. It is however possible to use categorical site covariates for grouping (continuous|categorical).
Furthermore, is is possible to use use nested random effects of species and another categorical site covariate (so that there is a random effect of species and an additional random effect of a categorical covariate within each species).


Derived quantities returned by the model are:
  
  \tabular{ll}{
    \code{Bpvalue} \tab Bayesian p-value (overall) \cr
    \code{Bpvalue_species} \tab Bayesian p-value (by species) \cr
    
    
    \code{Nspecies} \tab Species richness (only in JAGS models)\cr
    \code{Nspecies_station} \tab Species richness at each sampling locations (only in JAGS models) \cr
    \code{Nspecies_Covariate} \tab Species richness by categorical covariate (when using \code{richnessCategories}, only in JAGS models) \cr
    
    
    \code{R2} \tab sum of Freeman-Tukey residuals of observed data within each species \cr
    \code{new.R2} \tab sum of Freeman-Tukey residuals of simulated data within each species \cr
    \code{R3} \tab Total sum of Freeman-Tukey residuals of observed data \cr 
    \code{new.R3} \tab Total sum of Freeman-Tukey residuals of simulated data \cr
    \emph{\code{Ntotal}} \tab Total metacommunity size (= observed species + n0) \cr
    \emph{\code{n0}} \tab Number of unseen species in metacommunity \cr
    \emph{\code{omega}} \tab Data augmentation parameter \cr
    \emph{\code{w}} \tab Metacommunity membership indicator for each species
  }

Quantities in \emph{italic} at the bottom are only returned in full data augmentation. \code{Nspecies} and \code{Nspecies_Covariate} are only returned in JAGS models (because Nimble models don't explicitly return latent occupancy status z).
}
\section{Parameter naming convention}{
 
The parameter names are assembled from building blocks. The nomenclature  is as follows:    
  \tabular{lll}{
    \bold{Name}           \tab \bold{Refers to}  \tab  \bold{Description}  \cr
    \bold{\code{alpha}}   \tab Submodel          \tab detection submodel  \cr
    \bold{\code{beta}}    \tab Submodel          \tab occupancy submode \cr
    \bold{\code{0}}       \tab Intercept         \tab denotes the intercepts (alpha0, beta0)  \cr
    \bold{\code{fixed}}   \tab Effect type       \tab fixed effects (constant across species)  \cr
    \bold{\code{indep}}   \tab Effect type       \tab independent effects (separate for each species)  \cr
    \bold{\code{ranef}}   \tab Effect type       \tab random effects (of species and/or other categorical covariates)  \cr
    \bold{\code{cont}}    \tab Covariate type    \tab continuous covariates \cr
    \bold{\code{categ}}   \tab Covariate type    \tab categorical covariates \cr
    \bold{\code{mean}}    \tab Hyperparameter    \tab mean of random effect \cr
    \bold{\code{sigma}}   \tab Hyperparameter    \tab standard deviation of random effect \cr
    \bold{\code{tau}}     \tab Hyperparameter    \tab precision of random effect (used internally, not returned)
  }


For example, a fixed intercept of occupancy (constant across species) is \bold{\code{beta0}}, and a fixed intercept of detection probability is \bold{\code{alpha0}}.


An occupancy probability intercept with a random effect of species is: 
  
\bold{\code{beta0.mean}} community mean of the occupancy probability intercept

\bold{\code{beta0.sigma}} standard deviation of the community occupancy probability intercept.

\bold{\code{beta0[1]}} occupancy probability intercept of species 1 (likewise for other species). 


For effects of site covariates, the pattern is:
  
\code{submodel.effectType.covariateType.CovariateName.hyperparameter}

For example:
  
\bold{\code{beta.ranef.cont.habitat.mean}} is the mean community effect of the continuous site covariate 'habitat' on occupancy probability.

\bold{\code{beta.ranef.cont.habitat[1]}} is the effect of continuous site covariate 'habitat' on occupancy probability of species 1.

Site-occasion covariates are denoted by ".obs" after the submodel, e.g.: 
  
\bold{\code{alpha.obs.fixed.cont.effort}} is the fixed effect of the continuous observation-level covariate 'effort' on detection probability
}

\examples{

\dontrun{
# the example below fits community occupancy models to the sample data in camtrapR
# models are fit both in JAGS and Nimble
# The data set only contains 5 species and 3 stations, so the results will be nonsense. 
# It is only a technical demonstration with the camtrapR workflow
# for more complete examples, see vignette 5

data("camtraps")

# create camera operation matrix
camop_no_problem <- cameraOperation(CTtable      = camtraps,
                                    stationCol   = "Station",
                                    setupCol     = "Setup_date",
                                    retrievalCol = "Retrieval_date",
                                    hasProblems  = FALSE,
                                    dateFormat   = "dmy"
)

data("recordTableSample")

# make list of detection histories
species_to_include <- unique(recordTableSample$Species)

  DetHist_list <- detectionHistory(
    recordTable         = recordTableSample,
    camOp                = camop_no_problem,
    stationCol           = "Station",
    speciesCol           = "Species",
    recordDateTimeCol    = "DateTimeOriginal",
    species              = species_to_include,
    occasionLength       = 7,
    day1                 = "station",
    datesAsOccasionNames = FALSE,
    includeEffort        = TRUE,
    scaleEffort          = TRUE,
    timeZone             = "Asia/Kuala_Lumpur"
)


# create some fake covariates for demonstration
sitecovs <- camtraps[, c(1:3)]
sitecovs$elevation <- c(300, 500, 600)   

# scale numeric covariates
sitecovs[, c(2:4)] <- scale(sitecovs[,-1])


# bundle input data for communityModel
data_list <- list(ylist = DetHist_list$detection_history,
                  siteCovs = sitecovs,
                  obsCovs = list(effort = DetHist_list$effort))


# create community model for JAGS
modelfile1 <- tempfile(fileext = ".txt")
mod.jags <- communityModel(data_list,
                           occuCovs = list(fixed = "utm_y", ranef = "elevation"),
                           detCovsObservation = list(fixed = "effort"),
                           intercepts = list(det = "ranef", occu = "ranef"),
                           modelFile = modelfile1)

summary(mod.jags)

# fit in JAGS
fit.jags <- fit(mod.jags,
                n.iter = 1000,
                n.burnin = 500,
                chains = 3)   
summary(fit.jags)

# response curves (= marginal effect plots)
plot_effects(mod.jags, 
             fit.jags, 
             submodel = "state")
plot_effects(mod.jags, 
             fit.jags, 
             submodel = "det")
             
# effect sizes plot
plot_coef(mod.jags, 
          fit.jags, 
          submodel = "state")
plot_coef(mod.jags, 
          fit.jags, 
          submodel = "det")              

# create community model for Nimble
modelfile2 <- tempfile(fileext = ".txt")
mod.nimble <- communityModel(data_list,
                             occuCovs = list(fixed = "utm_x", ranef = "utm_y"),
                             detCovsObservation = list(fixed = "effort"),
                             intercepts = list(det = "ranef", occu = "ranef"),
                             modelFile = modelfile2, 
                             nimble = TRUE)      # set nimble = TRUE

# load nimbleEcology package 
# currently necessary to do explicitly, to avoid additional package dependencies
require(nimbleEcology)

# fit uncompiled model in Nimble
fit.nimble.uncomp <- fit(mod.nimble, 
                         n.iter = 10, 
                         chains = 1)

# fit compiled model in Nimble
fit.nimble.comp <- fit(mod.nimble, 
                       n.iter = 5000, 
                       n.burnin = 2500,
                       chains = 3, 
                       compile = TRUE)

# parameter summary statistics
summary(fit.nimble.comp)


# response curves (= marginal effect plots)
plot_effects(mod.nimble, 
             fit.nimble.comp, 
             submodel = "state")
plot_effects(mod.nimble, 
             fit.nimble.comp, 
             submodel = "det")

# effect sizes plot
plot_coef(mod.nimble, 
          fit.nimble.comp, 
          submodel = "state")
plot_coef(mod.nimble, 
          fit.nimble.comp, 
          submodel = "det")   

# traceplots
plot(fit.nimble.comp)


}

}
\references{
Kéry, M., and J. A. Royle. "Applied hierarchical modelling in ecology - Modeling distribution, abundance and species richness using R and BUGS." Volume 1: Prelude and Static Models. Elsevier/Academic Press, 2016.
}
\author{
Juergen Niedballa
}

\docType{data}
\name{recordTableIndividualSampleMultiSeason}
\alias{recordTableIndividualSampleMultiSeason}
\title{Sample single-species multi-season record table with custom metadata from camera trap images}
\format{A data frame with 31 rows and 17 variables}
\usage{
data(recordTableIndividualSampleMultiSeason)
}
\description{
Sample single-species multi-season record table with individual IDs from the tagged sample images in the package. Generated with function \code{\link{recordTableIndividual}}, then duplicated to simulate a second year.
}
\details{


The variables are as follows:

\tabular{ll}{
\code{Station} \tab Camera trap station ID \cr
\code{Species} \tab Species ID \cr
\code{Individual} \tab Individual ID \cr
\code{DateTimeOriginal} \tab Date and time as extracted from image \cr
\code{Date} \tab record date \cr
\code{Time} \tab record time of day \cr
\code{delta.time.secs} \tab time difference to first species record at a station (seconds) \cr
\code{delta.time.mins} \tab time difference to first species record at a station (minutes) \cr
\code{delta.time.hours} \tab time difference to first species record at a station (hours) \cr
\code{delta.time.days} \tab time difference to first species record at a station (days) \cr
\code{Directory} \tab Image directory \cr
\code{FileName} \tab image filename \cr
\code{HierarchicalSubject} \tab content of the HierarchicalSubject image metadata tag \cr
\code{Model} \tab camera model extracted from image metadata \cr
\code{Make} \tab camera make extracted from image metadata \cr
\code{metadata_Species} \tab content of custom image metadata tag "Species" (see HierarchicalSubject) \cr
\code{metadata_individual} \tab content of custom image metadata tag "individual" (see HierarchicalSubject) \cr
}

}

\examples{
# example data were created as follows:
data(recordTableIndividualSample)

recordTableIndividualSample_season2 <- recordTableIndividualSample[1:10, ]
recordTableIndividualSample_season2$DateTimeOriginal <- gsub("2009", "2010",
    recordTableIndividualSample_season2$DateTimeOriginal)
recordTableIndividualSampleMultiSeason <- rbind(recordTableIndividualSample,
    recordTableIndividualSample_season2)
}
\keyword{datasets}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/communityModel.R
\docType{class}
\name{commOccu-class}
\alias{commOccu-class}
\title{commOccu objects}
\value{
\code{commOccu} object
}
\description{
commOccu objects
}
\section{Slots}{

\describe{
\item{\code{modelText}}{JAGS model code as a character vector (made up of code chunks, use cat() to print)}

\item{\code{params}}{Parameters to monitor in the model runs}

\item{\code{inits_fun}}{Function to create start values for the MCMC chains. It being a function ensures different values in each chain}

\item{\code{data}}{List with data needed to run the model (detection & effort matrices, site covariates, number of species / stations / occasions)}

\item{\code{input}}{Input data_list (unchanged)}

\item{\code{nimble}}{logical indicator for whether it is a Nimble model}

\item{\code{modelFile}}{Path of the text file containing the model code}

\item{\code{covariate_info}}{Data frame containing information about covariates. Only used internally in plot_* and predict methods}

\item{\code{model}}{character indicating whether it is a standard "Occupancy" model or Royle-Nichols ("RN") occupancy model}
}}

\note{
The \code{data} slot is a list of model input data. While the exact content depends on function input, it can be summarized as:

\tabular{ll}{
\code{y} \tab array of detection histories. Dimensions are: y[species, station, occasion] \cr
\code{effort_binary} \tab matrix of binary (1/0) survey effort. Only used to ensure p = 0 when effort = 0. Dimensions are: effort_binary[station, occasion] \cr
\code{site-occasion covariates} \tab The required content of data_list$obsCovs as named matrices with dimensions [station, occasion] \cr
\code{site covariates} \tab The required columns of data_list$siteCovs as named vectors (length = number of stations) \cr
\code{M} \tab Number of species \cr
\code{J} \tab Number of stations \cr
\code{maxocc} \tab Number of occasions \cr
}

For categorical site-occasion covariates, an addition matrix containing an integer representation of the character matrix with suffix "_integer" is stored in the data slot.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/recordTable.R
\name{recordTable}
\alias{recordTable}
\title{Generate a species record table from camera trap images and videos}
\usage{
recordTable(
  inDir,
  IDfrom,
  cameraID,
  camerasIndependent,
  exclude,
  minDeltaTime = 0,
  deltaTimeComparedTo,
  timeZone,
  stationCol,
  writecsv = FALSE,
  outDir,
  metadataHierarchyDelimitor = "|",
  metadataSpeciesTag,
  additionalMetadataTags,
  removeDuplicateRecords = TRUE,
  returnFileNamesMissingTags = FALSE,
  eventSummaryColumn,
  eventSummaryFunction,
  video
)
}
\arguments{
\item{inDir}{character. Directory containing station directories. It must
either contain images in species subdirectories (e.g.
inDir/StationA/SpeciesA) or images with species metadata tags (without
species directories, e.g. inDir/StationA).}

\item{IDfrom}{character. Read species ID from image metadata ("metadata") of
from species directory names ("directory")?}

\item{cameraID}{character. Where should the function look for camera IDs:
'filename', 'directory'. 'filename' requires images renamed with
\code{\link{imageRename}}. 'directory' requires a camera subdirectory within
station directories (station/camera/species). Can be missing.}

\item{camerasIndependent}{logical. If \code{TRUE}, species records are
considered to be independent between cameras at a station.}

\item{exclude}{character. Vector of species names to be excluded from the
record table}

\item{minDeltaTime}{integer. Time difference between records of the same
species at the same station to be considered independent (in minutes)}

\item{deltaTimeComparedTo}{character. For two records to be considered
independent, must the second one be at least \code{minDeltaTime} minutes
after the last independent record of the same species
(\code{"lastIndependentRecord"}), or \code{minDeltaTime} minutes after the
last record (\code{"lastRecord"})?}

\item{timeZone}{character. Must be a value returned by
\code{\link[base:timezones]{OlsonNames}}}

\item{stationCol}{character. Name of the camera trap station column.
Assuming "Station" if undefined.}

\item{writecsv}{logical. Should the record table be saved as a .csv?}

\item{outDir}{character. Directory to save csv to. If NULL and
\code{writecsv = TRUE}, recordTable will be written to \code{inDir}.}

\item{metadataHierarchyDelimitor}{character. The character delimiting
hierarchy levels in image metadata tags in field "HierarchicalSubject".
Either "|" or ":".}

\item{metadataSpeciesTag}{character. In custom image metadata, the species
ID tag name.}

\item{additionalMetadataTags}{character. Additional camera model-specific
metadata tags to be extracted. (If possible specify tag groups as returned
by \code{\link{exifTagNames}})}

\item{removeDuplicateRecords}{logical. If there are several records of the
same species at the same station (also same camera if cameraID is defined)
at exactly the same time, show only one?}

\item{returnFileNamesMissingTags}{logical. If species are assigned with
metadata and images are not tagged, return a few file names of these images
as a message?}

\item{eventSummaryColumn}{character. A column in the record table (e.g. from
a metadata tag) by to summarise non-independent records (those within
\code{minDeltaTime} of a given record) with a user-defined function
(\code{eventSummaryFunction})}

\item{eventSummaryFunction}{character. The function by which to summarise
\code{eventSummaryColumn} of non-independent records, e.g. "sum", "max"
(optional)}

\item{video}{list. Contains information on how to handle video data
(optional). See details.}
}
\value{
A data frame containing species records and additional information
about stations, date, time and (optionally) further metadata.
}
\description{
Generates a record table from camera trap images or videos. Images/videos
must be sorted into station directories at least. The function can read
species identification from a directory structure (Station/Species or
Station/Camera/Species) or from image metadata tags.
}
\details{
The function can handle a number of different ways of storing images, and
supports species identification by moving images into species directories as
well as metadata tagging. In every case, images need to be stored into
station directories. If images are identified by moving them into species
directories, a camera directory is optional: "Station/Species/XY.JPG" or
"Station/Camera/Species/XY.JPG". Likewise, if images are identified using
metadata tagging, a camera directory can be used optionally:
"Station/XY.JPG" or "Station/Camera/XY.JPG".

If images are identified by metadata tagging, \code{metadataSpeciesTag}
specifies the metadata tag group name that contains species identification
tags. \code{metadataHierarchyDelimitor} is "|" for images tagged in DigiKam
and images tagged in Adobe Bridge / Lightroom with the default settings. It
is only necessary to change it if the default was changed in these programs.

\code{minDeltaTime} is a criterion for temporal independence of species
recorded at the same station. Setting it to 0 will make the function return
all records. Setting it to a higher value will remove records that were
taken less than \code{minDeltaTime} minutes after the last record
(\code{deltaTimeComparedTo = "lastRecord"}) or the last independent record
(\code{deltaTimeComparedTo = "lastIndependentRecord"}).

\code{removeDuplicateRecords} determines whether duplicate records
(identical station, species, date/time, (and camera if applicable)) are
all returned (FALSE) or collapsed into a single unique record (TRUE).

\code{camerasIndependent} defines if the cameras at a station are to be
considered independent. If \code{TRUE}, records of the same species taken by
different cameras are considered independent (e.g. if they face different
trails). Use \code{FALSE} if both cameras face each other and possibly
\code{TRUE} ).

\code{exclude} can be used to exclude "species" directories containing
irrelevant images (e.g. "team", "blank", "unidentified"). \code{stationCol}
can be set to match the station column name in the camera trap station table
(see \code{\link{camtraps}}).

Many digital images contain Exif metadata tags such as "AmbientTemperature"
or "MoonPhase" that can be extracted if specified in \code{metadataTags}.
Because these are manufacturer-specific and not standardized, function
\code{\link{exifTagNames}} provides a vector of all available tag names.
Multiple names can be specified as a character vector as: \code{c(Tag1,
Tag2, ...)}. The metadata tags thus extracted may be used as covariates in
modelling species distributions.

\code{eventSummaryColumn} and \code{eventSummaryFunction} can be used to
extract summary statistics for independent sampling events. For example, you
assigned a "count" tag to your images, indicating the number of individuals
in a picture. In a sequence of pictures taken within 1 minute, most pictures
show one individual, but one image shows two individuals. You tagged the
images accordingly (count = 1 or count = 2) and run \code{recordTable}. Set
\code{eventSummaryColumn = "count"} and \code{eventSummaryFunction = "max"}
to obtain the maximum number of \code{count} in all images within
\code{minDeltaTime} minutes of a given record. The results is in a new
column, in this example \code{count_max}. You can also calculate several
statistics at the same time, by supplying vectors of values, e.g.
\code{eventSummaryColumn = c("count", "count", "camera")} and
\code{eventSummaryFunction = c("min", "max", "unique")} to get minimum and
maximum count and all unique camera IDs for that event. Note that
\code{eventSummaryColumn} and \code{eventSummaryFunction} must be of same
length.

Argument \code{video} is a named list with 2 or 4 items. 2 items
(\code{file_formats}, \code{dateTimeTag}) are always required, and are
sufficent if \code{IDfrom = "directory"}. In that case, no digiKam tags will
be returned.  To return digiKam tags, two additional items are required
(\code{db_directory}, \code{db_filename}). This is essential when using
\code{IDfrom = "metadata"}. When using \code{IDfrom = "directory"}, it is
optional, but allows to extract metadata tags assigned to videos in digiKam.
This workaround is necessary because digiKam tags are not written into video
metadata, but are only saved in the digiKam database. So in contrast to JPG
images, they can not be extracted with ExifTool. It also requires that
\code{inDir} is in your digiKam database.

The items of argument \code{video} are:

\tabular{ll}{ \code{file_formats} \tab The video formats to extract (include
"jpg" if you want .JPG image metadata) \cr \code{dateTimeTag} \tab the
metadata tag to extract date/time from (use \code{\link{exifTagNames}} to
find out which tag is suitable) \cr \code{db_directory} \tab The directory
containing digiKam database (optional if \code{IDfrom = "directory"}) \cr
\code{db_filename} \tab The digiKam database file in \code{db_directory}
(optional if \code{IDfrom = "directory"}) \cr }

See the examples below for for how to specify the argument \code{video}.
}
\note{
The results of a number of other function will depend on the output of
this function (namely on the arguments \code{exclude} for excluding species
and \code{minDeltaTime}/ \code{deltaTimeComparedTo} for temporal
independence):

\tabular{l}{ \code{\link{detectionMaps}} \cr \code{\link{detectionHistory}}
\cr \code{\link{activityHistogram}} \cr \code{\link{activityDensity}} \cr
\code{\link{activityRadial}} \cr \code{\link{activityOverlap}} \cr
\code{\link{activityHistogram}} \cr \code{\link{surveyReport}} \cr }
}
\section{Warning }{
 Custom image metadata must be organised hierarchically
(tag group - tag; e.g. "Species" - "Leopard Cat"). Detailed information on
how to set up and use metadata tags can be found in
\href{https://CRAN.R-project.org/package=camtrapR/vignettes/camtrapr2.pdf}{vignette
2: Species and Individual Identification}.

Custom image metadata tags must be written to the images. The function
cannot read tags from .xmp sidecar files. Make sure you set the preferences
accordingly. In DigiKam, go to Settings/Configure digiKam/Metadata. There,
make sure "Write to sidecar files" is unchecked.

Please note the section about defining argument \code{timeZone} in the
vignette on data extraction (accessible via
\code{vignette("DataExtraction")} or online
(\url{https://cran.r-project.org/package=camtrapR/vignettes/camtrapr3.pdf})).
}

\examples{



\dontrun{   # the examples take too long to pass CRAN tests


# set directory with camera trap images in station directories
wd_images_ID_species <- system.file("pictures/sample_images_species_dir", 
                                    package = "camtrapR")

if (Sys.which("exiftool") != ""){        # only run these examples if ExifTool is available


rec_table1 <- recordTable(inDir               = wd_images_ID_species,
                       IDfrom                 = "directory",
                       minDeltaTime           = 60,
                       deltaTimeComparedTo    = "lastRecord",
                       writecsv               = FALSE,
                       additionalMetadataTags = c("EXIF:Model", "EXIF:Make")
)
# note argument additionalMetadataTags: it contains tag names as returned by function exifTagNames

rec_table2 <- recordTable(inDir               = wd_images_ID_species,
                       IDfrom                 = "directory",
                       minDeltaTime           = 60,
                       deltaTimeComparedTo    = "lastRecord",
                       exclude                = "UNID",
                       writecsv               = FALSE,
                       timeZone               = "Asia/Kuala_Lumpur",
                       additionalMetadataTags = c("EXIF:Model", "EXIF:Make", "NonExistingTag"),
                       eventSummaryColumn     = "EXIF:Make",
                       eventSummaryFunction   = "unique"
                       )
                       
# note the warning that the last tag in "additionalMetadataTags" ("NonExistingTag") was not found


any(rec_table1$Species == "UNID")    # TRUE
any(rec_table2$Species == "UNID")    # FALSE


# here's how the removeDuplicateRecords argument works

rec_table3a <- recordTable(inDir              = wd_images_ID_species,
                       IDfrom                 = "directory",
                       minDeltaTime           = 0,
                       exclude                = "UNID",
                       timeZone               = "Asia/Kuala_Lumpur",
                       removeDuplicateRecords = FALSE
)

rec_table3b <- recordTable(inDir              = wd_images_ID_species,
                       IDfrom                 = "directory",
                       minDeltaTime           = 0,
                       exclude                = "UNID",
                       timeZone               = "Asia/Kuala_Lumpur",
                       removeDuplicateRecords = TRUE
)


anyDuplicated(rec_table3a[, c("Station", "Species", "DateTimeOriginal")])   # got duplicates
anyDuplicated(rec_table3b[, c("Station", "Species", "DateTimeOriginal")])   # no duplicates

# after removing duplicates, both are identical:
whichAreDuplicated <- which(duplicated(rec_table3a[,c("Station", "Species", "DateTimeOriginal")]))
all(rec_table3a[-whichAreDuplicated,] == rec_table3b)


### extracting species IDs from metadata

wd_images_ID_species_tagged <- system.file("pictures/sample_images_species_tag", 
                                           package = "camtrapR")

rec_table4 <- recordTable(inDir               = wd_images_ID_species_tagged,
                       IDfrom                 = "metadata",
                       metadataSpeciesTag     = "Species",
                       exclude                = "unidentified")


###  Including videos
# sample videos are not included in package

# with videos, IDfrom = "directory", not extracting digiKam metadata

rec_table4 <- recordTable(inDir  = wd_images_ID_species,
                          IDfrom = "directory",
                          video  = list(file_formats = c("jpg", "mp4"),
                                        dateTimeTag  = "QuickTime:CreateDate")
)

# with videos, IDfrom = "metadata", extracting digiKam metadata

rec_table5 <- recordTable(inDir  = wd_images_ID_species,
                          IDfrom = "metadata",
                          metadataSpeciesTag = "Species",
                          video  = list(file_formats = c("jpg", "mp4", "avi", "mov"),
                                        dateTimeTag  = "QuickTime:CreateDate",
                                        db_directory = "C:/Users/YourName/Pictures",
                                        db_filename = "digikam4.db")
)

} else {  
# show function output if ExifTool is not available
message("ExifTool is not available. Cannot test function. Loading recordTableSample instead")
data(recordTableSample)
}

}

}
\references{
Phil Harvey's ExifTool \url{https://exiftool.org/}
}
\author{
Juergen Niedballa
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/speciesAccum.R
\name{speciesAccum}
\alias{speciesAccum}
\title{Species Accumulation Curves for Camera Trap Data}
\usage{
speciesAccum(
  CTtable,
  recordTable,
  speciesCol,
  recordDateTimeCol,
  recordDateTimeFormat = "ymd HMS",
  setupCol,
  dateFormat = "ymd",
  stationCol,
  assemblageCol = NULL,
  q = 0,
  x_unit = c("station", "survey_day", "station_day"),
  knots = 40,
  conf = 0.95,
  nboot = 50,
  endpoint = NULL
)
}
\arguments{
\item{CTtable}{data.frame containing the camera trap deployment information.}

\item{recordTable}{data.frame containing the camera trap records.}

\item{speciesCol}{character. Name of the column specifying species names in recordTable}

\item{recordDateTimeCol}{character. Name of the column containing date and time information in recordTable}

\item{recordDateTimeFormat}{character. Format of column \code{recordDateTimeCol} in \code{recordTable}}

\item{setupCol}{character. Name of the column containing camera setup dates in CTtable}

\item{dateFormat}{character. Format of column \code{setupCol} in \code{CTtable}}

\item{stationCol}{character. Name of the column containing station IDs in both tables}

\item{assemblageCol}{character. Optional. Name of column in recordTable for grouping data into separate assemblages}

\item{q}{numeric. The order of diversity measure. Default is 0 (species richness)}

\item{x_unit}{character. Whether to use "station" or "day" as sampling unit. Default is "station"}

\item{knots}{numeric. number of values along x axis for which values are computed}

\item{conf}{numeric. confidence interval}

\item{nboot}{numeric. number of replications}

\item{endpoint}{integer. Sample size used as endpoint for rarefaction/extrapolation (in \code{\link[iNEXT]{iNEXT}})}
}
\value{
An object of class "iNEXT" containing:
\itemize{
  \item DataInfo - data information
  \item iNextEst - diversity estimates for rarefied and extrapolated samples
  \item AsyEst - asymptotic diversity estimates
}
}
\description{
Generates species accumulation, rarefaction and extrapolation curves
   from camera trap data using the \pkg{iNEXT} package (Chao et al. 2014). The function
   creates sampling effort-based accumulation curves with sampling units being either
   camera trap stations or days.
   
   Note that these curves are based on observed detections only and do not account 
   for imperfect detection. Species may be present but not detected, leading to 
   underestimation of true species richness. For analyses that explicitly account 
   for imperfect detection, consider using occupancy-based approaches (see 
   \code{\link{communityModel}}).
}
\details{
The function provides three types of curves:
\itemize{
  \item Sample-size-based rarefaction/extrapolation curve
  \item Sample completeness curve
  \item Coverage-based rarefaction/extrapolation curve
}

While these curves provide useful insights into sampling completeness and species 
richness patterns, they should be interpreted with caution in camera trap studies. 
Unlike occupancy models, they do not account for:
\itemize{
  \item Imperfect detection (species present but not detected)
  \item Variation in detection probability among species
  \item Spatial variation in species occurrence
  \item Temporal variation in species activity
}
}
\note{
Requires package \pkg{iNEXT}.
}
\examples{
\dontrun{
# Basic usage with stations as sampling units
result <- speciesAccum(
  CTtable = cams,
  recordTable = recs,
  speciesCol = "Species",
  recordDateTimeCol = "DateTime", 
  setupCol = "Setup_date",
  stationCol = "Station",
  q = 0,
  x_unit = "station"
)

# Plot results
ggiNEXT(result, type = 1)  # Sample-size-based R/E curve
ggiNEXT(result, type = 2)  # Sample completeness curve
ggiNEXT(result, type = 3)  # Coverage-based R/E curve

# With assemblage grouping and days as sampling units
result_by_assemblage <- speciesAccum(
  CTtable = cams,
  recordTable = recs,
  speciesCol = "Species",
  recordDateTimeCol = "DateTime",
  setupCol = "Setup_date",
  stationCol = "Station",
  assemblageCol = "Season",
  q = 0,
  x_unit = "day"
)
}

}
\references{
Chao, A., Gotelli, N. J., Hsieh, T. C., Sander, E. L., Ma, K. H., Colwell, R. K., 
& Ellison, A. M. (2014). Rarefaction and extrapolation with Hill numbers: 
A framework for sampling and estimation in species diversity studies. 
Ecological Monographs, 84(1), 45-67.
}
\seealso{
\code{\link{surveyDashboard}} for interactive species accumulation analysis
}
\author{
Juergen Niedballa
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/detectionHistory.R
\name{detectionHistory}
\alias{detectionHistory}
\title{Species detection histories for occupancy analyses}
\usage{
detectionHistory(
  recordTable,
  species,
  camOp,
  output = c("binary", "count"),
  stationCol = "Station",
  speciesCol = "Species",
  recordDateTimeCol = "DateTimeOriginal",
  recordDateTimeFormat = "ymd HMS",
  occasionLength,
  minActiveDaysPerOccasion,
  maxNumberDays,
  day1 = "survey",
  buffer,
  includeEffort = TRUE,
  scaleEffort = FALSE,
  occasionStartTime = "deprecated",
  datesAsOccasionNames = FALSE,
  timeZone,
  writecsv = FALSE,
  outDir,
  unmarkedMultFrameInput
)
}
\arguments{
\item{recordTable}{data.frame. the record table created by
\code{\link{recordTable}}}

\item{species}{character. species name(s) for which to compute detection histories. Can 
be either a single species name (for use with \pkg{unmarked}/\pkg{ubms}) or a vector 
of multiple species names (for input to \code{\link{communityModel}})}

\item{camOp}{The camera operation matrix as created by
\code{\link{cameraOperation}}}

\item{output}{character. Return binary detections ("binary") or counts of
detections ("count")}

\item{stationCol}{character. name of the column specifying Station ID in
\code{recordTable}}

\item{speciesCol}{character. name of the column specifying species in
\code{recordTable}}

\item{recordDateTimeCol}{character. name of the column specifying date and
time in \code{recordTable}}

\item{recordDateTimeFormat}{character. Format of column
\code{recordDateTimeCol} in \code{recordTable}}

\item{occasionLength}{integer. occasion length in days}

\item{minActiveDaysPerOccasion}{integer. minimum number of active trap days
for occasions to be included (optional)}

\item{maxNumberDays}{integer. maximum number of trap days per station
(optional)}

\item{day1}{character. When should occasions begin: station setup date
("station"), first day of survey ("survey"), a specific date (e.g.
"2015-12-31")?}

\item{buffer}{integer. Makes the first occasion begin a number of days after
station setup. (optional)}

\item{includeEffort}{logical. Compute trapping effort (number of active
camera trap days per station and occasion)?}

\item{scaleEffort}{logical. scale and center effort matrix to mean = 0 and
sd = 1?}

\item{occasionStartTime}{(DEPRECATED) integer. time of day (the full hour)
at which to begin occasions. Please use argument \code{occasionStartTime} in
\code{\link{cameraOperation}} instead.}

\item{datesAsOccasionNames}{If \code{day1 = "survey"}, occasion names in the
detection history will be composed of first and last day of that occasion.}

\item{timeZone}{character. Must be a value returned by
\code{\link[base:timezones]{OlsonNames}}}

\item{writecsv}{logical. Should the detection history be saved as a .csv?}

\item{outDir}{character. Directory into which detection history .csv file is
saved}

\item{unmarkedMultFrameInput}{logical. Return input for multi-season
occupancy models in unmarked (argument "y" in
\code{\link[unmarked]{unmarkedMultFrame}}?}
}
\value{
If a single species is provided (typical for \pkg{unmarked}/\pkg{ubms} analyses), 
returns a list with either 1, 2 or 3 elements depending on the value of \code{includeEffort} 
and \code{scaleEffort}:
\item{detection_history}{A species detection matrix}
\item{effort}{A matrix giving the number of active camera trap days per station and 
occasion (= camera trapping effort). Only returned if \code{includeEffort = TRUE}}
\item{effort_scaling_parameters}{Scaling parameters of the effort matrix. Only 
returned if \code{includeEffort} and \code{scaleEffort} are \code{TRUE}}

If multiple species are provided (for use with \code{\link{communityModel}}), returns 
a similar list structure but with \code{detection_history} containing a named list of 
detection matrices, one for each species. The effort matrix is identical for all 
species and thus returned only once.
}
\description{
This function generates species detection histories that can be used in
single-species occupancy analyses with packages \link[unmarked:unmarked-package]{unmarked} 
and \pkg{ubms}, as well as multi-species/community occupancy models via 
\code{\link{communityModel}}. It generates detection histories in different formats, 
with adjustable occasion length and occasion start time.
}
\details{
The function creates species detection matrices in two possible formats:
detection-by-date or detection-by-occasion. The start of detection histories 
is controlled by \code{day1}:
\itemize{
  \item \code{"station"}: Each station's history begins on its setup day
  \item \code{"survey"}: All stations begin on the first day of the survey
  \item A specific date (e.g., \code{"2015-12-31"}): All stations begin on this date
}
Dates must be in "YYYY-MM-DD" format if specified directly.


Two output formats are available via the \code{output} parameter:
\itemize{
  \item \code{"binary"}: Records detection (1) or non-detection (0)
  \item \code{"count"}: Records the number of detections per occasion
}


The \code{includeEffort} parameter determines how camera operation affects the output:
\itemize{
  \item If \code{FALSE}: Periods when cameras were not operational or only 
  partly operational appear as NA in the detection history. This may lose 
  species record from incomplete occasions.
  \item If \code{TRUE}: Incomplete occasions are retained. Outputs contain a 
  separate effort matrix that can be used as an observation covariate in occupancy models.
}
It is generally advisable to include effort as a covariate to account for 
uneven sampling effort.


\code{occasionLength} controls how many days are aggregated into each sampling occasion.
Note that \code{occasionStartTime} has moved to \code{\link{cameraOperation}} to ensure
proper calculation of daily effort.

The values of \code{stationCol} in \code{recordTable} must be matched by the
row names of \code{camOp} (case-insensitive), otherwise an error is raised.

For date/time formatting, \code{recordDateTimeFormat} accepts two syntax styles:
\itemize{
  \item Base R style (using \%): e.g., "\%Y-\%m-\%d \%H:\%M:\%S"
  \item lubridate style: e.g., "ymd HMS"
}
\pkg{lubridate} will be used if there are no "\%" characters in
\code{recordDateTimeFormat}. The default and recommended format is 
"YYYY-MM-DD HH:MM:SS" (e.g., "2014-09-30 22:59:59").


For multi-season studies where \code{sessionCol} was used in \code{\link{cameraOperation}},
the function automatically detects this structure. Set \code{unmarkedMultFrameInput = TRUE}
to format output for \code{unmarkedMultFrame}, with rows representing sites and columns
ordered by season-major, occasion-minor (e.g., season1-occasion1, season1-occasion2, etc.).
}
\section{Warning}{
 Setting \code{output = "count"} returns a count of
detections, not individuals. These counts are not suitable for abundance modeling
(e.g., N-mixture models) as they do not represent individual animals.

For important information about the \code{timeZone} parameter, please refer to
the "Data Extraction" vignette (\code{vignette("DataExtraction")} or online at
\url{https://cran.r-project.org/package=camtrapR/vignettes/camtrapr3.pdf}).
}

\examples{


# define image directory
wd_images_ID <- system.file("pictures/sample_images_species_dir", package = "camtrapR")

# load station information
data(camtraps)

# create camera operation matrix
camop_no_problem <- cameraOperation(CTtable      = camtraps,
                                    stationCol   = "Station",
                                    setupCol     = "Setup_date",
                                    retrievalCol = "Retrieval_date",
                                    hasProblems  = FALSE,
                                    dateFormat   = "dmy"
)

\dontrun{
if (Sys.which("exiftool") != ""){        # only run this function if ExifTool is available
recordTableSample <- recordTable(inDir               = wd_images_ID,
                                 IDfrom              = "directory",
                                 minDeltaTime        = 60,
                                 deltaTimeComparedTo = "lastRecord",
                                 exclude             = "UNID",
                                 timeZone            = "Asia/Kuala_Lumpur"
)
}
}
data(recordTableSample)    # load the record history, as created above


# compute detection history for a species

# without trapping effort
DetHist1 <- detectionHistory(recordTable         = recordTableSample,
                            camOp                = camop_no_problem,
                            stationCol           = "Station",
                            speciesCol           = "Species",
                            recordDateTimeCol    = "DateTimeOriginal",
                            species              = "VTA",
                            occasionLength       = 7,
                            day1                 = "station",
                            datesAsOccasionNames = FALSE,
                            includeEffort        = FALSE,
                            timeZone             = "Asia/Kuala_Lumpur"
)

DetHist1                     # this is a list with 1 element
DetHist1$detection_history   # this is the contained detection/non-detection matrix


# with effort / using base R to define recordDateTimeFormat
DetHist2 <- detectionHistory(recordTable          = recordTableSample,
                             camOp                = camop_no_problem,
                             stationCol           = "Station",
                             speciesCol           = "Species",
                             recordDateTimeCol    = "DateTimeOriginal",
                             species              = "VTA",
                             occasionLength       = 7,
                             day1                 = "station",
                             datesAsOccasionNames = FALSE,
                             includeEffort        = TRUE,
                             scaleEffort          = FALSE,
                             timeZone             = "Asia/Kuala_Lumpur"
)

DetHist2$detection_history  # detection history  (alternatively, use: DetHist2[[1]])
DetHist2$effort             # effort (alternatively, use: DetHist2[[2]])

# with effort / using lubridate package to define recordDateTimeFormat
DetHist2_lub <- detectionHistory(recordTable          = recordTableSample,
                                 camOp                = camop_no_problem,
                                 stationCol           = "Station",
                                 speciesCol           = "Species",
                                 recordDateTimeCol    = "DateTimeOriginal",
                                 recordDateTimeFormat = "ymd HMS",
                                 species              = "VTA",
                                 occasionLength       = 7,
                                 day1                 = "station",
                                 datesAsOccasionNames = FALSE,
                                 includeEffort        = TRUE,
                                 scaleEffort          = FALSE,
                                 timeZone             = "Asia/Kuala_Lumpur"
)    

DetHist2_lub$detection_history  # detection history  (alternatively, use: DetHist2_lub[[1]])
DetHist2_lub$effort             # effort (alternatively, use: DetHist2_lub[[2]])


# multi-season detection history

# load multi-season data
data(camtrapsMultiSeason)
data(recordTableSampleMultiSeason)

# multi-season camera operation matrix
camop_season <- cameraOperation(CTtable          = camtrapsMultiSeason,
                                    stationCol   = "Station",
                                    setupCol     = "Setup_date",
                                    sessionCol   = "session",
                                    retrievalCol = "Retrieval_date",
                                    hasProblems  = TRUE,
                                    dateFormat   = "dmy"
)

# multi-season detection history
DetHist_multi_season <- detectionHistory(recordTable      = recordTableSampleMultiSeason,
                            camOp                  = camop_season,
                            stationCol             = "Station",
                            speciesCol             = "Species",
                            species                = "VTA",
                            occasionLength         = 10,
                            day1                   = "station",
                            recordDateTimeCol      = "DateTimeOriginal",
                            includeEffort          = TRUE,
                            scaleEffort            = FALSE,
                            timeZone               = "UTC",
                            unmarkedMultFrameInput = TRUE
)

DetHist_multi_season


# Multi-species example for community occupancy analysis with communityModel()
DetHist_multi_species <- detectionHistory(recordTable = recordTableSample,
                                 species = c("VTA", "PBE", "EGY"),   # multiple species
                                 camOp = camop_no_problem,
                                 stationCol = "Station",
                                 speciesCol = "Species",
                                 recordDateTimeCol = "DateTimeOriginal",
                                 occasionLength = 7,
                                 day1 = "station",
                                 includeEffort = TRUE,
                                 scaleEffort = FALSE,
                                 timeZone = "Asia/Kuala_Lumpur"
)


# bundle input data for communityModel
data_list <- list(ylist = DetHist_multi_species$detection_history,
                  siteCovs = camtraps,
                  obsCovs = list(effort = DetHist_multi_species$effort))


\dontrun{

# create community model
mod.jags <- communityModel(data_list,
                           ...) # model specification 
}

}
\author{
Juergen Niedballa
}

\docType{data}
\name{recordTableSampleMultiSeason}
\alias{recordTableSampleMultiSeason}
\title{Sample multi-season species record table from camera trap images}
\format{A data frame with 78 rows and 11 variables}
\usage{
data(recordTableSampleMultiSeason)
}
\description{
Sample multi-season species record table from camera trap images generated from the sample images in the package with the function \code{\link{recordTable}}. Season 2009 is the same as \code{\link{recordTableSample}}, season 2010 was simulated by adding 1 year to these records.
}
\details{


The variables are as follows:

\tabular{ll}{
\code{Station} \tab Camera trap station ID \cr
\code{Species} \tab Species ID \cr
\code{DateTimeOriginal} \tab Date and time as extracted from image \cr
\code{Date} \tab record date \cr
\code{Time} \tab record time of day \cr
\code{delta.time.secs} \tab time difference to first species record at a station (seconds) \cr
\code{delta.time.mins} \tab time difference to first species record at a station (minutes) \cr
\code{delta.time.hours} \tab time difference to first species record at a station (hours) \cr
\code{delta.time.days} \tab time difference to first species record at a station (days) \cr
\code{Directory} \tab Image directory \cr
\code{FileName} \tab image filename \cr
}

 
 
}
\keyword{datasets}

\examples{
# data were created with the following code:

data(recordTableSample)
recordTableSample_season2 <- recordTableSample

# substitute 2009 with 2010
recordTableSample_season2$DateTimeOriginal <- gsub("2009", "2010",
    recordTableSample_season2$DateTimeOriginal)
# combine with season 2009
recordTableSampleMultiSeason <- rbind(recordTableSample, recordTableSample_season2)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/filterRecordTable.R
\name{filterRecordTable}
\alias{filterRecordTable}
\title{Filter species record table for temporal independence}
\usage{
filterRecordTable(
  recordTable,
  minDeltaTime = 0,
  deltaTimeComparedTo,
  speciesCol = "Species",
  stationCol,
  cameraCol,
  camerasIndependent,
  recordDateTimeCol = "DateTimeOriginal",
  recordDateTimeFormat = "ymd HMS",
  removeDuplicateRecords = TRUE,
  exclude,
  timeZone,
  writecsv = FALSE,
  outDir,
  eventSummaryColumn,
  eventSummaryFunction,
  quiet = FALSE
)
}
\arguments{
\item{recordTable}{data frame as created by \code{\link{recordTable}}.}

\item{minDeltaTime}{integer. Time difference between records of the same
species at the same station to be considered independent (in minutes)}

\item{deltaTimeComparedTo}{character. For two records to be considered
independent, must the second one be at least \code{minDeltaTime} minutes
after the last independent record of the same species
(\code{"lastIndependentRecord"}), or \code{minDeltaTime} minutes after the
last record (\code{"lastRecord"})?}

\item{speciesCol}{character. name of the column specifying species in 
\code{recordTable}}

\item{stationCol}{character. Name of the camera trap station column.
Assuming "Station" if undefined.}

\item{cameraCol}{character. Name of the column specifying cameras in 
\code{recordTable} (optional).}

\item{camerasIndependent}{logical. If \code{TRUE}, species records are
considered to be independent between cameras at a station.}

\item{recordDateTimeCol}{character. Name of the column specifying date and 
time in \code{recordTable}.}

\item{recordDateTimeFormat}{character. Format of column 
\code{recordDateTimeCol} in \code{recordTable}}

\item{removeDuplicateRecords}{logical. If there are several records of the
same species at the same station (also same camera if cameraID is defined)
at exactly the same time, show only one?}

\item{exclude}{character. Vector of species names to be excluded from the
record table}

\item{timeZone}{character. Must be a value returned by
\code{\link[base:timezones]{OlsonNames}}}

\item{writecsv}{logical. Should the record table be saved as a .csv?}

\item{outDir}{character. Directory to save csv to. If NULL and
\code{writecsv = TRUE}, recordTable will be written to \code{inDir}.}

\item{eventSummaryColumn}{character. A column in the record table (e.g. from
a metadata tag) by to summarise non-independent records (those within
\code{minDeltaTime} of a given record) with a user-defined function
(\code{eventSummaryFunction})}

\item{eventSummaryFunction}{character. The function by which to summarise
\code{eventSummaryColumn} of non-independent records, e.g. "sum", "max"
(optional)}

\item{quiet}{logical. If TRUE, suppress printing of progress.}
}
\value{
A data frame containing species records and additional information
about stations, date, time, filtered for temporal independence.
}
\description{
Filter species record table for temporal independence
}
\examples{

if (Sys.which("exiftool") != ""){        # only run example if ExifTool is available

# set directory with camera trap images in station directories
wd_images_ID_species <- system.file("pictures/sample_images_species_dir", 
                                    package = "camtrapR")
                                    
# create record table without temporal filtering
rec_table <- recordTable(inDir                  = wd_images_ID_species,
                        IDfrom                 = "directory",
                        minDeltaTime           = 0,
                        exclude                = "UNID",
                        timeZone               = "Asia/Kuala_Lumpur",
                        removeDuplicateRecords = TRUE
                        )
                        
# filter for 60 minutes temporal independence
rec_table_filt <- filterRecordTable(recordTable         = rec_table, 
                                    minDeltaTime        = 60, 
                                    stationCol          = "Station",
                                    deltaTimeComparedTo = "lastIndependentRecord")
 nrow(rec_table)
 nrow(rec_table_filt)
 
 }
}
\author{
Juergen Niedballa
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/readWildlifeInsights.R
\name{readWildlifeInsights}
\alias{readWildlifeInsights}
\title{Import Wildlife Insights data to camtrapR}
\usage{
readWildlifeInsights(
  directory = NULL,
  zipfile = NULL,
  deployment_file = NULL,
  image_file = NULL
)
}
\arguments{
\item{directory}{character. Path to folder containing CSV files exported from Wildlife Insights.}

\item{zipfile}{character. Path to a ZIP file exported from Wildlife Insights.}

\item{deployment_file}{character. Path to the deployments CSV file.}

\item{image_file}{character. Path to the images CSV file.}
}
\value{
A list containing three elements:
\describe{
  \item{CTtable}{The full camera trap table, based on deployments.csv}
  \item{CTtable_aggregated}{An aggregated version of the camera trap table, with one row per station}
  \item{recordTable}{The record table, based on images.csv with additional columns from deployments.csv}
}
}
\description{
This function imports camera trap data from Wildlife Insights into a format compatible with camtrapR.
It can read data from a directory containing CSV files, from a ZIP file, or from individual CSV files.
}
\examples{
\dontrun{
# Reading from a directory
wi_data <- readWildlifeInsights(directory = "path/to/csv/files")

# Reading from a ZIP file
wi_data <- readWildlifeInsights(zipfile = "path/to/wildlife_insights_export.zip")

# Reading from separate CSV files
wi_data <- readWildlifeInsights(deployment_file = "path/to/deployments.csv",
                                image_file = "path/to/images.csv")
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/plot.R
\name{plot_coef,commOccu-method}
\alias{plot_coef,commOccu-method}
\alias{plot_coef}
\title{Plot effect sizes of covariates in community occupancy model}
\usage{
\S4method{plot_coef}{commOccu}(
  object,
  mcmc.list,
  submodel = "state",
  speciesSubset,
  ordered = TRUE,
  combine = FALSE,
  outdir,
  level = c(outer = 0.95, inner = 0.75),
  colorby = "significance",
  scales = "free_y",
  community_lines = FALSE,
  ...
)
}
\arguments{
\item{object}{\code{commOccu} object}

\item{mcmc.list}{mcmc.list. Output of \code{fit} called on a \code{commOccu} object}

\item{submodel}{character. Submodel to get plots for. Can be \code{"det"} or \code{"state"}}

\item{speciesSubset}{character. Species to include in coefficient plots.}

\item{ordered}{logical. Order species in plot by median effect (TRUE) or by species name (FALSE)}

\item{combine}{logical. Combine multiple plots into one plot (via facets)?}

\item{outdir}{character. Directory to save plots to (optional)}

\item{level}{numeric. Probability mass to include in the uncertainty interval (two values named "outer" and "inner", in that order). Second value (= inner interval) will be plotted thicker.}

\item{colorby}{character. Whether to color estimates by \code{"significance"} (of the effect estimates), or \code{"Bayesian p-value"} (of the species). Currently allows only \code{"significance"}.}

\item{scales}{character. Passed to \code{\link[ggplot2]{facet_grid}}. Can be \code{"free"} to scale x axes of effect estimates independently, or "free_y" to scale all x axes identically.}

\item{community_lines}{logical. Add faint vertical lines to the plot indicating median community effect (solid line) and its confidence interval (first value from \code{level}).}

\item{...}{additional arguments for \code{\link[ggplot2]{ggsave}} - only relevant if \code{outdir} is defined.}
}
\value{
A list of ggplot objects (one list item per covariate).
}
\description{
Plot effect sizes for all species in a community (multi-species) occupancy model. Currently only supports continuous covariates, not categorical covariates.
}
\details{
Users who wish to create their own visualizations can use the data stored in the ggplot output. It is accessed via e.g. \code{output$covariate_name$data}
}

\docType{data}
\name{recordTableSample}
\alias{recordTableSample}
\title{Sample species record table from camera trap images}
\format{A data frame with 39 rows and 11 variables}
\usage{
data(recordTableSample)
}
\description{
Sample species record table from camera trap images generated from the sample images in the package with the function \code{\link{recordTable}} .
}
\details{


The variables are as follows:

\tabular{ll}{
\code{Station} \tab Camera trap station ID \cr
\code{Species} \tab Species ID \cr
\code{DateTimeOriginal} \tab Date and time as extracted from image \cr
\code{Date} \tab record date \cr
\code{Time} \tab record time of day \cr
\code{delta.time.secs} \tab time difference to first species record at a station (seconds) \cr
\code{delta.time.mins} \tab time difference to first species record at a station (minutes) \cr
\code{delta.time.hours} \tab time difference to first species record at a station (hours) \cr
\code{delta.time.days} \tab time difference to first species record at a station (days) \cr
\code{Directory} \tab Image directory \cr
\code{FileName} \tab image filename \cr
}
 
 
}
\keyword{datasets}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/appendSpeciesNames.R
\name{appendSpeciesNames}
\alias{appendSpeciesNames}
\title{Add or remove species names from JPEG image filenames}
\usage{
appendSpeciesNames(
  inDir,
  IDfrom,
  hasCameraFolders,
  metadataSpeciesTag,
  metadataHierarchyDelimitor = "|",
  removeNames = FALSE,
  writecsv = FALSE
)
}
\arguments{
\item{inDir}{character. Directory containing camera trap images sorted into
station subdirectories (e.g. inDir/StationA/)}

\item{IDfrom}{character. Read species ID from image metadata ("metadata") of
from species directory names ("directory")?}

\item{hasCameraFolders}{logical. Do the station subdirectories of
\code{inDir} have camera-subdirectories (e.g. inDir/StationA/CameraA1;
inDir/StationA/CameraA2)?}

\item{metadataSpeciesTag}{character. The species ID tag name in image
metadata (if IDfrom = "metadata").}

\item{metadataHierarchyDelimitor}{character. The character delimiting
hierarchy levels in image metadata tags in field "HierarchicalSubject".
Either "|" or ":".}

\item{removeNames}{logical. remove appended species names?}

\item{writecsv}{logical. write csv table containing old and new file names
into \code{inDir}?}
}
\value{
A \code{data.frame} containing the old and new file names and
directories.
}
\description{
Add or remove species names from JPEG image filenames. It makes it easier to
find images of a species.
}
\details{
Species names can be appended or removed from image filenames. Before
running the function, you may want to run
\code{\link{checkSpeciesIdentification}} to detect possible
misidentifications. As an example, the function would change an image file
name from "StationA__2015-05-41__20-59-59(1).JPG" to
"StationA__2015-05-41__20-59-59(1)__Species Name.JPG". If species names were
appended several times by accident, they can all be removed by running the
function with \code{removeNames = TRUE}
}
\examples{

\dontrun{

# copy sample images to another location (so we don't mess around in the package directory)
wd_images_ID <- system.file("pictures/sample_images_species_dir", package = "camtrapR")
file.copy(from = wd_images_ID, to = getwd(), recursive = TRUE)       
wd_images_ID_copy <- file.path(getwd(), "sample_images_species_dir")

# append species names
SpecNameAppend1 <- appendSpeciesNames(inDir            = wd_images_ID_copy,
                                      IDfrom           = "directory",
                                      hasCameraFolders = FALSE,
                                      removeNames      = FALSE,
                                      writecsv         = FALSE)
  
SpecNameAppend1

# remove species names
SpecNameRemove1 <- appendSpeciesNames(inDir            = wd_images_ID_copy,
                                      IDfrom           = "directory",
                                      hasCameraFolders = FALSE,
                                      removeNames      = TRUE,
                                      writecsv         = FALSE)
  
SpecNameRemove1
}

}
\author{
Juergen Niedballa
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/activityHistogram.R
\name{activityHistogram}
\alias{activityHistogram}
\title{Plot histogram of single-species activity}
\usage{
activityHistogram(
  recordTable,
  species,
  allSpecies = FALSE,
  speciesCol = "Species",
  recordDateTimeCol = "DateTimeOriginal",
  recordDateTimeFormat = "ymd HMS",
  plotR = TRUE,
  writePNG = FALSE,
  plotDirectory,
  createDir = FALSE,
  pngMaxPix = 1000,
  ...
)
}
\arguments{
\item{recordTable}{data.frame. the record table created by
\code{\link{recordTable}}}

\item{species}{Name of the single species for which to create a histogram of
activity}

\item{allSpecies}{logical. Create plots for all species in \code{speciesCol}
of \code{recordTable}? Overrides argument \code{species}}

\item{speciesCol}{character. name of the column specifying species names in
\code{recordTable}}

\item{recordDateTimeCol}{character. name of the column specifying date and
time in \code{recordTable}}

\item{recordDateTimeFormat}{character. format of column
\code{recordDateTimeCol} in \code{recordTable}}

\item{plotR}{logical. Show plots in R graphics device?}

\item{writePNG}{logical. Create pngs of the plots?}

\item{plotDirectory}{character. Directory in which to create png plots if
\code{writePNG = TRUE}}

\item{createDir}{logical. Create \code{plotDirectory}?}

\item{pngMaxPix}{integer. image size of png (pixels along x-axis)}

\item{\dots}{additional arguments to be passed to function
\code{\link[graphics]{hist}}}
}
\value{
It returns invisibly a vector of species record date and time in
\code{POSIXlt} format. If \code{allSpecies == TRUE}, all species' vectors
are returned in an invisible named list.
}
\description{
The function generates a histogram of species diel activity in 1-hour
intervals.
}
\details{
Activity is calculated from the time of day of records. The date is ignored.

\code{recordDateTimeFormat} defaults to the "YYYY-MM-DD HH:MM:SS"
convention, e.g. "2014-09-30 22:59:59". \code{recordDateTimeFormat} can be
interpreted either by base-R via \code{\link[base]{strptime}} or in
\pkg{lubridate} via \code{\link[lubridate]{parse_date_time}} (argument
"orders"). \pkg{lubridate} will be used if there are no "\%" characters in
\code{recordDateTimeFormat}.

For "YYYY-MM-DD HH:MM:SS", \code{recordDateTimeFormat} would be either
"\%Y-\%m-\%d \%H:\%M:\%S" or "ymd HMS". For details on how to specify date
and time formats in R see \code{\link[base]{strptime}} or
\code{\link[lubridate]{parse_date_time}}.
}
\note{
If you have a sufficiently large number of records you may wish to
consider using \code{\link{activityDensity}} instead. Please be aware that
this function (like the other activity... function of this package) use
clock time. If your survey was long enough to see changes in sunrise and
sunset times, this may result in biased representations of species activity.
}
\examples{


# load record table
data(recordTableSample)

# generate activity histogram
species4activity <- "VTA"    # = Viverra tangalunga, Malay Civet

activityHistogram (recordTable = recordTableSample,
                   species     = species4activity,
                   allSpecies = FALSE)

}
\seealso{
\code{\link{activityDensity}}, \code{\link{activityRadial}},
\code{\link{activityOverlap}}
}
\author{
Juergen Niedballa
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/aggregateStations.R
\name{aggregateStations}
\alias{aggregateStations}
\title{Aggregate Camera Trap Table to Station Level}
\usage{
aggregateStations(
  CTtable,
  stationCol,
  cameraCol = NULL,
  setupCol = NULL,
  retrievalCol = NULL,
  dateFormat = NULL,
  quiet = FALSE
)
}
\arguments{
\item{CTtable}{A data frame or `sf` object representing the camera trap table.
Each row typically represents a unique camera deployment.}

\item{stationCol}{A character string specifying the name of the column in \code{CTtable}
that contains the unique station identifiers.}

\item{cameraCol}{A character string specifying the name of the column in \code{CTtable}
that contains the unique camera (deployment) identifiers.}

\item{setupCol}{character. name of the column containing camera setup dates
in \code{CTtable}}

\item{retrievalCol}{character. name of the column containing camera
retrieval dates in \code{CTtable}}

\item{dateFormat}{character. The format of columns \code{setupCol} and
\code{retrievalCol} (and potential problem columns) in \code{CTtable}. Must
be interpretable by either \code{as.Date} or the "orders" argument
\code{\link[lubridate]{parse_date_time}} in \pkg{lubridate}. Can be a date
or a date-time.}

\item{quiet}{logical. If TRUE, print message if no aggregation is needed.}
}
\value{
A data frame with one row per unique station, containing aggregated information.
         Spatial information (for `sf` objects) is preserved.
}
\description{
Aggregates a camera trap table from a station-camera (location-deployment) level to a station (location) level.
This function is useful when modeling or analysis is conducted at the station level,
when multiple cameras were deployed at a single station or a single camera had multiple deployments.
}
\details{
The aggregation logic handles different column types as follows:
\itemize{
  \item \strong{Station ID (specified by `stationCol`):} Remains unique for each row in the output.
  \item \strong{Camera ID (if present):} A character vector (comma-separated)
        of unique camera IDs deployed at that station.
  \item \strong{Setup/Retrieval Dates:} The earliest setup date and the latest retrieval date
        for all deployments within a station will be retained.
  \item \strong{Numeric Columns:} The `mean` of all values for that station will be calculated.
  \item \strong{Logical Columns:} The `mean` (which effectively calculates the proportion of `TRUE`s)
        of all values for that station will be calculated.
  \item \strong{Character Columns (other than Camera ID):} A character vector of unique values
        (comma-separated) for that station will be created.
  \item \strong{Factor Columns:} Similar to character columns, unique levels will be combined.
  \item \strong{Geometry Column (`sf` objects):} Centroid (midpoint) of all unique points at a 
         station is calculated.
}

It is recommended to inspect the aggregated output carefully, especially for columns
with mixed data types or specific aggregation requirements not covered by the defaults.
}
\note{
The function is mainly intended for aggregating covariates to station level. 
         It does currently not handle camera malfunction (via columns `ProblemX_from` / `ProblemX_to`)
         and does not provide proper handling of problem columns.
         Use \code{\link{cameraOperation}} to aggregate camera trap tables to station level for 
         analyses while accounting for camera malfunction / problem periods.
}
